/***************************************************************************
	eXynth GUI library
	(C) 2010 Anti Sullin / Artec Design <anti.sullin@artecdesign.ee>

	Based on xynth library:
    (C) 2003 - 2008 by Alper Akcan <distchx@yahoo.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/


#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H
#include FT_CACHE_H
#include FT_CACHE_CHARMAP_H
#include FT_CACHE_SMALL_BITMAPS_H

#include "exynth.h"
#include "font.h"
#include "image.h"

/* ------------- CACHE SUBSYSTEM ----------------- */

#define FT_CACHE_MAX_FACES	0
#define FT_CACHE_MAX_SIZES	0
#define FT_CACHE_MAX_DATA	(512 * 1024)

// These types are actually pointers:
static FT_Library  library = NULL;
static FTC_Manager ftc_manager = NULL;
static FTC_SBitCache sbitsCache = NULL;

struct fontNameList {
	struct fontNameList *next;
	char name[1]; //allocated with real size
};

// linked list of pointer to font name translation, required by FT.
// this list is only appended, anything is never removed.
struct fontNameList *fontIDs = NULL;

static FT_Error s_font_face_requester(FTC_FaceID face_id,
                         FT_Library lib,
                         FT_Pointer __attribute__((unused)) request_data, FT_Face * aface)
{
	struct fontNameList *fontName = (struct fontNameList *) face_id;

	return FT_New_Face(lib, fontName->name, 0, aface);
}

/* Initialize the font caching subsystem.
 * This creates static FT cache objects that will speed up the font rendering.
 */
int s_font_cache_init (void)
{
	if (!library) {
		if (FT_Init_FreeType(&library)) {
			//debugf(0, "FT_Init_FreeType");
			s_font_cache_uninit();
			return -1;
		}
	}
	if (!ftc_manager) {
		if (FTC_Manager_New(library, FT_CACHE_MAX_FACES, FT_CACHE_MAX_SIZES,
				FT_CACHE_MAX_DATA, s_font_face_requester, NULL, &ftc_manager)) {
			//debugf(0, "Initializing FT cache manager failed!\n");
			s_font_cache_uninit();
			return -2;
		}
	}
	if (!sbitsCache) {
		if (FTC_SBitCache_New(ftc_manager, &sbitsCache)) {
			//debugf(0, "Initializing FT sbit cache failed!\n");
			s_font_cache_uninit();
			return -3;
		}
	}

	return 0;
}

int s_font_cache_uninit (void)
{
	struct fontNameList *p, *n;

	if (ftc_manager) {
		FTC_Manager_Done(ftc_manager);
		ftc_manager = NULL;
		sbitsCache = NULL;
	}
	if (library) {
		FT_Done_FreeType(library);
		library = NULL;
	}
	for (p = fontIDs; p; p = n) {
		n = p->next;
		free(p);
	}
	fontIDs = NULL;

	return 0;
}

struct fontNameList* getFontId(const char *name) {
	struct fontNameList* p;
	for (p = fontIDs; p; p = p->next) {
		if (!strcmp(name, p->name)) return p;
	}

	// add it to list
	p = malloc(sizeof(struct fontNameList) + strlen(name));
	if (!p) return NULL;
	strcpy(p->name, name);
	p->next = fontIDs;
	fontIDs = p;

	return p;
}

/* ------------- FONT SUBSYSTEM ----------------- */

#define DEFAULT_FONT_SIZE 10

struct s_font_ft_s {
	FTC_ImageTypeRec fontType;
	int ascent, descent, width, min_x, max_x;
	int rendered, calculated;
};

int s_font_init (s_font_t **font, const char *name)
{
	FT_Face aface;

	if (!ftc_manager) return -1;

	*font = (s_font_t *) calloc(1, sizeof(s_font_t));
	if (!*font) return -1;
	(*font)->ft = (s_font_ft_t *) calloc(1, sizeof(s_font_ft_t));
	if (!(*font)->ft) {
		s_font_uninit(*font);
		return -2;
	}

	if (name[0] == '/') {
		(*font)->name = (char *) malloc(strlen(name) + 1);
		strcpy((*font)->name, name);
	} else {
		(*font)->name = (char *) malloc(strlen(FONT_PATH) + strlen(name) + 2);
		if (!((*font)->name)) {
			s_font_uninit(*font);
			return -2;
		}
		sprintf((*font)->name, "%s/%s", FONT_PATH, name);
	}

    (*font)->ft->fontType.face_id = getFontId((*font)->name);
    if (!((*font)->ft->fontType.face_id)) {
		s_font_uninit(*font);
		return -3;
    }

    (*font)->ft->fontType.width = 0;
    (*font)->ft->fontType.height = DEFAULT_FONT_SIZE;
    (*font)->ft->fontType.flags = FT_LOAD_DEFAULT;

	// try to load the font face to check its availability
	if (FTC_Manager_LookupFace(ftc_manager, (*font)->ft->fontType.face_id, &aface)) {
		//debugf(0, "Font %s not found", (*font)->name);
		s_font_uninit(*font);
		return -3;
	}

	if (s_image_init(&((*font)->glyph.img))) {
		//debugf(0, "Font image init failed");
		s_font_uninit(*font);
		return -4;
	}

	return 0;
}

int s_font_uninit (s_font_t *font)
{
	if (font == NULL) {
		return 0;
	}
	s_image_uninit(font->glyph.img);
	free(font->str);
	free(font->name);
	free(font->ft);
	free(font);
	return 0;

}

int s_font_set_size (s_font_t *font, int size)
{
	FTC_ScalerRec scaler;
	FT_Size asize;
	FT_Fixed scale;

	if (!ftc_manager) return -1;
	if (font->size == size) return 0;

	font->size = size;
	font->ft->fontType.height = size;
	font->ft->fontType.width = 0;
	font->ft->rendered = 0;
	font->ft->calculated = 0;

	scaler.face_id = font->ft->fontType.face_id;
	scaler.width = 0;
	scaler.height = size;
	scaler.pixel = 1;

	if (FTC_Manager_LookupSize( ftc_manager, &scaler, &asize)) {
		return -1;
	}

#define FT_FLOOR(X) ((X & -64) / 64)
#define FT_CEIL(X)  (((X + 63) & -64) / 64)

	scale = asize->metrics.y_scale;
	font->ascent  = FT_CEIL(asize->metrics.ascender);
	font->descent = FT_CEIL(asize->metrics.descender);
	font->height  = font->ascent - font->descent + /* baseline */ 1;
	font->lineskip = FT_CEIL(asize->metrics.height);
	font->underline_offset = FT_FLOOR(FT_MulFix(asize->face->underline_position, scale));
	font->underline_height = FT_FLOOR(FT_MulFix(asize->face->underline_thickness, scale));

	return 0;
}

int s_font_set_str (s_font_t *font, const char *str)
{
	if (font->str) {
		if (str && !strcmp(font->str, str)) return 0;
		free(font->str);
		font->str = NULL;
	}
	if (str) {
		font->str = strdup(str);
	}
	font->ft->rendered = 0;
	font->ft->calculated = 0;

	return 0;
}

int s_font_set_rgb (s_font_t *font, int r, int g, int b)
{
	font->rgb = ((r & 0xFF) << 0x10) | ((g & 0xFF) << 0x08) | (b & 0xFF);
	font->ft->rendered = 0;
	font->ft->calculated = 0;
	return 0;
}

int s_font_utf8_to_unicode (unsigned short *unicode, const char *utf8, int l)
{
	unsigned short ch;
	int ret = 1;

	if (!l) return 0;
	ch = ((const unsigned char *) utf8)[0];
	if (ch >= 0xF5) {
		return 0;
	} else if (ch >= 0xF0) {
		if (l<4) return 0;
		if ((utf8[1] & 0xC0) != 0x80) return 0;
		if ((utf8[2] & 0xC0) != 0x80) return 0;
		if ((utf8[3] & 0xC0) != 0x80) return 0;
		ch = (unsigned short) (utf8[0] & 0x07) << 18;
		ch |= (unsigned short) (utf8[1] & 0x3F) << 12;
		ch |= (unsigned short) (utf8[2] & 0x3F) << 6;
		ch |= (unsigned short) (utf8[3] & 0x3F);
		ret = 4;
	} else if (ch >= 0xE0) {
		if (l<3) return 0;
		if ((utf8[1] & 0xC0) != 0x80) return 0;
		if ((utf8[2] & 0xC0) != 0x80) return 0;
		ch = (unsigned short) (utf8[0] & 0x0F) << 12;
		ch |= (unsigned short) (utf8[1] & 0x3F) << 6;
		ch |= (unsigned short) (utf8[2] & 0x3F);
		ret = 3;
	} else if (ch >= 0xC2) {
		if (l<2) return 0;
		if ((utf8[1] & 0xC0) != 0x80) return 0;
		ch = (unsigned short) (utf8[0] & 0x1F) << 6;
		ch |= (unsigned short) (utf8[1] & 0x3F);
		ret = 2;
	} else if (ch >= 0x80) {
		return 0;
	}
	if (unicode) *unicode = ch;

	return ret;
}

int s_font_update_measurements (s_font_t *font)
{
	FT_Face aface;

	int len = strlen(font->str);
	char *p = font->str;

	int ascent = 0, descent = 0, pen_x = 0, min_x = INT_MAX, max_x = INT_MIN;

	if (!ftc_manager) return -1;
	if (font->ft->calculated) return 0;

	if (FTC_Manager_LookupFace(ftc_manager, font->ft->fontType.face_id, &aface)) {
		return -3;
	}

	while (len > 0) {
		FTC_SBit	sbit;
		FTC_Node	anode;
		unsigned short c = 0;
		int 		cnt;

		// get next char unicode number
		cnt = s_font_utf8_to_unicode(&c, p, len);
		if (cnt <= 0) {
			len -= 1;
			p += 1;
			continue;
		}
		len -= cnt;
		p += cnt;

		// lookup character face ID from unicode (default) charmap
		FT_UInt faceId = FT_Get_Char_Index(aface, c);

		// read char from freetype
		if (FTC_SBitCache_Lookup(sbitsCache, &font->ft->fontType, faceId, &sbit, &anode))
			continue;

		// read measurements
		if (max_x < pen_x + sbit->left + sbit->width) max_x = pen_x + sbit->left + sbit->width;
		if (min_x > pen_x + sbit->left) min_x = pen_x + sbit->left;
		pen_x += sbit->xadvance;
		if (ascent < sbit->top) ascent = sbit->top;
		if (descent < sbit->height - sbit->top)
                	descent = sbit->height - sbit->top;

		// release glyph
		FTC_Node_Unref(anode, ftc_manager);
	}

	font->ft->ascent = ascent;
	font->ft->descent = descent;
	font->ft->width = max_x - min_x +1;
	font->ft->min_x = min_x;
	font->ft->max_x = max_x;
	font->ft->calculated = 1;

	return 0;
}

int s_font_get_glyph_pos (s_font_t *font, int *pos)
{
	FT_Face aface;
	int len = strlen(font->str);
	char *p = font->str;
	int pen_x = 0, n;

	if (!ftc_manager) return -1;
	if (font->ft->rendered) return 0;

	// update bounding box size
	if (s_font_update_measurements(font)) return -1;
	font->glyph.img->w = font->ft->width;
	font->glyph.img->h = font->ft->ascent + font->ft->descent;

	// realloc bitmap
	free(font->glyph.img->rgba);
	font->glyph.img->rgba = NULL;
	if (!font->glyph.img->w || !font->glyph.img->h) return -2;
	font->glyph.img->rgba  = (unsigned int *) malloc(font->glyph.img->w * font->glyph.img->h * sizeof(unsigned int));
	if (!font->glyph.img->rgba) return -3;
	n = font->glyph.img->w * font->glyph.img->h;
	while (n--) font->glyph.img->rgba[n] = 255; // all opaque

	// font info for xynth
	font->glyph.xMax = font->ft->max_x;
	font->glyph.xMin = font->ft->min_x;
	font->glyph.yMax = font->ft->ascent;
	font->glyph.yMin = -font->ft->descent;
	font->glyph.height = font->height;
	font->glyph.lineskip = font->lineskip;
	font->glyph.size = font->size;

	if (FTC_Manager_LookupFace(ftc_manager, font->ft->fontType.face_id, &aface)) {
		return -4;
	}

	// draw the image!
	while (len > 0) {
		FTC_SBit	sbit;
		FTC_Node	anode;
		unsigned short c = 0;
		int 		cnt, width, xpos, ypos, height, xpos2, ypos2;
		int 		x,y;
		unsigned int rgb;

		// get next char unicode number
		cnt = s_font_utf8_to_unicode(&c, p, len);
		if (cnt <= 0) {
			len -= 1;
			p += 1;
			continue;
		}
		len -= cnt;
		p += cnt;

		// store the character x position if required
		if (pos) *pos++ = pen_x;

		// lookup character face ID from unicode (default) charmap
		FT_UInt faceId = FT_Get_Char_Index(aface, c);

		// read char from freetype
		if (FTC_SBitCache_Lookup(sbitsCache, &font->ft->fontType, faceId, &sbit, &anode))
			continue;

		// set up draw position
		xpos = pen_x + sbit->left - font->ft->min_x;
        width = sbit->width;
        if (xpos >= font->glyph.img->w) goto next;
        if (xpos + width > font->glyph.img->w) width = font->glyph.img->w - xpos;

        ypos = font->ft->ascent - sbit->top;
        height = sbit->height;
        if (ypos >= font->glyph.img->h) goto next;
        if (ypos + height > font->glyph.img->h) height = font->glyph.img->h - ypos;

        ypos2 = ypos;
        if (ypos2 < 0) ypos2 = 0;
        xpos2 = xpos;
        if (xpos2 < 0) xpos2 = 0;

        unsigned char *src_r = sbit->buffer + (ypos2 - ypos) * sbit->pitch;
        unsigned int *dst_r = font->glyph.img->rgba + ypos * font->glyph.img->w + xpos;

        // draw glyph
        rgb = (font->rgb << 8) & 0xFFFFFF00;
		for (y = ypos2; y < ypos + height; y++) {
			unsigned char *src = src_r;
			unsigned int *dst = dst_r;
			for (x = xpos2; x < xpos + width; x++) {
				if (*src) {
					*dst = rgb | (~*src & 0xff);
				}
				src++;
				dst++;
			}

			src_r += sbit->pitch;
			dst_r += font->glyph.img->w;
		}

		// release glyph
next:
		pen_x += sbit->xadvance;
		FTC_Node_Unref(anode, ftc_manager);

		// is there a point to continue?
		if (pos == NULL && pen_x > font->glyph.img->w) break;
	}

	// store the post-last-character x position if required
	if (pos) *pos = pen_x;

	return 0;
}



int s_font_get_width_pos (s_font_t *font, const char *str, int *pos)
{
	FT_Face aface;
	int len = strlen(str);
	const char *p = str;

	if (!ftc_manager) return -1;
	int pen_x = 0, min_x = INT_MAX, max_x = INT_MIN;

	if (FTC_Manager_LookupFace(ftc_manager, font->ft->fontType.face_id, &aface)) {
		return -3;
	}

	while (len > 0) {
		FTC_SBit	sbit;
		FTC_Node	anode;
		unsigned short c = 0;
		int 		cnt;

		// get next char unicode number
		cnt = s_font_utf8_to_unicode(&c, p, len);
		if (cnt <= 0) {
			len -= 1;
			p += 1;
			continue;
		}
		len -= cnt;
		p += cnt;

		// store the character x position if required
		if (pos) *pos++ = pen_x;

		// lookup character face ID from unicode (default) charmap
		FT_UInt faceId = FT_Get_Char_Index(aface, c);

		// read char from freetype
		if (FTC_SBitCache_Lookup(sbitsCache, &font->ft->fontType, faceId, &sbit, &anode))
			continue;

		// read measurements
		if (max_x < pen_x + sbit->left + sbit->width) max_x = pen_x + sbit->left + sbit->width;
		if (min_x > pen_x + sbit->left) min_x = pen_x + sbit->left;
		pen_x += sbit->xadvance;

		// release glyph
		FTC_Node_Unref(anode, ftc_manager);
	}

	// store the post-last-character x position if required
	if (pos) *pos = pen_x;

	return max_x - min_x +1;;
}
