/***************************************************************************
	eXynth GUI library
	(C) 2010 Anti Sullin / Artec Design <anti.sullin@artecdesign.ee>

	Based on xynth library:
    (C) 2003 - 2008 by Alper Akcan <distchx@yahoo.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/


#include <unistd.h>
#include <poll.h>
#include <pthread.h>
#include <errno.h>
#include <fcntl.h>

#include "../exynth.h"
#include "input.h"

//#define DEBUG
#include "../debug.h"

#define HZ	100

/* Initialize input thread */
int s_input_init()
{
	s_exynth_t *e = get_exynth();
	debug("s_input_init()\n");
	if (pipe(e->selfpipe)!= 0) return -1;
	fcntl (e->selfpipe[0],F_SETFL,fcntl(e->selfpipe[0],F_GETFL)|O_NONBLOCK);
	fcntl (e->selfpipe[1],F_SETFL,fcntl(e->selfpipe[1],F_GETFL)|O_NONBLOCK);

	if (pthread_create(&(e->thread), NULL, s_input_loop, NULL) != 0)
		return -2;
	return 0;
}

/* Shut down input thread */
int s_input_uninit()
{
	int ret;
	s_exynth_t *e = get_exynth();
	debug("s_input_uninit()\n");
	pthread_cancel(e->thread);
	ret = pthread_join(e->thread, NULL);
	input_touch_uninit();
	input_keypad_uninit();
	return 0;
}

/* Wake up the input thread */
void s_input_wakeup()
{
	s_exynth_t *e = get_exynth();
	// good old self-pipe hack
	debug("s_input_wakeup()\n");
	write(e->selfpipe[1], "",1);
}

/* Clean up self-pipe */
void s_input_selfpipe_flush()
{
	s_exynth_t *e = get_exynth();
	char dummy;
	while (read(e->selfpipe[0],&dummy,sizeof(dummy)) > 0);
}

/* Input loop */
typedef void(*handler_t)(void);
void * s_input_loop (void *arg)
{
	int fdcount = 0, i;
	struct pollfd fds[3];
	handler_t handler[3];
	long long start, stop;
	s_exynth_t *e = get_exynth();

	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

	/* Initialize self pipe*/
	fds[fdcount].fd = e->selfpipe[0];
	fds[fdcount].events = POLLIN;
	handler[fdcount] = s_input_selfpipe_flush;
	fdcount++;

	/* Initialize keypad */
	fds[fdcount].fd = input_keypad_init();
	fds[fdcount].events = POLLIN;
	handler[fdcount] = input_keypad_update;
	if (fds[fdcount].fd >= 0) {
		debug("s_input_loop: initialized keypad\n");
		fdcount++;
	}

	/* Initialize touchscreen */
	fds[fdcount].fd = input_touch_init();
	fds[fdcount].events = POLLIN;
	handler[fdcount] = input_touch_update;
	if (fds[fdcount].fd >= 0) {
		debug("s_input_loop: initialized touchscreen\n");
		fdcount++;
	}

	if (fdcount == 0) return NULL;

	start = s_gettimeofday();
	while (!e->exit) {
		int ret;
		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
		pthread_testcancel();

		// poll until wakeup or timeout
		for (i=0; i< fdcount; i++)
			fds[i].revents = 0;
		s_input_selfpipe_flush();
		int time = s_timer_next_wakeup();

		// compensate previous timer handler run-time
		stop = s_gettimeofday();
		if (time > 0){
			if (time > (stop-start))
				// nb! stop-start may be negative!
				time -= (stop-start);
			else
				time = 0;

			// do not wait less than scheduler granularity
			if (time < (1000/HZ)) time = 0;
		}

		// do not schedule a poll if time too short
		if (time != 0) {
			debug("s_input_loop: polling input\n");
			ret = poll(fds, fdcount, time);
			debug("s_input_loop: poll returned\n");

			if (ret == -1 && errno == EINTR) continue;
			if (ret == -1) break;
		}
		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

		// handle input (only revents if polled, otherwise all)
		pthread_mutex_lock(&e->mutex);
		for (i=0; i < fdcount; i++) {
			debug("poll %d, revents %x\n", i, fds[i].revents);
			if (time && fds[i].revents == 0) continue;
			(handler[i])();
		}

		// handle timer
		stop = s_gettimeofday();
		// always at least 1 unit to compensate inaccuracy
		if (stop <= start) stop = start + 1;

		s_timer_handle(start, stop);
		start = stop;

		stop = s_gettimeofday();

		// handle delayed window destruction
		s_window_perform_quit();

		pthread_mutex_unlock(&e->mutex);
	}
	return NULL;
}
