/***************************************************************************
	eXynth GUI library
	(C) 2010 Anti Sullin / Artec Design <anti.sullin@artecdesign.ee>

	Based on xynth library:
    (C) 2003 - 2008 by Alper Akcan <distchx@yahoo.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include "exynth.h"
#include "surface.h"
#include "driver.h"
#include "grlib.h"

//#define DEBUG
#include "debug.h"

// initialize client window surface
int s_surface_init (s_window_t *window)
{
	s_surface_t *s;
	debug("s_surface_init(win=%p)\n", window);
	s = (s_surface_t *) calloc(1, sizeof(s_surface_t));
	if (s == NULL) goto err0;

	window->surface = s;
	s_exynth_t *e = get_exynth();
	s_surface_defparams(window->surface, e->root_surface);
	return 0;
err0:
	return -1;
}

// initialize server surface, call after initializing fb
int s_server_surface_init (void)
{
	debug("s_server_surface_init\n");
	s_exynth_t *e = get_exynth();
	e->root_surface = (s_surface_t *) calloc(1, sizeof(s_surface_t));
	if (e->root_surface == NULL) return -1;
	e->fb_surface = (s_surface_t *) calloc(1, sizeof(s_surface_t));
	if (e->fb_surface == NULL) {
		free(e->root_surface);
		e->root_surface = NULL;
		return -2;
	}
	return 0;
}

// set up a window surface
int s_surface_create (s_window_t *window)
{
	debug("s_surface_create(win=%p)\n", window);
	window->surface->vbuf = (unsigned char *) calloc(sizeof(char), window->surface->width *
	                                                                 window->surface->height *
	                                                                 window->surface->bytesperpixel);
	if (window->surface->vbuf == NULL) return -1;
	window->surface->mode = SURFACE_VIRTUAL;
	return 0;
}

// set up server surface
int s_server_surface_create (void)
{
	debug("s_server_surface_create\n");
	s_exynth_t *e = get_exynth();
	e->root_surface->matrix = (unsigned char *) calloc(sizeof(char),
			e->root_surface->width * e->root_surface->height);
	if (e->root_surface->matrix == NULL) return -1;

	e->root_surface->linear_buf_width = e->root_surface->width;
	e->root_surface->linear_buf_pitch = e->root_surface->width;
	e->root_surface->linear_buf_height = e->root_surface->height;

	e->fb_surface->width = e->root_surface->width;
	e->fb_surface->height = e->root_surface->height;
	e->fb_surface->bitsperpixel = e->root_surface->bitsperpixel;

	// If we need to rotate, prepare an extra virtual buffer for non-rotated
	// composite image; we'll rotate and copy it to fb when needed.
#ifdef FB_ROTATE
	e->root_surface->vbuf = (unsigned char *) calloc(sizeof(char),
			e->root_surface->width * e->root_surface->height * e->root_surface->bytesperpixel);
	if (e->root_surface->vbuf == NULL) {
		free(e->root_surface->matrix);
		return -2;
	}
	s_surface_defparams(e->fb_surface, e->root_surface);
	e->fb_surface->linear_buf = e->root_surface->linear_buf;
	e->root_surface->linear_buf = e->root_surface->vbuf;
	e->root_surface->mode = SURFACE_VIRTUAL | SURFACE_REAL;
	e->fb_surface->mode = SURFACE_REAL;

	if (FB_ROTATE == 90 || FB_ROTATE == 270) {
		e->root_surface->width = e->fb_surface->height;
		e->root_surface->height = e->fb_surface->width;
		e->root_surface->linear_buf_width = e->root_surface->width;
		e->root_surface->linear_buf_height = e->root_surface->height;
		e->root_surface->linear_buf_pitch = e->root_surface->width;
	}
#else
	e->root_surface->mode = SURFACE_REAL;
#endif

	debug("root surface: %d x %d\n", e->root_surface->width,
			e->root_surface->height);
	return 0;
}

// un-initialize a window surface
void s_surface_uninit (s_window_t *window)
{
	debug("s_surface_uninit(win=%p)\n", window);
	if (window->surface == NULL) {
		return;
	}

	if (window->surface->vbuf) free(window->surface->vbuf);
	free(window->surface);
	window->surface = NULL;
}

// un-initialize server surface
int s_server_surface_uninit (void)
{
	debug("s_server_surface_uninit\n");
	s_exynth_t *e = get_exynth();
	if (e->root_surface) {
		if (e->root_surface->matrix) free(e->root_surface->matrix);
		if (e->root_surface->vbuf) free(e->root_surface->vbuf);
		free(e->root_surface);
		e->root_surface = NULL;
	}
	if (e->fb_surface) free(e->fb_surface);
	e->fb_surface = NULL;
	return 0;
}

// copy parameters from one to another surface
void s_surface_defparams (s_surface_t *dst, s_surface_t *src)
{
	dst->bytesperpixel = src->bytesperpixel;
	dst->bitsperpixel = src->bitsperpixel;
	dst->colors = src->colors;
	dst->blueoffset = src->blueoffset;
	dst->greenoffset = src->greenoffset;
	dst->redoffset = src->redoffset;
	dst->bluelength = src->bluelength;
	dst->greenlength = src->greenlength;
	dst->redlength = src->redlength;
	dst->bluemask = src->bluemask;
	dst->greenmask = src->greenmask;
	dst->redmask = src->redmask;

	dst->linear_buf = src->linear_buf;
	dst->matrix = src->matrix;

	dst->linear_mem_size = src->linear_mem_size;
	dst->linear_buf_width = src->linear_buf_width;
	dst->linear_buf_pitch = src->linear_buf_pitch;
	dst->linear_buf_height = src->linear_buf_height;
}

// expose the changes in window buffer to fb
// if all=1, do not check matrix
void s_surface_changed (s_window_t *window, s_rect_t *changed, int all)
{
	int x;
	int y;
	s_rect_t coor;

	debug("s_surface_changed(win=%p, x=%d, y=%d, w=%d, h=%d, all=%d)\n", window,
			changed->x, changed->y, changed->w, changed->h, all);

	// window real surface not set up?
	if (!(window->surface->mode & SURFACE_REAL)) return;
	// window not registered in priority map?
	if (window->visible == 0) return;

	// translate to window coordinate system
	x = changed->x - window->surface->buf.x;
	y = changed->y - window->surface->buf.y;
	if (s_rect_clip_real(window->surface, x, y, changed->w, changed->h, &coor)) {
		return;
	}

	// copy rect to linear buffer (coordinates in window coordinate system)
	if (all)
		bpp_putbox_od(window->surface, coor.x, coor.y, coor.w, coor.h,
	             window->surface->vbuf + ((coor.y * window->surface->width) + coor.x) * window->surface->bytesperpixel,
	             window->surface->width);
	else
		bpp_putbox_o(window->surface, window->id, coor.x, coor.y, coor.w, coor.h,
		             window->surface->vbuf + ((coor.y * window->surface->width) + coor.x) * window->surface->bytesperpixel,
		             window->surface->width);

#ifdef FB_ROTATE
	// copy rotated region from fake linear buffer to real framebuffer
	{
		char *src;
		char *dst;
		s_rect_t rotate;
		s_surface_t surface;
		s_exynth_t *e = get_exynth();

		src = malloc(coor.w * coor.h * e->root_surface->bytesperpixel);
		if (src == NULL) return;
		dst = malloc(coor.w * coor.h * e->root_surface->bytesperpixel);
		if (dst == NULL) {
			free(src);
			return;
		}

		// translate back to absolute coordinate system
		coor.x += window->surface->buf.x;
		coor.y += window->surface->buf.y;

		bpp_getbox(e->root_surface, coor.x, coor.y, coor.w, coor.h, src);
		s_getsurfacevirtual(&surface, e->fb_surface->width, e->fb_surface->height,
				e->fb_surface->bitsperpixel, NULL);
		s_rotatebox(&surface, &coor, src, &rotate, dst, FB_ROTATE);
		bpp_putbox_od(e->fb_surface, rotate.x, rotate.y, rotate.w, rotate.h, dst, rotate.w);

		free(src);
		free(dst);
	}
#endif
}

// expose black background to fb
void s_server_surface_changed (s_rect_t *changed, int all)
{
	s_exynth_t *e = get_exynth();
	debug("s_server_surface_changed(x=%d, y=%d, w=%d, h=%d, all=%d)\n",
			changed->x, changed->y, changed->w, changed->h, all);

	bpp_fillbox_o(e->root_surface, 0, changed->x, changed->y, changed->w, changed->h, 0);

#ifdef FB_ROTATE
	{
		char *src;
		char *dst;
		s_rect_t rotate;
		s_surface_t surface;

		src = malloc(changed->w * changed->h * e->root_surface->bytesperpixel);
		if (src == NULL) return;
		dst = malloc(changed->w * changed->h * e->root_surface->bytesperpixel);
		if (dst == NULL) {
			free(src);
			return;
		}

		bpp_getbox(e->root_surface, changed->x, changed->y, changed->w, changed->h, src);
		s_getsurfacevirtual(&surface, e->fb_surface->width, e->fb_surface->height,
				e->fb_surface->bitsperpixel, NULL);
		s_rotatebox(&surface, changed, src, &rotate, dst, FB_ROTATE);
		bpp_putbox_od(e->fb_surface, rotate.x, rotate.y, rotate.w, rotate.h, dst, rotate.w);

		free(src);
		free(dst);
	}
#endif
}
