/***************************************************************************
	eXynth GUI library
	(C) 2010 Anti Sullin / Artec Design <anti.sullin@artecdesign.ee>

	Based on xynth library:
    (C) 2003 - 2008 by Alper Akcan <distchx@yahoo.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <sys/time.h>

#include "exynth.h"
#include "list.h"
#include "input/input.h"

//#define DEBUG
#include "debug.h"

// allocate a new timer, initialize it
int s_timer_init (s_timer_t **timer)
{
	debug("s_timer_init()\n");
	*timer = (s_timer_t *) malloc(sizeof(s_timer_t));
	if (!*timer) return -1;
	(*timer)->timeval = -1;
	(*timer)->interval = -1;
	(*timer)->cb = NULL;
	(*timer)->data = NULL;
	(*timer)->window = NULL;
	return 0;
}

// de-allocate a timer
int s_timer_uninit (s_timer_t *timer)
{
	debug("s_timer_uninit(%p)\n", timer);
	free(timer);
	return 0;
}

// initialize timer subsystem
int s_timers_init (void)
{
	debug("s_timers_init()\n");
	s_exynth_t *e = get_exynth();
	s_list_sinit(&e->timer_list);
	return 0;
}

// de-initialize timer subsystem
int s_timers_uninit (void)
{
	debug("s_timers_uninit()\n");
	return 0;
}

// remove a timer from active list
int s_timer_del (s_timer_t *timer)
{
	int r;
	s_exynth_t *e = get_exynth();
	debug("s_timer_del(%p)\n", timer);

	if (timer == NULL) return -1;

	r = s_list_remove(&e->timer_list, s_list_get_pos(&e->timer_list, timer));

	return r;
}

// remove all timers belonging to a window from list
int s_timer_del_wnd (void *window)
{
	int r = -1;
	s_list_node_t *node, *pnode;
	s_exynth_t *e = get_exynth();
	debug("s_timer_del_wnd(%p)\n", window);

	pnode = NULL;
	node = s_list_first(&e->timer_list);

	while (node) {
		s_timer_t *timer = (s_timer_t *) s_list_element(node);
		if (timer && timer->window == window) {
			s_list_del_next(&e->timer_list, pnode);
			r = 0;
			if (pnode)
				node = s_list_next(pnode);
			else
				node = s_list_first(&e->timer_list);
			continue;
		}
		pnode = node;
		node = s_list_next(node);
	}

	return r;
}

// add a timer to list
int s_timer_add (s_timer_t *timer)
{
	s_exynth_t *e = get_exynth();
	debug("s_timer_add(%p)\n", timer);

	if (timer->timeval < 0) return -1;
	if (timer->cb == NULL && timer->window == NULL)
		return -2;

	timer->interval = timer->timeval;
	timer->startval = s_gettimeofday();

	if (s_list_get_pos(&e->timer_list, timer) < 0) {
		if (s_list_add(&e->timer_list, timer, -1) <0) {
			return -3;
		}
	}
	s_input_wakeup();

	return 0;
}

// --------------------------------------------------------------------
// Timer handler, running in server thread

// Next timer timeout in ms, -1 if none scheduled
int s_timer_next_wakeup ()
{
	int timeout = -1;
	s_list_node_t *node, *pnode;
	s_timer_t *tmr;
	s_exynth_t *e = get_exynth();

	debug("s_timer_next_wakeup()\n");

	pnode = NULL;
	node = s_list_first(&e->timer_list);
	while (node) {
		tmr = (s_timer_t *) s_list_element(node);
		if (tmr->interval <= 0) {
			if (tmr->timeval > 0) {
				tmr->interval = tmr->timeval;
			} else {
				s_list_del_next(&e->timer_list, pnode);
				if (pnode)
					node = s_list_next(pnode);
				else
					node = s_list_first(&e->timer_list);
				continue;
			}
		}

		if (timeout <= 0 || tmr->timeval < timeout)
			timeout = tmr->interval;

		pnode = node;
		node = s_list_next(node);
	}
	debug("s_timer_next_wakeup(): %d ms\n", timeout);
	return timeout;
}

// Handle a timer
void s_timer_handle(long long start, long long stop)
{
	s_event_t ev;
	s_list_node_t *node, *pnode;
	s_timer_t *tmr;
	int changed;
	s_exynth_t *e = get_exynth();
	debug("s_timer_handle(time %d)\n", time);

	// pass 1: update interval
	node = s_list_first(&e->timer_list);
	while (node) {
		long long time;
		tmr = (s_timer_t *) s_list_element(node);
		if (start < tmr->startval)
			time = stop - tmr->startval;
		else
			time = stop - start;

		tmr->interval -= time;
		node = s_list_next(node);
	}
	// pass 2: handle timers where interval is negative
	do {
		pnode = NULL;
		node = s_list_first(&e->timer_list);
		changed = 0;
		while (node) {
			tmr = (s_timer_t *) s_list_element(node);
			if (tmr->interval <= 0) {
				changed = 1;

				if (tmr->timeval > 0) {
					// if timeval > 0, restart
					tmr->interval = tmr->timeval;
				} else {
					// otherwise, delete from list
					s_list_del_next(&e->timer_list, pnode);
				}

				// callback; after callback the pointer may
				// have been free'd or the list may have changed!
				if (tmr->cb) {
					debug("s_timer_handle: callback for timer %p\n", tmr);
					tmr->cb(tmr->window, tmr);
				} else {
					ev.type = TIMER_EVENT;
					ev.data.timer = tmr;
					s_window_event(tmr->window, &ev);
				}
				// loop again from beginning as the list may have changed
				break;
			}
			pnode = node;
			node = s_list_next(node);
		}
	} while (changed);
}

// get current timestamp in ms
long long s_gettimeofday (void)
{
	long long tsec;
	long long tusec;

	struct timespec now_ts;
	if (clock_gettime( CLOCK_MONOTONIC, &now_ts ) < 0) {
		return -1;
	}

	tsec = ((long long) now_ts.tv_sec) * 1000;
	tusec = ((long long) now_ts.tv_nsec) / 1000 / 1000;
	return tsec + tusec;
}
