/***************************************************************************
	eXynth GUI library
	(C) 2010 Anti Sullin / Artec Design <anti.sullin@artecdesign.ee>

	Based on xynth library:
    (C) 2003 - 2008 by Alper Akcan <distchx@yahoo.com>
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>

#include "exynth.h"
#include "window.h"
#include "surface.h"
#include "server/server.h"
#include "input/input.h"
#include "timer.h"

//#define DEBUG
#include "debug.h"

// hide a window, remove from window stack
void s_window_hide (s_window_t *window)
{
	int id;
	debug("s_window_hide(%p)\n", window);
	if (window == NULL) return;
	id = window->id;
	if (id == 0) return;
	window->visible = 0;
	s_server_pri_del(id);
}

// prepare to show a window, add it to window stack
// with visible=false. Required to maintain the
// correct order of windows if parent creates a
// child early
void s_window_show_prepare(s_window_t *window)
{
	int id;
	debug("s_window_show_prepare(%p)\n", window);
	if (window == NULL) return;
	id = window->id;
	if (id == 0) return;

	s_server_pri_focus_prepare(id);
}

// commit the show (change stack, matrix and
// possibly re-draw part of fb)
void s_window_show (s_window_t *window)
{
	int id;
	debug("s_window_show(%p)\n", window);
	if (window == NULL) return;
	id = window->id;
	if (id == 0) return;
	window->visible = 1;
	s_server_pri_focus(id);
}

// move window
void s_window_set_coor (s_window_t *window, int form, int x, int y, int w, int h)
{
	s_rect_t rect = {x, y, w, h};
	s_rect_t old;
	int id;
	debug("s_window_set_coor(%p, x=%d, y=%d, w=%d, h=%d)\n", window, x, y, w, h);
	if (window == NULL) return;
	id = window->id;
	if (id == 0) return;
	old = window->surface->buf;
	window->surface->buf = rect;

	s_server_pri_move(id, &old);
}

// set alwaysontop flag. effective after next s_window_show
void s_window_set_alwaysontop (s_window_t *window, int alwaysontop)
{
	debug("s_window_setalwaysontop(%p, %d)\n", window, alwaysontop);
	window->alwaysontop = alwaysontop;
}

// register a new client window
int s_window_new (s_window_t *window, S_WINDOW type, s_window_t *parent)
{
	int id;
	s_event_t ev;
	s_exynth_t *e = get_exynth();

	debug("s_window_new(%p, type=%d, parent=%p)\n", window, type, parent);
	if (!(type & (WINDOW_MAIN | WINDOW_CHILD | WINDOW_MODAL))) {
		debug("s_window_new: invalid window type\n");
		goto err;
	}
	window->type = type;
	if (window->surface->width <= 0) {
		window->surface->width = window->surface->linear_buf_width;
	}
	if (window->surface->height <= 0) {
		window->surface->height = window->surface->linear_buf_height;
	}
	if (window->type & (WINDOW_CHILD | WINDOW_MODAL)) {
		window->parent = parent;
	}

	// already inited
	id = s_server_id_find(window);
	if (id >= 0) {
		debug("s_window_new: window (%p) already in server id list\n", window);
		goto err;
	}

	// allocate id
	id = s_server_id_alloc();
	if (id < 0) {
		debug("s_window_new: id allocation failed\n");
		goto err;
	}
	window->id = id;
	window->pri = -1;

	// allocate surface
	if (s_surface_create(window) != 0) {
		debug("s_window_new: creating surface failed\n");
		goto err1;
	}
	window->surface->mode |= SURFACE_REAL;

	// set up id info
	e->clients[id] = window;

	if (window->type & WINDOW_MODAL) {
		window->parent->num_modal++;
	}

	// notify
	ev.type = CREATED_EVENT;
	ev.data.idwnd = id;
	s_window_event(window, &ev);

	return 0;
err1:
	s_server_id_free(id);
err:
	return 1;
}

// handle a window event
int s_window_event(s_window_t *window, s_event_t *event)
{
	debug("s_window_event(%p, %p (type %08x)\n", window, event, event->type);
	if (window == NULL || event == NULL) return -1;
	if (window->atevent == NULL) return -2;

	// block the modal parent if needed
	if (window->num_modal != 0 && !(window->type & WINDOW_INPUT)) {
		if ((event->type & EV_CODE) == KEYBD_EVENT) return 0;
		if ((event->type & EV_CODE) == MOUSE_EVENT) return 0;
	}

	window->atevent(window, event);
	return 0;
}

// allocate and initialize a window
int s_window_init (s_window_t **window)
{
	s_window_t *w;

	debug("s_window_init(%p)\n", window);

	w = (s_window_t *) calloc(1, sizeof(s_window_t));
	if (s_surface_init(w))       { goto err1; }

	w->alwaysontop = 0;
	w->mouse_entered = 0;
	*window = w;

	return 0;

err1:
	free(w);
	return -1;
}

// free a window. called automatically after s_window_quit.
// call when s_window_add fails.
void s_window_uninit (s_window_t *window)
{
	debug("s_window_uninit(%p)\n", window);
	if (window == NULL) return;

	s_surface_uninit(window);

	free(window);
}

// user requests close - enqueue it.
// We can't destroy the window right away, there may be multiple references
// to that window in cpu stack. enqueue it and delete the window when it's safe
void s_window_quit (s_window_t *window)
{
	s_exynth_t *e = get_exynth();
	debug("s_window_quit(%p)\n", window);

	// close all children first
	if (window->num_modal) {
		int i;
		for (i = 0; i < S_CLIENTS_MAX; i++) {
			if (e->clients[i] != NULL && e->clients[i]->parent == window) {
				s_window_quit(e->clients[i]);
			}
		}
	}

	// and then, the window itself
	if (s_list_get_pos(&(e->exitlist), window) >= 0) return;
	s_list_add(&(e->exitlist), window, -1);
	s_input_wakeup();
}

// go through the list and perform the quit operations (called by event loop)
void s_window_perform_quit(void)
{
	s_exynth_t *e = get_exynth();
	s_list_node_t *n = s_list_first(&(e->exitlist));
	while (n) {
		s_window_t *w = (s_window_t*) s_list_element(n);
		s_window_exit(w);
		s_list_remove(&(e->exitlist), 0);
		n = s_list_first(&(e->exitlist));
	}
}

// exit a window. Called on s_window_quit.
void s_window_exit (s_window_t *window)
{
	s_list_node_t *node, *pnode;
	s_event_t ev;
	int id;
	debug("s_window_exit(%p)\n", window);
	if (window == NULL) return;
	id = window->id;
	if (id == 0) return;

	// Remove timers
	s_timer_del_wnd(window);

	// Clean-up matrix, send expose event
	window->visible = 0;
	s_server_pri_del(id);

	// Un-initialize client
	s_server_id_free(id);

	if ((window->parent != NULL) && s_server_id_find(window->parent) >= 0) {
		// Notify parent
		ev.type = CHCLOSED_EVENT;
		ev.data.child = window;
		s_window_event(window->parent, &ev);

		// Notify window
		ev.type = CLOSED_EVENT;
		s_window_event(window, &ev);

		if (window->type & WINDOW_MODAL) {
			// Reduce children count
			window->parent->num_modal--;
		}
	} else {
		// Notify window
		ev.type = CLOSED_EVENT;
		s_window_event(window, &ev);
	}

	// Notify wait list
	pnode = NULL;
	node = s_list_first(&window->waitlist);
	while (node) {
		s_window_wait_t *wait = (s_window_wait_t *) s_list_element(node);
		s_list_del_next(&window->waitlist, pnode);

		pthread_mutex_lock(&wait->mut);
		wait->closed = 1;
		pthread_mutex_unlock(&wait->mut);
		pthread_cond_signal(&wait->cond);
		// free() is performed by waiting thread

		if (pnode)
			node = s_list_next(pnode);
		else
			node = s_list_first(&window->waitlist);
	}

	// free the window struct
	s_window_uninit(window);
}

// wait until window closes
// do not call from window manager thread, you'll deadlock!
// do not take exynth mutex!
int s_window_wait (s_window_t *window)
{
	s_window_wait_t wait;
	s_exynth_t *e = get_exynth();
	debug("s_window_wait(%p)\n", window);
	wait.closed = 0;
	if (window == NULL) return 1;

	if (pthread_mutex_init(&wait.mut, NULL) != 0) goto err1;
	if (pthread_cond_init(&wait.cond, NULL) != 0) goto err2;

	// add to the window waitqueue
	pthread_mutex_lock(&e->mutex);
	if (s_server_id_find(window) < 0) {
		// window not registered, nothing to wait to
		debug("s_window_wait: window does not exits\n");
		wait.closed = 1;
		pthread_mutex_unlock(&e->mutex);
		goto err3;
	}
	if (s_list_add(&window->waitlist, &wait, 0) < 0) {
		debug("s_window_wait: failed adding window to wait list\n");
		pthread_mutex_unlock(&e->mutex);
		goto err3;
	}
	pthread_mutex_unlock(&e->mutex);

	// wait until released
	pthread_mutex_lock(&wait.mut);
	while (wait.closed == 0) {
		debug("s_window_wait: waiting for window to exit\n");
		if (pthread_cond_wait(&wait.cond, &wait.mut)) {
			pthread_mutex_unlock(&wait.mut);
			return -1;
		}
	}
	pthread_mutex_unlock(&wait.mut);

	debug("s_window_wait: done\n");
	// cleanup
err3:
	pthread_cond_destroy(&wait.cond);
err2:
	pthread_mutex_destroy(&wait.mut);
err1:
	return wait.closed;
}

// register window event handler function
void s_window_atevent (s_window_t *window, void (*f) (s_window_t *, s_event_t *))
{
	debug("s_window_atevent(%p)\n", window);
	window->atevent = f;
}

// broadcast an event to all windows
void s_window_broadcast_event(s_event_t *event)
{
	s_exynth_t *e = get_exynth();
	int i;
	for (i=S_CLIENTS_MAX-1; i>=0; i--) {
		if (e->clients[i] == NULL) continue;
		s_window_event(e->clients[i], event);
	}
}
