/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include "GuiCheckBox.h"
#include "GuiImageCache.h"
#include "GuiFrame.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor for the GUI check-box objects. Initializes check-box
 * state to 'unchecked', 'not grayed out'.
 *
 * Initially, none of the check-box images are loaded, so one has to initialize
 * them separately via loadEnabled(), loadDisabled() and loadGrayed() methods.
 * Automatic image initialization can be performed via GuiTemplate class.
 *
 * \param idResource Unique primitive identifier.
 * \param pFrame Pointer to a parent frame containing the primitive.
 */
GuiCheckBox::GuiCheckBox(int idResource, GuiFrame* pFrame)
		: GuiPrimitive(idResource, pFrame)
{
	imgOn = NULL;
	imgOff = NULL;
	isChecked = false;
	isGrayedOut = false;
	showChecked = false;
}

/**
 * \brief Class destructor
 *
 * Destructor for the check-box primitive objects.
 * Unloads currently loaded check-box images, if any.
 *
 */
GuiCheckBox::~GuiCheckBox()
{
	unload();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Load check-box image for the checked state
 *
 * Loads an image for the checked face of the check-box primitive.
 *
 * \param fileOn Name of the file containing the face image.
 * Parameter should either contain a full path or be relative to the
 * working directory.
 *
 * \return Returns true if the check-box image for the checked state was loaded
 * successfully.
 */
bool GuiCheckBox::loadEnabled(const char* fileOn)
{
	if(imgOn != NULL) GuiImageCache::closeImage(imgOn);

	imgOn = GuiImageCache::openImage(fileOn);

	return imgOn != NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Load check-box image for the unchecked state
 *
 * Loads an image for the unchecked face of the check-box primitive.
 *
 * \param fileOff Name of the file containing the face image.
 * Parameter should either contain a full path or be relative to the
 * working directory.
 *
 * \return Returns true if the check-box image for the unchecked state was loaded
 * successfully.
 */
bool GuiCheckBox::loadDisabled(const char* fileOff)
{
	if(imgOff != NULL) GuiImageCache::closeImage(imgOff);

	imgOff = GuiImageCache::openImage(fileOff);

	return imgOff != NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Load check-box image for the grayed-out state
 *
 * Loads an image for the grayed-out face of the check-box primitive.
 *
 * \param fileGrayed Name of the file containing the face image.
 * Parameter should either contain a full path or be relative to the
 * working directory.
 *
 * \return Returns true if the check-box image for the grayed-out state was loaded
 * successfully.
 */
bool GuiCheckBox::loadGrayed(const char* fileGrayed)
{
	if(imgGrayed != NULL) GuiImageCache::closeImage(imgGrayed);

	imgGrayed = GuiImageCache::openImage(fileGrayed);

	return imgGrayed != NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unload images
 *
 * Releases resources associated with loaded check-box images.
 */
void GuiCheckBox::unload()
{
	GuiImageCache::closeImage(imgOn);
	GuiImageCache::closeImage(imgOff);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the primitive interface
 *
 * Called by the parent frame whenever a primitive redraw is required.
 * \param drawSurface Pointer to a graphical surface where the primitive should
 * draw itself to.
 * \param paintArea The rectangular area of the parent frame surface that
 * should be re-painted by the primitive.
 */
void GuiCheckBox::onPaint(GuiSurface *drawSurface, s_rect_t &paintArea)
{
	if(!isVisible) return;

	s_rect_t surfRect;
	if(s_rect_intersect(&paintArea, &drawArea, &surfRect)) return;

	s_rect_t relRect = { surfRect.x - drawArea.x, surfRect.y - drawArea.y,
						surfRect.w, surfRect.h };

	s_image_t *img;
	if (isGrayedOut) img = imgGrayed;
	else img = showChecked ? imgOn : imgOff;

	if (img) drawSurface->drawImagePart(surfRect.x, surfRect.y, &relRect, img);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Selection change function of the primitive interface
 *
 * Called by the parent container whenever the selection is being
 * changed. Selection change inside a frame typically results of
 * a user touchscreen manipulation.
 *
 * \param type Specifies the new selection state of the given primitive
 */
void GuiCheckBox::onSelectionChanged(SELECTION type)
{
	if (isGrayedOut) return;

	// on every press, change the check state to the opposite
	switch (type) {
	case PRESSED:
		showChecked = !isChecked;
		break;
	case SELECTED:
		isChecked = showChecked;
		break;
	case UNKNOWN:
		showChecked = isChecked;
		break;
	default:
		break;
	}

	redraw();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set current check state
 *
 * Sets current check state of the check-box primitive to a specified
 * value. Also updates current selection of the containing frame.
 *
 * \param check New check state, 'true' if the check-box must go to checked state, 'false' if check-box must
 * be unchecked.
 */
void GuiCheckBox::setCheck(bool check)
{
	isChecked = check;
	showChecked = check;

	if (check) {
		setSelection(GuiPrimitive::SELECTED,0,0);
		pHolder->setFocusedItem(idResource);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get current check state
 *
 * Returns current check state of the check-box primitive.
 *
 * \return Returns true if the check-box is checked, false otherwise.
 */
bool GuiCheckBox::getCheck()
{
	bool check = isChecked;

	return check;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set current grayed-out state
 *
 * Sets current grayed-out state of the check-box primitive to a specified
 * value. Also updates current selection of the containing frame.
 * If a check box is grayed out, the grayed-out face is shown and the check
 * state is not changed when an user event is received.
 *
 * \param grayedOut New grayed out state, 'true' if the check-box must go to grayed-out state
 */
void GuiCheckBox::setGrayed(bool grayedOut)
{
	isGrayedOut = grayedOut;

	redraw();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get current grayed-out state
 *
 * Returns current grayed-out state of the check-box primitive.
 *
 * \return Returns true if the check-box is grayed out, false otherwise.
 */
bool GuiCheckBox::getGrayed()
{
	return isGrayedOut;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor for the GUI radio button objects. Initializes radio button
 * state to 'unchecked', 'not grayed out'.
 *
 * GuiRadioBox class is derived from the GuiCheckBox parent and inherits all
 * check-box methods except for the selection logic which is customized.
 *
 * Initially, none of the radio-box images are loaded,
 * so one has to initialize them separately via loadEnabled(),
 * loadDisabled() and loadGrayed() methods.
 *
 * \param idResource Unique primitive identifier.
 * \param pFrame Pointer to a parent frame containing the primitive.
 */
GuiRadioBox::GuiRadioBox(int idResource, GuiFrame* pFrame)
		: GuiCheckBox(idResource, pFrame)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Selection change function of the primitive interface
 *
 * Called by the parent container whenever the selection is being
 * changed. Selection change inside a frame typically results of
 * a user touchscreen manipulation.
 *
 * \param type Specifies the new selection state of the given primitive
 */
void GuiRadioBox::onSelectionChanged(SELECTION type)
{
	// set the checked state according to selection
	switch (type) {
	case PRESSED:
		if (isGrayedOut) return;
		showChecked = true;
		break;
	case SELECTED:
		if (isGrayedOut) return;
		isChecked = true;
		showChecked = true;
		break;
	case UNSELECTED:
		isChecked = false;
		showChecked = false;
	case UNKNOWN:
		showChecked = isChecked;
	}

	redraw();
}
