#ifndef GUICHECKBOX_H_
#define GUICHECKBOX_H_

#include "GuiPrimitive.h"

#define GUICHECKBOX(f,p) getPrimitive<GuiCheckBox>(f,p)

/**
 * \brief Checkbox primitive
 *
 * Implements a standard check-box element. The check box
 * can either be 'grayed out' or active. Visually, the check-box has three
 * distinct states, being checked and unchecked when active or grayed out.
 * Three image files are thus associated with the check-box, covering all states.
 *
 * When the check-box is initially pressed, the checked state is applied and an
 * appropriate image is drawn immediately. The check-box remains checked if
 * the pressure leaves the check-box within its area. If the pressure is
 * dragged outside of the check-box and released outside, the check-box restores
 * its unchecked state.
 *
 * All images associated with the check-box must have the same dimensions.
 * Images may have arbitrary transparent regions via the PNG alpha channel.
 *
 * The check-box primitive does not provide any text and is fully defined
 * using the above two images. If required, a text box can be placed next to check
 * box and the touch area of the check box can be expanded to react on the text
 * box area as well.
 *
 * XML code below defines a check-box primitive:
 *
 * \code
 *	<item id="ID" type="check" x="X" y="Y" width="W" height="H">
 * 		<init on="/path/to/on" off="/path/to/off" grayed="/path/to/grayed" />
 * 		<touch x="tX" y="tY" width="tW" height="tH"> />
 * 	</item>
 * \endcode
 *
 * Where
 * - ID represents the primitive unique identifier
 * - X, Y, W, H is the set of primitive coordinates in the parent frame
 * - /path/to/on specifies the location of an image file for the checked state
 * - /path/to/off specifies the location of an image file for the unchecked state
 * - /path/to/grayed specifies the location of an 'grayed-out' image file
 * - tX, tY, tW, tH are the coordinates of the touchable area in the parent frame
 *
 */

class GuiCheckBox : public GuiPrimitive
{
public:
	GuiCheckBox(int idResource, GuiFrame* pFrame);
	virtual ~GuiCheckBox();

	bool loadEnabled(const char* fileOn);
	bool loadDisabled(const char* fileOff);
	bool loadGrayed(const char* fileOff);
	void unload();

	void setCheck(bool check);
	void setGrayed(bool check);
	bool getCheck();
	bool getGrayed();

protected:

	virtual void onPaint(GuiSurface *drawSurface, s_rect_t &paintArea);
	virtual void onSelectionChanged(SELECTION type);

	/** Flags whether the checkbox is checked or not **/
	bool		isChecked;
	/** Used to show the visual reaction of checking/unchecking **/
	bool		showChecked;
	/** Flags whether the checkbox is grayed out or not **/
	bool		isGrayedOut;

private:
	s_image_t*	imgOn;
	s_image_t*	imgOff;
	s_image_t*	imgGrayed;
};

/**
 * \brief Radio button primitive
 *
 * Implements a standard radio button element. The radio button
 * can either be 'grayed out' or active. Visually, the radio button has three
 * distinct states, being checked or unchecked when active - or grayed out.
 * Three image files are thus associated with the radio button, covering all states.
 * The difference between a check-box and a button is that check-boxes are
 * fully independent of each other, while radio buttons are grouped by the frame
 * and only one radio button can be selected in its group at the same time.
 *
 * As in the case of the check-box primitive, three image files are associated
 * with the radio button, covering both states.
 *
 * When the radio button is initially pressed, the checked state is applied and an
 * appropriate image is drawn immediately. The radio button remains checked if
 * the pressure leaves the radio button within its area. At the same time the last
 * selected radio button in the group is unselected, ensuring that the current
 * selection is the only one. If the pressure is dragged outside of the
 * radio button and released outside, the radio button restores its unchecked state.
 *
 * All images associated with the radio button must have the same dimensions.
 * Images may have arbitrary transparent regions via the PNG alpha channel.
 *
 * The radio button primitive does not provide any text and is fully defined
 * using the above three images.
 *
 * XML code below defines a radio button primitive:
 *
 * \code
 *	<item id="ID" type="radio" x="X" y="Y" width="W" height="H">
 * 		<init on="/path/to/on" off="/path/to/off" grayed="/path/to/grayed" />
 * 		<touch x="tX" y="tY" width="tW" height="tH"> />
 * 	</item>
 * \endcode
 *
 * Where
 * - ID represents the primitive unique identifier
 * - X, Y, W, H is the set of primitive coordinates in the parent frame
 * - /path/to/on specifies the location of an image file for the selected state
 * - /path/to/off specifies the location of an image file for the unselected state
 * - /path/to/grayed specifies the location of an 'grayed-out' image file
 * - tX, tY, tW, tH are the coordinates of the touchable area in the parent frame
 *
 */

class GuiRadioBox : public GuiCheckBox
{
public:

	GuiRadioBox(int idResource, GuiFrame* pFrame);

protected:

	virtual void onSelectionChanged(SELECTION type);
};

#endif /*GUICHECKBOX_H_*/
