#include <stdlib.h>
#include <stdio.h>

#include "GuiDrawArea.h"

#include "GuiWindow.h"

/**
 * \brief Default constructor
 *
 * Default constructor of the drawing area. Sets drawing starting poisiton to
 * top-left corner.
 *
 * \param idResource Unique primitive identifier.
 * \param pFrame Pointer to a parent frame containing the primitive.
 */
GuiDrawArea::GuiDrawArea(int idResource, GuiFrame* pFrame)
		: GuiPrimitive(idResource, pFrame)
{
	imageData = NULL;
	x0 = 0;
	y0 = 0;
}

/**
 * \brief Object destructor
 */
GuiDrawArea::~GuiDrawArea()
{

}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize the drawing area
 *
 * GuiDrawArea is initialized when first pixel is drawn!
 * After that, the memory will not be re-initialized when size is changed!
 *
 * \return Returns true if the initialization was successful, false otherwise
 */
bool GuiDrawArea::initArea()
{
	if (imageData) {
		s_image_uninit(imageData);
		imageData = NULL;
	}

	s_image_init(&imageData);
	if (!imageData) return false;

	imageData->rgba = (unsigned int *)calloc(drawArea.w * drawArea.h, sizeof(unsigned int *));
	imageData->w = drawArea.w;
	imageData->h = drawArea.h;

	if (imageData->rgba) {
		for (int i=0; i<imageData->w*imageData->h; i++) imageData->rgba[i]=0xFF;
		return true;
	}

	s_image_uninit(imageData);
	imageData = NULL;
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Clear drawing area
 *
 * Clears the drawing area by setting all the pixels to 0xff
 */
void GuiDrawArea::clear()
{
	if (!imageData) {
		if (!initArea()) return;
	}
	for (int i=0; i<imageData->w*imageData->h; i++) imageData->rgba[i]=0xFF;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set pixel color
 *
 * Sets the color of the pixels
 * \param rgb Pointer to string containing the RGB color value to use
 * \param x X coordinate of the pixel
 * \param y Y coordinate of the pixel
 */
void GuiDrawArea::setPixel(const char* rgb, int x, int y)
{
	setPixel(GuiSurface::stringToRGB(rgb), x, y);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set pixel color
 *
 * Sets the color of the pixels
 * \param col RGB color structure containing RGB values to use
 * \param x X coordinate of the pixel
 * \param y Y coordinate of the pixel
 * \param alpha Transparency level
 */
void GuiDrawArea::setPixel(RGB_COLOR col, int x, int y, signed char alpha)
{
	if (!imageData) {
		if (!initArea()) return;
	}

	if (x<0 || y<0 || x>=imageData->w || y >= imageData->h) return;

	imageData->rgba[x+ y*imageData->w] = (col.red << 0x18) | (col.green << 0x10) | (col.blue << 0x08) | (alpha & 0xFF);
	x0 = x;
	y0 = y;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Draw line using specified color
 * Draws a line from (x0,y0) to (x1,y1) using specified color
 *
 * \param rgb Pointer to string containing RGB values to use
 * \param x1 X coordinate of the line end-point
 * \param y1 Y coordinate of the line end-point
 */
void GuiDrawArea::lineTo(const char *rgb, int x1, int y1)
{
	lineTo(GuiSurface::stringToRGB(rgb), x1, y1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Draw line using specified color
 * Draws a line from (x0,y0) to (x1,y1) using specified color
 *
 * \param col RGB color structure that contains color value to use for drawing
 * \param x1 X coordinate of the line end-point
 * \param y1 Y coordinate of the line end-point
 */
void GuiDrawArea::lineTo(RGB_COLOR col, int x1, int y1) {
	int ox1 = x1;
	int oy1 = y1;
	bool steep = abs(y1 - y0) > abs(x1 - x0);
	if (steep) {
		int tmp = x0;
		x0 = y0;
		y0 = tmp;
		tmp = x1;
		x1 = y1;
		y1 = tmp;
	}
	if (x0 > x1) {
		int tmp = x0;
		x0 = x1;
		x1 = tmp;
		tmp = y0;
		y0 = y1;
		y1 = tmp;
	}

	int deltax = x1 - x0;
	int deltay = abs(y1 - y0);
	int error = deltax / 2;
	int ystep;
	int y = y0;

	if (y0 < y1)
		ystep = 1;
	else
		ystep = -1;

	for (int x = x0+1; x < x1; x++) {
		if (steep)
			setPixel(col, y, x);
		else
			setPixel(col, x, y);
		error = error - deltay;
		if (error < 0) {
			y += ystep;
			error += deltax;
		}
	}
	setPixel(col, ox1, oy1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the primitive interface
 *
 * Called by the parent frame whenever a primitive redraw is required.
 * \param drawSurface Pointer to a graphical surface where the primitive should
 * draw itself to.
 * \param paintArea The rectangular area of the parent frame surface that
 * should be re-painted by the primitive.
 */
void GuiDrawArea::onPaint(GuiSurface *drawSurface, s_rect_t &paintArea)
{
	if(!isVisible || imageData == NULL) return;

	s_rect_t surfRect;
	if(s_rect_intersect(&paintArea, &drawArea, &surfRect)) return;

	s_rect_t relRect = { surfRect.x - drawArea.x, surfRect.y - drawArea.y, surfRect.w, surfRect.h };
	drawSurface->drawImagePart(surfRect.x, surfRect.y, &relRect, imageData);
}
