/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include "GuiLibrary.h"
#include "GuiEditBox.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default constructor
 * Creates the edit box primitive and positions the cursor to the default location
 *
 * \param idResource Unique primitive identifier.
 * \param pFrame Pointer to a parent frame containing the primitive.
 */
GuiEditBox::GuiEditBox(int idResource, GuiFrame* pFrame)
		: GuiMultiTextBox(idResource, pFrame)
{
	cursor = 4;
	cursorLine = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default destructor
 *
 * Destroys the edit box object
 */
GuiEditBox::~GuiEditBox()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set cursor position
 *
 * Sets the cursor position
 * \param pos Position where the cursor should be placed at
 */
void GuiEditBox::setCursor(size_t pos)
{
	cursor = pos;
	cursorLine = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set cursor position
 *
 * Sets the cursor position
 * \param x X coordinate where the cursors should be placed at
 * \param y Y coordinate where the cursor should be placed at
 */
void GuiEditBox::setCursor(int x, int y)
{
	if (!lineList) return;

	x -= drawArea.x;
	y -= drawArea.y;

	if(!isRendered) renderText();

	if(!isRendered)	return;

	// calculate vertical placement of the whole block
	int oy = 0;
	switch(vertAlign)
	{
	case TOP:
		oy = 0;
		break;
	case BOTTOM:
		oy = drawArea.h - textHeight;
		break;
	case VCENTER:
		oy = (drawArea.h - textHeight) / 2;
		break;
	}

	// find the touch line
	int line = 0;
	std::list<textLine>::iterator itl = lineList->begin();
	for(std::list<s_font_t*>::iterator it = fontData.begin();
			it !=fontData.end() && itl != lineList->end();
			itl++, line++, oy += lineHeight)
	{
		s_font_t* font = *it;
		it ++;
		if(!font) continue;

		// check if touch was on this line or this is last line
		if (y > oy + lineHeight && it != fontData.end()) continue;
		cursorLine = line;

		// calculate text X offset for h align
		int ox = 0;
		switch(horzAlign)
		{
		case LEFT:
			ox = 0;
			break;
		case RIGHT:
			ox = drawArea.w - font->glyph.img->w;
			break;
		case HCENTER:
			ox = (drawArea.w - font->glyph.img->w) / 2;
			break;
		}

		// find the cursor row
		cursor = itl->startChar + itl->numChars;
		cursorPos = itl->posX[itl->numChars];
		for (size_t i = 0; i<itl->numChars; i++){
			if (x - ox + font->glyph.xMin < (itl->posX[i] + itl->posX[i+1])/2) {
				cursor = itl->startChar + i;
				cursorPos = itl->posX[i];
				break;
			}
		}

		break;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Select the text
 * Selects the text in the edit box
 *
 * \param type Primitive selection state
 * \param x X coordinate of the selection
 * \param y Y coordinate of the selection
 */
void GuiEditBox::setSelection(SELECTION type, int x, int y)
{
	if (type == PRESSED) setCursor(x,y);
	GuiPrimitive::setSelection(type, x, y);
	redraw();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get cursor position
 *	Get the current position of the cursor
 * \return Returns the current cursor position
 */
size_t GuiEditBox::getCursor()
{
	return cursor;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Insert the text into cursor position
 * Inserts the text into cursor position and moves the cursor behind the inserted text
 * \param t Pointer to the text being inserted
 */
void GuiEditBox::insertText(const char *t)
{
	size_t pos = GuiLibrary::utf8pos2pos(text.c_str(), cursor);
	if (pos >= text.length())
		text += t;
	else
		text.insert(pos,t);
	setText(text.c_str());

	// move cursor behind our inserted text block
	size_t tlen = GuiLibrary::utf8strlen(t);
	cursor += tlen;
	cursorLine = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Delete the text
 * Deletes the text from the cursor position
 */
void GuiEditBox::delText()
{
	if (text.empty() || cursor <= 0) {
		return;
	}
	size_t pos = GuiLibrary::utf8pos2pos(text.c_str(), cursor);
	if (pos <= 0) return;
	if (pos >= text.length()) pos = text.length();

	// find previous UTF-8 char start
	size_t pos2 = pos-1;
	size_t len = text.length() - pos2;
	while (pos2 > 0) {
		if (s_font_utf8_to_unicode(NULL, text.c_str()+pos2, len) > 0) break;
		pos2--;
		len++;
	}

	// delete the last utf-8 char
	text.erase(pos2, pos - pos2);
	setText(text.c_str());

	if (cursor > 0) cursor--;
	cursorLine = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Sets the edit box text
 * Sets the text that will be displayed in edit boc
 * \param t Text to be shown
 */
void GuiEditBox::setText(const char *t)
{
	// create local copy of the text
	text = t;
	setText(text.c_str());

	// set cursor to the end of the text
	cursor = GuiLibrary::utf8strlen(t);
	cursorLine = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get edit box text
 * Retrieves the text that edit box holds
 *
 * \return Returns the edit box text as string
 */
std::string GuiEditBox::getText()
{
	std::string tmp = text;
	return tmp;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Calculates the cursor position
 * Calculates the position (X coordinate) and the line number where the
 * cursor is located at
 */
void GuiEditBox::calcCursorPos()
{
	cursorLine = 0;
	cursorPos = 0;
	if (!lineList) return;

	size_t line = 0;
	for(std::list<textLine>::iterator it = lineList->begin();
			it != lineList->end(); it++, line++) {
		cursorLine = line;
		if (cursor <= it->startChar + it->numChars) {
			if (cursor >= it->startChar)
				cursorPos = it->posX[cursor - it->startChar];
			else
				cursorPos = it->posX[0];
			break;
		} else {
			cursorPos = it->posX[it->numChars];
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the primitive interface
 *
 * Called by the parent frame whenever a primitive redraw is required.
 * \param drawSurface Pointer to a graphical surface where the primitive should
 * draw itself to.
 *
 * \param paintArea The rectangular area of the parent frame surface that
 * should be re-painted by the primitive.
 */
void GuiEditBox::onPaint(GuiSurface *drawSurface, s_rect_t &paintArea)
{
	// return immediately if the primitive is not visible
	if(!isVisible) return;

	// check if the multi-line text is rendered
	if(!isRendered) renderText();

	if (cursorLine < 0) calcCursorPos();

	// sanity check, do we have any data to draw?
	if(!isRendered || fontData.empty()) return;

	// clip the paintArea to control area
	s_rect_t repaintRect;
	if(s_rect_intersect(&drawArea, &paintArea, &repaintRect)) return;

	// calculate vertical placement of the whole block
	int oy = 0;
	switch(vertAlign)
	{
	case TOP:
		oy = 0;
		break;
	case BOTTOM:
		oy = drawArea.h - textHeight;
		break;
	case VCENTER:
		oy = (drawArea.h - textHeight) / 2;
		break;
	}

	// draw the text
	int line = 0;
	for(std::list<s_font_t*>::iterator it = fontData.begin();
			it !=fontData.end(); it ++, line++)
	{
		s_font_t* font = *it;
		if(!font) {
			oy += lineHeight;
			continue;
		}
		// baseline alignment compensation within lines
		int oy2 = font->ascent - font->glyph.yMax + 1;

		int ox = 0;
		switch(horzAlign)
		{
		case LEFT:
			ox = 0;
			break;
		case RIGHT:
			ox = drawArea.w - font->glyph.img->w;
			break;
		case HCENTER:
			ox = (drawArea.w - font->glyph.img->w) / 2;
			break;
		}

		s_rect_t textArea = { drawArea.x + ox, drawArea.y + oy + oy2,
				drawArea.w - ox, font->height - oy2};


		s_rect_t surfRect;
		if(!s_rect_intersect(&repaintRect, &textArea, &surfRect)) {
			s_rect_t sourceRect = { surfRect.x - textArea.x, surfRect.y - textArea.y,
					surfRect.w, surfRect.h };

#ifdef DEBUG
			printf("GuiMultiTextBox::onPaint: line, %d, %d, %d, %d to %d, %d\n",
					font->glyph.img->x, font->glyph.img->y,
					font->glyph.img->w, font->glyph.img->h,
					surfRect.x, surfRect.y);
#endif

			drawSurface->drawImagePart(surfRect.x, surfRect.y,
					&sourceRect, font->glyph.img);
		}

		if (line == cursorLine) {
			s_rect_t cursorRect = { drawArea.x + ox + cursorPos - font->glyph.xMin,
					drawArea.y + oy, CURSOR_WIDTH, font->ascent - font->descent};
			if (ox + cursorPos - font->glyph.xMin > drawArea.w)
				cursorRect.x = drawArea.w - CURSOR_WIDTH;
			if(!s_rect_intersect(&repaintRect, &cursorRect, &surfRect)) {
				drawSurface->fillRect(&surfRect, drawSurface->makeColor(&fontColor));
			}
		}

		oy += lineHeight;
	}
}
