/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include "GuiEffectWindow.h"

#define SLIDE_HORZ_STEP		12
#define SLIDE_VERT_STEP		12
#define EFFECT_DELAY		5

//#define DEBUG

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor for the effect-enabled window object.
 * Effect mutex is created here. Initially, all effects are disabled.
 *
 * \param idResource Unique identifier of the primitive.
 */
GuiEffectWindow::GuiEffectWindow(int idResource)
			: GuiWindow(idResource)
{
	originalPos.x = originalPos.y = originalPos.w = originalPos.h = 0;
	effectTimer = 0;

	openEffectType = OPEN_EFFECT_NONE;
	closeEffectType = CLOSE_EFFECT_NONE;
	showWindowType = SHOW_WINDOW;

	effectVertSpeed = SLIDE_HORZ_STEP;
	effectHorzSpeed = SLIDE_VERT_STEP;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destructor for the effect-enabled window object. Also destroys the effect
 * mutex.
 */
GuiEffectWindow::~GuiEffectWindow()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start effect timer
 *
 * Starts the animation timer with the pre-defined interval. Each time the
 * timer fires, a next image from the animation list is displayed by the
 * primitive. The animation is always started from the first frame.
 *
 * \param effect Effect type that should be generated when timer fires.
 *
 * \return Returns true if the timer was started successfully.
 */

bool GuiEffectWindow::startEffectTimer(int effect)
{
	stopEffectTimer();

	if (wndHandle == NULL) return false;

	if (s_timer_init(&effectTimer) != 0) return false;
	effectTimer->window = wndHandle;
	effectTimer->cb = effectTimerHandlerProc;
	effectTimer->data = (void*)effect;
	effectTimer->timeval = EFFECT_DELAY;
	if (s_timer_add(effectTimer) != 0) {
		s_timer_uninit(effectTimer);
		effectTimer = NULL;
		return false;
	}
	refreshRealSurface = false;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stops animation
 *
 * Stops the animation and destroys its timer.
 */

void GuiEffectWindow::stopEffectTimer()
{
	if(effectTimer) {
		s_timer_del(effectTimer);
		s_timer_uninit(effectTimer);
		effectTimer = NULL;
		refreshRealSurface = true;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Show or hide window
 *
 * Controls the window visibility, i.e. shows or hides the window.
 * onShowWindow() virtual function is called internally to notify the derived
 * classes about the new state.
 * If an effect is set up, it is launched.
 *
 * \param state Set to true to make the window visible.
 * \return Returns true if the window visibility has been updated successfully.
 */

bool GuiEffectWindow::showWindow(bool state)
{
	if(wndHandle == NULL || !isReady) return false;

	// add the window to priority list, visible=0
	// (to maintain correct window order when more
	// windows are created in onInitialize().
	if(state) s_window_show_prepare(wndHandle);

	// on first show, initialize window
	if (!isInitialized) {
		isInitialized = true;
		onInitialize();
	}

	// perform custom show/hide actions
	if(!onShowWindow(state ? SHOW_WINDOW : HIDE_WINDOW, 0)) {
		return true;
	}

	if(state) {
		showWindowType = SHOW_WINDOW;
		startShowEvent();
		s_window_show(wndHandle);
	}
	else {
		showWindowType = HIDE_WINDOW;
		if (!startHideEvent())
			s_window_hide(wndHandle);
	}

	onPostShowWindow(state ? SHOW_WINDOW : HIDE_WINDOW, 0);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close window
 *
 * Sends a command to the window instructing it to close and release all
 * contained GUI primitives. Also, virtual onShowWindow() function is called
 * in the calling thread to notify the derived classes about state change.
 * If an effect is set up, the effect is launched instead and the
 * function returns. In this case, the window is not actually closed once
 * this function returns!
 *
 * \param retCode Specifies the window return code. Return code may be
 * retrieved after the window is closed for further processing.
 *
 * \return Returns true if the closing procedure was initiated.
 */
bool GuiEffectWindow::closeWindow(int retCode)
{
	#ifdef DEBUG
	printf("GuiEffectWindow(%p)::closeWindow: id %d code %d\n", this, idResource, retCode);
	#endif

	if(wndHandle == NULL || !isReady) return false;

	if(!onShowWindow(CLOSE_WINDOW, retCode))
	{
		#ifdef DEBUG
		printf("GuiWindow(%p)::closeWindow: postponed\n", this);
		#endif
		return true;
	}
	returnCode = retCode;

	showWindowType = CLOSE_WINDOW;
	if(startHideEvent()) return true;

	s_window_quit(wndHandle);

	onPostShowWindow(CLOSE_WINDOW, retCode);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * Start a show-window effect.
 *
 * \return Returns true if the effect was started successfully
 */
bool GuiEffectWindow::startShowEvent()
{
	originalPos = windowPos;

	stopEffectTimer();

	switch(openEffectType)
	{
	case OPEN_EFFECT_NONE:
		return false;
	case OPEN_EFFECT_SLIDE_BOTTOM:
		windowPos.y = -windowPos.h;
		break;
	case OPEN_EFFECT_SLIDE_TOP:
		windowPos.y = wndHandle->surface->height;
		break;
	case OPEN_EFFECT_SLIDE_LEFT:
		windowPos.x = wndHandle->surface->width;
		break;
	case OPEN_EFFECT_SLIDE_RIGHT:
		windowPos.x = -windowPos.w;
		break;
	}

	setWindowPos(windowPos);

	#ifdef DEBUG
	printf("GuiEffectWindow::onShowWindow: starting timer for effect %d\n", openEffectType);
	#endif
	return startEffectTimer(openEffectType);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * Start a hide-window effect.
 *
 * \return Returns true if the effect was started successfully
 */
bool GuiEffectWindow::startHideEvent()
{
	originalPos = windowPos;

	stopEffectTimer();

	if(closeEffectType == CLOSE_EFFECT_NONE) return false;

	#ifdef DEBUG
	printf("GuiEffectWindow::onShowWindow: starting timer for effect %d\n", closeEffectType);
	#endif
	return startEffectTimer(closeEffectType);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Timer event handler procedure
 *
 * Function is called by exynth whenever a timer is fired
 * \param window Pointer to the window structure the timer is associated with
 * \param timer Pointer to the timer structure
 */
void GuiEffectWindow::effectTimerHandlerProc(s_window_t *window, s_timer_t *timer)
{
	s_window_t *handle = timer->window;
	GuiEffectWindow *pWindow = (GuiEffectWindow*)handle->data;
	if (!pWindow) return;

	pWindow->effectHandler((int)timer->data);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Timer event handler
 *
 * Function is called whenever a timer is fired. In the
 * context of effect-enabled windows, the timer is used to
 * implement window slide-in and slide-out animations.
 *
 * \param effect Specifies the effect type to handle
 */
void GuiEffectWindow::effectHandler(int effect)
{
	#ifdef DEBUG
	printf("GuiEffectWindow::effectHandler: %d\n", effect);
	#endif

	switch(effect)
	{
	case OPEN_EFFECT_SLIDE_BOTTOM:

		if(windowPos.y >= originalPos.y)
		{
			stopEffectTimer();
			break;
		}

		windowPos.y += originalPos.y - windowPos.y < effectVertSpeed ?
				originalPos.y - windowPos.y : effectVertSpeed;

		if(windowPos.y >= originalPos.y)
			refreshRealSurface = true;

		setWindowPos(windowPos);
		break;

	case OPEN_EFFECT_SLIDE_RIGHT:

		if(windowPos.x >= originalPos.x)
		{
			stopEffectTimer();
			break;
		}

		windowPos.x += originalPos.x - windowPos.x < effectHorzSpeed ?
				originalPos.x - windowPos.x : effectHorzSpeed;

		if(windowPos.x >= originalPos.x)
			refreshRealSurface = true;


		#ifdef DEBUG
		printf("GuiEffectWindow::onTimer: open %d %d %d %d\n",
				windowPos.x, windowPos.y, windowPos.w, windowPos.h);
		#endif
		setWindowPos(windowPos);
		break;

	case OPEN_EFFECT_SLIDE_TOP:

		if(windowPos.y <= originalPos.y)
		{
			stopEffectTimer();
			break;
		}

		windowPos.y -= windowPos.y - originalPos.y < effectVertSpeed ?
				windowPos.y - originalPos.y : effectVertSpeed;

		if(windowPos.y <= originalPos.y)
			refreshRealSurface = true;

		setWindowPos(windowPos);
		break;

	case OPEN_EFFECT_SLIDE_LEFT:

		if(windowPos.x <= originalPos.x)
		{
			stopEffectTimer();
			break;
		}

		windowPos.x -= windowPos.x - originalPos.x < effectHorzSpeed ?
				windowPos.x - originalPos.x : effectHorzSpeed;

		if(windowPos.x <= originalPos.x)
			refreshRealSurface = true;

		setWindowPos(windowPos);
		break;

	case CLOSE_EFFECT_SLIDE_BOTTOM:

		if(windowPos.y >= wndHandle->surface->height)
		{
			stopEffectTimer();

			if(showWindowType == HIDE_WINDOW)
				s_window_hide(wndHandle);
			else if(showWindowType == CLOSE_WINDOW) {
				s_window_quit(wndHandle);
				onPostShowWindow(CLOSE_WINDOW, returnCode);
			}
			break;
		}

		windowPos.y += effectVertSpeed;
		setWindowPos(windowPos);
		break;

	case CLOSE_EFFECT_SLIDE_RIGHT:

		if(windowPos.x >= wndHandle->surface->width)
		{
			stopEffectTimer();

			if(showWindowType == HIDE_WINDOW)
				s_window_hide(wndHandle);
			else if(showWindowType == CLOSE_WINDOW) {
				s_window_quit(wndHandle);
				onPostShowWindow(CLOSE_WINDOW, returnCode);
			}
			break;
		}

		windowPos.x += effectHorzSpeed;

		#ifdef DEBUG
		printf("GuiEffectWindow::onTimer: close %d %d %d %d\n",
				windowPos.x, windowPos.y, windowPos.w, windowPos.h);
		#endif

		setWindowPos(windowPos);
		break;

	case CLOSE_EFFECT_SLIDE_TOP:

		if(windowPos.y + windowPos.h <= 0)
		{
			stopEffectTimer();

			if(showWindowType == HIDE_WINDOW)
				s_window_hide(wndHandle);
			else if(showWindowType == CLOSE_WINDOW) {
				s_window_quit(wndHandle);
				onPostShowWindow(CLOSE_WINDOW, returnCode);
			}
			break;
		}

		windowPos.y -= effectVertSpeed;
		setWindowPos(windowPos);
		break;

	case CLOSE_EFFECT_SLIDE_LEFT:

		if(windowPos.x + windowPos.w <= 0)
		{
			stopEffectTimer();

			if(showWindowType == HIDE_WINDOW)
				s_window_hide(wndHandle);
			else if(showWindowType == CLOSE_WINDOW) {
				s_window_quit(wndHandle);
				onPostShowWindow(CLOSE_WINDOW, returnCode);
			}
			break;
		}

		windowPos.x -= effectHorzSpeed;
		setWindowPos(windowPos);
		break;
	}

	#ifdef DEBUG
	printf("GuiEffectWindow::effectHandler: %d unlocked\n", effect);
	#endif
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set window effects
 *
 * Specifies the type of window opening and closing effects. Can also disable
 * effects completely.
 *
 * \param openEffect Sets the window opening effect, slide-in animations are
 * currently supported.
 * \param closeEffect Sets the window closing effect, slide-out animations are
 * currently supported.
 */
void GuiEffectWindow::setEffects(OPEN_EFFECT_TYPE openEffect, CLOSE_EFFECT_TYPE closeEffect)
{
	openEffectType = openEffect;
	closeEffectType = closeEffect;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set effect speed
 *
 * Updates the speed of window effects or animations in both vertical and
 * horizontal direction.
 *
 * \param vertSpeed Specifies the step of a vertical slide animation,
 * in pixels
 * \param horzSpeed Specifies the step of a horizontal slide animation,
 * in pixels
 */
void GuiEffectWindow::setEffectSpeed(int vertSpeed, int horzSpeed)
{
	effectVertSpeed = vertSpeed;
	effectHorzSpeed = horzSpeed;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Checks if an effect is running
 *
 * \return Returns true if a window effect is currently running
 */
bool GuiEffectWindow::effectInProgress()
{
	return (effectTimer != 0);
}
