/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <time.h>
#include <stdio.h>

#include "GuiFrame.h"
#include "GuiWindow.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor for the GUI window frame object.
 * Initializes the frame geometry to zero, resets currently focused item,
 * sets visibility flag to true and viewpoint offset to zero.
 *
 * With the above attributes, the frame is completely empty. User must either
 * manually populate the frame with instantiated GUI primitives by calling
 * addPrimitive() method, or rely on automatic GUI skin function provided
 * by GuiTemplate class.
 *
 * \param idResource Unique identifier of the window frame
 * \param pParent Pointer to a window that contains the given frame
 */
GuiFrame::GuiFrame(int idResource, GuiWindow *pParent)
{
	this->idResource = idResource;

	parentWindow = pParent;
	focusedItem = RES_ID_INVALID;

	frameArea.x = frameArea.y = 0;
	frameArea.w = frameArea.h = 0;
	offsetX = offsetY = 0;
	virtWidth = virtHeight = 0;
	updateRequired = true;
	isVisible = true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destructor for the GUI window frame object.
 * Deletes any contained remaining primitives which were not released
 * before for some reason.
 */
GuiFrame::~GuiFrame()
{
	list<GuiPrimitive*>::iterator it;
	for(it = primitiveList.begin(); it != primitiveList.end(); it ++)
	{
		GuiPrimitive* pItem = (GuiPrimitive*)*it;
		delete pItem;
	}

	primitiveList.clear();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Add GUI primitive to frame
 *
 * Inserts the provided primitive at the end of the primitive list of the
 * frame.
 *
 * \param pItem Pointer to a GUI primitive object to be added
 * \return Returns a number of primitives currently associated with the given
 * frame.
 */
int GuiFrame::addPrimitive(GuiPrimitive *pItem)
{
	primitiveList.push_back(pItem);
	return primitiveList.size();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get a primitive by identifier
 *
 * Retrieves a primitive object by its unique identifier from the list.
 *
 * \param idResource Identifier of a primitive to be retrieved. If multiple
 * primitives have the same identifier, the first one is returned.
 * \return Returns the first primitive object that has a provided ID.
 */
GuiPrimitive* GuiFrame::getPrimitive(int idResource)
{
	list<GuiPrimitive*>::iterator it;
	for(it = primitiveList.begin(); it != primitiveList.end(); it ++)
	{
		if(((GuiPrimitive*)*it)->getResId() == idResource)
			return *it;
	}

	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Remove a primitive by identifier
 *
 * Erases a GUI primitive from the primitive list of the frame. The first
 * primitive with the provided identifier is erased, the primitive object is
 * also deleted automatically.
 *
 * \param idResource Identifier of a primitive to be removed
 * \return Returns true if the primitive was found and erased from the list.
 */
bool GuiFrame::removePrimitive(int idResource)
{
	list<GuiPrimitive*>::iterator it;
	for(it = primitiveList.begin(); it != primitiveList.end(); it ++)
	{
		GuiPrimitive* pItem = (GuiPrimitive*)*it;
		if(pItem->getResId() == idResource)
		{
			// remove it from the list
			primitiveList.erase(it);
			delete pItem;

			// remove selection too
			if(focusedItem == idResource) focusedItem = RES_ID_INVALID;

			return true;
		}
	}
	return false;
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Remove a range of primitives
 *
 * Erases a GUI primitives from the primitive list of the frame. The primitive
 * objects are also deleted automatically.
 *
 * \param minId Identifier of first primitive to be removed
 * \param maxId Identifier of last primitive to be removed
 * \return Returns true if primitives were found and erased from the list.
 */
bool GuiFrame::removePrimitives(int minId, int maxId)
{
	bool ret = false;

	list<GuiPrimitive*>::iterator it;
	for(it = primitiveList.begin(); it != primitiveList.end();)
	{
		GuiPrimitive* pItem = (GuiPrimitive*)*it;
		int id = pItem->getResId();
		if(id >= minId && id <= maxId)
		{
			// remove it from the list
			it = primitiveList.erase(it);
			delete pItem;

			// remove selection too
			if(focusedItem == id) focusedItem = RES_ID_INVALID;

			ret = true;
		} else {
			it ++;
		}
	}
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set viewpoint offset
 *
 * Sets the viewpoint coordinates of the given frame. Viewpoint offset is only
 * relevant for frames that have virtual dimensions larger than the area
 * displayed in the parent window. Given function is useful for implementing
 * scrollable GUI elements. Offset must be always positive.
 *
 * \param x X coordinate of the viewpoint offset in a virtual frame.
 * \param y Y coordinate of the viewpoint offset in a virtual frame.
 */
void GuiFrame::setOffset(int x, int y)
{
	offsetX = x;
	offsetY = y;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get current viewpoint offset
 *
 * \param pX Pointer to an integer that receives current X coordinate of the
 * viewpoint offset. If NULL value is specifed, X coordinate is not returned.
 * \param pY Pointer to an integer that receives current Y coordinate of the
 * viewpoint offset. If NULL value is specifed, Y coordinate is not returned.
 */
void GuiFrame::getOffset(int *pX, int *pY)
{
	if(pX) *pX = offsetX;
	if(pY) *pY = offsetY;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set frame coordinates
 *
 * Updates coordinates of the given window frame. Coordinates are calculated in
 * relation to the upper left corner of the parent window.
 * The visible are of the frame is bounded by the given rectangle.
 *
 * \note The frame is not redrawn automatically.
 *
 * \param frameArea Frame coordinate rectangle
 */
void GuiFrame::setGeometry(s_rect_t &frameArea)
{
	if (frameArea.w < 0 || frameArea.h < 0) return;
	this->frameArea = frameArea;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get frame coordinates
 *
 * Reads coordinates of the given window frame. Coordinates are returned in
 * relation to the upper left corner of the parent window.
 * The visible are of the frame is bounded by the given rectangle.
 *
 * \param frameArea Reference to the destination frame coordinate rectangle
 */
void GuiFrame::getGeometry(s_rect_t &frameArea)
{
	frameArea = this->frameArea;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get virtual frame size
 *
 * Retrieves the virtual size of the given frame. Virtual size can be larger
 * than the actual frame geometry, this may be used to implement scrollable
 * interface elements.
 *
 * \param pW Pointer to an integer that receives the width of the virtual frame.
 * If NULL value is specified, the width is not returned.
 * \param pH Pointer to an integer that receives the height of the virtual frame.
 * If NULL value is specified, the height is not returned.
 */
void GuiFrame::getVirtSize(int *pW, int *pH)
{
	if(pW) *pW = virtWidth;
	if(pH) *pH = virtHeight;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set virtual frame size
 *
 * Updates virtual size of the frame.
 *
 * \note The frame is not redrawn automatically.
 *
 * \param w New width of the virtual frame. If negative, then ignored.
 * \param h New height of the virtual frame. If negative, then ignored.
 */
void GuiFrame::setVirtSize(int w, int h)
{
	if (w >= 0) virtWidth = w;
	if (h >= 0) virtHeight = h;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set visibility flag
 *
 * Sets the frame visibility flag.
 *
 * \param flag Specifies if the frame is displayed or hidden. Hidden frames
 * ignore all painting requests.
 */
void GuiFrame::setVisible(bool flag)
{
	isVisible = flag;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get frame size
 *
 * Retrieves the actual size of the given frame that is visible in the window.
 *
 * \param pW Pointer to an integer that receives the width of the frame.
 * If NULL value is specified, the width is not returned.
 * \param pH Pointer to an integer that receives the height of the frame.
 * If NULL value is specified, the height is not returned.
 */
void GuiFrame::getSize(int *pW, int *pH)
{
	if(pW) *pW = frameArea.w;
	if(pH) *pH = frameArea.h;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Redraw the frame
 *
 * Performs the full frame redrawing.
 *
 * \return Returns true if the redraw operation is successful.
 */
bool GuiFrame::redrawFrame()
{
	s_rect_t redrawRect = { -1, -1, -1, -1 };
	return redrawFrame(redrawRect);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Redraw the frame partially
 *
 * Performs a partial redrawing of the given frame limited by the rectangular
 * area. Redraw request is passed to the parent window that in turn executes
 * painting procedures for all frame in the given area.
 *
 * \param updateArea The rectangular area of the frame that should be updated
 *
 * \return Returns true if the redraw operation is successful.
 */
bool GuiFrame::redrawFrame(s_rect_t updateArea)
{
	if (updateArea.w <= 0 || updateArea.h <= 0) {
		updateArea.x = offsetX;
		updateArea.y = offsetY;
		updateArea.w = frameArea.w;
		updateArea.h = frameArea.h;
	}

	// offset the frame position and calculate absolute coords
	updateArea.x = updateArea.x - offsetX + frameArea.x;
	updateArea.y = updateArea.y - offsetY + frameArea.y;

	s_rect_t redrawRect;
	// calculate the intersection with the frame view area, abort if no drawing needed
	if(s_rect_intersect(&updateArea, &frameArea, &redrawRect))
		return false;

	updateRequired = true;	// set the update flag now

	// call the window redraw function
	parentWindow->redrawWindow(redrawRect);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the frame interface
 *
 * Paint method is called by the parent window when frame redraw is required
 * in the given area.
 *
 * \param pSurface Pointer to a graphical surface where the frame should
 * draw itself to.
 * \param paintArea The rectangular area of the surface that should be
 * re-painted by the frame.
 *
 * \note Coordinates of the painting area are relative to the upper left corner of the provided surface.
 */
void GuiFrame::onPaint(GuiSurface *pSurface, s_rect_t &paintArea)
{
	if(!isVisible) return;

	// calculate the update rectangle, abort if no intersection
	s_rect_t absRect;
	if(s_rect_intersect(&paintArea, &frameArea, &absRect))
		return;

	// calculate relative coordinates in the target pSurface
	s_rect_t relRect = { absRect.x, absRect.y, absRect.w, absRect.h };

	list<GuiPrimitive*>::iterator itItem;

	// loop through all primitives in the list, painting them onto the virtual surface
	for(itItem = primitiveList.begin(); itItem != primitiveList.end(); itItem ++)
	{
		#ifdef DEBUG
		clock_t start = clock();
		#endif

		GuiPrimitive* pItem = (GuiPrimitive*)*itItem;

		// calculate real item position from its "virtual" rectangle
		s_rect_t oldPos = pItem->drawArea;
		s_rect_t newPos = {	oldPos.x - offsetX + frameArea.x,
							oldPos.y - offsetY + frameArea.y,
							oldPos.w, oldPos.h };

		// set temporary item coordinates relative to pSurface
		pItem->drawArea = newPos;

		pItem->onPaint(pSurface, relRect);
		pItem->drawArea = oldPos;

		#ifdef DEBUG
		printf("GuiFrame::onPaint: drawn resource %d (%d, %d, %d, %d) in %d ms\n",
			pItem->getResId(), newPos.x, newPos.y, newPos.w, newPos.h,
			(int)((clock() - start) / (CLOCKS_PER_SEC / 1000)));
		#endif
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize frame surface
 *
 * GuiFrame implementation does not provide a buffered frame surface. Given
 * method initializes all available GUI primitives contained by the frame.
 *
 * \param virtWidth Virtual width of the frame
 * \param virtHeight Virtual height of the frame
 *
 * \return Returns true if all components were initialized successfully.
 */
bool GuiFrame::initSurface(int virtWidth, int virtHeight)
{
	if (virtWidth <=0 || virtHeight <=0) return false;
	this->virtWidth = virtWidth;
	this->virtHeight = virtHeight;

	updateRequired = true;

	// initialize all primitives in the frame
	list<GuiPrimitive*>::iterator it;
	for(it = primitiveList.begin(); it != primitiveList.end(); it ++)
		((GuiPrimitive*)*it)->onInitialize();

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get a touchable primitive for given coordinates
 *
 * Retrieves a GUI primitive that has a touch region covering the given point.
 * Each primitive contained by frame may have a non-zero touch region. This
 * area is used for implementing of the touchscreen user interface.
 *
 * \param x X coordinate of the touch point
 * \param y Y coordinate of the touch point
 * \return Returns a primitive object ID, or RES_ID_INVALID if no touchable
 * primitive exists at that location.
 */
int GuiFrame::getPrimitiveTouch(int x, int y)
{
	if (!isVisible) return RES_ID_INVALID;

	// offset given coordinates with current frame position
	x -= (frameArea.x - offsetX);
	y -= (frameArea.y - offsetY);

	list<GuiPrimitive*>::reverse_iterator itItem;
	// reverse loop through all primitives (last ones are on top)
	for(itItem = primitiveList.rbegin(); itItem != primitiveList.rend(); itItem ++)
	{
		GuiPrimitive* pItem = (GuiPrimitive*)*itItem;
		// ignore invisible elements
		if (!pItem->getVisible()) continue;

		s_rect_t itemArea;
		pItem->getTouchArea(itemArea);
		// test if the point is within the rect
		if((x >= itemArea.x) && (x < (itemArea.x + itemArea.w)) &&
			(y >= itemArea.y) && (y < (itemArea.y + itemArea.h)))
		{
			// found a valid element, return it
			return pItem->getResId();
		}
	}

	return RES_ID_INVALID;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch event handler
 *
 * Called from GuiWindow on touch event if the touched primitive belongs to
 * this frame. The primitive will be notified of the selection change.
 *
 * \param item touched item
 * \param eventType event type received
 * \param x touch x coordinate
 * \param y touch y coordinate
 * \param onWin true if the touch is within the window area
 * \return Returns bitmask of notification event flags which events should be posted
 */
int GuiFrame::touch(TouchItem item, int eventType, int x, int y, bool onWin)
{
	int ret = NOTIFY_NONE;

	x -= frameArea.x;
	y -= frameArea.y;
	bool onFrame = onWin && x >= 0 && x < frameArea.w &&
				y >= 0 && y < frameArea.h;
	x += offsetX;
	y += offsetY;

	if (!isVisible) {
		untouch(item);
		return ret;
	}

	GuiPrimitive *pPrim = getPrimitive(item.idPrimitive);
	if (!pPrim) return ret;

	s_rect_t itemArea;
	pPrim->getTouchArea(itemArea);

	bool onPrim = onFrame &&
		((x >= itemArea.x) && (x < (itemArea.x + itemArea.w)) &&
		(y >= itemArea.y) && (y < (itemArea.y + itemArea.h)));

	GuiPrimitive::SELECTION sel = GuiPrimitive::UNKNOWN;
	if (onPrim) {
		if (eventType & MOUSE_RELEASED) {
			sel = GuiPrimitive::SELECTED;
			ret |= NOTIFY_TOUCH;
			setFocusedItem(item.idPrimitive);
		} else
			sel = GuiPrimitive::PRESSED;
	}

	GuiPrimitive::SELECTION lastSel = pPrim->getSelection();
	pPrim->setSelection(sel, x, y);

	if (sel == GuiPrimitive::PRESSED && lastSel != GuiPrimitive::PRESSED)
		ret |= NOTIFY_PRESS;
	if (sel != GuiPrimitive::PRESSED && lastSel == GuiPrimitive::PRESSED)
		ret |= NOTIFY_RELEASE;

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch event cleanup handler
 *
 * Called for cleanup to avoid leaving some primitives to PRESSED state
 *
 * \param item touched item
 * \return Returns bitmask of notification event flags which events should be posted
 */
int GuiFrame::untouch(TouchItem item)
{
	int ret = NOTIFY_NONE;

	GuiPrimitive *pPrim = getPrimitive(item.idPrimitive);
	if (!pPrim) return ret;

	GuiPrimitive::SELECTION lastSel = pPrim->getSelection();
	pPrim->setSelection(GuiPrimitive::UNKNOWN, 0, 0);

	if (lastSel == GuiPrimitive::PRESSED) ret |= NOTIFY_RELEASE;
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set focused primitive
 *
 * Updates the identifier of the focused primitive of the given frame.
 * Also updates the selection state of the previously selected primitive.
 *
 * \param idPrimitive Unique identifier of the primitive that is now in focus.
 *
 * \note Don't forget to set selection status of the new focused primitive to GuiPrimitive::SELECTED
 * in the caller function, otherwise unselecting gets filtered out!
 */
void GuiFrame::setFocusedItem(int idPrimitive)
{
	// remove focus from a previously selected item in the same frame
	if(focusedItem != RES_ID_INVALID && focusedItem != idPrimitive) {
		GuiPrimitive *focusedPrimitive = getPrimitive(focusedItem);
		if (focusedPrimitive)
			focusedPrimitive->setSelection(GuiPrimitive::UNSELECTED,0,0);
	}

	// set new focus
	// NB! don't forget to set selection status of the new focused primitive
	// to GuiPrimitive::SELECTED in the caller function, otherwise
	// unselecting gets filtered out!
	focusedItem = idPrimitive;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor for the GUI window frame object.
 *
 * \param idResource Unique identifier of the primitive.
 * \param pParent Pointer to the parent window the frame should draw itself into
 */
GuiBufferedFrame::GuiBufferedFrame(int idResource, GuiWindow *pParent)
				: GuiFrame(idResource, pParent)
{
	virtSurface = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destructor of the buffered frame. Also destroys the virtual painting surface.
 */
GuiBufferedFrame::~GuiBufferedFrame()
{
	if(virtSurface) delete virtSurface;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize frame surface
 *
 * Initializes virtual frame surface and paints all contained GUI primitives
 * to that surface. GUI elements are also initialized.
 *
 * \param virtWidth Virtual width of the surface.
 * \param virtHeight Virtual height of the surface.
 *
 * \return Returns true if all primitives are initialized and painted
 * successfully.
 */
bool GuiBufferedFrame::initSurface(int virtWidth, int virtHeight)
{
	GuiSurface *parentSurface = parentWindow->getSurface();
	virtSurface = new GuiSurface(virtWidth, virtHeight,
									parentSurface->getBitsPerPixel());

	s_rect_t rc = {0, 0, virtWidth, virtHeight};

	// fill with blackness
	virtSurface->fillRect(&rc, 0);

	// call parent initialization
	GuiFrame::initSurface(virtWidth, virtHeight);

	// perform an inital painting of all frame primitives
	list<GuiPrimitive*>::iterator itItem;
	s_rect_t virtRect = { 0, 0, frameArea.w, frameArea.h };
	for(itItem = primitiveList.begin(); itItem != primitiveList.end(); itItem ++)
	{
		GuiPrimitive* pItem = (GuiPrimitive*)*itItem;
		pItem->onPaint(virtSurface, virtRect);
	}

	return virtSurface != NULL;
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the frame interface
 *
 * Paint method is called by the parent window when frame redraw is required
 * in the given area.
 *
 * \param pSurface Pointer to a graphical surface where the frame should
 * draw itself to.
 * \param paintArea The rectangular area of the surface that should be
 * re-painted by the frame.
 *
 * \note Coordinates of the painting area are relative to the upper left corner of the provided surface.
 */
void GuiBufferedFrame::onPaint(GuiSurface *pSurface, s_rect_t &paintArea)
{
	if(!isVisible) return;

	// calculate the update rectangle, abort if no intersection
	s_rect_t absRect;
	if(s_rect_intersect(&paintArea, &frameArea, &absRect)) return;

	// calculate relative coordinates in the target pSurface
	s_rect_t relRect = { absRect.x - paintArea.x, absRect.y - paintArea.y,
						absRect.w, absRect.h };

	// calculate relative coordinates in the virtual frame surface
	s_rect_t virtRect = {	absRect.x - frameArea.x + offsetX,
							absRect.y - frameArea.y + offsetY,
							absRect.w, absRect.h };

	if(updateRequired)
	{
		// copy the backround from the original paint surface
		virtSurface->copySurfacePart(offsetX, offsetY, &relRect, pSurface);

		list<GuiPrimitive*>::iterator itItem;
		// loop through all primitives in the list, painting them onto the virtual surface
		for(itItem = primitiveList.begin(); itItem != primitiveList.end(); itItem ++)
		{
			#ifdef DEBUG
			clock_t start = clock();
			#endif

			GuiPrimitive* pItem = (GuiPrimitive*)*itItem;
			pItem->onPaint(virtSurface, virtRect);

			#ifdef DEBUG
			printf("GuiBufferedFrame::onPaint: drawn resource %d (%d, %d, %d, %d) in %d ms\n",
				pItem->getResId(), virtRect.x, virtRect.y, virtRect.w, virtRect.h,
				(int)((clock() - start) / (CLOCKS_PER_SEC / 1000)));
			#endif
		}

		updateRequired = false;
	}

	// copy the virtual frame surface back onto the original surface
	pSurface->copySurfacePart(relRect.x, relRect.y, &virtRect, virtSurface);
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Return number of primitives
 *
 * \return Returns the number of primitives the frame contains
 */
int GuiFrame::getNumPrimitives()
{
	return primitiveList.size();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Return the parent window handle
 *
 * \return Returns the pointer to the parent window that contains the frame
 */
GuiWindow * GuiFrame::getParentWindow()
{
	return parentWindow;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Return the unique ID of the frame
 *
 * \return Returns the unique identifier of the frame
 */
int GuiFrame::getResId()
{
	return idResource;
}
