#ifndef GUIFRAME_H_
#define GUIFRAME_H_

#include "GuiSurface.h"
#include "GuiPrimitive.h"

#include <list>
using std::list;

class GuiWindow;
class GuiPrimitive;

/**
 * \brief Window frame implementation
 *
 * Implements a container for GUI primitives referred to as the
 * frame. Any window may have a number of frames which may overlap each other.
 * The frame may contain a number of primitives, each implementing some
 * visualization function. The idea of the frame is:
 *
 * - to provide a virtual drawing space, simplifying implementation of
 *   scrollable elements
 * - to group elements logically with stand-alone selection settings,
 *   e.g. a group of radio boxes
 * - to provide a container for graphical primitives
 *
 * GuiFrame does not have a virtual painting surface used for buffered
 * drawing. Please refer to GuiBufferedFrame if your application requires
 * a buffered frame.
 *
 * XML code below defines a window frame:
 *
 * \code
 *	<frame id="ID" x="X" y="Y" width="W" height="H" buf="V" vx="VX" vy="VY">
 * 		<!-- primitives are defined here -->
 * 	</frame>
 * \endcode
 *
 * Where
 * - ID represents the unique identifier of the frame
 * - X, Y, W, H is the set of frame coordinates in the parent window
 * - V is the buffering flag (optional). When set to "y", enables the
 * - surface buffering. Refer to GuiBufferedFrame for more details.
 * - vx is the virtual width of the frame (optional)
 * - vy is the virtual height of the frame (optional)
 *
 * Virtual dimensions of the frame should be only specified for scrollable
 * frames.
 *
 * \author Andrei Birjukov, Anti Sullin
 */

class GuiFrame
{
friend class GuiWindow;

public:
	GuiFrame(int idResource, GuiWindow *pParent);
	virtual ~GuiFrame();

	int addPrimitive(GuiPrimitive *pItem);
	GuiPrimitive* getPrimitive(int idResource);
	bool removePrimitive(int idResource);
	bool removePrimitives(int minId, int maxId);
	int getNumPrimitives();
	void setFocusedItem(int idPrimitive);

	int getPrimitiveTouch(int x, int y);


	/** Types of notification events **/
	enum NOTIFY {
		/** No notification **/
		NOTIFY_NONE = 0,

		/** Notify touch event (short press and release) **/
		NOTIFY_TOUCH = 0x01,

		/** Notify press event (i.e. when dragging initiated) **/
		NOTIFY_PRESS = 0x02,

		/** Notify release event (i.e. when dragging is finished) **/
		NOTIFY_RELEASE = 0x04,
	};
	int touch(TouchItem item, int eventType, int x, int y, bool onWin);
	int untouch(TouchItem item);

	void setOffset(int x, int y);
	void getOffset(int *pX, int *pY);
	void setGeometry(s_rect_t &frameArea);
	void getGeometry(s_rect_t &frameArea);
	void getVirtSize(int *pX, int *pY);
	void getSize(int *pX, int *pY);
	void setVirtSize(int w, int h);
	void setVisible(bool flag);
	GuiWindow *getParentWindow();
	bool redrawFrame();
	bool redrawFrame(s_rect_t updateArea);
	int getResId();

	virtual bool initSurface(int virtWidth, int virtHeight);

protected:

	virtual void onPaint(GuiSurface *pSurface, s_rect_t &paintArea);

	/** Coordinates and size of the frame area **/
	s_rect_t frameArea;

	/** Unique ID of the frame **/
	int idResource;

	/** Virtual width of the frame **/
	int virtWidth;

	/** Virtual height of the frame **/
	int virtHeight;

	/** Viewpoint offset Y coordinate of the frame **/
	int offsetX;

	/** Viewpoint offset X coordinate of the frame **/
	int offsetY;

	/** List of primitives the frame contains **/
	list<GuiPrimitive*> primitiveList;

	/** Unique ID of the focused item **/
	int focusedItem;

	/** Pointer to the parent window that holds the frame **/
	GuiWindow* parentWindow;

	/** Set to true if frame needs updating, false otherwise **/
	bool updateRequired;

	/** Set to true if frame should be visible, false otherwise **/
	bool isVisible;
};

/**
 * \brief Window buffered frame implementation
 *
 * Implements a buffered frame that provides a separate drawing
 * surface for faster scrolling. The buffered frame class is based on the
 * standard GuiFrame class and inherits most of its methods and attributes.
 *
 * GuiBufferedFrame implements its own painting function to take advantage
 * of the buffered painting mechanism.
 *
 * \author Andrei Birjukov, Anti Sullin
 */

class GuiBufferedFrame : public GuiFrame
{
public:
	GuiBufferedFrame(int idResource, GuiWindow *pParent);
	virtual ~GuiBufferedFrame();

	virtual bool initSurface(int virtWidth, int virtHeight);

protected:
	virtual void onPaint(GuiSurface *pSurface, s_rect_t &paintArea);

private:
	GuiSurface* virtSurface;
};

#endif /*GUIFRAME_H_*/
