/*
 * GuiLibrary.cpp
 *
 *  Created on: Mar 31, 2009
 *      Author: andrei
 */

/**
 * \mainpage
 *
 * Welcome to the \b eXynthoo documentation.
 *
 * eXynthoo is a GUI class library that exists as an extension to the eXynth GUI
 * engine. The library is designed for use by C++ applications on the client
 * side of the GUI engine and is mostly targeted to embedded devices with
 * simple static interfaces.
 *
 * The goals of the eXynthoo library are following:
 * - Implement a set of standard elements of the user interface (GUI
 * primitives)
 * - Provide a layer of isolation between the code of an application and
 * the design of a user interface
 * - Enable for a fully customizable GUI in terms of look and feel that supports
 * transparent primitives
 * - Provide building blocks for static and scrollable windows
 * - Implement an additional window sub-layering via window frames
 * - Implement interface for touchscreen and hardware button input
 * - Provide a convenient interface to timers
 *
 * \author Andrei Birjukov, Anti Sullin @ Artec Design LLC
 */

#include <string.h>
#include <exynth/exynth.h>
#include <exynth/font.h>
#include <exynth/event.h>

#include "GuiLibrary.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize GUI library
 *
 * Initializes the GUI library and starts event handler
 * thread. It must be initialized once at the start of the application.
 *
 * \return Returns true if the library has been initialized successfully
 */
bool GuiLibrary::initialize()
{
	return (s_exynth_init() == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Shut down GUI library
 *
 * Shuts down GUI library and its thread.
 *
 * \return Returns true if the library was shut down successfully
 */
bool GuiLibrary::shutdown()
{
	return (s_exynth_uninit() == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Lock GUI library mutex
 *
 * The GUI library mutex will be locked. This mutex is locked in the library
 * before handling any event, so by locking this mutex you'll guarantee the
 * thread-safety of GUI operations.
 *
 * Lock this mutex if you need to access any GUI function from another thread.
 * Avoid deadlocks as all event handling is done with this mutex locked!
 * (ie avoid using a separate mutex for Observer (listener) pattern because
 * when de-registering listener with GUI mutex locked and waiting for the
 * listener mutex the listener thread may be waiting for with listener mutex
 * taken for GUI mutex to broadcast something).
 *
 * \note This is the only system mutex that exists and protects the GUI thread.
 * When doing some work in the GUI thread, never call this method since it guarantees
 * a deadlock situation! Use this method only from main thread (i.e. program main() function)
 *
 * \return Returns true if mutex was locked
 */
bool GuiLibrary::lock()
{
	return (s_exynth_lock() == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Lock GUI library mutex or return when locking failed within a
 *  specified time
 *
 * \param ms Timeout in millseconds
 * \return Returns true if mutex was locked
 */
bool GuiLibrary::timedlock(int ms)
{
    struct timespec abs_timeout;
    struct timeval tv;
    gettimeofday(&tv, NULL);

    abs_timeout.tv_sec  = tv.tv_sec + ms/1000;
    abs_timeout.tv_nsec = tv.tv_usec*1000 + (ms % 1000)*1000*1000;
    if (abs_timeout.tv_nsec >= 1000*1000*1000) {
    	abs_timeout.tv_nsec -= 1000*1000*1000;
    	abs_timeout.tv_sec++;
    }

	return (s_exynth_timedlock(&abs_timeout) == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unlock GUI library mutex
 *
 * See GuiLibrary::lock() for description.
 *
 * \return Returns true if mutex was unlocked
 */
bool GuiLibrary::unlock()
{
	return (s_exynth_unlock() == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Broadcast a system event to all windows
 *
 * Send a application-defined system event to all windows.
 * Use in platform library to send events.
 *
 * \param code Code of the event being broadcasted
 * \param data Event description
 */
void GuiLibrary::broadcastSystemEvent(int code, void* data)
{
	s_event_t event;
	event.type = SYSTEM_EVENT;
	event.data.user.code = code;
	event.data.user.data = data;
	s_window_broadcast_event(&event);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Broadcast a system event to all windows
 *
 * Send a application-defined user event to all windows.
 * Use in application code to send events to windows.
 *
 * \param code Code of the event being broadcasted
 * \param data Event description
 */
void GuiLibrary::broadcastUserEvent(int code, void* data)
{
	s_event_t event;
	event.type = USER_EVENT;
	event.data.user.code = code;
	event.data.user.data = data;
	s_window_broadcast_event(&event);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Count UTF-8 characters in string
 *
 * Counts the number of UTF-8 characters in a string.
 * Invalid characters are skipped.
 *
 * \param text Pointer to string whose characters should be calculated
 *
 * \return Returns the number of valid UTF-8 characters
 */
size_t GuiLibrary::utf8strlen(const char *text)
{
	const char *p = text;
	size_t len = strlen(text);
	int ucnt = 0;

	while (len > 0)
	{
		// get next char unicode number
		int cnt = s_font_utf8_to_unicode(NULL, p, len);
		if (cnt <= 0) {
			len -= 1; p += 1;
			continue;
		}
		len -= cnt;
		p += cnt;
		ucnt++;
	}
	return ucnt;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Return byte offset of n-th UTF-8 character in string
 *
 * Locates the n-th character in UTF-8 string
 * and returns its byte offset from string start.
 * If n is larger than number of characters in the string,
 * offset to string terminating '\\0' byte will be returned.
 *
 * \param text Pointer to string where the character should be located
 * \param n Character position that should be returned
 *
 * \return Returns n-th UTF-8 character byte offset in string
 */
 size_t GuiLibrary::utf8pos2pos(const char *text, size_t n)
 {
		const char *p = text;
		size_t len = strlen(text);
		size_t ucnt = 0;

		while (len > 0 && ucnt < n)
		{
			// get next char unicode number
			int cnt = s_font_utf8_to_unicode(NULL, p, len);
			if (cnt <= 0) {
				len -= 1; p += 1;
				continue;
			}
			len -= cnt;
			p += cnt;
			ucnt++;
		}
		return p - text;
 }
