/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "GuiMultiTextBox.h"

//#define DEBUG

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default constructor
 *
 * Default constructor for the primitive. Sets the text content, font size and
 * font color to null values.
 *
 * Also sets the rendered flag to false
 *
 * \param idResource Unique identificator of the primitive.
 * \param pFrame Pointer to the frame where the primitive should be placed into
 */
GuiMultiTextBox::GuiMultiTextBox(int idResource, GuiFrame* pFrame)
					: GuiPrimitive(idResource, pFrame)
{
	lineList = NULL;
	isRendered = false;
	userText = NULL;

	fontSize = 0;
	fontColor.red = fontColor.green = fontColor.blue = 0;

	horzAlign = LEFT;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default destructor of the object
 *
 * Uninitializes the font data
 *
 */
GuiMultiTextBox::~GuiMultiTextBox()
{
	while (!fontData.empty()) {
		s_font_t* font = fontData.back();
		fontData.pop_back();
		if (font) s_font_uninit(font);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the text box font style
 *
 * Sets the font style (type and size) of the text box
 * \param fontName Pointer to the font name that should be used
 * \param fontSize Size of the font that should be used
 *
 * \note This method sets the rendered flag to false
 */
void GuiMultiTextBox::setFont(const char* fontName, int fontSize)
{
	this->fontName = fontName;
	this->fontSize = fontSize;
	isRendered = false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the text box font color
 *
 * Sets the font color of the text box
 * \param pColor Pointer to the color string that should be used
 * \note This method sets the rendered flag to false
 */
void GuiMultiTextBox::setColor(const char *pColor)
{
	fontColor = GuiSurface::stringToRGB(pColor);
	isRendered = false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the text box content
 *
 * Fill the text box width the data provided
 * \param boxText Pointer to text that should be shown
 *
 * \note This method sets the rendered flag to false and does not cause the text box to be redrawn!
 */
void GuiMultiTextBox::setText(const char* boxText)
{
	if (!drawArea.w) return;
	userText = boxText; // only for string position calculations, not dereferenced

	wrapper.setWidth(drawArea.w-1);
	wrapper.setFont(fontName.c_str(), fontSize);
	wrapper.setText(userText);
	lineList = wrapper.getLineList();
	lineHeight = wrapper.getLineHeight();
	textHeight = wrapper.getTotalHeight();

	isRendered = false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Renders the text in the text box
 *
 * Causes the text box to be rendered using the specified font style
 * \note This method does not repaint the primitive, it simply renderes the text accordingly
 */
void GuiMultiTextBox::renderText()
{
	// clean up previous contents
	while (!fontData.empty()) {
		s_font_t* font = fontData.back();
		fontData.pop_back();
		if (font) s_font_uninit(font);
	}

	if (!lineList) return;

	// render the lines
	for(std::list<textLine>::iterator it = lineList->begin();
			it != lineList->end(); it++) {
		std::string	textLine(it->start, it->len);

		s_font_t* font;
		if(s_font_init(&font, fontName.c_str())) return;
		s_font_set_size(font, fontSize);
		s_font_set_rgb(font, fontColor.red, fontColor.green, fontColor.blue);

		s_font_set_str(font, textLine.c_str());
		s_font_get_glyph(font);

		fontData.push_back(font);
	}

	isRendered = true;
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the primitive interface
 *
 * Called by the parent frame whenever a primitive redraw is required.
 * \param drawSurface Pointer to a graphical surface where the primitive should
 * draw itself to.
 * \param paintArea The rectangular area of the parent frame surface that
 * should be re-painted by the primitive.
 *
 * \note If the text is not rendered yet, this method automatically renders it
 */
void GuiMultiTextBox::onPaint(GuiSurface *drawSurface, s_rect_t &paintArea)
{
	// return immediately if the primitive is not visible
	if(!isVisible) return;

	// check if the multi-line text is rendered
	if(!isRendered) renderText();

	// sanity check, do we have any data to draw?
	if(!isRendered || fontData.empty()) return;

	// clip the paintArea to control area
	s_rect_t repaintRect;
	if(s_rect_intersect(&drawArea, &paintArea, &repaintRect)) return;

	// calculate vertical placement of the whole block
	int oy = 0;
	switch(vertAlign)
	{
	case TOP:
		oy = 0;
		break;
	case BOTTOM:
		oy = drawArea.h - textHeight;
		break;
	case VCENTER:
		oy = (drawArea.h - textHeight) / 2;
		break;
	}

	// draw the text
	for(std::list<s_font_t*>::iterator it = fontData.begin();
			it !=fontData.end(); it ++)
	{
		s_font_t* font = *it;
		if(!font) {
			oy += lineHeight;
			continue;
		}
		// baseline alignment compensation within lines
		int oy2 = font->ascent - font->glyph.yMax + 1;

		int ox = 0;
		switch(horzAlign)
		{
		case LEFT:
			ox = 0;
			break;
		case RIGHT:
			ox = drawArea.w - font->glyph.img->w;
			break;
		case HCENTER:
			ox = (drawArea.w - font->glyph.img->w) / 2;
			break;
		}

		s_rect_t textArea = { drawArea.x + ox, drawArea.y + oy + oy2,
				drawArea.w - ox, font->height - oy2};


		s_rect_t surfRect;
		if(s_rect_intersect(&repaintRect, &textArea, &surfRect)) {
			oy += font->lineskip;
			continue;
		}

		s_rect_t sourceRect = { surfRect.x - textArea.x, surfRect.y - textArea.y,
							surfRect.w, surfRect.h };

		#ifdef DEBUG
		printf("GuiMultiTextBox::onPaint: line, %d, %d to %d, %d\n",
				font->glyph.img->w, font->glyph.img->h,
				surfRect.x, surfRect.y);
		#endif

		drawSurface->drawImagePart(surfRect.x, surfRect.y,
									&sourceRect, font->glyph.img);

		oy += lineHeight;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the text alignment in the text box
 *
 * Sets both the horizontal and vertical alignment of the text in the text box
 * \param horz Horizontal alignment type
 * \param vert Vertical aligment type
 */
void GuiMultiTextBox::setAlignment(TextHAlignment horz, TextVAlignment vert)
{
	horzAlign = horz;
	vertAlign = vert;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get the required text box size
 *
 * Gets the required size of the text box to render the text
 * \param rect Memory location of the structure where to store the calculated data
 */
void GuiMultiTextBox::getRequiredSize(s_rect_t &rect)
{
	rect.x = rect.y = 0;
	rect.w = drawArea.w;
	rect.h = textHeight;
}
