/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <exynth/exynth.h>
#include <memory.h>
#include <string.h>

#include "GuiPrimitive.h"
#include "GuiFrame.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Implements default constructor for the GUI primitive object.
 * Initializes selection state to UNSELECTED and primitive coordinates to zeros.
 * The primitive is visible by default.
 *
 * \param idResource Unique primitive identifier.
 * \param pFrame Pointer to a parent frame containing the primitive.
 */
GuiPrimitive::GuiPrimitive(int idResource, GuiFrame *pFrame)
{
	this->idResource = idResource;
	selection = UNSELECTED;
	lastselection = UNSELECTED;

	drawArea.x = drawArea.y = 0;
	drawArea.w = drawArea.h = 0;

	touchArea.x = touchArea.y = 0;
	touchArea.w = touchArea.h = 0;

	pHolder = pFrame;
	isVisible = true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Implements a destructor for the GUI primitive object.
 */
GuiPrimitive::~GuiPrimitive()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set touch area coordinates
 *
 * Sets coordinates of the touchable area of the given primitive. Coordinates
 * should be set relatively to the upper left corner of the parent frame that
 * contains the primitive. The frame uses the primitive touch area to implement
 * the touchscreen user interface.
 *
 * \param touchArea Rectangle structure containing coordinates of the
 * touchable area.
 */
void GuiPrimitive::setTouchArea(s_rect_t &touchArea)
{
	if (touchArea.w < 0 || touchArea.h < 0) return;
	this->touchArea = touchArea;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get touch area coordinates
 *
 * Retrieves coordinates of the touchable area of the given primitive. The
 * returned coordinates are set relatively to the parent frame that contains
 * the primitive.
 *
 * \param touchArea Reference to the rectangle structure that receives
 * coordinates of the touchable area.
 */
void GuiPrimitive::getTouchArea(s_rect_t &touchArea)
{
	touchArea = this->touchArea;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set primitive position
 *
 * Updates coordinates of the primitive in the parent frame, in relation to the
 * upper left corner of the frame. If the frame has a virtual dimension greater
 * than the displayed one, the primitive coordinates are still set relatively
 * to the frame upper left corner, regardless of the current viewpoint offset.
 *
 * \note The position change does not trigger automatic primitive redrawing,
 * so user must call the redraw() method separately.
 *
 * \param drawArea Rectangle structure containing coordinates of the new
 * position.
 */
void GuiPrimitive::setPosition(s_rect_t &drawArea)
{
	if (drawArea.w < 0 || drawArea.h < 0) return;
	this->drawArea = drawArea;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get primitive position
 *
 * Retrieves coordinates of the primitive in the parent frame, in relation to
 * the upper left corner of the frame.
 *
 * \param drawArea Reference to the rectangle structure that receives the
 * coordinates of the current position.
 */
void GuiPrimitive::getPosition(s_rect_t &drawArea)
{
	drawArea = this->drawArea;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set selection
 *
 * Updates selection state of the primitive. The onSelectionChanged() virtual
 * function will be called when selection is updated notifying the primitive
 * implementation about the change.
 *
 * \note The selection change does not usually trigger automatic primitive
 * redrawing, so user must call the redraw() method separately.
 *
 * \param type Specifies the new selection state.
 * \param x Contains the X-coordinate of the touch point relative to the window
 * \param y Contains the Y-coordinate of the touch point relative to the window
 */
void GuiPrimitive::setSelection(SELECTION type, int x, int y)
{
	#ifdef DEBUG
	printf("GuiPrimitive::setSelection: resId %d set to %d\n",
			idResource, type);
	#endif

	// is the selection changed?
	if(selection != type)
	{
		// backup the last selection in case the user wants to release the mouse outside the frame
		if (type == PRESSED && (selection == SELECTED || selection == UNSELECTED)) {
			lastselection = selection;
		}
		// backup the last selection in case the user wants to release the mouse outside the frame
		if (type == UNKNOWN)
			selection = lastselection;
		else
			selection = type;

		onSelectionChanged(type);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get selection
 *
 * Retrieves the current selection state of the primitive.
 * \return Returns the current selection state.
 */
GuiPrimitive::SELECTION GuiPrimitive::getSelection()
{
	SELECTION sel = selection;

	return sel;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Redraw primitive fully
 *
 * Performs a full primitive redrawing. Parent frame is requested to redraw the
 * current primitive location area.
 */
void GuiPrimitive::redraw()
{
	// request the frame redraw at the given primitive area
	pHolder->redrawFrame(drawArea);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set visibility
 *
 * Updates the primitive visibility.
 *
 * \note The visibility change does not trigger automatic primitive redrawing,
 * so user must call the redraw() method separately.
 *
 * \param flag Specifies if the primitive is visible or not. When set to false,
 * the primitive is not visible and implementation should ignore all repainting requests.
 */
void GuiPrimitive::setVisible(bool flag)
{
	isVisible = flag;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get visibility
 *
 * Retrieve current visibility state of the given primitive
 * \return Returns true if the primitive is visible
 */
bool GuiPrimitive::getVisible()
{
	bool ret = isVisible;

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Return unique ID of the primitive
 * \return Returns the primitive unique identification
 */
int GuiPrimitive::getResId()
{
	return idResource;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize function of the primitive interface
 *
 * This virtual function is called by the parent frame when the primitive is
 * being initialized. Specific implementation of the primitive may override
 * this function if such notification is required.
 */
void GuiPrimitive::onInitialize()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting function of the primitive interface
 *
 * This virtual function is called by the parent frame whenever a primitive
 * redraw is required. Specific implementation of the primitive must override
 * this function to perform appropriate painting.
 *
 * \param drawSurface Pointer to a graphical surface where the primitive should
 * draw itself to.
 *
 * \param paintArea The rectangular area of the parent frame surface that
 * should be re-painted by the primitive.
 */
void GuiPrimitive::onPaint(GuiSurface *drawSurface, s_rect_t &paintArea)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Selection change function of the primitive interface
 *
 * This virtual function is called by the parent container whenever the
 * selection has been changed. Selection change inside a frame is typically
 * a result of a user touchscreen manipulation. Specific implementation of the
 * primitive may override this function if such notification is required.
 *
 * \param type Specifies the new selection state of the given primitive
 */
void GuiPrimitive::onSelectionChanged(SELECTION type)
{
}
