/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "GuiSurface.h"

#include <exynth/grlib.h>

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor for new surface
 *
 * Constructor for the GUI primitive object that creates a new virtual surface
 * of the given dimension.
 *
 * \param width Specifies the width of the virtual surface
 * \param height Specifies the height of the virtual surface
 * \param bpp Specifies the number of bits per pixels for the virtual surface
 */
GuiSurface::GuiSurface(int width, int height, int bpp)
{
	pSurface = (s_surface_t*)malloc(sizeof(s_surface_t));
	virtBuf = (char*) calloc(width * height, ((bpp-1)/8+1));
	s_getsurfacevirtual(pSurface, width, height, bpp, virtBuf);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor for the existing surface
 *
 * Constructor for the GUI primitive object that attaches to an existing
 * eXynth surface. Virtual buffer of the surface is initialized to NULL.
 *
 * \param pSurface Pointer to the surface descriptor that object is being attached to
 */
GuiSurface::GuiSurface(s_surface_t *pSurface)
{
	this->pSurface = pSurface;
	this->virtBuf = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destructor of the GUI surface object. Also releases a virtual buffer if it
 * was created.
 */
GuiSurface::~GuiSurface()
{
	if(virtBuf != NULL)
	{
		free(virtBuf);
		free(pSurface);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get surface width
 *
 * Gets the width of the surface.
 *
 * \return Returns the width of the surface in pixels
 */
int GuiSurface::getWidth()
{
	return pSurface->width;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get surface height
 *
 * Gets the height of the surface.
 *
 * \return Returns the height of the surface in pixels
 */
int GuiSurface::getHeight()
{
	return pSurface->height;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get surface bits per pixel
 *
 * Gets the color depth of the surface
 *
 * \return Returns the number of bits per pixel of the surface
 */
int GuiSurface::getBitsPerPixel()
{
	return pSurface->bitsperpixel;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set pixel RGB
 *
 * Writes the given RGB coefficients to a certain pixel
 *
 * \param x X coordinate of the destination pixel
 * \param y Y coordinate of the destination pixel
 * \param rgb RGB coefficients to be written to the destination pixel
 */
void GuiSurface::setPixel(int x, int y, int rgb)
{
	s_setpixel(pSurface, x, y, rgb);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Draw a line
 *
 * Draws a line on the surface.
 *
 * \param x1 X coordinate of the starting point of the line
 * \param y1 Y coordinate of the starting point of the line
 * \param x2 X coordinate of the ending point of the line
 * \param y2 Y coordinate of the ending point of the line
 * \param rgb RGB coefficients of the line color
 */
void GuiSurface::drawLine(int x1, int y1, int x2, int y2, int rgb)
{
	s_line(pSurface, x1, y1, x2, y2, rgb);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Draw rectangle
 *
 * Drwas the rectangle with a specified color on the surface.
 *
 * \param pRect Coordinates to the rectangle that will be drawn. Note that
 * the coordinates are calculated from the top left corner.
 * \param rgb RGB value to the color the rectangle should be filled with
 *
 */
void GuiSurface::fillRect(s_rect_t *pRect, int rgb)
{
	s_fillbox(pSurface, pRect->x, pRect->y, pRect->w, pRect->h, rgb);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Draw image
 *
 * Draws a bitmap image into the specified position on the surface.
 *
 * \param x X coordinate of the upper left corner where the image should be placed
 * \param y Y coordinate of the upper left corner where the image should be placed
 * \param pBitmap Pointer to the bitmap image
 */
void GuiSurface::drawImage(int x, int y, s_image_t *pBitmap)
{
	s_putboxrgba(pSurface, x, y, pBitmap->w, pBitmap->h, pBitmap->rgba);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Draw part of the image
 *
 * Draws part of the bitmap image specified by rectangle into the destination position marked by
 * its upper left corner coordinates.
 *
 * \param destX X coordinate of the starting point of destination drawing
 * \param destY Y coordinate of the starting point of destination drawing
 * \param imgPart Rectangular area of the image that must be drawn
 * \param pBitmap Pointer to the bitmap image
 *
 */
void GuiSurface::drawImagePart(int destX, int destY,
								s_rect_t *imgPart, s_image_t *pBitmap)
{
	#ifdef DEBUG
	printf("GuiSurface::drawImagePart: dest (%d, %d, %d, %d) src (%d, %d, %d, %d)\n",
			destX, destY, pBitmap->w, pBitmap->h, imgPart->x, imgPart->y, imgPart->w, imgPart->h);
	#endif

	s_putboxpartrgba(pSurface, destX, destY,
					imgPart->w, imgPart->h,
					pBitmap->w, pBitmap->h,
					pBitmap->rgba, imgPart->x, imgPart->y);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Calculates the pixel value in the current color space
 *
 *\param r The red component of the desired color
 *\param g The green component of the desired color
 *\param b The blue component of the desired color
 *
 * \return Returns the integer value of the calculated pixel
 */
int GuiSurface::makeColor(int r, int g, int b)
{
	return s_rgbcolor(pSurface, r, g, b);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Calculates the pixel value in the current color space
 *
 * \overload int GuiSurface::makeColor(RGB_COLOR *rgb)
 *
 * \param rgb Pointer to the RGB color structure, where individual color values are stored
 *
 * \return Returns the integer value of the calculated pixel
 */
int GuiSurface::makeColor(RGB_COLOR *rgb)
{
	return s_rgbcolor(pSurface, rgb->red, rgb->green, rgb->blue);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Convert RGB color string into RGB color structure
 *
 * Converts the specified RGB color string into RGB color structure
 *
 * \param str Textual representation of RGB color
 *
 * \return Returns structure RGB_COLOR with the filled-in color components
 */
RGB_COLOR GuiSurface::stringToRGB(const char *str)
{
	unsigned int r=0, g=0, b=0;
	sscanf(str, "%02x%02x%02x", &r, &g, &b);

	RGB_COLOR ret = {r, g, b};
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Copies specified surface object buffer onto the current one
 *
 * Copies the specified surface object pSrc buffer onto the current surface into the
 * position specified by x and y coordinates
 *
 * \param x X coordinate of the destination point where to place the surface contents
 * \param y Y coordinate of the destination point where to place the surface contents
 * \param pSrc Pointer to the source surface object
 *
 */
void GuiSurface::copySurface(int x, int y, GuiSurface *pSrc)
{
	s_putbox(pSurface, x, y, pSrc->pSurface->width,
				pSrc->pSurface->height, pSrc->pSurface->vbuf);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Copies specified surface object buffer onto the current one
 *
 * Copies the specified surface object pSrc buffer onto the current surface starting from
 * the top-left corner position
 *
 * \param pSrc Pointer to the source surface object
 */
void GuiSurface::copySurface(GuiSurface *pSrc)
{
	copySurface(0, 0, pSrc);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Copies part of the specified surface object buffer onto the current one
 *
 * Copies part of the specified surface object pSrc buffer onto the current surface into the
 * position specified by x and y coordinates
 *
 * \param destX X coordinate of the destination point where to place the surface contents
 * \param destY Y coordinate of the destination point where to place the surface contents
 * \param srcPart Pointer to the rectangle specifies which part will be copied
 * \param pSrc Pointer to the source surface object
 */
void GuiSurface::copySurfacePart(int destX, int destY,
									s_rect_t *srcPart, GuiSurface *pSrc)
{
	s_putboxpart(pSurface, destX, destY,
				srcPart->w, srcPart->h,
				pSrc->pSurface->width, pSrc->pSurface->height,
				pSrc->pSurface->vbuf,
				srcPart->x, srcPart->y);
}
