/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *                        : (c) 2007-2008 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#include <sys/stat.h>

#include "GuiTemplate.h"
#include "GuiImage.h"
#include "GuiCheckBox.h"
#include "GuiDrawArea.h"
#include "GuiEditBox.h"
#include "GuiTextBox.h"
#include "GuiMultiTextBox.h"
#include "GuiTextMatrix.h"
#include "GuiButton.h"
#include "GuiRectangle.h"
#include "GuiAnimation.h"
#include "GuiProgressBar.h"
#include "GuiScrollBar.h"

#define MAX_PATH 128

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor for the GUI template object. DOM tree is empty at the
 * moment of creation.
 */
GuiTemplate::GuiTemplate()
{
	rootNode = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destructor for the GUI template objects.
 * Also releases the DOM tree and frees resources.
 */
GuiTemplate::~GuiTemplate()
{
	if(rootNode) s_xml_node_uninit(rootNode);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Load skin file
 *
 * Loads GUI window templates from the specified XML skin file. The DOM tree is
 * built and kept in memory until explicitly unloaded or the template object
 * is destroyed.
 *
 * \param fileName Specifies the path to the XML skin file
 * \return Returns true if the skin file is loaded successfully.
 */
bool GuiTemplate::loadFile(const char* fileName)
{
	char absFileName[MAX_PATH];

	if(rootNode) s_xml_node_uninit(rootNode);
	rootNode = NULL;

	// resolve PATH
	if (!lookupPath(fileName, absFileName, sizeof(absFileName))) return false;

	// load the skin file
	int ret = s_xml_parse_file(&rootNode, absFileName);
	if(ret != 0)
	{
		s_xml_node_uninit(rootNode);
		rootNode = NULL;
		return false;
	}

	// check the first XML node
	if(strcmp(rootNode->name, "guiskin"))
	{
		s_xml_node_uninit(rootNode);
		rootNode = NULL;
		return false;
	}

	// check the skin version
	char *ver = s_xml_node_get_attr_value(rootNode, "version");
	if(strcmp(ver, "1.0"))
	{
		s_xml_node_uninit(rootNode);
		rootNode = NULL;
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unload skin
 *
 * Releases DOM tree and frees resources.
 */
void GuiTemplate::unload()
{
	s_xml_node_uninit(rootNode);
	rootNode = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set platform identifier
 *
 * Sets the platform identifier. This can be used in all XML tags to filter
 * out some XML content if platform="p" attribute does not match value
 * set with setPlatform() function. This way it is possible to preserve a unified
 * skin file across multiple platforms.
 *
 * \param plat Platform identifier
 */
void GuiTemplate::setPlatform(std::string plat)
{
	platform = plat;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get root resources from DOM tree
 *
 * Root resources are not tied to any window. A GuiResource object can be
 * created and passed to getRootResource() to parse out the root resources.
 *
 * \param pRes Pointer to a resource object for which resources are loaded
 * \return Returns the number of resources loaded.
 */
int GuiTemplate::getRootResource(GuiResource *pRes)
{
	if(rootNode == NULL) return 0;
	return initResources(rootNode, pRes);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize window components
 *
 * Given method builds the user interface for the specified window from an
 * XML template skin file. The window is looked up from
 * the loaded DOM tree based on its unique identifier. Frames and primitives
 * are automatically instantiated and inserted into the given window.
 *
 * \param pWnd Pointer to a window object for which GUI elements are loaded
 * \return Returns true if the window interface is built successfully.
 */
bool GuiTemplate::initWindowLayout(GuiWindow *pWnd)
{
	if(rootNode == NULL) return false;
	s_list_node_t *listNode = rootNode->nodes.node;

	// find the right window first
	for(int nodeItem = 0; nodeItem < rootNode->nodes.nb_elt; nodeItem ++, listNode = (s_list_node_t*)listNode->next)
	{
		s_xml_node_t *xmlNode = (s_xml_node_t*)listNode->element;

		// what we got here is a window
		if(strcmp(xmlNode->name, "window")) continue;

		char* pfm = s_xml_node_get_attr_value(xmlNode, "platform");
		if (pfm && strcmp(pfm, platform.c_str())) continue;

		// grab window ID
		char* resId = s_xml_node_get_attr_value(xmlNode, "id");
		int iResId = (resId != NULL) ? atoi(resId) : -1;

		// found a window with the given resource ID
		if(pWnd->getResId() != iResId) continue;

		// grab window properties
		char* wndX = s_xml_node_get_attr_value(xmlNode, "x");
		char* wndY = s_xml_node_get_attr_value(xmlNode, "y");
		char* wndWidth = s_xml_node_get_attr_value(xmlNode, "width");
		char* wndHeight = s_xml_node_get_attr_value(xmlNode, "height");
		char* wndOnTop = s_xml_node_get_attr_value(xmlNode, "ontop");

		// validate properties
		if(wndX == NULL || wndY == NULL ||
			wndWidth == NULL || wndHeight == NULL) return -1;

		// set window properties
		s_rect_t wndPos = { atoi(wndX), atoi(wndY), atoi(wndWidth), atoi(wndHeight) };
		pWnd->setWindowPos(wndPos);
		if(wndOnTop) pWnd->setAlwaysOnTop(atoi(wndOnTop));

		// init window resources
		GuiResource *pRes = dynamic_cast<GuiResource*>(pWnd);
		if (pRes) initResources(xmlNode, pRes);

		// add frames to the window
		return initFrames(xmlNode, pWnd) > 0;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize window resources for the given window
 *
 * Loads window resources for the given window from the DOM node.
 *
 * \param pNode Pointer to a DOM node containing the elements of the given window
 * \param pRes Pointer to a GuiResource object that requires initialization
 * \return Returns a number of resources loaded.
 */
int GuiTemplate::initResources(s_xml_node_t *pNode, GuiResource *pRes)
{
	int numRes = 0;
	s_list_node_t *listNode = pNode->nodes.node;
	if (!pRes) return 0;

	for(int nodeItem = 0; nodeItem < pNode->nodes.nb_elt; nodeItem++, listNode = (s_list_node_t*)listNode->next)
	{
		s_xml_node_t *xmlNode = (s_xml_node_t*)listNode->element;

		// String resource?
		if(strcmp(xmlNode->name, "string")) continue;

		char* pfm = s_xml_node_get_attr_value(xmlNode, "platform");
		if (pfm && strcmp(pfm, platform.c_str())) continue;

		char* resId = s_xml_node_get_attr_value(xmlNode, "id");
		if(resId)
		{
			int iResId = atoi(resId);

			char* val = s_xml_node_get_attr_value(xmlNode, "val");
			pRes->setString(iResId, val);

			numRes++;
		}
	}

	return numRes;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize frames of the given window
 *
 * Loads and initializes GUI frames for the given window. Calls to primitive
 * initialization function for each valid frame.
 *
 * \param pNode Pointer to a DOM node containing the elements of the given window
 * \param pWnd Pointer to a window that requires initialization
 * \return Returns a number of frames inserted into the window.
 */
int GuiTemplate::initFrames(s_xml_node_t *pNode, GuiWindow *pWnd)
{
	int numFrames = 0;
	s_list_node_t *listNode = pNode->nodes.node;
	for(int nodeItem = 0; nodeItem < pNode->nodes.nb_elt; nodeItem++, listNode = (s_list_node_t*)listNode->next)
	{
		s_xml_node_t *xmlNode = (s_xml_node_t*)listNode->element;

		// what we got here is a frame of a window
		if(strcmp(xmlNode->name, "frame")) continue;

		char* pfm = s_xml_node_get_attr_value(xmlNode, "platform");
		if (pfm && strcmp(pfm, platform.c_str())) continue;

		// grab frame properties
		char* frameX = s_xml_node_get_attr_value(xmlNode, "x");
		char* frameY = s_xml_node_get_attr_value(xmlNode, "y");
		char* frameWidth = s_xml_node_get_attr_value(xmlNode, "width");
		char* frameHeight = s_xml_node_get_attr_value(xmlNode, "height");
		char* virtWidth = s_xml_node_get_attr_value(xmlNode, "vx");
		char* virtHeight = s_xml_node_get_attr_value(xmlNode, "vy");
		char* frameBuf = s_xml_node_get_attr_value(xmlNode, "buf");
		char* frameVisible = s_xml_node_get_attr_value(xmlNode, "visible");

		// grab frame ID
		char* resId = s_xml_node_get_attr_value(xmlNode, "id");
		int iResId = (resId != NULL) ? atoi(resId) : -1;

		// validate properties
		if(frameX == NULL || frameY == NULL ||
			frameWidth == NULL || frameHeight == NULL) return -1;

		if(virtWidth == NULL) virtWidth = frameWidth;
		if(virtHeight == NULL) virtHeight = frameHeight;

		// initialize frame object
		GuiFrame *pFrame;
		if(frameBuf && !strcmp(frameBuf, "y"))
		{
			// create a buffered frame
			pFrame = new GuiBufferedFrame(iResId, pWnd);
		}
		else
		{
			// create a non-buffered, or transparent frame
			pFrame = new GuiFrame(iResId, pWnd);
		}

		s_rect_t frameArea = { atoi(frameX), atoi(frameY),
							atoi(frameWidth), atoi(frameHeight) };

		pFrame->setGeometry(frameArea);

		if(frameVisible)
			pFrame->setVisible(strcmp(frameVisible, "y") == 0);

		// add primitives to the frame
		initPrimitives(xmlNode, pFrame);

		// initialize the frame surface
		if(!pFrame->initSurface(atoi(virtWidth), atoi(virtHeight)))
		{
			delete pFrame;
			return -1;
		}

		// finally, add the frame to the parent window
		pWnd->addFrame(pFrame);
		numFrames ++;
	}

	return numFrames;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Add extra primitives to the existing frame
 *
 * Creates and initializes GUI primitives for the given frame.
 *
 * \param pFrame Pointer to a frame that requires initialization
 * \param minId Initialize only primitives with id higher or equal than minId
 * \param maxId Initialize only primitives with id lower or equal than maxId
 * \param offset Add this number to the primitive id
 * \return Returns a number of primitives inserted into the frame.
 */
int GuiTemplate::addPrimitives(GuiFrame *pFrame, int minId, int maxId, int offset)
{
	int primitivesAdded = 0;
	if (pFrame == NULL || !pFrame->getParentWindow()) return 0;
	int frameId = pFrame->getResId();
	int parentWinId = pFrame->getParentWindow()->getResId();

	if(rootNode == NULL) return false;
	s_list_node_t *listNode = rootNode->nodes.node;

	// find the right window first
	for(int nodeItem = 0; nodeItem < rootNode->nodes.nb_elt; nodeItem++, listNode = (s_list_node_t*)listNode->next)
	{
		s_xml_node_t *xmlNode = (s_xml_node_t*)listNode->element;

		if(strcmp(xmlNode->name, "window")) continue;

		char* pfm = s_xml_node_get_attr_value(xmlNode, "platform");
		if (pfm && strcmp(pfm, platform.c_str())) continue;

		// is this the right window?
		char* resId = s_xml_node_get_attr_value(xmlNode, "id");
		int iResId = (resId != NULL) ? atoi(resId) : -1;
		if (iResId != parentWinId) continue;

		s_list_node_t *fListNode = xmlNode->nodes.node;
		for(int fNodeItem = 0; fNodeItem < xmlNode->nodes.nb_elt; fNodeItem++, fListNode = (s_list_node_t*)fListNode->next)
		{
			s_xml_node_t *fXmlNode = (s_xml_node_t*)fListNode->element;
			char* fResId = s_xml_node_get_attr_value(fXmlNode, "id");
			int fIResId = (fResId != NULL) ? atoi(fResId) : -1;

			char* pfm = s_xml_node_get_attr_value(fXmlNode, "platform");
			if (pfm && strcmp(pfm, platform.c_str())) continue;

			// is this the right frame?
			if(!strcmp(fXmlNode->name, "frame") && fIResId == frameId)
			{
				primitivesAdded += initPrimitives(fXmlNode, pFrame, false, minId, maxId, offset);
			}
		}
	}

	return primitivesAdded;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize primitives for the given frame
 *
 * Initializes GUI primitives for the given frame.
 *
 * \param pNode Pointer to a DOM node containing the elements to be initialized
 * \param pFrame Pointer to a frame that requires initialization
 * \param onLoad Skip initialization of primitives marked as onLoad="n"
 * \param minId Initialize only primitives with id higher or equal than minId
 * \param maxId Initialize only primitives with id lower or equal than maxId
 * \param offset Add this number to the primitive id
 * \return Returns a number of primitives inserted into the frame.
 */
int GuiTemplate::initPrimitives(s_xml_node_t *pNode, GuiFrame *pFrame, bool onLoad,
		int minId, int maxId, int offset)
{
	int numPrimitives = 0;
	s_list_node_t *listNode = pNode->nodes.node;
	for(int nodeItem = 0; nodeItem < pNode->nodes.nb_elt; nodeItem++, listNode = (s_list_node_t*)listNode->next)
	{
		s_xml_node_t *xmlNode = (s_xml_node_t*)listNode->element;

		// what we got here is a GUI primitive
		if(strcmp(xmlNode->name, "item")) continue;

		char* pfm = s_xml_node_get_attr_value(xmlNode, "platform");
		if (pfm && strcmp(pfm, platform.c_str())) continue;

		char* itemRes = s_xml_node_get_attr_value(xmlNode, "id");
		if (!itemRes) continue;

		int itemId = atoi(itemRes);
		if (itemId < minId || itemId > maxId) continue;

		// load only if
		// a) onLoad == false (we're adding primitives, not initializing
		//                      the window for the first time)
		// b) onLoadAttr == NULL (the item is not restricted by onLoad="" attribute)
		// c) onLoadAttr != "n"  (the item is not blocked for loading on window generation)
		char* onLoadAttr = s_xml_node_get_attr_value(xmlNode, "onLoad");
		if (!onLoad || !onLoadAttr || strcmp(onLoadAttr, "n")) {
			if (initPrimitive(xmlNode, pFrame, itemId + offset))
				numPrimitives ++;
		}
	}

	return numPrimitives;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \overload int GuiTemplate::initPrimitives(s_xml_node_t *pNode, GuiFrame *pFrame)
 */
int GuiTemplate::initPrimitives(s_xml_node_t *pNode, GuiFrame *pFrame)
{
	return initPrimitives(pNode, pFrame, true, INT_MIN, INT_MAX, 0);
}



////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize a GUI primitive and add it to the frame
 *
 * Creates and initializes GUI primitive and adds it to the frame.
 *
 * \param xmlNode Pointer to a DOM node containing the item
 * \param pFrame Pointer to a frame to add the primitive to
 * \param dstId Resource id of the primitive to be added
 * \return Returns true if primitive was inserted successfully
 */
bool GuiTemplate::initPrimitive(s_xml_node_t *xmlNode, GuiFrame *pFrame, int dstId)
{
	char absFileName[MAX_PATH];

	// grab primitive properties
	char* itemX = s_xml_node_get_attr_value(xmlNode, "x");
	char* itemY = s_xml_node_get_attr_value(xmlNode, "y");
	char* itemWidth = s_xml_node_get_attr_value(xmlNode, "width");
	char* itemHeight = s_xml_node_get_attr_value(xmlNode, "height");
	char* itemType = s_xml_node_get_attr_value(xmlNode, "type");
	char* itemVisible = s_xml_node_get_attr_value(xmlNode, "visible");

	// validate properties
	if(itemX == NULL || itemY == NULL || itemWidth == NULL ||
		itemHeight == NULL || itemType == NULL) return false;

	// extract primitive init and touch child nodes
	s_xml_node_t *xmlInit = NULL;
	s_xml_node_t *xmlTouch = NULL;
	s_list_node_t *childNode = xmlNode->nodes.node;

	for(int nodeParams=0; nodeParams < xmlNode->nodes.nb_elt; nodeParams++)
	{
		s_xml_node_t *xmlTemp = (s_xml_node_t*)childNode->element;
		if(!strcmp(xmlTemp->name, "init")) {
			char* pfm = s_xml_node_get_attr_value(xmlTemp, "platform");
			if (!pfm || !strcmp(pfm, platform.c_str()))
				xmlInit = xmlTemp;
		}
		else if(!strcmp(xmlTemp->name, "touch")) {
			char* pfm = s_xml_node_get_attr_value(xmlTemp, "platform");
			if (!pfm || !strcmp(pfm, platform.c_str()))
				xmlTouch = xmlTemp;
		}
		childNode = (s_list_node_t*)childNode->next;
	}

	// initialize the primitive object
	GuiPrimitive *pItem = NULL;

	// initialize the primitive position
	s_rect_t itemArea = { atoi(itemX), atoi(itemY),
						atoi(itemWidth), atoi(itemHeight) };

	// create an appropriate primitive basing on specified type
	if(!strcmp(itemType, "image"))
	{
		if(xmlInit == NULL) return false;
		pItem = new GuiImage(dstId, pFrame);
		pItem->setPosition(itemArea);

		char* fileName = s_xml_node_get_attr_value(xmlInit, "file");

		if(fileName) {
			if (lookupPath(fileName, absFileName, sizeof(absFileName)))
				((GuiImage*)pItem)->loadImage(absFileName);
		}
	}
	else if(!strcmp(itemType, "button"))
	{
		if(xmlInit == NULL) return false;

		char* fileDown = s_xml_node_get_attr_value(xmlInit, "down");
		char* fileUp = s_xml_node_get_attr_value(xmlInit, "up");

		if(fileUp == NULL && fileDown == NULL) return false;
		pItem = new GuiButton(dstId, pFrame);
		pItem->setPosition(itemArea);

		if(fileUp) {
			if (lookupPath(fileUp, absFileName, sizeof(absFileName)))
				((GuiButton*)pItem)->loadReleased(absFileName);
		}
		if(fileDown) {
			if (lookupPath(fileDown, absFileName, sizeof(absFileName)))
				((GuiButton*)pItem)->loadPressed(absFileName);
		}
	}
	else if(!strcmp(itemType, "rect"))
	{
		if(xmlInit == NULL) return false;
		pItem = new GuiRectangle(dstId, pFrame);
		pItem->setPosition(itemArea);

		char* colForegnd = s_xml_node_get_attr_value(xmlInit, "fore");

		if(colForegnd) ((GuiRectangle*)pItem)->setForeGround(colForegnd);
	}
	else if(!strcmp(itemType, "check"))
	{
		if(xmlInit == NULL) return false;

		char* fileOn = s_xml_node_get_attr_value(xmlInit, "on");
		char* fileOff = s_xml_node_get_attr_value(xmlInit, "off");
		char* fileGrayed = s_xml_node_get_attr_value(xmlInit, "grayed");
		char* checkState = s_xml_node_get_attr_value(xmlInit, "check");

		if(fileOn == NULL || fileOff == NULL) return false;
		pItem = new GuiCheckBox(dstId, pFrame);
		pItem->setPosition(itemArea);

		if (fileOn) {
			if (lookupPath(fileOn, absFileName, sizeof(absFileName)))
				((GuiCheckBox*)pItem)->loadEnabled(absFileName);
		}
		if (fileOff) {
			if (lookupPath(fileOff, absFileName, sizeof(absFileName)))
				((GuiCheckBox*)pItem)->loadDisabled(absFileName);
		}
		if (fileGrayed) {
			if (lookupPath(fileGrayed, absFileName, sizeof(absFileName)))
				((GuiCheckBox*)pItem)->loadGrayed(absFileName);
		}
		if(checkState) ((GuiCheckBox*)pItem)->setCheck(atoi(checkState) == 1);
	}
	else if(!strcmp(itemType, "radio"))
	{
		if(xmlInit == NULL) return false;

		char* fileOn = s_xml_node_get_attr_value(xmlInit, "on");
		char* fileOff = s_xml_node_get_attr_value(xmlInit, "off");
		char* fileGrayed = s_xml_node_get_attr_value(xmlInit, "grayed");
		char* checkState = s_xml_node_get_attr_value(xmlInit, "check");

		if(fileOn == NULL || fileOff == NULL) return false;
		pItem = new GuiRadioBox(dstId, pFrame);
		pItem->setPosition(itemArea);

		if (fileOn) {
			if (lookupPath(fileOn, absFileName, sizeof(absFileName)))
				((GuiRadioBox*)pItem)->loadEnabled(absFileName);
		}
		if (fileOff) {
			if (lookupPath(fileOff, absFileName, sizeof(absFileName)))
				((GuiRadioBox*)pItem)->loadDisabled(absFileName);
		}
		if (fileGrayed) {
			if (lookupPath(fileGrayed, absFileName, sizeof(absFileName)))
				((GuiRadioBox*)pItem)->loadGrayed(absFileName);
		}
		if(checkState) ((GuiRadioBox*)pItem)->setCheck(atoi(checkState) == 1);
	}
	else if(!strcmp(itemType, "text"))
	{
		char *textFont = NULL, *textSize = NULL, *textAlign = NULL;
		if(xmlInit) {
			textFont = s_xml_node_get_attr_value(xmlInit, "font");
			textSize = s_xml_node_get_attr_value(xmlInit, "size");
			textAlign = s_xml_node_get_attr_value(xmlInit, "align");
		}

		if (textFont)
			if (!lookupPath(textFont, absFileName, sizeof(absFileName))) return false;

		pItem = new GuiTextBox(dstId, pFrame);
		pItem->setPosition(itemArea);

		char* textColor = s_xml_node_get_attr_value(xmlInit, "color");
		char* textStr = s_xml_node_get_attr_value(xmlInit, "text");

		if(textFont && textSize) ((GuiTextBox*)pItem)->setFont(absFileName, atoi(textSize));
		if(textColor) ((GuiTextBox*)pItem)->setColor(textColor);
		if(textStr) ((GuiTextBox*)pItem)->setText(textStr);
		if(textAlign && strlen(textAlign) >= 2)
		{
			// default text position is top-left
			GuiTextBox::TextHAlignment hPos = GuiTextBox::LEFT;
			GuiTextBox::TextVAlignment vPos = GuiTextBox::TOP;

			switch(textAlign[0])	// determine the horizontal position
			{
			case 'l':
				hPos = GuiTextBox::LEFT;
				break;
			case 'r':
				hPos = GuiTextBox::RIGHT;
				break;
			case 'c':
				hPos = GuiTextBox::HCENTER;
				break;
			}

			switch(textAlign[1])	// determine the vertical position
			{
			case 't':
				vPos = GuiTextBox::TOP;
				break;
			case 'b':
				vPos = GuiTextBox::BOTTOM;
				break;
			case 'c':
				vPos = GuiTextBox::VCENTER;
				break;
			case 'n':
				vPos = GuiTextBox::VNORMAL;
				break;
			}

			((GuiTextBox*)pItem)->setAlignment(hPos, vPos);
		}
	}
	else if(!strcmp(itemType, "multitext"))
	{
		char *textFont = NULL, *textSize = NULL, *textAlign = NULL;
		if(xmlInit) {
			textFont = s_xml_node_get_attr_value(xmlInit, "font");
			textSize = s_xml_node_get_attr_value(xmlInit, "size");
			textAlign = s_xml_node_get_attr_value(xmlInit, "align");
		}

		if (textFont)
			if (!lookupPath(textFont, absFileName, sizeof(absFileName))) return false;

		pItem = new GuiMultiTextBox(dstId, pFrame);
		pItem->setPosition(itemArea);

		char* textColor = s_xml_node_get_attr_value(xmlInit, "color");
		char* textStr = s_xml_node_get_attr_value(xmlInit, "text");

		if(textFont && textSize) ((GuiMultiTextBox*)pItem)->setFont(absFileName, atoi(textSize));
		if(textColor) ((GuiMultiTextBox*)pItem)->setColor(textColor);
		if(textStr) ((GuiMultiTextBox*)pItem)->setText(textStr);
		if(textAlign && strlen(textAlign) >= 2)
		{
			// default text position is top-left
			GuiMultiTextBox::TextHAlignment hPos = GuiMultiTextBox::LEFT;
			GuiMultiTextBox::TextVAlignment vPos = GuiMultiTextBox::TOP;

			switch(textAlign[0])	// determine the horizontal position
			{
			case 'l':
				hPos = GuiMultiTextBox::LEFT;
				break;
			case 'r':
				hPos = GuiMultiTextBox::RIGHT;
				break;
			case 'c':
				hPos = GuiMultiTextBox::HCENTER;
				break;
			}

			switch(textAlign[1])	// determine the vertical position
			{
			case 't':
				vPos = GuiMultiTextBox::TOP;
				break;
			case 'b':
				vPos = GuiMultiTextBox::BOTTOM;
				break;
			case 'c':
				vPos = GuiMultiTextBox::VCENTER;
				break;
			}

			((GuiMultiTextBox*)pItem)->setAlignment(hPos, vPos);
		}
	}
	else if(!strcmp(itemType, "matrix"))
	{
		if(xmlInit == NULL) return false;

		char* textFont = s_xml_node_get_attr_value(xmlInit, "font");
		char* textSize = s_xml_node_get_attr_value(xmlInit, "size");

		if(textFont == NULL || textSize == NULL) return false;
		if (!lookupPath(textFont, absFileName, sizeof(absFileName))) return false;

		pItem = new GuiTextMatrix(dstId, pFrame, atoi(itemWidth), atoi(itemHeight));
		pItem->setPosition(itemArea);

		char *staticTextColor =
			s_xml_node_get_attr_value(xmlInit, "static_text_color");
		char *inactiveFieldColor =
			s_xml_node_get_attr_value(xmlInit, "inactive_field_color");
		char *activeFieldColor =
			s_xml_node_get_attr_value(xmlInit, "active_field_color");
		char *selectionColor =
			s_xml_node_get_attr_value(xmlInit, "selection_color");
		char *selectionBgColor =
			s_xml_node_get_attr_value(xmlInit, "selection_bg_color");

		char *cols = s_xml_node_get_attr_value(xmlInit, "cols");
		char *rows = s_xml_node_get_attr_value(xmlInit, "rows");

		((GuiTextMatrix*)pItem)->setFont(absFileName, atoi(textSize));
		((GuiTextMatrix*)pItem)->setColor(staticTextColor,
											inactiveFieldColor,
											activeFieldColor,
											selectionColor,
											selectionBgColor);

		if ((cols != NULL) && (rows != NULL))
			((GuiTextMatrix*)pItem)->setDimensions(atoi(cols), atoi(rows));
	}
	else if(!strcmp(itemType, "edit"))
	{
		char *textFont = NULL, *textSize = NULL, *textAlign = NULL;
		if(xmlInit) {
			textFont = s_xml_node_get_attr_value(xmlInit, "font");
			textSize = s_xml_node_get_attr_value(xmlInit, "size");
			textAlign = s_xml_node_get_attr_value(xmlInit, "align");
		}

		if (textFont)
			if (!lookupPath(textFont, absFileName, sizeof(absFileName))) return false;

		pItem = new GuiEditBox(dstId, pFrame);
		pItem->setPosition(itemArea);

		char* textColor = s_xml_node_get_attr_value(xmlInit, "color");
		char* textStr = s_xml_node_get_attr_value(xmlInit, "text");

		if(textFont && textSize) ((GuiEditBox*)pItem)->setFont(absFileName, atoi(textSize));
		if(textColor) ((GuiEditBox*)pItem)->setColor(textColor);
		if(textStr) ((GuiEditBox*)pItem)->setText(textStr);
		if(textAlign && strlen(textAlign) >= 2)
		{
			// default text position is top-left
			GuiEditBox::TextHAlignment hPos = GuiEditBox::LEFT;
			GuiEditBox::TextVAlignment vPos = GuiEditBox::TOP;

			switch(textAlign[0])	// determine the horizontal position
			{
			case 'l':
				hPos = GuiEditBox::LEFT;
				break;
			case 'r':
				hPos = GuiEditBox::RIGHT;
				break;
			case 'c':
				hPos = GuiEditBox::HCENTER;
				break;
			}

			switch(textAlign[1])	// determine the vertical position
			{
			case 't':
				vPos = GuiEditBox::TOP;
				break;
			case 'b':
				vPos = GuiEditBox::BOTTOM;
				break;
			case 'c':
				vPos = GuiEditBox::VCENTER;
				break;
			}

			((GuiEditBox*)pItem)->setAlignment(hPos, vPos);
		}	}
	else if(!strcmp(itemType, "animation"))
	{
		if(xmlInit == NULL) return false;
		pItem = new GuiAnimation(dstId, pFrame);
		pItem->setPosition(itemArea);
		childNode = xmlNode->nodes.node;

		for(int nodeParams=0;
				nodeParams < xmlNode->nodes.nb_elt; nodeParams++)
		{
			s_xml_node_t *xmlTemp = (s_xml_node_t*)childNode->element;
			if(!strcmp(xmlTemp->name, "frame"))
			{
				char* fileName = s_xml_node_get_attr_value(xmlTemp, "file");
				if(fileName)
					if (lookupPath(fileName, absFileName, sizeof(absFileName)))
						((GuiAnimation*)pItem)->addFrame(absFileName);
			}

			childNode = (s_list_node_t*)childNode->next;
		}

		char* frameDelay = s_xml_node_get_attr_value(xmlInit, "delay");
		char* loopCount = s_xml_node_get_attr_value(xmlInit, "loop");
		if(frameDelay) ((GuiAnimation*)pItem)->setFrameDelay(atoi(frameDelay));
		if(loopCount) ((GuiAnimation*)pItem)->setLoopCount(atoi(loopCount));
	}
	else if(!strcmp(itemType, "progress"))
	{
		if(xmlInit == NULL) return false;

		char* fileBack = s_xml_node_get_attr_value(xmlInit, "back");
		char* fileBar = s_xml_node_get_attr_value(xmlInit, "bar");

		if(fileBar == NULL) return false;
		if (!lookupPath(fileBar, absFileName, sizeof(absFileName))) return false;

		pItem = new GuiProgressBar(dstId, pFrame);
		pItem->setPosition(itemArea);

		((GuiProgressBar*)pItem)->loadBar(absFileName);
		if(fileBack) {
			if (lookupPath(fileBack, absFileName, sizeof(absFileName)))
				((GuiProgressBar*)pItem)->loadBackground(absFileName);
		}
	}
	else if(!strcmp(itemType, "scroll"))
	{
		char absFileName2[MAX_PATH];
		if(xmlInit == NULL) return false;

		char* fileBack = s_xml_node_get_attr_value(xmlInit, "back");
		char* fileBar = s_xml_node_get_attr_value(xmlInit, "bar");
		char* fileStart = s_xml_node_get_attr_value(xmlInit, "start");
		char* fileStop = s_xml_node_get_attr_value(xmlInit, "stop");

		if(fileBar == NULL) return false;
		if (!lookupPath(fileBar, absFileName, sizeof(absFileName))) return false;
		pItem = new GuiScrollBar(dstId, pFrame);
		pItem->setPosition(itemArea);

		((GuiScrollBar*)pItem)->loadBarFill(absFileName);
		if(fileBack) {
			if (lookupPath(fileBack, absFileName, sizeof(absFileName)))
				((GuiScrollBar*)pItem)->loadBackground(absFileName);
		}
		if(fileStart && fileStop) {
			if (lookupPath(fileStart, absFileName, sizeof(absFileName)) &&
			lookupPath(fileStop, absFileName2, sizeof(absFileName2)))
				((GuiScrollBar*)pItem)->loadBarEdges(absFileName, absFileName2);
		}

	}
	else if(!strcmp(itemType, "drawarea"))
	{
		pItem = new GuiDrawArea(dstId, pFrame);
		pItem->setPosition(itemArea);
	}

	if(pItem == NULL) return false;

	// set visibility flag
	pItem->setVisible(!(itemVisible != NULL && !strcmp(itemVisible, "n")));

	// extract data for the touch region, if any
	if(xmlTouch != NULL)
	{
		char* touchX = s_xml_node_get_attr_value(xmlTouch, "x");
		char* touchY = s_xml_node_get_attr_value(xmlTouch, "y");
		char* touchWidth = s_xml_node_get_attr_value(xmlTouch, "width");
		char* touchHeight = s_xml_node_get_attr_value(xmlTouch, "height");

		if(touchX && touchY && touchWidth && touchHeight)
		{
			s_rect_t touchArea = { atoi(touchX), atoi(touchY),
									atoi(touchWidth), atoi(touchHeight) };
			pItem->setTouchArea(touchArea);
		}
		else xmlTouch = NULL;
	}

	// finally, add the primitive to the frame
	pFrame->addPrimitive(pItem);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Add a default path to template engine
 *
 * Add a default path to the end of the path list.
 * The template engine will go through this list and resolve any file name
 * paths if required
 *
 * \param path The path to be added
 */
void GuiTemplate::addPath(std::string path)
{
	defaultPathList.push_back(path);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get the absolute path of the file
 *
 * Go through path list and try to find the absolute path of the file specified.
 *
 * \param file File name
 * \param ret Buffer for return data
 * \param count Return data buffer size
 * \return Returns true if path was resolved.
 */
bool GuiTemplate::lookupPath(const char* file, char* ret, size_t count)
{
	struct stat sb;

	if (!count) return false;
	ret[0] = 0;

	// absolute path?
	if (file[0] == '/') {
		if (stat(file, &sb) != 0) {
			return false;
		}

		if (S_ISREG(sb.st_mode)) {
			strncpy(ret, file, count);
			ret[count-1] = 0;
			return true;
		}
	}
	else {
		PATHLIST_T::iterator it;
		for ( it=defaultPathList.begin() ; it != defaultPathList.end(); it++ ) {
			// relative path - try all the paths
			snprintf(ret, count, "%s/%s", (*it).c_str(), file);
			ret[count-1] = 0;

			if (stat(ret, &sb) != 0) continue;

			if (S_ISREG(sb.st_mode)) {
				return true;
			}
		}
	}

	ret[0] = 0;
	return false;
}
