#ifndef GUITEMPLATE_H_
#define GUITEMPLATE_H_

#include <string>
#include <list>

#include <exynth/exynth.h>
#include <exynth/xml.h>

#include "GuiWindow.h"
#include "GuiResource.h"

#include <list>

#define PATHLIST_T std::list<std::string>

/**
 * \brief Skinnable GUI implementation
 *
 * Given class implements the window skin function. Template object loads an
 * XML file describing the look of the user interface. Once application
 * windows are created, their GUI primitives may be initialized by loading an
 * appropriate window template. With such approach, the major part of the user
 * interface design may remain separate from the application logic and
 * actual code.
 *
 * Once instantiated, the template object builds a DOM tree from the
 * specified XML GUI skin file. The windows implemented via GuiWindow class
 * may then load their content from a defined template using their unique
 * window identifiers. The templates in the skin file determine the frames
 * and primitives that apply to corresponding windows.
 *
 * Below is a structure of XML code that is used to define user interface
 * for multiple windows:
 *
 * \code
 * <?xml version="1.0"?>
 * <guiskin version="1.0">
 *	<window id="1" ...>
 * 		<frame id="1" ...>
 *			<item id="1" type="button" ...>
 * 				<init ... />
 * 			</item>
 * 			<item id="2" type="image" ...>
 * 				...
 * 			</item>
 * 		</frame>
 * 		<frame id="2" ...>
 * 			...
 * 		</frame>
 * 	</window>
 *	<window id="2" ...>
 * 		<frame id="1" ...>
 * 			...
 * 		</frame>
 * 	</window>
 * </guiskin>
 * \endcode
 *
 * For more information about the structure of the 'window' tag, please refer
 * to the GuiWindow class documentation. For information about the 'frame' tag,
 * refer to the GuiFrame class documentation. For information about primitive
 * tags, refer to a specific primitive class documentation.
 *
 * \author Andrei Birjukov
 */

class GuiTemplate
{
public:
	GuiTemplate();
	virtual ~GuiTemplate();

	bool loadFile(const char* fileName);
	bool initWindowLayout(GuiWindow *pWnd);
	void unload();
	void setPlatform(std::string plat);

	int addPrimitives(GuiFrame *pFrame, int minId, int maxId, int offset);
	int getRootResource(GuiResource *pRes);

	void addPath(std::string path);

	bool lookupPath(const char* file, char* ret, size_t count);
private:

	int initFrames(s_xml_node_t *pNode, GuiWindow *pWnd);
	int initResources(s_xml_node_t *pNode, GuiResource *pRes);
	int initPrimitives(s_xml_node_t *pNode, GuiFrame *pFrame);
	int initPrimitives(s_xml_node_t *pNode, GuiFrame *pFrame, bool onLoad,
			int minId, int maxId, int offset);

	bool initPrimitive(s_xml_node_t *xmlNode, GuiFrame *pFrame, int dstId);

	s_xml_node_t *rootNode;
	PATHLIST_T defaultPathList;
	std::string platform;
};

#endif /*GUITEMPLATE_H_*/
