#ifndef GUITEXTBOX_H_
#define GUITEXTBOX_H_

#include <string>

#include <exynth/font.h>

#include "GuiPrimitive.h"

#define GUITEXTBOX(f,p) getPrimitive<GuiTextBox>(f,p)

/**
 * \brief Textbox primitive
 *
 * Implements a single-line textbox. The specified font
 * is rendered onto the painting surface with selected size and color.
 * The vertical and horizontal text alignment can also be specified.
 *
 * XML code below defines a text box primitive:
 *
 * \code
 *	<item id="ID" type="text" x="X" y="Y" width="W" height="H">
 * 		<init text="T" font="F" size="S" color="RRGGBB" align="HV" />
 * 		<touch x="tX" y="tY" width="tW" height="tH"> />
 * 	</item>
 * \endcode
 *
 * Where
 * - ID represents the primitive unique identifier
 * - X, Y, W, H is the set of primitive coordinates in the parent frame
 * - T contains the initial text of the text box
 * - F specifies the font name that should match an available TTF font file name
 * - S specifies the font size
 * - RRGGBB specifies the font color encoded as hex RGB (e.g. "6699cc")
 * - HV sets the text alignment within the box coordinates. The first character
 *  specifies the horizontal alignment, the second character specifies the
 *	vertical alignment. See alignment options below for proper switches. So for example the "lc"
 *	means to align text horizontally left and vertically middle.
 * - tX, tY, tW, tH are the coordinates of the touchable area in the parent frame
 *
 * Possible horizontal alignment options are:
 * - "l" - left
 * - "r" - right
 * - "c" - center
 *
 * Possible vertical alignment options are:
 * - "t" - top
 * - "b" - bottom
 * - "c" - middle
 * - "n" - baseline (the usual way to display text)
 *
 * \note Touchable area is optional and usually required only for button type elements.
 *
 * \author Andrei Birjukov
 */

class GuiTextBox : public GuiPrimitive
{
public:

	/** Horizontal alignment type */
	enum TextHAlignment
	{
		/** Align left */
		LEFT,
		/** Align right */
		RIGHT,
		/** Align center */
		HCENTER
	};

	/** Vertical alignment type */
	enum TextVAlignment
	{
		/** Align top */
		TOP,
		/** Align bottom */
		BOTTOM,
		/** Align middle */
		VCENTER,
		/** Baseline alignment */
		VNORMAL

	};

	/** Identifies the font style, being a separate TTF */
	enum FontStyle
	{
		/** Regular font style */
		REGULAR,
		/** Bold font style */
		BOLD,
		/** Italic font style */
		ITALIC
	};

	/**
	 * Define a rich line of text and contains font size, style,
	 * color, text alignment, along with the actual text.
	 */
	struct FormattedTextLine
	{
		/** Identifier of the text line, optional */
		int					lineId;
		/** Horizontal alignment of text */
		TextHAlignment		textAlign;
		/** Specifies the style of font: bold, italic, regular */
		FontStyle			fontStyle;
		/** Specifies the size of font, in pixels */
		int					fontSize;
		/** Specifies the color of font, RGB structure */
		RGB_COLOR			fontColor;

		/** Displayed text string in UTF-8 format */
		std::string 		lineText;
	};

	GuiTextBox(int idResource, GuiFrame* pFrame);
	virtual ~GuiTextBox();

	bool setFont(const char* fontName, int fontSize);
	void setColor(const char *pColor);
	void setColor(RGB_COLOR *pColor);
	void setAlignment(TextHAlignment horz, TextVAlignment vert);

	void setText(const char* boxText);
	bool setText(FormattedTextLine *pText);

	void getRequiredSize(s_rect_t &rect);
	void getAlignment(TextHAlignment *pHorz, TextVAlignment *pVert);

protected:

	virtual void onPaint(GuiSurface *drawSurface, s_rect_t &paintArea);

private:
	s_font_t*		fontData;
	TextHAlignment	horzAlign;
	TextVAlignment	vertAlign;
};

#endif /*GUITEXTBOX_H_*/
