/***************************************************************************
 *   Copyright            : (c) 2007-2009 Artec Design LLC                 *
 *                        : (c) 2007-2009 Andrei Birjukov                  *
 *                        : (c) 2007-2009 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <exynth/font.h>

#include "GuiTextWrapper.h"

//#define DEBUG

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default constructor
 *
 * Default constructor of the class. Initializes the width to maximum available and
 * sets the X coordinate position to NULL.
 */
GuiTextWrapper::GuiTextWrapper() {
	xPos = NULL;
	width = INT_MAX;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default destructor
 *
 * Releases all the allocated text line structures and frees the X coordinate position pointer
 */
GuiTextWrapper::~GuiTextWrapper() {
	freeLines();
	if (xPos) {
		free(xPos);
		xPos = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Gets the list of allocated text lines
 *
 * \return Returns the list of allocated text lines
 */
std::list<textLine> *GuiTextWrapper::getLineList()
{
	return &textLines;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get the height of the text line
 *
 * \return Returns the height of the rendered text line
 */
int GuiTextWrapper::getLineHeight()
{
	return lineHeight;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief
 *
 * \return
 */
int GuiTextWrapper::getTotalHeight()
{
	return totalHeight;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the width of the line
 *
 * Sets the width of one text line
 * \param w Width to be set
 */
void GuiTextWrapper::setWidth(int w)
{
	width = w;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the font used for rendering
 *
 * Sets the font name and size that will be used for text rendering
 * \param fontName Name of the font to be used
 * \param fontSize Size of the font to be used
 */
void GuiTextWrapper::setFont(const char* fontName, int fontSize)
{
	this->fontName = fontName;
	this->fontSize = fontSize;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Load the text into the wrapper
 *
 * Sets the text that will be used for wrapping
 * \param text Pointer to the text being loaded
 */
void GuiTextWrapper::setText(const char *text) {
	freeLines();
	if (!updateWidths(text)) return;

	const char *p = text;
	size_t len = strlen(text);
	int *x = xPos;

	start.s = text;
	start.x = x;
	wrap.s = NULL;
	skipSpace = false;

	while (len > 0)
	{
		// get next char unicode number
		unsigned short c = 0;
		int cnt = s_font_utf8_to_unicode(&c, p, len);
		if (cnt <= 0) {
			len -= 1; p += 1;
			continue;
		}

		pos curr = {p,x}, next = {p+cnt, x+1};
		// process the char
		addChar(c, curr, next);

		len -= cnt;
		p += cnt;
		x++;
	}

	// add the remaining data
	pos curr = {p,x};
	addLine(start, curr);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Updates the character coordinates
 *
 * Updates the array of each character's X coordinates according to the loaded font information
 * \param text Pointer to text where to get characters from
 * \return Returns true if update was successful, false otherwise
 */
bool GuiTextWrapper::updateWidths(const char *text) {
	s_font_t *font;

	if (xPos) {
		free(xPos);
		xPos = NULL;
	}

	if(s_font_init(&font, fontName.c_str())) {
		return false;
	}
	s_font_set_size(font, fontSize);
	lineHeight = font->lineskip;
	asc2descHeight = font->ascent - font->descent;

	xPos = (int*) malloc(sizeof(int)*(strlen(text)+1));
	if (!xPos) {
		s_font_uninit(font);
		return false;
	}
	int tw = s_font_get_width_pos(font, text, xPos);
	s_font_uninit(font);

	if (tw < 0) {
		free(xPos);
		xPos = NULL;
		return false;
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Add a character to the line at the specified position
 *
 * \param c Character being added
 * \param curr Current position where to place the character
 * \param next Next position that will be occupied
 */
void GuiTextWrapper::addChar(unsigned short c, pos curr, pos next) {
#ifdef DEBUG
	printf("addchar [%c] start{%p, %d} wrap{%p} curr{%p, %d} next{%p, %d}\n", c&0xFF,
			start.s, *start.x, wrap.s, curr.s, *curr.x, next.s, *next.x);
#endif
	if (c == '\n') {
		// wrap here
		addLine(start, curr);
		skipSpace = false;
		start = next;
		wrap.s = NULL;
		return;
	}

	if (c < 0x7F && isspace(c & 0x7F)) {
		if (skipSpace) { // remove extra spaces after wrap at next line start
			start = next;
			return;
		}
		if (start.s == curr.s) return; // always allow first char on line

		// check if the char fits, wrap before it if needed
		if (*next.x - *start.x > width) {
			addLine(start, curr);
			start = next;
			wrap.s = NULL;
			skipSpace = true;
			return;
		}

		// set wrap point after this char
		wrap = next;
		return;
	}

	// non-space character
	skipSpace = false;
	if (start.s == curr.s) return; // always allow first char on line

	// check if the char fits
	if (*next.x - *start.x > width) {
		// first word on the line - wrap here
		if (!wrap.s) {
			addLine(start, curr);
			skipSpace = false;
			start = curr;
			wrap.s = NULL;
			return;
		}
		// otherwise, wrap at last wrap point
		addLine(start, wrap);
		skipSpace = false;
		start = wrap;
		wrap.s = NULL;
		return;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Adds a line of text to line list
 *
 * \param start Line start poisition
 * \param end Line end position
 */
void GuiTextWrapper::addLine(pos start, pos end) {
#ifdef DEBUG
	printf("addLine\n");
#endif
	textLine l;
	l.start = start.s;
	l.len = end.s - start.s;
	int cnt = end.x - start.x + 1; // nr of elements +1
	l.posX = (int*)malloc(sizeof(int) * cnt);
	if (l.posX) {
		for (int i=0; i<cnt; i++) {
			l.posX[i] = start.x[i] - start.x[0];
		}
	}
	l.posY = totalHeight;
	l.startChar = start.x - xPos;
	l.numChars = end.x - start.x;

	if (!totalHeight)
		totalHeight = asc2descHeight;
	else
		totalHeight += lineHeight;

	textLines.push_back(l);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Free all the lines stored
 *
 * Frees all the lines stored in line list
 */
void GuiTextWrapper::freeLines() {
	while (!textLines.empty()) {
		textLine l = textLines.back();
		textLines.pop_back();
		if (l.posX) free(l.posX);
	}
	totalHeight = 0;
	skipSpace = false;
}
