#ifndef GUITEXTWRAPPER_H_
#define GUITEXTWRAPPER_H_

#include <string>
#include <list>
#include <exynth/exynth.h>

/** Holds the information about one line of thext */
struct textLine {

	/** Pointer to the start location in original string */
	const char *start;

	/** Length of the text */
	size_t len;			// number of bytes to use from *start

	/** Number of UTF-8 encoded characters on line */
	size_t numChars;	// number of UTF-8 characters on this line

	/** Position of the first UTF-8 encoded character */
	size_t startChar;	// *start utf-8 character position

	/** Array of X coordinates of all the characters and the position after the last character */
	int *posX;			// array of X positions of all characters + X position after the last character

	/** Y coordinate of the line */
	int posY;			// y position of the line
};

/**
 * Implements the helper class to create and manager multiline text.
 * The text is originally stored in a variable as a whole, this class creates
 * data structures that wrap the text into lines. Each structure contains information about one
 * line of text.
 *
 * For proper wrapping the font name and size must be known.
 */
class GuiTextWrapper
{
public:
	GuiTextWrapper();
	~GuiTextWrapper();

	void setWidth(int w);
	void setFont(const char* fontName, int fontSize);
	void setText(const char *text);

	std::list<textLine> *getLineList();
	int	getLineHeight();
	int getTotalHeight();

protected:

	/** Holds the information about the current position in string */
	struct pos {

		/** Pointer to the currently selected character */
		const char *s;

		/** Pointer to the X-coordinate value in the character coordinates array */
		int *x;
	};

	bool updateWidths(const char *text);
	void addChar(unsigned short c, pos curr, pos next);
	void addLine(pos start, pos end);
	void freeLines();

	/** Line width */
	int				width;

	/** Name of the font that will be used for rendering */
	std::string		fontName;

	/** Font size that will be used for rendering */
	int				fontSize;

	/** Array of character X coordinate positions in string */
	int				*xPos;

	/** Height of the line */
	int				lineHeight;

	/** Total height of all the lines */
	int 			totalHeight;

	/** Maximum height the glyphs might use in this line */
	int				asc2descHeight;

	/** Position where the wrap should occur */
	pos				wrap;

	/** Position where to start text wrapping */
	pos				start;

	/** If true, the spaces at the beginning of line are not calculated for wrapping and skipped */
	bool			skipSpace;

	/** List of text lines */
	std::list<textLine>	textLines;
};

#endif /* GUITEXTWRAPPER_H_ */
