/***************************************************************************
 *   Copyright            : (c) 2007-2008 Artec Design LLC                 *
 *                        : (c) 2007-2008 Andrei Birjukov                  *
 *                        : (c) 2007-2008 Anti Sullin                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include "GuiWindow.h"

#include <exynth/input/input.h>

//#define DEBUG
//#define DEBUG_TIMER

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Default constructor of the generic GUI window object.
 * Initializes window location and size to zeros, also resets window style,
 * return code and a number of flags.
 *
 * \param idResource Unique window identifier
 */
GuiWindow::GuiWindow(int idResource)
{
	this->idResource = idResource;
	wndHandle = NULL;
	wndSurface = NULL;
	touchItem.idFrame = RES_ID_INVALID;
	touchItem.idPrimitive = RES_ID_INVALID;

	windowPos.x = windowPos.y = 0;
	windowPos.w = windowPos.h = 0;

	isEnabled = true;
	isReady = false;
	isAlwaysOnTop = false;
	selfDestruct = false;
	refreshRealSurface = true;
	isInitialized = false;

	wndStyle = 0;
	returnCode = 0;

	#ifdef DEBUG
	printf("GuiWindow(%p)::GuiWindow: new window %d\n", this, idResource);
	#endif
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destructor for the generic window objects.
 * If the given window is a child or a popup window, the destructor will
 * block until the window thread terminates.
 *
 * Also releases window descriptor, surface and deletes frames, if
 * there are any remaining for some reason. Self destructable will perform
 * deletion of the window object automatically, if selfDestruct flag is set.
 */

GuiWindow::~GuiWindow()
{
	#ifdef DEBUG
	printf("GuiWindow(%p)::~GuiWindow: id %d\n", this, idResource);
	#endif

	// now we can kill the window safely
	list<GuiFrame*>::iterator it;
	for(it = frameList.begin(); it != frameList.end(); it ++)
	{
		GuiFrame* pFrame = (GuiFrame*)*it;
		delete pFrame;
	}

	frameList.clear();

	if(wndHandle) s_window_uninit(wndHandle);
	if(wndSurface) delete wndSurface;

	#ifdef DEBUG
	printf("GuiWindow(%p)::~GuiWindow: id %d terminated\n", this, idResource);
	#endif
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Create window
 *
 * Initializes window data and calls eXynth engine, creating a new window and
 * connecting it to the server. The created window afterwards needs to be shown
 * using showWindow() and redrawn using redrawWindow().
 *
 * \param style Window style flags
 * - GUI_STYLE_MAIN creates a main window
 * - GUI_STYLE_CHILD window is associated with the parent window but the parent is not blocked. Child
 *   windows are automatically deleted when the parent is destroyed
 * - GUI_STYLE_MODAL specifies that the window is modal, deactivates the parent
 *   window and blocks its input until the modal window is closed
 * - GUI_STYLE_INPUT specifies that the window will receive all input events
 *   even if it is not in focus
 * - GUI_SYTLE_MONITOR window will be notified of window stack changes
 *
 * \param pParent Pointer to the parent window. Required for child, modal
 * and tooltip windows
 *
 * \return Returns true if the window has been created successfully.
 */

bool GuiWindow::createWindow(int style, GuiWindow *pParent)
{
	if (wndHandle != NULL) {
		return false;
	}

	// modal, child and tooltip styles are only possible if there is a parent
	if((style & (GUI_STYLE_MODAL | GUI_STYLE_CHILD)) && pParent == NULL) {
		#ifdef DEBUG
		printf("GuiWindow(%p)::createWindow: Cannot create modal/child window: parent not found\n", this);
		#endif
		return false;
	}

	if (s_window_init(&wndHandle) != 0) {
		#ifdef DEBUG
		printf("GuiWindow(%p)::createWindow: s_window_init failed\n", this);
		#endif
		return false;
	}

	int xynthStyle = 0;

	if(style & GUI_STYLE_MAIN) xynthStyle |= WINDOW_MAIN;
	if(style & GUI_STYLE_CHILD) xynthStyle |= WINDOW_CHILD;
	if(style & GUI_STYLE_MODAL) xynthStyle |= WINDOW_MODAL;
	if(style & GUI_STYLE_INPUT) xynthStyle |= WINDOW_INPUT;
	if(style & GUI_STYLE_MONITOR) xynthStyle |= WINDOW_MONITOR;

	s_window_t *parentHandle = (pParent == NULL) ? NULL : pParent->wndHandle;
	if (s_window_new(wndHandle, (S_WINDOW)xynthStyle, parentHandle) != 0) {
		#ifdef DEBUG
		printf("GuiWindow(%p)::createWindow: s_window_new failed\n", this);
		#endif
		return false;
	}

	if(wndSurface) delete wndSurface;
	wndSurface = new GuiSurface(wndHandle->surface);

	wndHandle->data = (void*)this;

	s_window_atevent(wndHandle, eventHandlerProc);

	wndParent = pParent;
	wndStyle = style;
	isReady = true;

	// child windows must automatically delete themselves
	if(wndStyle & (GUI_STYLE_CHILD | GUI_STYLE_MODAL))  selfDestruct = true;

	#ifdef DEBUG
	printf("GuiWindow(%p)::createWindow: id %d style %x\n", this, idResource, wndStyle);
	#endif

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set window position and size
 *
 * Updates the window coordinate rectangle, including window position and
 * dimensions. Using negative x and y coordinates, window may be placed
 * outside the screen. Width and height must be always positive.
 *
 * \param wndPos New window coordinate rectangle (x, y, w, h). Coordinates
 * should start with (0, 0) at the upper left corner of the screen.
 * \return Returns true if the new window coordinates are set successfully.
 */

bool GuiWindow::setWindowPos(s_rect_t &wndPos)
{
	if(wndHandle == NULL || !isReady) {
		return false;
	}
	if(wndPos.w < 0 || wndPos.h < 0) {
		return false;
	}

	s_window_set_coor(wndHandle, 0, wndPos.x, wndPos.y,
						wndPos.w, wndPos.h);

	windowPos = wndPos;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Show or hide window
 *
 * Controls the window visibility, i.e. shows or hides the window.
 * onShowWindow() virtual function is called internally to notify the derived
 * classes about the new state.
 *
 * \param state Set to true to make the window visible.
 * \return Returns true if the window visibility has been updated successfully.
 */

bool GuiWindow::showWindow(bool state)
{
	if(wndHandle == NULL || !isReady) {
		return false;
	}

	// add the window to priority list, visible=0
	// (to maintain correct window order when more
	// windows are created in onInitialize().
	if(state) s_window_show_prepare(wndHandle);

	// on first show, initialize window
	if (!isInitialized) {
		isInitialized = true;
		onInitialize();
	}

	// perform custom show/hide actions
	if(!onShowWindow(state ? SHOW_WINDOW : HIDE_WINDOW, 0)) {
		return true;
	}

	if(state) s_window_show(wndHandle);
	else s_window_hide(wndHandle);

	onPostShowWindow(state ? SHOW_WINDOW : HIDE_WINDOW, 0);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set always-on-top property
 *
 * Sets or resets the window always-on-top flag, i.e. fixes the window at the
 * top-most position in the current window Z-order.
 * Set this before createWindow() call; later updates have no effect.
 *
 * \param flag Specifies if the window is always displayed on top.
 * \return Returns true if the window state has been updated successfully.
 */

bool GuiWindow::setAlwaysOnTop(bool flag)
{
	if(wndHandle == NULL || !isReady) return false;

	isAlwaysOnTop = flag;
	s_window_set_alwaysontop(wndHandle, (int)flag);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable or disable window
 *
 * Sets the window enable flag, i.e. activates or deactivates the window.
 * Deactivated windows do not process any keypad or touchscreen events.
 *
 * \param flag Specifies if the window is active
 * \return Returns true if the window state has been updated successfully.
 */

bool GuiWindow::setEnabled(bool flag)
{
	if(wndHandle == NULL || !isReady) return false;
	isEnabled = flag;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Redraw window completely
 *
 * Performs the complete window redraw.
 *
 * \return Returns true if the redrawing is completed successfully.
 */
bool GuiWindow::redrawWindow()
{
	s_rect_t updateRect = { 0, 0, windowPos.w, windowPos.h };

	bool ret = redrawWindow(updateRect);
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Redraw window partially
 *
 * Redraws a selected area of the window. Redraw area is given as rectangle.
 * Redraw call blocks until all GUI primitives in the are have been repainted.
 *
 * \param updateArea Specifies the rectangular window area to be redrawn.
 * Coordinates should start with (0, 0) from the upper left corner of the window
 * \return Returns true if the redrawing is completed successfully.
 */
bool GuiWindow::redrawWindow(s_rect_t &updateArea)
{
	if(wndHandle == NULL || !isReady) return false;

	onPaint(updateArea);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close window
 *
 * Sends a command to the window instructing it to close and release all
 * contained GUI primitives. Also, virtual onShowWindow() function is called
 * in the calling thread to notify the derived classes about state change.
 *
 * \param retCode Specifies the window return code. Return code may be
 * retrieved after the window is closed for further processing.
 *
 * \return Returns true if the closing procedure was initiated.
 */
bool GuiWindow::closeWindow(int retCode)
{
	#ifdef DEBUG
	printf("GuiWindow(%p)::closeWindow: id %d code %d\n", this, idResource, retCode);
	#endif

	if(wndHandle == NULL || !isReady) return false;

	if(!onShowWindow(CLOSE_WINDOW, retCode))
	{
		#ifdef DEBUG
		printf("GuiWindow(%p)::closeWindow: postponed\n", this);
		#endif
		return true;
	}

	returnCode = retCode;
	s_window_quit(wndHandle);

	onPostShowWindow(CLOSE_WINDOW, retCode);

	#ifdef DEBUG
	printf("GuiWindow(%p)::closeWindow: done\n", this);
	#endif

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Wait for window closure
 *
 * Blocks until the window is closed.

 * For waiting for another window, declare it as modal/child
 * and wait for child_close event.
 *
 * \note Do not lock eXynth mutex!
 * \note Do not use from window event thread!
 *
 * \return Returns true after the window is closed.
 */
bool GuiWindow::waitForClose()
{
	return (s_window_wait(wndHandle) == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Add a new frame to the window
 *
 * Inserts a pointer to the specified frame to the end of the frame list.
 * \param pItem Pointer to the frame object that should be inserted
 *
 * \return Returns the number of contained frames after insertion.
 */
int GuiWindow::addFrame(GuiFrame *pItem)
{
	frameList.push_back(pItem);
	int ret = frameList.size();

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get frame by identifier
 *
 * Retrieves a GUI frame object given a unique frame identifier.
 *
 * \param idFrame Specifies the frame identifier.
 * \return Returns a pointer to the GUI frame object. Returns NULL if no frames
 * with the specified identifier were found.
 */
GuiFrame* GuiWindow::getFrame(int idFrame)
{
	GuiFrame* pRet = NULL;

	list<GuiFrame*>::iterator it;
	for(it = frameList.begin(); it != frameList.end(); it ++)
	{
		if(((GuiFrame*)*it)->getResId() == idFrame)
		{
			pRet = *it;
			break;
		}
	}

	return pRet;
}
////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Find primitive with touch area at {x,y}
 *
 * Retrieves the topmost primitive with touch area at {x,y} relative to window
 * top left corner.
 *
 * \param x X coordinate
 * \param y Y coordinate
 * \return Returns the frame ID and primitive ID as struct TouchItem. If not
 * found, returns {RES_ID_INVALID, RES_ID_INVALID}.
 */
TouchItem GuiWindow::findTouchItem(int x, int y)
{
	list<GuiFrame*>::reverse_iterator itFrame;
	for(itFrame = frameList.rbegin(); itFrame != frameList.rend(); itFrame ++)
	{
		GuiFrame* pFrame = (GuiFrame*)*itFrame;
		if (!pFrame->isVisible) continue;	// skip invisible frames

		int idFrame = pFrame->getResId();
		int idPrim = pFrame->getPrimitiveTouch(x, y);

		if (idPrim != RES_ID_INVALID) {
			TouchItem ret = {idFrame, idPrim};
			return ret;
		}
	}

	TouchItem ret = {RES_ID_INVALID, RES_ID_INVALID};
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Remove frame
 *
 * Deletes the first frame with the specified identifier from the list.
 * The frame resources and all contained primitives are released automatically.
 *
 * \param idFrame Specifies the identifier of the frame to be deleted
 * \return Returns true if a frame was found and deleted. Returns false if no
 * frames with the given identifier were found.
 */
bool GuiWindow::removeFrame(int idFrame)
{
	list<GuiFrame*>::iterator it;
	for(it = frameList.begin(); it != frameList.end(); it ++)
	{
		GuiFrame* pItem = (GuiFrame*)*it;
		if(pItem->getResId() == idFrame)
		{
			frameList.erase(it);
			delete pItem;
			return true;
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialization event handler
 *
 * Called when the window is being initialized. Classes that are derived from this base class
 * should override this method to provide additional initialization functionality.
 */
void GuiWindow::onInitialize()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Painting event handler
 *
 * Called when the window graphical content requires redrawing. The implementation traverses the list of
 * contained GUI frames and performs their repainting to the window surface,
 * given the coordinates of a rectangular drawing area.
 *
 * \param paintArea Reference to the rectangle structure that contains
 * coordinates of the area that needs drawing.
 */
void GuiWindow::onPaint(s_rect_t &paintArea)
{
	list<GuiFrame*>::iterator itFrame;

	#ifdef DEBUG
	clock_t start = clock();
	#endif

	for(itFrame = frameList.begin(); itFrame != frameList.end(); itFrame ++)
	{
		GuiFrame* pFrame = (GuiFrame*)*itFrame;
		pFrame->onPaint(wndSurface, paintArea);
	}

	// If we have some kind of periodic full refresh (ie window moving effect in progress),
	// we can disable the refreshing for every minor change and synchronize the
	// display by that
	if (refreshRealSurface && wndHandle != NULL && isReady) {
		// finally, copy the virtual surface data onto the window "real" surface
		s_rect_t paintRect = { paintArea.x + windowPos.x, paintArea.y + windowPos.y,
				paintArea.w, paintArea.h };
		s_surface_changed(wndHandle, &paintRect, 0);
	}

	#ifdef DEBUG
	printf("GuiWindow(%p)::onPaint: drawn (%d, %d, %d, %d) in %d ms\n", this,
			paintArea.x, paintArea.y, paintArea.w, paintArea.h,
			(int)((clock() - start) / (CLOCKS_PER_SEC / 1000)));
	#endif
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch down event handler
 *
 * Implements the default behavior for
 * the touch down event. The frame list is traversed and the first primitive
 * found with a matching touchable region is updated with the new selection
 * state.
 *
 * User windows may need to override this event handler to implement different
 * handling.
 *
 * \param x Contains the X coordinate of the touched point
 * \param y Contains the Y coordinate of the touched point
 */
void GuiWindow::onTouchDown(int x, int y)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch up event handler
 *
 * Implements the default behavior for
 * the touch up event. If the pointer is released within the same primitive
 * area that it was pressed in, the onTouchItem() event handler is called.
 * The state of current selection is also maintained.
 *
 * User windows may need to override this event handler to implement different
 * handling.
 *
 * \param x Contains the X coordinate of the released point
 * \param y Contains the Y coordinate of the released point
 */
void GuiWindow::onTouchUp(int x, int y)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch move event handler
 *
 * Called when a touch move event is
 * received. User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param x Contains the X coordinate of the current touch point
 * \param y Contains the Y coordinate of the current touch point
 */

void GuiWindow::onTouchMove(int x, int y)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch enter event handler
 *
 * Called when a touch enter event is
 * received. User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param x Contains the X coordinate of the touch point
 * \param y Contains the Y coordinate of the touch point
 */

void GuiWindow::onTouchEntered(int x, int y)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch primitive event handler
 *
 * Called when a specific primitive is
 * touched. The primitive must be pressed and released for this event to be
 * generated. User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param item Contains the descriptor of a GUI primitive that was touched
 */

void GuiWindow::onTouchItem(TouchItem item)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Press primitive event handler
 *
 * Called when a specific primitive is
 * pressed. User windows may need to override this event handler to implement
 * appropriate processing. Press events occur when the touchscreen is pressed
 * at the item location or when the touch is pressed at the item, moved out
 * and back.
 *
 * \param item Contains the descriptor of a GUI primitive that was pressed
 */

void GuiWindow::onPressItem(TouchItem item)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Release primitive event handler
 *
 * Called when a specific primitive is
 * released. User windows may need to override this event handler to implement
 * appropriate processing. Release events occur when the touchscreen is released
 * or when the touch is moved out of the item touch area.
 *
 * \param item Contains the descriptor of a GUI primitive that was released
 */

void GuiWindow::onReleaseItem(TouchItem item)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch exit event handler
 *
 * Called when a touch exit event is
 * received. User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param x Contains the X coordinate of the touch point
 * \param y Contains the Y coordinate of the touch point
 */
void GuiWindow::onTouchExited(int x, int y)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Key press event handler
 *
 * Called when a key down event is
 * received. User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param keyCode Specifies the identifier of the pressed key
 * \return Returns true if the event was handled; there may be more handlers in
 * the chain that are not called when true was returned
 */
bool GuiWindow::onKeyDown(int keyCode)
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Key release event handler
 *
 * Called when a key up event is
 * received. User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param keyCode Specifies the identifier of the released key
 * \return Returns true if the event was handled; there may be more handlers in
 * the chain that are not called when true was returned
 */
bool GuiWindow::onKeyUp(int keyCode)
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Window closure event handler
 *
 * Called after the window is closed.
 * The event handler loop is about to be terminated, but all frames and
 * primitives are still available for access.
 *
 * User windows may need to override this event handler to implement
 * appropriate processing.
 */
void GuiWindow::onClose()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Window visibility event handler
 *
 * Called before the window is shown, hidden, or closed.
 *
 * User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param type Specifies the type of new window visibility
 * \param extCode Specifies the extended code. Extended code is determined
 * only when type is CLOSE_WINDOW and contains the return code
 *
 * \return If true is returned by the handler, the caller proceeds with its
 * show, hide or closure operation. In case of false, the operation is aborted.
 */
bool GuiWindow::onShowWindow(SHOW_WINDOW_TYPE type, int extCode)
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Window visibility event handler
 *
 * Called after the window is shown, hidden, or closed.
 *
 * User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param type Specifies the type of new window visibility
 * \param extCode Specifies the extended code. Extended code is determined
 * only when type is CLOSE_WINDOW and contains the return code
 */
void GuiWindow::onPostShowWindow(SHOW_WINDOW_TYPE type, int extCode)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Modal window termination event handler
 *
 * Called when a child modal window is closed. The data of the child modal window is available for access.
 *
 * User windows may need to override this event handler to implement
 * appropriate processing.
 *
 * \param pModal Pointer to the child modal window being closed.
 * \return Return true if the parent window should be re-activated
 */

bool GuiWindow::onModalFinished(GuiWindow *pModal)
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Window stack change event
 *
 * Called when window stack changes (new window is added, window order is changed or
 * a window is removed)
 * Window must be created with WINDOW_MONITOR type to enable this event.
 *
 * User windows may need to override this event handler to implement
 * appropriate processing.
 */

void GuiWindow::onWindowStackChange()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Main event handler procedure
 *
 * Implements the window event handler procedure. Given static function is
 * called by the eXynth engine when a window receives a new event.
 *
 * \param window Pointer to the window descriptor that receives an event
 * \param event Pointer to the event descriptor
 */
void GuiWindow::eventHandlerProc(s_window_t *window, s_event_t *event)
{
	GuiWindow *pWindow = (GuiWindow*)window->data;
	if (pWindow == NULL) return;

	pWindow->eventHandler(event);

	if (event->type & CLOSED_EVENT) {
		bool selfDestruct = pWindow->selfDestruct;
		if(selfDestruct) delete pWindow;
	}
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Main event handler
 *
 * Decodes event flags and calls respective handlers
 *
 * \param event Pointer to the event descriptor
 */
void GuiWindow::eventHandler(s_event_t *event)
{
	if (event->type & CLOSED_EVENT) {
		isReady = false;
		// Nothing should access xynth resources here anymore.
		// All window class functions are still available until
		// the window object is destructed.
		onClose();
		// set window handle to null as xynth un-initializes everything after
		// we return.
		wndHandle = NULL;
		// pWindow may be self-destructed as we return from here
	}

	if (event->type & TIMER_EVENT) timerHandler(event);

	if (event->type & MONITOR_EVENT) onWindowStackChange();

	if (event->type & CHCLOSED_EVENT)
	{
		GuiWindow *pChild = (GuiWindow*)event->data.child->data;
		if (pChild) onModalFinished(pChild);
	}
	if (event->type & SYSTEM_EVENT)
	{
		onSystemEvent(event->data.user.code, event->data.user.data);
	}
	if (event->type & USER_EVENT)
	{
		onUserEvent(event->data.user.code, event->data.user.data);
	}

	if(isEnabled)
	{
		if (event->type & MOUSE_EVENT) {
			touchHandler(event);

			// Is the click in this window?
			// If the previous click was in this window, we get another notification
			// with MOUSE_EXITED to un-select all controls that were left selected etc.
			// We better not over-react.
			if (event->data.mouse.x > windowPos.x &&
					event->data.mouse.x < windowPos.x + windowPos.w &&
					event->data.mouse.y > windowPos.y &&
					event->data.mouse.y < windowPos.y + windowPos.h)
			{
				if (event->type & MOUSE_OVER) {
					onTouchMove(event->data.mouse.x, event->data.mouse.y);
				}

				if (event->type & MOUSE_PRESSED) {
					onTouchDown(event->data.mouse.x, event->data.mouse.y);
				}

				if (event->type & MOUSE_RELEASED) {
					onTouchUp(event->data.mouse.x, event->data.mouse.y);
				}

				if (event->type & MOUSE_ENTERED) {
					onTouchEntered(event->data.mouse.x, event->data.mouse.y);
				}

				if (event->type & MOUSE_EXITED) {
					onTouchExited(event->data.mouse.x, event->data.mouse.y);
				}
			}
			else
			{
				// input windows still wants to receive mouse events
				if(wndStyle & GUI_STYLE_INPUT)
				{
					if (event->type & MOUSE_OVER) {
						onTouchMove(event->data.mouse.x, event->data.mouse.y);
					}

					if (event->type & MOUSE_PRESSED) {
						onTouchDown(event->data.mouse.x, event->data.mouse.y);
					}

					if (event->type & MOUSE_RELEASED) {
						onTouchUp(event->data.mouse.x, event->data.mouse.y);
					}
				}
			}
		}

		if (event->type & KEYBD_EVENT) {
			if(event->type & KEYBD_PRESSED)
			{
				if (onKeyDown(event->data.keyboard.keycode))
					event->data.keyboard.handled = 1;
			}

			if(event->type & KEYBD_RELEASED)
			{
				if (onKeyUp(event->data.keyboard.keycode))
					event->data.keyboard.handled = 1;
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch event handler
 *
 * Implements default behavior for the touch operations on controls.
 *
 * \param event Pointer to the eXynth event structure
 */
void GuiWindow::touchHandler(s_event_t *event)
{
	int x = event->data.mouse.x - windowPos.x;
	int y = event->data.mouse.y - windowPos.y;

	bool onWin = x >= 0 && x < windowPos.w &&
				y >= 0 && y < windowPos.h;

	// on MOUSE_PRESSED within the window, we lock on to
	// another primitive
	if (onWin && (event->type & MOUSE_PRESSED)) {
		if (touchItem.idPrimitive != RES_ID_INVALID) {
			GuiFrame *pFrame = getFrame(touchItem.idFrame);
			if (pFrame) {
				int notify = pFrame->untouch(touchItem);
				notifyTouch(notify, touchItem);
			}
		}
		touchItem = findTouchItem(x,y);
	}

	// send the touch event to the primitive's frame for
	// further processing
	if (touchItem.idPrimitive == RES_ID_INVALID) return;
	GuiFrame *pFrame = getFrame(touchItem.idFrame);
	if (!pFrame) return;

	int notify = pFrame->touch(touchItem, event->type, x, y, onWin);
	TouchItem notifyItem = touchItem;

	// if this was a release event, unlock from this primitive
	if (event->type & MOUSE_RELEASED)
		touchItem.idPrimitive = RES_ID_INVALID;

	// Notify the application if required
	notifyTouch(notify, notifyItem);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Touch notification handler
 *
 * Forwards touch events to user application
 *
 * \param notify Bitmask of events to notify of
 * \param item Item that is the subject of the notification
 */
void GuiWindow::notifyTouch(int notify, TouchItem item)
{
	if (notify & GuiFrame::NOTIFY_TOUCH) onTouchItem(item);
	if (notify & GuiFrame::NOTIFY_PRESS) onPressItem(item);
	if (notify & GuiFrame::NOTIFY_RELEASE) onReleaseItem(item);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start a new timer
 *
 * Starts a new window timer that will fire every 'ms' milliseconds and call
 * the timer event handler.
 * If such timer is already running, false is returned. Starting timer is
 * thread-safe from any thread, no additional locking is required.
 *
 * \param ms Timer period in milliseconds
 * \param idTimer Unique identifier of the timer. This value is passed to a
 * timer event handler and may be used by the caller.
 *
 * \return Returns true when timer was started successfully, false otherwise.
 */

bool GuiWindow::startTimer(int ms, int idTimer)
{
	if(wndHandle == NULL || !isReady) return false;

	#ifdef DEBUG_TIMER
	printf("GuiWindow(%p)::startTimer: starting timer %d, %d ms\n", this, idTimer, ms);
	#endif

	// is timer already running?
	if (timerMap.find(idTimer) != timerMap.end()) {
		return false;
	}

	// initialize new timer
	TIMER_DATA timer;
	timer.pTimer = NULL;
	if (s_timer_init(&timer.pTimer) != 0) {
		return false;
	}

	timer.pTimer->data = (void*)idTimer;
	timer.pTimer->timeval = (long long)ms;
	timer.pTimer->interval = (long long)ms;
	timer.pTimer->window = wndHandle;

	// register the timer in xynth
	if(s_timer_add(timer.pTimer) < 0)
	{
		#ifdef DEBUG_TIMER
		printf("GuiWindow(%p)::startTimer: failed adding timer\n", this);
		#endif

		s_timer_uninit(timer.pTimer);
		return false;
	}

	// add to map
	timerMap[idTimer] = timer;

	#ifdef DEBUG_TIMER
	printf("GuiWindow(%p)::startTimer: done\n", this);
	#endif
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop an existing timer
 *
 * Terminates the given timer.
 * \param idTimer Unique identifier of the timer that should be stopped
 *
 * \return Returns true if the timer was successfully stopped and timer
 * resourced were released.
 */
bool GuiWindow::stopTimer(int idTimer)
{
	if(wndHandle == NULL || !isReady) {
		return false;
	}

	#ifdef DEBUG_TIMER
	printf("GuiWindow(%p)::stopTimer: stopping timer %d\n", this, idTimer);
	#endif

	// is this timer running?
	if (timerMap.find(idTimer) == timerMap.end()) {
		return false;
	}

	TIMER_DATA timer = timerMap[idTimer];

	// un-register in xynth
	if(s_timer_del(timer.pTimer) < 0)
	{
		#ifdef DEBUG_TIMER
		printf("GuiWindow(%p)::stopTimer: failed removing timer\n", this);
		#endif
		return false;
	}

	// erase from map
	timerMap.erase(idTimer);

	// release memory
	if(s_timer_uninit(timer.pTimer) != 0)
	{
		#ifdef DEBUG_TIMER
		printf("GuiWindow(%p)::stopTimer: failed releasing timer\n", this);
		#endif
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Timer handler
 *
 * Sub-handler for timer events. The procedure further
 * calls an appropriate timer event handler of the window or an associated primitive.
 *
 * \param event Pointer to event structure
 */
void GuiWindow::timerHandler(s_event_t *event)
{
	#ifdef DEBUG_TIMER
	printf("GuiWindow::timerHandler\n");
	#endif

	int idTimer = (int)event->data.timer->data;

	// is this timer running?
	if (timerMap.find(idTimer) == timerMap.end()) {
		return;
	}

	onTimer(idTimer);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Timer event handler
 *
 * Called when a timer associated with a window is fired.
 *
 * \param idTimer Specifies the identifier of the timer that was set during
 * creation
 */
void GuiWindow::onTimer(int idTimer)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief System event handler
 *
 * Called when a global system event is received.
 *
 * \param eventCode Specifies the identifier of the system event
 * \param data Specifies pointer to data that is broadcasted with the event;
 * 	do not free this pointer!
 */
void GuiWindow::onSystemEvent(int eventCode, void *data)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief User event handler
 *
 * Called when a user event is received.
 *
 * \param eventCode Specifies the identifier of the user event
 * \param data Specifies pointer to data that is broadcasted with the event;
 * 	do not free this pointer!
 */
void GuiWindow::onUserEvent(int eventCode, void *data)
{
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Submit a key event to the queue
 *
 * Submits a simulated key event, that gets distributed to
 * windows. Either key press or release events may be submitted. An appropriate
 * onKeyXXX() event handler will be called after the event is dispatched.
 *
 * \param keyCode Specifies the key identifier
 * \param isPressed Specifies whether the key is pressed or released
 * \return Returns true if the key event was sent successfully
 */
bool GuiWindow::submitKeyEvent(int keyCode, bool isPressed)
{
	input_keyboard_parse_submit(keyCode, (isPressed)?1:0);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Find a primitive
 *
 * Searches for the primitive with the specified identifiers.
 *
 * \param item Specifies the frame and primitive identifiers
 * \return Returns a pointer to the primitive object or NULL if no primitive
 * is found.
 */
GuiPrimitive* GuiWindow::findPrimitive(TouchItem item)
{
	GuiFrame *pFrame = getFrame(item.idFrame);
	if (!pFrame) {
		return NULL;
	}
	GuiPrimitive* pPrim = pFrame->getPrimitive(item.idPrimitive);
	return pPrim;
}
