#ifndef GUIWINDOW_H_
#define GUIWINDOW_H_

#include "GuiSurface.h"
#include "GuiFrame.h"
#include "GuiPrimitive.h"

#include <exynth/exynth.h>
#include <list>
#include <map>
using std::list;
using std::map;

#define GUI_MAX_GROUPS		4

#define GUI_STYLE_MAIN		0x01
#define GUI_STYLE_CHILD		0x02
#define GUI_STYLE_MODAL		0x04
#define GUI_STYLE_INPUT		0x80
#define GUI_STYLE_MONITOR	0x40

class GuiFrame;
class GuiPrimitive;
class GuiWindow;

/** Contains data of a timer belonging to this window object. */
struct TIMER_DATA
{
	/** Pointer to the eXynth timer structure */
	s_timer_t *pTimer;
};

/**
 * \brief Implements a generic GUI window
 *
 * Implements a generic window, being the basic building block of
 * the GUI. The window can be also viewed as a wrapper for the s_window_t type
 * of the eXynth engine and its related API.
 *
 * The window class is a container for GUI frames, that in turn are containers
 * for window primitives. The content of the window may be automatically
 * initialized by means of the skin function provided by GuiTemplate class.
 * In such case, the contained frames and related primitives will be instantiated
 * automatically and inserted into the window.
 *
 * The window class implements timers (see startTimer(), stopTimer()
 * methods) by wrapping eXynth s_timer_t and its related API.
 * Timer ID's:
 *  * 0-499:    guiOO internal (effectwindow etc)
 *  * 500-999:  guiOO internal primitives (animation etc)
 *  * 1000+:    user-allocation
 *
 * User should implement GUI windows by creating classes inherited from
 * GuiWindow (or its derivatives, GuiEffectWindow and GuiScrollWindow) and
 * implement required event handlers.
 *
 * XML code below defines a window:
 *
 * \code
 *	<window id="ID" x="X" y="Y" width="W" height="H">
 * 		<!-- frames are defined here -->
 * 		<frame ...>
 * 			<!-- primitives are defined here -->
 * 		</frame>
 * 	</window>
 * \endcode
 *
 * Where
 * - ID represents the unique identifier of the window
 * - X, Y, W, H is the set of coordinates of the window on the screen
 *
  * \author Andrei Birjukov, Anti Sullin
 */

class GuiWindow
{
public:

	/** Window visibility identifier that identifies window show/hide operation */
	enum SHOW_WINDOW_TYPE
	{
		/** The window is about to be hidden */
		HIDE_WINDOW,
		/** The window is about to be displayed */
		SHOW_WINDOW,
		/** The window is about to be closed */
		CLOSE_WINDOW
	};

	GuiWindow(int idResource);
	virtual ~GuiWindow();

	bool createWindow(int style, GuiWindow *pParent=NULL);
	bool setWindowPos(s_rect_t &wndPos);
	virtual bool showWindow(bool state);
	bool setAlwaysOnTop(bool flag);
	bool setEnabled(bool flag);
	bool redrawWindow();
	bool redrawWindow(s_rect_t &updateArea);
	virtual bool closeWindow(int retCode=0);
//	bool startModal(GuiWindow* pChild);
	bool waitForClose();

	/**
	 * \brief Get the unique ID of the window
	 * \return Returns the unique identifier of the window object
	 */
	int getResId() { return idResource; }

	/**
	 * \brief Get the window surface pointer
	 * \return Returns the pointer to the window surface object
	 */
	GuiSurface* getSurface() { return wndSurface; }

	/**
	 * \brief Get the window handle
	 * \return Returns pointer to the window handle associated with the window object
	 */
	s_window_t* getHandle() { return wndHandle; }

	/**
	 * \brief Get the window return code
	 * \return Returns the return code of the window object
	 */
	int getReturnCode() { return returnCode; }

	/**
	 * \brief Get always on top flag
	 * \return Returns true, if the window always on top flag is set, false otherwise
	 */
	bool getAlwaysOnTop() { return isAlwaysOnTop; }

	/**
	 * \brief Get enabled flag
	 * \return Returns true, if the window state is enabled, false otherwise
	 */
	bool getEnabled() { return isEnabled; }

	int addFrame(GuiFrame *pItem);
	GuiFrame* getFrame(int idFrame);
	bool removeFrame(int idFrame);

	/**
	 * \brief Get the number of frames in window
	 * \return Returns the number of frames associated with this window object
	 */
	int getNumFrames() { return frameList.size(); }

	/**
	 * \brief Get the specific primitive in a frame contained in window
	 *
	 * Get the primitive from one of the frames that are associated with the window
	 * \param idFrame Unique identifier of the frame where the primitive is
	 * \param idPrimitive Unique identifier of the primitive to be retrieved
	 * \return Returns pointer to the primitive or NULL if the primitive was not found
	 */
	template <class P>
	P* getPrimitive(int idFrame, int idPrimitive) {
		GuiFrame *frm = getFrame(idFrame);
		if (!frm) return NULL;
		GuiPrimitive *prm = frm->getPrimitive(idPrimitive);
		if (!prm) return NULL;
		return dynamic_cast<P*>(prm);
	}

	bool startTimer(int ms, int idTimer);
	bool stopTimer(int idTimer);

	static bool submitKeyEvent(int keyCode, bool isPressed);

	virtual void onInitialize();
	virtual void onPaint(s_rect_t &paintArea);
	virtual void onTouchDown(int x, int y);
	virtual void onTouchUp(int x, int y);
	virtual void onTouchMove(int x, int y);
	virtual void onTouchItem(TouchItem item);
	virtual void onPressItem(TouchItem item);
	virtual void onReleaseItem(TouchItem item);
	virtual void onTouchEntered(int x, int y);
	virtual void onTouchExited(int x, int y);
	virtual bool onKeyDown(int keyCode);
	virtual bool onKeyUp(int keyCode);
	virtual void onClose();
	virtual bool onShowWindow(SHOW_WINDOW_TYPE type, int extCode);
	virtual void onPostShowWindow(SHOW_WINDOW_TYPE type, int extCode);
	virtual bool onModalFinished(GuiWindow *pModal);
	virtual void onTimer(int idTimer);
	virtual void onWindowStackChange();
	virtual void onSystemEvent(int eventCode, void *data);
	virtual void onUserEvent(int eventCode, void *data);

protected:
	static void eventHandlerProc(s_window_t *window, s_event_t *event);
	virtual void eventHandler(s_event_t *event);

	void touchHandler(s_event_t *event);
	void timerHandler(s_event_t *event);

	virtual void notifyTouch(int notify, TouchItem item);
	GuiPrimitive* findPrimitive(TouchItem dPrim);
	TouchItem findTouchItem(int x, int y);

	/** Window unique identifier, set only in constructor */
	int					idResource;

	/** Pointer to the window descriptor (Xynth structure), locked by window mutex */
	s_window_t*			wndHandle;

	/** Pointer to the window surface object, locked by window mutex */
	GuiSurface*			wndSurface;

	/** List of contained window frames, protected by window mutex */
	list<GuiFrame*>		frameList;

	/** Map of active timers, protected by window mutex */
	map<int, TIMER_DATA> timerMap;

	/** Position of the window in relation to the upper left corner of the screen, protected by window mutex */
	s_rect_t			windowPos;
	/** Specifies currently touched primitive, used only in window event thread */
	TouchItem			touchItem;

	/** Pointer to a parent window object */
	GuiWindow*			wndParent;

	/** Specifies if the window is enabled, i.e. ready to receive user input */
	bool				isEnabled;
	/** Specifies if the window is shutting down after the closeWindow() was called */
	bool				isReady;
	/** Specifies if the window is always on top */
	bool				isAlwaysOnTop;
	/** Specifies if the window performs self destruction after closure */
	bool				selfDestruct;
	/** Specifies if every redraw should be followed by expose automatically */
	bool				refreshRealSurface;
	/** Specifies if onInitialize has been run */
	bool				isInitialized;

	/** Contains window style flags */
	int					wndStyle;
	/** Contains window return code */
	int					returnCode;
};

#endif /*GUIWINDOW_H_*/
