#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <termios.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/poll.h>

#include <platform.h>

#include "gsm.h"

//#define DEBUG

#ifdef DEBUG
#define debug		printf
#else
inline static void debug(char* fmt, ...) {}
#endif

static int powerDriver = -1;
static int modemPort = -1;

static char mainBuf[1024];
int mainBufCnt=0;
//static char lineBuf[512];

static char *modemResponseCodes[] =
{
	"\nOK\r\n",
	"\nCONNECT\r\n",
	"\nRING\r\n",
	"\nNO CARRIER\r\n",
	"\nERROR\r\n",
	"\nNO ANSWER\r\n",
	"\n+CMS ERROR:",
	"\n+CME ERROR:",
	NULL
};
static int modemResponseSizes[] = { 5, 10, 7, 13, 8, 12, 12, 12, 0 };

////////////////////////////////////////////////////////////////////////////////
//
int modemInit(const char* portName)
{
	struct termios attr;
	long ioflags;

	powerDriver = open(POWER_DEVICE_NAME, O_RDWR | O_NONBLOCK);
	if(powerDriver == -1)
	{
		debug("Failed opening power control driver\n");
		return -1;
	}

	modemPort = open(portName, O_RDWR | O_NDELAY);
	if(modemPort == -1)
	{
		debug("Failed opening modem port %s\n", portName);
		return -2;
	}

	ioflags = fcntl(modemPort, F_GETFL);
	tcgetattr(modemPort, &attr);

	attr.c_oflag = attr.c_lflag = 0;
	attr.c_cflag = CLOCAL | CREAD | HUPCL;
	attr.c_iflag = IGNBRK;

	memset(attr.c_cc, 0, sizeof(attr.c_cc));
	attr.c_cc[VMIN] = 1;

	attr.c_cflag |= (CRTSCTS | CSIZE | CS8 | CSTOPB);
	attr.c_iflag |= (IXON | IXANY | IXOFF);

	cfsetispeed(&attr, B115200);
	cfsetospeed(&attr, B115200);

	if(tcsetattr(modemPort, TCSANOW, &attr))
	{
		debug("Failed setting port config\n");
		return -3;
	}

	return fcntl(modemPort, F_SETFL, ioflags & ~O_NDELAY);
}

////////////////////////////////////////////////////////////////////////////////
//
int modemClose()
{
	if(close(powerDriver)) return -1;
	if(close(modemPort)) return -2;
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemPowerUp()
{
	int ret;
	int i;

	// enable serial port
	ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_OFF, 1);
	if (ret != 0) return -1;

	ret = ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD);
	if(ret == 1) return 0;

	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_PWR);
	usleep(100*1000);

	for (i=0; i<5; i++) {
		// generate a power-on pulse
		ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_ON);
		usleep(1000*1000);
		ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_ON);
		debug("modemPowerUp: ON low\n");

		// control the power state
		ret = ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD);
		debug("modemPowerUp: state %d\n", ret);

		if(ret == 1) return 0;

		usleep(100*1000);
	}

	debug("Failed powering up the GSM module\n");
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemPowerDown()
{
	int i, ret;

	ret = ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD);
	if(ret == 0) return 0;

	// disable serial port
	ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_ON, 1);
	if (ret != 0) return -1;

	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_ON);
	usleep(1500*1000);
	ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_ON);

	for(i=0; i<200; i++)
	{
		if(ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD) == 0)
		{
			ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_PWR);
			return 0;
		}

		usleep(50*1000);
	}

	ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_PWR);
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemSendCommand(const char* pCmd)
{
	usleep(20*1000);	// force delay between commands
	tcflush(modemPort, TCIOFLUSH);

	write(modemPort, (void*)pCmd, strlen(pCmd));
	debug("modemSendCommand: [%s]\n", pCmd);
	return write(modemPort, (void*)"\r", 1);
}

////////////////////////////////////////////////////////////////////////////////
//
int modemReadAnswer(char *pAnswer, int answerSize, int timeOut)
{
	int bytes = 0;
	struct pollfd pfd;

	pfd.fd = modemPort;
	pfd.revents = 0;
	pfd.events = POLLIN;

	mainBufCnt = 0;

	while(1)
	{
		if(poll(&pfd, 1, timeOut) < 0)
		{
			debug("Error polling port for input data\n");
			return AT_IO_FAIL;
		}

		if(!(pfd.revents && pfd.events)) return AT_TIMEOUT;
		timeOut = 500; // next cycle with 500ms timeout

   		ioctl(modemPort, FIONREAD, &bytes);
   		if(bytes > sizeof(mainBuf) - mainBufCnt - 1) bytes = sizeof(mainBuf) - mainBufCnt - 1;

   		bytes = read(modemPort, (void*)mainBuf+mainBufCnt, bytes);
		if(bytes < 0) break;

		mainBufCnt += bytes;
		mainBuf[mainBufCnt] = 0;

		debug("modemReadAnswer: [%s]\n", mainBuf);

		int valid = AT_PARTIAL, n;
		for(n = 0; modemResponseCodes[n] != NULL; n ++)
		{
			if(mainBufCnt >= modemResponseSizes[n] &&
				!strcmp(mainBuf + mainBufCnt - modemResponseSizes[n], modemResponseCodes[n]))
			{
				valid = n;
				break;
			}
		}

		if(valid != AT_PARTIAL)
		{
			if(pAnswer && answerSize)
			{
				*pAnswer = 0;

				if (mainBufCnt-2+1 <= answerSize)
					memcpy(pAnswer, mainBuf+2, mainBufCnt-2+1);
			}

			return valid;
		}
	}

	debug("Error reading from port\n");
	return AT_IO_FAIL;
}

////////////////////////////////////////////////////////////////////////////////
//
#if 0
int modemReadLine(char *pAnswer, int timeOut)
{
	int bytes = 0, total = 0;
	struct pollfd pfd;

	const char *answerLF = "\r\n";
	const int lenLF = strlen(answerLF);

	pfd.fd = modemPort;
	pfd.revents = 0;
	pfd.events = POLLIN;

	memset(mainBuf, 0, sizeof(mainBuf));

	while(1)
	{
		if(ioctl(modemPort, FIONREAD, &bytes) == 0)
		{
			if(bytes == 0)
			{
				if(poll(&pfd, 1, timeOut) < 0)
				{
					debug("Error polling port for input data\n");
					return AT_IO_FAIL;
				}

				if(!(pfd.revents && pfd.events)) return AT_TIMEOUT;
			}
		}

		ioctl(modemPort, FIONREAD, &bytes);
		if(bytes > sizeof(lineBuf)) return AT_SMALL_BUFFER;

		memset(lineBuf, 0, sizeof(lineBuf));
		bytes = read(modemPort, (void*)lineBuf, bytes);
		if(bytes < 0) break;

		debug("modemReadLine: [%s]\n", lineBuf);

		strcat(mainBuf, lineBuf);
		total += bytes;

		int valid = AT_PARTIAL, n;
		for(n = 0; modemResponseCodes[n] != NULL; n ++)
		{
			if(total >= modemResponseSizes[n] &&
				!strcmp(mainBuf + total - modemResponseSizes[n], modemResponseCodes[n]))
			{
				valid = n;
				break;
			}
		}

		if(valid != AT_PARTIAL)
		{
			if(pAnswer)
			{
				strncpy(pAnswer, mainBuf, total - lenLF);
				pAnswer[total - lenLF] = 0;
			}
			return valid;
		}

		if(total >= lenLF &&
				!strcmp(&mainBuf[total - lenLF], answerLF))
		{
			if(pAnswer)
			{
				strncpy(pAnswer, mainBuf, total - lenLF);
				pAnswer[total - lenLF] = 0;
			}
			return AT_PARTIAL;
		}
	}

	debug("Error reading from port\n");
	return AT_IO_FAIL;
}
#endif
////////////////////////////////////////////////////////////////////////////////
//
int modemHandShake()
{
	int i;
	for(i=0; i<25; i++)
	{
		modemSendCommand("AT");
		if(modemReadAnswer(NULL, 0, 250) == AT_OK)
		{
			modemSendCommand("AT+IPR=115200");
			if(modemReadAnswer(NULL, 0, 500) == AT_OK)
			{
				modemSendCommand("ATE0");
				return modemReadAnswer(NULL, 0, 500) == AT_OK ? 0 : -3;
			}
			else
			{
				debug("Failed communicating to the modem with a fixed baudrate\n");
				return -2;
			}
		}
	}

	debug("Failed establishing communication with modem\n");
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemReadRTC(struct tm* pDate)
{
	char answer[128];
	int year, mon, day, hour, min, sec, fields;

	modemSendCommand("AT+CCLK?");
	if(modemReadAnswer(answer, sizeof(answer), 1000) != AT_OK)
	{
		debug("Error reading modem RTC\n");
		return -1;
	}

	fields = sscanf(answer, "+CCLK: \"%02d/%02d/%02d,%02d:%02d:%02d\"",
		&year, &mon, &day, &hour, &min, &sec);

	if(fields != 6)
	{
		debug("Error parsing modem RTC answer\n");
		debug("Data: %s\n", answer);
		return -2;
	}

	pDate->tm_year = year + 100;
	pDate->tm_mon = mon - 1;
	pDate->tm_mday = day;
	pDate->tm_hour = hour;
	pDate->tm_min = min;
	pDate->tm_sec = sec;

	pDate->tm_wday = 0;
	pDate->tm_yday = 0;

	pDate->tm_gmtoff = 0;
	pDate->tm_isdst = -1;
	pDate->tm_zone = NULL;
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemWriteRTC(struct tm* pDate)
{
	char str[128];
	sprintf(str, "AT+CCLK=\"%02d/%02d/%02d,%02d:%02d:%02d%c%02d\"",
		pDate->tm_year % 100, pDate->tm_mon + 1, pDate->tm_mday,
		pDate->tm_hour, pDate->tm_min, pDate->tm_sec,
		pDate->tm_gmtoff < 0 ? '-' : '+',
		abs(pDate->tm_gmtoff) / 3600);

	modemSendCommand(str);

	if(modemReadAnswer(NULL, 0, 1000) != AT_OK)
	{
		debug("Error writing modem RTC\n");
		return -1;
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemReadADC(int idADC)
{
	char str[64];
	int mVolt = 0;

	sprintf(str, "AT#ADC=%d,2", idADC);
	modemSendCommand(str);

	if(modemReadAnswer(str, sizeof(str), 1000) != AT_OK)
	{
		debug("Error querying modem ADC\n");
		return -1;
	}

	if(sscanf(str, "#ADC: %d", &mVolt) != 1)
	{
		debug("Failed parsing modem ADC reply\n");
		return -2;
	}

	return mVolt;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemSetPowerMode(int idMode)
{
	char str[64];

	sprintf(str, "AT+CFUN=%d", idMode);
	modemSendCommand(str);
	if(modemReadAnswer(str, sizeof(str), 1000) != AT_OK)
	{
		debug("Error setting modem functionality\n");
		return -1;
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//
int modemSetBand(int freqBand)
{
	char str[64];
	switch(freqBand)
	{
		case FREQ_AUTODETECT:

			modemSendCommand("AT#AUTOBND=1");
			return modemReadAnswer(NULL, 0, 1000) == AT_OK ? 0 : -1;

		case FREQ_900X1800:
		case FREQ_900X1900:
		case FREQ_850X1800:
		case FREQ_850X1900:

			sprintf(str, "AT#BND=%d", freqBand);
			modemSendCommand(str);
			return modemReadAnswer(NULL, 0, 1000) == AT_OK ? 0 : -2;
	}

	return -1;
}
