/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <platform.h>

//#define DEBUG
#include "Debug.h"

#include "exynthoo/GuiLibrary.h"
#include "LibSpatha.h"
#include "PowerManager.h"
#include "BarcodeReader.h"

#define DEFAULT_SCANNER_TIMEOUT		5000

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the barcode reader object.
 */
BarcodeReader::BarcodeReader()
{
	devHandle = -1;
	readerActive = false;
	interval = DEFAULT_SCANNER_TIMEOUT;
}

/**
 * \brief Class destructor
 *
 * Terminates the barcode reader object.
 */
BarcodeReader::~BarcodeReader()
{
	if (devHandle != -1) shutdown();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reader initialization routine
 *
 * Initializes the barcode reader. Powers up Opticon module up and sends
 * initialization commands via serial port.
 *
 * \return Returns true if initialization was successful.
 */
bool BarcodeReader::initialize()
{
	if (devHandle != -1) return false;
	devHandle = 1; // dummy

	PowerManager *pm = PowerManager::getPowerManager();

	if (!pm->scannerPowerUp())
	{
		debug("Failed powering up the scanner\n");
		return false;
	}

	debug("Barcode scanner initialized\n");

	if (!SerialPort::open(BARCODE_SCANNER_DEVICE))
	{
		debug("Failed opening barcode scanner port: %s\n",
				BARCODE_SCANNER_DEVICE);
		return false;
	}

	setSpeed(9600);
	setCharBits(8);
	setParity(S_PARITY_NONE);
	setStopBits(1);
	setFlowControl(S_FLOW_NONE);

	Thread::sleep(100);

	const char *cmdTriggerCont = "\x1bYM\x0d";
	SerialPort::writeData((uint8_t*) cmdTriggerCont, strlen(cmdTriggerCont));
	SerialPort::drainOutput();

	return Thread::start();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reader shutdown routine
 *
 * Deinitializes barcode reader. Powers down Opticon reader.
 */
void BarcodeReader::shutdown()
{
	stopTimer();

	SerialPort::close();
	devHandle = -1;
	Thread::stop();

	PowerManager *pm = PowerManager::getPowerManager();
	pm->scannerPowerDown();

	debug("Barcode scanner terminated\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Prepare for reading
 *
 * Prepares the Opticon module for reading, enables aiming mode. When aiming,
 * the laser beam is enabled and focused to a single point.
 */
void BarcodeReader::initRead()
{
	PowerManager *pPM = PowerManager::getPowerManager();
	pPM->scannerSetAim(true);

	debug("Barcode scanner pointing...\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start reading
 *
 * Starts reading a barcode by triggering the Opticon module. When reading, the
 * laser beam is enabled and scans a line horizontally.
 */
void BarcodeReader::startRead()
{
	PowerManager *pPM = PowerManager::getPowerManager();
	pPM->scannerSetAim(false);

	startTimer();

	if (!readerActive)
	{
		const char *cmdTrigger = "\x1bZ\x0d";
		SerialPort::writeData((uint8_t*) cmdTrigger, strlen(cmdTrigger));
		SerialPort::drainOutput();
		readerActive = true;
		debug("Barcode scanner reading...\n");
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop reading
 *
 * Stops barcode reading. Detriggers the Opticon module and disables laser beam.
 */
void BarcodeReader::stopRead()
{
	PowerManager *pPM = PowerManager::getPowerManager();
	pPM->scannerSetAim(false);

	if (readerActive)
	{
		const char *cmdDetrigger="\x1bY\x0d";
		SerialPort::writeData((uint8_t*)cmdDetrigger, strlen(cmdDetrigger));
		SerialPort::drainOutput();
		readerActive = false;
		debug("Barcode scanner stopped.\n");
	}

	stopTimer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Timer event handler
 *
 *	Stops reading after a timeout
 */
void BarcodeReader::onTimer()
{
	stopRead();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reads the actual data
 *
 * Waits until the data becomes available from the device.
 * Then reads data from the the bar code reader and stores it in the internal buffer.
 */
void BarcodeReader::run()
{
	debug("BarcodeReader::run(): starting\n");
	char buf[2 * BARCODE_MAX_BYTES];

	while (!Thread::stopRequested())
	{
		// check if there is data available on serial
		if (!SerialPort::isInputPending(200)) continue;

		int bytesTotal = 0;
		while (true)
		{
			if (SerialPort::isInputPending(100) > 0)
			{
				int len = SerialPort::readData((uint8_t*) &buf[bytesTotal],
						BARCODE_MAX_BYTES - bytesTotal);

				debug("Read data, %d bytes\n", len);
				if (len < 0)	// non-recoverable failure
				{
					debug("error reading from port\n");
					break;
				}

				if ((len == 0) || (len > BARCODE_MAX_BYTES - bytesTotal))
				{
					// error is not fatal, so continue reading normally
					bytesTotal = 0;
					continue;
				}
				else bytesTotal += len;
			}
			else break;
		}

		if (Thread::stopRequested()) break;

		buf[bytesTotal] = 0;
		if (bytesTotal > 0)
		{
			std::string barcode = buf;
			size_t firstLoc = barcode.find_first_not_of(" \t\v\f\r\n");
			if (firstLoc == std::string::npos) continue;
			size_t lastLoc = barcode.find_last_not_of(" \t\v\f\r\n");
			if (lastLoc == std::string::npos) continue;
			if (lastLoc <= firstLoc) continue;
			const char *trimBuf = barcode.substr(firstLoc,
					lastLoc - firstLoc + 1).c_str();
			bytesTotal = strlen(trimBuf);

			memcpy(buf, trimBuf, bytesTotal);
			buf[bytesTotal] = 0;
			readerActive = false;

			debug("Barcode data read: %s\n", buf);
			submit(buf);
		}
	}

	debug("BarcodeReader::run(): finished\n");
}

/**
 * \brief Submit the data to event system
 *
 * Broadcasts the data read into the event system.
 * The EV_READER system event will be triggered.
 *
 * \param buf Pointer to the buffer that holds the data read
 */
void BarcodeReader::submit(const char *buf) {
	ReaderData data;
	data.reader = READER_BARCODE;
	data.pId = NULL;
	data.pData = buf;
	data.handled = 0;

	while (1) {
		if (GuiLibrary::timedlock(1000)) break;
		if (stopRequested()) return;
	}
	stopTimer();
	GuiLibrary::broadcastSystemEvent(EV_READER, (void*)&data);
	GuiLibrary::unlock();
}
