/*
 * BlueGigaBluetoothModule.cpp
 *
 *  Created on: 21.06.2011
 *      Author: ivari
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "BlueGigaBluetoothModule.h"
#include "PowerManager.h"
#include "BluetoothSPPConnection.h"

//#define DEBUG
//#define DEBUG_BYTES
//#define DEBUG_COMM
//#define DEBUG_MUX
#include "Debug.h"

BluetoothModule::ModemAnswers BlueGigaBluetoothModule::answers[] = {
		{ "READY.", OK },
		{ "SYNTAX ERROR", ERROR },
		{ "INQUIRY 0", ERROR },
		{ "SET CONTROL MUX 1", OK },
		{ NULL, NUL }
};

BlueGigaBluetoothModule::BlueGigaBluetoothModule(const char *portName) : BluetoothModule() {

	commPortName = portName;
	moduleEnabled = true;
	muxEnabled = false;
	btPort = NULL;
	pModule = this;
}

BlueGigaBluetoothModule::~BlueGigaBluetoothModule() {
}

bool BlueGigaBluetoothModule::shutdown() {
	return close();
}

bool BlueGigaBluetoothModule::open() {

	if(btPort) return false;

	lock();
	btPort = new BluetoothPort();
	if(!btPort->open(commPortName.c_str())) {
		delete btPort;
		btPort = NULL;
		return false;
	}

	btPort->setSpeed(115200);
	btPort->setFlowControl(S_FLOW_NONE);
	btPort->setCharBits(8);
	btPort->setStopBits(1);
	btPort->setParity(S_PARITY_NONE);
	btPort->initReader(this);

	unlock();

	debug("Opened BT port %s\n", commPortName.c_str());

	return true;
}

bool BlueGigaBluetoothModule::close() {

	lock();

	if(btPort) {
		if(!btPort->close()) {
			unlock();
			return false;
		}

		delete btPort;
		btPort = NULL;
	}

	unlock();

	return true;
}

void BlueGigaBluetoothModule::pollIncomingData(int timeOut) {
	lock();

	readAnswer(NULL, timeOut);

	unlock();
}

std::string BlueGigaBluetoothModule::getLocalMac() {
	std::string res;

	lock();

	if(writeCommand("SET", strlen("SET"))) {
		readAnswer(&res, 1000);
	}

	unlock();

	debug("%s\n", res.c_str());

	std::string token = "SET BT BDADDR ";

	size_t pos = res.find(token);
	if(pos != std::string::npos && res.length() >= pos + token.length() + 17)
		res = res.substr(pos + token.length(), 17);
	else
		res = "";

	return res;
}

bool BlueGigaBluetoothModule::disableEcho() {
	std::string cmd = "SET CONTROL ECHO 5";
	std::string res;

	lock();

	if(writeCommand(cmd.c_str(), cmd.length())) {
		readAnswer(&res, 1000);

		unlock();

		return res.length() == 0 || res.compare(cmd) == 0;
	}

	unlock();
	return false;
}

bool BlueGigaBluetoothModule::setLAP(const char* lapID) {
	std::string cmd = "SET BT LAP ";
	cmd.append(lapID);

	std::string res;

	lock();

	if(writeCommand(cmd.c_str(), cmd.length())) {
		bool ret = readAnswer(&res, 1000);
		if(ret && res.length() == cmd.length() + 2) {

			unlock();
			return true;
		}

		unlock();
		return !ret && muxEnabled;
	}

	unlock();

	return false;
}

bool BlueGigaBluetoothModule::setAuth(char *pin) {
	std::string cmd = "SET BT AUTH * ";
	cmd.append(pin);

	std::string res;

	lock();

	if(writeCommand(cmd.c_str(), cmd.length())) {
		readAnswer(&res, 1000);
		unlock();

		return res.length() == 0 || res.compare(cmd) == 0;
	}

	unlock();

	return false;
}

bool BlueGigaBluetoothModule::isAlive() {
	std::string res;
	int retryCount = 5;

	lock();

	//Check if the modem is ready
	while(retryCount > 0) {
		readAnswer(&res, 1000);
		if(getAnswerType() == BluetoothModule::OK) break;

		retryCount--;
	}

	unlock();

	return retryCount>0;
}

bool BlueGigaBluetoothModule::ping() {
	std::string res;
	char buf[16];

	for (unsigned int i = 0; i<MAX_BT_CONNECTIONS; i++) {
		if(connections[i] == NULL) continue;

		snprintf(buf, sizeof(buf), "PING %d", getLocalPort(connections[i]));
		buf[sizeof(buf)-1]='\0';

		res.clear();

		lock();

		if(writeCommand(buf, strlen(buf))) {
			readAnswer(&res, 1000);
		}

		unlock();

		if(res.length()>0) break;
	}

	return res.length()>0;
}

std::string BlueGigaBluetoothModule::getVersion() {
	std::string res;

	lock();

	if(writeCommand("INFO", strlen("INFO"))) {
		readAnswer(&res, 1000);
	}

	unlock();

	return res;
}

int BlueGigaBluetoothModule::getSignalStrength() {
	int strength = -128;
	char buf[64];
	std::string answer;


	for(unsigned int i=0; i<MAX_BT_CONNECTIONS; i++) {
		if(connections[i] == NULL || !connections[i]->isConnected()) continue;

		snprintf(buf,sizeof(buf),"RSSI %d", getLocalPort(connections[i]));
		buf[sizeof(buf)-1]='\0';

		answer.clear();
		lock();

		if(writeCommand(buf, strlen(buf))) {
			if (!readAnswer(&answer, 250)) {
				unlock();
				continue;
			}

			size_t pos = answer.rfind(" ");
			if(pos == std::string::npos) {
				//Answer is empty - connection is dropped. Close socket
				unlock();

				//setConnectionState(connections[i], SPP_DISCONNECTED);
				continue;
			}

			int value = atoi(answer.substr(pos, answer.length()-2).c_str());

			//Store minimal value
			if(value < strength || strength == -128) strength = value;
		}

		unlock();

	}

	return (strength + 100);
}

std::string BlueGigaBluetoothModule::queryActiveConnections() {
	std::string res;

	lock();

	if(writeCommand("LIST", strlen("LIST"))) {
		size_t len;

		do {
			len = res.length();
			readAnswer(&res,1000);
		}
		while(res.length() > len);
	}

	unlock();

	return res;
}

std::vector<BluetoothModule::Device> BlueGigaBluetoothModule::scanDevices() {
	std::vector<BluetoothModule::Device> devices;
	std::string ret;
	size_t start=0, end=0, spos, epos;

	lock();

	char buf[32] = "INQUIRY 5 NAME";
	if(writeCommand(buf, strlen(buf))) {
		size_t len;
		do {
			len = ret.length();
			readAnswer(&ret, 5000);
		} while(ret.length() > len);
	}

	unlock();

	//debug("Inquiry response: %s\n", ret.c_str());

	while(end < ret.length()) {
		struct Device dev;

		end = ret.find("\n", start);
		if(end == std::string::npos) break;

		std::string line = ret.substr(start,end-start);

		//NAME <MAC> <NAME>
		if((spos = line.find("NAME ")) != std::string::npos) {
			spos = line.find(" ", spos+1);
			epos = line.find(" ", spos+1);

			for(unsigned int i=0; i<devices.size(); i++) {
				if(devices[i].mac.compare(line.substr(spos+1, epos-spos-1)) == 0) {
					spos = epos + 1;
					devices[i].name = line.substr(spos, line.length()-spos-1);
					break;
				}
			}
		}

		//INQUIRY <MAC> <CLASS>
		else if(line.size() > 17 && (spos = line.find("INQUIRY ")) != std::string::npos) {
			spos = line.find(" ", spos+1);
			epos = line.find(" ", spos+1);
			dev.mac = line.substr(spos+1, epos-spos-1);

			spos = epos + 1;
			dev.classID = atoi(line.substr(spos, line.length()-spos-1).c_str());

			devices.push_back(dev);
		}

		start=end+1;
	}

	return devices;
}

bool BlueGigaBluetoothModule::enableMUX() {

	if(muxEnabled) return true;

	lock();

	char buf[32] = "SET CONTROL MUX 1";
	std::string res;

	if(writeCommand(buf, strlen(buf))) {
		readAnswer(&res, 1000);
	}

	if(getAnswerType() == OK) {
		debug("MUX enabled: %s\n", res.c_str());
		muxEnabled = true;
		unlock();
		return true;
	}

	unlock();

	return false;
}

bool BlueGigaBluetoothModule::disableMUX() {

	lock();

	char buf[32] = "SET CONTROL MUX 0";
	if(writeCommand(buf, strlen(buf))) {
		readAnswer(NULL, 1000);
	}

	if(getAnswerType() == OK) {
		debug("MUX disabled\n");
		muxEnabled = false;
		unlock();
		return true;
	}

	unlock();

	return false;
}

bool BlueGigaBluetoothModule::sendSPPData(int connid, const void *data, size_t len) {

	lock();

	if(writeCommand((const char *)data, len, getLocalPort(connections[connid]))) {
		unlock();
		btPort->startRead();
		return true;
	}

	unlock();

	return false;
}

bool BlueGigaBluetoothModule::addSDPRecord(int localPort, void* data, size_t len) {
	return false;
}

bool BlueGigaBluetoothModule::connectSPP(int idConn, unsigned char *dstAddr, unsigned char dstPort) {

	char buf[128];
	char pin[17];

	BluetoothSPPConnection *conn = connections[idConn];
	conn->getPin(pin, sizeof(pin));

	//Set auth if found
	if(!setAuth(pin)) {
		debug("Incorrect PIN %s\n", pin);
		return false;
	}

	//aargh, this is really ugly :(
	snprintf(buf, sizeof(buf), "CALL %02x:%02x:%02x:%02x:%02x:%02x 1101 RFCOMM", dstAddr[0],dstAddr[1],dstAddr[2],dstAddr[3],dstAddr[4],dstAddr[5]);
	buf[sizeof(buf)-1]='\0';

	std::string res;
	size_t pos;

	lock();

	if(writeCommand(buf, strlen(buf))) {
		readAnswer(&res, 5000);
	}

	unlock();

	if((pos = res.find("CONNECT")) != std::string::npos) {
		if((pos = res.find(" ")) == std::string::npos) return false;

		int channel = atoi(res.substr(pos+1, 1).c_str());
		debug("Connected %d via channel %d\n", idConn, channel);
		setConnectionState(connections[idConn], SPP_CONNECTED);
		setLocalPort(connections[idConn], channel);

		return true;
	}

	return false;
}

bool BlueGigaBluetoothModule::listenSPP(int idConn) {
	return false;
}

bool BlueGigaBluetoothModule::disconnectSPP(int idConn) {

	char buf[64];
	int channel = getLocalPort(connections[idConn]);

	snprintf(buf, sizeof(buf), "CLOSE %d", channel);
	buf[sizeof(buf)-1]='\0';

	std::string res;

	lock();

	if(writeCommand(buf, strlen(buf))) {
		readAnswer(&res, 5000);
	}

	unlock();

	snprintf(buf, sizeof(buf), "NO CARRIER %d ERROR 0", channel);
	buf[sizeof(buf)-1]='\0';

	return res.find(buf) != std::string::npos;
}

void BlueGigaBluetoothModule::setEnabled(bool enabled) {
	moduleEnabled = enabled;
}

bool BlueGigaBluetoothModule::lock() {
	return btMutex.lock();
}

void BlueGigaBluetoothModule::unlock() {
	btMutex.unlock();
}

BluetoothModule::AnswerType BlueGigaBluetoothModule::getAnswerType() {
	return btPort->getAnswerType();
}


//////////////////////////////////////////////////////
// Private methods
//////////////////////////////////////////////////////


bool BlueGigaBluetoothModule::writeCommand(const char *cmd, size_t cmdlen, int channel) {
	if (!btPort) return false;

	if(muxEnabled) {
		char frame[128];
		size_t len;

		if((len=createMuxFrame(frame, sizeof(frame), cmd, cmdlen, channel)) == 0) return false;
		size_t bytesWritten = btPort->write(frame, len);

#ifdef DEBUG_BYTES
		debug("Sending data via ch %d: [", channel);
		for(unsigned int i=0; i<len; i++) printf("%0X ", frame[i]);
		printf("]\n%s\n", cmd);
#endif

		return bytesWritten == len;
	}
	else {
		size_t bytesWritten = btPort->write(cmd, cmdlen);
		bytesWritten += btPort->write("\r",1);
		return bytesWritten == cmdlen+1;
	}
}

bool BlueGigaBluetoothModule::readAnswer(std::string *pAnswer, int timeOut) {
	if (!btPort) return false;

	char rxBuf[1024];
	size_t bytesRead = 0;

	memset(rxBuf, 0, sizeof(rxBuf));
	if(!btPort->isInputPending(timeOut)) return false;

	while (btPort->isInputPending(timeOut) && bytesRead < sizeof(rxBuf)) {

		//Read everything to the buffer
		bytesRead = bytesRead + btPort->readData(rxBuf + bytesRead, sizeof(rxBuf) - bytesRead);
		rxBuf[bytesRead] = '\0';

		// check for timeout
		if (bytesRead == 0) return false;

		//Check if we are in mux mode
		//If not, do not overwrite the mux flag, we might have collected garbage!
		if(rxBuf[0] == 0xBF) muxEnabled = true;

		if(muxEnabled) {
			//Check that we have complete frame
			if (bytesRead<5 || bytesRead-5 < rxBuf[3]) continue;

#ifdef DEBUG_BYTES
			debug("Got data: [");
			for(unsigned int i = 0; i < bytesRead; i++) printf("%0X ", rxBuf[i]);
			printf("] (%d bytes)\n %s\n", bytesRead, rxBuf+3);
#endif
			if(parseMuxFrame(pAnswer, rxBuf, 0, bytesRead)==0) break;
			if(btPort->getAnswerType() != NUL) break;
		}
		else {
			size_t alen = setAnswerType(rxBuf, strlen(rxBuf));
			if(pAnswer) (*pAnswer).append(rxBuf, strlen(rxBuf) - alen);
#ifdef DEBUG_BYTES
			debug("Result data (PLAIN): %s, answer %d\n", rxBuf, btPort->getAnswerType());
#endif
			if(btPort->getAnswerType() != NUL) break;
		}
	}

	return true;
}


size_t BlueGigaBluetoothModule::setAnswerType(const char *buf, size_t len) {
	int answer = 0;

	btPort->setAnswerType(NUL);

	while(answers[answer].type !=  BluetoothModule::NUL) {
		size_t alen = strlen(answers[answer].answer);

		//answer length + \r\n must be less than data length
		if(len - alen -2 < 0) continue;


		if(strncmp(buf + (len-alen-2), answers[answer].answer, alen) == 0) {
			btPort->setAnswerType(answers[answer].type);
#ifdef DEBUG_COMM
			debug("Answer type set to %d\n", answers[answer].type);
#endif

			return alen+2;
		}

		answer++;
	}

	return 0;
}


size_t BlueGigaBluetoothModule::createMuxFrame(char *buf, size_t buflen, const char *cmd, size_t cmdlen, int channel) {
	if(buflen < cmdlen + 4) return 0;

	memset(buf, 0, buflen);

	int pos=0;

	buf[pos++] = 0xBF;				//SOF
	buf[pos++] = channel;			//Command channel
	buf[pos++] = 0;					//FLAGS
	buf[pos++] = cmdlen;			//Data length

	memmove(buf+pos, cmd, cmdlen);
	buf[pos+cmdlen] = buf[1]^0xff;	//Channel XOR 0xFF

	return pos+cmdlen+1;
}

size_t BlueGigaBluetoothModule::parseMuxFrame(std::string *buf, char *frame, size_t pos, size_t len) {
	size_t flags, datalen, alen=0, lastpos = pos;
	int link;

#ifdef DEBUG_MUX
	debug("Starting MUX parser\n");
#endif

	//Detect SOF if we are at the beginning
	while(pos < len) {
		if(frame[pos++] == 0xBF) {
#ifdef DEBUG_MUX
			debug("Found SOF at %d\n", pos-1);
#endif
			break;
		}
	}

#ifdef DEBUG_MUX
	debug("POS: %d LEN: %d\n", pos, len);
#endif

	//We might have had partial data, return last pos
	if(pos == len) {
#ifdef DEBUG_MUX
		debug("End of stream reached\n");
#endif
		return lastpos;
	}

	link = frame[pos++];
	if(link != 0xFF && link>0x08) {
#ifdef DEBUG_MUX
		debug("Mux error: Incorrect link ID %x\n", link);
#endif
		return 0;
	}

	flags = frame[pos++];
	datalen = frame[pos++];

#ifdef DEBUG_MUX
	debug("Data length is %d\n", datalen);
#endif

	if(frame[pos + datalen] != (link^0xFF)) {
		debug("Mux error: Incorrect EOF %x != %x\n", frame[pos + datalen], link^0xff);
		return 0;
	}

	//Control frame? Set answer type then
	if(link == 0xFF) {
#ifdef DEBUG_MUX
		debug("Found control frame, setting answer type\n");
#endif
		alen = setAnswerType(frame + pos, datalen);
		if(buf) (*buf).append(frame + pos, datalen - alen);
	}
	//Data frame. Find proper SPP and store in its fifo
	else {
		unsigned int i;

		for(i=0; i<MAX_BT_CONNECTIONS; i++) {
			if(connections[i] == NULL) continue;
			if(getLocalPort(connections[i]) == link) break;
		}

		sendIncomingData(connections[i], frame + pos, datalen-alen);
	}


	return parseMuxFrame(buf, frame, pos + datalen + 1, len);
}


