/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : anti.sullin at artecdesign dot ee              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <string.h>

#include "BluetoothModule.h"
#include "BluetoothSPPConnection.h"

//#define DEBUG
#include "Debug.h"

BluetoothModule *BluetoothModule::pModule = NULL;

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Virtual constructor for BluetoothModule class
 */
BluetoothModule::BluetoothModule()
{
	for(unsigned int i = 0; i < MAX_BT_CONNECTIONS; i++) connections[i] = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Terminates the LMX Bluetooth module object.
 */
BluetoothModule::~BluetoothModule()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the pin code for incoming connection request
 * Sets the pin code when incoming connection request is received.
 *
 * \param newpin Pointer to pin code to use for the connection
 * \return Returns true when pin code was successfully set, false otherwise
 */
bool BluetoothModule::setIncomingPin(const char *newpin)
{
	if (strlen(newpin)+1 > sizeof(incomingPin)) return false;
	strncpy(incomingPin, newpin, sizeof(incomingPin));
	incomingPin[sizeof(incomingPin)-1] = 0;
	return true;
}
////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Create a new SPP connection
 *
 * Creates a new SPP connection with given parameters; returns pointer
 * to that connection object.
 *
 * \param dstAddr Destination BDADDR to connect to
 * \param dstPort Destination service port to connect to
 * \param pin Destination device's PIN code
 *
 * \return Returns pointer to new connection object or NULL on error.
 */
BluetoothSPPConnection *BluetoothModule::createSPPConnection(const unsigned char *dstAddr, int dstPort, const char *pin) {
	BluetoothSPPConnection *conn;

	conn = new BluetoothSPPConnection(pModule, false);
	if (!conn) return NULL;

	if (!conn->setDestAddr(dstAddr) ||
			!conn->setDestPort(dstPort) ||
			!conn->setPin(pin) ||
			!conn->registerSPP()) {
		delete conn;
		return NULL;
	}

	return conn;
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Listen for a new SPP connection
 *
 * Creates a new listening SPP connection with given parameters; returns pointer
 * to that connection object.
 * The object is in disconnected state and flips to connected state as soon as any
 * incoming connection is received.
 *
 * \param localPort Service port to listen on
 *
 * \return Returns pointer to new connection object or NULL on error.
 */
BluetoothSPPConnection *BluetoothModule::createIncomingSPPConnection(int localPort) {
	BluetoothSPPConnection *conn;

	conn = new BluetoothSPPConnection(pModule, true);
	if (!conn) return NULL;

	if (!conn->setLocalPort(localPort) ||
			!conn->registerSPP()) {
		delete conn;
		return NULL;
	}

	return conn;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Register SPP connection
 *
 * Register a new SPP connection pointer in LMXBluetoothModule.
 * \param conn Pointer to the bluetooth connection object
 *
 * \return Returns connection id or -1 on error
 */
int BluetoothModule::registerSPPConnection(BluetoothSPPConnection* conn)
{
	connMutex.lock();
	for (int i = MAX_BT_INCOMING_CONNECTIONS; i< MAX_BT_CONNECTIONS; i++) {
		if (connections[i]) continue;

		connections[i] = conn;
		connMutex.unlock();
		return i;
	}
	connMutex.unlock();
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Register listening SPP connection
 *
 * Register a new SPP connection pointer in LMXBluetoothModule.
 * \param conn Pointer to the bluetooth connection object
 *
 * \return Returns connection id or -1 on error
 */
bool BluetoothModule::registerIncomingSPPConnection(BluetoothSPPConnection* conn)
{
	int i = conn->getLocalPort() -1;
	debug("Registering incoming connection for port %d", i);
	if (i < 0 || i >= MAX_BT_CONNECTIONS) return false;

	connMutex.lock();
	if (connections[i]) {
		debug("Connection %d busy\n", i);
		connMutex.unlock();
		return false;
	}

	connections[i] = conn;
	connMutex.unlock();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unregister a bluetooth SPP connection
 *
 * Unregisters a bluetooth SPP connection in the LMXBluetoothModule
 * \param idConn Connection ID that should be unregistered
 */
void BluetoothModule::deregisterSPPConnection(int idConn)
{
	if (idConn < 0 || idConn >= MAX_BT_CONNECTIONS) return;
	connMutex.lock();
	connections[idConn] = NULL;
	connMutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Friendship is not inherited, so friend protected functions must be called
// from here, not from our derived classes!

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Forward incoming data to BluetoothSPPConnection object
 *
 * Forwards incoming SPP data to the buffer in the connection object.
 *
 * \param pConn Pointer to the connection object that must receive the data
 * \param buf Pointer to the buffer which contents must be forwarded
 * \param len Buffer size
 */
void BluetoothModule::sendIncomingData(BluetoothSPPConnection *pConn, void *buf, size_t len)
{
	pConn->incomingData(buf, len);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Forward connection state to BluetoothSPPConnection object
 *
 * Forwards connection state events to connection object.
 *
 * \param pConn Pointer to the connection object that must receive the connection state events
 * \param state The new state that will be set
 */
void BluetoothModule::setConnectionState(BluetoothSPPConnection *pConn, SPP_CONN_STATUS state)
{
	if(pConn == NULL) return;

	pConn->setConnectionState(state);
}

void BluetoothModule::setLocalPort(BluetoothSPPConnection *pConn, int port) {
	if(pConn == NULL) return;

	pConn->setLocalPort(port);
}

int BluetoothModule::getLocalPort(BluetoothSPPConnection *pConn) {
	if(pConn == NULL) return 0;

	return pConn->getLocalPort();
}

void BluetoothModule::deleteSPPConnection(BluetoothSPPConnection *pConn) {
	delete pConn;
}
