#ifndef BLUETOOTHMODULE_H_
#define BLUETOOTHMODULE_H_

#include <string>
#include <vector>

#include "Thread.h"

#define MAX_BT_CONNECTIONS	7
#define MAX_BT_INCOMING_CONNECTIONS	2
#define MAX_BT_PIN	17

#define CONNECT_TIMEOUT		20000
#define DISCONNECT_TIMEOUT	1000

/** Bluetooth module connection states */
enum SPP_CONN_STATUS {
	/** Connection failed */
	SPP_CONN_FAILED = -1,
	/** Module is disconnected */
	SPP_DISCONNECTED = 0,
	/** Module is connected */
	SPP_CONNECTED = 1,
	/** Module is set to listen for pairing requests */
	SPP_LISTENING = 2,
};

// forward declaration
class BluetoothSPPConnection;

/**
 * \brief Bluetooth module class
 *
 * Implements the generic Bluetooth subsystem interfaces. Other classes will be derived
 * from it to support specific Bluetooth modules.
 */
class BluetoothModule
{
public:
	enum AnswerType { NUL, OK, ERROR };
	struct ModemAnswers {
		const char * answer;
		AnswerType type;
	};

	struct Device {
		std::string name;
		std::string mac;
		int classID;
	};

	virtual ~BluetoothModule();

	/**
	 * Reader implementation should override this function to implement a hardware
	 * uninitialization procedure. Reader manager calls this function during
	 * system shutdown.
	 *
	 * \return Returns true if the shutdown was successful, false otherwise
	 */
	virtual bool shutdown() = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * reading device MAC address.
	 *
	 * \return Returns MAC address as string
	 */
	virtual std::string getLocalMac() = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * retrieving reader's bluetooth protocol version.
	 *
	 * \return Returns version number as string
	 */
	virtual std::string getVersion() = 0;

	/**
	 * Reader implementation should override this function to set the device into
	 * enabled or disabled state.
	 *
	 * \param enabled If set to true, the bluetooth device will be enabled. If set to false,
	 * the bluetooth device will be disabled.
	 */
	virtual void setEnabled(bool enabled) = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * connecting the device over SPP (Serial Port Protocol).
	 *
	 * \param idConn Connection ID to be used
	 * \param dstAddr Address of the destination device
	 * \param dstPort Port of the destination device
	 *
	 * \return Returns true if the connection was established successfully, false otherwise
	 */
	virtual bool connectSPP(int idConn, unsigned char *dstAddr, unsigned char dstPort) = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * putting the device into listening mode
	 *
	 * \param idConn Connection ID that should be used for listening incoming packets
	 * \return Returns true if the device is in listening mode, false otherwise
	 */
	virtual bool listenSPP(int idConn) = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * disconnecting the SPP link
	 *
	 * \param idConn Connection ID that should be disconnected
	 * \return Returns true if the device was successfully connected, false otherwise
	 */
	virtual bool disconnectSPP(int idConn) = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * storing a new service record in local service discovery database
	 *
	 * \param localPort Local port used by the service
	 * \param data Pointer to the data to be used
	 * \param len Data length
	 *
	 * \return Returns true if the record was successfully created
	 *
	 */
	virtual bool addSDPRecord(int localPort, void* data, size_t len) = 0;

	/**
	 * Reader implementation should override this function to implement method for
	 * sending the SDP data to over the bluetooth connection
	 * \param connid Connection ID, returned on SPP connection registration
	 * \param data Data to be sent over SPP link
	 * \param len Data length
	 *
	 * \return Returns true if data was sent out, false on error
	 */
	virtual bool sendSPPData(int connid, const void *data, size_t len) = 0;

	virtual bool open() = 0;
	virtual AnswerType getAnswerType() { return NUL; };

	virtual bool isAlive() { return false; }
	virtual bool setLAP(const char* lapID) { return false; }
	virtual bool ping() { return false; }
	virtual bool enableMUX() { return false; }
	virtual bool disableMUX() { return false; }
	virtual bool disableEcho() { return false; }
	virtual bool setAuth(char * pin) { return false; }
	virtual std::vector<Device> scanDevices() { return std::vector<Device>(); }
	virtual int getSignalStrength() { return -1; }

	virtual void pollIncomingData(int timeOut) {};

	/**
	 * \brief Return the bluetooth module object
	 * Returns pointer to the instance of BluetoothModule class
	 *
	 * \return Returns pointer to the BluetoothModule instance
	 */
	static BluetoothModule *getModule() { return pModule; }

	friend class BluetoothSPPConnection;

	int registerSPPConnection(BluetoothSPPConnection* conn);
	bool registerIncomingSPPConnection(BluetoothSPPConnection* conn);
	void deregisterSPPConnection(int idConn);
	void deleteSPPConnection(BluetoothSPPConnection *pConn);

	BluetoothSPPConnection *createSPPConnection(const unsigned char *dstAddr, int dstPort, const char *pin);
	BluetoothSPPConnection *createIncomingSPPConnection(int localPort);

	bool setIncomingPin(const char *pin);




protected:
	BluetoothModule();
	void sendIncomingData(BluetoothSPPConnection *pConn, void *buf, size_t len);
	void setConnectionState(BluetoothSPPConnection *pConn, SPP_CONN_STATUS state);

	void setLocalPort(BluetoothSPPConnection *pConn, int port);
	int getLocalPort(BluetoothSPPConnection *pConn);

	/** Static variable to hold the pointer to the instance of BluetoothModule class */
	static BluetoothModule *pModule;

	/** Mutex to protect SPP connection list */
	Mutex					connMutex; // protect SPP connections list

	/** List of bluetooth connections associated with the BluetoothModule class instance */
	BluetoothSPPConnection* connections[MAX_BT_CONNECTIONS];

	/** Incoming PIN code */
	char incomingPin[MAX_BT_PIN];
private:

};

#endif /*BLUETOOTHMODULE_H_*/
