/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : anti.sullin at artecdesign dot ee              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <string.h>

#include <sys/time.h>
#include <stdio.h>

#include "BluetoothSPPConnection.h"

//#define DEBUG
#include "Debug.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Virtual constructor for BluetoothSPPConnection class
 *
 * \param pBtModule Pointer to the bluetooth module object
 * \param incoming Set to true, if the connection will be initiated by the other party
 */
BluetoothSPPConnection::BluetoothSPPConnection(BluetoothModule *pBtModule, bool incoming)
{
	debug("BTSPPConn:connect, %p, %d\n", pBtModule, incoming);
	this->pBtModule = pBtModule;
	this->incoming = incoming;
	fifo_r = 0;
	fifo_w = 0;
	connStatus = SPP_DISCONNECTED;
	connId = -1;
	for (unsigned int i=0; i<sizeof(dstAddr); i++) dstAddr[i]=0xFF;
	pin[0] = 0;
	localPort = 0;
	dstPort = 0;
	channel = 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Unregisters and uninitializes the BluetoothSPPConnection object
 */
BluetoothSPPConnection::~BluetoothSPPConnection()
{
	if (pBtModule && connId != -1) {
		disconnect();
		pBtModule->deregisterSPPConnection(connId);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send SPP data
 *
 * Send data over Bluetooth SPP link. Returns nonzero if the data was sent.
 *
 * \param buf Pointer to the buffer that contains the data that will be sent
 * \param len Buffer length
 *
 * \note This does not guarantee that the receiver has received it, just that it was
 * transmitted!
 *
 * \return Returns number of bytes transmitted.
 */
ssize_t BluetoothSPPConnection::writeData(const void *buf, size_t len)
{
	if (connStatus != SPP_CONNECTED) return false;

	if (pBtModule->sendSPPData(connId, buf, len))
		return len;
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get Rx FIFO count
 *
 * Returns the number of unread bytes in Rx FIFO.
 *
 * \return Returns number of bytes in Rx FIFO.
 */
int BluetoothSPPConnection::getFifoCnt()
{
	int cnt = fifo_w - fifo_r;
	if (cnt < 0) cnt += BT_SPP_FIFO_SIZE;

	return cnt;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle incoming data
 *
 * Friend class BluetoothModule uploads Rx data
 *
 * \param vbuf Pointer to the buffer that contains the data
 * \param len Buffer size
 */
void BluetoothSPPConnection::incomingData(const void *vbuf, size_t len)
{
	size_t tail = BT_SPP_FIFO_SIZE - fifo_w;
	size_t free = BT_SPP_FIFO_SIZE - 1 - getFifoCnt(); // up to SIZE-1 bytes
	char *buf = (char*) vbuf;
	if (!free) return;

	// truncate buffer if too much data
	if (len > free) len = free;

	// do we fit to the tail?
	if (len <= tail) {
		memcpy(fifo+fifo_w, buf, len);
	} else {
		memcpy(fifo+fifo_w, buf, tail);
		memcpy(fifo, buf+tail, len-tail);
	}

	// move w pointer
	if (fifo_w + len >= BT_SPP_FIFO_SIZE)
		fifo_w = fifo_w + len - BT_SPP_FIFO_SIZE;
	else
		fifo_w = fifo_w + len;

	bufSem.post();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read Rx data from FIFO
 *
 * Read received data from FIFO to vbuf. Up to buflen bytes will be read
 *
 * \param vbuf Pointer to the buffer where to store the data
 * \param buflen Size of the buffer
 *
 * \return Returns number of bytes received
 */
ssize_t BluetoothSPPConnection::readData(void *vbuf, size_t buflen)
{
	size_t tail = BT_SPP_FIFO_SIZE - fifo_r;
	size_t cnt = getFifoCnt();
	char *buf = (char*) vbuf;

	// up to buflen bytes
	if (cnt > buflen) cnt = buflen;
	if (!cnt) return 0;

	// do we fit to the tail?
	if (cnt <= tail) {
		memcpy(buf, fifo+fifo_r, cnt);
	} else {
		memcpy(buf, fifo+fifo_r, tail);
		memcpy(buf+tail, fifo ,cnt-tail);
	}

	// move r pointer
	if (fifo_r + cnt >= BT_SPP_FIFO_SIZE)
		fifo_r = fifo_r + cnt - BT_SPP_FIFO_SIZE;
	else
		fifo_r = fifo_r + cnt;

	return cnt;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Wait for Rx data
 *
 * Waits until there is at least 1 byte in Rx FIFO before the timeout occurs.
 * \param timeout Timeout in milliseconds
 *
 * \return Returns the number of bytes in FIFO.
 */
size_t BluetoothSPPConnection::numInputPending(int timeout)
{
	if (fifo_r != fifo_w) return getFifoCnt();

	bufSem.pend(timeout);
	return getFifoCnt();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Wait for Rx data
 *
 * Waits until there is at least 1 byte in Rx FIFO before timeout occurs.
 * \param timeout Timeout in milliseconds
 *
 * \return Returns true if data is in buffer, false otherwise
 */
bool BluetoothSPPConnection::isInputPending(int timeout)
{
	if (fifo_r != fifo_w) return getFifoCnt();

	bufSem.pend(timeout);
	return getFifoCnt() > 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief ReOpen
 *
 * Not implemented for BluetoothSPPConnection, port can not be reOpened
 *
 * \return Returns always false
 */
bool BluetoothSPPConnection::reOpen()
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief isPortOpen
 *
 * Not implemented for BluetoothSPPConnection, port is always open
 *
 * \return Returns always true
 */
bool BluetoothSPPConnection::isPortOpen()
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Flush Rx FIFO
 *
 * Empties Rx FIFO buffer.
 */
void BluetoothSPPConnection::flushInput()
{
	fifo_r = fifo_w;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Flush serial port output
 *
 * \note This function is not implemented yet
 */

void BluetoothSPPConnection::flushOutput()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Drain the serial port
 *
 * \note This function is not implemented yet
 */
void BluetoothSPPConnection::drainOutput()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set connection state
 *
 * Called by friend class BluetoothModule to notify of connection state change.
 *
 * \param connStatus New status of the bluetooth connection
 */
void BluetoothSPPConnection::setConnectionState(SPP_CONN_STATUS connStatus)
{
	debug("setConnectionState %d\n", connStatus);

	if (incoming && connStatus != SPP_CONNECTED) connStatus = SPP_LISTENING;
	this->connStatus = connStatus;
	linkStatusSem.post();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Register the SPP connection
 *
 * Registers the SPP connection in the BluetoothModule. This will enable
 * the BluetoothModule to send Rx data and status updates to us.
 *
 * \return Returns true if SPP connection was successfully registered, false otherwise.
 */
bool BluetoothSPPConnection::registerSPP()
{
	int ret;
	debug("register SPP\n");
	if (!pBtModule) {
		debug("BT module not found\n");
		return false;
	}

	if (incoming) {
		if (!pBtModule->registerIncomingSPPConnection(this)) return false;
		connId = localPort-1;
		return true;
	}
	ret = pBtModule->registerSPPConnection(this);
	if (ret == -1) {
		debug("Could not register SPP connection\n");
		return false;
	}

	connId = ret;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get the destination address of the bluetooth device
 *
 * \param pAddr Buffer where to store the destination address, unsigned char[6]
 *
 * \return Returns true if the address was successfully stored into buffer, false otherwise
 * Returns false if the module waits for incoming connections.
 */
bool BluetoothSPPConnection::getDestAddr(unsigned char *pAddr)
{
	if (incoming) return false;
	if (dstAddr[0] == 0xFF && dstAddr[1] == 0xFF && dstAddr[2] == 0xFF &&
			dstAddr[3] == 0xFF && dstAddr[4] == 0xFF && dstAddr[5] == 0xFF)
		return false;
	memcpy(pAddr, dstAddr, sizeof(dstAddr));
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set destination address of the bluetooth device
 *
 * \param pAddr Destination address, unsigned char[6]
 *
 * \return Returns true if set, false on error.
 * Returns false if the module waits for incoming connections.
 */
bool BluetoothSPPConnection::setDestAddr(const unsigned char *pAddr)
{
	if (incoming) return false;

	size_t len;

	//Most likely we have a mac of XX:XX:XX:XX:XX:XX, try to parse
	if((len=strlen((char *)pAddr))==17) {
		unsigned int i=0;
		unsigned int low, high;

		while(i<6) {
			if(sscanf((char *)pAddr+i*3, "%1X", &high) == 0) return false;
			if(sscanf((char *)pAddr+i*3+1, "%1X", &low) == 0) return false;
			dstAddr[i] = (high << 4 & 0xFF) | (low & 0x0F);
			i++;
		}
	}
	else
		memcpy(dstAddr, pAddr, sizeof(dstAddr));

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get destination port of the remote bluetooth device
 *
 * \return Returns destination port or 0 if not set.
 */
int BluetoothSPPConnection::getDestPort()
{
	return dstPort;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get local bluetooth device port
 *
 * \return Returns local port or 0 if not set.
 */
int BluetoothSPPConnection::getLocalPort()
{
	return localPort;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set destination port for remote bluetooth device
 *
 * \param destPort Destination port of the remote bluetooth device
 *
 * \return Returns true if set, false on error
 */
bool BluetoothSPPConnection::setDestPort(int destPort)
{
	if (incoming) return false;
	if (destPort < 0 || destPort > MAX_BT_CONNECTIONS) return false;
	this->dstPort = destPort;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set local bluetooth device port
 *
 * \param localPort Port of the local bluetooth device
 *
 * \return Returns true if set, false on error
 */
bool BluetoothSPPConnection::setLocalPort(int localPort)
{
	//if (!incoming) return false;
	if (localPort < 0 || localPort > MAX_BT_CONNECTIONS) return false;
	this->localPort = localPort;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get bluetooth PIN
 *
 * Returns Bluetooth PIN set for this connection.
 * \note Bluetooth PIN is a unicode string!
 *
 * \param pin Buffer to store PIN
 * \param pinlen Pin buffer length
 */
void BluetoothSPPConnection::getPin(char *pin, size_t pinlen)
{
	strncpy(pin, this->pin, pinlen);
	pin[pinlen-1] = 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set bluetooth PIN
 *
 * Sets Bluetooth PIN set for this connection.
 * \note Bluetooth PIN is a unicode string!
 *
 * \param newpin Bluetooth pin as zero-terminated unicode string.
 *
 * \return Returns true if set, false on error
 */
bool BluetoothSPPConnection::setPin(const char *newpin)
{
	if (strlen(newpin)+1 > sizeof(pin)) return false;
	strncpy(pin, newpin, sizeof(pin));
	pin[sizeof(pin)-1] = 0;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Connect to remote device
 *
 * Initiate a pairing process and wait until connected.
 * This may take up to CONNECT_TIMEOUT ms. Do not shorten the
 * time, Bluetooth devices may be in low power mode and wake up
 * once in 10 seconds to check for new connections.
 *
 * \return Returns true if connected successfully, false on error.
 */
bool BluetoothSPPConnection::connect()
{
	struct timeval start, now;
	if (incoming) return false;
	if (connStatus == SPP_CONNECTED) return true;

	debug("connect start\n");

	if (!pBtModule->connectSPP(connId, dstAddr, dstPort))
		return false;

	gettimeofday(&start, 0);
	while (1) {
		if (connStatus == SPP_CONNECTED) return true;
		if (connStatus == SPP_CONN_FAILED) return false;
		linkStatusSem.pend(CONNECT_TIMEOUT/2);

		if (connStatus == SPP_CONNECTED) return true;
		if (connStatus == SPP_CONN_FAILED) return false;
		gettimeofday(&now, 0);
		if ((((now.tv_sec - start.tv_sec) * 1000) +
				((now.tv_usec - start.tv_usec) / 1000)) > CONNECT_TIMEOUT) {
			debug("Timeout sending command\n");
			break;
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start listening of incoming connections
 *
 * \return Returns true if initiated successfully, false on error.
 */
bool BluetoothSPPConnection::listen()
{
	if (!incoming) return false;
	if (connStatus == SPP_LISTENING) return true;
	if (connStatus == SPP_CONNECTED) return true;

	debug("listen start\n");

	if (!pBtModule->listenSPP(connId)) return false;
	connStatus = SPP_LISTENING;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Disconnect from remote device
 *
 * Disconnect a Bluetooth SPP connection.
 */
bool BluetoothSPPConnection::disconnect()
{
	return pBtModule->disconnectSPP(connId);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get connection status
 *
 * Returns true if still connected, false otherwise.
 * The connection may be dropped by remote host or by bad network coverage!
 */
bool BluetoothSPPConnection::isConnected()
{
	return connStatus == SPP_CONNECTED;
}
