/*
 * EventBroadcaster.h
 *
 *  Created on: Dec 8, 2008
 *      Author: andrei
 */

#ifndef EVENTBROADCASTER_H_
#define EVENTBROADCASTER_H_

#include "Thread.h"

#include <list>


//#define DEBUG_EVENTBROADCAST
#ifdef DEBUG_EVENTBROADCAST
	#define debugebc(fmt, args...)<--->do {\
			printf("(%s:%d) %s: ", __FILE__, __LINE__, __FUNCTION__); \
			printf(fmt, ## args); \
			} while(0);
#else 
	#define debugebc(fmt...)
#endif 


/**
 * \brief Event broadcasting class template
 *
 * This class is a template for classes that implement event broadcasting
 * to listeners. The template provides methods to manage a list of listeners.
 * The implementation may then enumerate registered listeners and call them
 * when appropriate.
 *
 * The class takes TListener as a template parameter, which identifies the
 * listener interface being managed by the event broadcaster.
 */
template <class TListener> class EventBroadcaster
{
public:

	/**
	 * \brief Class constructor
	 *
	 * Initializes the event broadcaster class.
	 */
	EventBroadcaster()
	{
		forwardIterator = eventListeners.end();
		reverseIterator = eventListeners.rend();
	}

	/**
	 * \brief Class destructor
	 *
	 * Destroys the event broadcaster class.
	 */
	virtual ~EventBroadcaster()
	{
		eventListeners.clear();
	}

	/**
	 * \brief Add new listener
	 *
	 * Adds a new listener of type TListener to the list of event listeners.
	 *
	 * \param pListener Pointer to the event listener class that
	 * implements TListener interface.
	 */
	void addListener(TListener *pListener)
	{
		debugebc("adding event listener %p\n", pListener);

		listenerMutex.lock();
		eventListeners.push_back(pListener);
		listenerMutex.unlock();
	}

	/**
	 * \brief Remove listener
	 *
	 * Deletes a listener of type TListener from the list of event listeners.
	 * The listener must be previously added by the addListener() call.
	 *
	 * \param pListener Pointer to the event listener object implementing TListener
	 * interface.
	 * \return Returns true if the listener was deleted successfully.
	 */
	bool removeListener(TListener *pListener)
	{
		debugebc("removing event listener %p\n", pListener);
		listenerMutex.lock();

		for(forwardIterator = eventListeners.begin();
			forwardIterator != eventListeners.end(); forwardIterator ++)
		{
			if(*forwardIterator == pListener)
			{
				eventListeners.erase(forwardIterator);
				listenerMutex.unlock();
				return true;
			}
		}

		listenerMutex.unlock();
		debugebc("failed\n");
		return false;
	}

	/**
	 * \brief Start listener enumeration
	 *
	 * Begins enumerating registered listeners. Enumeration is supported in
	 * forward and reverse directions. After the call to beginEnumListeners()
	 * succeeds, the caller should proceed to getNextListener() and finally
	 * endEnumListeners().
	 *
	 * \param enumForward Specifies whether the enumeration is in forward direction.
	 * \return Returns true if there is at least one listener to enumerate.
	 */
	bool beginEnumListeners(bool enumForward = true)
	{
		listenerMutex.lock();
		if (eventListeners.empty())
		{
			listenerMutex.unlock();
			return false;
		}

		if (enumForward)
		{
			forwardIterator = eventListeners.begin();
			reverseIterator = eventListeners.rend();
		}
		else
		{
			forwardIterator = eventListeners.end();
			reverseIterator = eventListeners.rbegin();
		}

		return true;
	}

	/**
	 * \brief Get next listener
	 *
	 * Retrieve the next listener from the list of event listeners.
	 * Note beginEnumListeners() must be called at the beginning of enumeration.
	 *
	 * \return Returns the pointer to the next registered listener object.
	 * Returns NULL if there are no more listeners in the list.
	 */
	TListener* getNextListener()
	{
		TListener *p = NULL;
		if (forwardIterator != eventListeners.end())
		{
			p = *forwardIterator;
			forwardIterator ++;
		}
		else if (reverseIterator != eventListeners.rend())
		{
			p = *reverseIterator;
			reverseIterator ++;
		}
		return p;
	}

	/** \brief End listener enumeration
	 *
	 * Stops enumerating registered listeners. Must be called at the end
	 * of enumeration to release the thread safety lock.
	 */
	void endEnumListeners()
	{
		listenerMutex.unlock();
	}

private:

	std::list<TListener*>	eventListeners;
	Mutex					listenerMutex;

	typename std::list<TListener*>::iterator forwardIterator;
	typename std::list<TListener*>::reverse_iterator reverseIterator;
};

#endif /* EVENTBROADCASTER_H_ */
