/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

/*
 * GprsConnection.cpp
 *
 *  Created on: Nov 19, 2008
 *      Author: andrei
 */

#include <sstream>
#include <platform.h>

//#define DEBUG
#include "Debug.h"

#include "GprsConnection.h"

#define GPRS_INTERFACE		"ppp0"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the GPRS network connection object.
 *
 * \param pSettings Points to a structure containing the GPRS configuration
 * parameters. The data is copied to a local storage.
 */
GprsConnection::GprsConnection(GPRS_CONNECTION_SETTINGS *pSettings) :
				NetworkConnection(COMM_MODULE_GPRS),
				gsmModule(GSM_PTY_COMMAND)
{
	cfg = *pSettings;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Terminates the GPRS network connection object.
 */
GprsConnection::~GprsConnection()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize the GPRS module
 *
 * Initializes the GPRS module and connects to the provider network. The data connection
 * must be opened separately by calling netConnect().
 *
 * This should be done before the application is fully loaded
 *
 * \return Returns true if initialization was successful and modem is working, false otherwise.
 */
bool GprsConnection::netInitialize()
{
	// GSM modem bootstrap object
	GsmModule gsmInit(GSM_MODULE_DEVICE);

	// power up the GSM module
	setStatus("Powering up GPRS module");
	debug("powering up gsm module\n");
	PowerManager *pPM = PowerManager::getPowerManager();
	if (!pPM->gprsPowerOn())
	{
		reportError("GSM module or power problem");
		return false;
	}

	// open the GSM module physical port
	if (!gsmInit.open())
	{
		reportError("GSM communication error");
		return false;
	}

	// set basic communication options
	setStatus("Initializing GPRS module");
	debug("init gsm module\n");
	if (!gsmInit.initComm())
	{
		reportError("GSM module initialization failed");
		return false;
	}

	// close the bootstrap GSM modem
	gsmInit.close();

	// start GSM07.10 port multiplexer
	setStatus("Initializing multiplexer");
	debug("init serial mux\n");
	const char *ports[] = { "/dev/ptyp0", "/dev/ptyp1", NULL };
	if (!serialMux.startMux(GSM_MODULE_DEVICE, ports))
	{
		reportError("GSM CMUX initialization error");
		return false;
	}

	// now, open the modem virtual command port
	if (!gsmModule.open())
	{
		reportError("GSM communication error");
		return false;
	}

	// disable power saving mode and set operational band
	gsmModule.powerSaveMode(GsmModule::POWER_FULL_SAVE);
	gsmModule.setFrequency(GsmModule::FREQ_900X1800);

	// set basic communication options
	setStatus("Initializing SIM card");
	debug("init sim card\n");
	if (!gsmModule.initSIM(cfg.simPin.c_str()))
	{
		reportError("SIM card not present or broken");
		return false;
	}

	setStatus("Initializing GPRS settings");
	debug("init gprs settings\n");
	if (!gsmModule.initGprs(cfg.gprsApn.c_str()))
	{
		reportError("GPRS configuration error");
		return false;
	}

	// seek operator network
	setStatus("Seeking for network");
	debug("seeking network\n");
	if (!gsmModule.searchNetwork(0))
	{
		debug("failed joining network\n");
		reportError("Unable to register with GSM network");
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Shuts down the GPRS module
 * Shuts down and powers off the GPRS module and multiplexer.
 *
 * \return Returns true if power down was successful
 */
bool GprsConnection::netShutdown()
{
	debug("closing GSM module\n");
	setStatus("Shutting down GPRS module");
	gsmModule.close();

	debug("shutting down CMUX\n");
	setStatus("Shutting down multiplexer");
	serialMux.stopMux();

	debug("powering down GSM module\n");
	PowerManager *pPM = PowerManager::getPowerManager();
	return pPM->gprsPowerOff();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Establishes the data connection
 *
 * Establishes the PPP data connection with the service provider
 *
 * \return Returns true, if the connection was established successfully
 */
bool GprsConnection::netConnect()
{
	setStatus("Initializing PPP session");
	debug("attaching gprs\n");
	if (!gsmModule.attachGprs())
	{
		debug("failed attaching gprs\n");
		reportError("GPRS attach procedure failed");
		return false;
	}

	GPRSPPP_PARAMETERS pppParams;
	pppParams.debugMode = false;
	pppParams.useAuth = !cfg.authUser.empty();
	pppParams.gprsUser = cfg.authUser.c_str();
	pppParams.gprsPass = cfg.authPass.c_str();
	pppParams.persistMode = false;

	if (!GprsPpp::start(GSM_PTY_DATA, &pppParams))
	{
		debug("failed starting GPRS PPP\n");
		reportError("Failed starting GPRS PPP");
		return false;
	}

	setStatus("GPRS is ready");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Terminates the PPP connection
 * Closes the PPP connection with the service provider.
 *
 * \return Returns true if the connection was terminated successfully
 */
bool GprsConnection::netDisconnect()
{
	// close ppp connection
	if(GprsPpp::isRunning())
	{
		setStatus("Terminating PPP session");
		if (!GprsPpp::stop())
		{
			debug("failed terminating PPP\n");
			reportError("Failed terminating GPRS PPP");
			return false;
		}
	}

	setStatus("GPRS is shut down");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Temporarily suspends the GPRS connection
 * Shuts down the GPRS command channel and terminates the PPP connection for power saving reasons.
 * It can be re-established with netResume() method.
 *
 * \return Returns true if the suspend was successful
 */
bool GprsConnection::netSuspend()
{
	debug("preparing to powersave\n");
	setStatus("Suspending GPRS connection");

	// close modem port temporarily
	if (!gsmModule.close())
	{
		debug("failed closing gsm module\n");
		reportError("failed closing GSM module");
		return false;
	}

	// close ppp connection
	if(GprsPpp::isRunning())
	{
		setStatus("Terminating PPP session");
		if (!GprsPpp::stop())
		{
			debug("failed terminating PPP\n");
			reportError("Failed terminating GPRS PPP");
			return false;
		}
	}

	setStatus("GPRS is suspended");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Resumes the connection
 * Resumes the GPRS connection from previously suspended state
 *
 * \return Returns true if the resume was successful
 */
bool GprsConnection::netResume()
{
	debug("returning from powersave\n");

	if (!GprsPpp::isRunning())
	{
		setStatus("Starting PPP session");
		GPRSPPP_PARAMETERS pppParams;
		pppParams.debugMode = false;
		pppParams.useAuth = !cfg.authUser.empty();
		pppParams.gprsUser = cfg.authUser.c_str();
		pppParams.gprsPass = cfg.authPass.c_str();
		pppParams.persistMode = false;

		if (!GprsPpp::start(GSM_PTY_DATA, &pppParams))
		{
			debug("failed starting GPRS PPP\n");
			reportError("Failed starting GPRS PPP");
			return false;
		}
	}
	else debug("gprs PPP already up!\n");

	// open temporarily closed modem port
	if (!gsmModule.open())
	{
		debug("failed opening gsm module\n");
		reportError("failed opening GSM module");
		return false;
	}

	setStatus("GPRS is ready");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Returns the status of the connection
 *
 * \return Returns true, if connection is active, false otherwise
 */
bool GprsConnection::isConnectionUp()
{
	return GprsPpp::isRunning();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get information about the network
 * Get general information about the network (i.e. signal strength etc.)
 *
 * \param netInfo Structure where the information will be stored
 * \return Returns true if the information was successfully stored
 */
bool GprsConnection::getNetworkInfo(GENERIC_NETWORK_INFO &netInfo)
{
	// read out network interface flags
	char ipAddr[16]="", macAddr[18]="";
	if (getInterfaceAddresses(GPRS_INTERFACE, ipAddr, macAddr))
	{
		netInfo.networkAddress = ipAddr;
		netInfo.hardwareAddress = macAddr;
	}

	// read out current GSM status
	GSM_NETWORK_INFO gsmInfo = { 0 };
	if (!gsmModule.readNetworkInfo(gsmInfo))
	{
		debug("failed reading GSM net info\n");
		reportError("Failed requesting network information");
		return false;
	}

	// check if there is network coverage
	if(gsmInfo.netStatus == 1 || gsmInfo.netStatus == 5)
	{
		netInfo.isConnected = true;
		netInfo.networkChannel = gsmInfo.rfChannel;
		netInfo.networkName = gsmInfo.networkName;

		// calculate 0..100% level
		if(gsmInfo.signalPower < -113) netInfo.signalStrength = 0;
		else if(gsmInfo.signalPower > -51) netInfo.signalStrength = 100;
		else netInfo.signalStrength = (gsmInfo.signalPower + 113) * 100 / 62;

		debug("Network %s @ %d, %d dBm (%d%%)\n", gsmInfo.networkName.c_str(),
				gsmInfo.rfChannel, gsmInfo.signalPower, netInfo.signalStrength);
	}
	else
	{
		netInfo.isConnected = false;
		netInfo.signalStrength = 0;
		netInfo.networkChannel = 0;
		netInfo.networkName.clear();
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get GSM module object
 *
 * \return Returns the pointer to the GSM/GPRS module object encapsulated by
 * this network connection class.
 */
GsmModule* GprsConnection::getGsmModule()
{
	return &gsmModule;
}
