/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

/*
 * GprsPpp.cpp
 *
 *  Created on: Sep 26, 2008
 *      Author: andrei
 */

#include <signal.h>
#include <sys/wait.h>
#include <errno.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <string.h>
#include <fstream>

#include "GprsPpp.h"

//#define DEBUG
#include "Debug.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start GPRS PPP
 *
 * Starts the GPRS PPP daemon (pppd). After the call returns successfully,
 * full IP connectivity should be available in the system via the ppp0
 * network interface.
 *
 * \param serialPort Points to a null-terminated string containing the serial
 * port name. The port may be either physical or virtual.
 * \param pParams Points to GPRS PPP configuration structure.
 * \return Returns true if the PPP daemon was started successfully and ppp0
 * interface is available.
 */
bool GprsPpp::start(const char* serialPort, GPRSPPP_PARAMETERS *pParams)
{
	const char *cmdArgs[32];
	int numCmdArgs = 0;

	cmdArgs[numCmdArgs ++] = "pppd";
	cmdArgs[numCmdArgs ++] = serialPort;

	if (pParams->useAuth)
	{
		cmdArgs[numCmdArgs ++] = "user";
		cmdArgs[numCmdArgs ++] = pParams->gprsUser;

		cmdArgs[numCmdArgs ++] = "password";
		cmdArgs[numCmdArgs ++] = pParams->gprsPass;
	}

	cmdArgs[numCmdArgs ++] = "nodetach";

	if (pParams->debugMode)
	{
		cmdArgs[numCmdArgs ++] = "debug";
		cmdArgs[numCmdArgs ++] = "record";
		cmdArgs[numCmdArgs ++] = "/tmp/traffic.log";

		cmdArgs[numCmdArgs ++] = "logfile";
		cmdArgs[numCmdArgs ++] = "/tmp/pppd_debug.log";
	}

	if(pParams->persistMode)
	{
		cmdArgs[numCmdArgs ++] = "persist";
		cmdArgs[numCmdArgs ++] = "maxfail";
		cmdArgs[numCmdArgs ++] = "0";
	}

	cmdArgs[numCmdArgs ++] = "call";
	cmdArgs[numCmdArgs ++] = GPRS_PROFILE_NAME;
	cmdArgs[numCmdArgs ++] = NULL;

	int pid = fork();
	if (pid == -1)
	{
		debug("failed forking process (%d)\n", errno);
		return false;
	}

	if (pid == 0)
	{
		setpgrp();
		if (execvp(DAEMON_EXECUTABLE, (char**) cmdArgs) < 0)
			debug("failed starting process (%d)\n", errno);
		_exit(-1);
	}

	debug("launched process: %d\n", pid);

	int retCount;
	for (retCount = DAEMON_START_TOUT / 2; retCount > 0; retCount --)
	{
		int fpid = getDaemonPID();
		debug("Waiting pppd to start, retCount = %d, PID = %d\n", retCount, fpid);
		if (fpid >= 0)
		{
			int s = socket(AF_INET, SOCK_STREAM, 0);
			if (s > 0)
			{
				struct ifreq ifr;
				strncpy(ifr.ifr_name, DAEMON_INTERFACE, IFNAMSIZ);

				// poll interface flags
				if (ioctl(s, SIOCGIFFLAGS, (char*) &ifr) == 0)
				{
					if (ifr.ifr_flags & IFF_UP)
					{
						close(s);
						break;
					}
				}
				else debug("failed getting interface flags\n");
				close(s);
			}
			else debug("failed creating socket\n");
		}

		usleep(200 * 1000);
	}

	if (retCount == 0)
	{
		debug("daemon start timeout\n");

		bool ret = kill(pid, SIGTERM) == 0;
		debug("stopping pid %d: %s (%d)\n", pid, ret ? "ok" : "failure", errno);

		retCount = 0;
		if (ret) // did the process receive the signal?
		{
			for (retCount = DAEMON_STOP_TOUT / 2; retCount > 0; retCount --)
			{
				// check that the process is running
				if (waitpid(pid, NULL, WNOHANG) > 0)
					break;
				usleep(200 * 1000);
			}
		}

		// correct termination failed, now kill the process
		if (retCount == 0)
		{
			debug("terminating daemon\n");
			kill(pid, SIGKILL);
		}

		// wait until the daemon quits
		waitpid(pid, NULL, 0);
		unlink(DAEMON_PID_FILE);
		return false;
	}

	debug("daemon PID = %d\n", getDaemonPID());
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop GPRS PPP
 *
 * Stops the GPRS PPP daemon (pppd), terminates PPP mode and releases the
 * selected serial port.
 *
 * \return Returns true if the PPP daemon was stopped successfully.
 */
bool GprsPpp::stop()
{
	int pid = getDaemonPID();
	if (pid < 0) return false;

	bool ret = kill(pid, SIGTERM) == 0;
	debug("stopping pppd pid %d: %s (%d)\n", pid, ret ? "ok" : "failure", errno);

	int retCount = 0;
	if (ret) // did the process receive the signal?
	{
		for (retCount = DAEMON_STOP_TOUT / 2; retCount > 0; retCount --)
		{
			// check that the process is running
			if (getDaemonPID() < 0) break;
			usleep(200 * 1000);
		}
	}

	// correct termination failed, now kill the process
	if (retCount == 0)
	{
		debug("terminating daemon\n");
		kill(pid, SIGKILL);
		unlink(DAEMON_PID_FILE);
	}

	// wait until the daemon quits
	waitpid(pid, NULL, 0);
	debug("daemon stopped\n");
	return true;

}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Check GPRS PPP status
 *
 * \return Returns true if GPRS PPP is active.
 */
bool GprsPpp::isRunning()
{
	return getDaemonPID() >= 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get the PID number of the daemon
 * \return Returns daemon PID number
 */
int GprsPpp::getDaemonPID()
{
	std::ifstream pidFile(DAEMON_PID_FILE);
	if (pidFile.is_open())
	{
		int pid = 0;
		pidFile >> pid;
		pidFile.close();
		return pid;
	}

	return -1;
}
