/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>

#include "Gsm0710Buffer.h"

//#define DEBUG
#include "Debug.h"

static const unsigned char CHECKSUM_TABLE[256] =
{
	0x00, 0x91, 0xE3, 0x72, 0x07, 0x96, 0xE4, 0x75, 0x0E, 0x9F, 0xED, 0x7C, 0x09, 0x98, 0xEA, 0x7B,
	0x1C, 0x8D, 0xFF, 0x6E, 0x1B, 0x8A, 0xF8, 0x69, 0x12, 0x83, 0xF1, 0x60, 0x15, 0x84, 0xF6, 0x67,
	0x38, 0xA9, 0xDB, 0x4A, 0x3F, 0xAE, 0xDC, 0x4D, 0x36, 0xA7, 0xD5, 0x44, 0x31, 0xA0, 0xD2, 0x43,
	0x24, 0xB5, 0xC7, 0x56, 0x23, 0xB2, 0xC0, 0x51, 0x2A, 0xBB, 0xC9, 0x58, 0x2D, 0xBC, 0xCE, 0x5F,
	0x70, 0xE1, 0x93, 0x02, 0x77, 0xE6, 0x94, 0x05, 0x7E, 0xEF, 0x9D, 0x0C, 0x79, 0xE8, 0x9A, 0x0B,
	0x6C, 0xFD, 0x8F, 0x1E, 0x6B, 0xFA, 0x88, 0x19, 0x62, 0xF3, 0x81, 0x10, 0x65, 0xF4, 0x86, 0x17,
	0x48, 0xD9, 0xAB, 0x3A, 0x4F, 0xDE, 0xAC, 0x3D, 0x46, 0xD7, 0xA5, 0x34, 0x41, 0xD0, 0xA2, 0x33,
	0x54, 0xC5, 0xB7, 0x26, 0x53, 0xC2, 0xB0, 0x21, 0x5A, 0xCB, 0xB9, 0x28, 0x5D, 0xCC, 0xBE, 0x2F,
	0xE0, 0x71, 0x03, 0x92, 0xE7, 0x76, 0x04, 0x95, 0xEE, 0x7F, 0x0D, 0x9C, 0xE9, 0x78, 0x0A, 0x9B,
	0xFC, 0x6D, 0x1F, 0x8E, 0xFB, 0x6A, 0x18, 0x89, 0xF2, 0x63, 0x11, 0x80, 0xF5, 0x64, 0x16, 0x87,
	0xD8, 0x49, 0x3B, 0xAA, 0xDF, 0x4E, 0x3C, 0xAD, 0xD6, 0x47, 0x35, 0xA4, 0xD1, 0x40, 0x32, 0xA3,
	0xC4, 0x55, 0x27, 0xB6, 0xC3, 0x52, 0x20, 0xB1, 0xCA, 0x5B, 0x29, 0xB8, 0xCD, 0x5C, 0x2E, 0xBF,
	0x90, 0x01, 0x73, 0xE2, 0x97, 0x06, 0x74, 0xE5, 0x9E, 0x0F, 0x7D, 0xEC, 0x99, 0x08, 0x7A, 0xEB,
	0x8C, 0x1D, 0x6F, 0xFE, 0x8B, 0x1A, 0x68, 0xF9, 0x82, 0x13, 0x61, 0xF0, 0x85, 0x14, 0x66, 0xF7,
	0xA8, 0x39, 0x4B, 0xDA, 0xAF, 0x3E, 0x4C, 0xDD, 0xA6, 0x37, 0x45, 0xD4, 0xA1, 0x30, 0x42, 0xD3,
	0xB4, 0x25, 0x57, 0xC6, 0xB3, 0x22, 0x50, 0xC1, 0xBA, 0x2B, 0x59, 0xC8, 0xBD, 0x2C, 0x5E, 0xCF
};

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the GSM 07.10 multiplexer buffer class.
 */
Gsm0710Buffer::Gsm0710Buffer()
{
	pRead = pWrite = dataBuffer;
	memset(dataBuffer, 0, MAX_BUFFER);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Terminates the GSM 07.10 multiplexer buffer class.
 */
Gsm0710Buffer::~Gsm0710Buffer()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Calculate checksum
 *
 * Calculates the checksum according to GSM 07.10 multiplexer specification.
 *
 * \param data Points to the data buffer.
 * \param len Specifies the length of the data in bytes.
 * \return Returns the checksum byte.
 */
unsigned char Gsm0710Buffer::makeCheckSum(const unsigned char *data, size_t len)
{
	unsigned char fcs = 0xff;
	while (len--)
		fcs = CHECKSUM_TABLE[fcs ^ *data++];
	return (0xff - fcs);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read frame from buffer
 *
 * Extracts a new frame from the GSM 07.10 buffer.
 *
 * \return Returns a pointer to the newly created frame class.
 *
 * \note If the buffer does not contain any frames, the NULL pointer is returned.
 */
Gsm0710Frame* Gsm0710Buffer::read()
{
	debug("read %p write %p\n", pRead, pWrite);

	// first, locate the beginning of frame
	bool foundFlag = false;
	while (getAllocated() > 0)
	{
		if (*pRead == FLAG)
		{
			foundFlag = true;
			break;
		}

		// seek to a next byte
		pRead = advanceBufPtr(pRead);
		debug("seeking start flag\n");
	}

	// do we have a started frame?
	if (!foundFlag)
	{
		debug("error locating frame start\n");
		return NULL;
	}

	// save the current read pointer, restore later if the complete frame
	// has not been received yet
	unsigned char *pHead = pRead;

	// now ignore empty frames
	while (getAllocated() > 0 && (*pRead == FLAG))
		pRead = advanceBufPtr(pRead);

	Gsm0710Frame *pFrame = NULL;
	// minimum acceptable frame size is 5 bytes (channel, type, length, fcs, flag)
	if (getAllocated() >= 5)
	{
		pFrame = new Gsm0710Frame();
		unsigned char fcs = 0xff;

		pFrame->dlci = (*pRead & 0xfc) >> 2;
		fcs = CHECKSUM_TABLE[fcs ^ *pRead];
		pRead = advanceBufPtr(pRead);

		pFrame->control = *pRead;
		fcs = CHECKSUM_TABLE[fcs ^ *pRead];
		pRead = advanceBufPtr(pRead);

		pFrame->length = (*pRead & 0xfe) >> 1;
		fcs = CHECKSUM_TABLE[fcs ^ *pRead];
		if ((*pRead & 0x01) == 0)
		{
			delete pFrame;
			debug("length of 2-byte unsupported, skipping frame\n");
			return read();
		}

		pRead = advanceBufPtr(pRead);
		if (getAllocated() < pFrame->length + 2)
		{
			delete pFrame;
			debug("incomplete frame, rewind\n");
			pRead = pHead;
			return NULL;
		}

		// parse frame data
		if (pFrame->length > 0)
		{
			pFrame->data = new unsigned char[pFrame->length];
			if (pFrame->data)
			{
				size_t bufTail = size_t((dataBuffer - pRead) + MAX_BUFFER);
				if (pFrame->length > bufTail)
				{
					memcpy(pFrame->data, pRead, bufTail);
					memcpy(pFrame->data + bufTail, dataBuffer, pFrame->length
							- bufTail);
					pRead = dataBuffer + (pFrame->length - bufTail);
				}
				else
				{
					memcpy(pFrame->data, pRead, pFrame->length);
					pRead += pFrame->length;
					if (pRead == dataBuffer + MAX_BUFFER)
						pRead = dataBuffer;
				}

				if (pFrame->matchControl(CONTROL_UI))
				{
					for (size_t i = 0; i < pFrame->length; i++)
						fcs = CHECKSUM_TABLE[fcs ^ pFrame->data[i]];
				}
			}
			else
			{
				debug("out of memory\n");
				pFrame->length = 0;
			}
		}

		// validate frame checksum
		if (CHECKSUM_TABLE[fcs ^ *pRead] != 0xcf)
		{
			debug("checksum error, frame dropped: %x\n", CHECKSUM_TABLE[fcs ^ *pRead]);
			delete pFrame;
			return read();
		}
		else
		{
			// validate end of frame flag
			pRead = advanceBufPtr(pRead);
			if (*pRead != FLAG)
			{
				delete pFrame;
				debug("frame is not properly terminated: %x\n", *pRead);
				pRead = pHead;
				return NULL;
			}

			pRead = advanceBufPtr(pRead);
		}
	}
	else
	{
		debug("incomplete frame, rewind\n");
		pRead = pHead;
	}

	return pFrame;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write raw data to buffer
 *
 * Appends raw GSM 07.10 multiplexer stream data to the buffer.
 * \note The data is written partially if the buffer does not have enough free capacity.
 *
 * \param data Points to the data buffer.
 * \param len Specifies the length of data in bytes.
 * \return Returns the number of bytes written to the buffer.
 */
size_t Gsm0710Buffer::write(const unsigned char *data, size_t len)
{
	size_t freeEnd = (size_t)((dataBuffer - pWrite) + MAX_BUFFER);
	size_t freeTotal = getFree();

	if (len > freeTotal)
		len = freeTotal;

	debug("read %p write %p size %d, tail %d, free %d\n",
			pRead, pWrite, len, freeEnd, freeTotal);

	if (len > freeEnd)
	{
		memcpy(pWrite, data, freeEnd);
		memcpy(dataBuffer, data + freeEnd, len - freeEnd);
		pWrite = dataBuffer + (len - freeEnd);
	}
	else
	{
		memcpy(pWrite, data, len);
		pWrite += len;

		if (pWrite == dataBuffer + MAX_BUFFER)
			pWrite = dataBuffer;
	}

	return len;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Advance the pointer to buffer
 *
 * Moves the pointer to the buffer forward.
 *
 * \param ptr Pointer to the buffer
 * \return Returns the new pointer location
 */
unsigned char* Gsm0710Buffer::advanceBufPtr(unsigned char *ptr)
{
	ptr++;
	if (ptr >= dataBuffer + MAX_BUFFER)
		ptr = dataBuffer;
	return ptr;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get buffer free space in bytes
 *
 * \return Returns the total size of the unallocated part of the buffer,
 * i.e. the number of bytes that would currently fit in the buffer.
 */
size_t Gsm0710Buffer::getFree()
{
	ptrdiff_t d = pRead - pWrite;
	if (pRead > pWrite)
		return (size_t)(d - 1);
	else
		return (size_t)(d + MAX_BUFFER - 1);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get buffer allocated space in bytes
 *
 * \return Returns the number of data bytes currently allocated in the buffer.
 */
size_t Gsm0710Buffer::getAllocated()
{
	ptrdiff_t d = pRead - pWrite;
	if (pRead > pWrite)
		return (size_t)(MAX_BUFFER - d);
	else
		return (size_t) -d;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the GSM 07.10 multiplexer frame object.
 */
Gsm0710Frame::Gsm0710Frame()
{
	dlci = 0;
	control = 0;
	data = NULL;
	length = 0;
}

/**
 * \brief Class destructor
 *
 * Terminates the GSM 07.10 multiplexer frame object.
 */
Gsm0710Frame::~Gsm0710Frame()
{
	if (data != NULL)
		delete[] data;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Compare UIH control frame type
 *
 * Compare the stored UIH control frame type with the one specified
 * \param type The frame type that the stored frame should match with
 *
 * \return Returns true if the specified control channel frame type matches the
 * currently stored frame type.
 */
bool Gsm0710Frame::matchCmdType(unsigned char type)
{
	if (data == NULL)
		return false;
	return ((data[0] & ~CR) == type);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Compare CMUX frame type
 *
 * Compare the stored CMUX frame type with the one specified
 * \param ctl The frame type that the stored frame should match with
 *
 * \return Returns true if the specified frame type matches the currently stored
 * frame type.
 */
bool Gsm0710Frame::matchControl(unsigned char ctl)
{
	return ((control & ~PF) == ctl);
}
