/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include <sys/ioctl.h>

#include <platform.h>

#include "Gsm0710Mux.h"

//#define DEBUG
//#define DEBUG_DUMP
//#define DEBUG_DUMP_HEX
//#define DEBUG_DUMP_ASCII
#include "Debug.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the GSM serial port multiplexer object.
 */
Gsm0710Mux::Gsm0710Mux()
{
	muxedPortNum = 0;
	physicalHandle = -1;

	strcpy(physicalPath, "");

	// clear virtual port handles and names
	for (int i = 0; i < VIRTUAL_PORTS_MAX; i ++)
	{
		strcpy(muxedPortMaster[i], "");
		muxedPortHandles[i] = -1;
	}

	// clear virtual channel state
	memset(channelState, 0, sizeof(channelState));
}

/**
 * \brief Class destructor
 *
 * Terminates the GSM serial port multiplexer object.
 */
Gsm0710Mux::~Gsm0710Mux()
{

}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start port multiplexer
 *
 * Initializes and launches the serial port multiplexer.
 * The GPRS module must be properly initialized before calling the multiplexer.
 * Prior initialization must include the following steps.
 * - Power up the module
 * - Serial port setup
 * - Initial handshake over serial port
 *
 * The above procedure is implemented by following classes.
 * - PowerManager, see gprsPowerOn() method
 * - GsmModule, see open() and initComm() methods
 *
 * The virtual serial ports created by the multiplexer are accessible via
 * /dev/ttyp0...ttyp3 devices. The opposite ends of these serial pipes must be
 * passed to the multiplexer as a parameter, /dev/ptyp0...ptyp3.
 *
 * \param physicalPort Points to the null-terminated string containing the name
 * of the modem physical serial port.
 * \param muxPorts Points to a null-terminated array of null-terminated strings
 * containing the names of the virtual serial ports. The application should
 * specify at least one port.
 * \return Returns true if the port multiplexer was started successfully.
 */
bool Gsm0710Mux::startMux(const char *physicalPort, const char **muxPorts)
{
	// sanity check, avoid double initialization
	if (physicalHandle != -1) return false;

	// open the physical (non-multiplexed) modem port
	if (!openPhysicalPort(physicalPort))
	{
		debug("unable to open modem port: %s\n", physicalPort);
		return false;
	}

	strncpy(physicalPath, physicalPort, sizeof(physicalPath));

	// initialize requested virtual channels
	for (int i = 0; i < VIRTUAL_PORTS_MAX && muxPorts[i] != NULL; i ++)
	{
		strncpy(muxedPortMaster[i], muxPorts[i], VIRTUAL_PATH_LEN);
		muxedPortNum ++;
	}

#ifdef SUPPORT_POWER_MANAGEMENT
		// register mux instance as a power manager listener
		PowerManager *pPM = PowerManager::getPowerManager();
		pPM->addListener(this);
#endif

	for (int i = 0; i < muxedPortNum; i++)
	{
		if (muxedPortHandles[i] < 0)
			openVirtualPort(i);
	}

	Thread::start();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop port multiplexer
 *
 * Stops the GSM serial port multiplexer and releases working thread.
 * All virtual ports are closed, further communication is not possible.
 *
 * \return Returns true if the multiplexer was stopped successfully.
 */
bool Gsm0710Mux::stopMux()
{
#ifdef SUPPORT_POWER_MANAGEMENT
	// unregister mux instance as a power manager listener
	PowerManager *pPM = PowerManager::getPowerManager();
	pPM->removeListener(this);
#endif

	// set the thread termination flag and join the thread
	Thread::stop();

	// close physical serial port
	if (physicalHandle != -1)
	{
		struct termios attr;
		// disable flow control
		tcgetattr(physicalHandle, &attr);
		attr.c_cflag &= ~CRTSCTS;
		tcsetattr(physicalHandle, TCSANOW, &attr);
		// flush the buffers
		tcflush(physicalHandle, TCIOFLUSH);
		// finally, close the port
		close(physicalHandle);
		physicalHandle = -1;
	}

	// clear multiplexer state variables
	muxedPortNum = 0;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Opens the modem physical serial port
 * Opens the serial port of the physical modem device.
 *
 * \param port Pointer to the null-terminated string with port name
 *
 * \return Returns true if the physical port was opened successfully
 */
bool Gsm0710Mux::openPhysicalPort(const char *port)
{
	// open serial port for the given name
	physicalHandle = open(port, O_RDWR| O_NOCTTY | O_NDELAY);

	if (physicalHandle != -1)
	{
		struct termios termCfg;
		fcntl(physicalHandle, F_SETFL, 0);

		// read terminal configuration
		tcgetattr(physicalHandle, &termCfg);

		// local mode, receiver enabled
		termCfg.c_cflag |= (CLOCAL | CREAD);

		// 8N1, no parity mode
		termCfg.c_cflag &= ~PARENB;
		termCfg.c_cflag &= ~CSTOPB;
		termCfg.c_cflag &= ~CSIZE;
		termCfg.c_cflag |= CS8;

		// hardware flow control enabled
		termCfg.c_cflag |= CRTSCTS;

		// clear flags
		termCfg.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
		termCfg.c_iflag &= ~(INLCR | ICRNL | IGNCR);

		termCfg.c_oflag &= ~OPOST;
		termCfg.c_oflag &= ~OLCUC;
		termCfg.c_oflag &= ~ONLRET;
		termCfg.c_oflag &= ~ONOCR;
		termCfg.c_oflag &= ~OCRNL;

		// apply new settings to tty
		tcsetattr(physicalHandle, TCSANOW, &termCfg);
		return true;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Opens the virtual port
 * Opens the specified virtual serial port
 *
 * \param portNumber Virtual port number to open
 * \return Returns true if virtual port was opened successfully, false otherwise
 */
bool Gsm0710Mux::openVirtualPort(int portNumber)
{
	muxedPortHandles[portNumber] =
			open(muxedPortMaster[portNumber], O_RDWR | O_NONBLOCK);

	if (muxedPortHandles[portNumber] < 0)
	{
		debug("unable to open virtual port: %s\n", muxedPortMaster[portNumber]);
		return false;
	}

	char *devName = ptsname(muxedPortHandles[portNumber]);

	if (devName == NULL)
	{
		debug("unable to read pty name: %s\n", muxedPortMaster[portNumber]);
		return false;
	}

	// read terminal settings
	termios tio;
	tcgetattr(muxedPortHandles[portNumber], &tio);
	// set output flags
	tio.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);
	tio.c_iflag &= ~(INLCR | ICRNL | IGNCR);

	// clear flags
	tio.c_oflag &= ~OPOST;
	tio.c_oflag &= ~OLCUC;
	tio.c_oflag &= ~ONLRET;
	tio.c_oflag &= ~ONOCR;
	tio.c_oflag &= ~OCRNL;

	// apply new settings to tty
	tcsetattr(muxedPortHandles[portNumber], TCSANOW, &tio);

	// corner case
	if (strcmp(devName, "/dev/ptmx") == 0)
	{
		// change owner and grant access
		grantpt(muxedPortHandles[portNumber]);
		unlockpt(muxedPortHandles[portNumber]);
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Closes the virtual port
 * Closes the specified virtual port
 *
 * \param portNumber Virtual port number to close
 * \return Returns true if virtual port was closed successfully, false otherwise
 */

bool Gsm0710Mux::closeVirtualPort(int portNumber)
{
	bool result = true;
	if (muxedPortHandles[portNumber] >= 0)
	{
		result = (close(muxedPortHandles[portNumber]) == 0);
		if (result)
		{
			debug("Closing channel number %d with handle %d SUCCEEDED\n",
					portNumber, muxedPortHandles[portNumber]);
		}
		else
		{
			debug("Closing channel number %d with handle %d FAILED\n",
					portNumber, muxedPortHandles[portNumber]);
		}
		muxedPortHandles[portNumber] = -1;
	}
	return result;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write data to physical port
 * Writes data to the modem device physical port
 *
 * \param channel Channel number that will be used for data transmission
 * \param input Data being sent
 * \param count Data length
 * \param type Frame type to be sent
 *
 * \return Returns number of bytes written
 */
size_t Gsm0710Mux::writeToPhysical(int channel, unsigned char *input,
		size_t count, unsigned char type)
{
	unsigned char prefix[5] =
	{ FLAG, EA | CR, 0, 0, 0 };
	unsigned char postfix[2] =
	{ 0xFF, FLAG};
	size_t prefix_length = 4, c;

	mutex.lock();

	// prepare GSM 07.10 header data
	prefix[1] = prefix[1] | ((63 & (unsigned char) channel) << 2);
	prefix[2] = type;

	// crop the frame
	count = count > MAX_FRAME_LEN ? MAX_FRAME_LEN : count;
	debug("channel %d: type %x, %d bytes\n", channel, type, count);

	// prepare length field
	if (count > 127)
	{
		prefix_length = 5;
		prefix[3] = ((127 & count) << 1);
		prefix[4] = (32640 & count) >> 7;
	}
	else
	{
		prefix[3] = 1 | (count << 1);
	}

	// prepare checksum value
	postfix[0] = Gsm0710Buffer::makeCheckSum(prefix + 1, prefix_length - 1);

	// write the header to port
	c = write(physicalHandle, prefix, prefix_length);
	if (channel == 0) debug("prefix channel %d: %d bytes (%d), type %d\n",
			channel, prefix_length, c, type);

	if (c != prefix_length)
	{
		debug("unable to write prefix, channel %d; wrote %d bytes\n", channel,
				c);
		mutex.unlock();
		return 0;
	}

	// if there is data available, write it
	if (count > 0)
	{
		c = write(physicalHandle, input, count);
		if (channel == 0) debug("data chan %d: %d/%d bytes\n", channel, count,
				c);

		if (count != c)
		{
			debug("unable to write payload, channel %d; wrote %d bytes\n",
					channel, c);
			mutex.unlock();
			return 0;
		}
	}

	// write the terminator
	c = write(physicalHandle, postfix, 2);
	if (c != 2)
	{
		debug("unable to write terminator, channel %d; wrote %d bytes.",
				channel, c);
		mutex.unlock();
		return 0;
	}

	mutex.unlock();
	return count;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get the data from virtual port
 * Read the data up to count from virtual port and write it to the buffer buf
 *
 * \param buf Pointer to the buffer to store virtual data
 * \param count Buffer size
 * \param port Virtual port number to read data from
 *
 * \return Returns number of bytes written to the buffer
 */
size_t Gsm0710Mux::handleVirtualData(unsigned char *buf, size_t count, int port)
{
	size_t written = 0;
	while (written < count)
	{
		// write the data in loop
		size_t remaining = count - written;
		size_t toWrite = remaining > MAX_FRAME_LEN ? MAX_FRAME_LEN : remaining;
		written += writeToPhysical(port + 1, buf + written, toWrite,
				CONTROL_UIH);
	}

	return written;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle data frames stored in buffer
 * Handle frames stored GSM 07.10 buffer
 *
 * \param inBuf Address of the buffer where the frames are stored
 * \param stopRequest Stop type flag requested to process
 *
 * \return Returns the number of frames extracted
 */
size_t Gsm0710Mux::handleSerialFrames(Gsm0710Buffer &inBuf, StopType &stopRequest)
{
	size_t framesExtracted = 0;
	Gsm0710Frame *pFrame;

	while ((pFrame = inBuf.read()) != NULL)
	{
		++ framesExtracted;
		debug("extracted frame %d (%d bytes): dlci %d, control %x\n",
				framesExtracted, pFrame->length, pFrame->dlci, pFrame->control);

		if (pFrame->matchControl(CONTROL_UI) || pFrame->matchControl(
				CONTROL_UIH))
		{
			// it is an information frame
			if (pFrame->dlci > 0)
			{
				// incoming data from logical channel
				writeToVirtual(pFrame->data, pFrame->length, pFrame->dlci);
			}
			else
			{
				// incoming command from control channel
				handleCommand(pFrame, stopRequest);
			}
		}
		else
		{
			// it is not an information frame
			switch (pFrame->control & ~PF)
			{
			case CONTROL_UA: // unnumbered acknowledgement

				if (channelState[pFrame->dlci].channelOpen)
				{
					debug("closing DLCI %d\n", pFrame->dlci);
					channelState[pFrame->dlci].channelOpen = false;
				}
				else
				{
					channelState[pFrame->dlci].channelOpen = true;
					if (pFrame->dlci == 0)
					{
						debug("opened control channel\n");
					}
					else
					{
						debug("opened DLCI %d\n", pFrame->dlci);
					}
				}
				break;

			case CONTROL_DM: // disconnected mode

				if (channelState[pFrame->dlci].channelOpen)
				{
					debug("DM received on channel %d\n", pFrame->dlci);
					channelState[pFrame->dlci].channelOpen = false;
				}
				else
				{
					if (pFrame->dlci == 0)
					{
						debug("DM on control channel, shutting down\n");
						stopRequest = STOPTYPE_NORMAL;
					}
					else
					{
						debug("could not open DLCI %d\n", pFrame->dlci);
					}
				}
				break;

			case CONTROL_DISC: // disconnect

				if (channelState[pFrame->dlci].channelOpen)
				{
					channelState[pFrame->dlci].channelOpen = false;
					writeToPhysical(pFrame->dlci, NULL, 0, CONTROL_UA| PF);
					if (pFrame->dlci == 0)
					{
						debug("closing control channel\n");
						stopRequest = STOPTYPE_SKIP_MODEM_CLEANUP;
					}
					else
					{
						debug("closed DLCI %d\n", pFrame->dlci);
					}
				}
				else
				{
					// the channel is closed already
					debug("DISC received but DLCI %d was already closed.\n",
							pFrame->dlci);
					writeToPhysical(pFrame->dlci, NULL, 0, CONTROL_DM| PF);
				}
				break;

			case CONTROL_SABM: // channel open request

				if (!channelState[pFrame->dlci].channelOpen)
				{
					if (pFrame->dlci == 0)
					{
						debug("opened control channel\n");
					}
					else
					{
						debug("opened DLCI %d\n", pFrame->dlci);
					}
				}
				else
				{
					// the channel is opened already
					debug("SABM received, but DLCI %d was already closed.\n",
							pFrame->dlci);
				}

				channelState[pFrame->dlci].channelOpen = true;
				writeToPhysical(pFrame->dlci, NULL, 0, CONTROL_UA| PF);
				break;
			}
		}

		delete pFrame;
	}

	return framesExtracted;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle command retrieved from frame
 * Handle the command stored in the GSM 07.10 frame
 *
 * \param pFrame Pointer to the frame that will be processed
 * \param stopRequest Stop type flag requested to process
 */
void Gsm0710Mux::handleCommand(Gsm0710Frame *pFrame, StopType &stopRequest)
{
	unsigned char type, signals;
	size_t length = 0, i, type_length;

	if (pFrame->length > 0)
	{
		// only single-byte types are supported
		type = pFrame->data[0];
		// skip bytes
		for (i = 0; (pFrame->length > i && (pFrame->data[i] & EA) == 0); i ++)
			;
		i ++;
		type_length = i;

		// extract the length of frame
		while (pFrame->length > i)
		{
			length = (length * 128) + ((pFrame->data[i] & 254) >> 1);
			if ((pFrame->data[i] & 1) == 1) break;
			i ++;
		}
		i ++;

		// handle MSC command
		if (pFrame->matchCmdType(TYPE_MSC))
		{
			if (i + 1 < pFrame->length)
			{
				int channel = ((pFrame->data[i ++] & 252) >> 2);
				if ((channel > 0) &&
					(channel <= muxedPortNum) &&
					(muxedPortHandles[channel - 1] >= 0))
				{
					signals = (pFrame->data[i]);
					debug("Flow control not supported, ignoring\n");
				}
			}
			else
			{
				debug("broken modem status command: %d, length %d, data %d\n",
						i, length, pFrame->length);
			}
		}
		// handle command
		else if ((type & CR) == CR)
		{
			bool isSupported = true;

			switch ((type & ~CR))
			{
			case TYPE_CLD:
				debug("received CLD from mobile station\n");
				stopRequest = STOPTYPE_SKIP_MODEM_CLEANUP;
				break;

			case TYPE_TEST:
				break;

			default:

				debug("received unknown command: %d\n", type);
				unsigned char *pResponse = new unsigned char[2 + type_length];
				pResponse[0] = TYPE_NSC;

				pResponse[1] = EA & ((127 & type_length) << 1);

				i = 2;
				while (type_length --)
				{
					pResponse[i] = pFrame->data[(i - 2)];
					i ++;
				}

				writeToPhysical(0, pResponse, i, CONTROL_UIH);
				delete[] pResponse;
				isSupported = false;
				break;
			}

			if (isSupported)
			{
				// send command acknowledgement
				pFrame->data[0] = pFrame->data[0] & ~CR;
				writeToPhysical(0, pFrame->data, pFrame->length, CONTROL_UIH);
			}
		}
		else
		{
			// received command acknowledgement
			if (pFrame->matchCmdType(TYPE_NSC))
			{
				debug("MS did not ack the command\n");
			}
			else
			{
				debug("command acked by MS\n");
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write data to virtual port
 * Writes data into multiplexer virtual port
 *
 * \param buf Pointer to buffer that contains data being written
 * \param count Buffer length
 * \param channel Channel number where to send the data
 *
 * \return Returns number of bytes written
 */
size_t Gsm0710Mux::writeToVirtual(unsigned char *buf, size_t count, int channel)
{
	if ((channel < 1) ||
		(channel > muxedPortNum) ||
		(muxedPortHandles[channel - 1] < 0))
	{
		return 0;
	}

	if (!isOutputReady(muxedPortHandles[channel - 1], 1000))
		return 0;
	return write(muxedPortHandles[channel - 1], buf, count);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize multiplexer
 * Initializes the multiplexer by sending the appropriate AT commands and prepares the
 * control channel and logical channels.
 *
 * \return Returns true if initialization was successful, false otherwise
 */
bool Gsm0710Mux::initMux()
{
	debug("initializing CMUX\n");
	if (!sendAT("AT\r", 200))
	{
		debug("no response to init AT command!\n");
		return false;
	}

	debug("entering MUX mode\n");
	if (!sendAT("AT+CMUX=0\r", 200))
	{
		debug("unable to enter multiplexed mode\n");
		return false;
	}

	debug("init control channel...\n");
	writeToPhysical(0, NULL, 0, CONTROL_SABM| PF);

	debug("init logical channels...\n");
	unsigned char v24[] =
	{ TYPE_MSC| CR, 5, 0, V24_RTC| V24_RTR };

	for (int i = 1; i <= muxedPortNum; i ++)
	{
		writeToPhysical(i, NULL, 0, CONTROL_SABM| PF);
		debug("connecting to virtual channel %d on port %s\n", i, physicalPath);

		v24[2] = (i << 2) | EA | CR;
		writeToPhysical(0, v24, sizeof(v24), CONTROL_UIH);
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send the appropriate AT command
 * Sends the AT commands to physical modem device
 *
 * \param cmd Pointer to null-terminated command being sent
 * \param tout Timeout for waiting the command to be processed
 *
 * \note If timeout is 0, there is no waiting. If timeout is -1, the polling request
 * blocks until the event occurs.
 *
 * \return Returns true if command was accepted, false otherwise
 */
bool Gsm0710Mux::sendAT(const char *cmd, int tout)
{
	char buf[128];

	// add a safety delay between modem AT commands
	usleep(20 * 1000);

	int wrote = write(physicalHandle, cmd, strlen(cmd));
	if (wrote < 0) return false;

	debug("tcdrain... ");
	tcdrain(physicalHandle);
	debug("done\n");

	pollfd pfd;
	pfd.fd = physicalHandle;
	pfd.events = POLLIN;
	pfd.revents = 0;

	if (poll(&pfd, 1, tout) > 0)
	{
		if ((pfd.revents & POLLIN) != 0)
		{
			Thread::sleep(100);
			memset(buf, 0, sizeof(buf));

			debug("read\n");

			int len = read(physicalHandle, buf, sizeof(buf));
			buf[len] = 0;

			if (len > 0 && strstr(buf, "OK\r\n")) return true;
		}
	}

	debug("exit\n");
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Detect if the output file descriptor is ready or not
 * Determines if the output file descriptor is ready or not within the timeOut period
 *
 * \param fd File descriptor that is polled
 * \param timeOut Timeout in milliseconds the polling waits the file descriptor to be ready
 *
 * \note If timeout is 0, there is no waiting. If timeout is -1, the polling request
 * blocks until the event occurs.
 *
 * \return Returns true if the file descriptor is ready, false otherwise
 */

bool Gsm0710Mux::isOutputReady(int fd, int timeOut)
{
	struct pollfd pfd;
	pfd.fd = fd;
	pfd.events = POLLOUT;
	pfd.revents = 0;
	return (poll(&pfd, 1, timeOut) == 1) &&
			((pfd.revents & POLLOUT) != 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get opened file descriptors
 * Get the list of opened file descriptors within timeout
 *
 * \param readyFds Pointer to array where to store the available file descriptors
 * \param timeout Number of milliseconds the polling function will wait for file descriptor to be ready
 *
 * \note If timeout is 0, there is no waiting. If timeout is -1, the polling request
 * blocks until the event occurs.
 *
 * \return Returns number of file descriptor stored in readyFds
 */

int Gsm0710Mux::getPendingFds(int *readyFds, int timeout)
{
	pollfd pfd[VIRTUAL_PORTS_MAX + 1];

	pfd[0].fd = physicalHandle;
	pfd[0].events = POLLIN;
	pfd[0].revents = 0;
	int count = 1;

	for (int i = 0; i < muxedPortNum; i++)
	{
		if (muxedPortHandles[i] >= 0)
		{
			pfd[count].fd = muxedPortHandles[i];
			pfd[count].events = POLLIN;
			pfd[count].revents = 0;
			count++;
		}
	}

	int resCount = 0;
	if (poll(pfd, count, timeout) > 0)
	{
		for (int i = 0; i < count; i++)
		{
			if ((pfd[i].revents & POLLIN) != 0)
			{
				readyFds[resCount] = pfd[i].fd;
				resCount++;
			}
		}
	}

	return resCount;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle data from physical modem device
 * Handles data coming from physical modem device and stored in GSM 07.10 buffer
 *
 * \param inBuf Address of the GSM 07.10 buffer where the data is stored
 * \param stopRequest Stop type flag requested to process
 */
void Gsm0710Mux::handlePhysicalInput(Gsm0710Buffer &inBuf, StopType &stopRequest)
{
	unsigned char buf[512];
	size_t readMax = inBuf.getFree();
	if (readMax > sizeof(buf))
		readMax = sizeof(buf);
	int len = read(physicalHandle, buf, readMax);

	// new data read from physical port
	if (len > 0)
	{
		debug("received %d bytes on physical port\n", len);
		debug_dump(buf, len);

		inBuf.write(buf, len);
		// parse incoming data and dispatch frames
		handleSerialFrames(inBuf, stopRequest);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle data coming from virtual input device
 * Handles the data stored in virtual input device
 *
 * \param fd File descriptor to the virtual input device where the data is stored
 */
void Gsm0710Mux::handleVirtualInput(int fd)
{
	unsigned char buf[512];
	// read data from a virtual serial port

	int i = 0;
	for (i = 0; i < muxedPortNum; i++)
	{
		if (fd == muxedPortHandles[i])
			break;
	}

	int len = read(fd, buf, sizeof(buf));
	if (len < 0)
	{
		debug("Read from ptty %d failed, errno = %d\n", i, errno);
		Thread::sleep(100);
	}
	else
	{
		debug("received %d bytes on ptty %d\n", len, i);
		debug_dump(buf, len);
		// process data from virtual port
		handleVirtualData(buf, len, i);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Starts the multiplexer thread
 * Starts the multiplexer thread, initializes the multiplexer, allocates the GSM 07.10 buffer and runs
 * polling event loop
 */
void Gsm0710Mux::run()
{
	debug("CMUX thread started\n");

	for (int i = 0; i < muxedPortNum; i++)
	{
		if (muxedPortHandles[i] < 0)
			openVirtualPort(i);
	}

	// initialize the multiplexer
	if (!initMux())
	{
		debug("CMUX init failed\n");
		return;
	}

	// allocate GSM 07.10 frame buffer
	Gsm0710Buffer inBuf;
	StopType stopRequest = STOPTYPE_NONE;

	// start the main loop
	while (!stopRequested())
	{
		int readyFds[VIRTUAL_PORTS_MAX + 1];
		int readyCount = getPendingFds(readyFds, 1000);

		for (int i = 0; i < readyCount; i++)
		{
			if (readyFds[i] == physicalHandle)
			{
				handlePhysicalInput(inBuf, stopRequest);
			}
			else
			{
				handleVirtualInput(readyFds[i]);
			}
		}

		if (stopRequest != STOPTYPE_NONE)
			break;
	}

	debug("Closing virtual channels...\n");

	// close available virtual channels
	for (int i = 0; i < muxedPortNum; i ++)
	{
		closeVirtualPort(i);
	}

	if (stopRequest != STOPTYPE_SKIP_MODEM_CLEANUP)
	{
		debug("Terminating multiplexer channels...");
		// multiplexer must terminate - close all channels
		for (int i = muxedPortNum - 1; i >= 0 ; i--)
		{
			// close virtual channels
			writeToPhysical(i, NULL, 0, CONTROL_DISC | PF);
			Thread::sleep(50);

		}

		for (int i = 0; i < SHUTDOWN_TIMEOUT; i++)
		{
			// finally, close the multiplexer
			unsigned char muxClose[2] = { TYPE_CLD | CR, 1 };
			writeToPhysical(0, muxClose, sizeof(muxClose), CONTROL_UIH);
			Thread::sleep(50);
		}
	}

	debug("Resetting channelStates\n");
	memset(channelState, 0, sizeof(channelState));

	debug("CMUX thread finished\n");
}

#ifdef SUPPORT_POWER_MANAGEMENT
////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle multiplexer power states
 * Handles multiplexer power states if SUPPORT_POWER_MANAGEMENT flag is defined.
 *
 * \param oldState Power state the multiplexer was
 * \param newState Power state the multiplexer will be put into
 *
 * \return Returns true if the change of power state was successful, false otherwise
 */
bool Gsm0710Mux::powerStateChange(PowerState oldState, PowerState newState)
{
	if (newState == LOW_POWER_SLEEP)
	{
		debug("CMUX entering powersave\n");

		Thread::stop();

		PowerManager *pPM = PowerManager::getPowerManager();
		pPM->gprsSetDTR(false);

		/*
		 unsigned char psc[] = { TYPE_PSC | CR, 3, 0 };
		 writeToPhysical(0, psc, sizeof(psc), CONTROL_UIH);

		 usleep(100*1000);

		 int status = 0;
		 ioctl(physicalHandle, TIOCMGET, &status);
		 status &= ~TIOCM_RTS;
		 ioctl(physicalHandle, TIOCMSET, &status);

		 debug("Modem state %x\n", status);
		 */
	}
	else if (oldState == LOW_POWER_SLEEP)
	{
		debug("CMUX leaving powersave\n");

		PowerManager *pPM = PowerManager::getPowerManager();
		pPM->gprsSetDTR(true);

		Thread::start();

		/*
		 int status = 0;
		 ioctl(physicalHandle, TIOCMGET, &status);
		 debug("Modem state %x\n", status);
		 status |= TIOCM_RTS;
		 ioctl(physicalHandle, TIOCMSET, &status);
		 */
	}

	return true;
}

#endif
