#ifndef GSM0710_MUX_H_
#define GSM0710_MUX_H_

#define SUPPORT_POWER_MANAGEMENT	// enables PM support

#include <poll.h>
#include "Thread.h"

#ifdef SUPPORT_POWER_MANAGEMENT
#include "PowerManager.h"
#endif

#include "Gsm0710Buffer.h"

#define VIRTUAL_PORTS_MAX	4
#define VIRTUAL_PATH_LEN	32

#define MAX_FRAME_LEN		31

#define SHUTDOWN_TIMEOUT	5

/** Structure to hold virtual channel state */
struct VIRTUAL_CHANNEL
{
	/** Set to true if the channel is opened */
	bool channelOpen;

	/** Channel state */
	unsigned char stateV24;
};

/**
 * \brief GSM port multiplexer class
 *
 * Implements the GSM 07.10 serial port multiplexer protocol.
 * The class creates multiple serial ports that physically operate on a single
 * serial port of the GPRS modem. The multiplexer class manages an internal
 * worker thread that translates virtual serial port communication.
 *
 * The port multiplexer is a power management aware class. When the mobile data
 * terminal is suspended to a low power state, the port multiplexer is properly
 * de-initialized to prepare the GPRS module for power saving.
 */

#ifndef SUPPORT_POWER_MANAGEMENT
class Gsm0710Mux
#else
class Gsm0710Mux : public PowerEventListener, protected Thread
#endif
{
public:

	Gsm0710Mux();
	virtual ~Gsm0710Mux();

	bool startMux(const char *physicalPort, const char **muxPorts);
	bool stopMux();

protected:

	/** Stopping types that can be requested */
	enum StopType
	{
		/** No stop requested */
		STOPTYPE_NONE,
		/** Requested normal stop (with modem cleanup) */
		STOPTYPE_NORMAL,
		/** Requested stop without modem cleanup */
		STOPTYPE_SKIP_MODEM_CLEANUP
	};

	virtual void run();

	bool initMux();
	bool openPhysicalPort(const char *port);
	bool openVirtualPort(int portNumber);
	bool closeVirtualPort(int portNumber);

	bool sendAT(const char *cmd, int tout);
	bool isOutputReady(int fd, int timeOut);
	int getPendingFds(int *readyFds, int timeout);

	size_t handleVirtualData(unsigned char *buf, size_t count, int port);
	size_t handleSerialFrames(Gsm0710Buffer &inBuf, StopType &stopRequest);

	void handleCommand(Gsm0710Frame *pFrame, StopType &stopRequest);

	size_t writeToPhysical(int channel, unsigned char *input, size_t count,
			unsigned char type);
	size_t writeToVirtual(unsigned char *buf, size_t count, int channel);
	void handlePhysicalInput(Gsm0710Buffer &inBuf, StopType &stopRequest);
	void handleVirtualInput(int fd);

#ifdef SUPPORT_POWER_MANAGEMENT
	virtual bool powerStateChange(PowerState oldState, PowerState newState) ;
#endif

	/** Number of initialized multiplexer ports */
	int muxedPortNum;

	/** Array of handles for initialized multiplexer ports */
	int muxedPortHandles[VIRTUAL_PORTS_MAX];

	/** Array of names for initialized multiplexer ports */
	char muxedPortMaster[VIRTUAL_PORTS_MAX][VIRTUAL_PATH_LEN];

	/** Handle to the physical modem port */
	int physicalHandle;

	/** Name of the physical modem port */
	char physicalPath[VIRTUAL_PATH_LEN];

	/** Array of virtual channel states */
	VIRTUAL_CHANNEL channelState[VIRTUAL_PORTS_MAX+1];

	/** Mutex to protect the multiplexer thread */
	Mutex mutex;
};

#endif /*GSM0710_MUX_H_*/
