/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>

#include "PowerManager.h"

#include "GsmModule.h"

//#define DEBUG
#include "Debug.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the GSM module object.
 *
 * \param portName Points to the null terminated string that contains the
 * modem communication port name.
 */
GsmModule::GsmModule(const char* portName)
{
	commPortName = portName;
	modemPort = NULL;
}

/**
 * \brief Class destructor
 *
 * Terminates the GSM module object.
 */
GsmModule::~GsmModule()
{
	if (modemPort) close();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Open GSM module port
 *
 * Opens the communication port of the GSM module. Communication port name is
 * specified in the class constructor.
 *
 * The communication port is automatically initialized to 115000 bps with
 * hardware flow control.
 *
 * \return Returns true if the communication port was opened successfully.
 */
bool GsmModule::open()
{
	debug("opening\n");
	if (modemPort) return false;

	lock();
	// create and open a new modem port object
	modemPort = new ModemPort();
	if (!modemPort->open(commPortName.c_str()))
	{
		debug("failed opening %s\n", commPortName.c_str());
		delete modemPort;
		modemPort = NULL;
		unlock();
		return false;
	}

	// initialize port settings
	modemPort->setSpeed(115200);
	modemPort->setFlowControl(S_FLOW_NONE);
	modemPort->setCharBits(8);
	modemPort->setStopBits(1);
	modemPort->setParity(S_PARITY_NONE);
	unlock();

	debug("done\n");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close GSM module port
 *
 * Closes the communication port of the GSM module.
 *
 * \return Returns true if the port was closed successfully.
 */
bool GsmModule::close()
{
	debug("closing\n");
	lock();

	if (!modemPort)
	{
		unlock();
		return true;
	}
	modemPort->close();

	delete modemPort;
	modemPort = NULL;
	unlock();

	debug("done\n");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send AT command
 *
 * Writes the specified AT command to the GSM module.
 * The carriage return is inserted automatically after sending the command.
 * All data currently pending in output and input buffers is flushed.
 *
 * \param atCmd Points to a null-terminated string containing the AT command.
 * \note There is no need to terminate the command with a carriage return.
 * \return Returns true if the command was sent successfully.
 */
bool GsmModule::writeCommand(const char* atCmd)
{
	if (!modemPort) return false;

	// allow at least 20 ms delay first
	Thread::sleep(20);
	modemPort->flushInput();
	modemPort->flushOutput();

	// send command text to the modem
	size_t bytesWritten = modemPort->write(atCmd);
	bytesWritten += modemPort->write("\r");

	debug("[%s]\n", atCmd);
	return bytesWritten == strlen(atCmd) + 1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read AT command answer
 *
 * Retrieves the full modem answer data for a previously sent AT command.
 * The function blocks until the answer is received or the timeout expires.
 *
 * \param pAnswer Points to the string that receives the full modem answer.
 * This parameter can be set to NULL if the full answer is not required.
 * \param timeOut Specifies the answer timeout, in milliseconds.
 * \return Returns modem answer type identifier. Returns AT_INVALID if the
 * timeout expired. See ModemAnswerType for more details.
 */

ModemAnswerType GsmModule::readAnswer(std::string *pAnswer, int timeOut)
{
	if (!modemPort) return AT_INVALID;
	char rxBuf[128];

	memset(rxBuf, 0, sizeof(rxBuf));
	// read the buffer from communication port
	int bytesRead = modemPort->read(rxBuf, sizeof(rxBuf), timeOut);
	// check for timeout
	if (bytesRead == 0)
	{
		debug("No response within %d ms\n", timeOut);
		return AT_INVALID;
	}

	debug("[%s]\n", rxBuf);

	// find the first occurence of newline
	char *pBuf = strchr(rxBuf, '\r');
	if (!pBuf)
	{ // no command echo received - data is not valid
		debug("Received answer is not valid: [%s]\n", rxBuf);
		return AT_INVALID;
	}

	pBuf ++; // proceed to the next character

	if (pAnswer) pAnswer->clear(); // reset the output string

	// remove all whitespace at the beginning
	while (*pBuf == '\n' || *pBuf == '\r' || *pBuf == '\t' || *pBuf == ' ')
		pBuf ++;

	// write to output
	if (pAnswer) *pAnswer = pBuf;
	return modemPort->getAnswer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read AT command answer
 *
 * Retrieves the modem answer type identifier for a previously sent AT command.
 * The function blocks until the answer is received or the timeout expires.
 *
 * \param timeOut Specifies the answer timeout, in milliseconds.
 * \return Returns modem answer type identifier. Returns AT_INVALID if the
 * timeout expired. See ModemAnswerType for more details.
 */

ModemAnswerType GsmModule::readAnswer(int timeOut)
{
	return readAnswer(NULL, timeOut);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize communication
 *
 * Performs the initial handshake with the GSM modem over the connected
 * communication port. This method should be called after open() call and before
 * any other AT command in order to achieve the best performance.
 *
 * This method executes the modem auto-bauding sequence, fixes
 * the baud rate, enables flow control and selects AT command interface.
 *
 * \return Returns true if the handshake succeeded.
 */
bool GsmModule::initComm()
{
	lock();
	ModemAnswerType ret;

	// initialize data comm with autobauding
	for (int i = 0; i < 5; i ++)
	{
		writeCommand("AT");
		ret = readAnswer(200);
		if (ret == AT_OK) break;
		Thread::sleep(200);
	}

	// enable extended error result codes
	writeCommand("AT+CMEE=1");
	ret = readAnswer(500);
	if (ret != AT_OK)
	{
		unlock();
		return false;
	}

	// set the fixed port speed
	writeCommand("AT+IPR=115200");
	ret = readAnswer(500);
	if (ret != AT_OK)
	{
		unlock();
		return false;
	}

	// set hardware flow control
	writeCommand("AT&K3");
	ret = readAnswer(500);
	// set hardware flow control on terminal side
	modemPort->setFlowControl(S_FLOW_HARD);
	if(ret != AT_OK)
	{
		Thread::sleep(500);
		writeCommand("AT&K3");
		ret = readAnswer(500);
		if(ret != AT_OK)
		{
			unlock();
			return false;
		}
	}

	// select Telit interface #2
	writeCommand("AT#SELINT=2");
	ret = readAnswer(500);

	unlock();
	return ret == AT_OK;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize SIM card
 *
 * Initializes the GSM SIM. If required, sends the specified PIN code to
 * enable the SIM. If SIM PIN is disabled, the passed PIN code is ignored.
 * The method will wait internally if the SIM is busy.
 *
 * \param pinCode Points to a null-terminated string that contains the PIN code.
 * \return Returns true if the SIM was initialized successfully.
 */
bool GsmModule::initSIM(const char* pinCode)
{
	lock();

	// first query the SIM presense
	const char* pinQuery = "AT+CPIN?";
	std::string cmd = pinQuery, answer;

	// we may have to repeat the procedure several times
	for (int i = 0; i < 10; i ++)
	{
		writeCommand(cmd.c_str());
		readAnswer(&answer, 5000);
		if (answerStartsWith(answer, "+CPIN: SIM PIN"))
		{
			debug("SIM PIN requested\n");
			cmd = "AT+CPIN=";
			cmd += pinCode;
			cmd += '\r';
		}
		else if (answerStartsWith(answer, "+CPIN: SIM PUK"))
		{
			// SIM is locked
			debug("SIM is locked, PUK requested\n");
			unlock();
			return false;
		}
		else if (answerStartsWith(answer, "+CPIN: READY"))
		{
			// ready to proceed, return
			debug("SIM PIN ready, proceeding\n");
			unlock();
			return true;
		}
		else if (answerStartsWith(answer, "OK\r\n"))
		{
			// ready to proceed, return
			debug("SIM PIN OK, proceeding\n");
			unlock();
			return true;
		}
		else if (answerStartsWith(answer, "+CME ERROR: 10"))
		{
			// SIM not present
			debug("SIM is not inserted, trying again\n");
			Thread::sleep(1000);
			cmd = pinQuery;
		}
		else if (answerStartsWith(answer, "+CME ERROR: 13"))
		{
			// SIM defective
			debug("SIM defect detected, trying again\n");
			Thread::sleep(500);
			cmd = pinQuery;
		}
		else if (answerStartsWith(answer, "+CME ERROR: 14"))
		{
			// SIM is busy
			debug("SIM is busy, trying again\n");
			Thread::sleep(750);
			cmd = pinQuery;
		}
		else if (answerStartsWith(answer, "+CME ERROR: 15"))
		{
			// SIM is wrong type
			debug("SIM is wrong type, aborting\n");
			unlock();
			return false;
		}
		else if (answerStartsWith(answer, "+CME ERROR: 16"))
		{
			// SIM PIN is not valid
			debug("Incorrect SIM PIN, aborting\n");
			unlock();
			return false;
		}
		else
		{
			// invalid answer
			debug("Invalid answer, retrying\n");
			Thread::sleep(500);
			cmd = pinQuery;
		}
	}

	unlock();
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize GPRS
 *
 * Initializes GPRS profile parameters.
 * Creates a PDP context and sets basic QoS levels.
 *
 * \param gprsApn Points to a null-terminated string that contains the GPRS APN.
 * \return Returns true if the GPRS profile was initialized successfully.
 */
bool GsmModule::initGprs(const char *gprsApn)
{
	std::string answer;
	std::ostringstream s;

	s << "AT+CGDCONT=1,\"IP\",\"" << gprsApn << "\",\"0.0.0.0\",0,0";

	int nRetry = 3;
	while (nRetry --)
	{
		lock();
		writeCommand(s.str().c_str());
		ModemAnswerType answerCode = readAnswer(&answer, 1000);

		if (answerCode != AT_OK)
		{
			debug("could not configure GPRS. Answer code: %d\n", answerCode);
			unlock();

			Thread::sleep(500);
			continue;
		}

		writeCommand("AT+CGQMIN=1,0,0,0,0,0");
		answerCode = readAnswer(&answer, 1000);

		if (answerCode != AT_OK)
		{
			debug("failed setting minimum QoS, answer code: %d\n", answerCode);
			unlock();
			continue;
		}

		writeCommand("AT+CGQREQ=1,0,0,0,0,0");
		answerCode = readAnswer(&answer, 1000);

		if (answerCode != AT_OK)
		{
			debug("failed setting requested QoS, answer code: %d\n", answerCode);
			unlock();
			continue;
		}

		unlock();
		debug("GPRS has been configured successfully.\n");
		return true;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Attach GPRS
 *
 * Attaches the mobile data terminal to the GPRS service.
 *
 * \return Returns true if GPRS was attached successfully.
 */
bool GsmModule::attachGprs()
{
	std::string answer;

	lock();
	writeCommand("AT+CGATT=1");
	ModemAnswerType answerCode = readAnswer(&answer, 5000);
	unlock();

	return answerCode == AT_OK;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Register in GSM network
 *
 * Finds a suitable operator and registers the mobile data terminal with its
 * network. Automatic network search is used based on provided SIM.
 *
 * The method blocks until the GSM module is registered with the network or
 * the specified timeout expires.
 *
 * \param waitMax Specifies the maximum wait time.
 * \return Returns true if a suitable network was found successfully.
 */
bool GsmModule::searchNetwork(int waitMax)
{
	lock();
	int nRetry = 3;

	// enable automatic search mode
	while (nRetry --)
	{
		writeCommand("AT+COPS=0,2");
		if (readAnswer(5000) == AT_OK) break;
		Thread::sleep(500);
	}

	if (nRetry < 0)
	{
		unlock();
		debug("failed enabling automatic search mode\n");
		return false;
	}

	// enable numeric operator codes
	writeCommand("AT+COPS=3,2");
	if (readAnswer(5000) != AT_OK)
	{
		unlock();
		debug("failed enabling numeric operator codes\n");
		return false;
	}

	// return immediately if we don't have to wait for network
	if (waitMax == 0)
	{
		unlock();
		return true;
	}

	// just try +CREG request several times
	// +CREG=1 unsolicited messages do not work quite well, avoid them for now

	int errorcounter = 0; // avoid very long delays in case the modem does not answer

	for (int i = 0; i < waitMax; i ++)
	{
		std::string s;
		writeCommand("AT+CREG?");
		if (readAnswer(&s, 10000) == AT_INVALID)
		{
			errorcounter ++;
			if (errorcounter >= 3)
			{
				debug("There are some serious problems with modem communication\n");
				unlock();
				return false;
			}
		}
		else errorcounter = 0;

		if (answerStartsWith(s, "+CREG:"))
		{
			const char* p = strchr(s.c_str(), ',');
			int netStatus = atoi(++ p);

			if (netStatus == 1)
			{
				debug("Operating in home operator network\n");
				unlock();
				return true;
			}
			else if (netStatus == 5)
			{
				debug("Operating in roaming operator network\n");
				unlock();
				return true;
			}
		}

		Thread::sleep(1000);
	}

	unlock();
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set operational band
 *
 * Selects GSM operational frequency band. Automatic frequency band detection
 * may be enabled, network searching time will be increased in such case.
 *
 * \param freqBand Specifies the frequency band.
 * \return Returns true if the band was set successfully.
 */
bool GsmModule::setFrequency(ModuleFreqBand freqBand)
{
	lock();

	ModemAnswerType ret = AT_INVALID;
	switch (freqBand)
	{
	case FREQ_AUTODETECT:

		// set band autodetection
		writeCommand("AT#AUTOBND=1");
		ret = readAnswer(1000);
		break;

	case FREQ_900X1800:
	case FREQ_900X1900:
	case FREQ_850X1800:
	case FREQ_850X1900:
	{
		std::ostringstream s;

		// set operational band of the module:
		s << "AT#BND=" << (int) freqBand;
		writeCommand(s.str().c_str());
		ret = readAnswer(5000);
	}
	}

	unlock();
	return ret == AT_OK;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set power mode
 *
 * Selects the power management mode of the GSM module.
 * Note that communication with the GSM modem is not possible in POWER_MINIMAL
 * and modes. The modem in POWER_MINIMAL mode will wake up on external events
 * and proceed to POWER_FULL mode.
 * Communication with the modem is possible in POWER_FULL_SAVE mode only if the
 * DTR signal is high. If DTR is low, the GSM module enters a power saving mode.
 *
 * \param pwrMode Specifies the new power mode.
 * \return Returns true if the power mode was set successfully.
 */
bool GsmModule::powerSaveMode(ModulePowerMode pwrMode)
{
	lock();

	ModemAnswerType ret;
	std::ostringstream s;

	// set module functionality:
	// 0=minimal, 1=full, 2=no TX, 4=no TX/RX, 5=power saving
	s << "AT+CFUN=" << (int) pwrMode;
	writeCommand(s.str().c_str());
	ret = readAnswer(5000);

	unlock();
	return ret == AT_OK;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get network information
 *
 * Requests the network information from the GSM module.
 * The network information structure is filled with available data.
 *
 * \param netInfo Contains a reference to the network information structure.
 * \return Returns true if the information was acquired successfully.
 */
bool GsmModule::readNetworkInfo(GSM_NETWORK_INFO &netInfo)
{
	lock();
	std::string str;

	writeCommand("AT+CREG?");
	if (readAnswer(&str, 1000) != AT_OK)
	{
		unlock();
		return false;
	}

	int unsolReg;
	if (sscanf(str.c_str(), "+CREG:%d,%d", &unsolReg, &netInfo.netStatus) != 2)
	{
		unlock();
		return false;
	}

	writeCommand("AT#MONI");
	if (readAnswer(&str, 1000) != AT_OK)
	{
		unlock();
		return false;
	}

	const char *headerText = "#MONI: ";
	const int headerLen = strlen(headerText);
	if (!answerStartsWith(str, headerText))
	{
		unlock();
		return false;
	}

	// locate and read out the network name
	size_t pos = str.find("BSIC:");
	if (pos > 0)
		netInfo.networkName = str.substr(headerLen, pos - headerLen - 1);
	else netInfo.networkName = "Unknown";

	// read out signal quality
	pos = str.find("RxQual:");
	if (pos > 0)
		netInfo.signalQuality = atoi(str.substr(pos + 7, 1).c_str());
	else netInfo.signalQuality = 0;

	// read out receiver level
	pos = str.find("PWR:");
	if (pos > 0)
		netInfo.signalPower = atoi(str.substr(pos + 4, 4).c_str());
	else netInfo.signalPower = 0;

	// read out channel identifier
	pos = str.find("ARFCN:");
	if (pos > 0)
		netInfo.rfChannel = atoi(str.substr(pos + 6, 3).c_str());
	else netInfo.rfChannel = 0;

	unlock();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Match modem answer
 *
 * Matches the beginning of a modem answer string to another string.
 *
 * \param answer Contains the full modem answer.
 * \param str Points to a null-terminated string containing the part that needs
 * to be compared.
 * \brief Returns true if the beginning of the answer string is identical with
 * the specified text string.
 */
bool GsmModule::answerStartsWith(std::string &answer, const char *str)
{
	return strncmp(answer.c_str(), str, strlen(str)) == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Lock GSM module
 *
 * Locks the GSM module and blocks simultaneous access by other threads.
 * A typical application should lock the module before sending an AT command and
 * unlock after receiving an answer.
 *
 * \return Returns true if the modem was locked successfully.
 */
bool GsmModule::lock()
{
	return modemMutex.lock();
}

/**
* \brief Unlock GSM module
*
* Unlocks the GSM module and makes it available for further communication.
*
* \return Returns true if the modem was unlocked successfully.
*/
void GsmModule::unlock()
{
	modemMutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set module RTC
 *
 * Sets integrated RTC of the GSM module to the specified date/time.
 *
 * \param pDate Points to the standard date/time structure.
 * \return Returns true if RTC was set successfully.
 */
bool GsmModule::setRTC(tm *pDate)
{
	lock();

	char str[128];
	char zone[5];
	sprintf(str, "AT+CCLK=\"%02d/%02d/%02d,%02d:%02d:%02d", pDate->tm_year
			- 100, pDate->tm_mon + 1, pDate->tm_mday, pDate->tm_hour,
			pDate->tm_min, pDate->tm_sec);

	if (pDate->tm_gmtoff >= 0)
		sprintf(zone, "+%02d\"", (int) pDate->tm_gmtoff);
	else sprintf(zone, "-%02d\"", (int) pDate->tm_gmtoff);

	strcat(str, zone);
	writeCommand(str);

	if (readAnswer(1000) != AT_OK)
	{
		unlock();
		return false;
	}

	unlock();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get module RTC
 *
 * Reads integrated RTC of the GSM module and returns its date/time.
 *
 * \param pDate Points to the standard date/time structure that receives the data.
 * \return Returns true if RTC was read successfully.
 */
bool GsmModule::getRTC(tm *pDate)
{
	lock();

	std::string str;
	writeCommand("AT+CCLK?");
	if (readAnswer(&str, 1000) != AT_OK)
	{
		unlock();
		return false;
	}

	unlock();

	int year, mon, day, hour, min, sec;
	int nFields = sscanf(str.c_str(),
			"+CCLK: \"%02d/%02d/%02d,%02d:%02d:%02d\"", &year, &mon, &day,
			&hour, &min, &sec);

	if (nFields != 6) return false;

	pDate->tm_year = year + 100;
	pDate->tm_mon = mon - 1;
	pDate->tm_mday = day;
	pDate->tm_hour = hour;
	pDate->tm_min = min;
	pDate->tm_sec = sec;

	pDate->tm_wday = 0;
	pDate->tm_yday = 0;

	pDate->tm_gmtoff = 0;
	pDate->tm_isdst = -1;
	pDate->tm_zone = NULL;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read firmware version
 *
 * Retrieves the firmware version of the GSM module.
 *
 * \return Returns the string containing the firmware version.
 */
std::string GsmModule::getVersion()
{
	if (modemVersion.empty())
	{
		lock();

		// query the hardware/firmware revision
		std::string rsp;
		writeCommand("AT+CGMR");
		ModemAnswerType ret = readAnswer(&rsp, 500);
		debug("Telit FW: %s\n", rsp.c_str());

		if (ret == AT_OK)
		{
			// modem firmware version is the first string
			size_t retPos = rsp.find('\r', 0);
			if (retPos != std::string::npos)
				modemVersion = rsp.substr(0, retPos);
		}

		unlock();
	}

	return modemVersion;
}
