#ifndef GSMMODULE_H_
#define GSMMODULE_H_

#include <time.h>
#include <string>

#include "Thread.h"
#include "ModemPort.h"

/** Structure to hold information about current GSM network status. */
struct GSM_NETWORK_INFO
{
	/** Specifies network status: 1 = home, 5 = roaming */
	int netStatus;
	/** Contains network operator name or alias */
	std::string networkName;
	/** Specifies the quality of reception, 0..7 */
	int signalQuality;
	/** Specifies the received signal power, in dBm */
	int signalPower;
	/** Contains network channel currently in use (ARFCN) */
	int rfChannel;
};

/**
 * \brief GSM module class
 *
 * This is a helper class that handles communication to the Telit GSM communication
 * module of the mobile data terminal. GsmModule class is designed to simplify
 * managing the GSM module and implements basic AT commands.
 */
class GsmModule
{
public:

	/** Available operational frequency bands of the GSM module. */
	enum ModuleFreqBand
	{
		/** Automatic detection of frequency band */
		FREQ_AUTODETECT=-1,
		/** European frequency bands, 900 Mhz and 1800 Mhz */
		FREQ_900X1800=0,
		/** Mixed frequency bands, 900 Mhz and 1900 Mhz */
		FREQ_900X1900,
		/** Mixed frequency bands, 850 Mhz and 1900 Mhz */
		FREQ_850X1800,
		/** American frequency bands, 850 Mhz and 1900 Mhz */
		FREQ_850X1900 };

	/** Available power modes of the GSM module. */
	enum ModulePowerMode
	{
		/** Minimal functionality mode, communication port is disabled */
		POWER_MINIMAL=0,
		/** Full functionality mode */
		POWER_FULL=1,
		/** Radio transmitter disabled */
		POWER_NOTX=2,
		/** Radio transmitter and receiver disabled */
		POWER_NORXTX=4,
		/** Full functionality mode with power saving */
		POWER_FULL_SAVE=5 };

	GsmModule(const char* portName);
	virtual ~GsmModule();

	bool open();
	bool close();

	bool lock();
	void unlock();

	bool writeCommand(const char* atCmd);
	ModemAnswerType readAnswer(std::string *pAnswer, int timeOut);
	ModemAnswerType readAnswer(int timeOut = DEFAULT_AT_TIMEOUT);

	bool initComm();
	bool initSIM(const char* pinCode);
	bool searchNetwork(int waitMax);
	bool initGprs(const char* gprsApn);
	bool attachGprs();

	bool setFrequency(ModuleFreqBand freqBand);
	bool powerSaveMode(ModulePowerMode pwrMode);

	bool readNetworkInfo(GSM_NETWORK_INFO &netInfo);

	bool setRTC(tm *pDate);
	bool getRTC(tm *pDate);

	/**
	 * \brief Get the communication port name
	 * \return Returns the name of the communication port as string
	 */
	std::string getPortName() { return commPortName; }
	std::string getVersion();

	static bool answerStartsWith(std::string &answer, const char *str);

private:

	std::string		commPortName;
	std::string		modemVersion;
	ModemPort*		modemPort;
	Mutex			modemMutex;
};

/** \file */
#endif /*GSMMODULE_H_*/
