/***************************************************************************
 *   Copyright            : (c) 2001 Adam Clauss                           *
 *                        : (c) 2001 Shane Hill                            *
 *                        : (c) 2007 Andrei Birjukov                       *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

// C++ Includes
#include <sstream>

// C Includes
#include <stdio.h>
#include <stdarg.h>
#include <ctype.h>

// Local Includes
#include "StdioFile.h"
#include "IniFile.h"

#define iniEOL "\r\n"

/**
 * \brief Class constructor
 *
 * Initializes the INI file object.
 *
 * \param iniPath Contains the path to the INI file.
 */
IniFile::IniFile(std::string const iniPath)
{
	SetPath(iniPath);
	caseInsensitive = true;
}

/**
 * \brief Read INI file
 *
 * Reads the INI file using the path specified in the constructor.
 *
 * \return Returns true if successful, false otherwise.
 */
bool IniFile::ReadFile()
{
	std::string keyname, valuename, value;
	std::string::size_type pLeft, pRight;

	StdioFile inFile;
	if (!inFile.open(path.c_str(), StdioFile::ModeRead)) return false;

	while (!inFile.isEOF())
	{
		std::string line = inFile.readString();
		// trim the linebreak, if any
		if (line.length() && line[line.length() - 1] == '\n')
			line = line.substr(0, line.length() - 1);

		if (line.length())
		{
			// To be compatible with Win32, check for existence of '\r'.
			// Win32 files have the '\r' and Unix files don't at the end of a line.
			// Note that the '\r' will be written to INI files from
			// Unix so that the created INI file can be read under Win32
			// without change.
			if (line[line.length() - 1] == '\r') line = line.substr(0,
					line.length() - 1);

			// Check that the user hasn't opened a binary file by checking the first
			// character of each line!
			if (line.length() != 0 && !isprint(line[0]))
			{
				inFile.close();
				return false;
			}
			if ((pLeft = line.find_first_of(";#[=")) != std::string::npos)
			{
				switch (line[pLeft])
				{
				case '[':
					if ((pRight = line.find_last_of("]")) != std::string::npos
							&& pRight > pLeft)
					{
						keyname = line.substr(pLeft + 1, pRight - pLeft - 1);
						AddKeyName(keyname);
					}
					break;

				case '=':
					valuename = line.substr(0, pLeft);
					value = line.substr(pLeft + 1);
					SetValue(keyname, valuename, value);
					break;

				case ';':
				case '#':
					if (!names.size())
						AddHeaderComment(line.substr(pLeft + 1));
					else AddKeyComment(keyname, line.substr(pLeft + 1));
					break;
				}
			}
		}
	}

	inFile.close();
	if (names.size()) return true;
	return false;
}

/**
 * \brief Write INI file
 *
 * Writes the INI file using the path specified in the constructor.
 *
 * \return Returns true if the file was written successfully.
 */
bool IniFile::WriteFile()
{
	unsigned commentID, keyID, valueID;

	// Normally you would use ofstream, but the SGI CC compiler has
	// a few bugs with ofstream. So ... fstream used.

	StdioFile outFile;
	if (!outFile.open(path.c_str(), StdioFile::ModeWrite)) return false;

	// Write header comments.
	for (commentID = 0; commentID < comments.size(); ++ commentID)
	{
		std::ostringstream s;
		s << ';' << comments[commentID] << iniEOL;
		outFile.writeString(s.str().c_str());
	}

	if (comments.size()) outFile.writeString(iniEOL);

	// Write keys and values.
	for (keyID = 0; keyID < keys.size(); ++ keyID)
	{
		std::ostringstream s;
		s << '[' << names[keyID] << ']' << iniEOL;

		// Comments.
		for (commentID = 0; commentID < keys[keyID].comments.size(); ++ commentID)
			s << ';' << keys[keyID].comments[commentID] << iniEOL;

		// Values.
		for (valueID = 0; valueID < keys[keyID].names.size(); ++ valueID)
			s << keys[keyID].names[valueID] << '='
			<< keys[keyID].values[valueID] << iniEOL;

		s << iniEOL;
		outFile.writeString(s.str().c_str());
	}

	outFile.close();
	return true;
}

/**
 * \brief Find INI key index
 *
 * Finds an index for the given key name.
 *
 * \param keyname Contains the name of the key.
 * \return Returns a valid key index, or noID on failure.
 */
long IniFile::FindKey(std::string const keyname) const
{
	for (unsigned keyID = 0; keyID < names.size(); ++ keyID)
		if (CheckCase(names[keyID]) == CheckCase(keyname)) return long(keyID);
	return noID;
}

/**
 * \brief Find INI value index
 *
 * Finds an index for the given value under the given key name.
 *
 * \param keyID Contains the key index.
 * \param valuename Contains the name of the value.
 * \return Returns a valid value index, or noID on failure.
 */
long IniFile::FindValue(unsigned const keyID, std::string const valuename) const
{
	if (!keys.size() || keyID >= keys.size()) return noID;

	for (unsigned valueID = 0; valueID < keys[keyID].names.size(); ++ valueID)
		if (CheckCase(keys[keyID].names[valueID]) == CheckCase(valuename)) return long(valueID);
	return noID;
}

/**
 * \brief Add INI key
 *
 * Inserts the new key (or section) to the INI file.
 *
 * \param keyname Contains the name of the new key.
 * \return Returns the key index.
 */
unsigned IniFile::AddKeyName(std::string const keyname)
{
	names.resize(names.size() + 1, keyname);
	keys.resize(keys.size() + 1);
	return (unsigned) (names.size() - 1);
}

/**
 * \brief Get INI key name
 *
 * \param keyID Specifies the key index.
 * \return Returns the name of the key.
 */
std::string IniFile::GetKeyName(unsigned const keyID) const
{
	if (keyID < names.size())
		return names[keyID];
	else return "";
}

/**
 * \brief Get number of INI values
 *
 * \param keyID Specifies the key index.
 * \return Returns the number of values associated with the given key.
 */
unsigned IniFile::GetNumValues(unsigned const keyID)
{
	if (keyID < keys.size()) return (unsigned) keys[keyID].names.size();
	return 0;
}

/**
 * \brief Get number of INI values
 *
 * \param keyname Specifies the name of the key.
 * \return Returns the number of values associated with the given key.
 */
unsigned IniFile::GetNumValues(std::string const keyname)
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return 0;
	return (unsigned) keys[keyID].names.size();
}

/**
 * \brief Get INI value name
 *
 * Retrieves the name of the value associated with the given key.
 *
 * \param keyID Specifies the key index.
 * \param valueID Specifies the value index.
 * \return Returns the name of the value.
 */
std::string IniFile::GetValueName(unsigned const keyID, unsigned const valueID) const
{
	if (keyID < keys.size() && valueID < keys[keyID].names.size()) return keys[keyID].names[valueID];
	return "";
}

/**
 * \brief Get INI value name
 *
 * Retrieves the name of the value associated with the given key.
 *
 * \param keyname Contains the name of the key.
 * \param valueID Specifies the value index.
 * \return Returns the name of the value.
 */
std::string IniFile::GetValueName(std::string const keyname,
		unsigned const valueID) const
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return "";
	return GetValueName(keyID, valueID);
}

/**
 * \brief Set INI value - by index
 *
 * Sets the given value of the specified key to text data.
 *
 * \param keyID Specifies the key index that contains the value.
 * \param valueID Specifies the value index.
 * \param value Contains the value text.
 * \return Returns true if the value was set successfully. On error returns false.
 */
bool IniFile::SetValue(unsigned const keyID, unsigned const valueID,
		std::string const value)
{
	if (keyID < keys.size() && valueID < keys[keyID].names.size())
		keys[keyID].values[valueID] = value;

	return false;
}

/**
 * \brief Set INI value - generic
 *
 * Sets the given value of the specified key to text data.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param value Contains the value text.
 * \param create Indicates if the value should be added if it does not exist.
 * \return Returns true if the value was set successfully. On error returns false.
 */
bool IniFile::SetValue(std::string const keyname, std::string const valuename,
		std::string const value, bool const create)
{
	long keyID = FindKey(keyname);
	if (keyID == noID)
	{
		if (create)
			keyID = long(AddKeyName(keyname));
		else return false;
	}

	long valueID = FindValue(unsigned(keyID), valuename);
	if (valueID == noID)
	{
		if (!create) return false;
		keys[keyID].names.resize(keys[keyID].names.size() + 1, valuename);
		keys[keyID].values.resize(keys[keyID].values.size() + 1, value);
	}
	else keys[keyID].values[valueID] = value;

	return true;
}

/**
 * \brief Set INI value - integer
 *
 * Sets the given value of the specified key to integer data.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param value Contains the integer value.
 * \param create Indicates if the value should be added if it does not exist.
 * \return Returns true if the value was set successfully. On error returns false.
 */
bool IniFile::SetValueI(std::string const keyname, std::string const valuename,
		int const value, bool const create)
{
	char svalue[MAX_VALUEDATA];

	sprintf(svalue, "%d", value);
	return SetValue(keyname, valuename, svalue, create);
}

/**
 * \brief Set INI value - floating point
 *
 * Sets the given value of the specified key to floating point data.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param value Contains the floating point value.
 * \param create Indicates if the value should be added if it does not exist.
 * \return Returns true if the value was set successfully. On error returns false.
 */
bool IniFile::SetValueF(std::string const keyname, std::string const valuename,
		double const value, bool const create)
{
	char svalue[MAX_VALUEDATA];

	sprintf(svalue, "%f", value);
	return SetValue(keyname, valuename, svalue, create);
}

/**
 * \brief Set INI value - printf style
 *
 * Sets the given value of the specified key to text data with formatting
 * field support.
 *
 * \param keyname Specifies the name of the key that contains the value
 * \param valuename Specifies the name of the value.
 * \param format Contains the value text with printf-style formatting fields.
 * \param ... Contains a variable number of arguments for formatting fields.
 * \return Returns true if the value was set successfully. On error returns false.
 */
bool IniFile::SetValueV(std::string const keyname, std::string const valuename,
		char *format, ...)
{
	va_list args;
	char value[MAX_VALUEDATA];

	va_start( args, format);
	vsprintf(value, format, args);
	va_end( args);
	return SetValue(keyname, valuename, value);
}

/**
 * \brief Get INI value - by index
 *
 * Retrieves the data of the given INI value associated with the specified key.
 *
 * \param keyID Specifies the key index that contains the value.
 * \param valueID Specifies the value index.
 * \param defValue Contains the default value data to be returned if the
 * value does not exist.
 * \return Returns the text data of the requested value.
 */
std::string IniFile::GetValue(unsigned const keyID, unsigned const valueID,
		std::string const defValue) const
{
	if (keyID < keys.size() && valueID < keys[keyID].names.size()) return keys[keyID].values[valueID];
	return defValue;
}

/**
 * \brief Get INI value - generic
 *
 * Retrieves the data of the given INI value associated with the specified key.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param defValue Contains the default value data to be returned if the
 * value does not exist.
 * \return Returns the text data of the requested value.
 */
std::string IniFile::GetValue(std::string const keyname,
		std::string const valuename, std::string const defValue) const
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return defValue;

	long valueID = FindValue(unsigned(keyID), valuename);
	if (valueID == noID) return defValue;

	return keys[keyID].values[valueID];
}

/**
 * \brief Get INI value - integer
 *
 * Retrieves the integer data of the given INI value associated with the
 * specified key.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param defValue Contains the default value data to be returned if the
 * value does not exist.
 * \return Returns the integer data of the requested value.
 */
int IniFile::GetValueI(std::string const keyname, std::string const valuename,
		int const defValue) const
{
	char svalue[MAX_VALUEDATA];

	sprintf(svalue, "%d", defValue);
	return atoi(GetValue(keyname, valuename, svalue).c_str());
}

/**
 * \brief Get INI value - floating point
 *
 * Retrieves the floating point data of the given INI value associated with
 * the specified key.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param defValue Contains the default value data to be returned if the
 * value does not exist.
 * \return Returns the floating point data of the requested value.
 */
double IniFile::GetValueF(std::string const keyname,
		std::string const valuename, double const defValue) const
{
	char svalue[MAX_VALUEDATA];

	sprintf(svalue, "%f", defValue);
	return atof(GetValue(keyname, valuename, svalue).c_str());
}

/**
 * \brief Get INI value - scanf-style
 *
 * Parses the data of the given INI value associated with the specified key.
 * Supports scanf-style formatting fields and a variable number of destinations.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value.
 * \param format Contains the data parsing format with printf-style formatting fields.
 * \param ... Contains a variable number of destination pointers for parsed fields.
 * \return Returns the floating point data of the requested value.
 */
unsigned IniFile::GetValueV(std::string const keyname,
		std::string const valuename, char *format, ...)
{
	std::string value;
	va_list  args;
	unsigned nVals;

	value = GetValue(keyname, valuename);
	if (!value.length()) return false;

	va_start( args, format);
	nVals = vsscanf( value.c_str(), format, args);
	va_end( args);

	return nVals;
}

/**
 * \brief Delete INI value
 *
 * Removes the given value assigned to a specified key.
 *
 * \param keyname Specifies the name of the key that contains the value.
 * \param valuename Specifies the name of the value to be removed.
 * \return Returns true if the value was deleted successfully.
 */
bool IniFile::DeleteValue(std::string const keyname,
		std::string const valuename)
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return false;

	long valueID = FindValue(unsigned(keyID), valuename);
	if (valueID == noID) return false;

	// This looks strange, but is neccessary.
	std::vector<std::string>::iterator npos = keys[keyID].names.begin() + valueID;
	std::vector<std::string>::iterator vpos = keys[keyID].values.begin() + valueID;
	keys[keyID].names.erase(npos, npos + 1);
	keys[keyID].values.erase(vpos, vpos + 1);

	return true;
}

/**
 * \brief Delete INI key
 *
 * Removes the specified key with all contained values.
 *
 * \param keyname Specifies the name of the key to be removed.
 * \return Returns true if the key was deleted successfully.
 */
bool IniFile::DeleteKey(std::string const keyname)
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return false;

	// Now hopefully this destroys the vector lists within keys.
	// Looking at <vector> source, this should be the case using the destructor.
	// If not, I may have to do it explicitly. Memory leak check should tell.
	// memleak_test.cpp shows that the following not required.
	//keys[keyID].names.clear();
	//keys[keyID].values.clear();

	std::vector<std::string>::iterator npos = names.begin() + keyID;
	std::vector<key>::iterator kpos = keys.begin() + keyID;
	names.erase(npos, npos + 1);
	keys.erase(kpos, kpos + 1);

	return true;
}

/**
 * \brief Reset the INI file
 *
 * Deletes all stored INI file data.
 */
void IniFile::Reset()
{
	// This loop not needed. The vector<> destructor seems to do
	// all the work itself. memleak_test.cpp shows this.
	//for ( unsigned i = 0; i < keys.size(); ++i) {
	//  keys[i].names.clear();
	//  keys[i].values.clear();
	//}
	names.clear();
	keys.clear();
	comments.clear();
}

/**
 * \brief Add INI header comment
 *
 * Adds a comment to the INI file header.
 * Header comments are those comments before the first key.
 *
 * \param comment Contains the comment text.
 */
void IniFile::AddHeaderComment(std::string const comment)
{
	comments.resize(comments.size() + 1, comment);
}

/**
 * \brief Get INI header comment
 *
 * Return a comment from the INI file header.
 *
 * \param commentID Specifies the comment line index.
 * \return Returns the requested comment line.
 */
std::string IniFile::GetHeaderComment(unsigned const commentID) const
{
	if (commentID < comments.size()) return comments[commentID];
	return "";
}

/**
 * \brief Delete INI header comment
 *
 * Delete a comment from the INI file header.
 *
 * \param commentID Specifies the comment line index.
 * \return Returns true if the line was deleted successfully.
 */
bool IniFile::DeleteHeaderComment(unsigned commentID)
{
	if (commentID < comments.size())
	{
		std::vector<std::string>::iterator cpos = comments.begin() + commentID;
		comments.erase(cpos, cpos + 1);
		return true;
	}
	return false;
}

/**
 * \brief Get number of key comments
 *
 * \param keyID Specifies the key index.
 * \return Returns the total number of key comments.
 */
unsigned IniFile::GetNumKeyComments(unsigned const keyID) const
{
	if (keyID < keys.size()) return (unsigned) keys[keyID].comments.size();
	return 0;
}

/**
 * \brief Get number of key comments
 *
 * \param keyname Specifies the name of the key.
 * \return Returns the total number of key comments.
 */
unsigned IniFile::GetNumKeyComments(std::string const keyname) const
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return 0;
	return (unsigned) keys[keyID].comments.size();
}

/**
 * \brief Add INI key comment
 *
 * Adds a comment to the specified key of the INI file.
 *
 * \param keyID Specifies the key index.
 * \param comment Contains the text of the comment.
 * \return Returns true if the comment was added succesfully.
 */
bool IniFile::AddKeyComment(unsigned const keyID, std::string const comment)
{
	if (keyID < keys.size())
	{
		keys[keyID].comments.resize(keys[keyID].comments.size() + 1, comment);
		return true;
	}
	return false;
}

/**
 * \brief Add INI key comment
 *
 * Adds a comment to the specified key of the INI file.
 *
 * \param keyname Specifies the name of the key.
 * \param comment Contains the text of the comment.
 * \return Returns true if the comment was added succesfully.
 */
bool IniFile::AddKeyComment(std::string const keyname, std::string const comment)
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return false;
	return AddKeyComment(unsigned(keyID), comment);
}

/**
 * \brief Get INI key comment
 *
 * Retrieves the comment related to the specified key of the INI file.
 *
 * \param keyID Specifies the key index.
 * \param commentID Specifies the comment line index.
 * \return Returns the text of the comment line.
 */
std::string IniFile::GetKeyComment(unsigned const keyID, unsigned const commentID) const
{
	if (keyID < keys.size() && commentID < keys[keyID].comments.size()) return keys[keyID].comments[commentID];
	return "";
}

/**
 * \brief Get INI key comment
 *
 * Retrieves the comment related to the specified key of the INI file.
 *
 * \param keyname Specifies the name of the key.
 * \param commentID Specifies the comment line index.
 * \return Returns the text of the comment line.
 */
std::string IniFile::GetKeyComment(std::string const keyname,
		unsigned const commentID) const
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return "";
	return GetKeyComment(unsigned(keyID), commentID);
}

/**
 * \brief Delete INI key comment
 *
 * Removes the comment related to the specific key of the INI file.
 *
 * \param keyID Specifies the key index.
 * \param commentID Specifies the comment line index.
 * \return Returns true if the comment was deleted successfully.
 */
bool IniFile::DeleteKeyComment(unsigned const keyID, unsigned const commentID)
{
	if (keyID < keys.size() && commentID < keys[keyID].comments.size())
	{
		std::vector<std::string>::iterator cpos = keys[keyID].comments.begin()
			+ commentID;
		keys[keyID].comments.erase(cpos, cpos + 1);
		return true;
	}
	return false;
}

/**
 * \brief Delete INI key comment
 *
 * Removes the comment related to the specific key of the INI file.
 *
 * \param keyname Specifies the name of the key.
 * \param commentID Specifies the comment line index.
 * \return Returns true if the comment was deleted successfully.
 */
bool IniFile::DeleteKeyComment(std::string const keyname,
		unsigned const commentID)
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return false;
	return DeleteKeyComment(unsigned(keyID), commentID);
}

/**
 * \brief Delete all INI key comments
 *
 * Removes all comments related to the specific key of the INI file.
 *
 * \param keyID Specifies the key index.
 * \return Returns true if all comments were deleted successfully.
 */
bool IniFile::DeleteKeyComments(unsigned const keyID)
{
	if (keyID < keys.size())
	{
		keys[keyID].comments.clear();
		return true;
	}
	return false;
}

/**
 * \brief Delete all INI key comments
 *
 * Removes all comments related to the specific key of the INI file.
 *
 * \param keyname Specifies the name of the key.
 * \return Returns true if all comments were deleted successfully.
 */
bool IniFile::DeleteKeyComments(std::string const keyname)
{
	long keyID = FindKey(keyname);
	if (keyID == noID) return false;
	return DeleteKeyComments(unsigned(keyID));
}

std::string IniFile::CheckCase(std::string s) const
{
	if (caseInsensitive) for (std::string::size_type i = 0; i < s.length(); ++ i)
		s[i] = tolower(s[i]);
	return s;
}
