#ifndef INIFILE_H_
#define INIFILE_H_

// IniFile.cpp:  Implementation of the CIniFile class.
// Written by:   Adam Clauss
// Email: cabadam@tamu.edu
// You may use this class/code as you wish in your programs.  Feel free to distribute it, and
// email suggested changes to me.
//
// Rewritten by: Shane Hill
// Date:         21/08/2001
// Email:        Shane.Hill@dsto.defence.gov.au
// Reason:       Remove dependancy on MFC. Code should compile on any
//               platform. Tested on Windows/Linux/Irix
//////////////////////////////////////////////////////////////////////

// C++ Includes
#include <string>
#include <vector>

// C Includes
#include <stdlib.h>

#define MAX_KEYNAME		64
#define MAX_VALUENAME	64
#define MAX_VALUEDATA	512

/**
 * \brief INI file helper class
 *
 * This class is a helper for handling INI files. It provides a number
 * of useful functions for reading and writing INI keys and their values.
 */
class IniFile
{
private:
	bool caseInsensitive;
	std::string path;
	struct key
	{
		std::vector<std::string> names;
		std::vector<std::string> values;
		std::vector<std::string> comments;
	};
	std::vector<key> keys;
	std::vector<std::string> names;
	std::vector<std::string> comments;
	std::string CheckCase(std::string s) const;

public:

	/** List of error codes */
	enum errors
	{
		/** INI key index not found */
		noID = -1
	};

	IniFile(std::string const iniPath = "");
	virtual ~IniFile()
	{
	}

	/**
	 * \brief Enable INI case sensitivity
	 *
	 * Enables the case sensitivity. All operations on INI file keys and values
	 * become case sensitive.
	 */
	void CaseSensitive()
	{
		caseInsensitive = false;
	}

	/**
	 * \brief Disable INI case sensitivity
	 *
	 * Disables the case sensitivity. All operations on INI file keys and values
	 * become case insensitive. This is the default behavior.
	 */
	void CaseInsensitive()
	{
		caseInsensitive = true;
	}

	/**
	 * \brief Set INI file path
	 *
	 * Sets the path of the INI file for reading and writing.
	 *
	 * \param newPath Contains the INI file name.
	 */
	void SetPath(std::string const newPath)
	{
		path = newPath;
	}

	/**
	 * \brief Get INI file path
	 *
	 * \return Returns the path of the INI file.
	 */
	std::string GetPath() const
	{
		return path;
	}

	bool ReadFile();
	bool WriteFile();
	void Reset();

	long FindKey(std::string const keyname) const;
	long FindValue(unsigned const keyID, std::string const valuename) const;

	/**
	 * \brief Get number of keys
	 *
	 * \return Returns the total number of keys currently in the INI file.
	 */
	unsigned GetNumKeys() const
	{
		return (unsigned) names.size();
	}

	unsigned AddKeyName(std::string const keyname);

	std::string GetKeyName(unsigned const keyID) const;

	unsigned GetNumValues(unsigned const keyID);
	unsigned GetNumValues(std::string const keyname);

	std::string GetValueName(unsigned const keyID, unsigned const valueID) const;
	std::string GetValueName(std::string const keyname, unsigned const valueID) const;

	std::string GetValue(unsigned const keyID, unsigned const valueID,
			std::string const defValue = "") const;
	std::string GetValue(std::string const keyname,
			std::string const valuename, std::string const defValue = "") const;
	int GetValueI(std::string const keyname, std::string const valuename,
			int const defValue = 0) const;

	/**
	 * \brief Get INI value - boolean
	 *
	 * Retrieves the boolean data of the given INI value associated with the
	 * specified key.
	 *
	 * \param keyname Specifies the name of the key that contains the value.
	 * \param valuename Specifies the name of the value.
	 * \param defValue Contains the default value data to be returned if the
	 * value does not exist.
	 * \return Returns the boolean data of the requested value.
	 */
	bool GetValueB(std::string const keyname, std::string const valuename,
			bool const defValue = false) const
	{
		return GetValueI(keyname, valuename, int(defValue)) ? true : false;
	}

	double GetValueF(std::string const keyname, std::string const valuename,
			double const defValue = 0.0) const;
	unsigned GetValueV(std::string const keyname, std::string const valuename,
			char *format, ...);

	bool SetValue(unsigned const keyID, unsigned const valueID,
			std::string const value);
	bool SetValue(std::string const keyname, std::string const valuename,
			std::string const value, bool const create = true);
	bool SetValueI(std::string const keyname, std::string const valuename,
			int const value, bool const create = true);

	/**
	 * \brief Set INI value - boolean
	 *
	 * Sets the given value of the specified key to boolean data.
	 *
	 * \param keyname Specifies the name of the key that contains the value.
	 * \param valuename Specifies the name of the value.
	 * \param value Contains the boolean value.
	 * \param create Indicates if the value should be added if it does not exist.
	 * \return Returns true if the value was set successfully. On error returns false.
	 */
	bool SetValueB(std::string const keyname, std::string const valuename,
			bool const value, bool const create = true)
	{
		return SetValueI(keyname, valuename, int(value), create);
	}

	bool SetValueF(std::string const keyname, std::string const valuename,
			double const value, bool const create = true);
	bool SetValueV(std::string const keyname, std::string const valuename,
			char *format, ...);

	bool DeleteValue(std::string const keyname, std::string const valuename);

	bool DeleteKey(std::string keyname);

	// Header comment functions.
	/**
	 * \brief Number of header comments
	 *
	 * \return Returns the total number of header comments.
	 */
	unsigned GetNumHeaderComments()
	{
		return (unsigned) comments.size();
	}

	void AddHeaderComment(std::string const comment);
	std::string GetHeaderComment(unsigned const commentID) const;
	bool DeleteHeaderComment(unsigned commentID);

	/**
	 * \brief Delete all header comments
	 *
	 * Clears all comments from the INI file header.
	 */
	void DeleteHeaderComments()
	{
		comments.clear();
	}

	// Key comment functions.
	// Key comments are those comments within a key. Any comments
	// defined within value names will be added to this list. Therefore,
	// these comments will be moved to the top of the key definition when
	// the CIniFile::WriteFile() is called.

	unsigned GetNumKeyComments(unsigned const keyID) const;
	unsigned GetNumKeyComments(std::string const keyname) const;
	bool AddKeyComment(unsigned const keyID, std::string const comment);
	bool AddKeyComment(std::string const keyname, std::string const comment);
	std::string GetKeyComment(unsigned const keyID, unsigned const commentID) const;
	std::string GetKeyComment(std::string const keyname, unsigned const commentID) const;
	// Delete a key comment.
	bool DeleteKeyComment(unsigned const keyID, unsigned const commentID);
	bool DeleteKeyComment(std::string const keyname, unsigned const commentID);
	// Delete all comments for a key.
	bool DeleteKeyComments(unsigned const keyID);
	bool DeleteKeyComments(std::string const keyname);
};

#endif /*INIFILE_H_*/
