/*
 * KeypadInputWindow.cpp
 *
 *  Created on: Oct 22, 2008
 *      Author: andrei
 */

#include <stdio.h>

//#define DEBUG
#include "Debug.h"
#include "PowerManager.h"

#include "KeypadInputWindow.h"

#ifdef USE_XYNTH

// keyWaitTimer, used for waiting of another press of the same key before advancing to next position
#define TIMER_ID_WAITKEYS		1000
// keyHoldTimer, used for detecting long key presses
#define TIMER_ID_HOLDKEYS		1001
// multitap and long-press delays in milliseconds
#define DELAY_MULTITAP			600
#define DELAY_LONGPRESS			1000

const unsigned char KeypadInputWindow::keyMapText[CHAR_INPUT_KEYS][2][12] =
{
	{	// keypad 0
		{ ' ', 0x0D, '(', ')', '<', '>', '[', ']', '{', '}', '0', 0x00 },
		{ ' ', 0x0D, '(', ')', '<', '>', '[', ']', '{', '}', '0', 0x00 }
	},
	{	// keypad 1
		{ '.', ',', ':', ';', '@', '#', '&', '!', '?', '"', '|', 0x00 },
		{ '.', ',', ':', ';', '@', '#', '&', '!', '?', '"', '|', 0x00 }
	},
	{	// keypad 2
		{ 'A', 'B', 'C', 0x8e, 0x92, 0x8f, '2', 0x00 },
		{ 'a', 'b', 'c', 0x84, 0x91, 0x86, '2', 0x00 }
	},
	{	// keypad 3
		{ 'D', 'E', 'F', '3', 0x00 },
		{ 'd', 'e', 'f', '3', 0x00 }
	},
	{	// keypad 4
		{ 'G', 'H', 'I', '4', 0x00 },
		{ 'g', 'h', 'i', '4', 0x00 }
	},
	{	// keypad 5
		{ 'J', 'K', 'L', '5', 0x00 },
		{ 'j', 'k', 'l', '5', 0x00 }
	},
	{	// keypad 6
		{ 'M', 'N', 'O', 0x99, 0xd8, 0xd5, '6', 0x00 },
		{ 'm', 'n', 'o', 0x94, 0xf8, 0xf5, '6', 0x00 }
	},
	{	// keypad 7
		{ 'P', 'Q', 'R', 'S', '7', 0x00 },
		{ 'p', 'q', 'r', 's', '7', 0x00 }
	},
	{	// keypad 8
		{ 'T', 'U', 'V', 0xdc, '8', 0x00 },
		{ 't', 'u', 'v', 0xfc, '8', 0x00 }
	},
	{	// keypad 9
		{ 'W', 'X', 'Y', 'Z', '9', 0x00 },
		{ 'w', 'x', 'y', 'z', '9', 0x00 }
	},
	{	// keypad .
		{ '-', '+', '*', '/', '%', '=', 0x80, '$', 0x9c, 0xb1, 0xab, 0x00 },
		{ '-', '+', '*', '/', '%', '=', 0x80, '$', 0x9c, 0xb1, 0xab, 0x00 }
	},
};

const unsigned char KeypadInputWindow::keyMapNum[CHAR_INPUT_KEYS] =
{ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.' };

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default constructor
 * Initializes the class and sets the keypad into uppercase character mode
 *
 * \param idResource Unique window identifier
 */
KeypadInputWindow::KeypadInputWindow(int idResource) :
	GuiWindow(idResource)
{
	shiftDown = false;
	shiftToggle = false;
	capsLock = true;
	numberMode = true;
	staticTextOnly = false;

	repkeyCounter = 0;
	repkeyKey = -1;
	holdKey = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 * Uninitializes the window
 */
KeypadInputWindow::~KeypadInputWindow()
{
}

////////////////////////////////////////////////////////////////////////////////
//
int KeypadInputWindow::inputKey(int keyCode)
{
	switch (keyCode)
	{
	case S_KEYCODE_ZERO:
		if (!shiftDown) return 0;
	case S_KEYCODE_ONE:
	case S_KEYCODE_TWO:
	case S_KEYCODE_THREE:
	case S_KEYCODE_FOUR:
	case S_KEYCODE_FIVE:
	case S_KEYCODE_SIX:
	case S_KEYCODE_SEVEN:
	case S_KEYCODE_EIGHT:
	case S_KEYCODE_NINE:
		if (!shiftDown) return keyCode - S_KEYCODE_ONE + 1;
	case S_KEYCODE_PERIOD:
		if (!shiftDown) return 10;
	}
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Key press event
 * Method is called when user presses a key (but does not release it)
 *
 * \param keyCode Identifier of the button the user is pressing
 * \note The keyCode is not the character code
 *
 * The following methods are invoked:
 *
 * - If the menu button is pressed, the onMenu() method is called
 * - If shift key is pressed, the shift flag is set
 * - If keypad input is disabled, no further keypresses are checked
 * - Execute the onKeyPress() method
 *
 * \return Returns true if the keypress was handled.
 */
bool KeypadInputWindow::onKeyDown(int keyCode)
{
	bool startKeyHoldTimer = false;
	debug("KeypadInputWindow::onKeyDown(): keyCode: %d\n", keyCode);

	// set shift flags here
	shiftToggle = (keyCode == S_KEYCODE_F11);
	if(shiftToggle) shiftDown = true;

	// check if menu button was pressed
	if (keyCode == S_KEYCODE_F12) onMenu();

	// check if we can process the keypress further
	if (!inputEnabled()) return true;

	stopKeyWaitTimer();
	if (repkeyKey > 0 && repkeyKey != keyCode) {
		// if any other key was pressed, then commit previous one
		onCharCommit();
		repkeyKey = -1;
		repkeyCounter = 0;
	}

	stopKeyHoldTimer();
	if (numberMode && (holdKey == S_KEYCODE_PERIOD)) {
		// commit the un-committed '.'
		onCharCommit();
	}
	holdKey = -1;

	if (staticTextOnly)
	{
		onStaticScreenKeyPress(keyCode);
		return true;
	}

	// some keys may be requested to send to server immediately and not to be processed
	if (!shiftDown)
	{
		if (specialFunction(keyCode)) return true;
	}

	// input keys
	int kbdInputKey = inputKey(keyCode);
	if (kbdInputKey >= 0)
	{
		if (numberMode)
		{
			if (keyCode == S_KEYCODE_PERIOD) {
				// do not commit '.' as it might turn to '-'
				if (onCharInput('.', INPUT_NO_CHANGE, false))
					startKeyHoldTimer = true;
			}
			else
				// entering a numeric character from keypad, submit immediately
				onCharInput(keyMapNum[kbdInputKey], INPUT_NO_CHANGE, true);
		}
		else
		{
			// set/advance the character index in the list
			if (repkeyKey == keyCode)
			{
				repkeyCounter ++;
				if (keyMapText[kbdInputKey][0][repkeyCounter] == 0)
					repkeyCounter = 0;
			}
			else repkeyCounter = 0;

			repkeyKey = keyCode;

			if (capsLock)
			{
				// preview the upper case alpha character
				if (onCharInput(keyMapText[kbdInputKey][0][repkeyCounter],
													INPUT_NO_CHANGE, false))
				{
					startKeyHoldTimer = true;
				}
				else
				{
					repkeyCounter = 0;
					repkeyKey = -1;
				}
			}
			else
			{
				// preview the lower case alpha character
				if (onCharInput(keyMapText[kbdInputKey][1][repkeyCounter],
													INPUT_NO_CHANGE, false))
				{
					startKeyHoldTimer = true;
				}
				else
				{
					repkeyCounter = 0;
					repkeyKey = -1;
				}
			}
		}
	} else
	{
		// other keys
		switch (keyCode)
		{
		case S_KEYCODE_F1:
		case S_KEYCODE_F2:
		case S_KEYCODE_F3:
		case S_KEYCODE_F4:
		case S_KEYCODE_F5:

			// F1..F5 if no shift is held down
			if (numberMode) onFunctionAccess(keyCode - S_KEYCODE_F1 + 1);
			// F6..F10 if shift is held down
			else onFunctionAccess(keyCode - S_KEYCODE_F1 + 6);
			break;

		case S_KEYCODE_DELETE:

			if (!shiftDown)
			{
				// delete one character
				if (onDelete(false))
					startKeyHoldTimer = true;
			}
			break;

		case S_KEYCODE_F11:

			startKeyHoldTimer = true;
			break;

		default:
			break;
		}
	}

	// start key hold timer to detect long presses
	if (startKeyHoldTimer) {
		debug("Starting key hold timer\n");
		holdKey = keyCode;
		startTimer(DELAY_LONGPRESS, TIMER_ID_HOLDKEYS);
	}

	// call virtual press method
	onKeyPress(keyCode, shiftDown);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable or disable the number mode
 * Enables or disables the numeric mode of the keypad. Triggers the onCharInput()
 * method with INPUT_NUMBER_MODE_CHANGE event.
 *
 * \param newMode Set to ture to enable numeric mode, false to disable it.
 */
void KeypadInputWindow::setNumberMode(bool newMode)
{
	numberMode = newMode;
	onCharInput(0, INPUT_NUMBER_MODE_CHANGE, false);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Key release event
 * Method is executed when user releases the pressed key
 *
 * \param keyCode Identifier of the button the user is pressing
 * \note The keyCode is not the character code
 *
 * The following methods are invoked:
 * - If shift is released, the shift flag will be reset
 * - If keypad is disabled no further processing occurs
 * - If shift is released after a delay put the keypad in number mode
 * - Execute the onKeyRelease() method
 *
 * \return Returns true if key release event was successfully handled.
 */
bool KeypadInputWindow::onKeyUp(int keyCode)
{
	debug("KeypadInputWindow::onKeyUp(): keyCode: %d\n", keyCode);

	// reset shift flag here if shift was released
	if (keyCode == S_KEYCODE_F11) shiftDown = false;

	// stop the timers
	stopKeyWaitTimer();
	if (repkeyKey > 0 && repkeyKey != keyCode) {
		// if any other key was released, then commit previous one
		onCharCommit();
		repkeyKey = -1;
		repkeyCounter = 0;
	}

	stopKeyHoldTimer();
	if (numberMode && (holdKey == S_KEYCODE_PERIOD)) {
		// commit the un-committed '.'
		onCharCommit();
	}
	holdKey = -1;

	// check if we can process the keypress further
	if (!inputEnabled() || staticTextOnly) return true;

	// if there is something to be waited for, start timer
	if (repkeyKey == keyCode)
	{
		debug("Starting key wait timer\n");
			startTimer(DELAY_MULTITAP, TIMER_ID_WAITKEYS);
	}
	else
	{
		// this should not happen
		if (repkeyKey>0) {
			onCharCommit();
			repkeyKey = -1;
			repkeyCounter = 0;
		}
	}

	switch (keyCode)
	{
	case S_KEYCODE_F11:

		if (shiftToggle)	// shift key is released, no other key pressed
		{
			// toggle between ABC and 123 mode
			setNumberMode(!numberMode);
		}
		break;
	}

	// call virtual release method
	onKeyRelease(keyCode, shiftDown);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Execute when the waiting timers are fired
 * Method is executed if any of the timers associated with keypad input are fired.
 *
 * There are two timers associated with the keypad:
 * - TIMER_ID_WAITKEYS - Waits for further keypresses and if the period elapses the character
 * selected will be committed using the onCharCommit() method
 * - TIMER_ID_HOLDKEYS - Waits to react to user event of holding down the key for a specific time.
 * For example holding down the shift key triggers case mode change, holding down delete triggers
 * onDelete() method to execute.
 *
 * \param idTimer Identification to the timer that was fired
 */
void KeypadInputWindow::onTimer(int idTimer)
{
	switch (idTimer)
	{
	case TIMER_ID_WAITKEYS:
		// enough of waiting... commit the character
		stopKeyWaitTimer();
		if (repkeyKey > 0) {
			onCharCommit();
			repkeyKey = -1;
			repkeyCounter = 0;
		}
		break;

	case TIMER_ID_HOLDKEYS:
		{
		bool stopTimer = false;

		// character input keys
		int kbdInputKey = inputKey(holdKey);
		if (!numberMode && kbdInputKey >= 0)
		{
			// We're keeping the key repeat position in repkeyCounter
			if (repkeyKey == holdKey) // double check the repkeyKey to avoid using wrong key buffer
			{
				capsLock = !capsLock;
				if (capsLock)
				{
					// preview the upper case alpha character
					if (!onCharInput(keyMapText[kbdInputKey][0][repkeyCounter],
												INPUT_CAPSLOCK_MODE_CHANGE, false))
						stopTimer = true;
				}
				else
				{
					// preview the lower case alpha character
					if (!onCharInput(keyMapText[kbdInputKey][1][repkeyCounter],
												INPUT_CAPSLOCK_MODE_CHANGE, false))
						stopTimer = true;
				}
			}
		}
		else
		{
			switch (holdKey) {
			case S_KEYCODE_F11:
				// hold shift while in character mode: ABC<->abc
				if (!numberMode) {
					capsLock = !capsLock;
					onCharInput(0, INPUT_CAPSLOCK_MODE_CHANGE, false);
					shiftToggle = false;
				}
				break;
			case S_KEYCODE_PERIOD:
				if (numberMode) {
					// special case: holding '.' in 123 mode brings us '-'
					onCharInput('-', INPUT_NO_CHANGE, true);
					stopTimer = true;
				}
				break;
			case S_KEYCODE_DELETE:
				if (!shiftDown) {
					// delete the rest
					onDelete(true);
					stopTimer = true;
				}
				break;
			default:
				// nothing to do with this key. outta here.
				stopTimer = true;
				break;
			}
		}

		if (stopTimer) {
			holdKey = -1;
			stopKeyHoldTimer();
		}

		break;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Process the window close event
 * Processes the window close event, stops all the active timers
 */
void KeypadInputWindow::onClose()
{
	stopKeyWaitTimer();
	stopKeyHoldTimer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Process the modal window close event
 * Processes the event of closing the modal child window, resets the shift flag
 * \param pModal Pointer to the modal window that is being closed
 *
 * \return Returns true
 */
bool KeypadInputWindow::onModalFinished(GuiWindow *pModal)
{
	debug("Running modal window cleanup\n");
	// assume that shift is not pressed
	shiftDown = false;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// NOTE: lock stopKeyWaitTimer before calling this function!

void KeypadInputWindow::stopKeyWaitTimer()
{
	debug("Stopping key repeat timer\n");
	stopTimer(TIMER_ID_WAITKEYS);
}

////////////////////////////////////////////////////////////////////////////////
//
// NOTE: lock stopKeyHoldTimer before calling this function!

void KeypadInputWindow::stopKeyHoldTimer()
{
	debug("Stopping caps timer\n");
	stopTimer(TIMER_ID_HOLDKEYS);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Return the current input mode
 * Returns the input mode the keypad is currently put into
 *
 * \return Returns input mode
 */
KeypadInputWindow::InputMode KeypadInputWindow::getInputMode()
{
	if (numberMode) return INPUT_123;
	else
	{
		if (capsLock) return INPUT_ABC;
		else return INPUT_abc;
	}
}

#endif
