/*
 * KeypadInputWindow.h
 *
 *  Created on: Oct 22, 2008
 *      Author: andrei
 */

#ifndef KEYPADINPUTWINDOW_H_
#define KEYPADINPUTWINDOW_H_

#include "LibSpatha.h"

#ifdef USE_XYNTH

#include <exynthoo/GuiWindow.h>

#define CHAR_INPUT_KEYS			12

/**
 * \brief Implements window with keypad access functionality
 * Creates a window that captures input from the keypad. Provides methods
 * for entering data using a mobile-style user experience.
 *
 * Keypad can be put into numeric, uppercase or lowercase mode. In character mode
 * different characters can be selected by pressing the appropriate keypad button
 * repeatedly. When using numeric mode, the numbers can be directly entered
 * without multiple keypresses.
 *
 * This class provides methods for handling the input events that will be
 * generated by use cases described above.
 */
class KeypadInputWindow: public GuiWindow
{
public:

	/** Keypad input modes */
	enum InputMode
	{
		/** Numeric mode */
		INPUT_123,
		/** Uppercase character mode */
		INPUT_ABC,
		/** Lowercase character mode */
		INPUT_abc,
		/** Input disabled */
		INPUT_DISABLED
	};

	/** Input mode changes */
	enum InputModeChange
	{
		/** No change occurred */
		INPUT_NO_CHANGE,
		/** Changed number mode to character mode or vice versa */
		INPUT_NUMBER_MODE_CHANGE,
		/** Changed character mode to uppercase or lowercase */
		INPUT_CAPSLOCK_MODE_CHANGE
	};

	KeypadInputWindow(int idResource);
	virtual ~KeypadInputWindow();

	InputMode getInputMode();

protected:

	/**
	 * \brief Key press event
	 * Implementation should override this function to implement method that will be
	 * executed when user presses (but does not release) the button on keypad.
	 *
	 * \param keyCode Identifier of the button being pressed
	 * \param shiftDown Set to true, if the shift button is also pressed
	 */
	virtual void onKeyPress(int keyCode, bool shiftDown) = 0;

	/**
	 * \brief Receive basic input only
	 * Implementation should override this function to implement method that will be
	 * executed when advanced keypad handling features are disabled and only basic keypresses
	 * are accepted.
	 *
	 * \param keyCode Identifier of the button that was pressed
	 */
	virtual void onStaticScreenKeyPress(int keyCode) = 0;

	/**
	 * \brief Key release event
	 * Implementation should override this function to implement method that will be
	 * executed when user has released the button on keypad.
	 *
	 * \param keyCode Identifier of the button being released
	 * \param shiftDown Set to true, if the shift button was also pressed
	 */
	virtual void onKeyRelease(int keyCode, bool shiftDown) = 0;

	/**
	 * \brief Function key press event
	 * Implementation should override this function to implement method that will be
	 * executed when user has pressed the one of the function keys (F1...F10) on keypad
	 *
	 * \param fn Identifier of the function key pressed
	 */
	virtual void onFunctionAccess(int fn) = 0;

	/**
	 * \brief Character input event
	 * Implementation should override this function to implement method that will be
	 * executed every time the user has pressed (and released) the key. The method captures
	 * every character the user has selected. With multiple keypresses on one button this method
	 * is called every time to preview the character currently chosen.
	 *
	 * \param in Character currently selected. This is the C char value, not the button identifier!
	 * \param inputModeChange Set to one of the input mode change values to specify if the input mode
	 * was changed during the keypress process (i.e. changing uppercase to lowercase)
	 * \param commitNow If set to true, this is the final character that will not be changed
	 *
	 * \return Returns true if the character input event was handled successfully.
	 *
	 */
	virtual bool onCharInput(unsigned char in,
					InputModeChange inputModeChange, bool commitNow) = 0;

	/**
	 * \brief Character commit event
	 * Implementation should override this function to implement method that will be
	 * executed every time the character commit timeout has passed. In that case the last
	 * temporary character (that was sent to onCharInput()) should be taken as the final
	 * character that will not change anymore.
	 */
	virtual void onCharCommit() = 0;

	/**
	 * \brief Special function keypress event
	 * Implementation should override this function to implement method that will be
	 * executed every time the user has pressed the key. This method can
	 * be used to trap special keypressed that should be handled differently. If the
	 * return code is false no further processing of the keypress will be done and the
	 * keypress is considered handled here.
	 *
	 * \param keyCode Identifier of the button being pressed
	 * \return Returns true if the keypress was a special function keypress and no further
	 * processing should be done. If set to false all other methods for handlng the keypress
	 * will be executed as well.
	 */
	virtual bool specialFunction(int keyCode) = 0;

	/**
	 * \brief Delete key press event
	 * Implementation should override this function to implement method that will be
	 * executed every time the user has pressed the DEL key. The method will detect if the
	 * delete key was pressed shortly or was held down for longer time.
	 *
	 * \param delAll Set to true if the DEL key was pressed for longer time (indicating that the user
	 * would like to delete all data (i.e. clear the field completely).
	 *
	 * \return Returns true if the key press event was handled successfully.
	 */
	virtual bool onDelete(bool delAll) = 0;

	/**
	 * \brief Menu key press event
	 * Implementation should override this function to implement method that will be
	 * executed every time the user has pressed the MENU key.
	 */
	virtual void onMenu() = 0;

	/**
	 * \brief General keypress event
	 * Implementation should override this function to implement method that will be
	 * executed every time the user has pressed the key on a keypad. This method is used
	 * to capture the keypresses and optionally discard them.
	 *
	 * \return Returns true if the keypress should be handled by all other methods. If returns
	 * false then the keypress event will be discarded.
	 */
	virtual bool inputEnabled() = 0;

	/**
	 * \brief Initialization event handler
	 * Implementation should override this function to implement method that will be
	 * executed when the window is being initialized.
	 */
	virtual void onInitialize() = 0;
	virtual bool onKeyDown(int keyCode);
	virtual bool onKeyUp(int keyCode);
	virtual void onTimer(int idTimer);
	virtual void onClose();
	virtual bool onModalFinished(GuiWindow *pModal);

	void setNumberMode(bool newMode);

	/** If set to true, no advanced keypad events are handled. With every keypress only the onStaticScreenKeyPress() is called */
	bool staticTextOnly;

private:

	void updateInputMode();
	int inputKey(int keyCode);

	int repkeyCounter;
	int repkeyKey;	// the key we are waiting again, -1 if none

	void stopKeyWaitTimer();

	int holdKey;	// the key we are holding down, -1 if none

	void stopKeyHoldTimer();

	bool shiftDown;
	bool shiftToggle;
	bool capsLock;
	bool numberMode;

	static const unsigned char keyMapNum[CHAR_INPUT_KEYS];
	static const unsigned char keyMapText[CHAR_INPUT_KEYS][2][12];
};

#endif
#endif /* KEYPADINPUTWINDOW_H_ */
