/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : anti.sullin at artecdesign dot ee              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <sstream>
#include <sys/time.h>
#include <stdio.h>
#include <string.h>

#include <platform.h>

#include "PowerManager.h"
#include "SysInfo.h"

#include "LMXBluetoothModule.h"
#include "BluetoothSPPConnection.h"

//#define DEBUG
#include "Debug.h"

#define MIN(X,Y) (((X) < (Y)) ? (X) : (Y))

#define BT_TX_TIMEOUT		100
#define READ_POLL_TIME		1000


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the LMX Bluetooth module object.
 *
 * \param portName Points to a null-terminated string containing the name of
 * the serial port the LMX module is connected to.
 */
LMXBluetoothModule::LMXBluetoothModule(const char* portName) :
	BluetoothModule(), Thread()
{
	btPort = NULL;

	for (int i = 0; i< MAX_BT_CONNECTIONS; i++) {
		connections[i] = NULL;
		sdpId[i] = NULL;
	}

	commPortName = portName;
	openPortMask = 0;
	activeSPPMask = 0;
	moduleEnabled = true;
	debug("constructor, %s\n", commPortName.c_str());
}

/**
 * \brief Class destructor
 *
 * Terminates the LMX Bluetooth module object.
 */
LMXBluetoothModule::~LMXBluetoothModule()
{
	debug("destructor started\n");
	if (btPort) closePort();
	debug("destructor closed\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Connect to Bluetooth module
 *
 * Opens the LMX Bluetooth module serial port and initializes port parameters.
 *
 * \return Returns true if the Bluetooth module was opened successfully.
 */
bool LMXBluetoothModule::openPort()
{
	debug("open...\n");
	if (btPort) return false;

	// create and open a new modem port object
	btPort = new LMXBluetoothPort();
	if (!btPort->open(commPortName.c_str()))
	{
		debug("failed opening %s\n", commPortName.c_str());
		delete btPort;
		btPort = NULL;
		return false;
	}

	// initialize port settings
	btPort->setSpeed(115200);
	btPort->setFlowControl(S_FLOW_HARD);
	btPort->setCharBits(8);
	btPort->setStopBits(1);
	btPort->setParity(S_PARITY_NONE);

	debug("done\n");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close Bluetooth module
 *
 * Closes the LMX Bluetooth module serial port.
 *
 * \return Returns true if the module port was closed successfully.
 */
bool LMXBluetoothModule::closePort()
{
	debug("close...\n");

	if (!btPort)
	{
		return false;
	}
	btPort->close();

	delete btPort;
	btPort = NULL;

	debug("done\n");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize communication with module
 *
 * Resets the LMX Bluetooth module and performs an initial handshake procedure.
 *
 * \return Returns true if the link was initialized successfully.
 */
bool LMXBluetoothModule::initComm()
{
	debug("initcomm\n");
	if (btPort) return false;
	if (!openPort()) return false;

	// reset the chip and wait for LMX_READY packet
	for (int i = 0; i < 5; i ++)
	{
		debug("initcomm try %d\n",i);
		btPort->sendCmd(LMX_CMD_RESET, NULL, 0, NULL, 0);

		uint8_t buf[MAX_BT_PKT_DATA_LEN + 1];
		int ret;
		unsigned char type, opcode;

		// Get packet
		ret = btPort->getPkt(&type, &opcode, buf, sizeof(buf), 250 );
		if (ret < 2) {
			debug("no packet, ret %d", ret);
			continue;
		}

		if (type != IND || opcode != LMX_READY) {
			debug("invalid type/opcode");
			continue;
		}
		std::string fwVer((char *)(buf+1), buf[0]);
		localFWVer=fwVer;

		debug("got fwver %s\n", fwVer.c_str());

		return true;
	}

	debug("initcomm failed\n");
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initializes LMX module settings
 *
 * Configures the LMX module settings
 *
 * \return Returns true if the module was configured successfully.
 */
bool LMXBluetoothModule::initModule()
{
	char buf[32];
	const char *serial = SysInfo::getSerialNumber();

	debug("initmodule\n");
	if (!initComm()) return false;

	buf[0] = 0;
	// no default pin - prompt each time
	if (!btPort->sendCmd(LMX_GAP_SET_FIXED_PIN, buf, 1, NULL, 0)) return false;
	// disable event filter
	if (!btPort->sendCmd(LMX_SET_EVENT_FILTER, buf, 1, NULL, 0)) return false;
	// remove all stored SDP records
	if (!btPort->sendCmd(LMX_DELETE_SDP_RECORDS, NULL, 0, NULL, 0)) return false;
	// non-automatic operating mode
	if (!btPort->sendCmd(LMX_WRITE_OPERATION_MODE, buf, 1, NULL, 0)) return false;

	// device class
	buf[2] = 0x08;
	buf[1] = 0x01;
	buf[0] = 0x14;
	if (!btPort->sendCmd(LMX_STORE_CLASS_OF_DEVICE, buf, 3, NULL, 0)) return false;

	// local name
	strncpy(buf+1, BLUETOOTH_NAME_PFX, sizeof(buf)-1);
	buf[sizeof(buf)-1] = 0;
	buf[0] = strlen(buf+1) + strlen(serial) +1;
	debug("BT name: [%s%s]\n", buf+1, serial);

	if (!btPort->sendCmd(LMX_GAP_WRITE_LOCAL_NAME, buf, strlen(buf), serial, strlen(serial)+1)) return false;

	if (!Thread::start()) return false;
	BluetoothModule::pModule = this;

	updateOpenPorts();

	debug("initmodule done\n");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize Bluetooth subsystem (static)
 *
 * Creates the LMX Bluetooth Module object, performs an initial handshake
 * procedure with the LMX Bluetooth module, starts BT thread
 *
 * \return Returns true if the bluetooth subsystem was initialized successfully.
 */
bool LMXBluetoothModule::initialize() //static
{
	debug("initialize %s\n", LMX_BLUETOOTH_DEVICE);
	if (BluetoothModule::pModule != NULL) return true;

	LMXBluetoothModule *pLMX = new LMXBluetoothModule(LMX_BLUETOOTH_DEVICE);

	if (pLMX->initModule()) return true;

	debug("initialize failed\n");

	delete pLMX;
	debug("returning\n");
	return false;
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Shut down Bluetooth subsystem (static)
 *
 * Disconnects all Bluetooth connections, stops BT thread.
 * You should delete all connection objects before shutting down LMXBluetoothModule!
 *
 * \return Returns true if the bluetooth subsystem was shut down successfully.
 */
bool LMXBluetoothModule::shutdown()
{
	debug("shutdown\n");
	for (int i = 0; i< MAX_BT_CONNECTIONS; i++) {
		if (connections[i] != NULL) return false;
	}

	Thread::stop();
	closePort();

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief send a LMX command and wait for reply
 *
 * \param cmd Command opcode
 * \param buf1 First part of message
 * \param len1 Length of the message first part
 * \param buf2 Second part of message
 * \param len2 Length of the message second part
 * \param timeout Timeout to wait for reply in ms
 * \param keepCFMMutex Set to true if lastCFMMutex should be left locked on successful exit
 *
 * Send a LMX commend to LMX module. Two data buffers can be passed with the
 * command, the buffer contents are concatenated.
 * The command will wait until confirm packet is received or timeout occurs.
 * The confirm packet contents are stored to lastCFM variable, protected by lastCFMMutex.
 * If the contents of the confirm packet are needed by the caller, keepCFMMutex can be
 * set to true to leave lastCFMMutex locked on successful exit. Then it is the responsibility
 * of the caller to unlock the mutex after the data has been examined. If the lastCFMMutex
 * is not locked, the next CFM reply may overwrite the buffer!
 *
 * \return Returns true if the command was sent successfully.
 */
bool LMXBluetoothModule::sendCmd(unsigned char cmd,
		const void *buf1, const size_t len1,
		const void *buf2, const size_t len2,
		int timeout, bool keepCFMMutex)
{
	struct timeval start, now;
	debug("sendCmd 0x%x\n", cmd);

	cmdMutex.lock();

	lastCFMMutex.lock();
	lastCFM.opcode = 0;
	lastCFMMutex.unlock();

	if (!btPort->sendCmd(cmd, buf1, len1, buf2, len2)) {
		cmdMutex.unlock();
		return false;
	}

	if (timeout==0) {
		cmdMutex.unlock();
		return true;
	}

	gettimeofday(&start, 0);
	while (1) {
		lastCFMMutex.lock();
		if (lastCFM.opcode == cmd) break;
		lastCFMMutex.unlock();

		lastCFMSem.pend(timeout);
		gettimeofday(&now, 0);
		if ((((now.tv_sec - start.tv_sec) * 1000) +
				((now.tv_usec - start.tv_usec) / 1000)) > timeout) {
			debug("Timeout sending command\n");
			cmdMutex.unlock();
			return false;
		}
	}

	if (!keepCFMMutex) lastCFMMutex.unlock();
	cmdMutex.unlock();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send SPP data over the bluetooth connection
 *
 * \param connid Connection ID, returned on SPP connection registration
 * \param data Data to be sent over SPP link
 * \param len Data length
 *
 * Data buffer given by *data parameter will be sent out over Bluetooth.
 * \note The return value shows if the data was sent out, not received by other end!
 *
 * \return Returns true if data was sent out, false on error
 */
bool LMXBluetoothModule::sendSPPData(int connid, const void *data, size_t len)
{
	bool ret = true;
	unsigned char buf[3];

	while (len) {
		size_t txlen = len;
		if (txlen > MAX_TXDATA) txlen = MAX_TXDATA;

		debug("sendSPPData conn %d, %d bytes\n", connid, len);
		buf[0] = connid + 1;
		buf[1] = txlen & 0xFF;
		buf[2] = (txlen >> 8) & 0xFF;

		ret = sendCmd(LMX_SPP_SEND_DATA, buf, sizeof(buf), data, txlen,
				BT_TX_TIMEOUT, true);
		if (!ret) return false;

		if (lastCFM.datalen != 2 || lastCFM.data[0] != 0)
			ret = false;

		lastCFMMutex.unlock();

		if (!ret) return false;

		data = (void*)((char*)data+txlen);
		len -= txlen;
	}

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send PIN code to destination MAC address for authorization
 *
 * \param mac MAC address to send pin code to
 *
 * The connection object corresponding to the MAC address will be located.
 * The PIN code will be read from the object and transmitted to destination
 * MAC for authorization.
 */
void LMXBluetoothModule::sendPin(unsigned char *mac)
{
	unsigned char addr[6];
	unsigned char buf[MAX_BT_PIN+7];
	debug("requested PIN for %02x:%02x:%02x:%02x:%02x:%02x\n", mac[5], mac[4], mac[3], mac[2], mac[1], mac[0]);
	connMutex.lock();

	// search MAC address in outgoing mac database
	for (int i = 0; i< MAX_BT_CONNECTIONS; i++) {
		if (!connections[i]) continue;
		if (!connections[i]->getDestAddr(addr)) continue;
		if (!memcmp(addr, mac, 6)) {
			memcpy(buf, mac, 6);
			connections[i]->getPin((char*)buf+7, sizeof(buf)-7);
			buf[6] = strlen((char*)buf+7);
			sendCmd(LMX_GAP_GET_PIN, buf, 7 + buf[6], NULL , 0, 0, false);

			connMutex.unlock();
			return;
		}
	}

	// incoming connection
	if (incomingPin[0]) {
		memcpy(buf, mac, 6);
		strncpy((char*)buf+7, incomingPin, sizeof(buf)-7);
		buf[sizeof(buf)-1] = 0;
		buf[6] = strlen((char*)buf+7);
		sendCmd(LMX_GAP_GET_PIN, buf, 7 + buf[6], NULL , 0, 0, false);
	}
	connMutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief LMX receiver thread
 * Starts the LMX receiver thread, sends out initialization packets and acts according
 * to link state flags
 */
void LMXBluetoothModule::run()
{
	unsigned char type = 0, opcode = 0;
	int cnt;
	unsigned char buf[MAX_BT_PKT_DATA_LEN+1];

	debug("thread started");

	while(true)
	{
		if (Thread::stopRequested()) break;

		cnt = btPort->getPkt(&type, &opcode, buf, sizeof(buf), READ_POLL_TIME);

		if (cnt<0) continue;

		/* Confirm packet - confirm a command */
		if (type == CFM) {
			lastCFMMutex.lock();
			lastCFM.opcode = opcode;
			lastCFM.datalen = MIN(cnt, MAX_LASTCFM);
			memcpy(lastCFM.data, buf, lastCFM.datalen);
			lastCFMMutex.unlock();
			lastCFMSem.post();
			continue;
		}
		/* Indicator packet */
		if (type == IND) {
			debug("IND: %02x\n", opcode);
			switch(opcode) {
			case LMX_SPP_INCOMING_DATA:
				connMutex.lock();
				if (cnt > 3 && buf[0] > 0 && buf[0] < MAX_BT_CONNECTIONS &&
						connections[buf[0]-1]) {
					sendIncomingData(connections[buf[0]-1], buf+3, cnt-3);
				}
				connMutex.unlock();
				break;
			case LMX_SPP_LINK_ESTABLISHED:
				connMutex.lock();
				if (cnt == 9 && buf[7] > 0 && buf[7] < MAX_BT_CONNECTIONS &&
						connections[buf[7]-1]) {
					if (buf[0] == 0)
						activeSPPMask |= 1<<(buf[7]-1);
					else
						activeSPPMask &= ~(1<<(buf[7]-1));
					setConnectionState(connections[buf[7]-1],
							(buf[0] == 0)? SPP_CONNECTED : SPP_CONN_FAILED);
				}
				connMutex.unlock();
				break;
			case LMX_SPP_INCOMING_LINK_ESTABLISHED:
				connMutex.lock();
				if (cnt == 7 && buf[6] > 0 && buf[6] < MAX_BT_CONNECTIONS &&
						connections[buf[6]-1]) {
					activeSPPMask |= 1<<(buf[6]-1);
					setConnectionState(connections[buf[6]-1], SPP_CONNECTED);
				}
				connMutex.unlock();
				break;
			case LMX_SPP_LINK_RELEASED:
				connMutex.lock();
				if (cnt == 2 && buf[1] > 0 && buf[1] < MAX_BT_CONNECTIONS &&
						connections[buf[1]-1]) {
					setConnectionState(connections[buf[1]-1], SPP_DISCONNECTED);
					activeSPPMask &= ~(1<<(buf[1]-1));
				}
				connMutex.unlock();
				break;
			case LMX_GAP_GET_PIN:
				if (cnt == 6) sendPin(buf);
				break;
			case LMX_GAP_ACL_ESTABLISHED:
				debug("connection from %02x:%02x:%02x:%02x:%02x:%02x\n", buf[5], buf[4], buf[3], buf[2], buf[1], buf[0]);
				break;
			default:
				break;
			}
			continue;
		}
	}
}

bool LMXBluetoothModule::updateOpenPorts()
{
	debug("setting opened port mask to %x\n", openPortMask);
	return sendCmd(LMX_SET_PORTS_TO_OPEN, &openPortMask, 4, NULL, 0, BT_TX_TIMEOUT, false);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get MAC address
 *
 * Retrieves the hardware MAC address of the LMX Bluetooth module.
 *
 * \return Returns the text string containing a formatted MAC address.
 */

std::string LMXBluetoothModule::getLocalMac()
{
	if (localMac.length() > 0)
		return localMac;

	bool ret;
	ret = sendCmd(LMX_GAP_READ_LOCAL_BDA, NULL, 0, NULL, 0, BT_TX_TIMEOUT, true);
	if (!ret) return "";

	if (lastCFM.datalen != 7 ||
			lastCFM.data[0] != 0 ) {
		lastCFMMutex.unlock();
		return "";
	}

	char buf[8];
	for (int i = 0; i < 6; ++i) {
		snprintf(buf, sizeof(buf), "%2.2X", lastCFM.data[6 - i]);
		buf[sizeof(buf)-1] = 0;
		localMac.append(buf);
		if (i<5) localMac.append(":");
	}

	lastCFMMutex.unlock();

	return localMac;
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initiate a SPP connection
 *
 * Initiates a SPP connection process for a connection id.
 *
 * \param idConn Connection ID (=local port), returned from connection object registration.
 * \param dstAddr Destination MAC address
 * \param dstPort Destination port
 *
 * \return Returns true if connection process was initiated, false on error.
 */
bool LMXBluetoothModule::connectSPP(int idConn, unsigned char *dstAddr, unsigned char dstPort)
{
	unsigned char buf[8];

	buf[0] = idConn+1;
	memcpy(buf+1, dstAddr, 6);
	buf[7] = dstPort;
	connMutex.unlock();

	openPortMask |= 1<<idConn;
	updateOpenPorts();

	if (sendCmd(LMX_SPP_ESTABLISH_LINK, buf, sizeof(buf), NULL, 0, CONNECT_TIMEOUT, false))
		return true;

	openPortMask &= ~(1<<idConn);
	updateOpenPorts();

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Puts the device into listening mode
 * Puts the device into listening mode, activates the proper SDP record
 *
 * \param idConn Connection ID that should be used for listening incoming packets
 * \return Returns true if the device is in listening mode, false otherwise
 */
bool LMXBluetoothModule::listenSPP(int idConn)
{
	activateSDPRecord(idConn, true);
	openPortMask |= 1<<idConn;
	updateOpenPorts();

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Request a SPP connection disconnect
 *
 * Initiates a SPP disconnect process
 *
 * \param idConn Connection ID
 *
 * \return Returns true if disconnect was initiated, false on error
 */
bool LMXBluetoothModule::disconnectSPP(int idConn)
{
	unsigned char buf[1];
	buf[0] = idConn+1;

	openPortMask &= ~(1<<idConn);
	updateOpenPorts();

	activateSDPRecord(idConn, false);

	return sendCmd(LMX_SPP_RELEASE_LINK, buf, sizeof(buf), NULL, 0, DISCONNECT_TIMEOUT, false);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stores a new service record in local SDP database
 * Stores a new service record in Service Discovery Database
 *
 * \param localPort Local port used by the service
 * \param data Pointer to the data to be used
 * \param len Data length
 *
 * \return Returns true if the record was successfully created
 */
bool LMXBluetoothModule::addSDPRecord(int localPort, void* data, size_t len)
{
	if (localPort < 0 || localPort >= MAX_BT_CONNECTIONS) return false;

	bool ret;
	unsigned char buf[5];

	buf[0] = localPort;
	buf[1] = 0x22;
	buf[2] = 0x44;
	buf[3] = len & 0xFF;
	buf[4] = (len >> 8) & 0xFF;

	ret = sendCmd(LMX_STORE_SDP_RECORD, buf, sizeof(buf), data, len, BT_TX_TIMEOUT, true);
	if (!ret) return false;

	if (lastCFM.datalen != 2 ||
			lastCFM.data[0] != 0 ) ret = false;
	sdpId[localPort] = lastCFM.data[1];
	debug("Added sdp record for port %d, id %x\n", localPort, sdpId[localPort]);

	lastCFMMutex.unlock();
	return ret;
}

/**
 * \brief Activates the proper service record
 * Activates the service record from Service Discovery Database
 *
 * \param localPort Local port used by the service
 * \param enable Set to true to enable the SDP record, false to disable it
 */
bool LMXBluetoothModule::activateSDPRecord(int localPort, bool enable) {
	if (localPort < 0 || localPort >= MAX_BT_CONNECTIONS) return false;

	bool ret;
	unsigned char buf[2];

	buf[0] = (enable)?1:0;
	buf[1] = sdpId[localPort];

	debug("activating (en %d) sdp record %d\n", enable, buf[1]);
	ret = sendCmd(LMX_ENABLE_SDP_RECORD, buf, sizeof(buf), NULL, 0, BT_TX_TIMEOUT, true);
	if (!ret) return false;

	if (lastCFM.datalen != 2 ||
			lastCFM.data[0] != 0 ) ret = false;

	lastCFMMutex.unlock();
	return ret;

}

//////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable or disable the module
 * Enables or disables the LMX Bluetooth module
 *
 * \param enabled Set to true to enable the module, false to disable it
 */
void LMXBluetoothModule::setEnabled(bool enabled)
{
	moduleEnabled = enabled;
	sendChipEnable(enabled);
}

bool LMXBluetoothModule::sendChipEnable(bool e)
{
	unsigned char buf[2];
	bool ret;
	buf[0] = (e)?1:0; // connectability
	buf[1] = (e)?1:0; // discoverability

	debug("chip enable %d\n", e);
	ret = sendCmd(LMX_GAP_SET_SCANMODE, buf, sizeof(buf), NULL, 0, BT_TX_TIMEOUT, true);
	if (!ret) return false;

	if (lastCFM.datalen != 1 ||
			lastCFM.data[0] != 0 ) ret = false;

	lastCFMMutex.unlock();
	return ret;
}

//////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Change module power state
 * Changes the module power state, if power management is enabled
 *
 * \param oldState Current power state of the module
 * \param newState The power state the module should be put into
 *
 * \return Returns true if the power state change was successful, false otherwise
 */
bool LMXBluetoothModule::powerStateChange(PowerState oldState, PowerState newState)
{
	if (newState == LOW_POWER_AWAKE || newState == LOW_POWER_SLEEP) {
		// block suspending until all bluetooth connections are closed
		if (activeSPPMask) return false;

		if (suspended) return true;
		sendChipEnable(false);
		suspended = true;
	} else {
		if (!suspended) return true;
		suspended = false;
		sendChipEnable(moduleEnabled);
	}

	return true;
}
