#ifndef LMXBLUETOOTHMODULE_H_
#define LMXBLUETOOTHMODULE_H_

#include <time.h>
#include <string>

#include "BluetoothModule.h"
#include "Thread.h"


/* --------- lmx98xx communication protocol -----------*/

// Opcodes
#define LMX_GAP_INQUIRY         	0x00
#define LMX_GAP_DEVICE_FOUND    	0x01
#define LMX_GAP_READ_LOCAL_NAME 	0x03
#define LMX_GAP_WRITE_LOCAL_NAME	0x04
#define LMX_GAP_READ_LOCAL_BDA  	0x05
#define LMX_GAP_SET_SCANMODE		0x06
#define LMX_GAP_SET_FIXED_PIN		0x17
#define LMX_GAP_ACL_ESTABLISHED		0x50
#define LMX_GAP_ACL_TERMINATED		0x51
#define LMX_GAP_GET_PIN				0x75

#define LMX_CHANGE_LOCAL_BDADDRESS 	0x27
#define LMX_STORE_CLASS_OF_DEVICE	0x28
#define LMX_ENABLE_SDP_RECORD		0x29
#define LMX_DELETE_SDP_RECORDS		0x2A
#define LMX_STORE_SDP_RECORD		0x31
#define LMX_WRITE_OPERATION_MODE	0x4A
#define LMX_SET_EVENT_FILTER		0x4E

#define LMX_READY					0x25
#define LMX_CMD_RESET 				0x26
#define LMX_ENTER_BLUETOOTH_MODE 	0x66

#define LMX_SPP_ESTABLISH_LINK		0x0A
#define LMX_SPP_LINK_ESTABLISHED	0x0B
#define LMX_SPP_INCOMING_LINK_ESTABLISHED	0x0C
#define LMX_SPP_RELEASE_LINK		0x0D
#define LMX_SPP_LINK_RELEASED		0x0E
#define LMX_SPP_SEND_DATA			0x0F
#define LMX_SPP_INCOMING_DATA		0x10
#define LMX_SET_PORTS_TO_OPEN		0x22

#define MAX_LASTCFM	100

#define MAX_TXDATA	128

#include "LMXBluetoothPort.h"
#include "PowerManager.h"
/**
 * \brief LMX Bluetooth module class
 *
 * Implements the National LMX-series Bluetooth module functionality.
 */
class LMXBluetoothModule: protected BluetoothModule, protected Thread, public PowerEventListener
{
public:

	LMXBluetoothModule(const char* portName);
	virtual ~LMXBluetoothModule();

	static bool initialize();
	virtual bool shutdown();

	virtual std::string getLocalMac();
	virtual std::string getVersion() { return localFWVer; }

	virtual void setEnabled(bool enabled);

	virtual bool powerStateChange(PowerState oldState, PowerState newState);
protected:
	virtual bool sendSPPData(int connid, const void *data, size_t len);

	virtual bool addSDPRecord(int localPort, void* data, size_t len);

	virtual bool connectSPP(int idConn, unsigned char *dstAddr, unsigned char dstPort);
	virtual bool listenSPP(int idConn);
	virtual bool disconnectSPP(int idConn);

	virtual bool open() { return true; }

private:
	virtual void run();
	bool sendChipEnable(bool e);

	bool openPort();
	bool closePort();

	bool initModule();
	bool initComm();
	bool sendCmd(unsigned char cmd,
			const void *buf1, const size_t len1,
			const void *buf2, const size_t len2,
			int timeout, bool keepCFMMutex);
	void sendPin(unsigned char *mac);

	bool updateOpenPorts();

	bool activateSDPRecord(int localPort, bool enable);

	LMXBluetoothPort*	btPort;

	std::string		commPortName;
	std::string		localMac;
	std::string		localFWVer;

	int openPortMask;
	int activeSPPMask;
	int suspended;

	bool moduleEnabled;

	Mutex			cmdMutex; // one concurrent command

	Mutex			lastCFMMutex; // protect struct lastCFM
	Semaphore		lastCFMSem; // notify of new CFM packet
	struct {					// last command confirmation response
		unsigned char opcode;
		unsigned char data[MAX_LASTCFM];
		int datalen;
	} lastCFM;

	unsigned char sdpId[MAX_BT_CONNECTIONS];
};

#endif /*LMXBLUETOOTHMODULE_H_*/
