/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#include "LMXBluetoothPort.h"

//#define DEBUG
#include "Debug.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default class constructor.
 *
 * Initializes the LMX Bluetooth port object.
 */
LMXBluetoothPort::LMXBluetoothPort()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default class constructor.
 *
 * Initializes the LMX Bluetooth port object and opens the specified serial port.
 *
 * \param portName Points to a null-terminated string containing the serial
 * port name.
 */
LMXBluetoothPort::LMXBluetoothPort(const char* portName)
{
	SerialPort::open(portName);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send command packet to Bluetooth chip
 *
 * Sends the specified command to LMX Bluetooth chip. The interface allows to send
 * 2 buffers of data that will be concatenated. Set buffer pointer to NULL if
 * buffer is not used.
 *
 * \param cmd Command opcode
 * \param buf1 Points to the buffer containing the beginning of command data or NULL.
 * \param len1 Specifies the length of data in the buffer 1.
 * \param buf2 Points to the buffer containing the end of command data or NULL.
 * \param len2 Specifies the length of data in the buffer 2.
 *
 * \return Returns 0 if the packet was sent successfully. Returns a negative
 * integer on error.
 */
bool LMXBluetoothPort::sendCmd(unsigned char cmd,
		const void *buf1, size_t len1,
		const void *buf2, size_t len2)
{
	struct bt_header hdr;
	unsigned long lret;
	size_t len;
	unsigned char etx = ETX;

	if (!buf1) len1 = 0;
	if (!buf2) len2 = 0;
	len = len1 + len2;

	// Test for length
	if (len < 0 || len > MAX_BT_PKT_DATA_LEN)
		return false;
	hdr.len = len;

	// Set STX, CMD
	hdr.stx = STX;
	hdr.pktype = REQ;
	hdr.opcode = cmd;

	// Calculate crc
	hdr.crc = (unsigned char) (0xff & (hdr.pktype + hdr.opcode + hdr.len));

	// Send
	lret = SerialPort::writeData(&hdr, sizeof(struct bt_header));
	if (lret != sizeof(struct bt_header)) return false;

	lret = SerialPort::writeData(buf1, len1);
	if (lret != len1) return false;

	lret = SerialPort::writeData(buf2, len2);
	if (lret != len2) return false;

	lret = SerialPort::writeData(&etx, 1);
	if (lret != 1) return false;

	debug("sent pkt type %c opcode %02x len %04x\n", hdr.pktype, hdr.opcode, hdr.len);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get packet from Bluetooth chip
 *
 * Reads a packet from the LMX Bluetooth chip and performs a few validity checks.
 * The call blocks until a packet is received or a timeout expires.
 *
 * \param type Pointer to variable where received packet type will be returned or NULL
 * \param opcode Pointer to variable where received packet opcode will be returned or NULL
 * \param vbuf Packet data buffer pointer
 * \param buflen Length of vbuf buffer
 * \param timeOut Timeout to wait for packet in ms
 *
 * \return Returns 0 if a valid packet was received. Returns -1 if there was no
 * packet received and timeout expired. Returns a negative integer less than -1
 * on error.
 */
int LMXBluetoothPort::getPkt(unsigned char *type, unsigned char *opcode, void *vbuf, size_t buflen, int timeOut) {
	long lret;
	size_t len, cnt=0;
	struct bt_header hdr = { 0, 0, 0, 0, 0};
	struct timeval start, now;
	bool waitSTX = true;
	char *buf = (char*)vbuf;

	// skip until start symbol
	// if the first symbol is not start symbol, then wait for {ETX,STX}
	gettimeofday(&start, 0);
	do {
		if (!isInputPending(timeOut)) {
			debug("Timeout 1\n");
			return -1;
		}
		lret = SerialPort::readData(&hdr, 1);
		if (lret != 1) {
			debug("Err 1\n");
			return -1;
		}

		if (waitSTX && hdr.stx == STX) break;

		if (hdr.stx == ETX) waitSTX = true;
		else waitSTX = false;

		gettimeofday(&now, 0);
		// Protection against incoming garbage ie. if the chip starts sending \0 characters etc.
		if ((((now.tv_sec - start.tv_sec) * 1000) +
		((now.tv_usec - start.tv_usec) / 1000)) > timeOut) {
			debug("Timeout 1a\n");
			return -1;
		}

	} while (1);

	// retrieve hdr
	while (isInputPending(timeOut)) {
		lret = SerialPort::readData( ((char*)&hdr) + 1 + cnt, sizeof(struct bt_header) -1 -cnt);
		if (lret<=0) {
			debug("Err 2\n");
			return -1;
		}
		cnt += lret;
		if (cnt >= sizeof(struct bt_header)-1) break;
	}
	if (cnt != sizeof(struct bt_header)-1) {
		debug("Timeout 2\n");
		return -1;
	}

	len = hdr.len + 1;
	if (len > buflen) {
		debug("Len err, len %d, buflen %d\n", len, buflen);
		return -2;
	}

	// retrieve the rest of the packet
	cnt = 0;
	while (isInputPending(timeOut)) {
		lret = SerialPort::readData(buf + cnt, len - cnt);
		if (lret<=0) {
			debug("Err 1\n");
			return -1;
		}
		cnt += lret;
		if(cnt >= len) break;
	}
	if (cnt != len) {
		debug("Timeout 3\n");
		return -1;
	}

	// check for etx
	if (buf[hdr.len] != ETX)
		return -3;

	// check for crc
	if (hdr.crc != ((unsigned char) (0xff & (hdr.pktype + hdr.opcode
			+ hdr.len))))
		return -4;

	if (opcode) *opcode = hdr.opcode;
	if (type) *type = hdr.pktype;

	debug("got pkt type %c opcode %02x len %04x\n", hdr.pktype, hdr.opcode, hdr.len);
	if (hdr.len) debug("data0: %02x\n", buf[0]);

	return hdr.len;
}
