/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include "LogFile.h"
#include "StdioFile.h"

#define WRITE_TO_STDOUT

/**
 * \brief Log file constructor
 *
 * Initializes the log file object.
 * \param fileName Specifies the log file name
 */
LogFile::LogFile(const char* fileName)
{
	this->fileName = fileName;
}

/**
 * \brief Log file destructor
 *
 * Destroys the log file object.
 */
LogFile::~LogFile()
{
}

/**
 * \brief Write message to log
 *
 * Writes a text message to the log file.
 * The resulting message will be saved as a separate line in the text file, as specified below.
 *
 * {TIMESTAMP}|{LEVEL}|{ID}|{TEXT}
 *
 * Line breaks are Windows style, in order for log files to be easily readable on all systems.
 *
 * \param logLevel Specifies message type.
 * \param entryId Contains the identifier of the module that generated the given message, or
 * any other identifier that helps classifying messages.
 * \param entryFmt Contains message text. May contain printf-style formatting fields.
 * \param ... Lists arguments for text format fields, printf-style.
 *
 * \return Returns true if writing the log entry was successful, false otherwise.
 */
bool LogFile::logPrintf(LogLevel logLevel, const char* entryId, const char* entryFmt, ...)
{
	bool ret;
	va_list arg;
	va_start(arg, entryFmt);

	ret = logvPrintf(logLevel, entryId, entryFmt, arg);

	va_end(arg);

	return ret;
}

/**
 * \brief Write message to log
 *
 * Writes a text message to the log file.
 * The resulting message will be saved as a separate line in the text file, as specified below.
 *
 * {TIMESTAMP}|{LEVEL}|{ID}|{TEXT}
 *
 * Linebreaks are Windows style, in order for log files to be easily readable on all systems.
 *
 * \param logLevel Specifies message type.
 * \param entryId Contains the identifier of the module that generated the given message, or
 * any other identifier that helps classifying messages.
 * \param entryFmt Contains message text. May contain vsprintf-style formatting fields.
 * \param arg Lists arguments for text format fields, vsprintf-style.
 *
 * \return Returns true if writing the log entry was successful, false otherwise
 */
bool LogFile::logvPrintf(LogLevel logLevel, const char* entryId, const char* entryFmt, va_list arg)
{
	char *buf = NULL;
	char bufTime[32] = "n/a";

	// entry time
	time_t t = time(NULL);
	struct tm time;
	if (localtime_r(&t, &time) != NULL)
		strftime(bufTime, sizeof(bufTime), "%Y-%m-%dT%H:%M:%S%z", &time);

	// entry level
	const char *strLevel = "";
	switch(logLevel) {
	case LOG_DEBUG: 	strLevel="D"; break;
	case LOG_INFO: 		strLevel="I"; break;
	case LOG_WARNING: 	strLevel="W"; break;
	case LOG_ERROR: 	strLevel="E"; break;
	case LOG_CRITICAL: 	strLevel="C"; break;
	default: break;
	}

	// entry text
	int ret = vasprintf(&buf, entryFmt, arg);
	if (ret < 0) return false;

	// make entry
	std::string str = bufTime;
	str += '|';
	str += strLevel;
	str += '|';
	str += entryId;
	str += '|';
	str += buf;
	str += "\r\n";

	// write to stdout?
#ifdef WRITE_TO_STDOUT
	printf("LOG: %s\n", buf);
#endif

	free(buf);

	// write to file?
	StdioFile out;
	if(!out.open(fileName.c_str(), StdioFile::ModeAppend))
		return false;
	out.writeString(str.c_str());
	out.close();
	return true;
}

/////////////////////////////////////////////////////////////////////
/**
 * \brief Log file constructor
 *
 * Initializes the log file object.
 * \param fileName Specifies the log file name
 */
RawLogFile::RawLogFile(const char* fileName)
{
	this->fileName = fileName;
}

/**
 * \brief Log file destructor
 *
 * Destroys the log file object.
 */
RawLogFile::~RawLogFile()
{
}

/**
 * \brief Write message to log
 *
 * Writes a text message to the log file.
 * The resulting message will be saved as a separate line in the text file
 *
 * Linebreaks are Windows style, in order for log files to be easily readable on all systems.
 *
 * \param entryFmt Contains message text. May contain printf-style formatting fields.
 * \param ... Lists arguments for text format fields, printf-style.
 *
 * \return Returns true if writing the log entry was successful, false otherwise.
 */
bool RawLogFile::logPrintf(const char* entryFmt, ...)
{
	bool ret;
	va_list arg;
	va_start(arg, entryFmt);

	ret = logvPrintf(entryFmt, arg);

	va_end(arg);

	return ret;
}

/**
 * \brief Write message to log
 *
 * Writes a text message to the log file.
 *
 * Linebreaks are Windows style, in order for log files to be easily readable on all systems.
 *
 * \param entryFmt Contains message text. May contain vsprintf-style formatting fields.
 * \param arg Lists arguments for text format fields, vsprintf-style.
 *
 * \return Returns true if writing the log entry was successful, false otherwise.
 */
bool RawLogFile::logvPrintf(const char* entryFmt, va_list arg)
{
	char *buf = NULL;

	// entry text
	int ret = vasprintf(&buf, entryFmt, arg);
	if (ret < 0 || buf == NULL) return false;

	// write to stdout?
#ifdef WRITE_TO_STDOUT
	printf("LOG: %s\n", buf);
#endif

	// write to file?
	StdioFile out;
	if(!out.open(fileName.c_str(), StdioFile::ModeAppend)) {
		free(buf);
		return false;
	}
	out.writeString(buf);
	out.writeString("\r\n");
	out.close();

	free(buf);
	return true;
}
