/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include "Media.h"

#include <stdio.h>
#include <string.h>
#include <sys/mount.h>

//#define DEBUG
#include "Debug.h"

#define MOUNT_RW	0x40000000

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Mount point table
 *
 * Lists all system mount points for appropriate media devices. Table entries
 * contain MountPoint structures, that specify the mount point path, media device
 * file, file system and the desired unmount state (full unmount or remount to
 * read-only).
 */
MountPoint Media::mountTable[] =
{
	{ "/mnt/mmc", "/dev/mmcblk0p1", "vfat", MEDIA_STATE_READONLY, 0, 0 },
	{ "", "", "", MEDIA_STATE_ERROR, 0, 0 }
};

Mutex Media::mountMutex;
bool Media::mountEnabled = true;

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable media mounter
 *
 * Enables automatic mounting of media. The files located on that media may
 * be opened when accessed through mountMedia() function or via StdioFile.
 */

void Media::mountEnable()
{
	mountMutex.lock();
	mountEnabled = true;
	mountMutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Disable media mounter
 *
 * Disables automatic mounting of media. The files located on that media
 * will fail to open, when accessed through mountMedia() function or via
 * StdioFile.
 */

void Media::mountDisable()
{
	mountMutex.lock();
	mountEnabled = false;
	mountMutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Mount media associated with file
 *
 * Mounts the media associated with the given file path, if required. The media
 * is mounted to a specific mount point with a fixed identifier, which will be
 * returned by the function on success. If the file is located on media that
 * does not require any special handling, the function will return immediately.
 *
 * \param fileName Contains full path to the file located on media.
 * \param writeFlag Specifies if file write access is required.
 * \return Returns a positive integer if the media has been mounted successfully.
 * The return value identifies the mount point associated with the media and
 * should be further passed to unmountMedia() function.
 * Returns MOUNT_NOT_REQUIRED = 0 if media mounting is not required for the
 * given file. Returns a negative integer (MOUNT_FAILED) if the call failed.
 */

int Media::mountMedia(const char* fileName, bool writeFlag)
{
	mountMutex.lock();

	if (!mountEnabled) {
		debug("mounting is DISABLED\n");
		mountMutex.unlock();
		return MOUNT_FAILED;
	}

	debug("file %s write %d\n", fileName, writeFlag);

	// look through the mount table
	for(int i=0; mountTable[i].unmountType != MEDIA_STATE_ERROR; i++)
	{
		if(belongsToMountPoint(mountTable[i].mountPoint, fileName))
		{
			// get current mounting state
			int state = getMediaState(mountTable[i].mountPoint);
			debug("current %s, state %d, ro %d, rw %d\n", mountTable[i].mountPoint, state,
					mountTable[i].num_writers, mountTable[i].num_writers);

			// if media is already mounted, check the mode
			if(	(state == MEDIA_STATE_READWRITE) ||
				(state == MEDIA_STATE_READONLY && !writeFlag))
			{
				debug("no mount required\n");
				mountMutex.unlock();
				if (writeFlag) {
					mountTable[i].num_writers++;
					return (i+1) | MOUNT_RW;
				}
				mountTable[i].num_readers++;
				return i+1;
			}

			// set mounting options
			unsigned long mountOpts = (state > MEDIA_STATE_UNMOUNT) ?
				MS_REMOUNT : 0;

			debug("performing mount on %s, %x\n",
					mountTable[i].devName, (unsigned int)mountOpts);

			// perform the mount system call
			if(!mount(mountTable[i].devName, mountTable[i].mountPoint,
						mountTable[i].fsType, mountOpts, ""))
			{
				// success, return mount point ID
				mountMutex.unlock();
				if (writeFlag) {
					mountTable[i].num_writers++;
					return (i+1) | MOUNT_RW;
				}
				mountTable[i].num_readers++;
				return i+1;
			}
			else
			{
				mountMutex.unlock();
				return MOUNT_FAILED;
			}
		}
	}

	debug("suitable mount point not found\n");
	mountMutex.unlock();
	return MOUNT_NOT_REQUIRED;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unmount media
 *
 * Unmount the media previously mounted with mountMedia(). The media may
 * not have any open files located on it in order to be successfully unmounted.
 *
 * \param mountPoint Contains mount point identifier returned by the mount method.
 * \return Returns true if the operation was successful, false otherwise
 */
bool Media::unmountMedia(int mountPoint)
{
	bool rw = false;
	if (mountPoint & MOUNT_RW) rw = true;
	mountPoint &= ~MOUNT_RW;

	// sanity check
	if(mountPoint < 1 || mountPoint >
		(int)(sizeof(mountTable)/sizeof(MountPoint))) return false;

	mountMutex.lock();

	// structure item is mount point ID minus 1
	mountPoint --;
	bool ok = false;

	if (rw) {
		if (mountTable[mountPoint].num_writers) {
			mountTable[mountPoint].num_writers--;
		}
	} else {
		if (mountTable[mountPoint].num_readers) {
			mountTable[mountPoint].num_readers--;
		}
	}

	// if there are writers left, do not un-mount
	if (mountTable[mountPoint].num_writers) {
		mountMutex.unlock();
		return ok;
	}

	// check the required unmount operation
	switch(mountTable[mountPoint].unmountType)
	{
	case MEDIA_STATE_UNMOUNT:

		debug("unmounting %d\n", mountPoint);

		// if there are readers left, umount as read-only
		if (mountTable[mountPoint].num_readers) {
			// perform a full unmount
			ok = umount(mountTable[mountPoint].mountPoint) == 0;
			break;
		}
		// fall through

	case MEDIA_STATE_READONLY:

		debug("remounting %d read-only\n", mountPoint);

		// perform a read-only re-mount
		ok = mount(	mountTable[mountPoint].devName,
					mountTable[mountPoint].mountPoint,
					mountTable[mountPoint].fsType,
					MS_REMOUNT | MS_RDONLY, "") == 0;
		break;
	}

	mountMutex.unlock();
	return ok;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unmount all media
 *
 * Unmounts all available media. Should be called before the system enters a
 * low-power state.
 *
 * \return Returns true if all media mount points have been successfully
 * unmounted.
 */
bool Media::unmountAll()
{
	mountMutex.lock();
	bool ok = true;

	for(int i=0; mountTable[i].unmountType != MEDIA_STATE_ERROR; i++)
	{
		if(getMediaState(mountTable[i].mountPoint) == MEDIA_STATE_UNMOUNT)
		{
			mountTable[i].num_readers = 0;
			mountTable[i].num_writers = 0;
			debug("%s is not mounted\n", mountTable[i].mountPoint);
			continue;
		}

		debug("unmounting %s\n", mountTable[i].mountPoint);

		// perform a full unmount
		ok = umount(mountTable[i].mountPoint) == 0;
		if(!ok)
		{
			debug("failed to unmount %s\n", mountTable[i].mountPoint);
			break;
		}

		mountTable[i].num_readers = 0;
		mountTable[i].num_writers = 0;
	}

	mountMutex.unlock();
	return ok;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get media state given a mount point.
 *
 * Retrieves current media mount state for the given mount point identifier.
 *
 * \param mountPoint Contains the mount point identifier, as returned by
 * mountMedia() function.
 * \return Returns MEDIA_STATE_READWRITE or MEDIA_STATE_READONLY if the media
 * is mounted, depending on its read/write setting. Returns MEDIA_STATE_UNMOUNT
 * if the media is not mounted. Returns MEDIA_STATE_ERROR on error.
 */
int Media::getMediaState(int mountPoint)
{
	// sanity check
	if(mountPoint < 1 || mountPoint >
		(int)(sizeof(mountTable)/sizeof(MountPoint))) return MEDIA_STATE_ERROR;

	return getMediaState(mountTable[mountPoint-1].mountPoint);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get media state given a file name
 *
 * Retrieves current media mount state for the given file name or a mount point
 * path.
 *
 * \param fileName Contains a mount point path or a full file path on the media.
 * \return Returns MEDIA_STATE_READWRITE or MEDIA_STATE_READONLY if the media
 * is mounted, depending on its read/write setting. Returns MEDIA_STATE_UNMOUNT
 * if the media is not mounted. Returns MEDIA_STATE_ERROR on error.
 */
int Media::getMediaState(const char *fileName)
{
	// open the system mounts file
	FILE* in = fopen("/proc/mounts", "r");
	if(!in) return MEDIA_STATE_ERROR;

	char buf[196]="";
	// read the file line by line
	while (fgets(buf, sizeof(buf), in))
	{
		char dev[65]="", point[33]="", fs[33]="", mode[65]="";
		// parse the row, drop it if invalid
		if (sscanf(buf, "%64s %32s %32s %64s", dev, point, fs, mode) != 4)
			continue;

		char *comma = strchr(mode, ',');
		if(comma) *comma = 0;

		if (belongsToMountPoint(point, fileName))
		{
			debug("%s -> %s, %s\n", dev, point, mode);

			fclose(in);
			if(!strcmp(mode, "ro")) return MEDIA_STATE_READONLY;
			else if(!strcmp(mode, "rw")) return MEDIA_STATE_READWRITE;
			else return MEDIA_STATE_ERROR;
		}
	}

	fclose(in);
	// reached the end, mount point not found
	return MEDIA_STATE_UNMOUNT;
}

////////////////////////////////////////////////////////////////////////////////
//
bool Media::belongsToMountPoint(const char *mountPoint, const char *fileName)
{
	int mountLen = strlen(mountPoint);
	int fileLen = strlen(fileName);

	if(mountLen > fileLen) return false;
	else if(mountLen == fileLen)
		return strcmp(mountPoint, fileName) == 0;

	if(!strncmp(mountPoint, fileName, mountLen) && fileName[mountLen] == '/')
		return true;

	return false;
}

