/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>

#include "ModemPort.h"

//#define DEBUG
#include "Debug.h"

const char* ModemPort::modemAnswerStrings[] =
{ "\nOK\r\n", "\nCONNECT\r\n", "\nRING\r\n", "\nNO CARRIER\r\n", "\nERROR\r\n",
		"\nNO ANSWER\r\n", "\n+CMS ERROR:", "\n+CME ERROR:", NULL };

size_t ModemPort::modemAnswerSizes[] =
{ 5, 10, 7, 13, 8, 12, 12, 12 };

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Default class constructor
 *
 * Initializes the modem port object.
 */
ModemPort::ModemPort()
{
	answerType = AT_INVALID;
	noCarrierDetect = 0;
}

/**
 * \brief Class constructor
 *
 * Initializes the modem port object and opens the specified serial port.
 * \param portName Points to a null-terminated string containing the serial
 * port name.
 */
ModemPort::ModemPort(const char* portName)
{
	answerType = AT_INVALID;
	noCarrierDetect = 0;
	SerialPort::open(portName);
}

/**
 * \brief Class destructor
 *
 * Destroys the modem port object.
 */
ModemPort::~ModemPort()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read modem response
 *
 * Reads the modem reply to a previously sent AT command.
 * The function blocks until the modem response is received or the timeout
 * expires.
 *
 * \param pData Pointer to a buffer that receives the command reply.
 * \param maxSize Specifies maximum length of the modem reply.
 * \param timeOut Specifies the modem response timeout, in milliseconds.
 * \return Returns number of characters received from the modem.
 */
size_t ModemPort::read(char *pData, size_t maxSize, int timeOut)
{
	size_t i = 0, lastLF;
	char readB;
	answerType = AT_INVALID; // reset the modem answer type
	if (!isInputPending(timeOut)) return 0;

	// check if there's data pending
	while (isInputPending(timeOut))
	{
		readData((uint8_t*) &readB, 1);
		debug("%c", readB);
		if (pData != NULL) pData[i] = readB;

		if ((readB == '\n') && (pData != NULL))
		{
			// seek for previous linebreak, if any
			for (lastLF = i - 1; lastLF > 0; lastLF --)
				if (pData[lastLF] == '\n') break;

			for (int n = 0; modemAnswerStrings[n]; n ++)
			{
				// check that the received string length equals to
				// the valid response length
				if ((i - lastLF + 1) < modemAnswerSizes[n]) continue;

				// check that the string matches
				if (!memcmp(&pData[lastLF], modemAnswerStrings[n],
						modemAnswerSizes[n]))
				{
					answerType = (ModemAnswerType) n;
					return ++ i;
				}
			}
		}

		if (++ i >= maxSize) break;
	}

	return i;
}

////////////////////////////////////////////////////////////////////////////////

bool ModemPort::checkNoCarrier(char inChar)
{
	if (inChar == modemAnswerStrings[3][noCarrierDetect])
	{
		noCarrierDetect ++;
		if (noCarrierDetect == modemAnswerSizes[3])
		{
			noCarrierDetect = 0;
			return true;
		}
	}
	else
	{
		noCarrierDetect = 0;
		if (inChar == modemAnswerStrings[3][0]) noCarrierDetect ++;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Wait for no-carrier condition
 *
 * Wait until the modem sends NO CARRIER.
 */
void ModemPort::waitNoCarrier(void)
{
	for (int i = 0; i < 50; i ++)
	{
		int count = numInputPending(100);
		char readB;
		while (count)
		{
			readData((uint8_t *) &readB, 1);
			if (checkNoCarrier(readB)) return;
			count --;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write AT command to the modem
 *
 * Sends the specified AT command to the modem.
 *
 * \param pData Pointer to the null terminated AT command string.
 * \return Returns number of characters written to modem port. On error, returns -1.
 */
ssize_t ModemPort::write(const char *pData)
{
	debug("Write:\n%s\n", pData);
	return writeData((uint8_t*) pData, strlen(pData));
}
