/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

/*
 * NetworkConnection.cpp
 *
 *  Created on: Nov 19, 2008
 *      Author: andrei
 */


#include <resolv.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <string.h>

#include <exynthoo/GuiLibrary.h>

//#define DEBUG
#include "Debug.h"

#include "LibSpatha.h"
#include "NetworkConnection.h"

#define NETWORK_POLL_INTERVAL	3
#define NET_REPAIR_WAIT			10
#define NETWORK_CONN_FAIL_RETRY	5

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the network connection object.
 * The network type identifier is stored for external use via getNetworkType()
 * method. It is recommended to use the platform specific types that match
 * the communication module type identifier of the mobile data terminal.
 *
 * \param netType Specifies the network type identifier.
 */
NetworkConnection::NetworkConnection(int netType)
					: networkType(netType)
{
	cntPollInfo = NETWORK_POLL_INTERVAL;
	lastEvent = NE_OFFLINE;
	currentState = OFFLINE;
	lastEventMsg = NULL;
	lastError = NULL;
}

/**
 * \brief Class destructor
 *
 * Terminates the network connection object.
 */
NetworkConnection::~NetworkConnection()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize network connection
 *
 * Initializes the network connection. The connection may be initialized only
 * once. Internally this method calls netInitialize() virtual function which
 * should be overridden by a specific network implementation.
 *
 * \return Returns true if the network connection was initialized successfully.
 */
bool NetworkConnection::initialize()
{
	stateMutex.lock();
	if (currentState != OFFLINE)
	{
		stateMutex.unlock();
		debug("connection already initialized\n");
		return false;
	}
	stateMutex.unlock();

	debug("initialize\n");
	return netInitialize();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Shutdown network connection
 *
 * Shuts down and uninitializes the network connection. The connection may be
 * shut down only if it is disconnected.
 *
 * Internally this method calls netShutdown() virtual function which should
 * be overridden by a specific network implementation.
 *
 * \return Returns true if the network connection was shut down successfully.
 */
bool NetworkConnection::shutdown()
{
	stateMutex.lock();
	if (currentState != OFFLINE)
	{
		debug("network connection not closed\n");
		stateMutex.unlock();
		return false;
	}
	stateMutex.unlock();

	debug("shutdown\n");
	return netShutdown();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Connect to network
 *
 * Connects to network. Starts the worker thread and launches the state machine
 * of the network connection.
 *
 * The function returns immediately, typically before the network is actually
 * connected. The caller may monitor network activity via NetworkEventListener
 * interface.
 *
 * \return Returns true if the connection was initiated successfully.
 */
bool NetworkConnection::connect()
{
	bool ret = Thread::start();
	debug("starting worker - %s\n", ret ? "OK" : "FAILED");

	PowerManager *pPM = PowerManager::getPowerManager();
	pPM->addListener(this);

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Disconnect network
 *
 * Disconnects from network. Stops the worker thread and calls netDisconnect()
 * internally.
 *
 * The function blocks until the network is disconnected, or an error occurs.
 *
 * \return Returns true if the connection was closed successfully.
 */
bool NetworkConnection::disconnect()
{
	PowerManager *pPM = PowerManager::getPowerManager();
	pPM->removeListener(this);

	debug("stopping worker\n");
	Thread::interrupt();
	if (!Thread::stop())
	{
		debug("stop failed\n");
		return false;
	}

	debug("disconnecting\n");
	if (!netDisconnect())
	{
		debug("failed disconnecting\n");
		return false;
	}

	currentState = OFFLINE;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Notify listeners about event
 *
 * Broadcasts the specified event by calling all registered listeners.
 *
 * \param e Specifies the event identifier to be delivered.
 */
void NetworkConnection::eventNotify(NetworkEvent e)
{
	lastEvent = e;
	if (beginEnumListeners())
	{
		while (NetworkEventListener *p = getNextListener())
			p->networkEvent(e);
		endEnumListeners();
	}
#ifdef USE_XYNTH
	while (1) {
		if (GuiLibrary::timedlock(1000)) break;
		if (stopRequested()) return;
	}
	GuiLibrary::broadcastSystemEvent(EV_NETWORK_STATE, (void*)e);
	GuiLibrary::unlock();
#endif
}

/**
 * \brief Get last network event (current status)
 * \return Returns the network event last occurred
 */
NetworkEvent NetworkConnection::getLastEvent()
{
	return lastEvent;
}

////////////////////////////////////////////////////////////////////////////////

bool NetworkConnection::powerStateChange(PowerState oldState, PowerState newState)
{
	bool ret = true;
	if (newState == LOW_POWER_SLEEP)
	{
		stateMutex.lock();
		if (currentState != STANDBY)
		{
			debug("suspend - postponing powersave\n");
			currentState = SUSPEND;
			ret = false;
		}
		else debug("entering powersave\n");
		stateMutex.unlock();
	}
	else if (newState == FULLY_OPERATIONAL)
	{
		stateMutex.lock();
		if (currentState == STANDBY)
		{
			debug("leaving powersave\n");
			currentState = RESUME;
		}
		stateMutex.unlock();
	}

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Worker thread
 *
 * Runs the state machine of the network connection. Tries to connect to
 * network initially. When connected, checks for network availability and
 * tries to restore the network connection in case of a failure.
 */
void NetworkConnection::run()
{
	while(true)
	{
		if (Thread::stopRequested()) break;

		stateMutex.lock();
		NetworkConnectionState state = currentState;
		stateMutex.unlock();

		debug("state %d\n", state);

		switch(state)
		{
		case OFFLINE:

			debug("connecting...\n");
			pollNetworkInfo();

			// connect to network now
			eventNotify(NE_CONNECTING);
			if (netConnect())
			{
				stateMutex.lock();
				currentState = CONNECTED;
				stateMutex.unlock();

				// re-initialize resolver for new network connection
				// this fix is for glibc 2.3.6, may not be required with newer
				res_init();

				eventNotify(NE_CONNECTED);
				break;
			}
			// wait a bit before trying again
			eventNotify(NE_OFFLINE);
			Thread::sleep(NETWORK_CONN_FAIL_RETRY * 1000);
			break;

		case SUSPEND:

			debug("suspending\n");
			// suspend network connection
			if (netSuspend())
			{
				stateMutex.lock();
				currentState = STANDBY;
				stateMutex.unlock();
				eventNotify(NE_OFFLINE);
				break;
			}
			// wait a bit before trying again
			Thread::sleep(NETWORK_CONN_FAIL_RETRY * 1000);
			break;

		case RESUME:

			debug("resuming\n");
			// resume network connection
			eventNotify(NE_CONNECTING);
			if (netResume())
			{
				stateMutex.lock();
				currentState = CONNECTED;
				stateMutex.unlock();
				eventNotify(NE_CONNECTED);
				break;
			}
			// wait a bit before trying again
			eventNotify(NE_OFFLINE);
			Thread::sleep(1000);
			break;

		case STANDBY:

			Thread::sleep(NETWORK_CONN_FAIL_RETRY * 1000);
			break;

		case REPAIR:
			// Wait for some time, maybe the connection will come back
			if (cntNetRepair < (NET_REPAIR_WAIT * 2)) {
				if (isConnectionUp())
				{
					stateMutex.lock();
					currentState = CONNECTED;
					stateMutex.unlock();
					eventNotify(NE_CONNECTED);
					break;
				}
				Thread::sleep(500);
				cntNetRepair++;
				break;
			}

			debug("repairing\n");
			// repair network connection
			if (netRepair())
			{
				stateMutex.lock();
				currentState = CONNECTED;
				stateMutex.unlock();
				eventNotify(NE_CONNECTED);
				break;
			}
			// wait a bit before trying again
			eventNotify(NE_OFFLINE);
			Thread::sleep(NETWORK_CONN_FAIL_RETRY * 1000);
			eventNotify(NE_CONNECTING);
			break;

		case CONNECTED:

			// check that network connection is fine
			if (!isConnectionUp())
			{
				debug("connection is down\n");

				stateMutex.lock();
				currentState = REPAIR;
				cntNetRepair = 0;
				stateMutex.unlock();
				eventNotify(NE_CONNECTING);
				break;
			}

			pollNetworkInfo();

			// all ok - sleep for a while
			Thread::sleep(1000);
			break;
		}

		if (Thread::stopRequested()) break;
	}

	debug("finished\n");
}

////////////////////////////////////////////////////////////////////////////////
// stubs
/**
 * \brief Suspend network connection
 *
 * Called internally when the mobile data terminal enters a power saving mode.
 * The implementation may override this method to take a specific action.
 *
 * \return Return true when ready to enter a power saving mode. Return false
 * to reject the request and abort power saving.
 */
bool NetworkConnection::netSuspend()
{
	return true;
}

/**
 * \brief Restore network connection
 *
 * Called internally when the mobile data terminal leaves a power saving mode.
 * The implementation may override this method to take a specific action.
 *
 * \return Return true when the connection is restored successfully.
 */
bool NetworkConnection::netResume()
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Repair network connection
 *
 * Called internally when the state machine of the network connection detects
 * that a network failure. The default behavior is to disconnect from network
 * by calling netDisconnect() and then re-connect by calling netConnect()
 * virtual methods.
 *
 * The implementation may override this method to take a specific action.
 *
 * \return Return true if the connection was repaired successfully.
 */

bool NetworkConnection::netRepair()
{
	if (!netDisconnect())
	{
		debug("repair failed - could not disconnect\n");
		return false;
	}

	if (!netConnect())
	{
		debug("repair failed - could not reconnect\n");
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////

void NetworkConnection::pollNetworkInfo()
{
	if (++ cntPollInfo > NETWORK_POLL_INTERVAL)
	{
		GENERIC_NETWORK_INFO netInfo;
		if (getNetworkInfo(netInfo))
		{
			// do not show signal strength for bogus channels
			int n = netInfo.isConnected ? netInfo.signalStrength : 0;

			if (beginEnumListeners())
			{
				while (NetworkEventListener *p = getNextListener())
					p->networkSignalQuality(n);
				endEnumListeners();
			}
#ifdef USE_XYNTH
			while (1) {
				if (GuiLibrary::timedlock(1000)) break;
				if (stopRequested()) return;
			}
			GuiLibrary::broadcastSystemEvent(EV_NETWORK_SIGNAL, (void*)n);
			GuiLibrary::unlock();
#endif
		}
		cntPollInfo = 0;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get interface addresses
 *
 * Retrieves IP and MAC addresses of the network connection interface.
 *
 * \param ifName Points to a null-terminated string containing the name of
 * the interface.
 * \param ipAddr Points to a string buffer that receives the IP address of the
 * network interface.
 * \param hwAddr Points to a string buffer that receives the hardware MAC
 * address of the network interface.
 *
 * \return Returns true if address was successfully stored in a buffer, false otherwise
 */
bool NetworkConnection::getInterfaceAddresses(const char* ifName,
						char *ipAddr, char *hwAddr)
{
	// get interface addresses
	int s = socket(AF_INET, SOCK_STREAM, 0);
	if (s > 0)
	{
		struct ifreq ifr;
		strncpy(ifr.ifr_name, ifName, IFNAMSIZ);

		if (ioctl(s, SIOCGIFADDR, (char*)&ifr) == 0)
		{
			in_addr_t addr = ((struct sockaddr_in*) &ifr.ifr_addr)->sin_addr.s_addr;
			sprintf(ipAddr, "%d.%d.%d.%d", addr & 0xff,
					addr>>8 & 0xff, addr>>16 & 0xff, addr>>24 & 0xff);
		}
		else debug("failed reading interface address\n");

		if (ioctl(s, SIOCGIFHWADDR, (char*)&ifr) == 0)
		{
			char addr[6];
			memcpy(addr, &ifr.ifr_hwaddr.sa_data, 6);
			sprintf(hwAddr, "%02X:%02X:%02X:%02X:%02X:%02X",
					addr[0], addr[1], addr[2], addr[3], addr[4], addr[5]);
		}
		else debug("failed reading interface hw address\n");

		close(s);
		return true;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Report network connection error
 * Reports the network connection error by broadcasting an EV_NETWORK_ERROR system event
 *
 * \param error Pointer to the error description
 */
void NetworkConnection::reportError(const char *error)
{
	debug(error);

	lastError = error;
	lastEventMsg = error;

	if (beginEnumListeners())
	{
		while (NetworkEventListener *p = getNextListener())
			p->networkError(error);
		endEnumListeners();
	}
#ifdef USE_XYNTH
	while (1) {
		if (GuiLibrary::timedlock(1000)) break;
		if (stopRequested()) return;
	}
	GuiLibrary::broadcastSystemEvent(EV_NETWORK_ERROR, (void*)error);
	GuiLibrary::unlock();
#endif
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Report the network status
 * Reports the network status by broadcasting an EV_NETWORK_STATUS system event
 *
 * \param status Pointer to the status description
 */
void NetworkConnection::setStatus(const char *status)
{
	debug("%s\n", status);

	lastStatus = status;
	lastEventMsg = status;

	if (beginEnumListeners())
	{
		while (NetworkEventListener *p = getNextListener())
			p->networkStatus(status);
		endEnumListeners();
	}
#ifdef USE_XYNTH
	while (1) {
		if (GuiLibrary::timedlock(1000)) break;
		if (stopRequested()) return;
	}
	GuiLibrary::broadcastSystemEvent(EV_NETWORK_STATUS, (void*)status);
	GuiLibrary::unlock();
#endif
}

////////////////////////////////////////////////////////////////////////////////
// overriding EventBroadcaster implementation because we must submit the last
// event automatically

void NetworkConnection::addListener(NetworkEventListener *pListener)
{
	EventBroadcaster<NetworkEventListener>::addListener(pListener);
	pListener->networkEvent(lastEvent);
}
