/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <linux/input.h>
#include <signal.h>
#include <sys/reboot.h>
#include <linux/rtc.h>
#include <errno.h>
#include <platform.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>

//#define DEBUG
#include "Debug.h"

#include "exynthoo/GuiLibrary.h"
#include "exynth/exynth.h"
#include "PowerManager.h"
#include "LibSpatha.h"
#include "Media.h"

#define BATTERY_INTEGR_COEFF		0.3

// the one and the only power manager object
PowerManager* PowerManager::powerManager = NULL;

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize power manager
 *
 * Creates the power manager singleton object and initializes it. Must be
 * called before any other power manager function.
 *
 * \note Device initialization is performed by calling the initialize() method
 *
 * \return Returns true if the power manager initialization was a success.
 */
bool PowerManager::initPowerManager()
{
	if(powerManager)
	{
		debug("Power manager is already created\n");
		return false;
	}

	powerManager = new PowerManager();
	if(!powerManager->initialize())
	{
		debug("PM failed init\n");
		delete powerManager;
		powerManager = NULL;
		return false;
	}

	debug("Power manager initialized successfully\n");
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Release power manager
 *
 * Uninitializes the power manager singleton object and deletes it. Must be
 * called at the end of the application life.
 *
 * \returns Returns true if the power manager was released successfully.
 */
bool PowerManager::deinitPowerManager()
{
	if(!powerManager)
	{
		debug("Power manager is not inialized\n");
		return false;
	}

	if(!powerManager->deinitialize())
	{
		debug("PM deinitialization failed\n");
		return false;
	}

	delete powerManager;
	powerManager = NULL;

	debug("Power manager deinitialized\n");
	return true;
}

/**
 * \brief Retrieve power manager
 *
 * Static function used to return a pointer to the one and the only power
 * manager object.
 *
 * \return Returns a pointer to the power manager singleton object.
 */
PowerManager* PowerManager::getPowerManager()
{
	return powerManager;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Presets power manager attributes. Maximum screen brightness is initially
 * set to 255.
 */
PowerManager::PowerManager()
{
	idleBegin = 0;
	currentState = POWER_DOWN;
	buzzerDriver = -1;
	powerDriver = -1;
	rtcDriver = -1;
	enableStateSwitch = true;
	displayEnabled = false;
	maxDisplayBrightness = 255;
	powerSaveSleep = POWER_SAVE_SLEEP;
	batteryVoltage = -1;
}

/**
 * \brief Class destructor
 *
 * Destroys the power manager object.
 */
PowerManager::~PowerManager()
{
}

////////////////////////////////////////////////////////////////////////////////

int PowerManager::findSpeakerDevice()
{
	char devFile[32];
	unsigned char bitMask[EV_MAX/8 + 1];

	for (int i = 0; i < 32; i ++)
	{
		snprintf(devFile, sizeof(devFile), "/dev/input/event%d", i);
		int fd = open(devFile, O_RDWR| O_NONBLOCK);
		if (fd == -1) continue;

		if (ioctl(fd, EVIOCGBIT(0, sizeof(bitMask)), bitMask) < 0)
		{
			close(fd);
			continue;
		}

		if (bitMask[EV_SND/8] & (1 << (EV_SND % 8)))
		{
			debug("Found an EV_SND capable device: %d\n", i);
			return fd;
		}
	}

	return -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialization routine
 * The actual routine to perform power manager initialization:
 * - Tries to open power management driver
 * - Finds speaker device
 * - Opens RTC device
 * - Sets power mode to FULLY_OPERATIONAL
 *
 * \return Returns true if initialization was performed successfully, false otherwise
 *
 */
bool PowerManager::initialize()
{
	powerDriver = open(POWER_DEVICE_NAME, O_RDWR| O_NONBLOCK);
	if (powerDriver == -1)
	{
		debug("Failed opening power management driver\n");
		return false;
	}

	buzzerDriver = findSpeakerDevice();
	if (buzzerDriver == -1)
	{
		debug("Failed opening buzzer driver\n");
		return false;
	}

	rtcDriver = open(RTC_DEVICE_NAME, O_RDWR | O_NONBLOCK);
	if(rtcDriver < 0)
	{
		debug("Failed opening Real-Time Clock device\n");
		return false;
	}

	// initialize RTC if not ready
	if (!rtcGetTime(NULL))
	{
		debug("Cold boot - starting RTC at 1st Jan 1970\n");
		struct tm localTime = { 0, 0, 0, 1, 0, 70, 0, 0, 0 };

		// set local time, requires super-user privileges
		time_t initTime = mktime(&localTime);
		stime(&initTime);

		// update RTC driver's time
		rtcSetTime(initTime);
	}


	idleBegin = s_gettimeofday()/1000;

	gprsSetDTR(true);
	oledEnable(true);
	oledFreeze(false);

	setPowerMode(FULLY_OPERATIONAL);

#ifdef USE_XYNTH
	s_exynth_set_idle_func(kickIdleProc);
#endif
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Uninitialization routine
 * The actual routine to perform device uninitialization. Closes
 * speaker device, power driver and RTC driver handles.
 *
 * \return Returns true if the uninitialization was performed successfully, false otherwise
 */
bool PowerManager::deinitialize()
{
	s_exynth_set_idle_func(NULL);

	oledFreeze(true);

	if (buzzerDriver >=0 ) close(buzzerDriver);
	if (powerDriver >=0 ) close(powerDriver);
	if (rtcDriver >=0 ) close(rtcDriver);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Deferring idle mode
 *
 * Instructs power manager that the device is not idle at the moment. Power
 * manager saves current timestamp that is further used for calculation of the
 * idle duration.
 *
 * \return Returns false if system was already idle, true otherwise
 */
bool PowerManager::kickIdle()
{
	int ret = false;
	powerMutex.lock();

	if (enableStateSwitch)
	{
		debug("Idle reset\n");
		idleBegin = s_gettimeofday()/1000;
		if (currentState != FULLY_OPERATIONAL) ret = true;
		setPowerMode(FULLY_OPERATIONAL);
	}

	powerMutex.unlock();
	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable power state switching
 *
 * Enables or disables power manager state switching.
 *
 * \param flag When set to false, setPowerState() method will not have any
 * effect. When set to true, setPowerState() operates normally.
 */
void PowerManager::enableTransitions(bool flag)
{
	powerMutex.lock();
	enableStateSwitch = flag;
	powerMutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get battery capacity and voltage
 *
 * Reads current battery capacity in percent. Alternatively, reads approximate
 * battery voltage for post-processing. Note that averaging is applied for successive
 * reads, to reduce voltage jitter and drift.
 *
 * \param pVoltage Pointer to integer that holds the approximated voltage level.
 * If the parameter is not NULL, the voltage is stored in millivolts.
 * \return Returns battery capacity in percent (0..100). In case of an error
 * returns -1.
 */
int PowerManager::getBatteryCapacity(int *pVoltage)
{
	int battDriver = open(VBATT_DEVICE_NAME, O_RDONLY);
	if (battDriver < 0)
	{
		debug("Failed opening battery node\n");
		return -1;
	}

	char buf[16] = "";
	ssize_t readBytes = read(battDriver, buf, sizeof(buf));
	close(battDriver);

	if (readBytes < 0) return -1;

	int mv = atoi(buf);
	if (mv <= 0)
	{
		debug("invalid input %d\n", mv);
		return -1;
	}

	// first readout - initialize stored value
	if (batteryVoltage < 0) batteryVoltage = mv;
	// subsequent reads - running average
	else batteryVoltage = (int)((float)batteryVoltage * (1 - BATTERY_INTEGR_COEFF) +
		(float)mv * BATTERY_INTEGR_COEFF);

	debug("instant %d average %d\n", mv, batteryVoltage);

	if (pVoltage) *pVoltage = batteryVoltage;

	if (batteryVoltage >= BATTERY_MAX_VOLTAGE) return 100;
	if (batteryVoltage <= BATTERY_MIN_VOLTAGE) return 0;

	return (batteryVoltage - BATTERY_MIN_VOLTAGE) * 100 / (BATTERY_MAX_VOLTAGE
			- BATTERY_MIN_VOLTAGE);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get battery voltage
 *
 * Reads current battery raw voltage for post-processing.
 *
 * \return Returns battery voltage in millivolts. Returns -1 in case of an error.
 */
int PowerManager::getBatteryVoltage()
{
	int battDriver = open(VBATT_DEVICE_NAME, O_RDONLY);
	if (battDriver < 0)
	{
		debug("Failed opening battery node\n");
		return -1;
	}

	char buf[16] = "";
	ssize_t readBytes = read(battDriver, buf, sizeof(buf));
	close(battDriver);

	if (readBytes < 0) return -1;

	int mv = atoi(buf);
	batteryVoltage = (mv < 0) ? -1 : mv;
	debug("%d, %d\n", mv, batteryVoltage);
	return batteryVoltage;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable LED outputs
 *
 * Enables or disables LED outputs of the data terminal. Note that the battery
 * charger circuitry also drives LED outputs, which may interfere with software.
 *
 * \param red Enables red LED.
 * \param green Enables green LED.
 */
void PowerManager::ledEnable(bool red, bool green)
{
	ioctl(powerDriver, red ? PWRCTRL_SET_GPIO_LO : PWRCTRL_SET_GPIO_HI,
			O_PIN_LED_R);
	ioctl(powerDriver, green ? PWRCTRL_SET_GPIO_LO : PWRCTRL_SET_GPIO_HI,
			O_PIN_LED_G);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set keypad backlight brightness
 *
 * Sets the keypad backlight brightness to absolute level instantly.
 *
 * \param level Specifies the keypad backlight brightness (0..255).
 */
void PowerManager::keybSetBackLight(unsigned char level)
{
	debug("keybSetBackLight: %d\n", level);
	ioctl(powerDriver, PWRCTRL_SET_KBDBL,  (unsigned char)level & 0xff);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Fade keypad backlight brightness
 *
 * Fades the keypad backlight brightness to the desired level fluently.
 *
 * \param level Specifies the destination keypad backlight brightness (0..255).
 */
void PowerManager::keybFadeBackLight(unsigned char level)
{
	debug("keybFadeBackLight: %d\n", level);
	ioctl(powerDriver, PWRCTRL_FADE_KBDBL_TO,  (unsigned char)level & 0xff);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set display brightness
 *
 * Sets the display brightness to the absolute level instantly.
 *
 * \param level Specifies the display brightness (0..255).
 * \return Returns true if the call was successful.
 */
bool PowerManager::oledSetBrightness(unsigned char level)
{
	displayMutex.lock();

	int oledDriver = open(OLED_DEVICE_NAME, O_RDWR | O_NONBLOCK);
	if (oledDriver < 0)
	{
		debug("Failed opening OLED device\n");
		displayMutex.unlock();
		return false;
	}

	int ret = ioctl(oledDriver, OLED_SET_BRT, (unsigned char)level & 0xff);
	debug("oledSetBrightness: %d %d\n", level, ret);

	close(oledDriver);
	displayMutex.unlock();
	return ret == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Fade display brightness
 *
 * Fades the display brightness to the specified level fluently.
 *
 * \param level Specifies the destination display brightness (0..255).
 * \return Returns true if the call was successful.
 */
bool PowerManager::oledFadeBrightness(unsigned char level)
{
	displayMutex.lock();

	int oledDriver = open(OLED_DEVICE_NAME, O_RDWR | O_NONBLOCK);
	if (oledDriver < 0)
	{
		debug("Failed opening OLED device\n");
		displayMutex.unlock();
		return false;
	}

	int ret = ioctl(oledDriver, OLED_FADE_BRT_TO, (unsigned char)level & 0xff);
	debug("oledFadeBrightness: %d %d\n", level, ret);

	close(oledDriver);
	displayMutex.unlock();
	return ret == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable display
 *
 * Enables or disables the display of the data terminal.
 * When the display is disabled, it is placed to a low-power mode.
 *
 * \param flag Specifies if the display should be enabled.
 * \return Returns true if operation was successful.
 */
bool PowerManager::oledEnable(bool flag)
{
	displayMutex.lock();

	// check if current state is the same
	if (flag == displayEnabled)
	{
		displayMutex.unlock();
		return true;
	}

	int oledDriver = open(OLED_DEVICE_NAME, O_RDWR | O_NONBLOCK);
	if (oledDriver < 0)
	{
		debug("Failed opening OLED device\n");
		displayMutex.unlock();
		return false;
	}

	int ret = ioctl(oledDriver, flag ? OLED_RESUME : OLED_SUSPEND);
	debug("oledEnable: %d - %d\n", flag, ret);
	close(oledDriver);

	displayEnabled = flag;
	displayMutex.unlock();
	return ret == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Freeze display
 *
 * Freezes the content of the display. May be used to avoid flickering,
 * for example.
 *
 * \param flag Set to true to freeze the display. Set to false to unfreeze afterwards.
 * \return Returns true if operation was successful.
 */
bool PowerManager::oledFreeze(bool flag)
{
	displayMutex.lock();

	// check if current state is the same
	if (flag == displayEnabled)
	{
		displayMutex.unlock();
		return true;
	}

	int oledDriver = open(OLED_DEVICE_NAME, O_RDWR | O_NONBLOCK);
	if (oledDriver < 0)
	{
		debug("Failed opening OLED device\n");
		displayMutex.unlock();
		return false;
	}

	int ret = ioctl(oledDriver, flag ? OLED_FREEZE : OLED_UNFREEZE);
	debug("oledFreeze: %d - %d\n", flag, ret);
	close(oledDriver);

	displayEnabled = flag;
	displayMutex.unlock();
	return ret == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set maximum display brightness
 *
 * Sets maximum display brightness. Does not have any effect immediately, as
 * the value is used for calculation of display brightnesses for
 * FULLY_OPERATIONAL and DISPLAY_DIMMED power modes.
 *
 * \param maxLevel Specifies maximum display brightness (0..255).
 */
void PowerManager::oledSetMaxBrightness(int maxLevel)
{
	maxDisplayBrightness = maxLevel;
}

/**
 * \brief Get USB VBus pin state
 *
 * \return Returns true if USB VBus pin is powered.
 */
bool PowerManager::isUsbConnected()
{
	int fd = open(VBUS_DEVICE_NAME, O_RDONLY);
	if (fd < 0)
	{
		debug("Failed opening VBus device\n");
		return false;
	}

	char buf[16] = "";
	ssize_t readBytes = read(fd, buf, sizeof(buf));
	close(fd);

	if (readBytes < 0) return false;
	if (buf[0] == '1') return true;

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Sound buzzer
 *
 * Generates a buzzer tone.
 *
 * \param freq Specifies buzzer tone frequency, in Hertz.
 * \param duration Specifies tone duration, in milliseconds.
 */
void PowerManager::buzzerPlay(int freq, int duration)
{
	input_event e;
	e.type = EV_SND;
	e.code = SND_TONE;
	e.value = freq;

	// start buzzing
	write(buzzerDriver, &e, sizeof(e));
	Thread::sleep(duration);

	// stop buzzing by setting frequency to zero
	e.value = 0;
	write(buzzerDriver, &e, sizeof(e));
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable radio interface of the communication module
 *
 * Enables or disables the radio interface of the data terminal communication
 * module. Currently implemented on WLAN-enabled devices.
 *
 * \param enable Specifies if the radio interface is to be enabled.
 */
void PowerManager::mpciEnableRadio(bool enable)
{
	ioctl(powerDriver,
			enable ? PWRCTRL_SET_GPIO_HI : PWRCTRL_SET_GPIO_LO, O_PIN_WLAN_RADIO_ON);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable communication module 3.3V power supply
 *
 * Enables or disables the 3.3V power supply to the data terminal communication
 * module. Currently implemented on WLAN-enabled devices.
 *
 * \param enable Specifies if the power supply is to be enabled.
 */
void PowerManager::mpciEnable3V3(bool enable)
{
	ioctl(powerDriver,
			enable ? PWRCTRL_SET_GPIO_HI : PWRCTRL_SET_GPIO_LO, O_PIN_MPCIE3VEN);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Power on GPRS module
 *
 * Powers up the GPRS module on GSM-enabled data terminals. The GPRS module
 * resulting state is sensed and reported back.
 *
 * \return Returns true if the GPRS module has reported a successful start.
 */
bool PowerManager::gprsPowerOn()
{
	const int pulseDuration = 1000;
	const int maxRetries = 3;

	// enable power to the module
	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_PWR);
	Thread::sleep(10);

	// enable serial port
	int ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_OFF, 1);
	if (ret != 0) return false;

	// check the initial power state
	ret = ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD);
	debug("initial state %d\n", ret);

	if (ret == 1) return true;	// module already on, return immediately

	for (int i = 0; i < maxRetries; i ++)
	{
		// generate a power-on pulse
		ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_ON);
		Thread::sleep(pulseDuration);
		ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_ON);

		// control the power state
		ret = ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD);
		debug("state %d\n", ret);

		if (ret == 1) return true;
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Power off GPRS module
 *
 * Powers down the GPRS module on GSM-enabled data terminals. The GPRS module
 * resulting state is sensed and reported back. Depending on network state,
 * the call may take variable amount of time. If the module does not report
 * a valid shutdown in 10 seconds, the module power supply is disconnected.
 *
 * \return Returns true.
 */
bool PowerManager::gprsPowerOff()
{
	const int pulseDuration = 2000;

	// disable serial port
	int ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_ON, 1);
	if (ret != 0) return false;

	// check the initial power state
	ret = ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD);
	debug("initial state %d\n", ret);

	if (ret == 1)	// check initial power state
	{
		ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_ON);
		Thread::sleep(pulseDuration);
		ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_ON);

		for (int i = 0; i < 100; i ++)
		{
			if (ioctl(powerDriver, PWRCTRL_GET_GPIO,
					I_PIN_GPRS_PWRGD) == 0) return true;
			Thread::sleep(100);
		}
	}

	ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_PWR);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reset GPRS module
 *
 * Generates a reset pulse to restart the GPRS module on GSM-enable data
 * terminals.
 */
void PowerManager::gprsReset()
{
	const int pulseDuration = 100;

	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_GPRS_RST);
	Thread::sleep(pulseDuration);
	ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_GPRS_RST);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set DTR line of GPRS module
 *
 * Enables or disables DTR signal on the GPRS module serial port interface.
 * DTR signal is used to control operation of the GPRS module, for example
 * its power saving mode.
 *
 * \param active Specifies the DTR line state.
 */
void PowerManager::gprsSetDTR(bool active)
{
	debug("active %d\n", active);
	ioctl(powerDriver, active ? PWRCTRL_SET_GPIO_LO : PWRCTRL_SET_GPIO_HI,
			O_PIN_GPRS_DTR);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get GPRS module state
 *
 * Reads current power state of the GPRS module on GSM-enabled data terminals.
 *
 * \return Returns true if the GPRS module is powered up.
 */
bool PowerManager::gprsGetPower()
{
	return ioctl(powerDriver, PWRCTRL_GET_GPIO, I_PIN_GPRS_PWRGD) == 1;
}


bool PowerManager::btPowerOn() {
	const int pulseDuration = 100;

	// enable power to the module
	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_BT_PWR);
	Thread::sleep(10);

	// enable serial port
	int ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_OFF, 1);
	if (ret != 0) return false;

	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_BT_RST);
	Thread::sleep(pulseDuration);
	ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_BT_RST);

	return true;
}


bool PowerManager::btPowerOff() {
	const int pulseDuration = 100;

	// disable serial port
	int ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_ON, 1);
	if (ret != 0) return false;

	ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_BT_RST);
	Thread::sleep(pulseDuration);

	ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_BT_PWR);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Power up barcode scanner
 *
 * Powers up a barcode scanner device on data terminals that have 1D or 2D
 * barcode scanner integrated into the sensor module.
 *
 * \return Returns true if the scanner was powered up successfully.
 */
bool PowerManager::scannerPowerUp()
{
	int ret;
	ret = ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_1D_AIM);
	if (ret != 0) return false;

	ret = ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_1D_PWR_N);
	if (ret != 0) return false;

	ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_OFF, 2);
	if (ret != 0) return false;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Power down barcode scanner
 *
 * Powers down a barcode scanner device on data terminals that have 1D or 2D
 * barcode scanner integrated into the sensor module.
 *
 * \return Returns true if the scanner was powered down successfully.
 */
bool PowerManager::scannerPowerDown()
{
	int ret;
	ret = ioctl(powerDriver, PWRCTRL_SET_GPIO_HI, O_PIN_1D_PWR_N);
	if (ret != 0) return false;

	ret = ioctl(powerDriver, PWRCTRL_SET_GPIO_LO, O_PIN_1D_AIM);
	if (ret != 0) return false;

	ret = ioctl(powerDriver, PWRCTRL_USART_TRISTATE_ON, 2);
	if (ret != 0) return false;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set barcode scanner aiming
 *
 * Enables or disables aiming of a barcode scanner device on data terminals
 * that have 1D or 2D barcode scanner integrated into the sensor module.
 *
 * \param flag Specifies if barcode scanner aiming is to be enabled.
 * \return Returns true if operation was a success.
 */
bool PowerManager::scannerSetAim(bool flag)
{
	int ret;
	ret = ioctl(powerDriver, flag ?
			PWRCTRL_SET_GPIO_LO : PWRCTRL_SET_GPIO_HI, O_PIN_1D_AIM);
	if (ret != 0) return false;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set device power mode
 *
 * Sets current power state of the data terminal. Executes all operations
 * required to enter the specified power state. All registered event listeners
 * are called and queried for their readiness to proceed to the new power state.
 *
 * The following states are supported: FULLY_OPERATIONAL, DISPLAY_DIMMED,
 * DISPLAY_MINIMAL, DISPLAY_DISABLED, LOW_POWER_SLEEP, POWER_DOWN.
 *
 * When LOW_POWER_SLEEP mode is specified, the system proceeds to a power saving
 * mode where it runs on a slow clock and awaits for a GPIO or RTC alarm
 * interrupt. In case of a GPIO interrupt (button press), the system proceeds
 * to FULLY_OPERATIONAL state. In case of an RTC alarm interrupt, the system
 * proceeds to a limited LOW_POWER_AWAKE state where it awaits further action.
 *
 * \param newState Specifies the new power state.
 * \return Returns true if the power mode was entered successfully.
 */
bool PowerManager::setPowerMode(PowerState newState)
{
	if (!enableStateSwitch || newState == currentState) return false;
	debug("%d\n", newState);

	// when going to a lower power saving state, call listeners in reverse
	// when going to a higher power saving state, call listeners in forward order
	if (beginEnumListeners(newState < currentState))
	{
		// notify active event listeners about power state change
		while (PowerEventListener *pl = getNextListener())
		{
			debug("Calling listener %p\n", pl);
			if(!pl->powerStateChange(currentState, newState))
			{
				// listener does not allow - abort state change
				endEnumListeners();
				return false;
			}
		}
		endEnumListeners();
	}

	currentState = newState;

	switch (newState)
	{
	case FULLY_OPERATIONAL:

		// TODO: set all settings, supporting arbitrary state transition sequence
		oledEnable(true);
		oledFadeBrightness(maxDisplayBrightness);
		break;

	case DISPLAY_DIMMED:

		oledFadeBrightness(maxDisplayBrightness / 2);
		break;

	case DISPLAY_MINIMAL:

		oledFadeBrightness(0);
		break;

	case DISPLAY_DISABLED:

		ledEnable(false, false);
		keybFadeBackLight(0);
		oledFadeBrightness(0);
		oledEnable(false);
		break;

	case LOW_POWER_AWAKE:
		break;

	case LOW_POWER_SLEEP:
	{
		keybSetBackLight(0);
		oledEnable(false);

		Media::mountDisable();
		if (!Media::unmountAll())
		{
			debug("Unmount failed\n");
			Media::mountEnable();
			setPowerMode(DISPLAY_DISABLED);
			break;
		}

		debug("Entering powersave...\n");

		FILE *f = fopen("/sys/power/state", "w");
		if (f == NULL)
		{
			debug("Kernel power state interface failure\n");
			Media::mountEnable();
			setPowerMode(DISPLAY_DISABLED);
			break;
		}

		// get duration to compensate CLOCK_MONOTONIC stall
		time_t sleepStart = time(NULL);

		// set alarm time
		if (powerSaveSleep > 0)
			rtcSetAlarm(powerSaveSleep);

		fputs("mem\n", f);
		fclose(f);

		Media::mountEnable();
		if(rtcGetAlarmState())
		{
			// compensate idle time
			idleBegin -= (time(NULL) - sleepStart);
			debug("Waking up on RTC alarm\n");
			setPowerMode(LOW_POWER_AWAKE);
		}
		else
		{
			debug("Waking up on GPIO\n");
			idleBegin = s_gettimeofday()/1000;
			setPowerMode(FULLY_OPERATIONAL);
		}

		break;
	}
	case POWER_DOWN:

		ledEnable(false, false);
		keybSetBackLight(0);

		gprsPowerOff();

		Media::mountDisable();
		if (!Media::unmountAll())
		{
			debug("Unmount failed\n");
			Media::mountEnable();
			setPowerMode(DISPLAY_DISABLED);
			break;
		}

		debug("Performing system shutdown\n");

		oledSetBrightness(0);
		oledEnable(false);


#ifdef USE_XYNTH
			GuiLibrary::lock();
			GuiLibrary::broadcastSystemEvent(EV_SHUTDOWN, NULL);
			GuiLibrary::unlock();
#endif
		break;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read temperature
 *
 * Reads internal temperature sensor.
 *
 * \return Returns current temperature in degrees Centigrade multiplied by 10.
 * In case of an error returns INT_MIN.
 */
int PowerManager::getTempReading()
{
	int tempDriver = open(TEMP_DEVICE_NAME, O_RDONLY);
	if (tempDriver < 0)
	{
		debug("Failed opening temperature sensor\n");
		return INT_MIN;
	}

	char buf[16] = "";
	ssize_t readBytes = read(tempDriver, buf, sizeof(buf));
	close(tempDriver);

	if (readBytes < 0) return INT_MIN;

	int rawTemp = atoi(buf);
	debug("raw %d\n", rawTemp);
	return rawTemp / 100;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get RTC alarm status
 *
 * Returns current RTC alarm status, as reported by kernel.
 *
 * \return Returns true if the RTC alarm has been triggered.
 */
bool PowerManager::rtcGetAlarmState()
{
	long status = 0;
	int ret = read(rtcDriver, &status, sizeof(status));
	debug("RTC status: %d %lx\n", ret, status);
	return ret > 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Enable RTC alarm
 *
 * Enables or disables RTC alarm.
 *
 * \param flag Specifies if the RTC alarm is to be enabled.
 * \return Returns true if operation was a success.
 */
bool PowerManager::rtcEnableAlarm(bool flag)
{
	if(ioctl(rtcDriver, flag ? RTC_AIE_ON : RTC_AIE_OFF) < 0)
	{
		debug("RTC ioctl RTC_AIE_xxx failed\n");
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set RTC alarm
 *
 * Initializes RTC alarm to trigger in a specified time interval. The alarm
 * may be used to bring the device out of a power saving mode.
 *
 * \param delay Specifies the time interval after which the alarm is triggered,
 * in seconds.
 * \return Retruns true if operation was a success.
 */
bool PowerManager::rtcSetAlarm(int delay)
{
	rtc_wkalrm alm;
	int ret;

	ret = ioctl(rtcDriver, RTC_RD_TIME, &alm.time);
	if (ret < 0)
	{
		debug("RTC_GET_TIME failed: %d, %d\n", ret, errno);
		return false;
	}

	debug("Current RTC: %02d/%02d/%04d %02d:%02d:%02d\n",
			alm.time.tm_mday, alm.time.tm_mon, alm.time.tm_year,
			alm.time.tm_hour, alm.time.tm_min, alm.time.tm_sec);

	tm timeStruct;
	timeStruct.tm_sec = alm.time.tm_sec;
	timeStruct.tm_min = alm.time.tm_min;
	timeStruct.tm_hour = alm.time.tm_hour;
	timeStruct.tm_mday = alm.time.tm_mday;
	timeStruct.tm_mon = alm.time.tm_mon;
	timeStruct.tm_year = alm.time.tm_year;
	timeStruct.tm_wday = alm.time.tm_wday;
	timeStruct.tm_yday = alm.time.tm_yday;
	timeStruct.tm_isdst = alm.time.tm_isdst;

	time_t timeSec = mktime(&timeStruct);
	timeSec += delay;
	if (localtime_r(&timeSec, &timeStruct) == NULL)
	{
		debug("localtime failed\n");
		return false;
	}

	alm.time.tm_sec = timeStruct.tm_sec;
	alm.time.tm_min = timeStruct.tm_min;
	alm.time.tm_hour = timeStruct.tm_hour;
	alm.time.tm_mday = timeStruct.tm_mday;
	alm.time.tm_mon = timeStruct.tm_mon;
	alm.time.tm_year = timeStruct.tm_year;
	alm.time.tm_wday = timeStruct.tm_wday;
	alm.time.tm_yday = timeStruct.tm_yday;
	alm.time.tm_isdst = timeStruct.tm_isdst;

	debug("Set alarm to: %02d/%02d/%04d %02d:%02d:%02d\n",
			alm.time.tm_mday, alm.time.tm_mon, alm.time.tm_year,
			alm.time.tm_hour, alm.time.tm_min, alm.time.tm_sec);

	alm.enabled = 1;
	alm.pending = 0;

	ret = ioctl(rtcDriver, RTC_WKALM_SET, &alm);
	if (ret < 0)
	{
		debug("RTC_SET_ALM failed: %d\n", ret);
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set RTC time
 *
 * Initializes the integrated RTC implemented by means of the CPU Real-Time
 * Timer. The RTC must be re-initialized every time the data terminal device
 * loses power supply for a long period.
 *
 * \param now Contains the current timestamp in Unix format.
 * \return Returns true if the RTC was set successfully
 */
bool PowerManager::rtcSetTime(time_t now)
{
	struct tm time;
	rtc_time rtc;

	if (localtime_r(&now, &time) == NULL)
	{
		debug("localtime failed\n");
		return false;
	}

	rtc.tm_sec = time.tm_sec;
	rtc.tm_min = time.tm_min;
	rtc.tm_hour = time.tm_hour;
	rtc.tm_mday = time.tm_mday;
	rtc.tm_mon = time.tm_mon;
	rtc.tm_year = time.tm_year;
	rtc.tm_wday = time.tm_wday;
	rtc.tm_yday = time.tm_yday;
	rtc.tm_isdst = time.tm_isdst;

	debug("Setting RTC to: %02d/%02d/%04d %02d:%02d:%02d\n",
			rtc.tm_mday, rtc.tm_mon, rtc.tm_year, rtc.tm_hour, rtc.tm_min, rtc.tm_sec);

	int ret = ioctl(rtcDriver, RTC_SET_TIME, &rtc);
	if (ret < 0)
	{
		debug("RTC_SET_TIME failed: %d, %d\n", ret, errno);
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get RTC time
 *
 * Reads current RTC time. The RTC must be running in order this call to succeed.
 *
 * \param pNow Points to the destination timestamp returned in Unix format.
 * \return Returns true if the RTC was read successfully.
 */
bool PowerManager::rtcGetTime(time_t *pNow)
{
	rtc_time rtc;
	int ret = ioctl(rtcDriver, RTC_RD_TIME, &rtc);
	if (ret < 0)
	{
		debug("RTC_RD_TIME failed: %d, %d\n", ret, errno);
		return false;
	}

	tm tmNow;

	tmNow.tm_sec = rtc.tm_sec;
	tmNow.tm_min = rtc.tm_min;
	tmNow.tm_hour = rtc.tm_hour;
	tmNow.tm_mday = rtc.tm_mday;
	tmNow.tm_mon = rtc.tm_mon;
	tmNow.tm_year = rtc.tm_year;
	tmNow.tm_wday = rtc.tm_wday;
	tmNow.tm_yday = rtc.tm_yday;
	tmNow.tm_isdst = rtc.tm_isdst;

	debug("Read RTC: %02d/%02d/%04d %02d:%02d:%02d\n",
			rtc.tm_mday, rtc.tm_mon, rtc.tm_year, rtc.tm_hour, rtc.tm_min, rtc.tm_sec);
	if (pNow != NULL) *pNow = mktime(&tmNow);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get current power state
 *
 * \return Returns current power manager state.
 */
PowerState PowerManager::getCurrentState()
{
	powerMutex.lock();
	PowerState state = currentState;
	powerMutex.unlock();

	return state;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get idle period duration
 *
 * Calculates and returns the duration of idle state. The idle duration is
 * calculated based on the last kickIdle() call instant.
 *
 * \return Returns idle period, in seconds.
 */
int PowerManager::getIdleDuration()
{
	powerMutex.lock();
	time_t now = s_gettimeofday()/1000;
	int ret = now - idleBegin;
	powerMutex.unlock();

	return ret > 0 ? ret : 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get ambient light intensity
 *
 * Reads integrated ambient light sensor data, combining both IR and visible
 * light spectra.
 *
 * \return Returns ambient light sensor intensity.
 */
int PowerManager::getAmbientLight()
{
	int lightDriver = open(LIGHT_DEVICE_NAME, O_RDONLY);
	if (lightDriver < 0)
	{
		debug("Failed opening I2C ambient light sensor\n");
		return -1;
	}

	char buf[16] = "";
	ssize_t readBytes = read(lightDriver, buf, sizeof(buf));
	close(lightDriver);

	if (readBytes < 0) return -1;

	int alReading = atoi(buf);
	debug("PowerManager::getAmbientLight: %d\n", alReading);
	return alReading;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set power saving sleep duration
 *
 * Sets the time period that the data terminal spends in power saving mode,
 * when entered. After the given period is elapsed, the device proceeds to
 * a limited LOW_POWER_AWAKE state and awaits further action.
 *
 * \param sec Specifies the power saving sleep duration, in seconds.
 */
void PowerManager::setPowerSaveSleep(int sec)
{
	powerSaveSleep = sec;
}

#ifdef USE_XYNTH
/**
 * \brief Deferring idle mode
 *
 * Instructs power manager that the device is not idle at the moment. Power
 * manager saves current timestamp that is further used for calculation of the
 * idle duration. Calls the kickIdle() method.
 *
 * \return Returns false if system was already idle, true otherwise
 */
int PowerManager::kickIdleProc() {
	//eat the keypress if it triggered coming out of sleep
	PowerManager *pm = PowerManager::getPowerManager();
	if (pm == NULL) return 0;
	if (pm->kickIdle()) return 1;

	return 0;
}
#endif
