/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <sys/stat.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <platform.h>

#include <exynthoo/GuiLibrary.h>

//#define DEBUG
#include "LibSpatha.h"
#include "Debug.h"
#include "PowerManager.h"
#include "RfidReader.h"

#define RFID_BUF_SIZE	2048

#define READER_PORTS {"/dev/rfid_nxp", "/dev/rfid_ams"}

#define DEFAULT_SCANNER_TIMEOUT		5000

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the RFID reader object.
 * \param t Reader type being used
 */
RfidReader::RfidReader(RfidReaderType t)
{
	devHandle = -1;
	readerActive = false;
	type = t;
	interval = DEFAULT_SCANNER_TIMEOUT;
}

/**
 * \brief Class destructor
 *
 * Terminates the RFID reader object.
 */
RfidReader::~RfidReader()
{
	if(devHandle != -1) shutdown();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reader initialization routine
 *
 * Initializes RFID reader interface, opens the driver for further communication.
 *
 * \return Returns true if initialization was successful.
 */
bool RfidReader::initialize()
{
	const char *ports[] = READER_PORTS;

	if (devHandle != -1) return false;
	devHandle = open(ports[type], O_RDONLY);

	if (devHandle == -1)
	{
		debug("Failed opening RFID reader device: %d\n", devHandle);
		return false;
	}

	close(devHandle);
	devHandle = -1;

	debug("RFID reader initialized\n");
	return Thread::start();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reader shutdown routine
 *
 * Shuts down the RFID reader interface, closes the driver.
 */
void RfidReader::shutdown()
{
	stopRead();
	Thread::requestStop();
	readerSem.post();
	Thread::stop();

	debug("RFID reader terminated\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Prepare for reading
 *
 * Triggers the RFID reader and initiates reading a tag.
 */
void RfidReader::initRead()
{
	readerActive = true;
	readerSem.post();
	debug("RFID reader scanning\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start reading
 *
 * Dummy function, currently does nothing.
 */
void RfidReader::startRead()
{
	debug("Dummy RFID reader start procedure\n");
	startTimer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop reading
 *
 * Instructs the RFID reader to stop reading tags.
 */
void RfidReader::stopRead()
{
	readerMutex.lock();
	if (devHandle != -1) {
		ioctl(devHandle, RFID_NXP_STOP_READ, 0);
	}
	readerMutex.unlock();
	debug("RFID reader stopped\n");

	stopTimer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Timer event handler
 *
 *	Stops reading after a timeout
 */
void RfidReader::onTimer()
{
	stopRead();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set read area
 *
 * Selects the blocks that should be read out from an RFID tag when a new read
 * procedure is initiated.
 *
 * \param b Specifies the memory region to read from RFID tag (see RfidReader.h
 * for details!)
 * \param first Specifies the first block number read from the RFID tag
 * \param last Specifies the last block number read from the RFID tag
 */
void RfidReader::setReadAreas(RfidBank b, int first, int last)
{
	bank = b;
	firstBlock = first;
	lastBlock = last;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Runs RFID reading thread
 * Executes the RFID reading process as a separate thread.
 */
void RfidReader::run()
{
	debug("RfidReader::run(): starting\n");
	PowerManager *pPM = PowerManager::getPowerManager();
	const char *ports[] = READER_PORTS;
	char buf[RFID_BUF_SIZE];

	while (!Thread::stopRequested())
	{
		readerSem.pend();
		if (readerActive && !Thread::stopRequested())
		{
			readerMutex.lock();
			devHandle = open(ports[type], O_RDONLY);
			readerMutex.unlock();
			if (devHandle == -1) break;

			if (type == RFID_UHF) {
				switch(bank) {
				case BANK_ID:
					ioctl(devHandle, RFID_AMS_SET_BLOCK_START, 0);
					ioctl(devHandle, RFID_AMS_SET_BLOCK_END, 0);
					ioctl(devHandle, RFID_AMS_SET_EPC_BANK, 0);
					break;
				case BANK_EPC_RES:
					ioctl(devHandle, RFID_AMS_SET_BLOCK_START, firstBlock);
					ioctl(devHandle, RFID_AMS_SET_BLOCK_END, lastBlock);
					ioctl(devHandle, RFID_AMS_SET_EPC_BANK, 0);
					break;
				case BANK_EPC_EPC:
					ioctl(devHandle, RFID_AMS_SET_BLOCK_START, firstBlock);
					ioctl(devHandle, RFID_AMS_SET_BLOCK_END, lastBlock);
					ioctl(devHandle, RFID_AMS_SET_EPC_BANK, 1);
					break;
				case BANK_EPC_TID:
					ioctl(devHandle, RFID_AMS_SET_BLOCK_START, firstBlock);
					ioctl(devHandle, RFID_AMS_SET_BLOCK_END, lastBlock);
					ioctl(devHandle, RFID_AMS_SET_EPC_BANK, 2);
					break;
				case BANK_DATA:
					ioctl(devHandle, RFID_AMS_SET_BLOCK_START, firstBlock);
					ioctl(devHandle, RFID_AMS_SET_BLOCK_END, lastBlock);
					ioctl(devHandle, RFID_AMS_SET_EPC_BANK, 3);
					break;
				}
			} else {
				ioctl(devHandle, RFID_NXP_SET_ALL_BLOCKS_OFF);
				if (bank == BANK_DATA) {
					for (int i = firstBlock; i <= lastBlock; i++)
						ioctl(devHandle, RFID_NXP_SET_BLOCK_ON, i);
				}
			}

			int len = read(devHandle, buf, RFID_BUF_SIZE - 1);
			if (Thread::stopRequested()) break;

			if (readerActive &&
				(len > 4) &&
				(len < RFID_BUF_SIZE) &&
				(buf[0] == 0))
			{
				int idLen = buf[1];
				int dataLen = buf[2] | buf[3]<<8;

				if (len == idLen + dataLen + 4)
				{
					// interpret as zero-terminated string
					buf[len] = 0;
					dataLen = strlen((char *)&buf[idLen + 4]);

					if ((idLen > 0) || (dataLen > 0))
					{
						char idBuf[2 * idLen + 1];
						idBuf[0] = 0;

						for (int i = 0; i < idLen; i++)
							sprintf((char *)&idBuf[i * 2], "%02X", buf[4 + i]);

						submit(idBuf, (char *)&buf[idLen + 4]);
					}
				}
			}

			readerMutex.lock();
			close(devHandle);
			devHandle = -1;
			readerMutex.unlock();

			readerActive = false;
		}
	}

	readerMutex.lock();
	if (devHandle != -1) {
		close(devHandle);
		devHandle = -1;
	}
	readerMutex.unlock();


	debug("RfidReader::run(): finished\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Submit the reading data to into event system
 * Broadcasts the data using an EV_READER system event.
 *
 * \param id Pointer to tag ID that was read
 * \param data Pointer to the user data stored in tag
 */
void RfidReader::submit(const char *id, const char *data) {
	ReaderData rdata;
	rdata.reader = (type == RFID_UHF)?READER_UHF_RFID:READER_RFID;
	rdata.pId = id;
	rdata.pData = data;
	rdata.handled = 0;

	while (1) {
		if (GuiLibrary::timedlock(1000)) break;
		if (stopRequested()) return;
	}
	stopTimer();
	GuiLibrary::broadcastSystemEvent(EV_READER, (void*)&rdata);
	GuiLibrary::unlock();
}
