#ifndef RFIDREADER_H_
#define RFIDREADER_H_

#include "exynthoo/GuiTimer.h"
#include "Thread.h"
#include "BasicReader.h"

enum RfidReaderType {
	RFID_HF = 0,
	RFID_UHF,
};

enum RfidBank
{
	BANK_ID = 0, 	// HF: tag id; 							UHF: ACK reply (PC + XPC + EPC)
	BANK_EPC_RES,	// HF: tag id; 							UHF: EPC RES bank[startblock - endblock]
	BANK_EPC_EPC,	// HF: tag id; 							UHF: EPC EPC bank[startblock - endblock]
	BANK_EPC_TID,	// HF: tag id; 							UHF: EPC TID bank[startblock - endblock]
	BANK_DATA		// HF: tag data[startblock - endblock]; UHF: EPC DATA bank[startblock - endblock]
};
/* UHF EPC GEN 2:
 * Decoding ACK reply
 * --------------------
 * See EPCGlobal TDS standard page 25 figure K (v1.4); XPC not included:
 * http://www.epcglobalinc.org/standards/tds/
 * See EPCGLobal UHFC1G2 standard:
 * http://www.epcglobalinc.org/standards/uhfc1g2
 *
 * ACK reply format is (PC [XPC1] [XPC2] EPC).
 * 16-bit PC is always present.
 * The bit 0x02 in PC[1] notifies the presence of 16-bit XPC1.
 * If XPC1 is present, the bit 0x01 in XPC[1] notifies the presence of 16-bit XPC2.
 * EPC code follows the PC and the optional XPC1, XPC2 fields.
 *
 * EPC code length is written to PC field; the driver verifies it and filters out invalid reads.
 * The data returned from driver is guaranteed to have correct length.
 *
 * If needed, the used has to strip the PC,XPC1,XPC2 fields himself.
 */

/*
 * UHF EPC GEN 2:
 * Reading data
 * --------------------
 * If the bank is not specified (start<=end), the reader reads only the EPC (ID) and fails when
 * EPC can not be read.
 * If the bank is set, then the reader reads both EPC and specified data region. If one of them
 * fails, the whole read fails and nothing is returned.
 * As most of the tags reply with error code when reading non-existing memory, the user must
 * set the block counts with care; if end count is beyond memory end, the read fails.
 */

/**
 * \brief RFID reader management class
 * Provides methods for managing the RFID  (both HF and UHF) readers.
 *
 * Implements appropriate methods from BasicReader class.
 *
 * Information about UHF EPC GEN2:
 * If the bank is not specified (start<=end), the reader reads only the EPC (ID) and fails when
 * EPC can not be read.
 * If the bank is set, then the reader reads both EPC and specified data region. If one of them
 * fails, the whole read fails and nothing is returned.
 * As most of the tags reply with error code when reading non-existing memory, the user must
 * set the block counts with care; if end count is beyond memory end, the read fails.
 *
 */
class RfidReader : public BasicReader, protected Thread, protected GuiTimer
{
public:
	/** RFID reader states */
	enum RfidState {
		/** Reader is idle */
		STATE_IDLE,
		/** The reading process is started */
		STATE_START_READ,
		/** Reading is in progress */
		STATE_READING,
		/** Reading is finished */
		STATE_STOP_READ,
	};

	RfidReader(RfidReaderType type);
	virtual ~RfidReader();

	virtual bool initialize();
	virtual void shutdown();

	virtual void initRead();
	virtual void startRead();
	virtual void stopRead();

	void setReadAreas(RfidBank bank, int firstBlock, int lastBlock);

protected:
	virtual void run();
	virtual void onTimer();

	void submit(const char *id, const char *data);

private:
	bool		readerActive;
	int			devHandle;
	Mutex		readerMutex;

	RfidBank	bank;
	int			firstBlock;
	int			lastBlock;

	Semaphore	readerSem;
	RfidReaderType type;
};

#endif /*RFIDREADER_H_*/
