/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <list>

//#define DEBUG
#include "Debug.h"
#include "ScannerInput.h"
#include "RfidReader.h"
#include "BarcodeReader.h"
#include "SysInfo.h"

ScannerInput* ScannerInput::scannerInput = NULL;

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 * Constructor of the class. The instance of this class should be created by
 * invoking the initialize() method instead.
 *
 * The class instance can be later retrieved by getScannerInput() method.
 */
ScannerInput::ScannerInput()
{
	for (int i = 0; i < MAX_READERS; i++) {
		pReaders[i] = NULL;
		scanInited[i] = false;
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 * Uninitializes the class. Deletes all the reader instances.
 */
ScannerInput::~ScannerInput()
{
	for(int i=0; i<MAX_READERS && pReaders[i] != NULL; i++)
	{
		// clean up, free memory
		delete pReaders[i];
	}

	debug("ScannerInput destroyed\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start scanner subsystem
 *
 * This function initializes scanners and starts the scanner threads.
 */
bool ScannerInput::initialize()
{
	if (scannerInput != NULL) return true;
	ScannerInput *pSc = new ScannerInput();
	if (!pSc->init()) {
		delete pSc;
		return false;
	}
	scannerInput = pSc;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop scanner subsystem
 *
 * This function shuts down the scanners
 */
void ScannerInput::shutdown()
{
	if (scannerInput == NULL) return;
	scannerInput->close();
	delete scannerInput;
	scannerInput = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get ScannerInput subsystem
 *
 * \return Returns pointer to ScannerInput
 */
ScannerInput* ScannerInput::getScannerInput()
{
	return scannerInput;
}

////////////////////////////////////////////////////////////////////////////////
//
void ScannerInput::detectReaders()
{
	SCANNER_INFORMATION *scanInfo = SysInfo::getScannerInformation();
	if (scanInfo == NULL) return;

	int barcodeReader = scanInfo->readerType[0];
	int rfidReader = scanInfo->readerType[1];

	// allocate readers
	if (barcodeReader == '1')
		pReaders[SCANNER_BARCODE] = new BarcodeReader();

	if (rfidReader == '1')
		pReaders[SCANNER_RFID] = new RfidReader(RFID_HF);
	else if (rfidReader == '2')
		pReaders[SCANNER_RFID] = new RfidReader(RFID_UHF);

}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Initialize scanner device
 * Performs per-device initialization. Calls the appropriate initialize() method
 * for each reader type and increments list of available readers if the call was
 * successful.
 *
 * \return Returns true if scanner devices were initialized successfully.
 */
bool ScannerInput::init()
{
	int numReaders = 0;

	// add supported readers
	detectReaders();

	for(int i=0; i<MAX_READERS; i++)
	{
		if (pReaders[i] == NULL) continue;

		debug("Initializing reader %d\n", i);
		if(!pReaders[i]->initialize()) return false;

		numReaders ++;
	}

	// fail if no readers were initialized
	if (numReaders == 0) return false;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Shutdown scanner devices
 * Performs per-device uninitialization by calling the appropriate shutdown() method
 * for each scanner type.
 */
void ScannerInput::close()
{
	for (int i = 0; i < MAX_READERS && pReaders[i] != NULL; i ++)
	{
		debug("Shutting down reader %d\n", i);
		pReaders[i]->shutdown();
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Prepare for scanning
 * Prepares the scanner by calling the appropriate initRead() method for that
 * scanner type.
 *
 * \param scanner Scanner device to be prepared
 *
 * \return Returns true if the device is prepared successfully.
 */
bool ScannerInput::initScan(int scanner)
{
	if ((scanner >= 0) && (scanner < MAX_READERS))
	{
		if (pReaders[scanner] != NULL) {
			debug("Calling initRead() of reader %d\n", scanner);
			pReaders[scanner]->initRead();
			scanInited[scanner] = true;
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start scanning
 * Starts the scanning process by calling the appropriate startRead() method for each
 * registered scanner device.
 *
 * \return Returns true if scanning was successful.
 */
bool ScannerInput::startScan()
{
	bool result = false;

	for (int i = 0; i < MAX_READERS; i++)
	{
		if ((pReaders[i] != NULL) && scanInited[i])
		{
			debug("Calling startRead() of reader %d\n", i);
			pReaders[i]->startRead();
			result = true;
		}
	}

	return result;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop scanning
 * Stops the scanning process by calling the appropriate stopRead() method for each
 * registered scanner device.
 *
 * \return Returns true if scanning result was successful.
 */
bool ScannerInput::stopScan()
{
	debug("stopScan\n");
	bool result = false;

	for (int i = 0; i < MAX_READERS; i ++)
	{
		if ((pReaders[i] != NULL) && scanInited[i])
		{
			debug("Calling stopRead() of reader %d\n", i);
			pReaders[i]->stopRead();
			scanInited[i] = false;

			result = true;
		}
	}

	return result;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set the RFID scan areas
 * Sets the RFID areas to be scanned for devices of RFID scanner type
 *
 * \param bank Specifies the memory region to read from RFID tag (see RfidReader.h
 * for details!)
 * \param first_block Specifies the first block number read from the RFID tag
 * \param last_block Specifies the last block number read from the RFID tag
 */

void ScannerInput::setRfidReadAreas(RfidBank bank, int first_block, int last_block)
{
	if ((pReaders[SCANNER_RFID] != NULL) &&	(!scanInited[SCANNER_RFID]))
	{
		((RfidReader *)pReaders[SCANNER_RFID])->setReadAreas(bank, first_block, last_block);
	}
}
