/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <sys/ioctl.h>
#include <sys/poll.h>

#include "SerialPort.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the serial class object.
 */

SerialPort::SerialPort()
{
	devFile = -1;
	hup = false;
	lastAttrValid = false;
}

/**
 * \brief Class destructor
 *
 * Destroys the serial class object.
 */
SerialPort::~SerialPort()
{
	if(devFile != -1) close();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Open serial port
 *
 * Opens the specified serial port and initializes it to default settings.
 * Port baud rate is not changed.
 *
 * \param portDev Specifies the serial port device name, e.g. /dev/ttyAT1.
 * \return Returns true if the serial port was opened successfully.
 */
bool SerialPort::open(const char* portDev)
{
	int cflags = O_RDWR | O_NDELAY;
	devFile = ::open(portDev, cflags);
	devName = portDev;
	if(devFile == -1) return false;

	hup = false;
	initConfig();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close serial port
 *
 * Closes the serial port, also flushes I/O before closing the descriptor.
 *
 * \return Returns true if the operation was a success.
 */
bool SerialPort::close()
{
	struct termios attr;
	// disable flow control
	tcgetattr(devFile, &attr);
	attr.c_cflag &= ~CRTSCTS;
	tcsetattr(devFile, TCSANOW, &attr);
	// flush the buffers
	tcflush(devFile, TCIOFLUSH);
	// finally, close the port
	int ret = ::close(devFile);
	devFile = -1;
	hup = true;
	return ret == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Reopen serial port
 *
 * Reopens the specified serial port and initializes it to last settings.
 * Port baud rate is not changed.
 *
 * \return Returns true if the serial port was reopened successfully.
 */
bool SerialPort::reOpen()
{
	hup = true;
	::close(devFile);
	devFile = -1;

	int cflags = O_RDWR | O_NDELAY;
	devFile = ::open(devName.c_str(), cflags);
	if(devFile == -1) return false;

	hup = false;
	initConfig();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Query port status
 *
 * Checks if port is open. Port may be closed by system using HUP
 *
 * \return Returns true if the serial port  is opened.
 */
bool SerialPort::isPortOpen()
{
	if (hup) return false;

	pollfd pfd;
	pfd.fd = devFile;
	pfd.events = POLLHUP;
	int status = 0;
	while (status < 1)
	{
		pfd.revents = 0;
		status = poll(&pfd, 1, 0);
		if (status < 1)
		{
			if (status == -1 && errno == EINTR) continue;
			else return false;
		}
	}

	if ((pfd.revents & POLLHUP)) {
		hup = true;
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read data from serial port
 *
 * Receives the specified number of bytes from the serial port.
 * The reading may stop prematurely if a timeout expires while receiving
 * the next character.
 *
 * \param pData Points to the data buffer.
 * \param maxLen Specifies the number of bytes to be read.
 *
 * \return Returns number of bytes actually received from the serial port.
 * Returns -1 on error.
 */
ssize_t SerialPort::readData(void *pData, size_t maxLen)
{
	return ::read(devFile, (void*)pData, maxLen);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write data to serial port
 *
 * Sends the specified number of bytes to the serial port.
 *
 * \param pData Points to the data buffer.
 * \param dataLen Specifies the number of bytes to write.
 * \return Returns number of bytes actually written to the serial port.
 */
ssize_t SerialPort::writeData(const void *pData, size_t dataLen)
{
	return ::write(devFile, (void*)pData, dataLen);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get number of pending bytes
 *
 * If no bytes are currently pending in the receive buffer, the function
 * will block until timeout expires, or the data arrives to serial port.
 * If there are bytes pending in the receive buffer already, the function
 * will return immediately.
 *
 * \param timeOut Specifies the timeout in milliseconds.
 * \return Returns number of bytes pending for read. Returns 0 if no bytes
 * are pending.
 */
size_t SerialPort::numInputPending(int timeOut)
{
	pollfd pfd;
	pfd.fd = devFile;
	pfd.revents = 0;
	if (!hup) pfd.events |= POLLHUP;

	int status = 0;
	while (status < 1)
	{
		pfd.events = POLLIN;
		status = poll(&pfd, 1, timeOut);
		if (status < 1)
		{
			if (status == -1 && errno == EINTR) continue;
			else return 0;
		}
		if ((pfd.revents & POLLHUP)) {
			hup = true;
			return false;
		}
	}

	if (pfd.revents & POLLIN)
	{
		int bytes;
    	if (ioctl(devFile, FIONREAD, &bytes) < 0) return 1;
    	else return bytes;
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Check if there is input data
 *
 * Expects incoming data from the serial port. The function blocks if there
 * is no data currently pending for receive, until the timeout expires or the
 * data arrives. The function returns immediately if there is data in the
 * receive buffer already.
 *
 * \param timeOut Specifies the timeout in milliseconds.
 * \return Returns true if there is data pending for reading.
 */
bool SerialPort::isInputPending(int timeOut)
{
	pollfd pfd;
	pfd.fd = devFile;
	pfd.events = POLLIN;
	if (!hup) pfd.events |= POLLHUP;

	int status = 0;
	while (status < 1)
	{
		pfd.revents = 0;
		status = poll(&pfd, 1, timeOut);
		if (status < 1)
		{
			if (status == -1 && errno == EINTR) continue;
			else return false;
		}
		if ((pfd.revents & POLLHUP)) {
			hup = true;
			return false;
		}
	}

	return status == 1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Configure serial port
 * Initiates serial port configuration.
 *
 * \note If there are already valid configuration attributes stored the attributes
 * are simply reapplied to the serial port.
 */
void SerialPort::initConfig(void)
{
	if (!lastAttrValid) {
		tcgetattr(devFile, &lastAttr);

		lastAttr.c_oflag = lastAttr.c_lflag = 0;
		lastAttr.c_cflag = CLOCAL | CREAD | HUPCL;
		lastAttr.c_iflag = IGNBRK;

		memset(lastAttr.c_cc, 0, sizeof(lastAttr.c_cc));
		lastAttr.c_cc[VMIN] = 1;

		lastAttr.c_cflag |= (CRTSCTS | CSIZE | PARENB | PARODD | CSTOPB);
		lastAttr.c_iflag |= (IXON | IXANY | IXOFF);

		tcsetattr(devFile, TCSANOW, &lastAttr);
		lastAttrValid = true;
	}

	tcsetattr(devFile, TCSANOW, &lastAttr);

	long ioflags = fcntl(devFile, F_GETFL);
	fcntl(devFile, F_SETFL, ioflags & ~O_NDELAY);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Flush serial port input
 *
 * Discards all serial port input currently pending for receive.
 */
void SerialPort::flushInput(void)
{
	tcflush(devFile, TCIFLUSH);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Flush serial port output
 *
 * Discards all serial port output currently pending for transmit.
 */
void SerialPort::flushOutput(void)
{
	tcflush(devFile, TCOFLUSH);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Drain the serial port
 *
 * Waits until all output written to the serial port has been transmitted.
 */
void SerialPort::drainOutput(void)
{
	tcdrain(devFile);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set baud rate
 *
 * Sets new baud rate of the serial port. Accepts the following rates:
 * 115200, 57600, 38400, 19200, 9600, 4800, 2400, 1200, 600, 300, 110
 *
 * \param speed Specifies the new baud rate.
 * \return Returns true if the operation was a success.
 */
bool SerialPort::setSpeed(unsigned long speed)
{
	unsigned long rate;

	switch(speed)
	{
	#ifdef B115200
	case 115200:
		rate = B115200;
		break;
	#endif
	#ifdef B57600
	case 57600:
		rate = B57600;
		break;
	#endif
	#ifdef B38400
	case 38400:
		rate = B38400;
		break;
	#endif
	case 19200:
		rate = B19200;
		break;
	case 9600:
		rate = B9600;
		break;
	case 4800:
		rate = B4800;
		break;
	case 2400:
		rate = B2400;
		break;
	case 1200:
		rate = B1200;
		break;
	case 600:
		rate = B600;
		break;
	case 300:
		rate = B300;
		break;
	case 110:
		rate = B110;
		break;
	default:
		return false;
	}

	tcgetattr(devFile, &lastAttr);
	cfsetispeed(&lastAttr, rate);
	cfsetospeed(&lastAttr, rate);
	tcsetattr(devFile, TCSANOW, &lastAttr);

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set flow control mode
 *
 * Sets the serial port flow control mode. Allowed modes are listed below.
 * - FLOW_NONE
 * - FLOW_SOFT
 * - FLOW_HARD
 * - FLOW_BOTH
 *
 * \param flow Specifies the new flow control mode.
 * \return Returns true if the operation was a success.
 */
bool SerialPort::setFlowControl(S_FLOW_CONTROL flow)
{
	tcgetattr(devFile, &lastAttr);

	lastAttr.c_cflag &= ~CRTSCTS;
	lastAttr.c_iflag &= ~(IXON | IXANY | IXOFF);

	switch(flow)
	{
	case S_FLOW_SOFT:
		lastAttr.c_iflag |= (IXON | IXANY | IXOFF);
		break;
	case S_FLOW_BOTH:
		lastAttr.c_iflag |= (IXON | IXANY | IXOFF);
	case S_FLOW_HARD:
		lastAttr.c_cflag |= CRTSCTS;
		break;
	case S_FLOW_NONE:
		break;
	default:
		return false;
	}

	tcsetattr(devFile, TCSANOW, &lastAttr);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set number of stop bits
 *
 * Sets number of stop bits for the serial port character.
 *
 * \param bits Specifies the number of stop bits, should be set to 1 or 2.
 * \return Returns true if the operation was a success.
 */
bool SerialPort::setStopBits(int bits)
{
	tcgetattr(devFile, &lastAttr);
	lastAttr.c_cflag &= ~CSTOPB;

	switch(bits) {
	case 1:
		break;
	case 2:
		lastAttr.c_cflag |= CSTOPB;
		break;
	default:
		return false;
	}

	tcsetattr(devFile, TCSANOW, &lastAttr);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set number of data bits
 *
 * Sets number of data bits for the serial port character. Valid options are
 * 5, 6, 7 or 8 data bits.
 *
 * \param bits Specifies the number of data bits.
 * \return Returns true if the operation was a success.
 */
bool SerialPort::setCharBits(int bits)
{
	tcgetattr(devFile, &lastAttr);
	lastAttr.c_cflag &= ~CSIZE;

	switch(bits)
	{
	case 5:
		lastAttr.c_cflag |= CS5;
		break;
	case 6:
		lastAttr.c_cflag |= CS6;
		break;
	case 7:
		lastAttr.c_cflag |= CS7;
		break;
	case 8:
		lastAttr.c_cflag |= CS8;
		break;
	default:
		return false;
	}

	tcsetattr(devFile, TCSANOW, &lastAttr);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set parity mode
 *
 * Sets the new parity control mode of the serial port. Valid modes are listed below.
 * - PARITY_NONE
 * - PARITY_EVEN
 * - PARITY_ODD
 *
 * \param parity Specifies the serial port parity check mode.
 * \return Returns true if the operation was a success.
 */
bool SerialPort::setParity(S_PARITY_CONTROL parity)
{
	tcgetattr(devFile, &lastAttr);
	lastAttr.c_cflag &= ~(PARENB | PARODD);

	switch(parity)
	{
	case S_PARITY_EVEN:
		lastAttr.c_cflag |= PARENB;
		break;
	case S_PARITY_ODD:
		lastAttr.c_cflag |= (PARENB | PARODD);
		break;
	case S_PARITY_NONE:
		break;
	default:
		return false;
	}
	tcsetattr(devFile, TCSANOW, &lastAttr);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Send break sequence
 *
 * Transmits a continuous stream of zero-valued bits for 0.25 seconds.
 */
void SerialPort::sendBreak(void)
{
	tcsendbreak(devFile, 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set DTR signal
 *
 * Set the serial port DTR signal to the specified state.
 *
 * \param enable Specifies the new DTR state
 */
void SerialPort::setDTR(bool enable)
{
	int status = 0;
	ioctl(devFile, TIOCMGET, &status);
	if(enable) status |= TIOCM_DTR;
	else status &= ~TIOCM_DTR;
	ioctl(devFile, TIOCMSET, &status);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set RTS signal
 *
 * Set the serial port RTS signal to the specified state.
 *
 * \param enable Specifies the new RTS state
 */
void SerialPort::setRTS(bool enable)
{
	int status = 0;
	ioctl(devFile, TIOCMGET, &status);
	if(enable) status |= TIOCM_RTS;
	else status &= ~TIOCM_RTS;
	ioctl(devFile, TIOCMSET, &status);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read CTS signal
 *
 * \return Returns current serial port CTS signal state.
 */
bool SerialPort::getCTS()
{
	int status = 0;
	ioctl(devFile, TIOCMGET, &status);
	return (status & TIOCM_CTS) != 0;
}
