/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include "StdioFile.h"
#include "Media.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor.
 *
 * Initializes the I/O file object.
 */
StdioFile::StdioFile()
{
	fileStream = NULL;
	mountPoint = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor.
 *
 * Destroys the I/O file object.
 */
StdioFile::~StdioFile()
{
	if(fileStream) close();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Open file
 *
 * Opens a stream file with the given file name. When required, the
 * corresponding media will be also mounted.
 *
 * \param fileName Contains the file name.
 * \param mode Specifies the file access mode.
 * \return Returns true if the file was opened successfully.
 */
bool StdioFile::open(const char* fileName, OpenMode mode)
{
	const char *openMode = NULL;
	switch(mode)	// determine stdio file mode
	{
	case ModeRead:
		openMode = "r";
		break;
	case ModeWrite:
		openMode = "w";
		break;
	case ModeReadWrite:
		openMode = "w+";
		break;
	case ModeAppend:
		openMode = "a";
		break;
	case ModeUpdate:
		openMode = "r+";
		break;
	}

	// try to mount the required media
	mountPoint = Media::mountMedia(fileName, mode != ModeRead);
	if(mountPoint == MOUNT_FAILED) return false;

	// finally, open the file
	fileStream = fopen(fileName, openMode);
	if (fileStream == NULL) Media::unmountMedia(mountPoint);
	return fileStream != NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close file
 *
 * Closes the stream file. If the media was mounted associated with this file,
 * it will be automatically unmounted, provided that no other files are open
 * on that media.
 *
 * \return Returns true if the file was successfully closed.
 */
bool StdioFile::close()
{
	if(!fileStream) return false;
	if(fclose(fileStream) == 0)
	{
		// try to unmount the media
		Media::unmountMedia(mountPoint);
		fileStream = NULL;
		return true;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get End-of-File flag
 *
 * \return Returns true if end of file has been reached.
 */
bool StdioFile::isEOF()
{
	return feof(fileStream) != 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write data to file
 *
 * Writes the given data buffer to file at the current position.
 *
 * \param pData Points to the data buffer.
 * \param dataLen Specifies the length of the data buffer.
 * \return Returns true if the buffer has been written successfully.
 */
bool StdioFile::writeData(const void* pData, size_t dataLen)
{
	return fwrite(pData, 1, dataLen, fileStream) == dataLen;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read data from file
 *
 * Reads the data from file to a specified buffer.
 *
 * \param pData Pointer to the data buffer.
 * \param maxLen Specifies the length of the data to be read.
 * \return Returns number of bytes actually read from the file.
 */
size_t StdioFile::readData(void* pData, size_t maxLen)
{
	return fread(pData, 1, maxLen, fileStream);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write string to file
 *
 * Writes the specified string to the file at the current position.
 *
 * \param strData Pointer to a null terminated character string.
 * \return Return true if the string has been written successfully.
 */
bool StdioFile::writeString(const char* strData)
{
	return fputs(strData, fileStream) >= 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read string from file
 *
 * Reads a string from the file.
 *
 * \return Returns the STL string object.
 */
std::string StdioFile::readString()
{
	char buf[256] = "";
	if(!fgets(buf, sizeof(buf), fileStream)) return "";
	else return buf;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get file size
 *
 * Get the length of the file
 *
 * \return Returns the length of the file in bytes or -1 in case of error.
 */
long StdioFile::getSize()
{
	long pos, ret;
	pos = ftell(fileStream);
	if (pos < 0) return -1;
	fseek(fileStream, 0, SEEK_END);
	ret = ftell(fileStream);
	fseek(fileStream, pos, SEEK_SET);
	return ret;
}
