/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <platform.h>
#include <string.h>

#include "SysInfo.h"

//#define DEBUG
#include "Debug.h"

#define SCANNER_INFO_VERSION	1

char SysInfo::serialNumber[MAX_VAR_SIZE] = "";
char SysInfo::modelType[MAX_VAR_SIZE] = "";
char SysInfo::hardwareRevision[MAX_VAR_SIZE] = "";
char SysInfo::firmwareVersion[MAX_VAR_SIZE] = "";
char SysInfo::envBuffer[MAX_ENV_SIZE] = "";
SCANNER_INFORMATION SysInfo::scannerInfo = { {0}, {0}, 0 };
unsigned char SysInfo::scannerInfoVersion = 0;
time_t SysInfo::manufTimeStamp = 0;
int SysInfo::commModuleType = -1;

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get device serial number
 *
 * Retrieves the unit serial number from data terminal onboard flash memory.
 * \note The first call reads the flash block, subsequent calls reuse cached data.
 *
 * Serial number should contain six decimal numbers.
 *
 * \return Returns the serial number of the mobile data terminal.
 */
const char* SysInfo::getSerialNumber()
{
	if(!strlen(serialNumber))
		getUbootEnv(UBOOT_VAR_SERIAL, serialNumber, sizeof(serialNumber));

	return serialNumber;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get device firmware version
 *
 * Retrieves the version of the data terminal firmware.
 * \note The first call reads the firmware version from file, subsequent calls
 * reuse cached data.
 *
 * \return Returns the firmware version of the mobile data terminal.
 */
const char* SysInfo::getFirmwareVersion()
{
	if(!strlen(firmwareVersion))
	{
		FILE *f = fopen(FILE_FIRMWARE_VER, "r");
		if(f)
		{
			fgets(firmwareVersion, sizeof(firmwareVersion), f);
			int len = strlen(firmwareVersion);
			while(-- len > 0)
			{
				if(firmwareVersion[len] == '\r' ||
					firmwareVersion[len] == '\n')
					firmwareVersion[len] = 0;
				else break;
			}
			fclose(f);
		}
	}

	return firmwareVersion;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get device hardware revision
 *
 * Retrieves the unit hardware revision from data terminal onboard flash memory.
 * \note The first call reads the flash block, subsequent calls reuse cached data.
 *
 * \return Returns the hardware revision of the mobile data terminal.
 */
const char* SysInfo::getHardwareRevision()
{
	if(!strlen(hardwareRevision))
		getUbootEnv(UBOOT_VAR_HWREV, hardwareRevision, sizeof(hardwareRevision));

	return hardwareRevision;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get device manufacture date
 *
 * Retrieves the unit manufacture timestamp from data terminal onboard flash
 * memory.
 * \note The first call reads the flash block, subsequent calls reuse
 * cached data.
 *
 * \return Returns the manufacture timestamp of the mobile data terminal.
 */
time_t SysInfo::getManufTimeStamp()
{
	if(!manufTimeStamp) {
		char str[MAX_VAR_SIZE] = "";
		getUbootEnv(UBOOT_VAR_MFTIMESTAMP, str, sizeof(str));
		manufTimeStamp = (time_t)strtoul(str, NULL, 10);
	}

	return manufTimeStamp;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get device model type
 *
 * Retrieves the unit model type from data terminal onboard flash memory.
 * \note The first call reads the flash block, subsequent calls reuse cached data.
 *
 * Defined values:
 * - ts8 -- Artec Triskan TS8 mobile data terminal
 *
 * \return Returns the model type identifier of the mobile data terminal.
 */
const char* SysInfo::getModelType()
{
	if (strlen(modelType)) return modelType;

	char tmp[16];
	getUbootEnv(UBOOT_VAR_MODEL, modelType, sizeof(modelType));
	strcat(modelType, "-");

	SCANNER_INFORMATION *pSI = getScannerInformation();

	if (pSI == NULL)
	{
		// default identifier
		sprintf(tmp, "%d000", getCommModule());
		strcat(modelType, tmp);
	}
	else
	{
		// insert GPRS/WiFi model identifier
		sprintf(tmp, "%d", getCommModule());
		strcat(modelType, tmp);
		// append reader identifier
		if (pSI->readerType[0] != 0)
		{
			strncat(modelType, pSI->readerType,
					sizeof(pSI->readerType));
		}
		else strcat(modelType, "000");
	}
	return modelType;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read bootloader environment variable
 *
 * Retrieves the value of the U-boot bootloader environment variable.
 * The data is located in the mobile terminal internal flash memory.
 * The bootloader environment variables are intended to store global
 * manufacturing and configuration data which is considered read-only during
 * a normal mobile terminal operation.
 *
 * \param varName Specifies the name of the environment variable.
 * \param pBuf Contains a pointer to the buffer where the variable value is copied to.
 * \param maxLen Specifies the maximum length of the returned data.
 * \return Returns true if the variable was read successfully. Returns false if
 * the variable was not found, or on a flash memory error.
 */
bool SysInfo::getUbootEnv(const char* varName, char* pBuf, int maxLen)
{
	int fd = open(FLASH_DEV, O_RDONLY);
	if(fd < 0)
	{
		debug("failed opening %s\n", FLASH_DEV);
		return false;
	}

	if(lseek(fd, FLASH_OFFSET, SEEK_SET) == -1)
	{
		debug("failed seeking flash\n");
		close(fd);
		return false;
	}

	size_t readBytes = read(fd, envBuffer, MAX_ENV_SIZE);
	if(readBytes != MAX_ENV_SIZE)
	{
		debug("failed reading flash\n");
		close(fd);
		return false;
	}

	char *startPtr = NULL;
	char currentEnv[256]="";

	for (startPtr = envBuffer; *startPtr != 0; startPtr += (strlen(currentEnv) + 1))
	{
		memset(currentEnv, 0, sizeof(currentEnv));
		strncpy(currentEnv, startPtr, sizeof(currentEnv)-1);
		size_t varLen = strlen(varName);

		char *pVal = strchr(currentEnv, '=');
		if (pVal == NULL)
		{
			debug("no key associated with %s\n", varName);
			continue; // invalid variable?
		}

		// check that variable length matches too
		if ((size_t)(pVal - currentEnv) != varLen) continue;

		if (!strncmp(currentEnv, varName, varLen))
		{
			strncpy(pBuf, ++pVal, maxLen);
			debug("variable %s is [%s]\n", varName, pBuf);
			close(fd);
			return true;
		}
	}

	close(fd);
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get scanner information
 *
 * Retrieves information about a scanner module attached to the mobile data
 * terminal.
 *
 * \note The first call reads the data from the scanner module,
 * subsequent calls reuse cached data, since the module is not hot-swappable.
 *
 * \return Returns scanner module descriptor. Returns NULL if the information
 * is not available or communication to scanner module fails.
 */
SCANNER_INFORMATION* SysInfo::getScannerInformation()
{
	if (scannerInfoVersion == 0)
	{
		int fd = open(READER_MODULE_EEPROM, O_RDONLY);
		if (fd < 0)
		{
			debug("Failed opening scanner EEPROM driver\n");
			return NULL;
		}

		ssize_t readBytes = read(fd, &scannerInfoVersion,
				sizeof(scannerInfoVersion));
		if (readBytes != sizeof(scannerInfoVersion))
		{
			debug("Failed reading scanner information\n");
			return NULL;
		}

		// check version identifier in scanner info block
		if (scannerInfoVersion != SCANNER_INFO_VERSION)
		{
			debug("Invalid scanner information\n");
			return NULL;
		}

		// read out scanner information block
		readBytes = read(fd, &scannerInfo, sizeof(scannerInfo));
		close(fd);

		if (readBytes != sizeof(scannerInfo))
		{
			debug("Failed reading scanner information\n");
			return NULL;
		}
	}

	return &scannerInfo;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get communication module type
 *
 * Retrieves the data terminal communication module type from the onboard
 * flash memory.
 * \note The first call reads the flash block, subsequent calls reuse cached data.
 *
 * Defined values are:
 * - 0 -- No wireless connectivity
 * - 1 -- GPRS module installed
 * - 5 -- WLAN module installed
 *
 * \return Returns the mobile data terminal communication moduele type identifier.
 */
int SysInfo::getCommModule()
{
	if (commModuleType < 0)
	{
		char buf[MAX_VAR_SIZE];
		if(getUbootEnv(UBOOT_VAR_COMM_TYPE, buf, sizeof(buf)))
		{
			commModuleType = atoi(buf);
		}
		// default is a GPRS-enabled unit
		else commModuleType = COMM_MODULE_GPRS;
	}

	return commModuleType;
}
