/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include "LibSpatha.h"

#ifdef USE_XYNTH

#include <exynthoo/GuiLibrary.h>

#include "PowerManager.h"
#include "SystemMonitor.h"

//#define DEBUG
#include "Debug.h"

#define DEFAULT_DISPLAY_DIM_DELAY		5
#define DEFAULT_DISPLAY_MIN_DELAY		20
#define DEFAULT_DISPLAY_OFF_DELAY		30
#define DEFAULT_POWERSAVE_DELAY			60
#define	DEFAULT_POWERDOWN_DELAY			3600

#define DEFAULT_BATTERY_INTERVAL		4*1000
#define DEFAULT_TEMPERATURE_INTERVAL	10*1000
#define DEFAULT_LIGHTING_INTERVAL		3*1000
#define DEFAULT_MAX_KEYPAD_BACKLIGHT	255

#define LIGHT_SENSOR_THRESHOLD1			10
#define LIGHT_SENSOR_THRESHOLD2			30

#define MAXIMAL_SLEEP_PERIOD	(15*60) // 15 minutes
#define MINIMAL_SLEEP_PERIOD	30		// 30 seconds

#define SERVICE_POLL_INTERVAL	1000

SystemMonitor* SystemMonitor::monitor = NULL;

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the system monitor object.
 * Local variables (displayDimDelay, displayOffDelay, batteryPollInterval,
 * tempPollInterval, lightPollInterval, keypadMaxBrightness) are set to their
 * default values.
 *
 * Instead of invoking this constructor directly, the method initialize() should be used.
 * It creates the instance of the class which can be later retrieved by invoking the
 * getMonitor() method.
 */
SystemMonitor::SystemMonitor() : GuiTimer()
{
	displayDimDelay = DEFAULT_DISPLAY_DIM_DELAY;
	displayMinDelay = DEFAULT_DISPLAY_MIN_DELAY;
	displayOffDelay = DEFAULT_DISPLAY_OFF_DELAY;
	powerSaveDelay = DEFAULT_POWERSAVE_DELAY;
	powerDownDelay = DEFAULT_POWERDOWN_DELAY;

	lastBatteryPoll = lastTempPoll = s_gettimeofday();

	batteryPollInterval = DEFAULT_BATTERY_INTERVAL;
	tempPollInterval = DEFAULT_TEMPERATURE_INTERVAL;
	lightPollInterval = DEFAULT_LIGHTING_INTERVAL;
	keypadMaxBrightness = DEFAULT_MAX_KEYPAD_BACKLIGHT;

	for (int i = 0; i<NUM_POWER_STATES; i++)
		powerStateCounter[i] = 0;

	interval = SERVICE_POLL_INTERVAL;

	usbConnected = false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Terminates the system monitor object.
 */
SystemMonitor::~SystemMonitor()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start system monitor
 *
 * This function initializes and starts the system monitor.
 *
 * \return Returns true if the system monitor was initialized successfully.
 */
bool SystemMonitor::initialize()
{
	if (monitor != NULL) return true;
	SystemMonitor *pMonitor = new SystemMonitor();
	if (!pMonitor->init()) {
		delete pMonitor;
		return false;
	}
	monitor = pMonitor;
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop system monitor
 *
 * This function shuts down the system monitor.
 */
void SystemMonitor::shutdown()
{
	if (monitor == NULL) return;
	monitor->close();
	delete monitor;
	monitor = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get system monitor object
 *
 * \return returns singleton system monitor object
 */
SystemMonitor* SystemMonitor::getMonitor() {
	return monitor;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start system monitor
 *
 * Starts the system monitor timer.
 *
 * \return Returns true if the system monitor was started successfully.
 */
bool SystemMonitor::init()
{
	return startTimer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop system monitor
 *
 * Shuts down the system monitor.
 */
void SystemMonitor::close()
{
	stopTimer();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Process system data
 * Process system data when the timer has fired
 */
void SystemMonitor::onTimer()
{
	PowerManager *pPm = PowerManager::getPowerManager();

	// kick idle when usb plugged in or disconnected
	if (pPm) {
		bool old = usbConnected;
		usbConnected =pPm->isUsbConnected();
		if (old != usbConnected) pPm->kickIdle();
	}

	// process battery voltage
	handleBattery();

	// process temperature sensor data
	handleTemperature();

	// process ambient lighting sensor data
	handleLighting();

	// handle power state switching
	handlePower();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set power state change delays
 * Sets the time after which various power events should happen.
 *
 * \param dim Time to dim the display to half power
 * \param min Time to dim the display to lowest setting
 * \param off Time to turn off the display
 * \param save Time the system should go to low-power sleep
 * \param down Time to shut down the system
 */
void SystemMonitor::setPowerDelays(int dim, int min, int off, int save, int down)
{
	debug("Power delays: %d, %d, %d, %d, %d\n", dim, min, off, save, down);
	displayDimDelay = dim;
	displayMinDelay = min;
	displayOffDelay = off;
	powerSaveDelay = save;
	powerDownDelay = down;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set keypad maximal brightness
 */
void SystemMonitor::setKeyboardBrightness(int b)
{
	keypadMaxBrightness = b;
	if (keypadMaxBrightness < 0) keypadMaxBrightness = 0;
	if (keypadMaxBrightness > 255) keypadMaxBrightness = 255;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get power state counters
 */
void SystemMonitor::getPowerCounters(int *counters)
{
	for (int i = 0; i<NUM_POWER_STATES; i++)
		counters[i] = powerStateCounter[i];
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief System idle monitor
 *
 * System idle monitor function is called by the system monitor periodically
 * to indicate that the system state remains idle.
 *
 * Default behavior is to automatically cycle through a number of power states.
 * When the displayDimDelay is elapsed, the display is dimmed. When the
 * displayOffDelay is elapsed, the display is disabled completely

 */
void SystemMonitor::handlePower()
{
	PowerManager *pPm = PowerManager::getPowerManager();

	PowerState state = pPm->getCurrentState();
	int idle = pPm->getIdleDuration();

	if (sleepStart != 0) {
		// we're back from sleep. log sleep duration
		powerStateCounter[LOW_POWER_SLEEP] =
						powerStateCounter[LOW_POWER_SLEEP] +
						(time(NULL)-sleepStart);
		sleepStart = 0;
	}

	switch(state)
	{
	case FULLY_OPERATIONAL:

		powerStateCounter[FULLY_OPERATIONAL] =
				powerStateCounter[FULLY_OPERATIONAL] +
				(SERVICE_POLL_INTERVAL / 1000);

		if(displayDimDelay <= 0) break;
		if(idle < displayDimDelay) break;

		// reduce OLED brightness
		pPm->setPowerMode(DISPLAY_DIMMED);
		break;

	case DISPLAY_DIMMED:

		powerStateCounter[DISPLAY_DIMMED] =
				powerStateCounter[DISPLAY_DIMMED] +
				(SERVICE_POLL_INTERVAL / 1000);

		if(displayMinDelay <= 0) break;
		if(idle < displayMinDelay) break;

		// reduce OLED brightness to minimum
		pPm->setPowerMode(DISPLAY_MINIMAL);
		break;

	case DISPLAY_MINIMAL:

		powerStateCounter[DISPLAY_MINIMAL] =
				powerStateCounter[DISPLAY_MINIMAL] +
				(SERVICE_POLL_INTERVAL / 1000);

		if(displayOffDelay <= 0) break;
		if(idle < displayOffDelay) break;

		// disable OLED screen
		pPm->setPowerMode(DISPLAY_DISABLED);
		break;

	case DISPLAY_DISABLED:

		powerStateCounter[DISPLAY_DISABLED] =
				powerStateCounter[DISPLAY_DISABLED] +
				(SERVICE_POLL_INTERVAL / 1000);

		if(powerSaveDelay <= 0) break;
		if(idle < powerSaveDelay) break;
		if(usbConnected) break;
		// do not break, time to go to sleep!

	case LOW_POWER_AWAKE: // <- sleepwalking
	{
		// go to sleep
		// already power-down time? (handled below)
		if (powerDownDelay > 0 && idle >= powerDownDelay) break;

		// go to sleep for determined time
		int powerSaveSleep = powerDownDelay - idle;
		// no powerDownDelay configured, use maximal period
		if (powerDownDelay <= 0)
			powerSaveSleep = MAXIMAL_SLEEP_PERIOD;
		// sanity check
		if (powerSaveSleep > MAXIMAL_SLEEP_PERIOD)
			powerSaveSleep = MAXIMAL_SLEEP_PERIOD;
		if (powerSaveSleep < MINIMAL_SLEEP_PERIOD)
			powerSaveSleep = MINIMAL_SLEEP_PERIOD;

		debug("Sleeping for %d sec\n", powerSaveSleep);
		pPm->setPowerSaveSleep(powerSaveSleep);
		sleepStart = time(NULL);
		pPm->setPowerMode(LOW_POWER_SLEEP);
		idle = pPm->getIdleDuration();
	}
		break;

	case LOW_POWER_SLEEP:
		// We're reaching this point only when the sleep
		// has not yet begun or wakeup has not yet ended.
		// Ignore.
		break;

	default:
		break;
	}

	// check if it is time to power down
	debug("idle: %d\n", idle);
	if((powerDownDelay > 0) && (idle >= powerDownDelay) && !usbConnected)
	{
		debug("power saving elapsed, shutdown\n");

		// shut everything down
		pPm->setPowerMode(POWER_DOWN);
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle battery data
 * Handle battery information when the system monitor polling timer has been fired
 */
void SystemMonitor::handleBattery()
{
	if (batteryPollInterval <= 0) return;

	time_t timeNow = s_gettimeofday();
	if(timeNow >= (lastBatteryPoll + batteryPollInterval))
	{
		lastBatteryPoll = timeNow;
		int currentBattery = 0;

		PowerManager *pPM = PowerManager::getPowerManager();
		int battCapacity = pPM->getBatteryCapacity(&currentBattery);

		if(battCapacity < 0)
		{
			debug("error reading battery capacity (%d)\n", battCapacity);
		}
		else
		{
			// call all listeners
			GuiLibrary::broadcastSystemEvent(EV_POWER_BATTERY, (void*)battCapacity);

			debug("battery capacity %d mV (%d%%)\n", currentBattery, battCapacity);

			if(battCapacity == 0)
			{
				// call all listeners
				GuiLibrary::broadcastSystemEvent(EV_POWER_LOWBAT, (void*)battCapacity);
			}
		}


	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle temperature data
 * Handle temperature information when the system monitor polling timer has been fired
 */
void SystemMonitor::handleTemperature()
{
	if (tempPollInterval <= 0) return;

	time_t timeNow = s_gettimeofday();
	if(timeNow >= (lastTempPoll + tempPollInterval))
	{
		lastTempPoll = timeNow;

		PowerManager *pPM = PowerManager::getPowerManager();
		int currentTemp = pPM->getTempReading();

		// call all listeners
		// NB! INT_MIN is sent on sensor error!
		GuiLibrary::broadcastSystemEvent(EV_TEMPERATURE, (void*)currentTemp);

		if (currentTemp >= 0) {
			debug("Current temperature: %d.%d 'C\n", currentTemp / 10, currentTemp % 10);
		} else {
			debug("Current temperature: -%d.%d 'C\n", (-currentTemp) / 10, (-currentTemp) % 10);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Handle lighting  data
 * Handle lighting information when the system monitor polling timer has been fired
 */
void SystemMonitor::handleLighting()
{
	if (lightPollInterval <= 0) return;

	time_t timeNow = time(NULL);
	if(timeNow >= (lastLightPoll + lightPollInterval))
	{
		PowerManager *pPM = PowerManager::getPowerManager();
		PowerState state = pPM->getCurrentState();

		if(state != FULLY_OPERATIONAL && state != DISPLAY_DIMMED) return;

		int lightLevel = pPM->getAmbientLight();
		debug("ambient light: %d\n", lightLevel);

		if(lightLevel > LIGHT_SENSOR_THRESHOLD2)
		{
			pPM->keybFadeBackLight(0);
		}
		else if(lightLevel > LIGHT_SENSOR_THRESHOLD1)
		{
			pPM->keybFadeBackLight(keypadMaxBrightness / 2);
		}
		else
		{
			pPM->keybFadeBackLight(keypadMaxBrightness);
		}
	}
}

#endif
