/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

/*
 * TcpSocket.cpp
 *
 *  Created on: Sep 29, 2008
 *      Author: andrei
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <netinet/tcp.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>

#include "TcpSocket.h"

//#define DEBUG
//#define DEBUG_DUMP
//#define DEBUG_DUMP_HEX

#include "Debug.h"

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the socket object. Socket timeouts are set to default values
 * of 30 seconds.
 */
TcpSocket::TcpSocket()
{
	recvTimeout = DEFAULT_RECV_TIMEOUT;
	sendTimeout = DEFAULT_SEND_TIMEOUT;
	socketHandle = -1;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destroys the socket object.
 */
TcpSocket::~TcpSocket()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Connect socket
 *
 * Creates a socket and connects to the specified host, given the host name
 * or IP address and the port number.
 *
 * \param host Contains the remote host name or IP address in text format.
 * \param port Specifies the remote port number.
 * \return Returns true if the socket was successfully connected.
 */
bool TcpSocket::connect(const char* host, int port)
{
	char service[16];
	sprintf(service, "%d", port);

	struct addrinfo addrHint, *addrInfo = NULL;
	memset(&addrHint, 0, sizeof(addrHint));
	addrHint.ai_family = AF_INET;
	addrHint.ai_socktype = SOCK_STREAM;

	// query host addresses
	if (getaddrinfo(host, service, &addrHint, &addrInfo) != 0)
	{
		debug("failed querying address info: %d\n", errno);
		return false;
	}

	// create socket
	socketHandle = ::socket(addrInfo->ai_family, addrInfo->ai_socktype,
			addrInfo->ai_protocol);

	if (socketHandle < 0)
	{
		debug("failed creating socket: %d\n", errno);
		freeaddrinfo(addrInfo);
		return false;
	}

	// initialize timeout values
	if (!setTimeouts(sendTimeout, recvTimeout))
	{
		close();
		debug("failed setting socket timeouts (%d)\n", errno);
		freeaddrinfo(addrInfo);
		return false;
	}

	// connect to host
	if (::connect(socketHandle, addrInfo->ai_addr, addrInfo->ai_addrlen) < 0)
	{
		close();
		debug("failed connecting to host: %s:%d (%d)\n", host, port, errno);
		freeaddrinfo(addrInfo);
		return false;
	}

	// cleanup
	freeaddrinfo(addrInfo);
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Close socket
 *
 * Closes the socket.
 *
 * \return Returns true if the operation was a success.
 */
bool TcpSocket::close()
{
	if (socketHandle >= 0)
	{
		int ret = ::close(socketHandle);
		socketHandle = -1;
		return ret == 0;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Check connection state
 *
 * \return Returns true if the socket is connected to a remote host.
 */
bool TcpSocket::isConnected()
{
	struct sockaddr hostAddr;
	socklen_t hostLen = sizeof(hostAddr);
	return ::getpeername(socketHandle, &hostAddr, &hostLen) == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Write socket data
 *
 * Writes data to the socket and sends data to remote host.
 *
 * \param pData Points to the data buffer about to be sent.
 * \param dataLen Specifies the length of the data in bytes.
 * \return Returns true if all data bytes were sent successfully.
 */
bool TcpSocket::writeData(unsigned char *pData, size_t dataLen)
{
	ssize_t dataSent = ::send(socketHandle, pData, dataLen, MSG_NOSIGNAL);

	debug_dump(pData, dataLen);
	debug("sent %d bytes\n", dataSent);

	return (size_t) dataSent == dataLen;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Read socket data
 *
 * Receives data from the socket originating from a remote host.
 * The function blocks until a portion of data has been received, or the
 * socket receive timeout expires.
 * Receive may return less data than specified by maxLen depending on
 * the underlying protocols.
 *
 * \param pData Points to the buffer that receives incoming data.
 * \param maxLen Specifies the maximum length of data to receive.
 * \return Returns the number of bytes received from the socket.
 */
size_t TcpSocket::readData(unsigned char* pData, size_t maxLen)
{
	ssize_t dataRecv = ::recv(socketHandle, pData, maxLen, 0);

	debug_dump(pData, dataRecv);

	debug("received %d bytes\n", dataRecv);
	return dataRecv < 0 ? 0 : (size_t) dataRecv;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set socket timeouts
 *
 * Sets new socket timeout values. The timeouts are effective immediately if
 * the socket is already connected, or cached until a new connection is made.
 *
 * \param sendTout Specifies the socket send timeout, in milliseconds.
 * \param recvTout Specifies the socket receive and connect timeout, in milliseconds.
 */
bool TcpSocket::setTimeouts(int sendTout, int recvTout)
{
	sendTimeout = sendTout;
	recvTimeout = recvTout;

	if(socketHandle == -1) return true;

	struct timeval tv;
	tv.tv_sec = recvTout / 1000;
	tv.tv_usec = (recvTout % 1000) * 1000;

	if (::setsockopt(socketHandle, SOL_SOCKET, SO_RCVTIMEO, (char*) &tv,
			sizeof(tv)) < 0)
	{
		debug("failed setting receive timeout: %d\n", errno);
		return false;
	}

	tv.tv_sec = sendTout / 1000;
	tv.tv_usec = (sendTout % 1000) * 1000;

	if (::setsockopt(socketHandle, SOL_SOCKET, SO_SNDTIMEO, (char*) &tv,
			sizeof(tv)) < 0)
	{
		debug("failed setting send timeout: %d\n", errno);
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Set socket keepalive
 *
 * Sets new socket keepalive parameters.
 *
 * \param keepIdle Specifies the time period after which the first keepalive is sent, in seconds.
 * \param keepInterval Specifies the time interval between consequent keepalives, in seconds.
 * \param keepCount Specifies the maximum number of keepalive probes sent before the connection is dropped.
 * \return Returns true if the keepalive settings have been set successfully.
 */
bool TcpSocket::setKeepalive(int keepIdle, int keepInterval, int keepCount)
{
	if (::setsockopt(socketHandle, IPPROTO_TCP, TCP_KEEPCNT, &keepCount,
			sizeof(keepCount)) < 0)
	{
		debug("failed setting TCP_KEEPCNT\n");
		return false;
	}

	if (::setsockopt(socketHandle, IPPROTO_TCP, TCP_KEEPIDLE, &keepIdle,
			sizeof(keepIdle)) < 0)
	{
		debug("failed setting TCP_KEEPIDLE\n");
		return false;
	}

	if (::setsockopt(socketHandle, IPPROTO_TCP, TCP_KEEPINTVL, &keepInterval,
			sizeof(keepInterval)) < 0)
	{
		debug("failed setting TCP_KEEPINTVL\n");
		return false;
	}

	int enableFlag = 1;
	if (::setsockopt(socketHandle, SOL_SOCKET, SO_KEEPALIVE, &enableFlag,
			sizeof(enableFlag)) < 0)
	{
		debug("failed enabling keepalive on socket\n");
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get pending input length
 *
 * \return Returns the number of unread bytes pending in the socket receive buffer.
 */
size_t TcpSocket::getPendingInput()
{
	int ret = 0;
	if(::ioctl(socketHandle, SIOCINQ, &ret) < 0)
	{
		debug("failed querying length of receive queue\n");
	}

	return (size_t)ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Get pending output length
 *
 * \return Returns the number of unsent bytes pending in the socket transmit buffer.
 */
size_t TcpSocket::getPendingOutput()
{
	int ret = 0;
	if(::ioctl(socketHandle, SIOCOUTQ, &ret) < 0)
	{
		debug("failed querying length of transmit queue\n");
	}

	return (size_t)ret;
}
