/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <time.h>
#include <pthread.h>
#include <signal.h>
#include <sys/time.h>

#include "Thread.h"

#define THREAD_STACK    (1024*512)

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the thread class attributes.
 */
Thread::Thread()
{
	threadHandle = 0;
	stopRequest = false;
	stopComplete = false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Destroys the thread class instance.
 */
Thread::~Thread()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Sleep
 *
 * Suspends the thread for a given amount of time. You may call this static
 * function from any location of your application code.
 *
 * \param msec Specifies the thread sleep time in milliseconds.
 * \return Returns true if the call was a success. Returns false if the thread
 * sleep was interrupted.
 */
bool Thread::sleep(int msec)
{
	timespec sleepTime =
	{ msec / 1000, msec % 1000 * 1000000 };
	return nanosleep(&sleepTime, NULL) == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Thread main loop
 *
 * Override this to implement your worker thread. Run function gets called by
 * the thread procedure. The thread is terminated after the run function
 * returns.
 */
void Thread::run()
{
}

////////////////////////////////////////////////////////////////////////////////

void Thread::signalHandler(int sig)
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Run the thread
 * Runs the thread. Initializes the thread and sets appropriate signal actions.
 *
 * \param pParam Pointer to thread parameters
 */
void* Thread::threadRun(void *pParam)
{
	Thread *pThis = (Thread*) pParam;

	struct sigaction act;
	sigset_t set;
	sigemptyset(&set);
	sigaddset(&set, SIGUSR1);

	// initialize thread signal mask
	pthread_sigmask(SIG_UNBLOCK, &set, NULL);

	sigemptyset(&act.sa_mask);
	act.sa_handler = signalHandler;

	// set signal action
	sigaction(SIGUSR1, &act, NULL);

	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);

	pThis->run();
	pThis->reportStopped();

	pthread_exit(NULL);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Start thread
 *
 * Launches the thread procedure and calls run function.
 *
 * \return Returns true if the call is successful.
 */
bool Thread::start()
{
	mutex.lock();
	if (threadHandle == 0)
	{
		stopRequest = false;
		stopComplete = false;

        pthread_attr_t attr;

		pthread_attr_init(&attr);
		pthread_attr_setstacksize(&attr, THREAD_STACK);

		int rc = pthread_create(&threadHandle, &attr, threadRun, (void*) this);

		pthread_attr_destroy(&attr);

		if (rc == 0)
		{
			mutex.unlock();
			return true;
		}
	}

	mutex.unlock();
	return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Stop thread
 *
 * Terminates the running thread in a controlled fashion. First, the thread
 * termination request flag is raised. Thread implementation must then detect
 * this condition and clean up returning from run() function.
 *
 * If the thread does not react to the termination request, it is terminated
 * in an uncontrolled way via pthread_cancel() call.
 *
 * \param timeOut Specifies the amount of time to wait for a controlled thread
 * shutdown in milliseconds.
 * \param interrupt If true, a SIGUSR1 will be sent to the thread to notify
 * that it should be shut down.
 * \return Returns true if the thread is stopped or terminated.
 */
bool Thread::stop(int timeOut, bool interrupt)
{
	mutex.lock();
	if (threadHandle == 0)
	{
		mutex.unlock();
		return false;
	}

	if (!stopComplete)
	{
		stopRequest = true;		// raise thread stop flag

		if (interrupt)
			pthread_kill(threadHandle, SIGUSR1); // interrupt the thread
			// to make it to come back from blocking reads etc.

		for (int i = 0; (i < timeOut) && !stopComplete; i += 100)
		{
			mutex.unlock();
			Thread::sleep(100);
			mutex.lock();
		}

		stopRequest = false;
		if (!stopComplete)		// thread failed to shut down politely
		{
			if (pthread_cancel(threadHandle) != 0)
			{
				mutex.unlock();
				threadHandle = 0;
				return false;
			}
		}
	}

	pthread_join(threadHandle, NULL);
	threadHandle = 0;

	mutex.unlock();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Request the stop of a thread
 *
 * Sets a flag that the thread should stop and does not block. Call Thread::stop()
 * afterwards to clean up after the thread to avoid memory leak!
 */
void Thread::requestStop()
{
	mutex.lock();
	if (threadHandle != 0)
		stopRequest = true;		// raise thread stop flag
	mutex.unlock();
	return;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Terminate thread
 *
 * Kills the thread in an uncontrolled fashion. Thread resources may not be
 * properly released, since the thread procedure is aborted at any cancellation
 * point, which may be arbitrary libc I/O function call.
 *
 * \return Returns true if the thread was terminated successfully.
 */
bool Thread::terminate()
{
	mutex.lock();
	if (threadHandle == 0)
	{
		mutex.unlock();
		return false;
	}

	if (pthread_cancel(threadHandle) != 0)
	{
		mutex.unlock();
		return false;
	}

	pthread_join(threadHandle, NULL);
	threadHandle = 0;

	mutex.unlock();
	return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Interrupt thread execution
 *
 * Sends SIGUSR1 signal to thread and causes POSIX compatible functions to
 * abort with EINTR. Note that thread must still be terminated via the
 * stop() call.
 *
 * \return Returns true if the signal was sent successfully.
 */
bool Thread::interrupt()
{
	mutex.lock();
	int ret = -1;
	if (threadHandle != 0)
	{
		ret = pthread_kill(threadHandle, SIGUSR1);
	}
	mutex.unlock();
	return ret >= 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Check shutdown flag
 *
 * Returns the thread procedure shutdown flag set by reportStopped() method
 * call.
 * \return Returns true if the thread shutdown flag is set.
 */
bool Thread::stopRequested()
{
	mutex.lock();
	bool result = stopRequest;
	mutex.unlock();
	return result;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Signal thread completion
 *
 * Signals thread completion and indicates a successful controlled
 * thread shutdown. Normally, the implementation does not need to call this
 * function, as it is done automatically by the thread procedure after
 * the run() function returns.
 */
void Thread::reportStopped()
{
	mutex.lock();
	stopRequest = false;
	stopComplete = true;
	mutex.unlock();
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the semaphore object. Initial state of the semaphore is
 * not posted.
 */
Semaphore::Semaphore()
{
	pthread_mutex_init(&mutex, NULL);
	pthread_cond_init(&cond, NULL);
}

/**
 * \brief Class destructor
 *
 * Terminates the semaphore object. The semaphore is not posted before
 * termination.
 */
Semaphore::~Semaphore()
{
	pthread_mutex_unlock(&mutex);
	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&cond);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Post semaphore
 *
 * Posts (or signals, releases) a semaphore. The first thread pending on the
 * semaphore will be released.
 */
void Semaphore::post()
{
	pthread_mutex_lock(&mutex);
	pthread_cond_signal(&cond);
	pthread_mutex_unlock(&mutex);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Pend on semaphore
 *
 * Pends on the semaphore (or waits, acquires the semaphore). The call will
 * block until another thread posts the semaphore.
 *
 * \return Returns true if the semaphore has been posted. Returns false if
 * the call has been interrupted.
 */
bool Semaphore::pend()
{
	pthread_mutex_lock(&mutex);

	bool ret = pthread_cond_wait(&cond, &mutex) == 0;
	pthread_mutex_unlock(&mutex);

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Pend on semaphore with timeout
 *
 * Pends on the semaphore (or waits, acquires the semaphore) for the given
 * amount of time. The call will block until another thread posts the
 * semaphore, or the timeout expires.
 *
 * \param ms Pending operation timeout, in milliseconds.
 * \return Returns true if the semaphore has been posted. Returns false on a
 * timeout condition or if the call has been interrupted.
 */
bool Semaphore::pend(int ms)
{
	pthread_mutex_lock(&mutex);

	timeval tp;
	gettimeofday(&tp, NULL);

	timespec ts;
	ts.tv_sec = tp.tv_sec + ms / 1000;
	long usec = tp.tv_usec + ((ms % 1000) * 1000);
	ts.tv_sec += usec / 1000000;
	ts.tv_nsec = (usec % 1000000) * 1000;

	bool ret = pthread_cond_timedwait(&cond, &mutex, &ts) == 0;
	pthread_mutex_unlock(&mutex);

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the mutex object. Initial state of the mutex is unlocked.
 */
Mutex::Mutex()
{
	pthread_mutex_init(&mutex, NULL);
}

/**
 * \brief Class destructor
 *
 * Terminates the mutex object. The mutex is unlocked before termination.
 */
Mutex::~Mutex()
{
	pthread_mutex_unlock(&mutex);
	pthread_mutex_destroy(&mutex);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Lock mutex
 *
 * Locks (or acquires) the mutex object by the calling thread. The call
 * will block until the mutex is unlocked by another thread.
 */
bool Mutex::lock()
{
	return pthread_mutex_lock(&mutex) == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unlock mutex
 *
 * Unlocks (or releases) the mutex object and releases all lock() calls.
 * The first thread that is pending on mutex lock will be released.
 */
void Mutex::unlock()
{
	pthread_mutex_unlock(&mutex);
}


////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the condition variable object.
 */
Cond::Cond()
{
	pthread_mutex_init(&mutex, NULL);
	pthread_cond_init(&cond, NULL);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Terminates the condition variable object. The mutex is unlocked before termination.
 */
Cond::~Cond()
{
	pthread_mutex_unlock(&mutex);
	pthread_mutex_destroy(&mutex);
	pthread_cond_destroy(&cond);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Signal a condition
 *
 * Unblocks a thread that is blocked by condition variable.
 * Mutex must be locked before calling signal(); unlock mutex
 * afterwards.
 */
void Cond::signal()
{
	pthread_cond_signal(&cond);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Wait for a condition
 *
 * Waits for a condition variable being signaled.
 * Mutex must be locked before calling wait(); wait unlocks the mutex
 * atomically while sleeping and re-locks after exiting from sleep;
 * unlock mutex afterwards.
 *
 * \return Returns true if the thread timed wait call executed successfully.
 */
bool Cond::wait()
{
	return (pthread_cond_wait(&cond, &mutex) == 0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Wait for a condition, up to specified milliseconds
 *
 * Waits for a condition variable being signaled.
 * Mutex must be locked before calling wait(); wait unlocks the mutex
 * atomically while sleeping and re-locks after exiting from sleep;
 * unlock mutex afterwards.
 *
 * \param ms Number of milliseconds to wait
 * \return Returns true if the thread timed wait call executed successfully.
 */
bool Cond::wait(int ms)
{
	struct timeval now;
	gettimeofday(&now, NULL);

	timespec ts;
	ts.tv_sec = now.tv_sec + ms / 1000;
	long usec = now.tv_usec + ((ms % 1000) * 1000);
	ts.tv_sec += usec / 1000000;
	ts.tv_nsec = (usec % 1000000) * 1000;

	bool ret = pthread_cond_timedwait(&cond, &mutex, &ts) == 0;

	return ret;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Lock the thread mutex
 * Aquire mutex lock for the thread
 *
 * \return Returns true if locking was successful.
 */
bool Cond::lock()
{
	return pthread_mutex_lock(&mutex) == 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Unlock the thread mutex
 * Release the mutex lock for the thread
 */
void Cond::unlock()
{
	pthread_mutex_unlock(&mutex);
}

