#ifndef THREAD_H_
#define THREAD_H_

#include <pthread.h>

/**
 * \brief Mutex class
 *
 * Implements pthread mutex functionality.
 */
class Mutex
{
public:

	Mutex();
	virtual ~Mutex();

	bool lock();
	void unlock();

private:

	pthread_mutex_t mutex;
};

/**
 * \brief Thread class
 *
 * Implements pthread thread functionality.
 */
class Thread
{
public:
	Thread();
	virtual ~Thread();

	static bool sleep(int msec);

	bool start();
	bool stop(int timeOut = 10000, bool interrupt = false);
	bool terminate();
	bool interrupt();

	void requestStop();

protected:

	virtual void run();
	static void* threadRun(void *pParam);

	bool stopRequested();
	void reportStopped();

	/** Thread instance handle */
	pthread_t threadHandle;

private:

	static void signalHandler(int sig);

	bool stopRequest;
	bool stopComplete;
	Mutex mutex;
};

/**
 * \brief Semaphore class
 *
 * Implements a semaphore based on pthread mutex and conditional.
 */
class Semaphore
{
public:

	Semaphore();
	virtual ~Semaphore();

	void post();
	bool pend();
	bool pend(int ms);

private:

	pthread_mutex_t mutex;
	pthread_cond_t cond;
};

/**
 * \brief Conditional variable class
 *
 * Wraps a pthread mutex and conditional, use for advanced cases of conditional wait.
 */
class Cond
{
public:

	Cond();
	virtual ~Cond();

	bool lock();
	void unlock();

	void signal();
	bool wait();
	bool wait(int ms);

private:

	pthread_mutex_t mutex;
	pthread_cond_t cond;
};


#endif /*THREAD_H_*/
