/***************************************************************************
 *   Copyright            : (c) 2008-2009 Artec Design LLC                 *
 *                        : (c) 2008-2009 Andrei Birjukov                  *
 *                        : (c) 2008-2009 Anti Sullin                      *
 *                        : (c) 2008-2009 Jaanus Sepp                      *
 *   Contact              : andrei.birjukov at artecdesign dot ee          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

/*
 * WlanConnection.cpp
 *
 *  Created on: Nov 19, 2008
 *      Author: andrei
 *..Modified on: 01.10.2015
 *      Modified by: Silver Kits <silver.kits@vecly.com>
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <math.h>
#include <signal.h>
#include <fstream>

#include <platform.h>

//#define DEBUG
#include "Debug.h"

#include "WlanConnection.h"

#define WLAN_MODULE_NAME       "rt3070sta"
#define WLAN_MODULE_PATH       "/lib/modules/wireless/" WLAN_MODULE_NAME ".ko"
#define USB_MODULE_NAME        "ohci-hcd"

#define QUOTEME_(x) #x
#define QUOTEME(x) QUOTEME_(x)

#define WLAN_INTERFACE          "ra0"
#define WLAN_IFUP_TIMEOUT       5
#define WLAN_CONNECT_TIMEOUT    10
#define WLAN_DHCP_TIMEOUT       10
#define DHCP_CLIENT_PID         "/tmp/udhcpc.pid"
#define DHCP_CLIENT_CMD         "udhcpc -nR -t " QUOTEME(WLAN_DHCP_TIMEOUT) " -p " DHCP_CLIENT_PID " -i " WLAN_INTERFACE

#define WPA_SUPPL_CONF          "/data/wpa_supplicant.conf"
#define WPA_SUPPL_PID           "/tmp/wpa_supplicant.pid"
#define WPA_SUPPL_CONTROL       "/tmp/wpa_supplicant.ctrl"
#define WPA_SUPPL_CMD_APP       "wpa_supplicant -B -Dwext -i" WLAN_INTERFACE " -P" WPA_SUPPL_PID " -C" WPA_SUPPL_CONTROL
#define WPA_SUPPL_CMD_CONF      "wpa_supplicant -B -Dwext -i" WLAN_INTERFACE " -P" WPA_SUPPL_PID " -C" WPA_SUPPL_CONTROL " -c" WPA_SUPPL_CONF

#define IP_SCRIPT_CMD           "/usr/share/udhcpc/default.script bound"

// following is a list of functions re-used from wpa_ctrl.c of wpa_supplicant under BSD license

struct WPA_CONTROL * wpa_ctrl_open(const char *ctrl_path, const char *iface)
{
    struct WPA_CONTROL *ctrl;
    static int counter = 0;
    int ret;
    int tries = 0;

    ctrl = (WPA_CONTROL*) malloc(sizeof(*ctrl));
    if (ctrl == NULL)
        return NULL;
    memset(ctrl, 0, sizeof(*ctrl));

    ctrl->s = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (ctrl->s == -1) {
        free(ctrl);
        return NULL;
    }

    ctrl->local.sun_family = AF_UNIX;
    counter++;
try_again:
    ret = snprintf(ctrl->local.sun_path, sizeof(ctrl->local.sun_path),
              "/tmp/wpa_ctrl_%d-%d", getpid(), counter);
    if (ret < 0 || (size_t) ret >= sizeof(ctrl->local.sun_path)) {
        close(ctrl->s);
        free(ctrl);
        return NULL;
    }
    tries++;
    if (bind(ctrl->s, (struct sockaddr *) &ctrl->local,
            sizeof(ctrl->local)) < 0) {
        if (errno == EADDRINUSE && tries < 2) {
            /*
             * getpid() returns unique identifier for this instance
             * of wpa_ctrl, so the existing socket file must have
             * been left by unclean termination of an earlier run.
             * Remove the file and try again.
             */
            unlink(ctrl->local.sun_path);
            goto try_again;
        }
        close(ctrl->s);
        free(ctrl);
        return NULL;
    }

    ctrl->dest.sun_family = AF_UNIX;
    snprintf(ctrl->dest.sun_path, sizeof(ctrl->dest.sun_path),
            "%s/%s", ctrl_path, iface);

    if (connect(ctrl->s, (struct sockaddr *) &ctrl->dest,
            sizeof(ctrl->dest)) < 0) {
        close(ctrl->s);
        unlink(ctrl->local.sun_path);
        free(ctrl);
        return NULL;
    }

    return ctrl;
}

void wpa_ctrl_close(struct WPA_CONTROL *ctrl)
{
    unlink(ctrl->local.sun_path);
    close(ctrl->s);
    free(ctrl);
}

int wpa_ctrl_request(struct WPA_CONTROL *ctrl, const char *cmd, size_t cmd_len,
             char *reply, size_t *reply_len,
             void (*msg_cb)(char *msg, size_t len))
{
    struct timeval tv;
    int res;
    fd_set rfds;
    const char *_cmd;
    char *cmd_buf = NULL;
    size_t _cmd_len;

    _cmd = cmd;
    _cmd_len = cmd_len;

    if (send(ctrl->s, _cmd, _cmd_len, 0) < 0) {
        free(cmd_buf);
        return -1;
    }
    free(cmd_buf);

    for (;;) {
        tv.tv_sec = 2;
        tv.tv_usec = 0;
        FD_ZERO(&rfds);
        FD_SET(ctrl->s, &rfds);
        res = select(ctrl->s + 1, &rfds, NULL, NULL, &tv);
        if (FD_ISSET(ctrl->s, &rfds)) {
            res = recv(ctrl->s, reply, *reply_len, 0);
            if (res < 0)
                return res;
            if (res > 0 && reply[0] == '<') {
                /* This is an unsolicited message from
                 * wpa_supplicant, not the reply to the
                 * request. Use msg_cb to report this to the
                 * caller. */
                if (msg_cb) {
                    /* Make sure the message is nul
                     * terminated. */
                    if ((size_t) res == *reply_len)
                        res = (*reply_len) - 1;
                    reply[res] = '\0';
                    msg_cb(reply, res);
                }
                continue;
            }
            *reply_len = res;
            break;
        } else {
            return -2;
        }
    }
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class constructor
 *
 * Initializes the WLAN connection object.
 *
 * \param pSettings Points to a structure containing the WLAN configuration
 * parameters. The data is copied to a local storage.
 */
WlanConnection::WlanConnection(WLAN_CONNECTION_SETTINGS *pSettings)
                : NetworkConnection(COMM_MODULE_WLAN)
{
    cfg = *pSettings;
    ifSocket = -1;
    wpaControl = NULL;
    lastError = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Class destructor
 *
 * Terminates the WLAN connection object.
 */
WlanConnection::~WlanConnection()
{
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Network initialization routine
 * Initializes the network connection
 */
bool WlanConnection::netInitialize()
{
    PowerManager *pPM = PowerManager::getPowerManager();
    pPM->mpciEnable3V3(true);
    pPM->mpciEnableRadio(true);

    setStatus("Loading 802.11 WLAN driver");
    if (!loadWlanModule())
    {
        debug("failed loading wlan module\n");
        reportError("Failed initializing 802.11 WLAN driver");
        return false;
    }

    ifSocket = socket(AF_INET, SOCK_STREAM, 0);
    if (ifSocket == -1)
    {
        debug("failed creating socket\n");
        reportError("Failed creating WLAN socket");
        return false;
    }

    for (int i=0; i<(WLAN_IFUP_TIMEOUT*4); i++) {
        Thread::sleep(250);

        ifreq ifr;
        memset(&ifr, 0, sizeof(ifreq));
        if (ifaceRequest(SIOCGIFFLAGS, &ifr)) return true;
    }

    return false;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Network shutdown routine
 * Shuts down the network connection
 *
 * \return Returns true if the network connection was shut down successfully.
 */
bool WlanConnection::netShutdown()
{
    close(ifSocket);
    ifSocket = -1;

    if (!unloadWlanModule())
    {
        debug("failed removing wlan module\n");
        return false;
    }

    PowerManager *pPM = PowerManager::getPowerManager();
    pPM->mpciEnableRadio(false);
    pPM->mpciEnable3V3(false);

    return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Establish network connection
 * Establish connection to the network
 *
 * \return Returns true if network connection was established successfully.
 */
bool WlanConnection::netConnect()
{
    ifreq ifr;
    memset(&ifr, 0, sizeof(ifreq));

    setStatus("Starting 802.11 WLAN module");
    if (!ifaceRequest(SIOCGIFFLAGS, &ifr))
    {
        debug("failed reading interface flags\n");
        reportError("Failed preparing to enable network interface");
        return false;
    }

    debug("enabling interface: %s\n", WLAN_INTERFACE);
    ifr.ifr_flags |= IFF_UP;    // interface up

    if (!ifaceRequest(SIOCSIFFLAGS, &ifr))
    {
        debug("failed writing interface flags\n");
        reportError("Failed enabling network interface");
        return false;
    }

    setStatus("Starting WPA supplicant");
    const char *cmd;
    if (cfg.securityMode == WLAN_SECURITY_WPASUPPLICANT)
        cmd = WPA_SUPPL_CMD_CONF;
    else
        cmd = WPA_SUPPL_CMD_APP;
    debug("starting wpa_supplicant: %s\n", cmd);
    if (system(cmd) != 0)
    {
        debug("failed starting wpa_supplicant\n");
        reportError("Failed initializing WPA supplicant");
        netDisconnect();
        return false;
    }

    debug("opening wpa control if\n");
    wpaControl = wpa_ctrl_open(WPA_SUPPL_CONTROL, WLAN_INTERFACE);
    if (wpaControl == NULL)
    {
        debug("failed opening wpa control if\n");
        reportError("Failed to communicate with WPA supplicant");
        netDisconnect();
        return false;
    }

    setStatus("Configuring network");
    debug("configuring network\n");
    if (!configureNetwork())
    {
        debug("failed configuring network\n");
        reportError("Failed to configure network settings");
        netDisconnect();
        return false;
    }

    setStatus("Connecting to network");
    int cnt;
    for (cnt = 0; cnt<WLAN_CONNECT_TIMEOUT; cnt++) {
        if (isConnectionUp()) break;
        Thread::sleep(1000);
    }
    if (!isConnectionUp()) {
        debug("failed connecting to network\n");
        reportError("Failed connecting to network");
        netDisconnect();
        return false;
    }

    if (cfg.ipAddr.s_addr != 0)
    {
        setStatus("Setting IP address");
        if (!setStaticIp())
        {
            debug("failed setting IP address\n");
            reportError("Failed setting IP address");
            netDisconnect();
            return false;
        }
    } else {
        setStatus("Waiting for DHCP IP");
        debug("starting dhcp client: %s\n", DHCP_CLIENT_CMD);
        if (system(DHCP_CLIENT_CMD) != 0)
        {
            debug("failed starting dhcp client\n");
            reportError("Failed obtaining DHCP IP lease");
            netDisconnect();
            return false;
        }
    }

    setStatus("Network is connected");
    return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 * \brief Disconnect from network
 * Disconnects the from the network
 *
 * \return Returns true if network was disconnected successfully.
 */
bool WlanConnection::netDisconnect()
{
    std::ifstream dhcpPidFile(DHCP_CLIENT_PID);
    if (dhcpPidFile.is_open())
    {
        int pid = 0;
        dhcpPidFile >> pid;
        dhcpPidFile.close();

        setStatus("Stopping DHCP");
        debug("stopping dhcp client process %d\n", pid);
        if (kill(pid, SIGTERM) != 0) {
            debug("failed stopping dhcp client\n");
        }
        else {
            for (int i=0; i<20; i++) {
                Thread::sleep(250);
                if (kill(pid, 0) != 0) break;
            }
        }
    }

    if (wpaControl != NULL)
    {
        debug("closing wpa control if\n");
        wpa_ctrl_close(wpaControl);
        wpaControl = NULL;
    }

    std::ifstream wpaPidFile(WPA_SUPPL_PID);
    if (wpaPidFile.is_open())
    {
        int pid = 0;
        wpaPidFile >> pid;
        wpaPidFile.close();

        debug("stopping wpa_supplicant process %d\n", pid);
        if (kill(pid, SIGTERM) != 0) {
            debug("failed stopping wpa_supplicant\n");
        }
        else {
            for (int i=0; i<20; i++) {
                Thread::sleep(250);
                if (kill(pid, 0) != 0) break;
            }
        }
    }

    setStatus("Stopping WLAN module");
    debug("bringing interface down\n");

    ifreq ifr;
    memset(&ifr, 0, sizeof(ifreq));

    if (!ifaceRequest(SIOCGIFFLAGS, &ifr))
    {
        debug("failed reading interface flags\n");
        reportError("Failed preparing to disable networking");
        return false;
    }

    ifr.ifr_flags &= ~IFF_UP;    // interface down

    if (!ifaceRequest(SIOCSIFFLAGS, &ifr))
    {
        debug("failed writing interface flags\n");
        reportError("Failed disabling networking");
        return false;
    }

    setStatus("Network is shut down");
    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::netSuspend()
{
    if (!netDisconnect())
    {
        debug("disconnect failed, aborting\n");
        reportError("Failed disconnecting from network for suspend");
        return false;
    }

    if (!unloadWlanModule())
    {
        debug("failed removing wlan module\n");
        reportError("Failed stopping WLAN module");
        return false;
    }

    PowerManager *pPM = PowerManager::getPowerManager();
    pPM->mpciEnable3V3(false);
    pPM->mpciEnableRadio(false);
    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::netResume()
{
    PowerManager *pPM = PowerManager::getPowerManager();
    pPM->mpciEnable3V3(true);
    pPM->mpciEnableRadio(true);

    if (!loadWlanModule())
    {
        debug("failed inserting wlan module\n");
        reportError("Failed re-loading WLAN module on resume");
        return false;
    }

    if (!netConnect())
    {
        debug("connect failed, aborting\n");
        reportError("Failed re-connecting to WLAN network on resume");
        return false;
    }
    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::isConnectionUp()
{
    ifreq ifr;
    memset(&ifr, 0, sizeof(ifreq));

    if (!ifaceRequest(SIOCGIFFLAGS, &ifr))
    {
        debug("failed reading interface flags\n");
        return false;
    }

    debug("flags %x\n", ifr.ifr_flags);
    if (!(ifr.ifr_flags & IFF_UP))
    {
        debug("interface is not up\n");
        return false;
    }

    std::string ssid = ifaceGetEssid();
    return matchSSID(ssid.c_str());
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::getNetworkInfo(GENERIC_NETWORK_INFO &netInfo)
{
    iwreq iwr;

    netInfo.networkName = ifaceGetEssid();
    netInfo.isConnected = matchSSID(netInfo.networkName.c_str());

    if (!ifaceWirelessRequest(SIOCGIWFREQ, &iwr))
    {
        debug("failed reading frequency\n");
        return false;
    }

    float freq = float(iwr.u.freq.m) * pow(10, iwr.u.freq.e) / 1000000;

    int wlanFreqInt = (int)round(freq);

    //2 is subtracted from next WLAN channel centre frequency to cover possible error cases.
    //Ch. 2 (centre freq. 2417 MHz) is set when freq. is bigger than 2415 MHz and smaller than 2420 MHz.
    //For channel 14 it is -6 because channel spacing between 13 and 14 is bigger.

    if (wlanFreqInt < 2412-2)
        netInfo.networkChannel = 0;
    else if (wlanFreqInt < 2417-2)
        netInfo.networkChannel = 1;
    else if (wlanFreqInt < 2422-2)
        netInfo.networkChannel = 2;
    else if (wlanFreqInt < 2427-2)
        netInfo.networkChannel = 3;
    else if (wlanFreqInt < 2432-2)
        netInfo.networkChannel = 4;
    else if (wlanFreqInt < 2437-2)
        netInfo.networkChannel = 5;
    else if (wlanFreqInt < 2442-2)
        netInfo.networkChannel = 6;
    else if (wlanFreqInt < 2447-2)
        netInfo.networkChannel = 7;
    else if (wlanFreqInt < 2452-2)
        netInfo.networkChannel = 8;
    else if (wlanFreqInt < 2457-2)
        netInfo.networkChannel = 9;
    else if (wlanFreqInt < 2462-2)
        netInfo.networkChannel = 10;
    else if (wlanFreqInt < 2467-2)
        netInfo.networkChannel = 11;
    else if (wlanFreqInt < 2472-2)
        netInfo.networkChannel = 12;
    else if (wlanFreqInt < 2484-6)
        netInfo.networkChannel = 13;
    else if (wlanFreqInt < 2484+6)
        netInfo.networkChannel = 14;
    else
        netInfo.networkChannel = 0;

    debug("net [%s](%d) frequency %i MHz channel %d\n", netInfo.networkName.c_str(),
            netInfo.isConnected, wlanFreqInt, netInfo.networkChannel);

    iw_statistics stats = { 0 };
    iwr.u.essid.pointer = (caddr_t) &stats;
    iwr.u.essid.length = sizeof(stats);
    iwr.u.essid.flags = 1;

    if (!ifaceWirelessRequest(SIOCGIWSTATS, &iwr))
    {
        debug("failed reading stats\n");
        return false;
    }

    signed char level = (signed char)stats.qual.level;
    if (level > -57)
        netInfo.signalStrength = 99;
    else if (level > -67)
        netInfo.signalStrength = 79;
    else if (level > -71)
        netInfo.signalStrength = 59;
    else if (level > -81)
        netInfo.signalStrength = 39;
    else if (level > -90)
        netInfo.signalStrength = 19;
    else
        netInfo.signalStrength = 0;

    debug("stats: level %d, noise %d, quality %d\n",
            stats.qual.level, stats.qual.noise, netInfo.signalStrength);

    char ipAddr[16]="", macAddr[18]="";
    if (getInterfaceAddresses(WLAN_INTERFACE, ipAddr, macAddr))
    {
        netInfo.networkAddress = ipAddr;
        netInfo.hardwareAddress = macAddr;
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::ifaceRequest(int type, ifreq *pIfr)
{
    if (ifSocket < 0) return false;

    strncpy(pIfr->ifr_name, WLAN_INTERFACE, IFNAMSIZ);

    if (ioctl(ifSocket, type, (char*)pIfr) != 0)
    {
        debug("interface request %d failed (%d)\n", type, errno);
        return false;
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::ifaceWirelessRequest(int type, iwreq *pIwr)
{
    if (ifSocket < 0) return false;

    strncpy(pIwr->ifr_name, WLAN_INTERFACE, IFNAMSIZ);

    if (ioctl(ifSocket, type, (char*)pIwr) != 0)
    {
        debug("interface request %d failed (%d)\n", type, errno);
        return false;
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
std::string WlanConnection::ifaceGetEssid()
{
    iwreq iwr;
    char buffer[64];
    iwr.u.essid.pointer = (caddr_t) buffer;
    iwr.u.essid.length = sizeof(buffer);
    iwr.u.essid.flags = 0;

    if (!ifaceWirelessRequest(SIOCGIWESSID, &iwr))
    {
        debug("failed getting essid\n");
        return "";
    }

    buffer[iwr.u.essid.length] = 0;
    return buffer;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::isModuleLoaded()
{
    char tmp[256];

    FILE *pfModules = fopen("/proc/modules", "r");
    if (!pfModules) {
        debug("Error reading /proc/modules");
        return false;
    }

    while (fgets(tmp, sizeof(tmp), pfModules))
    {
        if (!strncmp(tmp, WLAN_MODULE_NAME " ", sizeof(WLAN_MODULE_NAME)))
        {
            fclose(pfModules);
            return true;
        }
    }

    fclose(pfModules);
    return false;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::loadWlanModule()
{
    debug("inserting wlan modules\n");
    if (isModuleLoaded()) return true;

    if (system("modprobe " USB_MODULE_NAME) != 0)
    {
        debug("failed inserting usb module\n");
        return false;
    }

    if (system("insmod " WLAN_MODULE_PATH) != 0)
    {
        debug("failed inserting wlan module\n");
        return false;
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::unloadWlanModule()
{
    debug("removing wlan modules\n");
    if (!isModuleLoaded()) return true;

    if (system("rmmod " WLAN_MODULE_NAME) != 0)
    {
        debug("failed removing wlan module\n");
        return false;
    }

    if (system("rmmod " USB_MODULE_NAME) != 0)
    {
        debug("failed removing usb module\n");
        return false;
    }
    return true;

}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::configureNetwork()
{
    char wepKey[32];
    snprintf(wepKey, sizeof(wepKey), "%d", cfg.defaultKey);
    wepKey[sizeof(wepKey)-1] = 0;

    struct WLAN_PARAMETERS
    {
        const char* param;
        const char* value;
    };
#define MAX_WLAN_PARAMETERS 10
    WLAN_PARAMETERS netParams[MAX_WLAN_PARAMETERS];
    int p=0;

    netParams[p].param = "ssid";
    netParams[p++].value = cfg.netSsid.c_str();

    switch (cfg.securityMode) {
    case WLAN_SECURITY_NONE:
        netParams[p].param = "key_mgmt";
        netParams[p++].value = "NONE";
        netParams[p].param = "auth_alg";
        netParams[p++].value = "OPEN";
        break;
    case WLAN_SECURITY_WEP:
        netParams[p].param = "key_mgmt";
        netParams[p++].value = "NONE";
        netParams[p].param = "auth_alg";
        netParams[p++].value = "SHARED";
        netParams[p].param = "wep_key0";
        netParams[p++].value = cfg.wepKeys[0].c_str();
        netParams[p].param = "wep_key1";
        netParams[p++].value = cfg.wepKeys[1].c_str();
        netParams[p].param = "wep_key2";
        netParams[p++].value = cfg.wepKeys[2].c_str();
        netParams[p].param = "wep_key3";
        netParams[p++].value = cfg.wepKeys[3].c_str();
        netParams[p].param = "wep_tx_keyidx";
        netParams[p++].value = wepKey;
        break;
    case WLAN_SECURITY_WPAPSK:
        netParams[p].param = "proto";
        netParams[p++].value = "WPA";
        netParams[p].param = "key_mgmt";
        netParams[p++].value = "WPA-PSK";
        netParams[p].param = "auth_alg";
        netParams[p++].value = "OPEN";
        netParams[p].param = "psk";
        netParams[p++].value = cfg.wpaKey.c_str();
        break;
    case WLAN_SECURITY_WPA2PSK:
        netParams[p].param = "proto";
        netParams[p++].value = "WPA2";
        netParams[p].param = "key_mgmt";
        netParams[p++].value = "WPA-PSK";
        netParams[p].param = "auth_alg";
        netParams[p++].value = "OPEN";
        netParams[p].param = "psk";
        netParams[p++].value = cfg.wpaKey.c_str();
        break;

    case WLAN_SECURITY_WPASUPPLICANT:
        // supplicant already configured from command-line
        return true;
    }

    std::string wpaReply;
    if (!wpaControlRequest("ADD_NETWORK", wpaReply)) {
        debug("failed adding network\n");
        return false;
    }

    int netId = atoi(wpaReply.c_str());
    debug("using network ID %d\n", netId);

    for (int i = 0; i < p; i ++)
    {
        const char *param = netParams[i].param;
        const char *value = netParams[i].value;

        if (strlen(value) == 0) continue;

        debug("setting parameter '%s' to '%s'\n", param, value);
        if (!wpaSetNetworkParameter(netId, param, value))
        {
            debug("failed setting %s\n", param);
            return false;
        }
    }

    if (!wpaControlRequest("ENABLE_NETWORK 0", wpaReply)) {
        debug("failed enabling network\n");
        return false;
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::wpaSetNetworkParameter(int netId,
        const char* pParam, const char* pValue)
{
    char str[256];
    snprintf(str, sizeof(str)-1, "SET_NETWORK %d %s %s", netId, pParam, pValue);

    std::string reply;
    bool ret = wpaControlRequest(str, reply);
    if (!ret) return false;

    return (reply.compare(0, 2, "OK") == 0);
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::wpaControlRequest(const char* pCmd, std::string &reply)
{
    if (wpaControl == NULL) return false;
    char replyBuf[256] = "";
    size_t replyLen = sizeof(reply) - 1;

    debug("wpa_ctrl command: %s\n", pCmd);
    int ret = wpa_ctrl_request(wpaControl, pCmd, strlen(pCmd), replyBuf, &replyLen, NULL);
    replyBuf[replyLen - 1] = 0;
    debug("returned %d (%d)\n", ret, replyLen);

    reply = replyBuf;
    return ret == 0;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::matchSSID(const char* ssid)
{
    std::string confSsid = cfg.netSsid;
    size_t ssidLen = confSsid.length();
    if (ssidLen > 1 && confSsid[0] == '"' && confSsid[ssidLen - 1] == '"')
        confSsid = confSsid.substr(1, ssidLen - 2);

    return confSsid == ssid;
}

////////////////////////////////////////////////////////////////////////////////
//
static int unputenv(const char *e)
{
    char *d = strdup(e);
    if (d == NULL) return -1;
    char *eq = strchr(d, '=');
    if (eq != NULL) *eq = 0;
    int ret = unsetenv(d);
    free(d);
    return ret;
}

////////////////////////////////////////////////////////////////////////////////
//
bool WlanConnection::setStaticIp()
{
    char tmp[32];
    char *env[6];
    bool ret = true;

    asprintf(&env[0], "interface=%s", WLAN_INTERFACE);

    inet_ntop(AF_INET, &(cfg.ipAddr), tmp, sizeof(tmp));
    tmp[sizeof(tmp)-1] = 0;
    asprintf(&env[1], "ip=%s",tmp);

    inet_ntop(AF_INET, &(cfg.netMask), tmp, sizeof(tmp));
    tmp[sizeof(tmp)-1] = 0;
    asprintf(&env[2], "subnet=%s",tmp);

    inet_ntop(AF_INET, &(cfg.gateway), tmp, sizeof(tmp));
    tmp[sizeof(tmp)-1] = 0;
    asprintf(&env[3], "router=%s",tmp);

    inet_ntop(AF_INET, &(cfg.dnsAddr), tmp, sizeof(tmp));
    tmp[sizeof(tmp)-1] = 0;
    asprintf(&env[4], "dns=%s",tmp);

    env[5] = NULL;

    // add environment
    char **e = env;
    while (*e != NULL) {
        putenv(*e);
        e++;
    }

    // launch script
    if (system(IP_SCRIPT_CMD) != 0) ret = false;;

    // release added environment
    e = env;
    while (*e != NULL) {
        if (unputenv(*e) == 0) free(*e);
        e++;
    }

    return ret;
}
