/*
 * WlanConnection.h
 *
 *  Created on: Nov 19, 2008
 *      Author: andrei
 */

#ifndef WLANCONNECTION_H_
#define WLANCONNECTION_H_

#include <sys/socket.h>
#include <sys/un.h>
#include <linux/wireless.h>
#include <netinet/in.h>
#include <string>

#include "NetworkConnection.h"

/** Encryption and authentication types for WLAN network connection. */
enum WLAN_SECURITY
{
	/** Open network, no authentication/encryption */
	WLAN_SECURITY_NONE=0,
	/** Legacy WEP security */
	WLAN_SECURITY_WEP,
	/** WPA-PSK authentication, TKIP encryption */
	WLAN_SECURITY_WPAPSK,
	/** WPA2-PSK authentication, AES encryption */
	WLAN_SECURITY_WPA2PSK,
	/** Use /data/wpasupplicant.conf for configuration */
	WLAN_SECURITY_WPASUPPLICANT
};
#define WLAN_SECURITY_COUNT 5
#define WLAN_SECURITY_NAMES1 { "NONE", "WEP", "WPA-PSK", "WPA2-PSK", "WPASUPPLICANT" }
#define WLAN_SECURITY_NAMES2 { "OPEN", "WEP", "WPA", "WPA2", "SUPPLICANT" }

/** Parameters required to connect to a specific WLAN. */
struct WLAN_CONNECTION_SETTINGS
{
	/** SSID, network name */
	std::string			netSsid;
	/** Network authentication type */
	int /*WLAN_SECURITY*/		securityMode;
	/** WEP network keys. */
	std::string			wepKeys[4];
	/** Default encryption key (0-3 in case of WEP) */
	int					defaultKey;
	/** WPA-PSK / WPA2-PSK network key. */
	std::string			wpaKey;
	/** IP address, DHCP if null; other network settings*/
	struct in_addr		ipAddr;
	/** Network mask */
	struct in_addr		netMask;
	/** Gateway address */
	struct in_addr		gateway;
	/** DNS server address */
	struct in_addr		dnsAddr;
};


/**
 * \brief Internal structure for control interface library
 *
 * This structure is used by the wpa_supplicant/hostapd control interface
 * library to store internal data. Programs using the library should not touch
 * this data directly. They can only use the pointer to the data structure as
 * an identifier for the control interface connection and use this as one of
 * the arguments for most of the control interface library functions.
 */
struct WPA_CONTROL {
	/** Handle to the control socket */
	int s;
	/** Local address */
	struct sockaddr_un local;
	/** Remote address */
	struct sockaddr_un dest;
};

/**
 * \brief WLAN connection class
 *
 * Implements the WLAN network connection.
 * Currently supported are the Ralink RT2870 and RT3070 WLAN modules with
 * Ralink drivers.
 */
class WlanConnection: public NetworkConnection
{
public:

	WlanConnection(WLAN_CONNECTION_SETTINGS *pSettings);
	virtual ~WlanConnection();

	virtual bool isConnectionUp();
	virtual bool getNetworkInfo(GENERIC_NETWORK_INFO &netInfo);

	virtual const char* getLastError() { return lastError; }

protected:

	virtual bool netInitialize();
	virtual bool netShutdown();
	virtual bool netConnect();
	virtual bool netDisconnect();
	virtual bool netSuspend();
	virtual bool netResume();

private:

	bool ifaceRequest(int type, ifreq *pIfr);
	bool ifaceWirelessRequest(int type, iwreq *pIwr);

	bool configureNetwork();
	bool wpaSetNetworkParameter(int netId, const char* pParam, const char* pValue);
	bool wpaControlRequest(const char* pCmd, std::string &reply);
	bool setStaticIp();

	std::string ifaceGetEssid();
	bool matchSSID(const char* ssid);

	bool isModuleLoaded();
	bool loadWlanModule();
	bool unloadWlanModule();

	WLAN_CONNECTION_SETTINGS	cfg;
	int							ifSocket;
	const char*					lastError;
	WPA_CONTROL*				wpaControl;
};

#endif /* WLANCONNECTION_H_ */
