/*
 * adc081s021.c
 *
 * Copyright (C) 2008 Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * Based on lm75.c by:
 * Copyright (C) 2006 Kaiwan N Billimoria <kaiwan@designergraphix.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//#define DEBUG

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/err.h>
#include <linux/sysfs.h>
#include <linux/hwmon.h>
#include <linux/mutex.h>
#include <linux/spi/spi.h>
#include <linux/delay.h>
#include <asm/gpio.h>

#define DRVNAME		"adc081s021"


// Voltage divider
#define R1			150
#define R2			196

#define VMAX		(3300 * (R1+R2) / R2)

#define PWR_DELAY_MS	10

struct adc081s021 {
	struct device *hwmon_dev;
	struct mutex lock;
	unsigned int pwr_pin;
};

/* sysfs hook function */
static ssize_t adc081s021_sense_vbat(struct device *dev,
		struct device_attribute *attr, char *buf)
{
	struct spi_device *spi = to_spi_device(dev);
	int status, val, raw;
	u8 rxbuf[2];
	struct adc081s021 *p_adc081s021 = dev_get_drvdata(&spi->dev);

	if (mutex_lock_interruptible(&p_adc081s021->lock))
		return -ERESTARTSYS;

	at91_set_gpio_output(p_adc081s021->pwr_pin, 0);
	msleep(PWR_DELAY_MS);

	/*
	 * spi_read() requires a DMA-safe buffer; so we use
	 * spi_write_then_read(), transmitting 0 bytes.
	 */
	status = spi_write_then_read(spi, NULL, 0, &rxbuf[0], 2);
	if (status < 0) {
		printk(KERN_WARNING
		"spi_write_then_read failed with status %d\n", status);
		goto out;
	}
	dev_dbg(dev, "rxbuf[1] : 0x%x rxbuf[0] : 0x%x\n", rxbuf[1], rxbuf[0]);

	at91_set_gpio_output(p_adc081s021->pwr_pin, 1);

	raw = ((rxbuf[0] << 2) | (rxbuf[1] >> 6)) & 0xFF;
	dev_dbg(dev, "raw=0x%x\n", raw);

	val = (raw * VMAX) / 255;
	status = sprintf(buf, "%d\n", val); /* millivolts */
out:
	mutex_unlock(&p_adc081s021->lock);
	return status;
}

static DEVICE_ATTR(in_input0, S_IRUGO, adc081s021_sense_vbat, NULL);

static ssize_t adc081s021_show_name(struct device *dev, struct device_attribute
			      *devattr, char *buf)
{
	return sprintf(buf, "adc081s021\n");
}

static DEVICE_ATTR(name, S_IRUGO, adc081s021_show_name, NULL);

/*----------------------------------------------------------------------*/

static int __devinit adc081s021_probe(struct spi_device *spi)
{
	struct adc081s021 *p_adc081s021;
	int status;

	spi->bits_per_word = 8;
	spi->mode = SPI_MODE_3;
	status = spi_setup(spi);
	if (status < 0)
		return status;

	p_adc081s021 = kzalloc(sizeof *p_adc081s021, GFP_KERNEL);
	if (!p_adc081s021)
		return -ENOMEM;

	p_adc081s021->pwr_pin = (unsigned int)spi->dev.platform_data;
	if (!p_adc081s021->pwr_pin) {
		status = -EINVAL;
		goto out_gpio_check_failed;
	}
	at91_set_gpio_output(p_adc081s021->pwr_pin, 1);

	mutex_init(&p_adc081s021->lock);

	/* sysfs hook */
	p_adc081s021->hwmon_dev = hwmon_device_register(&spi->dev);
	if (IS_ERR(p_adc081s021->hwmon_dev)) {
		dev_dbg(&spi->dev, "hwmon_device_register failed.\n");
		status = PTR_ERR(p_adc081s021->hwmon_dev);
		goto out_dev_reg_failed;
	}
	dev_set_drvdata(&spi->dev, p_adc081s021);

	if ((status = device_create_file(&spi->dev, &dev_attr_in_input0))
	 || (status = device_create_file(&spi->dev, &dev_attr_name))) {
		dev_dbg(&spi->dev, "device_create_file failure.\n");
		goto out_dev_create_file_failed;
	}

	return 0;

out_dev_create_file_failed:
	device_remove_file(&spi->dev, &dev_attr_in_input0);
	hwmon_device_unregister(p_adc081s021->hwmon_dev);
out_dev_reg_failed:
	dev_set_drvdata(&spi->dev, NULL);
out_gpio_check_failed:
	kfree(p_adc081s021);
	return status;
}

static int __devexit adc081s021_remove(struct spi_device *spi)
{
	struct adc081s021 *p_adc081s021 = dev_get_drvdata(&spi->dev);

	device_remove_file(&spi->dev, &dev_attr_in_input0);
	device_remove_file(&spi->dev, &dev_attr_name);
	hwmon_device_unregister(p_adc081s021->hwmon_dev);
	dev_set_drvdata(&spi->dev, NULL);
	kfree(p_adc081s021);

	return 0;
}

static struct spi_driver adc081s021_driver = {
	.driver = {
		.name	= "adc081s021",
		.owner	= THIS_MODULE,
	},
	.probe	= adc081s021_probe,
	.remove	= __devexit_p(adc081s021_remove),
};

static int __init init_adc081s021(void)
{
	return spi_register_driver(&adc081s021_driver);
}

static void __exit cleanup_adc081s021(void)
{
	spi_unregister_driver(&adc081s021_driver);
}

module_init(init_adc081s021);
module_exit(cleanup_adc081s021);

MODULE_AUTHOR("Anti Sullin");
MODULE_DESCRIPTION("adc081s021 driver");
MODULE_LICENSE("GPL");
