/*
 * Driver for keys on GPIO lines capable of generating interrupts.
 *
 * Copyright 2007 Jaanus Sepp <jaanus.sepp@artecdesign.ee>
 * Copyright 2007 Anti Sullin <anti.sullin@artecdesign.ee>
 * Copyright 2005 Phil Blundell
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//--------------------------------------------------------------------------------

//#define DEBUG

//--------------------------------------------------------------------------------

#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <linux/gpio_keyb.h>

#include <asm/gpio.h>

//--------------------------------------------------------------------------------

// debug macros

#define d_err(format, arg...)  \
  dev_err(gpio_keyb_state.dev, format, ## arg)

#define d_warn(format, arg...)  \
  dev_warn(gpio_keyb_state.dev, format, ## arg)

#define d_info(format, arg...)  \
  dev_info(gpio_keyb_state.dev, format, ## arg)

#define d_notice(format, arg...)  \
  dev_notice(gpio_keyb_state.dev, format, ## arg)

#define d_dbg(format, arg...)  \
  dev_dbg(gpio_keyb_state.dev, format, ## arg)

//--------------------------------------------------------------------------------

#define GPIO_KEYB_POLL_MS     50       // keyboard poll interval

// Global state variables for the driver

struct gpio_keyb {
  struct platform_device *pdev;
  struct device *dev;                  // for debug messages (dev_warn, etc)
  int suspended;
	struct timer_list	timer;             // for timer functions
} gpio_keyb_state;

/************************************************ FUNCTIONAL PART **************************************/

static void gpio_keyb_poll(unsigned long handle) {
  static unsigned long inp_last_state;   // for detecting input changes
  int i, j, state, keyseq;

  struct gpio_keyb_platform_data *pdata = gpio_keyb_state.dev->platform_data;
  struct input_dev *input = platform_get_drvdata(gpio_keyb_state.pdev);

  // process direct pin buttons
  for (i = 0; i < pdata->n_direct_buttons; i++) {
    int gpio = pdata->buttons[i].gpio_input;
    state = (gpio_get_value(gpio) != 0) ^ (pdata->buttons[i].active_low_input);
    if (state != ((inp_last_state & (1 << i)) != 0)) {
      input_report_key(input, pdata->buttons[i].keycode, state);
      if (state) d_dbg("PRESSED: %s\n", pdata->buttons[i].desc); else
                 d_dbg("RELEASED: %s\n", pdata->buttons[i].desc);
      inp_last_state ^= (1 << i);
      input_sync(input);
    }
  }

  // process matrix buttons

  // release all matrix output pins
  // Turn the idle output pins to inputs. Otherwize it would be possible to short
  // two output pins with different logical levels with specific key combinations.
  for (i = 0; i < pdata->n_matrix_rows; i++) {
    gpio_direction_input(pdata->rows[i].gpio_output);
  }

  // scan through the matrix
  for (i = 0; i < pdata->n_matrix_rows; i++) {
    gpio_direction_output(pdata->rows[i].gpio_output, !pdata->rows[i].active_low_output); // assert a row
    udelay(1);
    // scan through the columns
    for (j = 0; j < pdata->n_matrix_cols; j++) {
      state = (gpio_get_value(pdata->cols[j].gpio_input) != 0) ^ (pdata->cols[j].active_low_input);
      keyseq = pdata->n_direct_buttons + i * pdata->n_matrix_cols + j;
      if (state != ((inp_last_state & (1 << keyseq)) != 0)) {
        input_report_key(input, pdata->rows[i].buttons[j].keycode, state);
        if (state) d_dbg("PRESSED: %s\n", pdata->rows[i].buttons[j].desc); else
                   d_dbg("RELEASED: %s\n", pdata->rows[i].buttons[j].desc);
        inp_last_state ^= (1 << keyseq);
        input_sync(input);
      }
    }
    gpio_direction_input(pdata->rows[i].gpio_output);  // release the row
  }

  // re-assert all output pins to detect next change
  for (i = 0; i < pdata->n_matrix_rows; i++) {
    gpio_direction_output(pdata->rows[i].gpio_output, !pdata->rows[i].active_low_output);
  }

  if (!gpio_keyb_state.suspended) mod_timer(&gpio_keyb_state.timer, jiffies + msecs_to_jiffies(GPIO_KEYB_POLL_MS));
}

//--------------------------------------------------------------------------------
// we need this routine only for wake-up

static irqreturn_t gpio_keyb_isr(int irq, void *dev_id) {
  return IRQ_HANDLED;
}

//--------------------------------------------------------------------------------

static int __devinit gpio_keyb_probe(struct platform_device *pdev) {
  struct gpio_keyb_platform_data *pdata = pdev->dev.platform_data;
  struct input_dev *input;
  int i = 0;
  int j = 0;
  int irq, error;
  char buf[8];

  // Save the platform device
  gpio_keyb_state.pdev = pdev;
  if (gpio_keyb_state.pdev == NULL) return EINVAL;

  // Get device. We need it for the log messages.
  gpio_keyb_state.dev = &pdev->dev;
  // we can't output a log message since our dev is NULL
  if (gpio_keyb_state.dev == NULL) return EINVAL;

  // allocate input device structure
  input = input_allocate_device();
  if (!input) {
    d_err("Error allocating input device structure\n");
    return -ENOMEM;
  }

  // initialize input device structure
  platform_set_drvdata(pdev, input);

  input->evbit[0] = BIT(EV_KEY);

  input->name = pdev->name;
  input->phys = "gpio-keyb/input0";
  input->dev.parent = &pdev->dev;

  input->id.bustype = BUS_HOST;
  input->id.vendor = 0x0001;
  input->id.product = 0x0001;
  input->id.version = 0x0100;

  // initialize direct input keys
  for (i = 0; i < pdata->n_direct_buttons; i++) {
    gpio_direction_input(pdata->buttons[i].gpio_input);    // set pin as input
    sprintf(buf, "DIR_%02d", i);                           // interrupt name in case the key definition does not provide it
    irq = gpio_to_irq(pdata->buttons[i].gpio_input);
    set_irq_type(irq, IRQ_TYPE_EDGE_BOTH);
    // set up interrupt
    error = request_irq(irq, gpio_keyb_isr, IRQF_SAMPLE_RANDOM,
             pdata->buttons[i].desc ? pdata->buttons[i].desc : buf, pdev);
    if (error) {
      d_err("Unable to claim irq %d; error %d\n", irq, error);
      goto direct_fail;
    }
    disable_irq(irq);                                      // we need IRQ only for wake-up
    set_bit(pdata->buttons[i].keycode, input->keybit);
  }

  // initialize matrix input keys
  for (i = 0; i < pdata->n_matrix_cols; i++) {
    gpio_direction_input(pdata->cols[i].gpio_input);       // set column pin as input
    sprintf(buf, "COL_%02d", i);                           // name the interrupt by the column
    irq = gpio_to_irq(pdata->cols[i].gpio_input);
    set_irq_type(irq, IRQ_TYPE_EDGE_BOTH);
    error = request_irq(irq, gpio_keyb_isr, IRQF_SAMPLE_RANDOM, buf, pdev);
    if (error) {
      d_err("Unable to claim irq %d; error %d\n", irq, error);
      goto matrix_fail;
    }

    disable_irq(irq);                                      // we need IRQ only for wake-up

    for (j = 0; j < pdata->n_matrix_rows; j++) {
      set_bit(pdata->rows[j].buttons[i].keycode, input->keybit); // register key codes this device can generate
    }
  }

  // register the device
  error = input_register_device(input);
  if (error) {
    d_err("Unable to register the input device\n");
    goto reg_fail;
  }

  // set the matrix output pins (rows)
  for (j = 0; j < pdata->n_matrix_rows; j++) {
    gpio_direction_output(pdata->rows[j].gpio_output, !pdata->rows[j].active_low_output);
  }

  device_init_wakeup(&pdev->dev, 1);
  gpio_keyb_state.suspended = 0;
	init_timer(&gpio_keyb_state.timer);
	gpio_keyb_state.timer.function = gpio_keyb_poll;
  mod_timer(&gpio_keyb_state.timer, jiffies + msecs_to_jiffies(GPIO_KEYB_POLL_MS));
  return 0;

reg_fail:
matrix_fail:
  for (i = i - 1; i >= 0; i--) free_irq(gpio_to_irq(pdata->cols[i].gpio_input), pdev);
  i = pdata->n_direct_buttons;        // free all direct key interrupts

direct_fail:
  for (i = i - 1; i >= 0; i--) free_irq(gpio_to_irq(pdata->buttons[i].gpio_input), pdev);

  input_free_device(input);
  return error;
}

//--------------------------------------------------------------------------------

static int __devexit gpio_keyb_remove(struct platform_device *pdev) {
  struct gpio_keyb_platform_data *pdata = pdev->dev.platform_data;
  struct input_dev *input = platform_get_drvdata(pdev);
  int i;

  // free direct key interrupts
  for (i = 0; i < pdata->n_direct_buttons; i++) {
    free_irq(gpio_to_irq(pdata->buttons[i].gpio_input), pdev);
  }

  // free matrix column interrupts
  for (i = 0; i < pdata->n_matrix_cols; i++) {
    free_irq(gpio_to_irq(pdata->cols[i].gpio_input), pdev);
  }

  gpio_keyb_state.suspended = 1;

  device_init_wakeup(&pdev->dev, 0);
  input_unregister_device(input);
  return 0;
}

//--------------------------------------------------------------------------------

#ifdef CONFIG_PM

static int gpio_keyb_suspend(struct platform_device *pdev, pm_message_t state) {
  struct gpio_keyb_platform_data *pdata = pdev->dev.platform_data;
  int i, irq;

  if (device_may_wakeup(&pdev->dev)) {
    // allow the direct key interrupts to wake up the system
    for (i = 0; i < pdata->n_direct_buttons; i++) {
      irq = gpio_to_irq(pdata->buttons[i].gpio_input);
      enable_irq(irq);
      enable_irq_wake(irq);
    }

    // allow the matrix column interrupts to wake up the system
    for (i = 0; i < pdata->n_matrix_cols; i++) {
      irq = gpio_to_irq(pdata->cols[i].gpio_input);
      enable_irq(irq);
      enable_irq_wake(irq);
    }
  }

  gpio_keyb_state.suspended = 1;

  return 0;
}

//--------------------------------------------------------------------------------

static int gpio_keyb_resume(struct platform_device *pdev) {
  struct gpio_keyb_platform_data *pdata = pdev->dev.platform_data;
  int i, irq;

  // disallow the direct key interrupts to wake up the system
  for (i = 0; i < pdata->n_direct_buttons; i++) {
    irq = gpio_to_irq(pdata->buttons[i].gpio_input);
    disable_irq(irq);
    disable_irq_wake(irq);
  }

  // disallow the matrix column interrupts to wake up the system
  for (i = 0; i < pdata->n_matrix_cols; i++) {
    irq = gpio_to_irq(pdata->cols[i].gpio_input);
    disable_irq(irq);
    disable_irq_wake(irq);
  }

  gpio_keyb_state.suspended = 0;
  mod_timer(&gpio_keyb_state.timer, jiffies + msecs_to_jiffies(GPIO_KEYB_POLL_MS));
  return 0;
}

#endif

//--------------------------------------------------------------------------------

struct platform_driver gpio_keyb_device_driver = {
  .probe    = gpio_keyb_probe,
  .remove    = __devexit_p(gpio_keyb_remove),

  #ifdef CONFIG_PM
    .suspend  = gpio_keyb_suspend,
    .resume    = gpio_keyb_resume,
  #endif

  .driver    = {
    .name  = "gpio-keyb",
  }
};

//--------------------------------------------------------------------------------

static int __init gpio_keyb_init(void) {
  return platform_driver_register(&gpio_keyb_device_driver);
}

//--------------------------------------------------------------------------------

static void __exit gpio_keyb_exit(void) {
  platform_driver_unregister(&gpio_keyb_device_driver);
}

//--------------------------------------------------------------------------------

module_init(gpio_keyb_init);
module_exit(gpio_keyb_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Keyboard driver for CPU GPIOs");
