/*
 * isl29001.c
 *
 * Copyright (C) 2008 Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * Based on lm75.c
 *  Copyright (c) 1998, 1999  Frodo Looijaard <frodol@dds.nl>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/i2c.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/err.h>
#include <linux/mutex.h>
#include <linux/delay.h>


/* The ISL29001 registers */
#define ISL29001_CMD_PWRDN		0x8C
#define ISL29001_CMD_RESET		0x0C
#define ISL29001_CMD_READ_CH1	0x00
#define ISL29001_CMD_READ_CH2	0x04
#define ISL29001_CMD_TEST_COMM	0x80


/* The ISL29001 registers */
#define ISL29001_REG_VAL_LSB	0x00
#define ISL29001_REG_VAL_MSB	0x01

/* Functions */
static int isl29001_probe(struct i2c_client *client,
	       const struct i2c_device_id *id);
static int isl29001_remove(struct i2c_client *client);

/* Driver data */
static const struct i2c_device_id isl29001_id[] = {
	{ "isl29001", 0 },
	{ }
};

static struct i2c_driver isl29001_driver = {
	.class		= I2C_CLASS_HWMON,
	.driver = {
		.name	= "isl29001",
	},
	.probe		= isl29001_probe,
	.remove		= __devexit_p(isl29001_remove),
	.id_table	= isl29001_id,
};


/* Client data */
struct isl29001_data {
	struct i2c_client	client;
	struct device 		*hwmon_dev;
	struct mutex		lock;
};

int isl29001_read(struct i2c_client *client) {
	unsigned char msgbuf[2];
	struct i2c_msg msg = { client->addr, client->flags | I2C_M_RD, 2, msgbuf };

	if (i2c_transfer(client->adapter, &msg, 1) < 0)
			return -1;

	return msgbuf[0] | (msgbuf[1] << 8);
}

int isl29001_write(struct i2c_client *client, char cmd) {
	struct i2c_msg msg = { client->addr, client->flags, 1, &cmd };

	if (i2c_transfer(client->adapter, &msg, 1) < 0)
			return -1;

	return 0;
}


/* SENSOR 1 RAW READING */
static ssize_t show_illum1(struct device *dev, struct device_attribute *da,
			 char *buf)
{
	int val;
	struct i2c_client *client = to_i2c_client(dev);
	struct isl29001_data *data = i2c_get_clientdata(client);

	mutex_lock(&data->lock);
	isl29001_write(client, ISL29001_CMD_RESET);
	msleep(1);
	isl29001_write(client, ISL29001_CMD_READ_CH1);
	msleep(120);
	val = isl29001_read(client);
	isl29001_write(client, ISL29001_CMD_PWRDN);
	mutex_unlock(&data->lock);

	return sprintf(buf, "%d\n", val);
}

/* SENSOR 2 RAW READING */
static ssize_t show_illum2(struct device *dev, struct device_attribute *da,
			 char *buf)
{
	int val;
	struct i2c_client *client = to_i2c_client(dev);
	struct isl29001_data *data = i2c_get_clientdata(client);

	mutex_lock(&data->lock);
	isl29001_write(client, ISL29001_CMD_RESET);
	msleep(1);
	isl29001_write(client, ISL29001_CMD_READ_CH2);
	msleep(120);
	val = isl29001_read(client);
	isl29001_write(client, ISL29001_CMD_PWRDN);
	mutex_unlock(&data->lock);

	return sprintf(buf, "%d\n", val);
}

/* ILLUMINATION IN LUX */
static ssize_t show_illum(struct device *dev, struct device_attribute *da,
			 char *buf)
{
	int val1, val2, val;
	struct i2c_client *client = to_i2c_client(dev);
	struct isl29001_data *data = i2c_get_clientdata(client);

	mutex_lock(&data->lock);
	isl29001_write(client, ISL29001_CMD_RESET);
	msleep(1);
	isl29001_write(client, ISL29001_CMD_READ_CH1);
	msleep(120);
	val1 = isl29001_read(client);
	isl29001_write(client, ISL29001_CMD_RESET);
	isl29001_write(client, ISL29001_CMD_READ_CH2);
	msleep(120);
	val2 = isl29001_read(client);
	isl29001_write(client, ISL29001_CMD_PWRDN);
	mutex_unlock(&data->lock);

	// val = 1.85 * (val1 - 7.5 * val2)
	// 1.85 = 37/20
	// full scale : 10000 lux
	val = val1 - (7 * val2) - (val2/2);
	val = 37 * val / 20;
	val = 10000 * val / 32768;

	return sprintf(buf, "%d\n", val);
}



static SENSOR_DEVICE_ATTR(illum1_input, S_IRUGO, show_illum1, NULL, 0);
static SENSOR_DEVICE_ATTR(illum2_input, S_IRUGO, show_illum2, NULL, 0);
static SENSOR_DEVICE_ATTR(illum_input, S_IRUGO, show_illum, NULL, 0);

static struct attribute *isl29001_attributes[] = {
	&sensor_dev_attr_illum1_input.dev_attr.attr,
	&sensor_dev_attr_illum2_input.dev_attr.attr,
	&sensor_dev_attr_illum_input.dev_attr.attr,
	NULL
};

static const struct attribute_group isl29001_group = {
	.attrs = isl29001_attributes,
};

static int isl29001_probe(struct i2c_client *client,
			const struct i2c_device_id *dev_id)
{
	struct isl29001_data *data;
	int err = 0;

	if (!(data = kzalloc(sizeof(struct isl29001_data), GFP_KERNEL))) {
		err = -ENOMEM;
		goto exit;
	}

	i2c_set_clientdata(client, data);

	mutex_init(&data->lock);

	/* Register sysfs hooks */
	if ((err = sysfs_create_group(&client->dev.kobj, &isl29001_group)))
		goto exit_free;

	data->hwmon_dev = hwmon_device_register(&client->dev);
	if (IS_ERR(data->hwmon_dev)) {
		err = PTR_ERR(data->hwmon_dev);
		goto exit_remove;
	}

	printk("ISL29001 light intensity sensor\n");
	return 0;

exit_remove:
	sysfs_remove_group(&client->dev.kobj, &isl29001_group);
exit_free:
	kfree(data);
exit:
printk("isl29001_probe failed: error %d\n", err);
	return err;
}

static int __devexit isl29001_remove(struct i2c_client *client)
{
	struct isl29001_data *data = i2c_get_clientdata(client);
	hwmon_device_unregister(data->hwmon_dev);
	sysfs_remove_group(&client->dev.kobj, &isl29001_group);
	kfree(data);
	return 0;
}

static int __init sensors_isl29001_init(void)
{
	return i2c_add_driver(&isl29001_driver);
}

static void __exit sensors_isl29001_exit(void)
{
	i2c_del_driver(&isl29001_driver);
}

MODULE_AUTHOR("Anti Sullin <anti.sullin@artecdesign.ee>");
MODULE_DESCRIPTION("ISL29001 driver");
MODULE_LICENSE("GPL");

module_init(sensors_isl29001_init);
module_exit(sensors_isl29001_exit);
