/*
 * Magtek intellihead magstripe card reader driver.
 *
 * Copyright (C) 2007 Anti Sullin / Artec Design <anti.sullin@artecdesign.ee>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 * reader is connected to:
 *  data -> an interrupt pin
 *  data, clk -> synchronous serial port td, tk
 *  power -> gpio pin
 *
 *
 * reader pulls down data pin to indicate new data.
 * magstripe_magtek_new interrupt is generated, interrupt configures ssc controller
 * ssc controller receives 232 bytes of data, 32 bits at once with magstripe_magtek_io interrupt
 * the data is parsed into 7-bit array and checksum checked
 * after last interrupt, magstripe reader is powered down and magstripe_magtek_timer is set
 * on magstripe_magtek_timer, reader is re-powered and interrupts are enabled
 *
 */

#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/err.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/magstripe_magtek.h>

#include <asm/irq.h>
#include <asm/delay.h>

#include <asm/arch/at91_ssc.h>
#include <asm/arch/gpio.h>

// ENABLE DEBUGGING MESSAGES
// Caution: enabling any printk's in ssc interrupt handler will cause data loss

//#define DEBUG
//#define DEBUG_RAW_DATA

#define DUAL_DIRECTION_READ

#define SSC_PUT_CR(port,v)	writel(v, (port).membase + AT91_SSC_CR)
#define SSC_GET_CR(port)	readl((port).membase + AT91_SSC_CR)
#define SSC_PUT_CMR(port,v)	writel(v, (port).membase + AT91_SSC_CMR)
#define SSC_GET_CMR(port)	readl((port).membase + AT91_SSC_CMR)
#define SSC_PUT_RCMR(port,v)	writel(v, (port).membase + AT91_SSC_RCMR)
#define SSC_GET_RCMR(port)	readl((port).membase + AT91_SSC_RCMR)
#define SSC_PUT_RFMR(port,v)	writel(v, (port).membase + AT91_SSC_RFMR)
#define SSC_GET_RFMR(port)	readl((port).membase + AT91_SSC_RFMR)

#define SSC_PUT_RHR(port,v)	writel(v, (port).membase + AT91_SSC_RHR)
#define SSC_GET_RHR(port)	readl((port).membase + AT91_SSC_RHR)

#define SSC_GET_SR(port)	readl((port).membase + AT91_SSC_SR)

#define SSC_PUT_IER(port,v)	writel(v, (port).membase + AT91_SSC_IER)
#define SSC_GET_IER(port)	readl((port).membase + AT91_SSC_IER)
#define SSC_PUT_IDR(port,v)	writel(v, (port).membase + AT91_SSC_IDR)
#define SSC_GET_IDR(port)	readl((port).membase + AT91_SSC_IDR)
#define SSC_PUT_IMR(port,v)	writel(v, (port).membase + AT91_SSC_IMR)
#define SSC_GET_IMR(port)	readl((port).membase + AT91_SSC_IMR)

// major is always 10 for misc devices
#define MAGSTRIPE_MAGTEK_MINOR 50
// see also include/linux/miscdevice.h for number allocation

// clock speed= master clock 200MHz / clk_div
// magtek reader allows up to 1MHz so div>200!
#ifdef DEBUG
#define SSC_CLK_DIV 4000
#else
#define SSC_CLK_DIV 500
#endif

#ifndef AT91_SSC_CKO_DURING_TRANSFERS
#define	AT91_SSC_CKO_DURING_TRANSFERS		(2 << 2)
#endif

#define	MAG_RESET_TIME	msecs_to_jiffies(250)

//#ifndef DUAL_DIRECTION_READ
// data count and buffer size (in 4bytes)
//single direction read 1842/8 = 232 (round up to next 4-byte integer)
//#define RX_COUNT 232
//#define RX_BUF_SIZE 266
//#else
//double direction read 3666/8 = 460
#define RX_COUNT 460
// (RX_COUNT * 8 / 7) round up
#define RX_BUF_SIZE 526
//#endif



#define TRACK_LEN 86
#define FIRST_TRACK_BEGIN 2

int track_begin[] = {
	FIRST_TRACK_BEGIN, //2, 432
	FIRST_TRACK_BEGIN + TRACK_LEN, //88, 346
	FIRST_TRACK_BEGIN + 2 * TRACK_LEN //174, 260
};

int track_begin_backwards[] = {
	FIRST_TRACK_BEGIN + 5 * TRACK_LEN, //2, 432
	FIRST_TRACK_BEGIN + 4 * TRACK_LEN, //88, 346
	FIRST_TRACK_BEGIN + 3 * TRACK_LEN //174, 260
};



static struct magstripe_magtek {
	int 	ssc_irq;
	int 	reader_irq;
	void __iomem 	*membase;

	char 	*rxbuf;
	char	*rxbuf_w;
	int 	data_shift;
	unsigned int	rx_count;
	volatile char 	rx_state; //0x01 track0, 0x02 track1, 0x04 track2 | 0x00 before track, 0x10 on track, 0x20 after track
	char	checksum;

	char	data_rdy; // 0x01 track0_rdy; 0x02 track1_rdy; 0x04 track2_rdy

	struct timer_list	timer;

	char state;
	char dcnt; // debounce counter

	int pwr_pin;
	int data_pin;

	spinlock_t rd_lock;
} info;

static DECLARE_WAIT_QUEUE_HEAD(read_wait_queue); /* Used for blocking read */

#define RX_STATE_POSITION_MASK 0xF0
	#define RX_STATE_BEFORE_TRACK 0x00
	#define RX_STATE_ON_TRACK 0x10
	#define RX_STATE_LRC 0x20
	#define RX_STATE_AFTER_TRACK 0x30
	#define RX_STATE_SEARCHING_START 0x40
// used also on data_rdy
#define RX_STATE_TRACK_MASK 0x07
	#define TRACK_0 0x01
	#define TRACK_1 0x02
	#define TRACK_2 0x04
#define RX_DIR_BACKWARD 0x08

#define DATA_RDY_TRACK_MASK 0x07
// flag that is set on new card swipe and cleared on data read
#define DATA_RDY_NEW_READ 0x10

#define STATE_READY 0
#define STATE_POWER_ON 1
#define STATE_POWER_OFF 2
#define STATE_BUSY 3
#define STATE_DEBOUNCE 4

#define DEBOUNCE_COUNT 5
#define DEBOUNCE_INTERVAL msecs_to_jiffies(50)

#define START_SENTINEL 0x45
#define END_SENTINEL 0x1F

// External interrupt handler.
//  An interrupt is generated with magstripe data line to
//  indicate that we have more data.
static irqreturn_t magstripe_magtek_new(int irq, void *arg)
{
#ifdef DEBUG
	printk("magtek received reader IRQ\n");
#endif

	if (gpio_get_value(info.data_pin) != 0) return IRQ_HANDLED;

	// If we're not ready, skip the interrupt
	if (info.state != STATE_READY) {
		return IRQ_HANDLED;
	}

	disable_irq(info.reader_irq);

	info.dcnt = DEBOUNCE_COUNT;
	info.state = STATE_DEBOUNCE;

	mod_timer(&info.timer, jiffies + DEBOUNCE_INTERVAL);

	return IRQ_HANDLED;
}

#ifdef DEBUG_RAW_DATA
static int print_data(void)
{
	int i;

	printk("rx_state: %x \n", info.rx_state);
	printk("data_ready: %x \n", info.data_rdy);

	for (i=0;i<	RX_BUF_SIZE; i+=4)
	{
		printk("%d: %02x %02x %02x %02x ",i, info.rxbuf[i],info.rxbuf[i+1],info.rxbuf[i+2],info.rxbuf[i+3]);
		printk("%c %c %c %c \n",(info.rxbuf[i] & 0x3F)+0x20,(info.rxbuf[i+1] & 0x3f)+0x20,(info.rxbuf[i+2]&0x3f)+0x20,(info.rxbuf[i+3]&0x3f)+0x20);
	}
}
#endif

// checks data parity and sets track valid bits in info.data_ready
// info.rd_lock should be taken!
void validate_data(char data)
{
	char tmp, parity;

	switch (info.rx_state & 0xF0)
	{
		case RX_STATE_BEFORE_TRACK:
			if (data == START_SENTINEL) {
				info.rx_state = (info.rx_state & ~RX_STATE_POSITION_MASK) | RX_STATE_ON_TRACK;
				info.checksum = data;
			}
			break;
		case RX_STATE_ON_TRACK:
			// we have 6-bit data. for parity calculation, convert to 3-bit
			tmp = data ^ (data >> 3);
			tmp &= 0x7;
			parity = (0x9669 >> tmp) & 1;
			if (parity<<6 == (data & 0x40))
			{
				info.checksum ^= data;
				if (data == END_SENTINEL) {
					info.rx_state = (info.rx_state & ~RX_STATE_POSITION_MASK) | RX_STATE_LRC;
				}
			}
			else
			{
				//parity error. we're done with this track
				info.rx_state = (info.rx_state & ~RX_STATE_POSITION_MASK) | RX_STATE_AFTER_TRACK;
#ifdef DEBUG
				printk("Magstripe parity error on track %d (data %x)\n", info.rx_state & 0x0F, data);
#endif
			}
			break;
		case RX_STATE_LRC: // LRC check digit
			tmp = data ^ (data >> 3);
			tmp &= 0x7;
			parity = (0x9669 >> tmp) & 1;
			if ((parity<<6 == (data & 0x40)) && ((data & 0x3F) == (info.checksum & 0x3F)))
			{
				info.data_rdy |= info.rx_state & 0x0F; //set current track valid bit
#ifdef DEBUG
				printk("Track %d valid\n", info.rx_state & 0x0F);
#endif
			}
			else
			{
#ifdef DEBUG
				printk("Magstripe checksum error on track %d  (calc %x, recv %x)\n",
						info.rx_state & 0x0F, info.checksum, data);
#endif
			}

			info.rx_state = (info.rx_state & ~RX_STATE_POSITION_MASK) | RX_STATE_AFTER_TRACK;
			break;
		default:
			break;
	}
}

// SSC RXRDY interrupt handler
// TODO: SSC OVERFLOW CHECK
static irqreturn_t magstripe_magtek_io(int irq, void *arg)
{
	unsigned long data;
	unsigned long dummy;
	int max_shift;

	// have we completed all reads?
	if (info.rx_count >= RX_COUNT || info.data_rdy & DATA_RDY_NEW_READ)

	{
		// stop the SSC
		SSC_PUT_IDR(info, AT91_SSC_RXRDY);
		SSC_PUT_CR(info, AT91_SSC_RXDIS);

		// turn off reader power for reset
		at91_set_gpio_output(info.pwr_pin, 0);
		info.state=STATE_POWER_OFF;

#ifdef DEBUG_RAW_DATA
		print_data();
#endif

		// SSC hw seems buggy. reset it after use.
		// one more dummy read to flush the receiver buffer
		dummy = SSC_GET_RHR(info);

		SSC_PUT_CR(info,AT91_SSC_SWRST);//reset the SSC
		SSC_PUT_CMR(info,SSC_CLK_DIV);// set clock divider
		SSC_PUT_RFMR(info, 31); // 32-bit reads

		// timer to turn reader power back on
		mod_timer(&info.timer, jiffies + MAG_RESET_TIME );

		// copy the direction bit
		info.data_rdy |= info.rx_state & RX_DIR_BACKWARD;

		// indicate that a read is complete
		info.data_rdy |= DATA_RDY_NEW_READ;

	    // wake up blocking read
		wake_up_interruptible (&read_wait_queue);

		return IRQ_HANDLED;
	} // no? then continue reading.


	data = ~(SSC_GET_RHR(info));

#ifdef DUAL_DIRECTION_READ
	if ((info.rx_state & RX_STATE_POSITION_MASK) != RX_STATE_SEARCHING_START)
	{
#endif
	// second half of a 7-bit word on 32-bit boundary
	*(info.rxbuf_w) |= (data<<(-info.data_shift)) & 0x7F;
	info.data_shift += 7;
	validate_data(*(info.rxbuf_w));
	info.rxbuf_w++;
#ifdef DUAL_DIRECTION_READ
	}
#endif

	max_shift = ((info.rx_state & RX_STATE_POSITION_MASK) == RX_STATE_SEARCHING_START)? 32 :(32-7);
	// 7-bit words within current 32 bits
	while (info.data_shift <= max_shift)
	{
#ifdef DUAL_DIRECTION_READ
		if ((info.rx_state & RX_DIR_BACKWARD) &&
				((info.rx_state & RX_STATE_POSITION_MASK) == RX_STATE_SEARCHING_START)) {
			if ((data >> (info.data_shift)) & 0x01) {
#ifdef DEBUG
				printk("track start: %d, %d, %d\n", info.rx_count, info.data_shift, info.rxbuf_w - info.rxbuf);
#endif
				info.rx_state = (info.rx_state & ~RX_STATE_POSITION_MASK) | RX_STATE_BEFORE_TRACK;
				max_shift = 32-7;
			}
			else
			{
				info.data_shift++;
				continue;
			}
		}
#endif
		//end of track 0
		if (info.rx_count==76 && info.data_shift==12) {
			info.data_shift=18;
			info.rx_state = TRACK_1 | RX_STATE_BEFORE_TRACK;
		}
		//end of track 1
		if (info.rx_count==152 && info.data_shift==12) {
			info.data_shift=18;
			info.rx_state = TRACK_2 | RX_STATE_BEFORE_TRACK;
		}
#ifdef DUAL_DIRECTION_READ
		//end of direction 1, start track 2
		if (info.rx_count==228 && info.data_shift==12) {
			// If at least one of the track has succeeded, we're done
			if (info.data_rdy) {
				 info.data_rdy |= DATA_RDY_NEW_READ;
			}
			else
			{
				// otherwise, try the other direction
				info.data_shift=18;
				info.rx_state = RX_DIR_BACKWARD | TRACK_2 | RX_STATE_SEARCHING_START;
				max_shift = 32;
			}
		}
		//end of track 2, start track 1
		if (info.rx_count==304 && info.data_shift>=(18-7) && (info.rx_state & TRACK_2)) {
			/*
			 	xxxx xxxN NNNN NNOO OOOO Oxxx xxxx xxxx  ds 11, shl 14, shr 25
				xxxx xxxN NNNN NNOO OOOO xxxx xxxx xxxx  ds 12, shl 14, shr 26
				xxxx xxxN NNNN NNOO OOOx xxxx xxxx xxxx  ds 13, shl 14, shr 27
				xxxx xxxN NNNN NNOO OOxx xxxx xxxx xxxx  ds 14, shl 14, shr 28
				xxxx xxxN NNNN NNOO Oxxx xxxx xxxx xxxx  ds 15, shl 14, shr 29
				xxxx xxxN NNNN NNOO xxxx xxxx xxxx xxxx  ds 16, shl 14, shr 30
				xxxx xxxN NNNN NNOx xxxx xxxx xxxx xxxx  ds 17, shl 14, shr 31
			 */
			*(info.rxbuf_w) = ((data<< (32-18))>>(info.data_shift +(32-18))) & 0x7F;
			info.data_shift += 7;
			validate_data(*(info.rxbuf_w));
#ifdef DEBUG
			printk("track1: %d, %d\n", info.rx_count, info.data_shift);
#endif
			info.data_shift=18;
			info.rx_state = RX_DIR_BACKWARD | TRACK_1 | RX_STATE_SEARCHING_START;
			info.rxbuf_w = &info.rxbuf[track_begin_backwards[1]];
		}
		//end of track 1, start track 0
		if (info.rx_count==380 && info.data_shift>=(18-7) && (info.rx_state & TRACK_1)) {
			*(info.rxbuf_w) = ((data<< (32-18))>>(info.data_shift +(32-18))) & 0x7F;
			info.data_shift += 7;
			validate_data(*(info.rxbuf_w));
#ifdef DEBUG
			printk("track0: %d, %d\n", info.rx_count, info.data_shift);
#endif
			info.data_shift=18;
			info.rx_state = RX_DIR_BACKWARD | TRACK_0 | RX_STATE_SEARCHING_START;
			info.rxbuf_w = &info.rxbuf[track_begin_backwards[0]];
		}

		//end of track 0
		if (info.rx_count==456 && info.data_shift>=(18-7) && (info.rx_state & TRACK_0)) {
			*(info.rxbuf_w) = ((data<< (32-18))>>(info.data_shift +(32-18))) & 0x7F;
			info.data_shift += 7;
			validate_data(*(info.rxbuf_w));
			info.rxbuf_w++;
#ifdef DEBUG
			printk("track0_end: %d, %d\n", info.rx_count, info.data_shift);
#endif
			info.rx_state = RX_DIR_BACKWARD | TRACK_0 | RX_STATE_AFTER_TRACK;
		}

#endif
		if ((info.rx_state & RX_STATE_POSITION_MASK) != RX_STATE_SEARCHING_START)
		{
			*(info.rxbuf_w) = (data>>(info.data_shift)) & 0x7F;
			info.data_shift += 7;
			validate_data(*(info.rxbuf_w));
			info.rxbuf_w++;
		}
	}

	if ((info.rx_state & RX_STATE_POSITION_MASK) != RX_STATE_SEARCHING_START)
	{
		// first half of 7-bit word on 32-bit boundary
		*(info.rxbuf_w) = (data >> (info.data_shift)) & 0x7F;
	}
	info.data_shift -= 32; // if state searching_start, then data_shift should be always 32 here.
	info.rx_count +=4;

	return IRQ_HANDLED;
}

// timer that turns reader power back on and enables reader IRQ
static void magstripe_magtek_timer(unsigned long handle)
{
	unsigned long flags;

	if (info.state == STATE_DEBOUNCE) {
		if (gpio_get_value(info.data_pin) != 0) {
			// we do not have new data - this was just some noise. Reset everything.
			info.state = STATE_POWER_OFF;
			at91_set_gpio_output(info.pwr_pin, 0);
			mod_timer(&info.timer, jiffies + MAG_RESET_TIME );
			return;
		}
		if (!info.dcnt) {
			spin_lock_irqsave(&info.rd_lock, flags);

			// configure SSC and data structures for new read
			info.state = STATE_BUSY;
			info.data_rdy=0;
			info.rxbuf_w = info.rxbuf;
			info.rx_count = 0;
			*(info.rxbuf_w)=0;
			info.data_shift = -3;  //drop first 18 bytes
			info.rx_state = TRACK_0 | RX_STATE_BEFORE_TRACK;

			spin_unlock_irqrestore(&info.rd_lock, flags);

			SSC_PUT_IER(info, AT91_SSC_RXRDY); //enable rxrdy int
			SSC_PUT_CR(info,AT91_SSC_RXEN);

			//clock MUST be switched on after RXEN, otherwise we lose first bit on half occasions
			SSC_PUT_RCMR(info, AT91_SSC_CKS_DIV|AT91_SSC_CKO_CONTINUOUS|AT91_SSC_CK_RISING|AT91_SSC_CKG_NONE|AT91_SSC_START_CONTINUOUS);
				//use divided clock, continuous receive clk, sample on rising edge
				// no clock gating with RF, sample continuously
		}
		else {
			info.dcnt--;
			mod_timer(&info.timer, jiffies + DEBOUNCE_INTERVAL);
		}
	}
	else
	if (info.state == STATE_POWER_OFF)
	{
		// turn on power
		at91_set_gpio_output(info.pwr_pin, 1);

		spin_lock_irqsave(&info.rd_lock, flags);
		info.state = STATE_POWER_ON;
		spin_unlock_irqrestore(&info.rd_lock, flags);

		mod_timer(&info.timer, jiffies + MAG_RESET_TIME );
	}
	else
	{
		spin_lock_irqsave(&info.rd_lock, flags);
		info.state = STATE_READY;
		spin_unlock_irqrestore(&info.rd_lock, flags);

		// enable reader irq
		enable_irq(info.reader_irq);
	}
}
/* ======================================= USERSPACE INTERFACE ==================================*/


// TODO: to be fully compatible to posix, handle segmented reads?
static ssize_t magstripe_magtek_read(struct file *file, char *buf, size_t count,
			 loff_t * ppos)
{
	ssize_t cnt=1;
	char *rd, *wr = buf+1;
	int state, track, i;
	unsigned long flags;

	// OUTPUT DATA FORMAT:
	// byte 1: flags
	//			0x07: valid tracks bitmask
	//			0x10: data ready == 1
	// bytes 2-n: tracks 1,2,3 data concatenated

	// All the data must be read within one operation.
	// The read buffer must be at least 253 bytes!
	// (maximum data length on card). The data is returned
	// as-is (ie if the card track is padded with spaces,
	// those are read, too.
	// Card data bytes are ASCII in range 0x20...0x5F!!!

	// if the user wanted 0 bytes, let him have it
	if (count == 0) return 0;

	// the user must provide a buffer long enough to accomodate full data
	if (count < 3*(TRACK_LEN-2) + 1) return -EINVAL;

#ifdef DEBUG_RAW_DATA
	// Reset buffer to see uninitialized areas.
	memset(info.rxbuf, 0xFF, RX_BUF_SIZE);
#endif

	// wait until we have some data available
	while (!(info.data_rdy & DATA_RDY_NEW_READ))
	{
		interruptible_sleep_on(&read_wait_queue);
		/* was the sleep interrupted by a signal? */
		if (signal_pending(current))
			return -ERESTARTSYS;
	}

	spin_lock_irqsave(&info.rd_lock, flags);

	// if not data ready then return EOF
	if (!(info.data_rdy & DATA_RDY_NEW_READ)) return 0;

	buf[0] = info.data_rdy;

#ifdef DEBUG
	printk("data_rdy: %x\n", info.data_rdy);
#endif

	for (track=0; track<3; track++)
	{
		//send only valid track data
		if (!(info.data_rdy & (1<<track))) continue;

		if (info.data_rdy & RX_DIR_BACKWARD)
			rd = info.rxbuf + track_begin_backwards[track];
		else
			rd = info.rxbuf + track_begin[track];

		state=RX_STATE_BEFORE_TRACK;

		for (i=0; i<TRACK_LEN && state != RX_STATE_AFTER_TRACK && cnt<count; i++)
		{
			switch(state)
			{
				case RX_STATE_BEFORE_TRACK:
					if (*rd == START_SENTINEL)
						state=RX_STATE_ON_TRACK;
					break;
				case RX_STATE_ON_TRACK:
					if (*rd == END_SENTINEL)
					{
						state=RX_STATE_AFTER_TRACK;
					}
					else
					{
						cnt++;
						*wr = ((*rd) & 0x3F)+0x20; //copy ascii data, strip checksum bit
						wr++;
					}
			}
			rd++;
		}
	}

	// this set of data is read. invalidate the data.
	info.data_rdy=0;

	spin_unlock_irqrestore(&info.rd_lock, flags);

	return cnt;
}

/* ======================================= INIT AND UNINIT ======================================*/

// device file operations
static struct file_operations magstripe_magtek_fops = {
      .owner = THIS_MODULE,
      .read = magstripe_magtek_read,
      //.write = magstripe_magtek_write,
      //.open = magstripe_magtek_open,
      //.release = magstripe_magtek_close,
      //.ioctl = magstripe_magtek_ioctl,
};

static struct miscdevice magstripe_magtek_miscdev = {
	.minor = MAGSTRIPE_MAGTEK_MINOR,
	.name = "magstripe_magtek",
	.fops = &magstripe_magtek_fops
};



// platform device probe
// TODO: check error handling!
static int __devinit magstripe_magtek_probe(struct platform_device *pdev)
{
	int ret=0;
	struct clk	*clk;
	struct resource *regs;
	int ssc_irq, reader_irq;
	struct magstripe_magtek_platform_data *pdata = pdev->dev.platform_data;

	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs)
		return -ENXIO;

	ssc_irq = platform_get_irq(pdev, 0);
	if (ssc_irq < 0)
		return ssc_irq;

	reader_irq = platform_get_irq(pdev, 1);
	if (reader_irq < 0)
		return reader_irq;

    clk = clk_get(NULL, "ssc1_clk");
    if (IS_ERR(clk))
		return PTR_ERR(clk);

	info.rxbuf = kmalloc(RX_BUF_SIZE, GFP_KERNEL);
	if (!info.rxbuf) {
		printk(KERN_ERR "Could not allocate data buffer\n");
		goto out1;
	}
	info.rxbuf_w = info.rxbuf;

	clk_enable(clk);

	info.membase = ioremap(regs->start, regs->end - regs->start + 1);
	if (!info.membase)
		goto out;

	init_timer(&info.timer);
	info.timer.function = magstripe_magtek_timer;

	info.pwr_pin = pdata->power_pin;
	info.data_pin = pdata->data_pin;

	// turn on reader power
	at91_set_gpio_output(info.pwr_pin, 1);

	info.state = STATE_READY;
	info.data_rdy=0;

	SSC_PUT_CR(info,AT91_SSC_SWRST);//reset the SSC
	SSC_PUT_CMR(info,SSC_CLK_DIV);// set clock divider

	SSC_PUT_RFMR(info, 31); // 32-bit reads
	SSC_PUT_RCMR(info, AT91_SSC_CKS_DIV|AT91_SSC_CKO_DURING_TRANSFERS|AT91_SSC_CK_RISING|AT91_SSC_CKG_NONE|AT91_SSC_START_CONTINUOUS);
		//use divided clock, continuous receive clk, sample on rising edge
		// no clock gating with RF, sample continuously

	// ssc controller irq
	ret = request_irq(ssc_irq, magstripe_magtek_io, 0, "magtek0", NULL);
	if (ret) {
		printk(KERN_ERR "unable to request ssc IRQ%d\n", ssc_irq);
		goto out;
	}
	info.ssc_irq = ssc_irq;

	// magstripe reader irq
	ret = request_irq(reader_irq, magstripe_magtek_new, IRQF_TRIGGER_LOW, "magtek1", NULL);
	if (ret) {
		printk(KERN_ERR "unable to request reader IRQ%d\n", reader_irq);
		goto out;
	}
	info.reader_irq = reader_irq;

	ret = misc_register(&magstripe_magtek_miscdev);
	if (ret) {
		printk(KERN_ERR "Can't misc_register on minor=%d\n", MAGSTRIPE_MAGTEK_MINOR);
		goto out;
	}

	printk(KERN_INFO "Magtek magstripe reader at %p (irq %i, %i)\n", info.membase, info.ssc_irq, info.reader_irq);


	return 0;
out:
	kfree(info.rxbuf);
out1:
	return ret;
}


// platform device remove
// TODO: this is not complete!
static int __devexit magstripe_magtek_remove(struct platform_device *pdev)
{
	free_irq(info.ssc_irq, NULL);
	free_irq(info.reader_irq, NULL);

	kfree(info.rxbuf);

	printk(KERN_INFO "Unregistered Magtek magstripe reader driver\n");
	return 0;
}


// TODO: suspend and resume
#ifdef CONFIG_PM
static int magstripe_magtek_suspend(struct platform_device *pdev, pm_message_t state)
{
	at91_set_gpio_output(info.pwr_pin, 0);
	return 0;
}

static int magstripe_magtek_resume(struct platform_device *pdev)
{
	at91_set_gpio_output(info.pwr_pin, 1);
	return 0;
}
#else
#define magstripe_magtek_suspend NULL
#define magstripe_magtek_resume NULL
#endif




static struct platform_driver magstripe_magtek_driver = {
	.probe		= magstripe_magtek_probe,
	.remove		= __devexit_p(magstripe_magtek_remove),
	.suspend	= magstripe_magtek_suspend,
	.resume		= magstripe_magtek_resume,
	.driver		= {
		.name	= "magstripe_magtek",
		.owner	= THIS_MODULE,
	},
};

static int __init magstripe_magtek_init(void)
{
	return platform_driver_register(&magstripe_magtek_driver);
}

static void __exit magstripe_magtek_exit(void)
{
	platform_driver_unregister(&magstripe_magtek_driver);
}

module_init(magstripe_magtek_init);
module_exit(magstripe_magtek_exit);

MODULE_AUTHOR("Anti Sullin");
MODULE_DESCRIPTION("Magtek magstripe reader driver");
MODULE_LICENSE("GPL");
