/*
 * CMEL C0283QGLC OLED panel control driver
 * Copyright (C) 2008 Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * The driver controls the CMEL C0283QGLC OLED panel via SPI bus
 *
 * User interface is provided via both write and ioctl interfaces.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//#define DEBUG 1

#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/spi/spi.h>
#include <linux/fs.h>
#include <asm/gpio.h>
#include <linux/delay.h>
#include <linux/oled.h>

// major is always 10 for misc devices
#define C0283QGLC_MINOR 52
// see also include/linux/miscdevice.h for number allocation


struct oled_conf {
	unsigned char reg;
	short val;
	unsigned char delay; // delay in ms after this byte was sent.
						// 0xFF if last byte.
						// 0xFE if needed to switch power on afterwards
};

#define R_TOP 0x6F
#define G_TOP 0x6F
#define B_TOP 0x79
#define R_BTM 0x0E
#define G_BTM 0x0E
#define B_BTM 0x0E

#define FADE_STEP 0x20
#define FADE_TIMESTEP (msecs_to_jiffies(25))
#define PWRON_DELAY (msecs_to_jiffies(50))

// ********************************** PRE-DEFINED COMMANDS ***************************************

// Freeze
struct oled_conf oled_freeze[] = {
		{0x02, 0x0012, 0xFF}
};
struct oled_conf oled_unfreeze[] = {
		{0x02, 0x0192, 0xFF}
};

// blank
struct oled_conf oled_blank[] = {
		{0x05, 0x0000, 0xFF}
};
struct oled_conf oled_unblank[] = {
		{0x05, 0x0001, 0xFF}
};

// standby
struct oled_conf oled_standby[] = {
		{0x05, 0x0000, 0x00},
		{0x10, 0x0001, 0xFF}
};
struct oled_conf oled_resume[] = {
		{0x10, 0x0000, 0x00},
		{0x05, 0x0001, 0xFE},
};

//brightness (reg values set with c0283qglc_oled_set_brightness)
struct oled_conf oled_brt[] = {
		{0x70, 0x0000, 0x00},
		{0x71, 0x0000, 0x00},
		{0x72, 0x0000, 0xFF},
};

struct oled_conf oled_set[] = {
		{0x00, 0x0000, 0xFF}
};


// draiveri globaalsed muutujad
static struct c0283qglc_oled {
	struct spi_device *spi;
	struct mutex lock;

	struct spi_transfer	xfer;
	struct spi_message	msg;

	struct oled_conf *xmit; //pointer to data currently being sent or 0 when idle
	int phase;
	struct timer_list	timer;

	int brt;
	int fade_dest;
	struct timer_list	fade_timer;

	int pwr_pin;
	int pwr;

	unsigned char *xmit_buf;

} info;

#define XMIT_LEN 3


static void c0283qglc_fill_buf(unsigned char *buf, struct oled_conf *data, int phase) {
	if (phase == 0) {
		buf[0] = 0x70;
		buf[1] = 0x00;
		buf[2] = data->reg;
	} else if (phase == 1) {
		buf[0] = 0x72;
		buf[1] = (data->val >> 8) & 0xFF;
		buf[2] = (data->val) & 0xFF;
	} else {
		buf[0] = 0x70;
		buf[1] = 0x00;
		buf[2] = 0x22;
	}
}

/* *********************************************** HIGH LEVEL FNC **************************************/
static void c0283qglc_oled_set_brightness(int brt) {
	oled_brt[0].val = R_BTM + ((R_TOP * brt / 0xFF) << 7);
	oled_brt[1].val = G_BTM + ((G_TOP * brt / 0xFF) << 7);
	oled_brt[2].val = B_BTM + ((B_TOP * brt / 0xFF) << 7);
}

/* *********************************************** SPI LOGIC **************************************/
// reads next command from array and transmits to oled
static void c0283qglc_oled_xmit_next(void)
{
	int status;

	if (!info.xmit) return;

	if (info.phase == 0) {
		info.phase = 1;
		c0283qglc_fill_buf(info.xmit_buf, info.xmit, info.phase);
	} else if (info.phase == 1) {
		if (info.xmit->delay >= 0xFE)
		{
			if (info.brt != info.fade_dest) {
				// do not re-enable RGB interface while fading.
				// we'll get called again until brt==fade_dest.
				info.xmit=NULL;
				return;
			}
			info.phase = 2;
			// Switch off LCDDEN.
			// We must send at least 2 VSYNC's before we start sending RGB data
			// to avoid GRAM corruption and an ugly glitch. So we block LCDDEN.
			at91_set_gpio_output(AT91_PIN_PB3, 0);
		}
		else
		{
			info.phase = 0;
			info.xmit++;
		}
		c0283qglc_fill_buf(info.xmit_buf, info.xmit, info.phase);
	} else
		return;

	status = spi_async(info.spi, &info.msg);
	if (status) {
		dev_err(&info.spi->dev, "spi_async --> %d\n",	status);
		info.xmit=NULL;
		mutex_unlock(&info.lock);
	}
}

// callback after successful SPI transmission
// this either sends new byte immediately or schedules timer callback for delay
static void c0283qglc_oled_spi_rx(void *dummy)
{
	if (!info.xmit) return;

	if (info.phase == 0)
		c0283qglc_oled_xmit_next();
	else if (info.phase == 1) {
		if (info.xmit->delay >= 0xFE)
			c0283qglc_oled_xmit_next();
		else
			mod_timer(&info.timer, jiffies + msecs_to_jiffies(info.xmit->delay+1));
	} else {
		// Wait and re-enable LCDDEN
		while (at91_get_gpio_value(AT91_PIN_PB0));
		while (!at91_get_gpio_value(AT91_PIN_PB0));
		at91_set_A_periph(AT91_PIN_PB3, 0);

		// Switch OLED panel on after a delay?
		if (info.xmit->delay == 0xFE) {
			info.xmit=NULL;
			mod_timer(&info.timer, jiffies + PWRON_DELAY);
		} else {
			info.xmit=NULL;
			mutex_unlock(&info.lock);
		}
	}
}

/* *********************************************** CALLBACKS **************************************/

// callback for delay timer
static void c0283qglc_oled_timer(unsigned long dummy)
{
	if (!info.xmit) {
		// The only time we get here while info.xmit is NULL,
		// is when we are doing poweron delay
		at91_set_gpio_output(info.pwr_pin, 1);
		info.pwr = 1;
		mutex_unlock(&info.lock);
		return;
	}
	c0283qglc_oled_xmit_next();
}

// fader timer handler.
static void c0283qglc_fade_timer(unsigned long dummy)
{
	int status;

	if (info.xmit) {
		// spi is busy... do another loop
		mod_timer(&info.fade_timer, jiffies + FADE_TIMESTEP);
		return;
	}

	if (info.brt > info.fade_dest) {
		if (info.brt - info.fade_dest < FADE_STEP)
			info.brt = info.fade_dest;
		else
			info.brt -= FADE_STEP;
	} else {
		if (info.fade_dest - info.brt < FADE_STEP)
			info.brt = info.fade_dest;
		else
			info.brt += FADE_STEP;
	}

	c0283qglc_oled_set_brightness(info.brt);
	info.xmit = oled_brt;

	info.phase = 0;
	c0283qglc_fill_buf(info.xmit_buf, info.xmit, info.phase);

	status = spi_async(info.spi, &info.msg);
	if (status) {
		dev_err(&info.spi->dev, "spi_async --> %d\n",	status);
		info.xmit = NULL;
		mutex_unlock(&info.lock);
	}

	if (info.brt != info.fade_dest)
		mod_timer(&info.fade_timer, jiffies + FADE_TIMESTEP); // continue?
}
/* *********************************************** USERSPACE INTERFC **************************************/

static ssize_t c0283qglc_oled_write(struct file *file, const char *buf, size_t count,
			 loff_t * ppos)
{
	int status, temp;

	if (count==0) return 0;

	if (mutex_lock_interruptible(&info.lock))
		return -ERESTARTSYS;

	if (info.xmit) { // are we still busy?
		mutex_unlock(&info.lock);
		return -EBUSY;
	}

	switch (buf[0]) {
	case 'S':
		at91_set_gpio_output(info.pwr_pin, 0);
		info.pwr = 0;
		info.xmit = oled_standby;
		break;
	case 's':
		info.xmit = oled_resume;
		break;
	case 'B':
		info.xmit = oled_blank;
		break;
	case 'b':
		info.xmit = oled_unblank;
		break;
	case 'F':
		info.xmit = oled_freeze;
		break;
	case 'f':
		info.xmit = oled_unfreeze;
		break;
	case 'L':
	case 'l':
		if (count < 3) return -EINVAL;
		if (sscanf(buf+1, "%2x", &temp)!=1) return -EINVAL;
		info.brt = temp & 0xFF;
		info.fade_dest = info.brt;
		c0283qglc_oled_set_brightness(info.brt);
		info.xmit = oled_brt;
		break;
	case 'R':
	case 'r':
		if (count < 7) return -EINVAL;
		if (sscanf(buf+1, "%6x", &temp)!=1) return -EINVAL;
		oled_set[0].reg=temp & 0xFF;
		oled_set[0].val=(temp >> 8) & 0xFFFF;
		info.xmit = oled_set;
		break;
	case 'A':
	case 'a':
		if (count < 3) return -EINVAL;
		if (sscanf(buf+1, "%2x", &temp)!=1) return -EINVAL;
		info.fade_dest = temp & 0xFF;
		info.xfer.cs_change = 0;
		c0283qglc_fade_timer(0);
		return count;
	}

	if (info.xmit) {
		info.phase = 0;
		c0283qglc_fill_buf(info.xmit_buf, info.xmit, info.phase);
		info.xfer.cs_change = 0;

		status = spi_async(info.spi, &info.msg);
		if (status) {
			dev_err(&info.spi->dev, "spi_async --> %d\n",	status);
			info.xmit = NULL;
			mutex_unlock(&info.lock);
		}
	} else
		mutex_unlock(&info.lock);

	return count;
}

static ssize_t c0283qglc_oled_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	      unsigned long arg)
{
	int status;

	if (mutex_lock_interruptible(&info.lock))
		return -ERESTARTSYS;

	if (info.xmit) { // are we still busy?
		mutex_unlock(&info.lock);
		return -EBUSY;
	}

	switch (cmd) {
	case OLED_SUSPEND:
		at91_set_gpio_output(info.pwr_pin, 0);
		info.pwr = 0;
		info.xmit = oled_standby;
		break;
	case OLED_RESUME:
		info.xmit = oled_resume;
		break;
	case OLED_BLANK:
		info.xmit = oled_blank;
		break;
	case OLED_UNBLANK:
		info.xmit = oled_unblank;
		break;
	case OLED_FREEZE:
		info.xmit = oled_freeze;
		break;
	case OLED_UNFREEZE:
		info.xmit = oled_unfreeze;
		break;
	case OLED_SET_BRT:
		info.brt = arg & 0xFF;
		info.fade_dest = info.brt;
		c0283qglc_oled_set_brightness(info.brt);
		info.xmit = oled_brt;
		break;
	case OLED_FADE_BRT_TO:
		info.fade_dest = arg & 0xFF;
		info.xfer.cs_change = 0;
		c0283qglc_fade_timer(0);
		return 0;
	}

	if (info.xmit) {
		info.phase = 0;
		c0283qglc_fill_buf(info.xmit_buf, info.xmit, info.phase);
		info.xfer.cs_change = 0;

		status = spi_async(info.spi, &info.msg);
		if (status) {
			dev_err(&info.spi->dev, "spi_async --> %d\n",	status);
			info.xmit = NULL;
			mutex_unlock(&info.lock);
		}
	} else
		mutex_unlock(&info.lock);

	return 0;
}

/* *********************************************** MISC DEV PROBE, REMOVE etc **************************************/




// misc device init params: file operations
static struct file_operations c0283qglc_oled_fops = {
      .owner = THIS_MODULE,
      .write = c0283qglc_oled_write,
      .ioctl = c0283qglc_oled_ioctl,
};

// misc dev init params
static struct miscdevice c0283qglc_oled_miscdev = {
	.minor = C0283QGLC_MINOR,
	.name = "c0283qglc_oled",
	.fops = &c0283qglc_oled_fops
};



static int c0283qglc_oled_suspend(struct spi_device *spi, pm_message_t message)
{
	// THIS HERE IS JUST FOR PRECAUTION. SUSPEND THE PANEL BEFORE GOING TO FULL SUSPEND.

	// This shuts down the OLED panel power. Otherwise, if the LCDC clock
	// is shut off, the panel goes full-on and this might be dangerous.
	// This DOES NOT turn off the OLED driver chip, so all DCDC's and
	// clocks are still running! Always suspend the driver chip before
	// going to suspend to minimize power consumption.
	at91_set_gpio_output(info.pwr_pin, 0);
	return 0;
}

static int c0283qglc_oled_resume(struct spi_device *spi)
{
	if (info.pwr)
		at91_set_gpio_output(info.pwr_pin, 1);
	return 0;
}


static int __devinit c0283qglc_oled_probe(struct spi_device *spi)
{
	int ret;
	int	err;
	memset(&info, 0, sizeof(info));

	spi->bits_per_word = 8;

	info.spi = spi;
	info.brt = 0xFF;
	info.fade_dest = 0xFF;
	spi->mode = SPI_MODE_3;
	err = spi_setup(spi);
	if (err < 0)
		return err;

	dev_set_drvdata(&spi->dev, &info);
	spi->dev.power.power_state = PMSG_ON;

	info.pwr_pin = (int)spi->dev.platform_data;
	at91_set_gpio_output(info.pwr_pin, 1);
	info.pwr = 1;

	spi_message_init(&info.msg);

	info.xfer.len = XMIT_LEN;

	spi_message_add_tail(&info.xfer, &info.msg);
	info.msg.complete = c0283qglc_oled_spi_rx;

	init_timer(&info.timer);
	info.timer.function = c0283qglc_oled_timer;

	init_timer(&info.fade_timer);
	info.fade_timer.function = c0283qglc_fade_timer;

	// alloc dma-safe buffer
	info.xmit_buf = kzalloc(XMIT_LEN, GFP_KERNEL);
	if (!info.xmit_buf) {
		err = -ENOMEM;
		goto out2;
	}
	info.xfer.tx_buf = info.xmit_buf;

	// register misc dev
	ret = misc_register(&c0283qglc_oled_miscdev);
	if (ret) {
		printk(KERN_ERR "Can't misc_register on minor=%d\n", C0283QGLC_MINOR);
		err = -EBUSY;
		goto out;
	}
	mutex_init(&info.lock);

	printk(KERN_INFO "c0283qglc oled SPI control\n");

	return 0;
out2:

out:

	return err;
}
static int __devexit c0283qglc_oled_remove(struct spi_device *spi)
{
	dev_dbg(&spi->dev, "unregistered c0283qglc oled driver\n");
	return 0;
}



/* *********************************************** KERNEL DRIVER STUFF **************************************/


static struct spi_driver c0283qglc_oled_driver = {
	.driver = {
		.name	= "c0283qglc_oled", // nimi peab vastama board'is seadistatud device'ga
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
	},
	.probe		= c0283qglc_oled_probe,
	.remove		= __devexit_p(c0283qglc_oled_remove),
	.suspend	= c0283qglc_oled_suspend,
	.resume		= c0283qglc_oled_resume,
};


static int __init c0283qglc_oled_init(void)
{
	return spi_register_driver(&c0283qglc_oled_driver);
}
module_init(c0283qglc_oled_init);


static void __exit c0283qglc_oled_exit(void)
{
	spi_unregister_driver(&c0283qglc_oled_driver);
}
module_exit(c0283qglc_oled_exit);

MODULE_DESCRIPTION("c0283qglc oled control driver");
MODULE_LICENSE("GPL");
