/*
 * External power switching driver
 * Copyright (C) 2008 Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * The driver implements interfaces to control external device power switching.
 *
 * PWM-driven power:
 *  - keyboard backlight
 *
 * ON/OFF switched power
 *  - gpio pins
 *
 *  TRISTATE control
 *  - serial ports
 *
 * User interface is provided via both write and ioctl interfaces.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

//#define DEBUG

#include <linux/clk.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <mach/at91_tc.h>
#include <asm/gpio.h>
#include <linux/pwrctrl.h>

// major is always 10 for misc devices
#define PWRCTRL_MINOR 55
// see also include/linux/miscdevice.h for number allocation

// MAX value for backlight
#define KBD_BL_MAX 0xFF
// fade timestep
#define KBD_BL_TIMESTEP (msecs_to_jiffies(25))
#define FADE_STEP	0x20

/* Register access macros */
#define tc_readl(reg) \
	readl((info.membase) + AT91_TC_##reg)
#define tc_writel(reg,value) \
	writel((value), (info.membase) + AT91_TC_##reg)

// draiveri globaalsed muutujad
static struct pwrctrl {
	void __iomem 	*membase;
	struct platform_device *pdev;


	unsigned char kbd_bl; //kbd backlight value
	unsigned char kbd_bl_dest; //kbd backlight fade destination
	struct timer_list	kbd_bl_timer; // fade timer

	struct pwrctrl_platform_data *data; //pin mapping

	struct clk	*fade_timer_clk;
} info;


/* *********************************************** LOW-LEVEL INTFC **************************************/

// ================================= KBD BL ==============================
// set kbd backlight value
inline static void pwrctrl_set_kbd_bl(char bl)
{
	char val = 0xFF - bl + 1; // 0 is always on, 1 is always off
					// 0xff wraps to 0
	tc_writel(RA, val);
	info.kbd_bl = bl;

	// switch KBBL converter on/off
	gpio_set_value(info.data->kbbl_en_pin, (bl != 0));
}

// initialize kbd backlight control
inline static void pwrctrl_init_kbd_bl(char bl)
{
	char val = bl+1; // 0 is always on, 1 is always off
					// 0xff wraps to 0

	tc_writel(IDR, 0xFF); // disable all int's
	tc_writel(CCR, AT91_TC_CLKEN); // enable clk
	tc_writel(CCR, AT91_TC_SWTRG); // trigger clk start
	tc_writel(CMR, AT91_TC_TIMER_CLOCK4 |
		AT91_TC_WAVE | AT91_TC_WAVESEL_UP_AUTO |
		AT91_TC_ACPA_CLEAR | AT91_TC_ACPC_SET | AT91_TC_ASWTRG_CLEAR);
	tc_writel(RA, val);	tc_writel(RC, 0xFF); // resolution 256 steps
}

// un-initialize kbd backlight control
inline static void pwrctrl_uninit_kbd_bl(void)
{
	tc_writel(RA, 0x01); // bl off
	tc_writel(CCR, AT91_TC_CLKDIS); // disable clk
	tc_writel(CCR, AT91_TC_SWTRG); // trigger output to go to inactive
}

// ================================= GPIO I/O ==============================
inline static void pwrctrl_set_gpio(int pin, int val)
{
	if (pin >= O_PIN_COUNT) return;

	switch (info.data->output_pins[pin].controller) {
		case CTRL_GPIO:
			gpio_set_value(info.data->output_pins[pin].pin, val);
			#ifdef DEBUG
			printk("Setting output %d (pin %d) to %d\n",pin, info.data->output_pins[pin].pin, val);
			#endif
		break;
		case CTRL_GPIO_OD:
			if (val)
				at91_set_gpio_input(info.data->output_pins[pin].pin, 0);
			else
				at91_set_gpio_output(info.data->output_pins[pin].pin, 0);
		break;
	}
}

inline static int pwrctrl_get_gpio(int pin)
{
	if (pin >= I_PIN_COUNT) return -EINVAL;

	switch (info.data->input_pins[pin].controller) {
		case CTRL_GPIO:
			return gpio_get_value(info.data->input_pins[pin].pin);
		break;
	}
	return -EINVAL;
}
// ================================= GPIO TRISTATE ==============================
inline static int pwrctrl_tristate_usart(int usart, int tristate)
{
	switch (usart) {
	case 0:
		if (tristate) {
			at91_set_gpio_input(AT91_PIN_PC8, 1); // TXD0
			at91_set_gpio_input(AT91_PIN_PC9, 1); // RXD0
			at91_set_gpio_input(AT91_PIN_PC10, 1); // RTS0
			at91_set_gpio_input(AT91_PIN_PC11, 1); // CTS0
		} else {
			at91_set_A_periph(AT91_PIN_PC8, 1); // TXD0
			at91_set_A_periph(AT91_PIN_PC9, 1); // RXD0
			at91_set_A_periph(AT91_PIN_PC10, 1); // RTS0
			at91_set_A_periph(AT91_PIN_PC11, 1); // CTS0
		}
		return 0;
	case 1:
		if (tristate) {
			at91_set_gpio_input(AT91_PIN_PC12, 1); // TXD1
			at91_set_gpio_input(AT91_PIN_PC13, 1); // RXD1
			at91_set_gpio_input(AT91_PIN_PA12, 1); // RTS1
			at91_set_gpio_input(AT91_PIN_PA13, 1); // CTS1
		} else {
			at91_set_A_periph(AT91_PIN_PC12, 1); // TXD1
			at91_set_A_periph(AT91_PIN_PC13, 1); // RXD1
			at91_set_B_periph(AT91_PIN_PA12, 1); // RTS1
			at91_set_B_periph(AT91_PIN_PA13, 1); // CTS1
		}
		return 0;
	case 2:
		if (tristate) {
			at91_set_gpio_input(AT91_PIN_PC14, 1); // TXD2
			at91_set_gpio_input(AT91_PIN_PC15, 1); // RXD2
			at91_set_gpio_input(AT91_PIN_PA15, 1); // RTS2
			at91_set_gpio_input(AT91_PIN_PA16, 1); // CTS2
		} else {
			at91_set_A_periph(AT91_PIN_PC14, 1); // TXD2
			at91_set_A_periph(AT91_PIN_PC15, 1); // RXD2
			at91_set_B_periph(AT91_PIN_PA15, 1); // RTS2
			at91_set_B_periph(AT91_PIN_PA16, 1); // CTS2
		}
		return 0;
	}
	return -EINVAL;
}

/* *********************************************** CALLBACKS **************************************/

// KBD backlight fader timer handler.
static void pwrctrl_kbd_bl_timer(unsigned long dummy)
{
	if (info.kbd_bl > info.kbd_bl_dest) {
		if (info.kbd_bl - info.kbd_bl_dest < FADE_STEP)
			info.kbd_bl = info.kbd_bl_dest;
		else
			info.kbd_bl -= FADE_STEP;
	} else {
		if (info.kbd_bl_dest - info.kbd_bl < FADE_STEP)
			info.kbd_bl = info.kbd_bl_dest;
		else
			info.kbd_bl += FADE_STEP;
	}
	pwrctrl_set_kbd_bl(info.kbd_bl); // set brightness


	if (info.kbd_bl != info.kbd_bl_dest)
		mod_timer(&info.kbd_bl_timer, jiffies + KBD_BL_TIMESTEP); // continue?
}

/* *********************************************** USERSPACE INTERFC **************************************/

// write interface to use in shell scripts
static ssize_t pwrctrl_write(struct file *file, const char *buf, size_t count,
			 loff_t * ppos)
{
	int temp;

	if (count==0) return 0;

	switch (buf[0])
	{
		case 'O':
		// Oxx: set gpio output xx high
			if (count<3) return -EINVAL;
			if (sscanf(buf+1, "%2x", &temp)!=1) return -EINVAL;
			pwrctrl_set_gpio(temp & 0xFF, 1);
		break;
		case 'o':
		// oxx: set gpio output xx high
			if (count<3) return -EINVAL;
			if (sscanf(buf+1, "%2x", &temp)!=1) return -EINVAL;
			pwrctrl_set_gpio(temp & 0xFF, 0);
		break;
		case 'L':
		// Lxx : set keyboard backlight brightness to xx
			if (count<3) return -EINVAL;
			if (sscanf(buf+1, "%2x", &temp)!=1) return -EINVAL;
			pwrctrl_set_kbd_bl(temp & 0xFF);
		break;
		case 'A':
		case 'a':
		// Axx : fade to xx
			if (count<3) return -EINVAL;
			if (sscanf(buf+1, "%2x", &temp)!=1) return -EINVAL;
			info.kbd_bl_dest = temp & 0xFF;
			mod_timer(&info.kbd_bl_timer, jiffies + KBD_BL_TIMESTEP);
		break;

		case 'T':
		// Tn: Tristate usart n
			if (count<2) return -EINVAL;
			if (sscanf(buf+1, "%1x", &temp)!=1) return -EINVAL;
			pwrctrl_tristate_usart(temp, 1);
		break;
		case 't':
		// tn: Un-Tristate usart n
			if (count<2) return -EINVAL;
			if (sscanf(buf+1, "%1x", &temp)!=1) return -EINVAL;
			pwrctrl_tristate_usart(temp, 0);
		break;

		default:
		return count;
	}


	return count;
}



static int pwrctrl_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		      unsigned long arg)
{
	switch (cmd) {
		case PWRCTRL_SET_GPIO_HI:
			pwrctrl_set_gpio(arg & 0xFF, 1);
			return 0;
		case PWRCTRL_SET_GPIO_LO:
			pwrctrl_set_gpio(arg & 0xFF, 0);
			return 0;
		case PWRCTRL_GET_GPIO:
			return pwrctrl_get_gpio(arg & 0xFF);
		case PWRCTRL_SET_KBDBL:
			pwrctrl_set_kbd_bl(arg & 0xFF);
			return 0;
		case PWRCTRL_FADE_KBDBL_TO:
			info.kbd_bl_dest = arg & 0xFF;
			mod_timer(&info.kbd_bl_timer, jiffies + KBD_BL_TIMESTEP);
			return 0;
		case PWRCTRL_USART_TRISTATE_ON:
			return pwrctrl_tristate_usart(arg & 0xFF, 1);
		case PWRCTRL_USART_TRISTATE_OFF:
			return pwrctrl_tristate_usart(arg & 0xFF, 0);
	}
	return -EINVAL;
}

/* *********************************************** MISC DEV PROBE, REMOVE etc **************************************/

// misc device init params: file operations
static struct file_operations pwrctrl_fops = {
      .owner = THIS_MODULE,
      .write = pwrctrl_write,
      .ioctl = pwrctrl_ioctl,
};

// misc dev init params
static struct miscdevice pwrctrl_miscdev = {
	.minor = PWRCTRL_MINOR,
	.name = "pwrctrl",
	.fops = &pwrctrl_fops
};


// suspend and resume
// NO PM IMPLEMENTED YET
#ifdef CONFIG_PM
static int pwrctrl_suspend(struct platform_device *pdev, pm_message_t state)
{
	gpio_set_value(info.data->kbbl_en_pin, 0);
	clk_disable(info.fade_timer_clk);
	return 0;
}

static int pwrctrl_resume(struct platform_device *pdev)
{
	clk_enable(info.fade_timer_clk);
	gpio_set_value(info.data->kbbl_en_pin, (info.kbd_bl != 0));
	return 0;
}
#else
#define pwrctrl_suspend NULL
#define pwrctrl_resume NULL
#endif



static int __devinit pwrctrl_probe(struct platform_device *pdev)
{
	int ret;
	int j;

	struct resource *regs;

	// just for caution- initialize struct info with zeroes
	memset(&info, 0, sizeof(info));

	// store pointer to device info for later use.
	info.pdev = pdev;

	// tc0 regs
	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs)
		return -ENXIO;

	info.membase = ioremap(regs->start, regs->end - regs->start + 1);
	if (!info.membase)
		return -ENOMEM;

	// initialize pointer to platform data struct
	info.data = pdev->dev.platform_data;

	//initialize KB backlight
	at91_set_gpio_output(info.data->kbbl_en_pin, 0);

	// initialize fade timer
	init_timer(&info.kbd_bl_timer);
	info.kbd_bl_timer.function = pwrctrl_kbd_bl_timer;

	// initialize gpio pins - direction output, default state
	for (j = 0; j < O_PIN_COUNT; j++) {
		if (info.data->output_pins[j].controller == CTRL_GPIO)
		{
			at91_set_gpio_output(info.data->output_pins[j].pin, info.data->output_pins[j].def);
		}
		else if (info.data->output_pins[j].controller == CTRL_GPIO_OD)
		{
			if (info.data->output_pins[j].def)
				at91_set_gpio_input(info.data->input_pins[j].pin, 0);
			else
				at91_set_gpio_output(info.data->output_pins[j].pin, 0);
		}
	}

	// initialize gpio pins - direction input
	for (j = 0; j < I_PIN_COUNT; j++) {
		if (info.data->input_pins[j].controller == CTRL_GPIO)
		{
			at91_set_gpio_input(info.data->input_pins[j].pin, 0);
		}
	}

	// initialize tc0 clock
    info.fade_timer_clk = clk_get(NULL, "tc0_clk");
    if (IS_ERR(info.fade_timer_clk))
		return PTR_ERR(info.fade_timer_clk);

	// try to register misc dev
	ret = misc_register(&pwrctrl_miscdev);
	if (ret) {
		#ifdef DEBUG
		printk(KERN_ERR "Can't misc_register on minor=%d\n", PWRCTRL_MINOR);
		#endif
		return -EBUSY;
	}

	// enable timer clock
	clk_enable(info.fade_timer_clk);
	// configure timer
	info.kbd_bl=0x00; // default value - the boot loader has already switched the bl on
	pwrctrl_init_kbd_bl(info.kbd_bl);


	// we're up&running.
	#ifdef DEBUG
	printk(KERN_INFO "External Power Control\n");
	#endif

	// disable serial ports pins initially, put into input mode to avoid leakage
	pwrctrl_tristate_usart(0, 1);
	pwrctrl_tristate_usart(1, 1);
	pwrctrl_tristate_usart(2, 1);
 	return 0;
}

static int __devexit pwrctrl_remove(struct platform_device *pdev)
{
	// stop timer, output to low level
	pwrctrl_uninit_kbd_bl();

	clk_disable(info.fade_timer_clk);

	dev_dbg(&pdev->dev, "unregistered pwrctrl driver\n");
	return 0;
}



/* *********************************************** KERNEL DRIVER STUFF **************************************/


static struct platform_driver pwrctrl_driver = {
	.driver = {
		.name	= "pwrctrl", // nimi peab vastama board'is seadistatud device'ga
		.owner	= THIS_MODULE,
	},
	.probe		= pwrctrl_probe,
	.remove		= __devexit_p(pwrctrl_remove),
	.suspend	= pwrctrl_suspend,
	.resume		= pwrctrl_resume,
};


static int __init pwrctrl_init(void)
{
	return platform_driver_register(&pwrctrl_driver);
}
module_init(pwrctrl_init);


static void __exit pwrctrl_exit(void)
{
	platform_driver_unregister(&pwrctrl_driver);
}
module_exit(pwrctrl_exit);

MODULE_DESCRIPTION("External Power Control driver");
MODULE_LICENSE("GPL");
