/*
 * Copyright (c) 2010 Artec Design LLC
 *   Jaanus Sepp <jaanus.sepp@artecdesign.ee>
 *   Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * Driver for Austria Microsystems RFID Reader chip
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. *
*/

//#define DEBUG

#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/mutex.h>
#include <linux/miscdevice.h>
#include <linux/spi/spi.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/rfid_ams.h>
#include <asm/gpio.h>

#include "ams.h"

//--------------------------------------------------------------------------------

#define AMS_MINOR                    58

#define set_gpio_output(pin, value) \
  at91_set_gpio_output(info.pdata->pin_assignment[RFID_AMS_PIN_##pin], (value) != 0)

#ifdef DEBUG
#define dbg(fmt, args...) printk(fmt, ##args)
#else
#define dbg(fmt, args...) do {} while(0)
#endif

#define AMS_SPI_BUF_SIZE			128

//--------------------------------------------------------------------------------

struct rfid_ams info;

DECLARE_WAIT_QUEUE_HEAD(irq_wait_queue); // main waits for irq transfers to complete
DEFINE_SPINLOCK(irq_lock);

static void ams_process_irq(void *context);

/* *********************************************** SPI I/O **************************************/
unsigned char ams_read(unsigned char addr)
{
	info.tx_buf[0] = (addr & AMS_ADDRESS_MASK) | AMS_READ;
	info.tx_buf[1] = 0;
	info.spi_trans.len = 2;
	spi_sync(info.p_spi_dev, &info.spi_mess);
	return info.rx_buf[1];
}

//-----------------------------------------------------------------------------
// BROKEN FOR REGISTERS 0x14...0x1E ! Use ams_read()!
int ams_read_cont(unsigned char addr, char *buf, int count)
{
	int ret = -ENOMEM;

	info.tx_buf[0] = (addr & AMS_ADDRESS_MASK) | AMS_READ;
	info.tx_buf[1] = 0; // Chip bug workaround

	info.tx_buf[2] = (addr & AMS_ADDRESS_MASK)	|
						AMS_READ |
						AMS_CONTINUOUS;
	if ((count > 0) && (count < AMS_SPI_BUF_SIZE)) {
		memset((char *)&info.tx_buf[3], 0, count);
		info.spi_trans.len = count + 3;
		ret = spi_sync(info.p_spi_dev, &info.spi_mess);
		memcpy(buf, (char *)&info.rx_buf[3], count);
	}
	return ret;
}

//-----------------------------------------------------------------------------
int ams_write_async(unsigned char addr, unsigned char data,
		void (*complete)(void *))
{
	info.spi_mess.complete = complete;
	info.tx_buf[0] = addr & AMS_ADDRESS_MASK;
	info.tx_buf[1] = data;
	info.tx_buf[2] = AMS_READ;		// Chip bug workaround
	info.spi_trans.len = 3;

	if (info.spi_mess.complete)
		return spi_async(info.p_spi_dev, &info.spi_mess);

	return spi_sync(info.p_spi_dev, &info.spi_mess);
}

//-----------------------------------------------------------------------------
int ams_write_cont_async(unsigned char addr, char *buf, int count,
		void (*complete)(void *))
{
	if ((count <= 0) || (count > AMS_SPI_BUF_SIZE - 1))	return -ENOMEM;

	info.spi_mess.complete = complete;
	info.tx_buf[0] = (addr & AMS_ADDRESS_MASK)	| AMS_CONTINUOUS;
	memcpy((char *)&info.tx_buf[1], buf, count);
	info.spi_trans.len = count + 1;

	if (info.spi_mess.complete)
		return spi_async(info.p_spi_dev, &info.spi_mess);
	return spi_sync(info.p_spi_dev, &info.spi_mess);
}

//-----------------------------------------------------------------------------
int ams_command_async(unsigned char cmd,
		void (*complete)(void *context))
{
	info.spi_mess.complete = complete;
	info.tx_buf[0] = cmd;
	info.tx_buf[1] = AMS_READ;		// Chip bug workaround
	info.spi_trans.len = 2;

	if (info.spi_mess.complete)
		return spi_async(info.p_spi_dev, &info.spi_mess);
	return spi_sync(info.p_spi_dev, &info.spi_mess);
}

//-----------------------------------------------------------------------------
int ams_command_write_async(unsigned char cmd,
		unsigned char address, unsigned char data,
		void (*complete)(void *context))
{
	info.spi_mess.complete = complete;
	info.spi_trans.len = 4;
	info.tx_buf[0] = cmd;
	info.tx_buf[1] = address & AMS_ADDRESS_MASK;
	info.tx_buf[2] = data;
	info.tx_buf[3] = AMS_READ;

	if (info.spi_mess.complete)
		return spi_async(info.p_spi_dev, &info.spi_mess);
	return spi_sync(info.p_spi_dev, &info.spi_mess);
}

//-----------------------------------------------------------------------------
int ams_command_write_cont_async(unsigned char cmd,
		unsigned char address, unsigned char *buf, unsigned char buf_len,
		void (*complete)(void *context))
{
	if (buf_len > AMS_SPI_BUF_SIZE - 3) return -ENOMEM;

	info.spi_mess.complete = complete;
	info.spi_trans.len = 1;
	info.tx_buf[0] = cmd;

	if (buf_len > 0) {
		info.tx_buf[1] =
			(address & AMS_ADDRESS_MASK) | AMS_CONTINUOUS;
		memcpy((char *)&info.tx_buf[2], buf, buf_len);
		info.spi_trans.len += (buf_len + 1);
	}

	if (info.spi_mess.complete)
		return spi_async(info.p_spi_dev, &info.spi_mess);
	return spi_sync(info.p_spi_dev, &info.spi_mess);
}
//-----------------------------------------------------------------------------
// Fifo handling: rxxbuf + last bytes left in fifo
int ams_read_fifo(char *dbuf, int count)
{
#if 1
	char buf[FIFO_SIZE+1];
	int len2;

	len2 = ams_read(AMS_REG_FIFOSTATUS);
	if (len2 & AMS_FIFOOVF) {
		return -1;
		//dbg("FIFO OVERFLOW!\n");
	}

	len2 &= AMS_FIFOCNT_MASK;
	if (len2 > FIFO_SIZE) {
		dbg("fifo cnt > fifo len?\n");
		len2 = FIFO_SIZE;
	}
	if (len2 > count) len2 = count;
	if (len2 == 0) return 0;

	ams_read_cont(AMS_REG_FIFO-1, buf, len2+1);
	memcpy(dbuf, buf+1, len2);

	return len2;
#else
	char buf[AMS_REG_FIFO - AMS_REG_FIFOSTATUS + FIFO_SIZE];
	int len;

	ams_read_cont(AMS_REG_FIFOSTATUS, buf, AMS_REG_FIFO - AMS_REG_FIFOSTATUS + FIFO_SIZE);

	if (buf[0] & AMS_FIFOOVF) {
		return -1;
		//dbg("FIFO OVERFLOW!\n");
	}

	len = buf[0] & AMS_FIFOCNT_MASK;
	if (len > FIFO_SIZE) {
		//dbg("fifo cnt > fifo len?\n");
		len = FIFO_SIZE;
	}
	if (len > count) len = count;
	if (len == 0) return 0;

	memcpy(dbuf, buf+AMS_REG_FIFO - AMS_REG_FIFOSTATUS, len);

	return len;

#endif
}

/* *********************************************** HW **************************************/

#define CLSYSAOPCTRL0 0x93
#define CLSYSAOPCTRL1 0x84
#define CLSYSAOPCTRL2 0x0A

#define PLLMAIN0	0x4F
#define PLLMAIN1	0xD8
#define PLLMAIN2	0x60 	// 6= 50 kHz reference frequency
#define PLLREF		50

//-----------------------------------------------------------------------------
void ams_on(void) {
	printk("ams on\n");
	set_gpio_output(PWR_N, 0);
	msleep(100);
	set_gpio_output(ENABLE, 1);
	msleep(100);
}

//-----------------------------------------------------------------------------
static unsigned long freqtable[NUM_BANDS][52] = {
		{ 865700, 865900, 866100, 866300, 866500, 866700, 866900, 867100, 867300, 867500 },
		{ 902250, 902750, 903250, 903750, 904250, 904750, 905250, 905750, 906250, 906750,
				907250, 907750, 908250, 908750, 909250, 909750, 910250, 910750, 911250, 911750,
				912250, 912750, 913250, 913750, 914250, 914750, 915250, 915750, 916250, 916750,
				917250, 917750, 918250, 918750, 919250, 919750, 920250, 920750, 921250, 921750,
				922250, 922750, 923250, 923750, 924250, 924750, 925250, 925750, 926250, 926750,
				927250, 927750 },
		{ 952000, 953000, 954000 },
};
static int freqcnt[NUM_BANDS] = {10, 52, 3};

void ams_set_next_frequency(void)
{
	if (info.band >= NUM_BANDS || info.band < 0) info.band = 0;

	// fixed channel?
	if (info.fixed_channel >= 0 && info.fixed_channel < freqcnt[info.band]) {
		if (info.channel == info.fixed_channel) return;
		info.channel = info.fixed_channel;
		ams_set_frequency(freqtable[info.band][info.channel]);
		return;
	}

	// do frequency hopping, advance lfsr
	if (info.lfsr == 0) info.lfsr = 1;
	info.lfsr = (info.lfsr >> 1) ^ (unsigned int)((0 - (info.lfsr & 1u)) & 0xd0000001u);
	if (info.channel == info.lfsr % freqcnt[info.band]) return;
	info.channel = info.lfsr % freqcnt[info.band];
	ams_set_frequency(freqtable[info.band][info.channel]);
}

int ams_get_freq(int band, int channel)
{
	if (band < 0 || band >= NUM_BANDS) return -EAGAIN;
	if (channel < 0 || channel >= freqcnt[band]) return -EAGAIN;
	return freqtable[band][channel];
}

//-----------------------------------------------------------------------------
// Sets the main PLL frequency
void ams_set_frequency(unsigned long frequency)
{
	// get the reference frequency
	unsigned char buf[8];
	unsigned char var;
	unsigned int i, j, x, reg_A, reg_B;
	unsigned long divisor;

	// Shut down the RF power
	var = ams_read(AMS_REG_STATUSCTRL);
	var &= ~(0x0D);
	ams_write(AMS_REG_STATUSCTRL, var);

	ams_command(AMS_CMD_BLOCKRX);

	divisor = frequency / PLLREF;

    i = 0x3FFF & (divisor >> 6); // division by 64
    x = (i << 6) + i;
    x -= divisor;
    j = i;
    do {
    	if (x >= 33) {
    		i--;
    		x -= 33;
    	}
    	if (x >= 32) {
    		j--;
    		x -= 32;
    	}
    } while (x >= 32);

    reg_A = i - x;
    reg_B = j + x;

    buf[2] = (PLLMAIN2 & 0xF0) | ((reg_B >> 6) & 0x0F);
	buf[1] = ((reg_B << 2) & 0xFC) | ((reg_A >> 8) & 0x03);
	buf[0] = reg_A & 0xFF;

	ams_write_cont(AMS_REG_PLLMAIN, buf, 3);
	ams_command(AMS_CMD_HOPMAINFREQ);

	if (info.chip_version >= 0x35) {
		buf[0] = CLSYSAOPCTRL0;
		buf[1] = CLSYSAOPCTRL1 & ~0x80;
		buf[2] = CLSYSAOPCTRL2;
		ams_write_cont(AMS_REG_CLSYSAOCPCTRL, buf, 3);
		buf[1] |= 0x80;   // set the auto Bit
		ams_write_cont(AMS_REG_CLSYSAOCPCTRL, buf, 3);
	}

	for(i=0; i<100; i++) {
		var	= ams_read(AMS_REG_AGCINTERNALSTATUS);
		if (var & 0x02) break;
		msleep(1);
	}
	if (i == 100) {
		printk(KERN_ERR "AS399x: Error: PLL did not lock\n");
		// do not enable tx/tx
		return;
	}

	ams_command(AMS_CMD_ENABLERX);

	if (info.chip_version >= CHIP_AS3992) {
		msleep(10);
		ams_command(AMS_CMD_TRIGGERRXFCALIBR);
		msleep(10);
	}

	// Switch ON the RF power
	var = ams_read(AMS_REG_STATUSCTRL);
	var |= 0x01;
	ams_write(AMS_REG_STATUSCTRL, var);

	// perform AGL
	var |= (info.autogain << 2) & 0x0C;
	ams_write(AMS_REG_STATUSCTRL, var);
}
//-----------------------------------------------------------------------------
void ams_set_rxgain(signed char gain) {
	unsigned char nr05, nr0a;
	if (info.chip_version < CHIP_AS3992) {
		if (gain > 0) gain = 0;
		if (gain < -7) gain = -7;
		nr05 = info.r05;
		nr0a = info.r0a & 0x1F;
		nr0a |= ((-gain) << 5) & 0xE0;
	} else {
		if (gain > 3) gain = 3;
		if (gain < -3) gain = -3;
		nr05 = info.r05 & 0x7F;
		nr0a = info.r0a & 0x3F;
		if (gain <= 0) {
			nr0a |= ((-gain) << 6) & 0xC0;
		} else {
			nr0a |= (gain << 6) & 0xC0;
			nr05 |= 0x80;
		}
	}

	if (nr05 != info.r05) {
		ams_write(AMS_REG_TRCALREGGEN2MISC, nr05);
		info.r05 = nr05;
	}
	if (nr0a != info.r0a) {
		ams_write(AMS_REG_RXSPECIAL2, nr0a);
		info.r0a = nr0a;
	}
}
//-----------------------------------------------------------------------------
void ams_configure(void)
{
	unsigned char buf[3];
	int rxlinkcfg[] = {0x00, 0x03, 0x06, 0x08, 0x09, 0x0C, 0x0F};
	int rxfilter[] = {0x00, 0x00, 0x01, 0x02, 0x02, 0x02, 0x03};
	int rxfilterv5[] = {0xFF, 0xFF, 0x3B, 0x3B, 0x30, 0x20, 0x00};
	int rxlinkfreqs[] = RXLINKFREQS;
	int taris[] = TARIS;

	int tpri, trcal, rtcal, rxwait, rxnoresp;

	if (info.dr == 0)
		trcal = 1000 * 1000 * 8 / rxlinkfreqs[info.rxlinkfreq] * 1000; // ns
	else
		trcal = 1000 * 1000 * 64 / rxlinkfreqs[info.rxlinkfreq] * 1000 / 3;
	rtcal = (3 * taris[info.tari]) / 2; // ns
	tpri = 1000 * 1000 * 1000 / rxlinkfreqs[info.rxlinkfreq]; // ns
	dbg("tari: %dns; rtcal: %dns\n", taris[info.tari], rtcal);
	dbg("tpri: %dns; trcal: %dns; rxlink: %dkHz\n", tpri, trcal, rxlinkfreqs[info.rxlinkfreq]/1000);

	// link timing parameters, according to epc c1g2 spec table 6.13 (t1min, t1max)
	// nxnoresp: add 10 tpri cycles, see AMS_REG_RXNORESPWAIT register description in ds.
	rxwait = (((rtcal > tpri*10)?rtcal:(tpri*10)) / 1000 * (100-FT) / 100) -2; // us
	rxnoresp = (((rtcal > tpri*10)?rtcal:(tpri*10)) / 1000 * (100+FT) / 100) +2 + tpri / 100; // us
	// just to be sure, take twice the timeout.
	rxnoresp = rxnoresp * 2;
	dbg("rxwait: %dus; rxnoresp %dus\n", rxwait, rxnoresp);

	rxwait = rxwait * 10 / 64; // in 6.4us steps
	if (rxwait > 0xFF) rxwait = 0xFF;
	rxnoresp = rxnoresp * 10/ 256 + 1; // in 25.6 us steps
	if (rxnoresp > 0xFF) rxnoresp = 0xFF;
	trcal = trcal / 100; // in 100ns steps

	if (info.m<=1) info.trext = 1; // short preamble is supported only for M4, M8
	dbg("modulation: %d; trext: %d\n", info.m, info.trext);

	ams_write(AMS_REG_STATUSCTRL, 0x02);   // Enable receiver
	// Receive using CRC, select EPC Gen2, Miller RX Decoding, Tari
	ams_write(AMS_REG_PROTOCOLCTRL, 0x00 | (info.tari & 0x03) | ((info.m << 2) & 0x0C));
	// EPC Gen2 TX Options: PW len = 0.5Tari, TX one len = 2.00Tari
	ams_write(AMS_REG_TXOPTGEN2, 0xF0);
	if (info.chip_version < CHIP_AS3992) { // AS3990, AS3991
		// EPC Gen2 RX Options: Link freq, trext preamble
		ams_write(AMS_REG_RXOPTGEN2, (info.trext?0x02:0x00) | ((rxlinkcfg[info.rxlinkfreq] << 4) & 0xF0));
	} else {
		// EPC Gen2 RX Options: Link freq, trext preamble, pilot reduce
		ams_write(AMS_REG_RXOPTGEN2, (info.trext?0x02:0x00) |
				((rxlinkcfg[info.rxlinkfreq] << 4) & 0xF0) | ((info.pilotreduce << 2) & 0xC0));
	}
	// TRcal Low ans Misc Registers, TRcal, do not decrease vdd_mix, no io_low, no open-drain
	ams_write(AMS_REG_TRCALREGGEN2, (trcal & 0xFF));
	info.r05 = 0x00 | ((trcal >> 8) & 0x0F);
	ams_write(AMS_REG_TRCALREGGEN2MISC, info.r05);
	// Delayed Transmission Wait Time Disabled
	ams_write(AMS_REG_TXREPLYINSLOT, 0x00);
    // RX No Response Wait Time
	ams_write(AMS_REG_RXNORESPWAIT, rxnoresp);
	// RX Wait Time
	ams_write(AMS_REG_RXWAITTIME, rxwait);

	info.r0a = 0;
	if (!(info.flags & AMS_DIFF_INPUT)) info.r0a |= 0x04;
	if (info.chip_version < CHIP_AS3992) { // AS3990, AS3991
		// RX Special Setting (undocumented bit 6 is set, filter settings)
		ams_write(AMS_REG_RXSPECIAL, 0x40 | 0x08 | ((info.m == 0)?0x04:0x00) | (rxfilter[info.rxlinkfreq] & 0x03));
		// RX Special Setting 2 (Set RxHyst, gain is set later)
		if (info.rxhyst > 0x03) info.rxhyst = 0x03;
		info.r0a |= (info.rxhyst << 3) & 0x18;
	} else {
		// RX Special Setting (filter settings)
		ams_write(AMS_REG_RXSPECIAL, rxfilterv5[info.rxlinkfreq]);
		// RX Special Setting 2 (Set RxHyst)
		info.r0a |= (info.rxhyst << 3) & 0x38;
		// Mixer extra gain +10dB
		if (info.rxflags & RX_MIXAMPL10) info.r0a |= 0x02;
	}
	// Mixer input attenuation -8dB
	if (info.rxflags & RX_MIXATTEN8) info.r0a |= 0x01;
	// RX Special Setting 2 (RxGain, RxHyst, input select, input attenuation / amplification)
	ams_write(AMS_REG_RXSPECIAL2, info.r0a);
	// fill in RxGain field
	ams_set_rxgain(info.rxgain);

	if (info.flags & AMS_EXTERNAL_PA) {
		// Regulator and IO control (PA reg 3V, diff out, 14mA low-out pre-driver)
		ams_write(AMS_REG_REGIOCONTROL, 0x52);

		// Modulator control (modulator connected to aux)
		buf[0] = 0x00;
		buf[1] = 0x3F;
		buf[2] = 0x20;
		ams_write_cont(AMS_REG_MODULATORCTRL, buf, 3);
	} else {
		// Regulator and IO control (PA reg 3V, diff out, 14mA main PA pre-driver)
		ams_write(AMS_REG_REGIOCONTROL, 0x58);

		// Modulator control (modulator connected to main)
		buf[0] = 0x00;
		buf[1] = 0x3F;
		buf[2] = 0x40;
		ams_write_cont(AMS_REG_MODULATORCTRL, buf, 3);
	}

	ams_write(AMS_REG_IRQMASK, IRQMASK_ALL);

	buf[0] = CLSYSAOPCTRL0;	// high cp current 20 MHz SYS clock
	buf[1] = CLSYSAOPCTRL1;
	buf[2] = CLSYSAOPCTRL2;
	ams_write_cont(AMS_REG_CLSYSAOCPCTRL, buf, 3);

	buf[0] = PLLMAIN0;
	buf[1] = PLLMAIN1;
	buf[2] = PLLMAIN2;
	ams_write_cont(AMS_REG_PLLMAIN, buf, 3);

	if (info.flags & AMS_EXTERNAL_OSC)
		buf[0] = 0x76;
	else
		buf[0] = 0x46;
	buf[1] = 0x18;
	buf[2] = 0xC2;
	ams_write_cont(AMS_REG_PLLAUX, buf, 3);

	ams_write(AMS_REG_RXLENGTHUP, 0x00);
	ams_write(AMS_REG_RXLENGTHLOW, 0x00);

	buf[0] = 0x00;
	buf[1] = 0x00;
	buf[2] = 0x00;
	ams_write_cont(AMS_REG_TESTSETTING, buf, 3);

	ams_set_next_frequency();

	msleep(50);
	buf[0] = ams_read(AMS_REG_AGCINTERNALSTATUS);
	if ((buf[0] & 0x01) != 0) {
		dbg(KERN_INFO "Oscillator stable\n");
	} else {
		dbg(KERN_INFO "Oscillator problem\n");
	}

	if ((buf[0] & 0x02) != 0) {
		dbg(KERN_INFO "PLL stable\n");
	} else {
		dbg(KERN_INFO "PLL problem\n");
	}
}

//-----------------------------------------------------------------------------
void ams_off(void)
{
	unsigned long flags;
	dbg("ams off\n");

	// avoid being stuck at read() with powered off chip
	info.stop_reading = 1;
	info.exit_irq = 1;


	// disable the irq to be sure that we do not get irq storm
	spin_lock_irqsave(&irq_lock, flags);
	if (!info.irq_disabled) {
		info.irq_disabled= 1;
		disable_irq_nosync(info.pdata->irq);
	}
	spin_unlock_irqrestore(&irq_lock, flags);

	info.channel = -1;

	// power down!
	set_gpio_output(ENABLE, 0); //disable chip
	set_gpio_output(PWR_N, 1); // power down 5V
}

/* ********************************************* ASYNC **************************************/
void ams_async_handler(void *context)
{
	int ret;
	cmd_t *chain = info.irq_cmd_chain;

	// End-of-chain. Fall out of processing.
	if (chain == NULL || *chain == NULL) {
		info.irq_async = 0;
		wake_up(&irq_wait_queue);
		return;
	}

	// Advance chain pointer
	info.irq_cmd_chain++;
	if (*info.irq_cmd_chain == NULL)
		info.irq_cmd_chain = NULL;

	// Call the next function in the chain
	ret = (*chain)();
	// Error? Fall out of processing
	if (ret != 0) {
		info.irq_flags |= SIRQ_GEN;
		info.irq_async = 0;
		wake_up(&irq_wait_queue);
	}
}

//-----------------------------------------------------------------------------
int ams_cmd_resetfifo(void)
{
	return ams_command_async(AMS_CMD_RESETFIFO, ams_async_handler);
}
//-----------------------------------------------------------------------------
void ams_cmd_read_rn16_complete(void *context)
{
	int status = info.rx_buf[1];
	int len = status & AMS_FIFOCNT_MASK;
	if (status & AMS_FIFOOVF || len < 2) {
		info.irq_flags |= SIRQ_GEN;
		info.irq_async = 0;
		wake_up(&irq_wait_queue);
		dbg("Exit @ read_reqrn_complete, status %04x\n", status);
		return;
	}
	info.rn16 = info.rx_buf[5];
	info.rn16 |= info.rx_buf[3] << 8;

	ams_async_handler(context);
}

int ams_cmd_read_rn16(void)
{
	info.tx_buf[0] = (AMS_REG_FIFOSTATUS & AMS_ADDRESS_MASK) | AMS_READ;
	info.tx_buf[1] = 0;
	info.tx_buf[2] = (AMS_REG_FIFO & AMS_ADDRESS_MASK) | AMS_READ;
	info.tx_buf[3] = 0;
	info.tx_buf[4] = (AMS_REG_FIFO & AMS_ADDRESS_MASK) | AMS_READ;
	info.tx_buf[5] = 0;

	info.spi_mess.complete = ams_cmd_read_rn16_complete;
	info.spi_trans.len = 6;

	return spi_async(info.p_spi_dev, &info.spi_mess);
}

/* ****************************************** IRQ **************************************/
// Prepare all variables for going from user to irq mode
void ams_prepare_irq()
{
	info.irq_flags = 0;
	info.irq_cmd_chain = NULL;
	info.exit_irq = 0;
	info.irq_done = IRQ_RX;
	info.irq_err = IRQ_ERR;

	// reset FIFO
	ams_command(AMS_CMD_RESETFIFO);

	// clear int flags
	ams_read(AMS_REG_IRQSTATUS);
}

// User mode: wait until async/irq mode finishes or timeout occurs
int ams_wait_irq_done(int timeout)
{
	int ret;

	// wait until commands and their ACKs are finished in IRQ handler
	ret = wait_event_interruptible_timeout(irq_wait_queue,
			((info.irq_flags & info.irq_err) ||
					(info.irq_cmd_chain == NULL && (info.irq_flags & info.irq_done))),
			timeout);
	ams_disable_irq_sync();

	if (ret < 0) {
		dbg("Transfer interrupted by signal\n");
		return -EINTR;
	}
	if (ret == 0) {
		dbg("Transfer timeout (cmd %p, flags %04x)\n",
				(info.irq_cmd_chain)?(*info.irq_cmd_chain):NULL,
				info.irq_flags);
		return -ETIMEDOUT;
	}
	if (info.irq_cmd_chain != NULL ||
			!(info.irq_flags & info.irq_done) ||
			(info.irq_flags & info.irq_err)) {
		dbg("Transfer error (cmd %p, flags %04x)\n",
				(info.irq_cmd_chain)?(*info.irq_cmd_chain):NULL,
				info.irq_flags);
		return -EIO;
	}

	return 0;
}

//-----------------------------------------------------------------------------
// IRQ wrappers for calling with spi async commands
// user mode -> irq mode
void ams_enable_irq(void *context)
{
	unsigned long flags;
	spin_lock_irqsave(&irq_lock, flags);
	if (info.irq_disabled && !info.exit_irq) {
		info.irq_disabled = 0;
		enable_irq(info.pdata->irq);
	}
	spin_unlock_irqrestore(&irq_lock, flags);
}

// async mode -> irq mode
void ams_enable_irq_unlock(void *context)
{
	unsigned long flags;
	info.irq_async = 0;

	spin_lock_irqsave(&irq_lock, flags);
	//dbg("enable_irq_unlock %d\n", info.pdata->irq);
	if (info.irq_disabled && !info.exit_irq) {
		info.irq_disabled = 0;
		enable_irq(info.pdata->irq);
	} //else dbg("Exit_irq ams_enable_irq_unlock\n");
	spin_unlock_irqrestore(&irq_lock, flags);

	wake_up(&irq_wait_queue);
}

//-----------------------------------------------------------------------------
// user mode: cancel all irq & async mode processing (join)
void ams_disable_irq_sync(void)
{
	unsigned long flags;
	info.exit_irq = 1;

	spin_lock_irqsave(&irq_lock, flags);
	//dbg("disable_irq %d\n", info.pdata->irq);
	if (!info.irq_disabled) {
		disable_irq(info.pdata->irq);
		info.irq_disabled = 1;
	}
	spin_unlock_irqrestore(&irq_lock, flags);

	// wait until the handler finishes
	wait_event(irq_wait_queue,(!info.irq_async));
}

//-----------------------------------------------------------------------------
// External IRQ handler second half, after status reg is read via SPI
static void ams_process_irq(void *context)
{
	int irq_status = info.irqflags_buf->rx[1];

	if (info.exit_irq) {
		dbg("Exit_irq ams_process_irq\n");
		info.irq_async = 0;
		info.irq_flags |= SIRQ_GEN;
	    wake_up(&irq_wait_queue);
		return;
	}

	// collect the flags for main thread
	info.irq_flags |= irq_status;

	// is the processing completed?
	if ((irq_status & info.irq_done) && (info.irq_cmd_chain == NULL)) {
		info.irq_async = 0;
		wake_up(&irq_wait_queue);
		return;
	}

	// If an error occurred, there is no point to continue
	if (irq_status & info.irq_err) {
		//dbg("irq_err: %x\n", irq_status);
		info.irq_async = 0;
		wake_up(&irq_wait_queue);
		return;
	}

	// when IRQ_TX rises, send CMD_RESETFIFO
	if ((info.irq_flags & IRQ_TX) && !(info.irq_flags & SIRQ_FIFORST_DONE)) {
		info.irq_flags |= SIRQ_FIFORST_DONE;
		if (ams_command_async(AMS_CMD_RESETFIFO, ams_enable_irq_unlock) != 0) {
			dbg("RESETFIFO failed");
			info.irq_flags |= SIRQ_GEN;
			info.irq_async = 0;
			wake_up(&irq_wait_queue);
			return;
		}
		return;
	}

	// If we're done and have a follow-up command chain, perform it
	if (info.irq_cmd_chain && (info.irq_flags & IRQ_DONE) == IRQ_RX) {
		info.irq_flags = 0;
		ams_async_handler(NULL);
		return;
	}

	// we're done in async; re-enable IRQs
	info.irq_async = 0;

	if (info.irq_disabled) {
		info.irq_disabled = 0;
		enable_irq(info.pdata->irq);
	}

	// wake up user waitqueue after the ISR has completed
	wake_up(&irq_wait_queue);
}


// External IRQ Handler
static irqreturn_t rfid_ams_isr(int irq, void *dev_id)
{
	if (info.irq_disabled)
	{
		dbg("IRQ Disabled\n");
		return IRQ_HANDLED;
	}

	// disable the irq while we read from SPI the status reg
	info.irq_disabled = 1;
	disable_irq_nosync(info.pdata->irq);

	// schedule the status reg read
	info.irq_async = 1;
	spi_async(info.p_spi_dev, &info.spi_mess_irqflags);

	return IRQ_HANDLED;
}

/* *********************************************** KERNEL DRIVER **************************************/
static struct miscdevice rfid_ams_miscdev =
{
	.minor = AMS_MINOR,
	.name = "rfid_ams",
	.fops = &rfid_ams_fops //declared in epc.c
};

static int __devinit rfid_ams_probe(struct spi_device *spi)
{
	int ret;
	int	err;
	unsigned char buf1[4];
	unsigned char buf2[4];
	const char *chip;

	memset(&info, 0, sizeof(struct rfid_ams));
	info.p_spi_dev = spi;

	/* Default settings */
	info.band = FREQ_EU;
	info.fixed_channel = -1; // frequency hopping

	info.m = 1; // 0:FM0, 1:M2, 2:M4, 3:M8
	info.rxlinkfreq = 1; // 0:40k, 1:80k, 2: 160k...
	info.dr = 0; // 0:8, 1:64/3
	info.tari = 1; // 0:6.2ns, 1:12.5ns, 2:25ns
	info.trext = 1; // 0:short, 1:long
	info.pilotreduce = 0; // reduce pilot tone detection, AS3992 only

	info.rxgain = 0; // decrease gain x*3dB (0...7)
	info.rxhyst = 1; // decrease hysteresis x*5dB (0...3)
	info.autogain = 1; // 0: off, 1:AGC, 2:AGL
	info.rxflags = RX_MIXAMPL10; // attenuation and extra amplification

	info.pdata = spi->dev.platform_data;
	if (info.pdata == NULL) {
		printk(KERN_ERR "RFID_AMS: Platform data is NULL\n");
		return -EINVAL;
	}
	info.flags = info.pdata->flags;

	/* Disable the chip */
	set_gpio_output(ENABLE, 0);
	spi->bits_per_word = 8;
	spi->mode = SPI_MODE_1;
	err = spi_setup(spi);

	if (err < 0)
		return err;

	/* Allocate memory for SPI message data buffers */
	info.tx_buf = (char *)kzalloc(AMS_SPI_BUF_SIZE, GFP_KERNEL);
	if (info.tx_buf == NULL) {
		printk(KERN_ERR "RFID_AMS: Can't get memory for tx_buf\n");
		return -ENOMEM;
	}

	info.rx_buf = (char *)kzalloc(AMS_SPI_BUF_SIZE, GFP_KERNEL);

	if (info.rx_buf == NULL) {
		printk(KERN_ERR "RFID_AMS: Can't get memory for rx_buf\n");
		kfree(info.tx_buf);
		return -ENOMEM;
	}

	info.irqflags_buf = kzalloc(sizeof(struct rfid_ams_spidata_irqflags), GFP_KERNEL);
	if (info.irqflags_buf == NULL) {
		printk(KERN_ERR "RFID_AMS: Can't get memory for irqflags_buf\n");
		kfree(info.tx_buf);
		kfree(info.rx_buf);
		return -ENOMEM;
	}

	/* Init the generic message */
	spi_message_init(&info.spi_mess);
	info.spi_trans.rx_buf = info.rx_buf;
	info.spi_trans.tx_buf = info.tx_buf;
	spi_message_add_tail(&info.spi_trans, &info.spi_mess);

	/* Init the IRQ flags reading message */
	spi_message_init(&info.spi_mess_irqflags);
	info.spi_trans_irqflags.rx_buf = info.irqflags_buf->rx;
	info.spi_trans_irqflags.tx_buf = info.irqflags_buf->tx;
	spi_message_add_tail(&info.spi_trans_irqflags, &info.spi_mess_irqflags);
	info.spi_mess_irqflags.complete = ams_process_irq;
	info.spi_mess_irqflags.context = NULL;
	info.irqflags_buf->tx[0] = (AMS_REG_IRQSTATUS & AMS_ADDRESS_MASK) | AMS_READ;
	info.irqflags_buf->tx[1] = 0;
	info.spi_trans_irqflags.len = 2;

	/* Register MISC device */
	ret = misc_register(&rfid_ams_miscdev);
	if (ret) {
		printk(KERN_ERR "RFID_AMS: Can't misc_register on minor=%d\n", AMS_MINOR);
		goto freebuffers;
	}

	/* Init the driver */
	ams_on();

	ret = ams_read(AMS_REG_VERSIONREG);

	/* Test interface */
	buf1[0] = 0x55;
	buf1[1] = 0xAA;
	buf1[2] = 0x00;
	buf1[3] = 0xFF;
	ams_write_cont(AMS_REG_PLLMAIN, buf1, 3);
	ams_read_cont(AMS_REG_PLLMAIN, buf2, 3);
	if (!memcmp(buf1, buf2, 3) == 0) {
		printk(KERN_INFO "RFID_AMS: Interface test FAILED (%02x %02x %02x %02x, ver %02x)\n", buf2[0], buf2[1], buf2[2], buf2[3], ret);
		ret = -EIO;
		goto poweroff;
	}

	info.chip_version = ams_read(AMS_REG_VERSIONREG);

	if (info.chip_version < 0x35)
		chip = "AS3990";
	else if (info.chip_version < 0x50)
		chip = "AS3991";
	else if (info.chip_version < 0x52)
		chip = "AS3992";
	else {
		printk("RFID_AMS: Unsupported chip rev %d\n", info.chip_version);
		ret = -EINVAL;
		goto poweroff;
	}
	printk(KERN_INFO "Austria Microsystems RFID driver; detected chip %s, rev %02x\n",
			chip, info.chip_version);

	/* Init the IRQ handler */
	ret = request_irq(info.pdata->irq, rfid_ams_isr, IRQF_TRIGGER_HIGH, "ams", NULL);
	if (ret != 0) {
		printk(KERN_ERR "RFID_AMS: unable to request RFID IRQ%d\n", info.pdata->irq);
		ret = -EINVAL;
		goto poweroff;
	}

	ams_off();

return 0;


poweroff:
	ams_off();
	misc_deregister(&rfid_ams_miscdev);

freebuffers:
	kfree(info.irqflags_buf);
	kfree(info.tx_buf);
	kfree(info.rx_buf);

	return ret;
}

//--------------------------------------------------------------------------------
static int __devexit rfid_ams_remove(struct spi_device *spi)
{
	ams_off();
	misc_deregister(&rfid_ams_miscdev);
	kfree(info.irqflags_buf);
	kfree(info.tx_buf);
	kfree(info.rx_buf);
	return 0;
}

//--------------------------------------------------------------------------------
static int rfid_ams_suspend(struct spi_device *spi, pm_message_t message)
{
	info.suspend_powerup = !info.stop_reading;
	ams_off();
	return 0;
}

//--------------------------------------------------------------------------------
static int rfid_ams_resume(struct spi_device *spi)
{
	if (info.suspend_powerup) { // if we were reading, power up the chip again
		ams_on();
		ams_configure();
	}
	return 0;
}

//--------------------------------------------------------------------------------
static struct spi_driver rfid_ams_driver =
{
	.driver = {
		.name	= "rfid_ams",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
	},
	.probe		= rfid_ams_probe,
	.remove		= __devexit_p(rfid_ams_remove),
	.suspend	= rfid_ams_suspend,
	.resume		= rfid_ams_resume,
};

//--------------------------------------------------------------------------------
static int __init rfid_ams_init(void)
{
	return spi_register_driver(&rfid_ams_driver);
}

module_init(rfid_ams_init);

//--------------------------------------------------------------------------------
static void __exit rfid_ams_exit(void)
{
	spi_unregister_driver(&rfid_ams_driver);
}

module_exit(rfid_ams_exit);

//--------------------------------------------------------------------------------

MODULE_DESCRIPTION("Austria Microsystems RFID driver");
MODULE_LICENSE("GPL");
