/*
 * Copyright (c) 2010 Artec Design LLC
 *   Jaanus Sepp <jaanus.sepp@artecdesign.ee>
 *   Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * Driver for Austria Microsystems RFID Reader chip
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. *
*/

//#define DEBUG

#include <linux/platform_device.h>
#include <linux/miscdevice.h>
#include <linux/spi/spi.h>
#include <linux/spinlock.h>
#include <linux/jiffies.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/rfid_ams.h>

#include "ams.h"
#include "epc.h"

#ifdef DEBUG
#define dbg(fmt, args...) printk(fmt, ##args)
#else
#define dbg(fmt, args...) do {} while(0)
#endif

#define CMD_TIMEOUT					msecs_to_jiffies(100)
#define DATA_TIMEOUT				msecs_to_jiffies(1000) // write: at least tx time + 25ms

/* *********************************************** ASYNC callbacks **************************************/
int ams_epc_cmd_ack(void)
{
	unsigned char buf[5];

	buf[0] = TXLEN_H(2,2);
	buf[1] = TXLEN_L(2,2);

	buf[2] = ((EPC_CMD_ACK << 6) & 0xC0) |	// CMD
				((info.rn16 >> 10) & 0x3F);
	buf[3] = (info.rn16 >> 2) & 0xFF;
	buf[4] = (info.rn16 << 6) & 0xC0;

	return ams_command_write_cont_async(AMS_CMD_TRANSM,
				AMS_REG_TXLENGTHUP, buf, 5,
				ams_enable_irq_unlock);
}

//-----------------------------------------------------------------------------
int ams_epc_cmd_reqrn(void)
{
	unsigned char buf[5];

	buf[0] = TXLEN_H(3, 0);
	buf[1] = TXLEN_L(3, 0);

	buf[2] = EPC_CMD_REQRN;
	buf[3] = (info.rn16 >> 8) & 0xFF;
	buf[4] = info.rn16 & 0xFF;

	return ams_command_write_cont_async(AMS_CMD_TRANSMCRC,
					AMS_REG_TXLENGTHUP, buf, 5,
					ams_enable_irq_unlock);
}
/* *********************************************** EPC helpers **************************************/
void ams_epc_select(unsigned char* mask, unsigned char len)
{
	unsigned char buf[64];
	unsigned char i;

	if (len > 31) len = 31;

	buf[0] = TXLEN_H(3+len,5);
	buf[1] = TXLEN_L(3+len,5);

	buf[2] = ((EPC_CMD_SELECT << 4) & 0xF0) |
				((EPC_TARGET_INV_S0 << 1) & 0x0E) |		// Target
				((0x00 >> 2 ) & 0x01);					// Action
	buf[3] = ((0x00 << 6) & 0xC0) |						// Action
				((EPC_MEM_EPC << 4) & 0x30) |			// MEMBANK
				((0x20 >> 4) & 0x0F);					// EBV
	buf[4] = ((0x20 << 4) & 0xF0) |						// EBV
				((len >> 1) & 0x0F);					// Length
	buf[5] = ((len << 7) & 0xF0);						// Length

	for (i = 5; i < (5+len); i++) {						// Mask
		buf[i] |= (0x0F & (*mask >> 4));
		buf[i+1] = 0xF0 & (*mask << 4);
		mask++;
	}

	buf[i] |= ((0x00 << 3) & 0x08);						// Disable Truncation

	ams_command_write_cont_async(AMS_CMD_TRANSMCRC,
					AMS_REG_TXLENGTHUP, buf, len + 6,
					ams_enable_irq);
}

//-----------------------------------------------------------------------------
static void ams_epc_query_standard(unsigned char q)
{
	unsigned char buf[2];

	buf[0] = ((info.dr << 5) & 0x20) |		// DR
				((info.m << 3) & 0x18) |	// M
				((info.trext)?0x04:0x00) | 	// TREXT
				((0x00 << 0) & 0x03);		// SEL
	buf[1] = ((0x00 << 6) & 0xC0) |			// SESSION
				((0x00 << 5) & 0x20) |  	// TARGET
				((q << 1) & 0x1E);			// Q
	ams_command_write_cont_async(AMS_CMD_QUERY,
			AMS_REG_FIFO, buf, 2,
			ams_enable_irq);
}

//-----------------------------------------------------------------------------
static void ams_epc_query_rep(void)
{
	ams_command_async(AMS_CMD_QUERYREP, ams_enable_irq);
}


//-----------------------------------------------------------------------------
static int ams_epc_readerror(int handle)
{
	char buf[6];
	int ret;
	// Read the tag response from buffer
	ret = ams_read_fifo(buf, sizeof(buf));
	if (ret != 5) return -EREMOTEIO;
	if (buf[1] != ((handle>>8) & 0xFF) || buf[2] != (handle & 0xFF))
		return -EREMOTEIO;

	dbg("Tag error %d\n", buf[0]);
	switch(buf[0]) {
	case 0x0:
		return -EREMOTEIO;
	case 0x03:
		return -ENOSPC;
	case 0x04:
		return -EROFS;
	case 0x0B:
		return -ENOSR;
	}
	return -EREMOTEIO;
}

/* *********************************************** EPC **************************************/
// Query for tags and open the first tag that responds.
// args: pId: a buffer to hold tag epc code
//       pIdLen: in: pId length; out: length of epc read
// returns: <0 on error; tag 16-bit rn on success
static int ams_epc_openfirst(unsigned char* pId, int *pIdLen)
{
#define QUERY_Q		1	// at least 1, 0 does not work

	unsigned int slot_count, slot;
	int ret, cnt;

	slot_count = 1 << QUERY_Q;

	// -------------------------------------------------------------------------------------
	// send select command - no reply
	ams_prepare_irq();
	info.irq_done = IRQ_NORX; //we'll get no reply on select
	info.irq_err = IRQ_ERR_NORX;

	ams_epc_select(NULL, 0); // send EPC select command and enable IRQs

	// wait until command is finished in IRQ handler
	ret = ams_wait_irq_done(CMD_TIMEOUT);
	if (ret != 0) return ret;

	// at least 300us delay
	msleep(1);

	// -------------------------------------------------------------------------------------
	// query the tags in timeslots
	for (slot = 0; slot < slot_count; slot ++) {
		// chain of commands to perform on query command answer:
		cmd_t chain[] = {ams_cmd_read_rn16, ams_cmd_resetfifo, ams_epc_cmd_ack, // read RN16, send ACK
				ams_cmd_resetfifo, ams_epc_cmd_reqrn, // ignore ACK response (no time to read it), send reqRN
				ams_cmd_read_rn16, ams_cmd_resetfifo, ams_epc_cmd_ack, NULL }; // read handle, send ACK

		ams_prepare_irq();
		info.irq_cmd_chain = chain;

		// send query or queryrep command and enable IRQ
		if (slot == 0)
			ams_epc_query_standard(QUERY_Q);
		else
			ams_epc_query_rep();

		// wait until commands and their ACKs are finished in IRQ handler
		ret = ams_wait_irq_done(CMD_TIMEOUT);
		if (ret == -EINTR) return ret;
		if (ret == -ETIMEDOUT) return ret;
		if (ret != 0) continue;

		// Read the tag ACK response from buffer
		cnt = ams_read_fifo(pId, *pIdLen);

		/*
		dbg("Got %d bytes from tag:\n", cnt);
		int k;
		for (k = 0; k < cnt; k++)
			dbg("%02X ", pId[k]);
		dbg("\n");
		*/

		if (cnt < 2) {
			dbg("QUERY: low resp count\n");
			continue;
		}

		*pIdLen = cnt;
		return info.rn16;
	}
	return -ENODEV;
}

//-----------------------------------------------------------------------------
// Request a new RN16 for a handle.
static int ams_epc_reqrn(int handle) {
	char rnbuf[3], buf[5];
	int ret;

	buf[0] = TXLEN_H(3, 0);
	buf[1] = TXLEN_L(3, 0);

	buf[2] = EPC_CMD_REQRN; //Command to tag: EPC_READ
	buf[3] = (handle >> 8) & 0xFF;
	buf[4] = handle & 0xFF;

	ams_prepare_irq();
	ams_command_write_cont_async(AMS_CMD_TRANSMCRC, AMS_REG_TXLENGTHUP, buf, 5,	ams_enable_irq);

	// wait until commands and their ACKs are finished in IRQ handler
	ret = ams_wait_irq_done(CMD_TIMEOUT);
	if (ret != 0) return ret;

	// Read the tag response from buffer (use larger buffer to check for size==2)
	ret = ams_read_fifo(rnbuf, sizeof(rnbuf));
	if (ret != 2) {
		//dbg("RN error, ret %d.\n", ret);
		return -EREMOTEIO;
	}

	// return rn in low 16 bits
	return (rnbuf[0] << 8) | (rnbuf[1]);
}

//-----------------------------------------------------------------------------
static int ams_epc_read(unsigned char* pData, int dataLen, int handle, int memBank, int wordPtr, int wordCount)
{
	char buf[8];
	int ret;

	if (dataLen < (wordCount+2)) return -EINVAL;

	// TODO: write support for EBV for multi-byte pointers
	// See UHF C1G2 standard, Annex A
	if (wordPtr > 0x7F) return -2;

	buf[0] = TXLEN_H(5, 2);
	buf[1] = TXLEN_L(5, 2);

	buf[2] = EPC_CMD_READ; //Command to tag: EPC_READ
	buf[3] = ((memBank << 6) & 0xC0) | ((wordPtr >> 2) & 0x3F);
	buf[4] = ((wordPtr << 6) & 0xC0) | ((wordCount >> 2) & 0x3F);
	buf[5] = ((wordCount << 6) & 0xC0) | ((handle >> 10) & 0x3F);
	buf[6] = (handle >> 2) & 0xFF;
	buf[7] = (handle << 6) & 0xC0;

	ams_prepare_irq();
	ams_command_write_cont_async(AMS_CMD_TRANSMCRCEHEAD, AMS_REG_TXLENGTHUP, buf, 8,
			ams_enable_irq);

	// wait until commands and their ACKs are finished in IRQ handler
	ret = ams_wait_irq_done(DATA_TIMEOUT);
	if (ret != 0) return ret;

	// tag replied an error?
	if (info.irq_flags & IRQ_HDR) {
		dbg("Read error reply from tag\n");
		return ams_epc_readerror(handle);
	}

	// Read the tag response from buffer
	ret = ams_read_fifo(pData, dataLen);

	// If we've picked up some trailing garbage to fifo (happens time to time),
	// don't use it.
	if (ret > (2*wordCount) + 2) ret = (2*wordCount) +2;
	if (ret != (2*wordCount) + 2) return -EREMOTEIO; // not enough data

	// Does the RN in reply match?
	if (pData[2*wordCount] != ((handle>>8) & 0xFF) ||
			pData[(2*wordCount)+1] != (handle & 0xFF)) {
		dbg("RN mismatch while reading\n");
		return -EREMOTEIO;
	}

	return ret - 2;
}

static int ams_epc_write(int handle, int memBank, int wordPtr, int data)
{
	char buf[9];
	int ret, rn;

	// TODO: write support for EBV for multi-byte pointers
	// See UHF C1G2 standard, Annex A
	if (wordPtr > 0x7F) return -EINVAL;

	rn = ams_epc_reqrn(handle);
	if (rn < 0) return rn;

	data = data ^ rn; // cover-coding

	buf[0] = TXLEN_H(6, 2);
	buf[1] = TXLEN_L(6, 2);

	buf[2] = EPC_CMD_WRITE; //Command to tag: EPC_WRITE
	buf[3] = ((memBank << 6) & 0xC0) | ((wordPtr >> 2) & 0x3F);
	buf[4] = ((wordPtr << 6) & 0xC0) | ((data >> 10) & 0x3F);
	buf[5] = (data >> 2) & 0xFF;
	buf[6] = ((data << 6) & 0xC0) | ((handle >> 10) & 0x3F);
	buf[7] = (handle >> 2) & 0xFF;
	buf[8] = (handle << 6) & 0xC0;

	// Disable NORX IRQ, otherwise it would block receiver
	ams_write(AMS_REG_IRQMASK, IRQMASK_NORXTIMEOUT);

	ams_prepare_irq();

	info.irq_err = IRQ_ERR_NOTIMEOUT; // ignore chip timeout IRQ, use our own;
	// chip timeout IRQ is too small (max 6ms), EPC C1G2 requires at least 25ms

	ams_command_write_cont_async(AMS_CMD_TRANSMCRCEHEAD, AMS_REG_TXLENGTHUP, buf, 9,
			ams_enable_irq);

	// wait until commands and their ACKs are finished in IRQ handler
	ret = ams_wait_irq_done(DATA_TIMEOUT);

	ams_write(AMS_REG_IRQMASK, IRQMASK_ALL);

	if (ret != 0) return ret;

	// tag replied an error?
	if (info.irq_flags & IRQ_HDR) {
		dbg("Read error reply from tag.\n");
		return ams_epc_readerror(handle);
	}

	// Read the tag response from buffer
	ret = ams_read_fifo(buf, sizeof(buf));

/*
	dbg("Got %d bytes from tag (rn %x, handle %x, irq %x):\n", ret, rn, handle, info.irq_flags);
	int k;
	for (k = 0; k < ret; k++)
		dbg("%02X ", buf[k]);
	dbg("\n");
*/

	if (ret < 2) return -EREMOTEIO;

	// Does the handle in reply match?
	if (buf[0] != ((handle>>8) & 0xFF) || buf[1] != (handle & 0xFF)) {
		//dbg("RN match err\n");
		return -EREMOTEIO;
	}

	return 0;
}


/* *********************************************** MISC DEV **************************************/
static ssize_t rfid_ams_read(struct file *file, char *buf, size_t count, loff_t * ppos)
{
#define MAX_ACK_RESP	70

	int ret = 0, ret2;
	int handle, i;
	struct rfid_ams_hdr retHdr;
	char tagResp[MAX_ACK_RESP]; //2B PC, 0-4B XPC, 0-62B EPC
	int epcLen, xpcLen, dataLen;
	char *buf_w = buf + sizeof(retHdr);

	if (count < 4) return 0;

	dataLen = 2*(info.dataEnd - info.dataStart);
	if (dataLen < 0) dataLen = 0;

	info.stop_reading = 0;

	while (!info.stop_reading) {
		retHdr.ret = -ECANCELED;
		retHdr.idLen = 0;
		retHdr.dataLen = 0;

		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			goto out;
		}

		// find first tag and set it to open state
		ret = sizeof(tagResp);
		handle = ams_epc_openfirst(tagResp, &ret);

		if (handle == -ETIMEDOUT) {
			dbg("Chip timeout while opening tag, resetting\n");
			ams_off();
			msleep(500);
			ams_on();
			ams_configure();
		}

		if (handle < 0)	{
			dbg("Open failed, ret %d\n", handle);
		}

		if (handle >= 0 && ret >= 2) { // parse tag response
			epcLen = (tagResp[0] >> 3) * 2; // in bytes
			xpcLen = 0;
			if (tagResp[0] & 0x02) {
				if (tagResp[3] & 0x01)
					xpcLen = 4;
				else
					xpcLen = 2;
			}

			if (ret >= 2 + epcLen + xpcLen
					&& count >= sizeof(retHdr) + 2 + epcLen + xpcLen + dataLen) {
				// successful epc read
				//dbg("Successful read, epcLen %d, xpcLen %d\n", epcLen, xpcLen);
				retHdr.ret = 0;
				retHdr.idLen = 2 + epcLen + xpcLen;
				retHdr.dataLen = 0;
				/* return the whole backscatter (PC + [XPC1] + [XPC2] + EPC) */
				if (copy_to_user(buf_w, tagResp, 2 + epcLen + xpcLen)){
					ret = -EFAULT;
					goto out;
				}
				buf_w += 2 + epcLen + xpcLen;

				// read tag data, 16 bit blocks, up to 8 blocks at once
				for (i = info.dataStart; i<info.dataEnd; i+=8) {
					int cnt = info.dataEnd - i;
					if (cnt > 8) cnt = 8;
					ret2 = ams_epc_read(tagResp, sizeof(tagResp), handle, info.dataBank, i, cnt);
					if (ret2 == cnt*2) {
						if (copy_to_user(buf_w, tagResp, ret2)){
							ret = -EFAULT;
							goto out;
						}
						buf_w += ret2;
						retHdr.dataLen += ret2;
					} else if (ret2 == -ETIMEDOUT) {
						dbg("Chip timeout while reading, resetting\n");
						ams_off();
						msleep(500);
						ams_on();
						ams_configure();
					} else {
						dbg("Read mem %x, %d-%d failed, ret %d\n", info.dataBank, i, i+cnt, ret2);
						ret = ret2;
						goto out;
					}
				}

				break;
			}
		}

		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			goto out;
		}
		if (file->f_flags & O_NONBLOCK) {
			ret = 0;
			goto out;
		}

		ams_set_next_frequency(); // do frequency hopping
		cpu_relax(); // optimization barrier!
		msleep(100);
	}

	if (copy_to_user(buf, &retHdr, sizeof(retHdr))) return -EFAULT;
	return sizeof(retHdr) + retHdr.idLen + retHdr.dataLen;

out:
	if (signal_pending(current)) {
		ret = -ERESTARTSYS;
	}
	if (!signal_pending(current))
		ams_set_next_frequency(); // do frequency hopping
	return ret;
}

//-----------------------------------------------------------------------------
static int rfid_ams_write_data(short handle, int bank, int offset, int count, const char *data) {
	int i, j, ret=0;
	for (i = 0; i < count; i++) {
		unsigned char datal, datah;
		if (get_user(datal, data++) || get_user(datah, data++))
			return -EINVAL;
		for (j=0; j<TRIES_WRITE; j++) {
			dbg("Writing mem %x, %d\n", bank, offset+i);
			ret = ams_epc_write(handle, bank, offset+i, datal | (datah << 8));
			dbg("Write ret %d\n", ret);
			if (ret >= 0) break;
			if (i==0 && j>=(TRIES_WRITE_FIRST-1)) break;
		}
		if (ret < 0) break;
	}
	return ret;
}

//-----------------------------------------------------------------------------
static ssize_t rfid_ams_write(struct file *file,
					const char *buf, size_t count, loff_t *pos)
{
	int ret = 0;
	int handle;
	struct rfid_ams_write_hdr hdr;
	char tagResp[MAX_ACK_RESP]; //2B PC, 0-4B XPC, 0-62B EPC

	if (count < sizeof(struct rfid_ams_write_hdr)) return -EINVAL;
	if (copy_from_user(&hdr, buf, sizeof(struct rfid_ams_write_hdr)))
		return -EINVAL;
	if (count < sizeof(struct rfid_ams_write_hdr) + hdr.len) return -EINVAL;

	info.stop_reading = 0;
	while (!info.stop_reading) {
		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			goto out;
		}
		// find first tag and set it to open state
		ret = sizeof(tagResp);
		handle = ams_epc_openfirst(tagResp, &ret);

		if (handle == -ETIMEDOUT) {
			dbg("Chip timeout while opening tag, resetting\n");
			ams_off();
			msleep(500);
			ams_on();
			ams_configure();
		}

		if (handle < 0)	{
			dbg("Open failed, ret %d\n", handle);
		}

		if (handle >= 0 && ret >= 2) {
			signed char tmp = 0;
			// If AGC is enabled, disable it and use the last gain AGC got when opening the tag.
			// Otherwise, AGC will turn up the gain too high and some noise triggers the receiver
			// before actual write cmd reply reception (up to 25ms) and we'll exit with preamble error.
			if (info.autogain == 1) {
				tmp = (ams_read(AMS_REG_AGCINTERNALSTATUS) >> 4) & 0x0F;
				ams_set_rxgain(info.rxgain - tmp);

				tmp = ams_read(AMS_REG_STATUSCTRL);
				tmp &= ~(0x0C);
				ams_write(AMS_REG_STATUSCTRL, tmp);
			}

			ret = rfid_ams_write_data(handle, hdr.bank, hdr.offset, hdr.len,
					buf + sizeof(struct rfid_ams_write_hdr));

			// re-enable AGC
			if (info.autogain == 1) {
				tmp |= (info.autogain << 2) & 0x0C;
				ams_write(AMS_REG_STATUSCTRL, tmp);

				ams_set_rxgain(info.rxgain);
			}


			if (ret >= 0) {
				// successful write, return original count
				ret = count;
				goto out;
			}
			// error?
			if (ret < 0) goto out;
			// otherwise, continue
		}

		if (signal_pending(current)) {
			ret = -ERESTARTSYS;
			goto out;
		}
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			goto out;
		}

		ams_set_next_frequency(); // do frequency hopping
		cpu_relax(); // optimization barrier!
		msleep(100);
	}

	return ret;

out:
	if (!signal_pending(current))
		ams_set_next_frequency(); // do frequency hopping
	return ret;
}

//-----------------------------------------------------------------------------

static int rfid_ams_ioctl(struct inode *inode, struct file *file,
								unsigned int cmd, unsigned long arg)
{
	switch(cmd)
	{
	case RFID_AMS_STOP_READ:
		info.stop_reading = 1;
		return 0;

	case RFID_AMS_SET_EPC_BANK:
		if (arg > 0x03) return -EINVAL;
		info.dataBank = arg & 0x03;
		return 0;

	case RFID_AMS_SET_BLOCK_START:
		// We do not support EBV yet, so limit to <127
		if (arg > 0x7F) return -EINVAL;
		info.dataStart = arg & 0x7F;
		return 0;

	case RFID_AMS_SET_BLOCK_END:
		// We do not support EBV yet, so limit to <127
		if (arg > 0x7F) return -EINVAL;
		info.dataEnd = arg & 0x7F;
		return 0;

	case RFID_AMS_SET_CHAN:
		if ((arg & 0xFF) > NUM_BANDS) return -EINVAL;
		info.band = arg & 0xFF;
		info.fixed_channel = (arg >> 16) & 0xFF;
		info.channel = -1;
		ams_set_next_frequency();
		return 0;

	case RFID_AMS_GET_CHAN:
		return ((info.channel << 16) & 0xFFFF0000) | (info.band & 0xFFFF);

	case RFID_AMS_GET_FREQ:
		return ams_get_freq(info.band, info.channel);

	case RFID_AMS_GET_RSSI:
	{
		int ret = ams_read(AMS_REG_RSSILEVELS);
		int tmp = ams_read(AMS_REG_AGCINTERNALSTATUS);
		ret |= ((tmp << 5) & 0x100);
		ret |= ((tmp << 12) & 0xF0000);
		return ret;
	}

	case RFID_AMS_GET_RDC:
	{
		int ret;
		ams_write(AMS_REG_ADCSEL, 0x01);
		ams_command(AMS_CMD_TRIGGERADCCON);
		udelay(10);
		ret = ams_read(AMS_REG_ADC) & 0xFF;
		ams_write(AMS_REG_ADCSEL, 0x02);
		ams_command(AMS_CMD_TRIGGERADCCON);
		udelay(10);
		ret |= (ams_read(AMS_REG_ADC) << 8) & 0xFF00;
		return ret;
	}

	case RFID_AMS_SET_RXTXPARAMS:
		if ((arg & 0xFF) > 2) return -EINVAL;
		if (((arg >> 8) & 0xFF) > 6) return -EINVAL;
		if (((arg >> 16) & 0xFF) > 1) return -EINVAL;
		info.tari = arg & 0xFF;
		info.rxlinkfreq = ((arg >> 8) & 0xFF);
		info.dr = ((arg >> 16) & 0xFF);
		// reload settings
		ams_configure();
		return 0;

	case RFID_AMS_GET_RXTXPARAMS:
	{
		int ret;
		ret = info.tari & 0xFF;
		ret |= (info.rxlinkfreq << 8) & 0xFF00;
		ret |= (info.dr << 16) & 0xFF0000;
		return ret;
	}

	case RFID_AMS_SET_RXMODPARAMS:
		if ((arg & 0xFF) > 3) return -EINVAL;
		if (((arg >> 8) & 0xFF) > 1) return -EINVAL;
		if (((arg >> 16) & 0xFF) > 3) return -EINVAL;
		info.m = arg & 0xFF;
		info.trext = ((arg >> 8) & 0xFF);
		info.pilotreduce = ((arg >> 16) & 0xFF);
		// reload settings
		ams_configure();
		return 0;

	case RFID_AMS_GET_RXMODPARAMS:
	{
		int ret;
		ret = info.m & 0xFF;
		ret |= (info.trext << 8) & 0xFF00;
		ret |= (info.pilotreduce << 16) & 0xFF0000;
		return ret;
	}

	case RFID_AMS_SET_RXGAINPARAMS:
		if (((arg >> 8) & 0xFF) > 7) return -EINVAL;
		if (((arg >> 16) & 0xFF) > 2) return -EINVAL;
		info.rxgain = arg & 0xFF;
		info.rxhyst = ((arg >> 8) & 0xFF);
		info.autogain = ((arg >> 16) & 0xFF);
		info.rxflags = ((arg >> 24) & 0xFF);
		// reload settings
		ams_configure();
		return 0;

	case RFID_AMS_GET_RXGAINPARAMS:
	{
		int ret;
		ret = info.rxgain & 0xFF;
		ret |= (info.rxhyst << 8) & 0xFF00;
		ret |= (info.autogain << 16) & 0xFF0000;
		ret |= (info.rxflags << 24) & 0xFF000000;
		return ret;
	}
	}

	return -EINVAL;
}

static int rfid_ams_open(struct inode *inode, struct file *file)
{
	ams_on();
	ams_configure();
	return 0;
}

static int rfid_ams_release(struct inode *inode, struct file *file)
{
	ams_off();
	return 0;
}


//--------------------------------------------------------------------------------
struct file_operations rfid_ams_fops =
{
      .owner = THIS_MODULE,
      .read = rfid_ams_read,
      .write = rfid_ams_write,
      .ioctl = rfid_ams_ioctl,
      .open = rfid_ams_open,
      .release = rfid_ams_release,
};
