/*
 * Copyright (c) 2007 Artec Design LLC
 *   Jaanus Sepp <jaanus.sepp@artecdesign.ee>
 *
 * Driver for NXP RFID Reader chip
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. *
*/

//--------------------------------------------------------------------------------

//#define DEBUG 1

//--------------------------------------------------------------------------------

#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/miscdevice.h>
#include <linux/spi/spi.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/rfid_nxp.h>
#include <asm/gpio.h>

//--------------------------------------------------------------------------------

#define RFID_NXP_MINOR                    51

// macros for I/O port manipulation

#define set_gpio_value(pin, value) \
  at91_set_gpio_value(nxpinfo.pdata->pin_assignment[RFID_NXP_PIN_##pin], (value) != 0)

#define get_gpio_value(pin) \
  at91_get_gpio_value(nxpinfo.pdata->pin_assignment[RFID_NXP_PIN_##pin])

#define set_gpio_output(pin, value) \
  at91_set_gpio_output(nxpinfo.pdata->pin_assignment[RFID_NXP_PIN_##pin], (value) != 0)

#define set_gpio_input(pin, use_pullup) \
  at91_set_gpio_input(nxpinfo.pdata->pin_assignment[RFID_NXP_PIN_##pin], (use_pullup) != 0)

#define set_A_periph(pin, use_pullup) \
  at91_set_A_periph(nxpinfo.pdata->pin_assignment[RFID_NXP_PIN_##pin], (use_pullup) != 0)

#define set_B_periph(pin, use_pullup) \
  at91_set_B_periph(nxpinfo.pdata->pin_assignment[RFID_NXP_PIN_##pin], (use_pullup) != 0)

#define RFID_NXP_TAG_RETRIES            0x03
#define RFID_NXP_FIFO_SIZE                64

#define RFID_NXP_REG_PAGESELECT         0x00   // selects the register page
#define RFID_NXP_REG_COMMAND            0x01   // starts and stops the command execution
#define RFID_NXP_REG_FIFODATA           0x02   // input and output of 64 byte FIFO buffer
#define RFID_NXP_REG_PRIMARYSTATUS      0x03   // status flags of the receiver, transmitter and FIFO buffer
#define RFID_NXP_REG_FIFOLENGTH         0x04   // number of bytes buffered in the FIFO
#define RFID_NXP_REG_SECONDARYSTATUS    0x05   // diverse status flags
#define RFID_NXP_REG_INTERRUPTEN        0x06   // control bits to enable and disable passing of interrupt requests
#define RFID_NXP_REG_INTERRUPTRQ        0x07   // interrupt request flags
#define RFID_NXP_REG_CONTROL            0x09   // diverse control flags (timer, power saving, ...)
#define RFID_NXP_REG_ERRORFLAG          0x0A   // error flags showing the error status of the last command executed
#define RFID_NXP_REG_COLLPOS            0x0B   // bit position of the first bit collision detected on the RF-interface
#define RFID_NXP_REG_TIMERVALUE         0x0C   // actual value of the timer
#define RFID_NXP_REG_CRCRESULTLSB       0x0D   // LSB of the CRC-coprocessor register
#define RFID_NXP_REG_CRCRESULTMSB       0x0E   // MSB of the CRC-coprocessor register
#define RFID_NXP_REG_BITFRAMING         0x0F   // adjustments for bit oriented frames
#define RFID_NXP_REG_TXCONTROL          0x11   // controls the logical behaviour of the antenna pins TX1 and TX2
#define RFID_NXP_REG_CWCONDUCTANCE      0x12   // selects the conductance of the antenna driver pins TX1 and TX2
#define RFID_NXP_REG_MODCONDUCTANCE     0x13   // defines the driver output conductance
#define RFID_NXP_REG_CODERCONTROL       0x14   // sets the clock rate and the coding mode
#define RFID_NXP_REG_MODWIDTH           0x15   // selects the width of the modulation pulse
#define RFID_NXP_REG_MODWIDTHSOF        0x16   // selects the width of the modulation pulse for SOF (I*Code Fast Mode)
#define RFID_NXP_REG_TYPEBFRAMING       0x17   // defines the framing for ISO14443B communication
#define RFID_NXP_REG_RXCONTROL1         0x19   // controls receiver behaviour
#define RFID_NXP_REG_DECODERCONTROL     0x1A   // controls decoder behaviour
#define RFID_NXP_REG_BITPHASE           0x1B   // selects the bit-phase between transmitter and receiver clock
#define RFID_NXP_REG_RXTHRESHOLD        0x1C   // selects the thresholds for the bit decoder
#define RFID_NXP_REG_BPSKDEMCONTROL     0x1D   // control BPSK receiver behaviour
#define RFID_NXP_REG_RXCONTROL2         0x1E   // controls decoder behaviour and defines the input source for the receiver
#define RFID_NXP_REG_CLOCKQCONTROL      0x1F   // controls clock generation for the 90' phase shifted Q-channel clock
#define RFID_NXP_REG_RXWAIT             0x21   // selects the time interval after transmission, before receiver starts
#define RFID_NXP_REG_CHANNELREDUNDANCY  0x22   // selects the mode of checking the data integrity on the RF-channel
#define RFID_NXP_REG_CRCPRESETLSB       0x23   // LSB of the pre-set value of the CRC register
#define RFID_NXP_REG_CRCPRESETMSB       0x24   // MSB of the pre-set value of the CRC register
#define RFID_NXP_REG_TIMESLOTPERIOD     0x25   // selects the time between automatically transmitted frames
#define RFID_NXP_REG_MFOUTSELECT        0x26   // selects internal signal applied to pin MFOUT
#define RFID_NXP_REG_FIFOLEVEL          0x29   // defines level for FIFO over- and underflow warning
#define RFID_NXP_REG_TIMERCLOCK         0x2A   // selects the divider for the timer clock
#define RFID_NXP_REG_TIMERCONTROL       0x2B   // selects start and stop conditions for the timer
#define RFID_NXP_REG_TIMERRELOAD        0x2C   // defines the pre-set value for the timer
#define RFID_NXP_REG_IRQPINCONFIG       0x2D   // configures the output stage of pin IRQ

// masks for RFID_NXP_REG_INTERRUPTRQ

#define RFID_NXP_REG_INTERRUPTRQ_SETIRQ  0x80   // determine if the flag bits will be set or cleared
#define RFID_NXP_REG_INTERRUPTRQ_ALL     0x3F   // all interrupt flags
#define RFID_NXP_REG_INTERRUPTRQ_TIMER   0x20   // timer IRQ flag
#define RFID_NXP_REG_INTERRUPTRQ_TX      0x10   // TX IRQ flag
#define RFID_NXP_REG_INTERRUPTRQ_RX      0x08   // RX IRQ flag
#define RFID_NXP_REG_INTERRUPTRQ_IDLE    0x04   // IDLE IRQ flag
#define RFID_NXP_REG_INTERRUPTRQ_HIALERT 0x02   // HiAlert IRQ flag
#define RFID_NXP_REG_INTERRUPTRQ_LOALERT 0x01   // LoAlert IRQ flag

// masks for RFID_NXP_REG_CONTROL

#define RFID_NXP_REG_CONTROL_STANDBY             0x20   // StandBy bit
#define RFID_NXP_REG_CONTROL_POWERDOWN           0x10   // PowerDown bit
#define RFID_NXP_REG_CONTROL_FLUSHFIFO           0x01   // set this bit to flush FIFO

// commands for register RFID_NXP_REG_COMMAND

#define RFID_NXP_REG_COMMAND_IDLE                0x00   // IDLE command
#define RFID_NXP_REG_COMMAND_TRANCEIVE           0x1E   // TRANCEIVE command
#define RFID_NXP_REG_COMMAND_STARTUP             0x3F   // STARTUP command activated by the NXP chip

// masks for RFID_NXP_REG_PRIMARYSTATUS
#define RFID_NXP_REG_PRIMARYSTATUS_MASK_MODEMSTATE   0x70
#define RFID_NXP_REG_PRIMARYSTATUS_SHIFT_MODEMSTATE  0x04

#define RFID_NXP_REG_PRIMARYSTATUS_MODEMSTATE_IDLE   0x00

// masks for RFID_NXP_REG_TXCONTROL
#define RFID_NXP_REG_TXCONTROL_TX1RFEN               0x01
#define RFID_NXP_REG_TXCONTROL_TX2RFEN               0x02

// masks for RFID_NXP_REG_DECODERCONTROL
#define RFID_NXP_REG_DECODERCONTROL_RXMULTIPLE       0x40

// masks for RFID_NXP_REG_CHANNELREDUNDANCY
#define RFID_NXP_REG_CHANNELREDUNDANCY_TXCRCEN       0x04
#define RFID_NXP_REG_CHANNELREDUNDANCY_RXCRCEN       0x08

// macros

#define RFID_NXP_FLUSHFIFO()            nxp_reg_modify(RFID_NXP_REG_CONTROL, RFID_NXP_REG_CONTROL_FLUSHFIFO, 0)
#define RFID_NXP_MODEMSTATE()           ((nxp_reg_read(RFID_NXP_REG_PRIMARYSTATUS) &   \
                                          RFID_NXP_REG_PRIMARYSTATUS_MASK_MODEMSTATE) >>    \
                                          RFID_NXP_REG_PRIMARYSTATUS_SHIFT_MODEMSTATE)
#define RFID_NXP_TRANSMITTER_ON()       nxp_reg_modify(RFID_NXP_REG_TXCONTROL,         \
                                          RFID_NXP_REG_TXCONTROL_TX1RFEN | RFID_NXP_REG_TXCONTROL_TX2RFEN, 0)
#define RFID_NXP_TRANSMITTER_OFF()      nxp_reg_modify(RFID_NXP_REG_TXCONTROL,         \
                                          0, RFID_NXP_REG_TXCONTROL_TX1RFEN | RFID_NXP_REG_TXCONTROL_TX2RFEN)
#define RFID_NXP_CRC_ENABLE()           nxp_reg_modify(RFID_NXP_REG_CHANNELREDUNDANCY, \
                                          RFID_NXP_REG_CHANNELREDUNDANCY_TXCRCEN | RFID_NXP_REG_CHANNELREDUNDANCY_RXCRCEN, 0)
#define RFID_NXP_CRC_DISABLE()          nxp_reg_modify(RFID_NXP_REG_CHANNELREDUNDANCY, \
                                          0, RFID_NXP_REG_CHANNELREDUNDANCY_TXCRCEN | RFID_NXP_REG_CHANNELREDUNDANCY_RXCRCEN)

//--------------------------------------------------------------------------------

#define CascLevel1           0x93
#define CascLevel2           0x95
#define CascLevel3           0x97

static const unsigned char cascConsts[] = { CascLevel1, CascLevel2, CascLevel3 };

#define iso14443AREQA        0x26     // REQA command
#define iso14443ARead        0x30     // read command
#define iso14443AHLTA        0x50     // HLTA command
#define iso14443AWUPA        0x52     // WUPA command
#define iso14443AWrite       0xA2     // write command

//--------------------------------------------------------------------------------

#define RFID_STATE_HARD_POWERDOWN    0
#define RFID_STATE_SOFT_POWERDOWN    1
#define RFID_STATE_READING           2

#define RFID_NXP_MAX_BLOCKS			256	// use only multiples of 32

struct rfid_nxp {
  struct rfid_nxp_plat_data *pdata;     // platform data that specifies the GPIO pins
  struct spi_device      *p_spi_dev;
  struct spi_message     spi_mess;
  struct spi_transfer    spi_trans;
  char *tx_buf;
  char *rx_buf;
  char state;
  unsigned long protocols_in_use;
  unsigned long blocks_in_use[RFID_NXP_MAX_BLOCKS >> 5];
  spinlock_t state_lock;
} nxpinfo;

/* *********************************************** FUNCTIONAL PART **************************************/

//-----------------------------------------------------------------------------

static unsigned char nxp_reg_read(unsigned char addr) {
  nxpinfo.tx_buf[0] = (addr << 1) | 0x80;
  nxpinfo.tx_buf[1] = 0;
  nxpinfo.spi_trans.len = 2;
  spi_sync(nxpinfo.p_spi_dev, &nxpinfo.spi_mess);
  return nxpinfo.rx_buf[1];
}

//-----------------------------------------------------------------------------

static void nxp_reg_write(unsigned char addr, unsigned char data) {
  nxpinfo.tx_buf[0] = addr << 1;
  nxpinfo.tx_buf[1] = data;
  nxpinfo.spi_trans.len = 2;
  spi_sync(nxpinfo.p_spi_dev, &nxpinfo.spi_mess);
}

//-----------------------------------------------------------------------------

static void nxp_reg_modify(unsigned char addr, unsigned char setbits, unsigned char clearbits) {
  unsigned char regvalue = nxp_reg_read(addr);
  regvalue |= setbits;
  regvalue &= ~clearbits;
  nxp_reg_write(addr, regvalue);
}

//-----------------------------------------------------------------------------

static char nxp_block_needed(int block)
{
	char result = 0;

	if (block < RFID_NXP_MAX_BLOCKS)
	{
		spin_lock(&nxpinfo.state_lock);
		result = (nxpinfo.blocks_in_use[block >> 5] & (1 << (block & 0x1F))) != 0;
		spin_unlock(&nxpinfo.state_lock);
	}

	return result;
}

//-----------------------------------------------------------------------------

static int nxp_remove_unneeded_blocks(char *buf,
		int first_block, int block_count, int block_size)
{
	int src;
	int dest = 0;
	int block = first_block;

	for (src = 0; src < block_count; src++)
	{
		if (nxp_block_needed(block))
		{
			if (src > dest)
				memcpy(&buf[dest * block_size], &buf[src * block_size], block_size);
			dest++;
		}

		block++;
	}

	return dest;
}

//-----------------------------------------------------------------------------

static void configure_iso15693(void) {
  nxp_reg_write(RFID_NXP_REG_RXCONTROL2, 0x01);
  nxp_reg_write(RFID_NXP_REG_CODERCONTROL, 0x2F);
  nxp_reg_write(RFID_NXP_REG_MODWIDTH, 0x3F);
  nxp_reg_write(RFID_NXP_REG_RXCONTROL1, 0x8B);
  nxp_reg_write(RFID_NXP_REG_DECODERCONTROL, 0x54);
  nxp_reg_write(RFID_NXP_REG_BITPHASE, 0xD2);
  nxp_reg_write(RFID_NXP_REG_RXTHRESHOLD, 0x68);
  nxp_reg_write(RFID_NXP_REG_CLOCKQCONTROL, 0x47);
  nxp_reg_write(RFID_NXP_REG_RXWAIT, 0x08);
  nxp_reg_write(RFID_NXP_REG_CHANNELREDUNDANCY, 0x20);
  nxp_reg_write(RFID_NXP_REG_CRCPRESETLSB, 0xFF);
  nxp_reg_write(RFID_NXP_REG_CRCPRESETMSB, 0xFF);
  nxp_reg_write(RFID_NXP_REG_TXCONTROL, 0x5B);
}

//-----------------------------------------------------------------------------

static void configure_iso14443a(void) {
  nxp_reg_write(RFID_NXP_REG_RXCONTROL2, 0x81);
  nxp_reg_write(RFID_NXP_REG_CODERCONTROL, 0x19);
  nxp_reg_write(RFID_NXP_REG_MODWIDTH, 0x13);
  nxp_reg_write(RFID_NXP_REG_RXCONTROL1, 0x73);
  nxp_reg_write(RFID_NXP_REG_DECODERCONTROL, 0x48);
  nxp_reg_write(RFID_NXP_REG_BITPHASE, 0xAD);
  nxp_reg_write(RFID_NXP_REG_RXTHRESHOLD, 0x88);
  nxp_reg_write(RFID_NXP_REG_CLOCKQCONTROL, 0x45);
  nxp_reg_write(RFID_NXP_REG_RXWAIT, 0x06);
  nxp_reg_write(RFID_NXP_REG_CHANNELREDUNDANCY, 0x03);
  nxp_reg_write(RFID_NXP_REG_CRCPRESETLSB, 0x63);
  nxp_reg_write(RFID_NXP_REG_CRCPRESETMSB, 0x63);
  nxp_reg_write(RFID_NXP_REG_TXCONTROL, 0x5B);
}

//-----------------------------------------------------------------------------
// NOTE: Other value pairs for CWCONDUCTANCE and MODCONDUCTANCE that were
// tested to work almost as well as the values below are:
// 0x25, 0x15 (modulation depth 17%)
// 0x28, 0x09 (modulation depth 21%)
// 0x28, 0x14 (modulation depth 28%)
// 0x34, 0x15 (modulation depth 21%)
// 0x1F, 0x0A (modulation depth 18%)

// NOTE2: It seems that the BITPHASE and MODWIDTH registers do NOT influence the
// ISO14443B tag access. But since nothing is 100% sure, just fix these registers
// to values that have been tested to work.

static void configure_iso14443b(void) {
  nxp_reg_write(RFID_NXP_REG_RXCONTROL2, 0x81);
  nxp_reg_write(RFID_NXP_REG_MODCONDUCTANCE, 0x15);
  nxp_reg_write(RFID_NXP_REG_CODERCONTROL, 0x20);
  nxp_reg_write(RFID_NXP_REG_MODWIDTH, 0x13);
  nxp_reg_write(RFID_NXP_REG_TYPEBFRAMING, 0x00);
  nxp_reg_write(RFID_NXP_REG_RXCONTROL1, 0x73);
  nxp_reg_write(RFID_NXP_REG_DECODERCONTROL, 0x59);
  nxp_reg_write(RFID_NXP_REG_BITPHASE, 0xBB);
  nxp_reg_write(RFID_NXP_REG_RXTHRESHOLD, 0x56);
  nxp_reg_write(RFID_NXP_REG_CLOCKQCONTROL, 0x56);
  nxp_reg_write(RFID_NXP_REG_RXWAIT, 0x03);
  nxp_reg_write(RFID_NXP_REG_CHANNELREDUNDANCY, 0x20);
  nxp_reg_write(RFID_NXP_REG_CRCPRESETLSB, 0xFF);
  nxp_reg_write(RFID_NXP_REG_CRCPRESETMSB, 0xFF);
  nxp_reg_write(RFID_NXP_REG_TXCONTROL, 0x4B);
}

//-----------------------------------------------------------------------------

static char nxp_tag_transaction(char *ptr, char bytes, char *outbuf, char max_outbytes, char use_crc) {
  int fifocount = 0;
  unsigned char result;

  if (bytes > RFID_NXP_FIFO_SIZE) return 0;
  if (use_crc) RFID_NXP_CRC_ENABLE(); else RFID_NXP_CRC_DISABLE();  // enable or disable CRC generator and checker
  RFID_NXP_FLUSHFIFO();                                  // flush FIFO

  nxpinfo.tx_buf[0] = RFID_NXP_REG_FIFODATA << 1;
  memcpy(&nxpinfo.tx_buf[1], ptr, bytes);
  nxpinfo.spi_trans.len = bytes + 1;
  spi_sync(nxpinfo.p_spi_dev, &nxpinfo.spi_mess);

  nxp_reg_write(RFID_NXP_REG_INTERRUPTRQ, RFID_NXP_REG_INTERRUPTRQ_RX);   // clear RX flag
  nxp_reg_write(RFID_NXP_REG_COMMAND, RFID_NXP_REG_COMMAND_TRANCEIVE);    // launch TRANCEIVE command

  // wait the command to complete
  msleep(30);
  result = nxp_reg_read(RFID_NXP_REG_INTERRUPTRQ);
  nxp_reg_write(RFID_NXP_REG_COMMAND, RFID_NXP_REG_COMMAND_IDLE);

  if ((result & RFID_NXP_REG_INTERRUPTRQ_RX) != 0) {
    fifocount = nxp_reg_read(RFID_NXP_REG_FIFOLENGTH);
    if ((fifocount > 0) && (fifocount < RFID_NXP_FIFO_SIZE)) {    // read out the FIFO
      if (fifocount > max_outbytes) {
        fifocount = 0;
      } else {
        nxpinfo.tx_buf[0] = (RFID_NXP_REG_FIFODATA << 1) | 0x80;
        if (fifocount > 1) memset(&(nxpinfo.tx_buf[1]), RFID_NXP_REG_FIFODATA << 1, fifocount - 1);
        nxpinfo.tx_buf[fifocount] = 0;
        nxpinfo.spi_trans.len = fifocount + 1;
        spi_sync(nxpinfo.p_spi_dev, &nxpinfo.spi_mess);
        memcpy(outbuf, nxpinfo.rx_buf + 1, fifocount);
      }
    }
  }

  return fifocount;
}

//-----------------------------------------------------------------------------

static void nxp_on(void) {
  int i;

  spin_lock(&nxpinfo.state_lock);
  switch (nxpinfo.state) {
    case RFID_STATE_HARD_POWERDOWN:
      set_gpio_output(PWR_N, 0);				// power pin
      set_gpio_output(PWRDWN, 0);                // nreset/pwrdwn pin: power up chip
      spin_unlock(&nxpinfo.state_lock);
      for (i = 0; i < 100; i++) {
    	  //if the chip is down/dead, we read from bus 00. So
    	  //read another register that is not 00, too
        if (nxp_reg_read(RFID_NXP_REG_COMMAND) == RFID_NXP_REG_COMMAND_IDLE &&
        		nxp_reg_read(RFID_NXP_REG_PAGESELECT) == 0x80) break;
        msleep(10);
      }
      if (i == 100) nxp_reg_write(RFID_NXP_REG_COMMAND, RFID_NXP_REG_COMMAND_IDLE);
      spin_lock(&nxpinfo.state_lock);               // NOTE: NO break, fall through
    case RFID_STATE_SOFT_POWERDOWN:
      nxpinfo.state = RFID_STATE_READING;
      spin_unlock(&nxpinfo.state_lock);
      nxp_reg_write(RFID_NXP_REG_COMMAND, RFID_NXP_REG_COMMAND_IDLE);
    	nxp_reg_write(RFID_NXP_REG_PAGESELECT, 0); // set linear 6-bit addressing mode
      nxp_reg_modify(RFID_NXP_REG_CONTROL, 0, RFID_NXP_REG_CONTROL_STANDBY | RFID_NXP_REG_CONTROL_POWERDOWN);
      nxp_reg_write(RFID_NXP_REG_CWCONDUCTANCE, 0x25);
      msleep(10);
      break;
    default:
      spin_unlock(&nxpinfo.state_lock);
      break;
  }
}

//-----------------------------------------------------------------------------

static void nxp_off(void) {
	RFID_NXP_TRANSMITTER_OFF();
  nxp_reg_modify(RFID_NXP_REG_CONTROL, RFID_NXP_REG_CONTROL_POWERDOWN, 0);
  spin_lock(&nxpinfo.state_lock);
  nxpinfo.state = RFID_STATE_SOFT_POWERDOWN;
  spin_unlock(&nxpinfo.state_lock);
}

//-----------------------------------------------------------------------------

static char iso14443a_reqa(void) {
  unsigned char result;

  RFID_NXP_CRC_DISABLE();                                // disable CRC generator and checker
  RFID_NXP_FLUSHFIFO();                                  // flush FIFO

  nxp_reg_write(RFID_NXP_REG_BITFRAMING, 7);             // write only 7 bits with next I/O
  nxp_reg_write(RFID_NXP_REG_FIFODATA, iso14443AREQA);   // write REQA to FIFO
  nxp_reg_write(RFID_NXP_REG_INTERRUPTRQ, RFID_NXP_REG_INTERRUPTRQ_RX);   // clear RX flag
  nxp_reg_write(RFID_NXP_REG_COMMAND, RFID_NXP_REG_COMMAND_TRANCEIVE);    // launch TRANCEIVE command

  // wait the command to complete
  msleep(10);
  result = nxp_reg_read(RFID_NXP_REG_INTERRUPTRQ);
  nxp_reg_write(RFID_NXP_REG_COMMAND, RFID_NXP_REG_COMMAND_IDLE);

  if ((result & RFID_NXP_REG_INTERRUPTRQ_RX) != 0) {
    if (nxp_reg_read(RFID_NXP_REG_FIFOLENGTH) == 2) return 1;  // we don't check ATQA
  }

  return 0;
}

//-----------------------------------------------------------------------------

static char iso14443a_get_id(unsigned char cascade_level, unsigned char *tag_id, char *is_last_level) {
  unsigned char buf[5];
  unsigned char i;
  unsigned char XorByte = 0;

  buf[0] = cascConsts[cascade_level];
  buf[1] = 0x20;

  if (nxp_tag_transaction(buf, 2, buf, 5, 0) != 5) return 0;   // wrong length

  for (i = 0; i < 5; i++) XorByte ^= buf[i];
  if (XorByte) return 0;               // checksum wrong
  if (buf[0] == 0x88) {                // more cascade levels
    if (cascade_level == 2) return 0;  // must be last level
    memcpy(&(tag_id[3 * cascade_level]), &(buf[1]), 3);
  } else {                             // last cascade level
    memcpy(&(tag_id[3 * cascade_level]), buf, 4);
    *is_last_level = 1;
  }

  return 1;
}

//-----------------------------------------------------------------------------

static char iso14443a_select(unsigned char cascade_level, unsigned char *tag_id, char *is_last_level) {
  unsigned char buf[7];
  unsigned char i = 0;
  unsigned char XorByte = 0;

  buf[0] = cascConsts[cascade_level];
  buf[1] = 0x70;               // select 56 valid bits

  if (!(*is_last_level)) {
    buf[2] = 0x88;
    i++;
  }
  memcpy(buf + 2 + i, tag_id + 3 * cascade_level, 4 - i);
  for (i = 0; i < 4; i++) XorByte ^= buf[2 + i];
  buf[6] = XorByte;
  if (nxp_tag_transaction(buf, 7, buf, 1, 1) != 1) return 0;
  if (buf[0] & 0x04) return 1;                             // UID not complete
  return 2;                                                // UID complete
}


//-----------------------------------------------------------------------------

static char iso14443a_inventory(unsigned char *tagId) {
  char is_last_level = 0;
  int i, j;
  unsigned char buf[10];

	if (!iso14443a_reqa()) return 0;
  for (i = 0; i < 3; i++) {
    if (!iso14443a_get_id(i, buf, &is_last_level)) return 0;
    switch (iso14443a_select(i, buf, &is_last_level)) {
      case 0:
        return 0;
      case 2:
        for (j = 0; j < 3 * i + 4; j++) tagId[j] = buf[3 * i + 3 - j];
        return 3 * i + 4;
    }
  }

  return 0;
}

//-----------------------------------------------------------------------------

// NB! Block 0 refers to the serial number area, normal user data
// starts from block 4. We don't need to prevent reading of these locations,
// but in case on some day we will implement block writing, we need to ensure
// that the lock bits and OTP bytes will not be accidentally written!

static char iso14443a_read_block(unsigned char block, char *buf)
{
	int i;

	for (i = 0; i < RFID_NXP_TAG_RETRIES; i++)
	{
		buf[0] = iso14443ARead;
		buf[1] = block;
		if (nxp_tag_transaction(buf, 2, buf, 16, 1) == 16) return 16;
	}

	return 0;
}

//-----------------------------------------------------------------------------

static int iso14443a_read_memory(char *buf)
{
	int current_block = 0;
	int bytes_read = 0;

	// currently we only support reading the memory of Mifare UltraLight
	// and this tag has 16 blocks of data.

	while (current_block < 16)
	{
		if (!nxp_block_needed(current_block))
		{
			// we don't need this block
			current_block++;
			continue;
		}

		if (iso14443a_read_block(current_block, &(buf[bytes_read])) != 16)
			return -1;			// communication error

		{
			// We got 4 blocks of data (16 bytes), check if we need all of it
			int cur_max_blocks;
			int cur_copied_blocks;

			cur_max_blocks = 4;
			if (current_block + cur_max_blocks > 16)
				cur_max_blocks = 16 - current_block;

			cur_copied_blocks =	nxp_remove_unneeded_blocks(&buf[bytes_read],
									current_block, cur_max_blocks, 4);

			bytes_read += cur_copied_blocks * 4;
			current_block += 4;
		}
	}

	return bytes_read;
}


//-----------------------------------------------------------------------------

static char iso15693_inventory(unsigned char *tagId) {
  unsigned char buf[10];
  int i;

  buf[0] = 0x26;  // flags
  buf[1] = 0x01;  // INVENTORY command
  buf[2] = 0x00;  // mask length

  if (nxp_tag_transaction(buf, 3, buf, 10, 1) != 10) return 0;
  if (buf[0] != 0) return 0;
  for (i = 0; i < 8; i++) tagId[i] = buf[9 - i];
  return 8;
}

//-----------------------------------------------------------------------------

static char iso15693_read_block(int first_block,
		int block_count, int block_size, char *data, char *tagId)
{
	unsigned char buf[64];
	int i, j;
	int count;

	for (i = 0; i < RFID_NXP_TAG_RETRIES; i++)
	{
		buf[0] = 0x22;                 // flags
		buf[1] = 0x23;                 // read multiple blocks command
		for (j = 0; j < 8; j++) buf[j + 2] = tagId[7 - j];  // command is ADDRESSED
		buf[10] = (unsigned char)(first_block & 0xFF);
		buf[11] = (unsigned char)((block_count - 1) & 0xFF);

		count = nxp_tag_transaction(buf, 12, buf, 64, 1);

		if (count != (block_count * block_size + 1)) continue;
		if (buf[0] != 0) continue;

		memcpy(data, &(buf[1]), count - 1);
		return count - 1;
	}

	return 0;
}

//-----------------------------------------------------------------------------

static int iso15693_read_memory(char *buf, char *tagId)
{
	int block_size;
	int block_count;
	int blocks_now;
	int cur_copied_blocks;
	int bytes_read = 0;
	int current_block = 0;

	switch (tagId[1])
	{
	case 0x04:					// Philips tag
		block_size  = 4;		// 4 bytes per block
		block_count = 28;		// 28 blocks, ie 112 bytes total
		break;

	case 0x05:					// Infineon tag
		switch (tagId[2])
		{
		case 0x00:
			block_size  = 4;	// 4 bytes per block
			block_count = 248;	// 248 blocks, ie 992 bytes total
			break;

		case 0x02:
			block_size  = 8;	// 8 bytes per block
			block_count = 29;	// 29 blocks, ie 232 bytes total
			break;

		default:
			block_size  = 0;
			block_count = 0;	// unknown tag, assume no readable memory
			break;
		}
		break;

	case 0x07:					// Texas Instruments tag
		block_size  = 4;		// 4 bytes per block
		block_count = 64;		// 64 blocks, ie 512 bytes
		break;

	default:
		block_size  = 0;
		block_count = 0;		// unknown tag, assume no readable memory
		break;
	}


	while (current_block < block_count)
	{
		if (!nxp_block_needed(current_block))
		{
			// we don't need this block
			current_block++;
			continue;
		}

		// max block count that fits into the NXP chip buffer
		blocks_now = 60 / block_size;

		// decrease if we have reached to the end of tag
		if (current_block + blocks_now > block_count)
			blocks_now = block_count - current_block;

		// we could read max_blocks_now, see if we actually need all of it
		while (!nxp_block_needed(current_block + blocks_now - 1))
		{
			blocks_now--;    // we don't need this block
		}

		if (iso15693_read_block(current_block, blocks_now,
				block_size, &(buf[bytes_read]), tagId) != blocks_now * block_size)
			return -1;         // communication error

		cur_copied_blocks = nxp_remove_unneeded_blocks(&buf[bytes_read],
							current_block, blocks_now, block_size);

		bytes_read += cur_copied_blocks * block_size;
		current_block += blocks_now;
	}

	return bytes_read;
}

//-----------------------------------------------------------------------------

static char iso14443b_select(void) {
  char buf[2];

  buf[0] = 0x06;  // initiate
  buf[1] = 0x00;
  if (nxp_tag_transaction(buf, 2, buf, 1, 1) == 1) {
    buf[1] = buf[0];
    buf[0] = 0x0E;
    if (nxp_tag_transaction(buf, 2, buf, 1, 1) == 1) return 1;
  }

  return 0;
}

//-----------------------------------------------------------------------------

static char iso14443b_inventory(unsigned char *tagId) {
  unsigned char buf[2];

  int i;

  if (!iso14443b_select()) return 0;

  for (i = 0; i < 4; i++) {
    buf[0] = 0x08;
    buf[1] = i;
    if (nxp_tag_transaction(buf, 2, buf, 2, 1) != 2) return 0;
    tagId[7 - 2 * i] = buf[0];
    tagId[6 - 2 * i] = buf[1];
  }

  return 8;
}

//-----------------------------------------------------------------------------

// NB! Block 0 refers to the UID area, normal user data starts from block 4.
// Block 15 refers to the OTP LOCK and CHIP ID area, so there is 11 blocks
// of normal user data area. We don't need to prevent reading from the system
// area, but in case on some day we will implement block writing, we need to ensure
// that the OTP LOCK bits will not be accidentally written!

static char iso14443b_read_block(unsigned char block, char *buf)
{
	int i;

	for (i = 0; i < RFID_NXP_TAG_RETRIES; i++)
	{
		buf[0] = 0x08;
		buf[1] = block;
		if (nxp_tag_transaction(buf, 2, buf, 2, 1) == 2)
			return 2;
	}

	return 0;
}

//-----------------------------------------------------------------------------

static int iso14443b_read_memory(char *buf)
{
	int current_block = 0;
	int bytes_read = 0;

	// currently we only support reading the memory of ST SR176 chip
	// and this tag has 16 blocks of data.

	while (current_block < 16)
	{
		if (!nxp_block_needed(current_block))
		{
			// we don't need this block
			current_block++;
			continue;
		}

		if (iso14443b_read_block(current_block, &(buf[bytes_read])) != 2)
			return -1;			// communication error

		bytes_read += 2;
		current_block++;
	}

	return bytes_read;
}

/* *********************************************** USERSPACE INTERFC **************************************/

static ssize_t rfid_nxp_read(struct file *file, char *buf, size_t count, loff_t * ppos)
{
	char idLen;
	char tagId[16];
	int dataLen = 0;
	char *data;
	#define MAX_DATA	1024
	char protocol = 0;
	char state_copy;
	unsigned long protocols_copy;

	if (count < 4) return 0;

	while(1)
	{
		spin_lock(&nxpinfo.state_lock);
		state_copy = nxpinfo.state;
		protocols_copy = nxpinfo.protocols_in_use;
		spin_unlock(&nxpinfo.state_lock);

		if (state_copy != RFID_STATE_READING) break;	// end requested
		if (protocols_copy == 0) break;					// no protocols selected

		if (signal_pending(current)) {
			return -ERESTARTSYS;
		}

		while ((protocols_copy & (1 << protocol)) == 0)
		{
			protocol++;
			if (protocol == 32)
			{
				RFID_NXP_TRANSMITTER_OFF();
				protocol = 0;

				if (file->f_flags & O_NONBLOCK) {
					return 0;
				}

				msleep(40);		// additional delay before starting a new round
			}
		}

		idLen = 0;
		dataLen = -1;

		data = kmalloc(MAX_DATA, GFP_KERNEL);
		if (!data) return -ENOMEM;

		switch (protocol)
		{
		case RFID_NXP_PROTOCOL_ISO15693:

			configure_iso15693();
			msleep(10);
			idLen = iso15693_inventory(tagId);
			if (idLen == 8) dataLen = iso15693_read_memory(data, tagId);
			if (dataLen == -1) {
				dataLen = 0;	// ignore error, then we can at least return the tag ID
			}
			
			break;

		case RFID_NXP_PROTOCOL_ISO14443A:

			configure_iso14443a();
			msleep(10);
			idLen = iso14443a_inventory(tagId);
			if (idLen == 7) dataLen = iso14443a_read_memory(data);
			if (dataLen == -1) {
				dataLen = 0;	// ignore error, then we can at least return the tag ID
			}
			
			break;

		case RFID_NXP_PROTOCOL_ISO14443B:

			configure_iso14443b();
			msleep(10);
			idLen = iso14443b_inventory(tagId);
			if (idLen == 8) dataLen = iso14443b_read_memory(data);
			if (dataLen == -1) {
				dataLen = 0;	// ignore error, then we can at least return the tag ID
			}
			
			break;
		}

		if ((idLen > 0) && (dataLen > -1))
		{
			buf[0] = 0;			// status = success
			buf[1] = idLen;
			*((short int *)(&(buf[2]))) = dataLen;

			count -= 4;			// exclude the header bytes

			// copy tag ID to output buffer
			if (count < idLen) idLen = count;
			count -= idLen;
			if (idLen > 0) memcpy(&(buf[4]), tagId, idLen);

			// copy tag data to output buffer
			if (count < dataLen) dataLen = count;
			if (dataLen > 0) memcpy(&(buf[idLen + 4]), data, dataLen);
			
			kfree(data);

			return idLen + dataLen + 4;
		}
		kfree(data);

		protocol++;
		msleep(10);
	}

	memset(buf, 0, 4);
	buf[0] = -ECANCELED;

	return 4;
}

//-----------------------------------------------------------------------------

// NOTE: The driver write interface is currently used for testing and tuning only, do NOT
// use this in the application!

static ssize_t rfid_nxp_write(struct file *file, const char *buf, size_t count, loff_t *pos) {
 /* int bitphase, rxthres;
  if (count < 5) count = -EINVAL; else {
    sscanf(buf, "%2x %2x", &bitphase, &rxthres);
    nxp_reg_write(RFID_NXP_REG_BITPHASE, bitphase);
    nxp_reg_write(RFID_NXP_REG_RXTHRESHOLD, rxthres);
  }
  printk("BitPhase = %02X, RxThreshold = %02X\n", nxp_reg_read(RFID_NXP_REG_BITPHASE), nxp_reg_read(RFID_NXP_REG_RXTHRESHOLD));*/
  return count;
}

//-----------------------------------------------------------------------------

static int rfid_nxp_ioctl(struct inode *inode, struct file *file,
								unsigned int cmd, unsigned long arg)
{
	int i;

	switch(cmd)
	{
	case RFID_NXP_STOP_READ:

		spin_lock(&nxpinfo.state_lock);
		nxpinfo.state = RFID_STATE_SOFT_POWERDOWN;
		spin_unlock(&nxpinfo.state_lock);
		return 0;

	case RFID_NXP_SET_PROTOCOLS:

		spin_lock(&nxpinfo.state_lock);
		nxpinfo.protocols_in_use = arg;
		spin_unlock(&nxpinfo.state_lock);
		return 0;

	case RFID_NXP_SET_BLOCK_ON:

		if (arg < RFID_NXP_MAX_BLOCKS)
		{
			spin_lock(&nxpinfo.state_lock);
			nxpinfo.blocks_in_use[arg >> 5] |= (1 << (arg & 0x1F));
			spin_unlock(&nxpinfo.state_lock);
			return 0;
		}
		break;

	case RFID_NXP_SET_BLOCK_OFF:

		if (arg < RFID_NXP_MAX_BLOCKS)
		{
			spin_lock(&nxpinfo.state_lock);
			nxpinfo.blocks_in_use[arg >> 5] &= ~(1 << (arg & 0x1F));
			spin_unlock(&nxpinfo.state_lock);
			return 0;
		}
		break;

	case RFID_NXP_SET_ALL_BLOCKS_ON:

		spin_lock(&nxpinfo.state_lock);
		for (i = 0; i < RFID_NXP_MAX_BLOCKS >> 5; i++)
			nxpinfo.blocks_in_use[i] = 0xFFFFFFFF;
		spin_unlock(&nxpinfo.state_lock);
		return 0;

	case RFID_NXP_SET_ALL_BLOCKS_OFF:

		spin_lock(&nxpinfo.state_lock);
		for (i = 0; i < RFID_NXP_MAX_BLOCKS >> 5; i++)
			nxpinfo.blocks_in_use[i] = 0;
		spin_unlock(&nxpinfo.state_lock);
		return 0;
	}

	return -EINVAL;
}

static int rfid_nxp_open(struct inode *inode, struct file *file)
{
	nxp_on();
	return 0;
}

static int rfid_nxp_release(struct inode *inode, struct file *file)
{
	nxp_off();
	return 0;
}


/* *********************************************** MISC DEV PROBE, REMOVE etc **************************************/

//--------------------------------------------------------------------------------

// misc device init params: file operations
static struct file_operations rfid_nxp_fops =
{
      .owner = THIS_MODULE,
      .read = rfid_nxp_read,
      .write = rfid_nxp_write,
      .open = rfid_nxp_open,
      .release = rfid_nxp_release,
      .ioctl = rfid_nxp_ioctl,
};

//--------------------------------------------------------------------------------

// misc dev init params
static struct miscdevice rfid_nxp_miscdev =
{
	.minor = RFID_NXP_MINOR,
	.name = "rfid_nxp",
	.fops = &rfid_nxp_fops
};

//--------------------------------------------------------------------------------

static int rfid_nxp_suspend(struct spi_device *spi, pm_message_t message)
{
	spin_lock(&nxpinfo.state_lock);
	nxpinfo.state = RFID_STATE_HARD_POWERDOWN;
	spin_unlock(&nxpinfo.state_lock);
	set_gpio_output(PWRDWN, 0); // nreset/pwrdwn pin: power down pin to avoid leakage
	set_gpio_output(PWR_N, 1); // power_n pin: shut down power
	return 0;
}

//--------------------------------------------------------------------------------

static int rfid_nxp_resume(struct spi_device *spi)
{
	return 0;
}

//--------------------------------------------------------------------------------

static int __devinit rfid_nxp_probe(struct spi_device *spi)
{
	int ret;
	int	err;

	memset(&nxpinfo, 0, sizeof(struct rfid_nxp));
	nxpinfo.protocols_in_use = (1 << RFID_NXP_PROTOCOL_ISO15693) |
							(1 << RFID_NXP_PROTOCOL_ISO14443A) |
							(1 << RFID_NXP_PROTOCOL_ISO14443B);
	nxpinfo.p_spi_dev = spi;

	{
		int i;
		for (i = 0; i < RFID_NXP_MAX_BLOCKS >> 5; i++)
			nxpinfo.blocks_in_use[i] = 0xFFFFFFFF;
	}

	nxpinfo.pdata = spi->dev.platform_data;
	if (nxpinfo.pdata == NULL)
	{
		printk(KERN_ERR "Platform data is NULL\n");
		return EINVAL;
	}

	spi->bits_per_word = 8;
	spi->mode = SPI_MODE_0;
	err = spi_setup(spi);
	if (err < 0)
		return err;


	nxpinfo.tx_buf = (char *)kzalloc(128, GFP_KERNEL);
	if (nxpinfo.tx_buf == NULL) return -ENOMEM;

	nxpinfo.rx_buf = (char *)kzalloc(128, GFP_KERNEL);
	if (nxpinfo.rx_buf == NULL)
	{
		kfree(nxpinfo.tx_buf);
		return -ENOMEM;
	}

	spi_message_init(&nxpinfo.spi_mess);
	nxpinfo.spi_trans.rx_buf = nxpinfo.rx_buf;
	nxpinfo.spi_trans.tx_buf = nxpinfo.tx_buf;
	nxpinfo.spi_trans.cs_change = 0;
	spi_message_add_tail(&nxpinfo.spi_trans, &nxpinfo.spi_mess);

	ret = misc_register(&rfid_nxp_miscdev);
	if (ret)
	{
		printk(KERN_ERR "Can't misc_register on minor=%d\n", RFID_NXP_MINOR);
		goto freebuffers;
	}

	spin_lock(&nxpinfo.state_lock);
	nxpinfo.state = RFID_STATE_HARD_POWERDOWN;
	spin_unlock(&nxpinfo.state_lock);
	set_gpio_output(PWRDWN, 0); // nreset/pwrdwn pin: avoid leak
	set_gpio_output(PWR_N, 1); // power_n pin: shut down power

	printk(KERN_INFO "NXP RFID reader\n");
	goto out;

freebuffers:
	kfree(nxpinfo.tx_buf);
	kfree(nxpinfo.rx_buf);

out:
	return 0;
}

//--------------------------------------------------------------------------------

static int __devexit rfid_nxp_remove(struct spi_device *spi)
{
	dev_dbg(&spi->dev, "unregistered rfid_nxp\n");
	misc_deregister(&rfid_nxp_miscdev);
	kfree(nxpinfo.rx_buf);
	kfree(nxpinfo.tx_buf);
	return 0;
}

/* *********************************************** KERNEL DRIVER STUFF **************************************/


// spi_driver - laiendatud platform_driver
static struct spi_driver rfid_nxp_driver =
{
	.driver = {
		.name	= "rfid_nxp",
		.bus	= &spi_bus_type,
		.owner	= THIS_MODULE,
	},
	.probe		= rfid_nxp_probe,
	.remove		= __devexit_p(rfid_nxp_remove),
	.suspend	= rfid_nxp_suspend,
	.resume		= rfid_nxp_resume,
};

//--------------------------------------------------------------------------------

static int __init rfid_nxp_init(void)
{
	return spi_register_driver(&rfid_nxp_driver);
}

module_init(rfid_nxp_init);

//--------------------------------------------------------------------------------

static void __exit rfid_nxp_exit(void)
{
	spi_unregister_driver(&rfid_nxp_driver);
}

module_exit(rfid_nxp_exit);

//--------------------------------------------------------------------------------

MODULE_DESCRIPTION("RFID NXP driver");
MODULE_LICENSE("GPL");
