/*
 * Generic AT91 Timer-connected beeper driver
 *
 * Copyright (C) 2007 Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * based on ixp4xx-beeper.c
 *  Copyright (C) 2005 Tower Technologies
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. *
 */

#include <linux/clk.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <mach/hardware.h>
#include <mach/at91_tc.h>
#include <mach/gpio.h>

MODULE_AUTHOR("Anti Sullin <anti.sullin@artecdesign.ee>");
MODULE_DESCRIPTION("at91 beeper driver");
MODULE_LICENSE("GPL");

static DEFINE_SPINLOCK(beep_lock);

/* Register access macros */
#define tc_readl(reg) \
	readl((info.membase) + AT91_TC_##reg)
#define tc_writel(reg,value) \
	writel((value), (info.membase) + AT91_TC_##reg)



// draiveri globaalsed muutujad
static struct at91_beeper {
	void __iomem 	*membase;
	struct platform_device *pdev;
	struct clk	*clk;
	unsigned int clkfreq;
	struct input_dev *input_dev;
	unsigned int shdwn_pin;
} info;



static void at91_spkr_control(unsigned int freq)
{
	unsigned int div;
	unsigned long flags;

	spin_lock_irqsave(&beep_lock, flags);

	if (freq) {
		div=info.clkfreq / freq;
		tc_writel(IDR, 0xFF); // disable all int's
		tc_writel(CCR, AT91_TC_CLKEN); // enable clk
		tc_writel(CCR, AT91_TC_SWTRG); // trigger clk start
		tc_writel(CMR, AT91_TC_TIMER_CLOCK4 |
			AT91_TC_WAVE | AT91_TC_WAVESEL_UP_AUTO |
			AT91_TC_ACPA_CLEAR | AT91_TC_ACPC_SET | AT91_TC_ASWTRG_CLEAR);
		tc_writel(RA, div>>1); // pulse width 50%
		tc_writel(RC, div); // frequency == timer length
	} else {
		tc_writel(CCR, AT91_TC_CLKDIS); // disable clk
		tc_writel(CCR, AT91_TC_SWTRG); // trigger output to go to inactive
	}
	spin_unlock_irqrestore(&beep_lock, flags);
}

static int at91_spkr_event(struct input_dev *dev, unsigned int type, unsigned int code, int value)
{
	if (type != EV_SND)
		return -1;

	switch (code) {
		case SND_BELL:
			if (value)
				value = 1000;
		case SND_TONE:
			break;
		default:
			return -1;
	}

	if (value < 20 || value > 32767) value = 0;

	at91_spkr_control(value);

	return 0;
}

static int __devinit at91_spkr_probe(struct platform_device *pdev)
{
	struct input_dev *input_dev;
	int err;
	struct resource *regs;

	input_dev = input_allocate_device();
	if (!input_dev)
		return -ENOMEM;

	input_dev->name = "at91-spkr",
	input_dev->phys = "at91/gpio";
	input_dev->id.bustype = BUS_HOST;
	input_dev->id.vendor  = 0x001f;
	input_dev->id.product = 0x0001;
	input_dev->id.version = 0x0100;
	input_dev->dev.parent = &pdev->dev;

	input_dev->evbit[0] = BIT(EV_SND);
	input_dev->sndbit[0] = BIT(SND_BELL) | BIT(SND_TONE);
	input_dev->event = at91_spkr_event;

	// store pointer to device info for later use.
	info.pdev = pdev;

	// tc2 regs
	regs = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!regs)
		return -ENXIO;

	info.membase = ioremap(regs->start, regs->end - regs->start + 1);
	if (!info.membase)
		return -ENOMEM;

	info.clk = clk_get(&pdev->dev, "tc1_clk");
	clk_enable(info.clk);
	info.clkfreq = clk_get_rate(info.clk) / 128;

	err = input_register_device(input_dev);

	if (err)
		return err;

	info.shdwn_pin = (unsigned int)(pdev->dev.platform_data);

	//turn on amp
	at91_set_gpio_output(info.shdwn_pin, 0);

	info.input_dev = input_dev;

	// we're up&running.
	printk(KERN_INFO "AT91 Speaker Driver\n");


	return 0;
}

static int __devexit at91_spkr_remove(struct platform_device *dev)
{
	input_unregister_device(info.input_dev);

	/* turn the speaker off */
	at91_spkr_control(0);

	// turn off amp
	at91_set_gpio_output(info.shdwn_pin, 1);

	return 0;
}

static void at91_spkr_shutdown(struct platform_device *dev)
{
	/* turn off the speaker */
	at91_spkr_control(0);

	// turn off amp
	at91_set_gpio_output(info.shdwn_pin, 1);
}

#ifdef CONFIG_PM

static int at91_spkr_suspend(struct platform_device *pdev, pm_message_t state) {

	clk_disable(info.clk);
	// turn off amp
	at91_set_gpio_output(info.shdwn_pin, 1);
	return 0;
}

static int at91_spkr_resume(struct platform_device *pdev) {
	clk_enable(info.clk);
	// turn on amp
	at91_set_gpio_output(info.shdwn_pin, 0);
	return 0;
}

#endif

static struct platform_driver at91_spkr_platform_driver = {
	.driver		= {
		.name	= "at91-spkr",
		.owner	= THIS_MODULE,
	},
	.probe		= at91_spkr_probe,
	.remove		= __devexit_p(at91_spkr_remove),
	.shutdown	= at91_spkr_shutdown,
  #ifdef CONFIG_PM
    .suspend   = at91_spkr_suspend,
    .resume    = at91_spkr_resume,
  #endif
};

static int __init at91_spkr_init(void)
{
	return platform_driver_register(&at91_spkr_platform_driver);
}

static void __exit at91_spkr_exit(void)
{
	platform_driver_unregister(&at91_spkr_platform_driver);
}

module_init(at91_spkr_init);
module_exit(at91_spkr_exit);
