/*
 * Copyright (c) 2007 Artec Design LLC
 *   Anti Sullin <anti.sullin@artecdesign.ee>
 *
 * Driver for AT91 WDT. This driver resets the WDT periodically and listens
 * for reset key combination.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. *
*/

//--------------------------------------------------------------------------------

#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/device.h>
#include <linux/wdt.h>
#include <mach/at91_wdt.h>
#include <mach/at91_rstc.h>
#include <asm/gpio.h>

// major is always 10 for misc devices
#define WDT_MINOR 57
// see also include/linux/miscdevice.h for number allocation

// watchdog reset cycle
#define WDT_RST_STEP (msecs_to_jiffies(1000))
// how long the user has to hold down (in WDT_RST_STEP times)
#define RESET_KEY_DELAY 4
//#define DEBUG



// draiveri globaalsed muutujad
static struct wdt {
	struct platform_device *pdev;
	int wdt_rst_enable;
	struct timer_list wdt_rst_timer; // watchdog reset timer
	int rst_cnt;
	struct wdt_rst_key *rst_keys;
	int num_rst_keys;
} info;

//--------------------------------------------------------------------------------
// macros for WDT register manipulation

#define wdt_reset() \
	writel(0xA5000001, AT91_VA_BASE_SYS + AT91_WDT_CR)

#define device_reset() \
	writel(AT91_RSTC_KEY | AT91_RSTC_PROCRST | AT91_RSTC_PERRST | AT91_RSTC_EXTRST, \
			AT91_VA_BASE_SYS + AT91_RSTC_CR)

#define get_key_state(k) \
	(gpio_get_value(k))

//--------------------------------------------------------------------------------
// Kernel timer, that triggers WDT reset
static void wdt_reset_timer(unsigned long dummy)
{
	int i, down=1;

	if (info.wdt_rst_enable) {
		wdt_reset();
		mod_timer(&info.wdt_rst_timer, jiffies + WDT_RST_STEP); // continue?
	}

	// Does the user hold the reset key combination
	if (info.num_rst_keys) {
		for (i=0; i<info.num_rst_keys; i++){
			if (info.rst_keys[i].active_low_input)
				down = down && !get_key_state(info.rst_keys[i].gpio_input);
			else
				down = down && get_key_state(info.rst_keys[i].gpio_input);
		}
		if (down) {
			info.rst_cnt++;
		}
		else
			info.rst_cnt=0;

		if (info.rst_cnt >= RESET_KEY_DELAY) device_reset();
	}
}

//--------------------------------------------------------------------------------
// FILE OPERATIONS

// IOCTL interface
static int wdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		      unsigned long arg)
{
	switch (cmd) {
		case WDT_ENABLE_RST:
			info.wdt_rst_enable = 1;
			mod_timer(&info.wdt_rst_timer, jiffies + WDT_RST_STEP);
			return 0;
		case WDT_DISABLE_RST:
			info.wdt_rst_enable = 0;
			return 0;
		case WDT_RESET:
			wdt_reset();
			return 0;
	}
	return -EINVAL;
}

/* *********************************************** MISC DEV PROBE, REMOVE etc **************************************/

// misc device init params: file operations
static struct file_operations wdt_fops = {
      .owner = THIS_MODULE,
      .ioctl = wdt_ioctl,
};

// misc dev init params
static struct miscdevice wdt_miscdev = {
	.minor = WDT_MINOR,
	.name = "wdt",
	.fops = &wdt_fops
};


static int __devinit wdt_probe(struct platform_device *pdev)
{
	int ret;
	struct wdt_platform_data *pdata = pdev->dev.platform_data;
	// just for caution- initialize struct info with zeroes
	memset(&info, 0, sizeof(info));
	info.wdt_rst_enable = 1;

	// store pointer to device info for later use.
	info.pdev = pdev;

	// reset key combination
	info.rst_keys = pdata->rst_keys;
	info.num_rst_keys = pdata->num_rst_keys;

	// first wdt reset
	wdt_reset();

	// try to register misc dev
	ret = misc_register(&wdt_miscdev);
	if (ret) {
		printk(KERN_ERR "Can't misc_register on minor=%d\n", WDT_MINOR);
		return -EBUSY;
	}

	// initialize wdt reset timer
	init_timer(&info.wdt_rst_timer);
	info.wdt_rst_timer.function = wdt_reset_timer;

	// set up timer for wdt reset
	mod_timer(&info.wdt_rst_timer, jiffies + WDT_RST_STEP);

	// we're up&running.
	printk(KERN_INFO "AT91 real-time timer.\n");

	return 0;
}

static int __devexit wdt_remove(struct platform_device *pdev)
{
	del_timer(&info.wdt_rst_timer);

	misc_deregister(&wdt_miscdev);

	dev_dbg(&pdev->dev, "unregistered wdt driver\n");
	return 0;
}

#ifdef CONFIG_PM
static int wdt_suspend(struct platform_device *pdev, pm_message_t state)
{
	wdt_reset();
	return 0;
}

static int wdt_resume(struct platform_device *pdev)
{
	wdt_reset();
	return 0;
}
#else
#define wdt_suspend NULL
#define wdt_resume NULL
#endif

/* *********************************************** KERNEL DRIVER STUFF **************************************/


static struct platform_driver wdt_driver = {
	.driver = {
		.name	= "wdt", // nimi peab vastama board'is seadistatud device'ga
		.owner	= THIS_MODULE,
	},
	.probe		= wdt_probe,
	.remove		= __devexit_p(wdt_remove),
	.suspend	= wdt_suspend,
	.resume		= wdt_resume,
};


static int __init wdt_init(void)
{
	return platform_driver_register(&wdt_driver);
}
module_init(wdt_init);


static void __exit wdt_exit(void)
{
	platform_driver_unregister(&wdt_driver);
}
module_exit(wdt_exit);

MODULE_DESCRIPTION("AT91 real-time timer and WDT driver");
MODULE_LICENSE("GPL");
